//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// See header file "GCCore.h" for usage.
//

#include "GCCore.h"

//
// Initialization -----------------------------------------------------
//
// Must be called by client application after initializing the toolbox,
// and before any other Gray Council calls are made.
//

OSErr InitGrayCouncil()
	{
	OSErr	result;
	SInt32	gestaltResponse;

	//
	// Initialize some basic globals.
	//
	
	// Default flags. Only live scrolling is on by default.
	gAGADefaults = kAGALiveScrolling;
	
	// Set global Color QuickDraw flag.
	result = ::Gestalt(gestaltQuickdrawVersion, &gestaltResponse);
	gAGAHasColorQD = (result == noErr) && (gestaltResponse != gestaltOriginalQD);
	
	// Check for presence of the Appearance Manager.
	result = ::Gestalt('appr', &gestaltResponse);
	Boolean hasAppearanceMgr = (result == noErr) && ((gestaltResponse & 0x00000001) != 0);
	
	// Check for presence of the Aaron or Kaleidoscope.
	Boolean	hasAaronOrKaleidoscope = (
		(::Gestalt('Aarn', &gestaltResponse) == noErr) ||
		(::Gestalt('Copl', &gestaltResponse) == noErr) ||
		(::Gestalt('Acid', &gestaltResponse) == noErr)
		);

	// Now init our global switches based on these things and the
	// client's wishes.
	
	if ((hasAppearanceMgr && gAppearanceManagerSavvyInit) ||
		(hasAaronOrKaleidoscope && gAaronKaleidoscopeSavvyInit))
		{
		gAGAStdControls = false;
		gAGAExtraControls = false;
		}
	
	// Allocate the group container.
	gGroupsContainer = new AGAGroupsContainer;

	if (gGroupsContainer == NULL)
		return memFullErr;
	else
		return noErr;
	}

//
// AGAObject ---------------------------------------------------------------
//

AGAObject::AGAObject(Rect* bounds)
	{
	// Construct enabled.

	mBounds = *bounds;
	mEnabled = true;
	mActive = true;
	mEnabledBackgroundColor = gAGARamp[r2];
	mDisabledBackgroundColor = gAGARamp[r2];
	}

AGAObject::~AGAObject()
	{
	}

void AGAObject::GetObjectBounds(Rect* bounds)
	{
	// Return current bounds.

	*bounds = mBounds;
	}

void AGAObject::SetObjectBounds(Rect* bounds, Boolean redraw)
	{
	// Change bounds, redraw if specified.

	mBounds = *bounds;
	
	if (redraw)
		this->DrawObject();
	}

void AGAObject::DrawObject()
	{
	// Override to draw particular type of object.
	}

Boolean AGAObject::ContainsMouse(Point mouseLocation)
	{
	// Return true if mouse will hit object.

	return ::PtInRect(mouseLocation, &mBounds);
	}

Boolean AGAObject::TrackMouse(Point /*mouseLocation*/)
	{
	if (! mEnabled)
		return false;

	// Process mouse tracking, return true if tracking
	// eventually "did something". Return false if
	// tracking had no ultimate effect, e.g. mouse
	// up outside a push button.

	Boolean	finishedIn = false;
	Boolean	wasIn = false;
	Boolean	isIn;
	Boolean	isButtonDown = true;	// ensure at least one time through
	Point	newMouseLocation;
	
	while (isButtonDown)
		{
		::GetMouse(&newMouseLocation);	// gives local coordinates
		
		isIn = this->HitTest(newMouseLocation);
		
		if (isIn != wasIn)
			this->SetTrackingState(isIn);
		
		wasIn = isIn;

		isButtonDown = ::StillDown();
		}
	
	if (isIn)	// if finished IN, restore normal appearance
		this->SetTrackingState(kTrackingOut);
	
	return isIn;
	}

void AGAObject::SetEnable(Boolean isEnabled, Boolean redraw)
	{
	// Set object enable state, redraw if specified.

	mEnabled = isEnabled;
	
	if (redraw)
		this->DrawObject();
	}

void AGAObject::SetActive(Boolean isActive, Boolean redraw)
	{
	// Set object enable state, redraw if specified.

	mActive = isActive;
	
	if (redraw)
		this->DrawObject();
	}

void AGAObject::SetBackgroundColors(RGBColor* enabledColor, RGBColor* disabledColor)
	{
	// Set colors to be used when erasing background.
	// Most views don't need this. Default is AGA ramp r2.
	// AGAStaticText and AGADisclosureTriangle need to
	// erase their background, so if your background is
	// not r2 (white, for example), call this to set it.
	// Also, objects contained in tab panel views need
	// r1 when enabled, and r2 when disabled, so that's
	// an even wackier case; the adapter classes handle
	// it for you.

	mEnabledBackgroundColor = *enabledColor;
	mDisabledBackgroundColor = *disabledColor;
	}

SInt32 AGAObject::GetValue()
	{
	return 0;	// overriden for objects that have a useful value
	}

void AGAObject::SetValue(SInt32 /*newValue*/, Boolean /*redraw*/)
	{
	// overriden for objects that have a useful value
	}

Boolean AGAObject::HitTest(Point mouseLocation)
	{
	// Return true if mouse is inside tracking area of
	// object. This function may be overridden for
	// classes that do not have simple rectangular
	// boundary hit testing.

	return ::PtInRect(mouseLocation, &mBounds);
	}

void AGAObject::SetTrackingState(Boolean /*isIn*/)
	{
	// Override to draw the object in the appropriate
	// state during tracking. If isIn is true, it means
	// that the mouse is now in the positive hit test
	// area of the object.
	}

void AGAObject::ApplyBackgroundColor()
	{
	// Use the proper background color and apply it as
	// the current foreground color suitable for a
	// subsequent PaintRect or similar call.

	if (mEnabled)
		::RGBForeColor(&mEnabledBackgroundColor);
	else
		::RGBForeColor(&mDisabledBackgroundColor);
	}

//
// MExclusiveObject ---------------------------------------------------------------
//
// This class is a mixin class for objects that can be included in a
// group of mutually exclusively selected objects, i.e. radio buttons
// or icon buttons. When any group member is selected, the others in
// the group are sent the TurnOff message.
//

MExclusiveObject::MExclusiveObject(UInt32 groupIDPart1, UInt32 groupIDPart2)
	{
	// Construct by setting the ID pair and adding to
	// the group.

	mGroupIDPart1 = groupIDPart1;
	mGroupIDPart2 = groupIDPart2;

	gGroupsContainer->AGAAddGroupMember(this);
	}

MExclusiveObject::~MExclusiveObject()
	{
	// Destruct by removing from the group.

	gGroupsContainer->AGARemoveGroupMember(this);
	}

void MExclusiveObject::SetGroupID(UInt32 groupIDPart1, UInt32 groupIDPart2)
	{
	gGroupsContainer->AGARemoveGroupMember(this);

	mGroupIDPart1 = groupIDPart1;
	mGroupIDPart2 = groupIDPart2;

	gGroupsContainer->AGAAddGroupMember(this);
	}

void MExclusiveObject::TurnOff()
	{
	// Override as needed.
	}

//
// AGATextStyle --------------------------------------------------------
//

void AGATextStyle::NormalizeTextStyle(TextStyle& aTextStyle, SInt16 systemFontNum, SInt16 applFontNum)
	{
	if (aTextStyle.tsFont == systemFontNum)
		aTextStyle.tsFont = ::GetSysFont();//systemFont;
	else if (aTextStyle.tsFont == applFontNum)
		aTextStyle.tsFont = ::GetAppFont();//applFont;
	}

AGATextStyle::AGATextStyle()
	{
	// Default constructor, use system font.

	mFontNum = ::GetSysFont();	// typically Chicago
	mFontSize = 0;			// application size, typically 12-point
	mFontStyle = normal;
	}

AGATextStyle::AGATextStyle(SInt16 fontNum, SInt16 fontSize, Style fontStyle)
	{
	// Construct with direct parameters.

	mFontNum = fontNum;
	mFontSize = fontSize;
	mFontStyle = fontStyle;
	}

AGATextStyle::AGATextStyle(StringPtr fontName, SInt16 fontSize, Style fontStyle)
	{
	// Construct by converting font name to font number.

	::GetFNum(fontName, &mFontNum);
	mFontSize = fontSize;
	mFontStyle = fontStyle;
	}

AGATextStyle::AGATextStyle(const TextStyle& textStyle)
	{
	// Construct by copying another text style.

	mFontNum = textStyle.tsFont;
	mFontSize = textStyle.tsSize;
	mFontStyle = textStyle.tsFace;
	}

void AGATextStyle::PrepareForDrawing() const
	{
	// Set the text pen for subsequent drawing.
	
	::TextFont(mFontNum);
	::TextSize(mFontSize);
	::TextFace(mFontStyle);
	}

//
// AGADrawingEnvironment ---------------------------------------------------------------
//

AGADrawingEnvironment::AGADrawingEnvironment()
	{
	// Save current environment, then cleanse the pen.

	this->SaveEnvironment();
	this->Cleanse();
	}

AGADrawingEnvironment::~AGADrawingEnvironment()
	{
	this->RestoreEnvironment();
	}

void AGADrawingEnvironment::SaveEnvironment()
	{
	// Save all text/pen attributes that we will change.

	GrafPtr	currentPort;

	::GetPort(&currentPort);

	::GetPenState(&mSavedPenState);
	::GetForeColor(&mSavedForeColor);
	::GetBackColor(&mSavedBackColor);
	mSavedFontNum = currentPort->txFont;
	mSavedFontSize = currentPort->txSize;
	mSavedFontStyle = currentPort->txFace;
	}

void AGADrawingEnvironment::RestoreEnvironment()
	{
	// Restore text/pen attributes that we saved.

	::SetPenState(&mSavedPenState);
	::RGBForeColor(&mSavedForeColor);
	::RGBBackColor(&mSavedBackColor);
	::TextFont(mSavedFontNum);
	::TextSize(mSavedFontSize);
	::TextFace(mSavedFontStyle);
	}

void AGADrawingEnvironment::Cleanse()
	{
	// Put the pen into a nice clean state.

	::PenNormal();							// pen size 1,1, solid black pattern, srcOr
	::RGBForeColor(&gAGARamp[rB]);			// black fore color
	::RGBBackColor(&gAGARamp[rW]);			// white back color
	gAGAStdSystemStyle.PrepareForDrawing();	// system font 12 point
	}

//
// AGATrackingIndicator ---------------------------------------------------------------
//
// This is the abstract superclass for scroll bars and sliders.
// They all have a range, and a value shown by a draggable indicator.
//

AGATrackingIndicator::AGATrackingIndicator(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue)
	: AGAObject(bounds)
	{
	// Construct with default options.

	mMinimum = minimum;
	mMaximum = maximum;
	mValue = initialValue;
	
	mPageSize = 1;
	mGhostValue = mValue;
	mLiveTracking = false;
	mIsProportional = false;
	mIsHorizontal = (bounds->right - bounds->left) > (bounds->bottom - bounds->top);
	mIsPressed = false;
	
	mNotificationRoutine = NULL;
	}

AGATrackingIndicator::AGATrackingIndicator(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, SInt32 pageSize, Boolean liveTracking, Boolean proportional)
	: AGAObject(bounds)
	{
	// Construct with specified options.

	mMinimum = minimum;
	mMaximum = maximum;
	mValue = initialValue;
	mPageSize = pageSize;
	mGhostValue = mValue;
	mLiveTracking = liveTracking;
	mIsProportional = proportional;
	mIsHorizontal = (bounds->right - bounds->left) > (bounds->bottom - bounds->top);
	mIsPressed = false;
	
	mNotificationRoutine = NULL;
	}

AGATrackingIndicator::~AGATrackingIndicator()
	{
	}

void AGATrackingIndicator::InstallNotificationRoutine(AGATrackingIndicatorNotifyPtr notificationRoutine, void* userData)
	{
	// Install the supplied function pointer and user data to
	// be called during notification.

	mNotificationRoutine = notificationRoutine;
	mUserData = userData;
	}

void AGATrackingIndicator::SetAttributes(Boolean isHorizontal, Boolean liveTracking, Boolean isProportional)
	{
	// Set options as specified.

	mIsHorizontal = isHorizontal;
	mLiveTracking = liveTracking;
	mIsProportional = isProportional;
	}

void AGATrackingIndicator::SetLiveTracking(Boolean liveTracking)
	{
	// Turn live tracking on or off.

	mLiveTracking = liveTracking;
	}

void AGATrackingIndicator::DrawObject()
	{
	// Call component drawing functions. These are all overridden
	// by subclasses.

	this->DrawBackground();
	this->DrawIndicatorTrack();
	this->DrawTrackEnds();
	this->DrawIndicator();
	}

Boolean AGATrackingIndicator::TrackMouse(Point mouseLocation)
	{
	if (! mEnabled)
		return false;

	// Track the mouse and return true if the indicator
	// value was changed.

	return this->TrackPart(mouseLocation, this->TrackTestPart(mouseLocation));
	}

Boolean AGATrackingIndicator::TrackPart(Point mouseLocation, SInt32 partHit)
	{
	// Track the mouse in the specified part and return true
	// if the indicator value was changed.

	if (partHit != kIndicatorPress)
		return false;

	SInt32	originalValue = mValue;
	Point	oldPoint = mouseLocation;
	Point	newPoint = mouseLocation;
	SInt32	oldValue = originalValue;
	SInt32	newValue = originalValue;
	Boolean	isButtonDown = true;	// ensure at least one time through
	
	mGhostValue = mValue;
	mIsPressed = true;
	this->DrawIndicator();	// will draw it pressed
	
	while (isButtonDown)
		{
		::GetMouse(&newPoint);	// gives local coordinates
		
		if (this->IsValidIndicatorTrackMouse(newPoint) &&
			! ::EqualPt(newPoint, oldPoint))
			{
			newValue = this->GetValueFromMouseDelta(oldValue, mouseLocation, newPoint);
		
			if (mLiveTracking)
				{
				SInt32	oldValue = mValue;

				this->SetRawValue(newValue, kRedraw);
				
				if (mValue != oldValue)
					this->NotifyValue();
				}
			else
				this->SetGhostValue(newValue, kRedraw);
			
			oldPoint = newPoint;
			}
		
		isButtonDown = ::StillDown();
		}
	
	mIsPressed = false;
	
	if (mLiveTracking)
		this->DrawIndicator();	// unpressed
	else
		{
		this->RemoveGhost(newValue);
		this->DrawIndicator();	// unpressed

		if (newValue != mValue)	// don't notify if no change
			{
			this->SetRawValue(newValue, kRedraw);	// will draw new indicator
			this->NotifyValue();
			}
		}
	
	return (mValue != originalValue);
	}

SInt32 AGATrackingIndicator::GetValue()
	{
	// Return the current indicator value.

	return mValue;
	}

void AGATrackingIndicator::SetGhostValue(SInt32 newValue, Boolean redraw)
	{
	// Set the ghost indicator value, and optionally redraw it
	// at its new position.

	if (newValue > mMaximum)
		newValue = mMaximum;
	else if (newValue < mMinimum)
		newValue = mMinimum;

	if (newValue != mGhostValue)
		{
		if (redraw)
			this->RemoveGhost(newValue);

		mGhostValue = newValue;
		
		if (redraw)
			this->DrawGhost();
		}
	}

void AGATrackingIndicator::SetValue(SInt32 newValue, Boolean redraw)
	{
	this->SetRawValue(newValue, redraw);
	}

void AGATrackingIndicator::SetRawValue(SInt32 newValue, Boolean redraw)
	{
	// Set the indicator value, and optionally redraw it
	// at its new position.

	if (newValue > mMaximum)
		newValue = mMaximum;
	else if (newValue < mMinimum)
		newValue = mMinimum;

	if (newValue != mValue)
		{
		if (redraw)
			this->RemoveIndicator(newValue);

		mValue = newValue;
		
		if (redraw)
			this->DrawIndicator();
		}
	}

void AGATrackingIndicator::GetRange(SInt32* minimum, SInt32* maximum)
	{
	// Return the current min/max range.

	*minimum = mMinimum;
	*maximum = mMaximum;
	}

void AGATrackingIndicator::SetRange(SInt32 newMinimum, SInt32 newMaximum, Boolean redraw)
	{
	// Set the min/max range, and optionally redraw.

	if ((newMinimum != mMinimum) ||
		(newMaximum != mMaximum))
		{
		if (redraw)
			this->RemoveIndicator(mValue);

		mMinimum = newMinimum;
		mMaximum = newMaximum;
		
		mValue = MaxSInt32(mMinimum, MinSInt32(mValue, mMaximum));

		if (redraw)
			this->DrawIndicator();
		}
	}

SInt32 AGATrackingIndicator::GetPageSize()
	{
	// Return the current indicator page size.

	return mPageSize;
	}

void AGATrackingIndicator::SetPageSize(SInt32 newPageSize, Boolean redraw)
	{
	// Set the indicator page size. Redraw is not
	// necessary if the indicator is not proportional.

	if (newPageSize != mPageSize)
		{
		if (redraw && mIsProportional)
			this->RemoveIndicator(mValue);

		mPageSize = newPageSize;
		
		if (redraw && mIsProportional)
			this->DrawIndicator();
		}
	}

SInt32 AGATrackingIndicator::TrackTestPart(Point /*mouseLocation*/)
	{
	// Must be overridden. Return the part value that the
	// specified mouse position would hit.

	return kNoTrackPress;
	}

Boolean AGATrackingIndicator::IsValidIndicatorTrackMouse(Point mouseLocation)
	{
	// Usually overridden. Return true if the specified mouse
	// position should be considered "in" for tracking purposes
	// after the initial hit has succeeded. Return false if the
	// specified mouse position is too far away from the control
	// to use it for the next momentary tracking position.

	return ::PtInRect(mouseLocation, &mBounds);
	}

SInt32 AGATrackingIndicator::GetValueFromMouseDelta(SInt32 originalValue, Point /*originalMouseLocation*/, Point /*newMouseLocation*/)
	{
	// Must be overridden. Return the indicator value that should be
	// applied assuming the supplied starting value and mouse hit point,
	// and the supplied current mouse tracking location.
	return originalValue;
	}

void AGATrackingIndicator::DrawBackground()
	{
	// Override to paint the background area so that the
	// indicator drawn at its old position will be wiped out.
	// Clipping will be already set up to avoid excess flicker.
	}

void AGATrackingIndicator::DrawIndicatorTrack()
	{
	// Override to draw the empty indicator track. Must include
	// all possible indicator coverage.
	}

void AGATrackingIndicator::DrawTrackEnds()
	{
	// Override to draw the track ends outside the possible
	// indicator coverage if not drawn in DrawIndicatorTrack.
	}

void AGATrackingIndicator::DrawIndicator()
	{
	// Override to draw the indicator at its current value/state.
	}

void AGATrackingIndicator::RemoveIndicator(SInt32 /*newValue*/)
	{
	// Override to draw/erase whatever is needed to remove the
	// current indicator in preparation for being drawn at
	// the specified new value. The new value is supplied so
	// that erasing can be kept to a minimal area.
	}

void AGATrackingIndicator::DrawGhost()
	{
	// Override to draw the indicator in ghost mode at
	// the current ghost value.
	}

void AGATrackingIndicator::RemoveGhost(SInt32 /*newValue*/)
	{
	// Override to draw/erase whatever is needed to remove the
	// ghost indicator in preparation for being drawn at
	// the specified new value. The new value is supplied so
	// that erasing can be kept to a minimal area.
	}

void AGATrackingIndicator::NotifyValue()
	{
	// Call the installed notification routine, if any, with
	// the installed user data and our new value.

	if (mNotificationRoutine != NULL)
		(*(mNotificationRoutine))(this, mValue, mUserData);
	}

//
// AGAOffscreenImage ------------------------------------------------------------
//
// AGAOffscreenImage holds the data needed to blast a clipped region
// of pixel values onto the current port. This is used by the AGARadioButton
// class to avoid the ultra-tedious pixel-by-pixel drawing of the grayscale
// radio button images.
//

AGAOffscreenImage::AGAOffscreenImage()
	{
	mImagePixMap = NULL;
	mClippingRegion = NULL;
	}

AGAOffscreenImage::~AGAOffscreenImage()
	{
	if (mImagePixMap != NULL)
		::DisposePixMap(mImagePixMap);

	if (mClippingRegion != NULL)
		::DisposeRgn(mClippingRegion);
	}

OSErr AGAOffscreenImage::CreateImageData(Point imageSize,
										SInt8* imageColorIndexes,
										RgnHandle clippingRegion,
										const RGBColor* outsideColor)
	{
	// Allocate an offscreen PixMap to contain the RGB values
	// specified.

	const UInt32 kImageDepth = 4;
	
	::SetRect(&mImageSourceRect, 0, 0, imageSize.h, imageSize.v);

	if (clippingRegion != NULL)
		{
		// Copy the supplied clipping region.

		mClippingRegion = ::NewRgn();
		if (mClippingRegion == NULL)
			return memFullErr;

		::CopyRgn(clippingRegion, mClippingRegion);
		OSErr	result = ::MemError();
		if (result != noErr)
			return result;
		}

	//
	// Allocate and set up the PixMap structure.
	//

	Ptr	imagePtr = ::NewPtr(kImageDepth * imageSize.h * imageSize.v);

	if (imagePtr == NULL)
		return memFullErr;

	mImagePixMap = ::NewPixMap();
	::HLockHi((Handle) mImagePixMap);
	
	PixMapPtr	pixMapPtr = *mImagePixMap;
	
	pixMapPtr->baseAddr = imagePtr;
	pixMapPtr->rowBytes = (kImageDepth * imageSize.h) | 0x8000;
	pixMapPtr->bounds = mImageSourceRect;
	pixMapPtr->pixelType = RGBDirect;
	pixMapPtr->pixelSize = 32;	// direct 32-bit color values
	pixMapPtr->cmpCount = 3;	// 3 color components (RGB)
	pixMapPtr->cmpSize = 8;		// 8 bits per component
	pixMapPtr->planeBytes = 0;

	if (pixMapPtr->pmTable != NULL)
		{
		// We're using direct color, so toss the table.
		::DisposeCTable(pixMapPtr->pmTable);
		pixMapPtr->pmTable = NULL;
		}
	
	pixMapPtr->pmReserved = 0;
	
	//
	// Finally, set the pixel RGB color values.
	//
	
	UInt32*	pixelPtr = (UInt32*) imagePtr;
	UInt32	numPixels = imageSize.h * imageSize.v;
	SInt8*	indexPtr = imageColorIndexes;
	
	for (UInt32 pixelIndex = 0; pixelIndex < numPixels; pixelIndex++)
		{
		SInt8		rampIndex = *indexPtr++;
		RGBColor	rampColor;
		
		if (rampIndex == OUT)
			rampColor = *outsideColor;
		else
			rampColor = gAGARamp[rampIndex];
		
		//
		// Construct 32-bit color value.
		// Assuming red is RRrr, blue is BBbb, green is GGgg,
		// we want 0x00RRGGBB, using high byte of each color.
		// This way we toss the low-order (LSB) color data.
		//

		UInt32	pixelValue =
				((((UInt32) rampColor.red) & 0x0000FF00) << 8) |
				(((UInt32) rampColor.green) & 0x0000FF00) |
				((((UInt32) rampColor.blue) & 0x0000FF00) >> 8);

		*pixelPtr++ = pixelValue;
		}

	return noErr;
	}

void AGAOffscreenImage::DrawImage(SInt16 imageHOrigin, SInt16 imageVOrigin)
	{
	// Blast the offscreen image data into the current
	// port at the specified coordinate location.

	RGBColor	savedBackColor;
	GrafPtr		currentPort;
	Rect		outputRect = mImageSourceRect;
	
	::GetBackColor(&savedBackColor);

	// Set to black-on-white before doing CopyBits or our colors may
	// be massaged on the way to the screen.
	::RGBBackColor(&gAGARamp[rW]);
	::RGBForeColor(&gAGARamp[rB]);
	
	::OffsetRect(&outputRect, imageHOrigin, imageVOrigin);
	
	if (mClippingRegion != NULL)
		::OffsetRgn(mClippingRegion, imageHOrigin, imageVOrigin);

	::GetPort(&currentPort);
	::CopyBits((BitMapPtr) *mImagePixMap,
		&(currentPort->portBits),
		&mImageSourceRect,
		&outputRect,
		srcCopy,
		mClippingRegion);
	
	if (mClippingRegion != NULL)
		::OffsetRgn(mClippingRegion, - imageHOrigin, - imageVOrigin);

	::RGBBackColor(&savedBackColor);
	}

void AGAOffscreenImage::DrawPattern(Rect* imageRect)
	{
	// Blast the offscreen image data into the current
	// port, starting at the specified coordinate location,
	// repeating tiled until the imageRect is filled.

	RGBColor	savedBackColor;
	GrafPtr		currentPort;
	Rect		outputRect;
	
	::GetBackColor(&savedBackColor);
	::RGBBackColor(&gAGARamp[rW]);	// need white BG for desired CopyBits behavior
	
	if (mClippingRegion != NULL)
		::OffsetRgn(mClippingRegion, imageRect->left, imageRect->top);

	::GetPort(&currentPort);
	
	outputRect.top = imageRect->top;
	outputRect.bottom = outputRect.top + mImageSourceRect.bottom;

	while (outputRect.top < imageRect->bottom)
		{
		outputRect.left = imageRect->left;
		outputRect.right = outputRect.left + mImageSourceRect.right;
		
		while (outputRect.left < imageRect->right)
			{
			::CopyBits((BitMapPtr) *mImagePixMap,
				&(currentPort->portBits),
				&mImageSourceRect,
				&outputRect,
				srcCopy,
				mClippingRegion);
			
			::OffsetRect(&outputRect, mImageSourceRect.right, 0);
			}

		::OffsetRect(&outputRect, 0, mImageSourceRect.bottom);
		}

	if (mClippingRegion != NULL)
		::OffsetRgn(mClippingRegion, - imageRect->left, - imageRect->top);

	::RGBBackColor(&savedBackColor);
	}

//
// GDIterator -----------------------------------------------------
//

RgnHandle GDIterator::mSavedClipping = ::NewRgn();

GDIterator::GDIterator()
	{
	this->Setup();
	}

GDIterator::~GDIterator()
	{
	this->Cleanup();
	}

void GDIterator::Setup()
	{
	// Save the current clipping so we can put it back in Cleanup
	// and use it as we iterate over each device.

	::GetClip(mSavedClipping);
	
	mIteratedYet = false;
	mCurrentGDHandle = NULL;
	}

Boolean GDIterator::NextDevice(Boolean& deep)
	{
	// If there's another device to draw on, get its GDHandle and deepness,
	// and return true. Otherwise, return false.
	
	Boolean	gotNextDevice;

	if (gAGAHasColorQD)
		{
		mCurrentGDHandle = this->FindNextValidDevice(mCurrentGDHandle, deep);
		gotNextDevice = (mCurrentGDHandle != NULL);
		}
	else
		{
		deep = false;
		gotNextDevice = ! mIteratedYet;
		
		mIteratedYet = true;
		}
	
	return gotNextDevice;
	}

void GDIterator::Cleanup()
	{
	// Restore clipping to what it was when we were constructed.

	::SetClip(mSavedClipping);
	}

void GDIterator::ClipFurtherToCurrentDevice()
	{
	// Clip further to prevent drawing outside the current device.

	this->ClipFurtherToDevice(mCurrentGDHandle);
	}

GDHandle GDIterator::FindNextValidDevice(GDHandle currentDevice, Boolean& deep)
	{
	// Return the GDHandle of the next valid device, if any, and indicate
	// whether it is "deep".
	// As a designed side-effect, when we find a next device, we restore
	// the original clipping and then re-clip further to this next device
	// so no drawing occurs on other devices.

	GDHandle	deviceToTest;
	GDHandle	deviceFound = NULL;
	
	deep = false;
	
	if (currentDevice == NULL)
		deviceToTest = ::GetDeviceList();
	else
		deviceToTest = ::GetNextDevice(currentDevice);

	while ((deviceToTest != NULL) && (deviceFound == NULL))
		{
		Boolean	isValidDevice = false;
		
		if (::TestDeviceAttribute(deviceToTest, screenDevice))
			if (::TestDeviceAttribute(deviceToTest, screenActive))
				{
				isValidDevice = true;
				
				SInt16	pixelDepth = (**((**deviceToTest).gdPMap)).pixelSize;
				Boolean	isColor = ::TestDeviceAttribute(deviceToTest, gdDevType);
				
				deep = (pixelDepth >= kDeepColorBits) || ((pixelDepth >= kDeepGrayscaleBits) && ! isColor);
				
				deviceFound = deviceToTest;
				}

		if (! isValidDevice)
			deviceToTest = ::GetNextDevice(deviceToTest);
		}
	
	if (deviceFound != NULL)
		{
		::SetClip(mSavedClipping);
		this->ClipFurtherToDevice(deviceFound);
		}

	return deviceFound;
	}

void GDIterator::ClipFurtherToDevice(GDHandle aDevice)
	{
	// Clip further to the specified device so no drawing will occur
	// on other devices.

	if (aDevice != NULL)
		{
		Rect	deviceRect = (**aDevice).gdRect;
		Point	drTopLeft;
		Point	drBotRight;
		
		::SetPt(&drTopLeft, deviceRect.left, deviceRect.top);
		::SetPt(&drBotRight, deviceRect.right, deviceRect.bottom);
		
		::GlobalToLocal(&drTopLeft);
		::GlobalToLocal(&drBotRight);
		
		::SetRect(&deviceRect, drTopLeft.h, drTopLeft.v, drBotRight.h, drBotRight.v);
		
		AGAClipFurther(&deviceRect);
		}
	}

//
// AGAGroupsContainer --------------------------------------------------------
//
// This class is used internally to maintain the MExclusiveObject
// behavior that allows radio buttons to automatically maintain
// standard radio button group XOR behavior.
//

AGAGroupsContainer::AGAGroupsContainer()
	{
	mGroupsArraySize = 0;
	mGroups = (AGAGroupMember**) ::NewHandle(0);
	
	this->ExpandArray();
	}

AGAGroupsContainer::~AGAGroupsContainer()
	{
	if (mGroups != NULL)
		{
		::DisposeHandle((Handle) mGroups);
		mGroups = NULL;
		}
	}

void AGAGroupsContainer::AGAAddGroupMember(MExclusiveObject* groupMember)
	{
	this->AddGroupMember(groupMember, true);
	}

void AGAGroupsContainer::AGARemoveGroupMember(MExclusiveObject* groupMember)
	{
	// If the supplied object has an ID pair, search for it in the
	// members array and zero it out.

	if ((mGroups != NULL) && (groupMember->mGroupIDPart1 != kNoGroupID) && (groupMember->mGroupIDPart2 != kNoGroupID))
		{
		AGAGroupMember*	memberPtr;

		::HLock((Handle) mGroups);

		memberPtr = *mGroups;

		for (UInt32 i = 0; i < mGroupsArraySize; i++)
			{
			if ((memberPtr->mGroupIDPart1 == groupMember->mGroupIDPart1) &&
				(memberPtr->mGroupIDPart2 == groupMember->mGroupIDPart2) &&
				(memberPtr->mGroupMember == groupMember))
				{
				memberPtr->mGroupIDPart1 = kNoGroupID;
				memberPtr->mGroupIDPart2 = kNoGroupID;
				memberPtr->mGroupMember = NULL;
				}
				
			memberPtr++;
			}

		::HUnlock((Handle) mGroups);
		}
	}

void AGAGroupsContainer::AGAHitGroupMember(MExclusiveObject* groupMember)
	{
	// If the supplied object has an ID pair, send every other
	// member of its group the TurnOff message.

	if ((mGroups != NULL) && (groupMember->mGroupIDPart1 != kNoGroupID) && (groupMember->mGroupIDPart2 != kNoGroupID))
		{
		AGAGroupMember*	memberPtr;

		::HLock((Handle) mGroups);

		memberPtr = *mGroups;

		for (UInt32 i = 0; i < mGroupsArraySize; i++)
			{
			if ((memberPtr->mGroupIDPart1 == groupMember->mGroupIDPart1) &&
				(memberPtr->mGroupIDPart2 == groupMember->mGroupIDPart2) &&
				(memberPtr->mGroupMember != groupMember))
				memberPtr->mGroupMember->TurnOff();
				
			memberPtr++;
			}

		::HUnlock((Handle) mGroups);
		}
	}

void AGAGroupsContainer::AddGroupMember(MExclusiveObject* groupMember, Boolean allowExpand)
	{
	// If the supplied object has an ID pair, add it to the
	// members array. Try to re-use a zeroed-out entry; if none
	// is left, expand the handle by another chunk.

	if ((mGroups != NULL) && (groupMember->mGroupIDPart1 != kNoGroupID) && (groupMember->mGroupIDPart2 != kNoGroupID))
		{
		AGAGroupMember*	memberPtr;
		Boolean			done = false;

		::HLock((Handle) mGroups);

		memberPtr = *mGroups;

		for (UInt32 i = 0; (i < mGroupsArraySize) && ! done; i++)
			{
			if ((memberPtr->mGroupIDPart1 == kNoGroupID) &&
				(memberPtr->mGroupIDPart2 == kNoGroupID))
				{
				memberPtr->mGroupIDPart1 = groupMember->mGroupIDPart1;
				memberPtr->mGroupIDPart2 = groupMember->mGroupIDPart2;
				memberPtr->mGroupMember = groupMember;
				done = true;
				}
				
			memberPtr++;
			}

		::HUnlock((Handle) mGroups);
		
		if (allowExpand && ! done)
			{
			this->ExpandArray();
			this->AddGroupMember(groupMember, false);	// false = recur once only!
			}
		}
	}

void AGAGroupsContainer::ExpandArray()
	{
	if (mGroups != NULL)
		{
		::SetHandleSize((Handle) mGroups, (mGroupsArraySize + kGroupArrayChunkSize) * sizeof(AGAGroupMember));
		
		if (::MemError() == noErr)
			{
			mGroupsArraySize += kGroupArrayChunkSize;
			
			for (UInt32 i = mGroupsArraySize - kGroupArrayChunkSize; i < mGroupsArraySize; i++)
				{
				(*mGroups)[i].mGroupIDPart1 = kNoGroupID;
				(*mGroups)[i].mGroupIDPart2 = kNoGroupID;
				(*mGroups)[i].mGroupMember = NULL;
				}
			}
		}
	}

//
// AGABackgroundPaint ------------------------------------------------
//

void AGABackgroundPaint(const Rect* backgroundBounds,
						Boolean drawFill,
						AGABackgroundKind backgroundKind,
						Boolean isActive,
						Boolean hasGrowBox)
	{
	if (backgroundKind == kNoBackgroundPaint)
		return;

	AGADrawingEnvironment	env;

	// Draw in the specified active/inactive state.

	enum { kBGFill, kBGActiveLight, kBGActiveShadow, kBGActiveCorner, kBGInactiveLight, kBGInactiveShadow, kBGInactiveCorner, kBGNumIndexes };

	UInt8	colorIndexes[kBGNumIndexes];

	switch (backgroundKind)
		{
		case kRaisedModelessBackground:
			colorIndexes[kBGFill] = r2;
			colorIndexes[kBGActiveLight] = rW;
			colorIndexes[kBGActiveShadow] = r6;
			colorIndexes[kBGActiveCorner] = r2;
			colorIndexes[kBGInactiveLight] = r2;
			colorIndexes[kBGInactiveShadow] = r2;
			colorIndexes[kBGInactiveCorner] = r2;
			break;

		case kRaisedModalBackground:
			colorIndexes[kBGFill] = r2;
			colorIndexes[kBGActiveLight] = rW;
			colorIndexes[kBGActiveShadow] = r6;
			colorIndexes[kBGActiveCorner] = r4;
			colorIndexes[kBGInactiveLight] = r2;
			colorIndexes[kBGInactiveShadow] = r2;
			colorIndexes[kBGInactiveCorner] = r2;
			break;

		case kFlatWindowBackground:
			colorIndexes[kBGFill] = r2;
			colorIndexes[kBGActiveLight] = r2;
			colorIndexes[kBGActiveShadow] = r2;
			colorIndexes[kBGActiveCorner] = r2;
			colorIndexes[kBGInactiveLight] = r2;
			colorIndexes[kBGInactiveShadow] = r2;
			colorIndexes[kBGInactiveCorner] = r2;
			break;
		}

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		Rect	r = *backgroundBounds;

		if (deep)
			{
			::InsetRect(&r, 1, 1);
			
			if (drawFill)
				{
				::RGBForeColor(&gAGARamp[colorIndexes[kBGFill]]);
			
				::PaintRect(&r);
				}
			
			if (isActive)
				::RGBForeColor(&gAGARamp[colorIndexes[kBGActiveCorner]]);
			else
				::RGBForeColor(&gAGARamp[colorIndexes[kBGInactiveCorner]]);

			::MoveTo(r.left - 1, r.bottom);
			::Line(0, 0);
			::MoveTo(r.right, r.top - 1);
			::Line(0, 0);
			
			if (isActive)
				::RGBForeColor(&gAGARamp[colorIndexes[kBGActiveLight]]);
			else
				::RGBForeColor(&gAGARamp[colorIndexes[kBGInactiveLight]]);

			::MoveTo(r.left - 1, r.bottom - 1);
			::LineTo(r.left - 1, r.top - 1);
			::LineTo(r.right - 1, r.top - 1);
			
			if (isActive)
				::RGBForeColor(&gAGARamp[colorIndexes[kBGActiveShadow]]);
			else
				::RGBForeColor(&gAGARamp[colorIndexes[kBGInactiveShadow]]);

			::MoveTo(r.left, r.bottom);
			
			if (hasGrowBox)
				{
				::LineTo(r.right - 15, r.bottom);
				::Line(0, -15);
				::Line(15, 0);
				}
			else
				::LineTo(r.right, r.bottom);

			::LineTo(r.right, r.top);
			}
		else	// 1-bit
			{
			if (drawFill)
				{
				::RGBBackColor(&gAGARamp[rW]);
				::EraseRect(&r);
				}
			}
		}
	}

//
// AGADrawSeparator ------------------------------------------------
//

void AGADrawSeparator(const Rect* bounds, AGASeparatorLayout layout)
	{
	Point	startPt;
	Point	endPt;
	Boolean	isHorizontal;
	
	if (layout == kSeparatorAutomatic)
		{
		if ((bounds->right - bounds->left) > (bounds->bottom - bounds->top))
			layout = kSeparatorTop;
		else
			layout = kSeparatorLeft;
		}
	
	switch (layout)
		{
		case kSeparatorLeft:
		case kSeparatorTop:

			startPt.h = bounds->left;
			startPt.v = bounds->top;
			
			break;

		case kSeparatorRight:

			startPt.h = bounds->right;
			startPt.v = bounds->top;
			
			break;

		case kSeparatorBottom:

			startPt.h = bounds->left;
			startPt.v = bounds->bottom;
			
			break;
		}

	switch (layout)
		{
		case kSeparatorLeft:

			endPt.h = bounds->left;
			endPt.v = bounds->bottom;

			break;

		case kSeparatorTop:

			endPt.h = bounds->right;
			endPt.v = bounds->top;

			break;

		case kSeparatorRight:
		case kSeparatorBottom:

			endPt.h = bounds->right;
			endPt.v = bounds->bottom;
			
			break;
		}
	
	isHorizontal = (layout == kSeparatorTop) || (layout == kSeparatorBottom);

	AGADrawingEnvironment	env;
	GDIterator				iter;
	Boolean					deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			::RGBForeColor(&gAGARamp[r7]);
		else
			{
			::RGBForeColor(&gAGARamp[rB]);
			PenPatGray();
			}

		::MoveTo(startPt.h, startPt.v);
		
		if (isHorizontal)
			::LineTo(endPt.h - 2, endPt.v);
		else
			::LineTo(endPt.h, endPt.v - 2);

		if (deep)
			{
			::RGBForeColor(&gAGARamp[rW]);
			::MoveTo(startPt.h + 1, startPt.v + 1);
			
			if (isHorizontal)
				::LineTo(endPt.h - 1, endPt.v + 1);
			else
				::LineTo(endPt.h + 1, endPt.v - 1);
			}
		else
			PenPatBlack();
		}
	}

//
// AGADrawBevelButton ------------------------------------------------
//

static void DrawMiniBevelEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	// Draw the button edges as specified for a
	// "mini" sized button.

	enum { TLFrame, TLInner, BRFrame, BRInner, CornerFrame, CornerInner, Fill, kNumEdgeColors };

	UInt8	colorIndexes[kNumEdgeColors];

	Rect	r = *bounds;
	
	if (isPressed)
		{
		colorIndexes[TLFrame] = rA2;
		colorIndexes[BRFrame] = r11;
		colorIndexes[CornerFrame] = r12;
		colorIndexes[TLInner] = r10;
		colorIndexes[BRInner] = r5;
		colorIndexes[CornerInner] = r8;
		colorIndexes[Fill] = r7;
		}
	else if (isEnabled)
		{
		colorIndexes[TLFrame] = isOn ? rA2 : r9;
		colorIndexes[BRFrame] = isOn ? r11 : rA1;
		colorIndexes[CornerFrame] = isOn ? r12 : r10;
		colorIndexes[TLInner] = isOn ? r10 : rW;
		colorIndexes[BRInner] = isOn ? r5 : r7;
		colorIndexes[CornerInner] = isOn ? r8 : r3;
		colorIndexes[Fill] = isOn ? r7 : r3;
		}
	else	// disabled
		{
		colorIndexes[TLFrame] = isOn ? r8 : r5;
		colorIndexes[BRFrame] = r6;
		colorIndexes[CornerFrame] = isOn ? r7 : r5;
		colorIndexes[TLInner] = isOn ? r4 : r2;
		colorIndexes[BRInner] = isOn ? r4 : r2;
		colorIndexes[CornerInner] = isOn ? r4 : r2;
		colorIndexes[Fill] = isOn ? r4 : r2;
		}
	
	// Draw the outer frame.

	::RGBForeColor(&gAGARamp[colorIndexes[TLFrame]]);
	::MoveTo(r.left, r.bottom - 2);
	::LineTo(r.left, r.top);
	::LineTo(r.right - 1, r.top);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRFrame]]);
	::Move(0, 1);
	::LineTo(r.right - 1, r.bottom - 1);
	::LineTo(r.left, r.bottom - 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	// Draw the 1st pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner]]);
	::MoveTo(r.left + 1, r.bottom - 3);
	::LineTo(r.left + 1, r.top + 1);
	::LineTo(r.right - 2, r.top + 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner]]);
	::Move(0, 1);
	::LineTo(r.right - 2, r.bottom - 2);
	::LineTo(r.left + 1, r.bottom - 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner]]);
	::Line(0, 0);

	// Fill the interior.

	::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
	::InsetRect(&r, 2, 2);
	::PaintRect(&r);
	::InsetRect(&r, -2, -2);
	}

static void DrawSmallBevelEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	// Draw the button edges as specified for a
	// "small" sized button.

	enum { TLFrame, TLInner1, TLInner2, BRFrame, BRInner1, BRInner2, CornerFrame, CornerInner1, CornerInner2, Fill, kNumEdgeColors };

	UInt8	colorIndexes[kNumEdgeColors];

	Rect	r = *bounds;
	
	if (isPressed)
		{
		colorIndexes[TLFrame] = rA2;
		colorIndexes[BRFrame] = r11;
		colorIndexes[CornerFrame] = rA2;
		colorIndexes[TLInner1] = r11;
		colorIndexes[BRInner1] = r4;
		colorIndexes[CornerInner1] = r8;
		colorIndexes[TLInner2] = r9;
		colorIndexes[BRInner2] = r6;
		colorIndexes[CornerInner2] = r7;
		colorIndexes[Fill] = r7;
		}
	else if (isEnabled)
		{
		colorIndexes[TLFrame] = isOn ? rA2 : r9;
		colorIndexes[BRFrame] = isOn ? r11 : rA1;
		colorIndexes[CornerFrame] = isOn ? rA2 : r9;
		colorIndexes[TLInner1] = isOn ? r11 : r3;
		colorIndexes[BRInner1] = isOn ? r4 : r8;
		colorIndexes[CornerInner1] = isOn ? r8 : r5;
		colorIndexes[TLInner2] = isOn ? r9 : rW;
		colorIndexes[BRInner2] = r6;
		colorIndexes[CornerInner2] = isOn ? r7 : r3;
		colorIndexes[Fill] = isOn ? r7 : r3;
		}
	else	// disabled
		{
		colorIndexes[TLFrame] = isOn ? r8 : r5;
		colorIndexes[BRFrame] = r6;
		colorIndexes[CornerFrame] = isOn ? r8 : r5;
		colorIndexes[TLInner1] = isOn ? r4 : r2;
		colorIndexes[BRInner1] = isOn ? r4 : r2;
		colorIndexes[CornerInner1] = isOn ? r4 : r2;
		colorIndexes[TLInner2] = isOn ? r4 : r2;
		colorIndexes[BRInner2] = isOn ? r4 : r2;
		colorIndexes[CornerInner2] = isOn ? r4 : r2;
		colorIndexes[Fill] = isOn ? r4 : r2;
		}
	
	// Draw the outer frame.

	::RGBForeColor(&gAGARamp[colorIndexes[TLFrame]]);
	::MoveTo(r.left, r.bottom - 2);
	::LineTo(r.left, r.top);
	::LineTo(r.right - 1, r.top);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRFrame]]);
	::Move(0, 1);
	::LineTo(r.right - 1, r.bottom - 1);
	::LineTo(r.left, r.bottom - 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	// Draw the 1st pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner1]]);
	::MoveTo(r.left + 1, r.bottom - 3);
	::LineTo(r.left + 1, r.top + 1);
	::LineTo(r.right - 2, r.top + 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner1]]);
	::Move(0, 1);
	::LineTo(r.right - 2, r.bottom - 2);
	::LineTo(r.left + 1, r.bottom - 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	// Draw the 2nd pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner2]]);
	::MoveTo(r.left + 2, r.bottom - 4);
	::LineTo(r.left + 2, r.top + 2);
	::LineTo(r.right - 3, r.top + 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner2]]);
	::Move(0, 1);
	::LineTo(r.right - 3, r.bottom - 3);
	::LineTo(r.left + 2, r.bottom - 3);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	// Fill the interior.

	::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
	::InsetRect(&r, 3, 3);
	::PaintRect(&r);
	::InsetRect(&r, -3, -3);
	}

static void DrawLargeBevelEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	// Draw the button edges as specified for a
	// "large" sized button.

	enum { TLFrame, TLInner1, TLInner2, TLInner3, BRFrame, BRInner1, BRInner2, BRInner3, CornerFrame, CornerInner1, CornerInner2, CornerInner3, Fill, kNumEdgeColors };

	UInt8	colorIndexes[kNumEdgeColors];

	Rect	r = *bounds;
	
	if (isPressed)
		{
		colorIndexes[TLFrame] = rA2;
		colorIndexes[BRFrame] = r11;
		colorIndexes[CornerFrame] = r12;
		colorIndexes[TLInner1] = rA1;
		colorIndexes[BRInner1] = r4;
		colorIndexes[CornerInner1] = r11;
		colorIndexes[TLInner2] = r11;
		colorIndexes[BRInner2] = r5;
		colorIndexes[CornerInner2] = r8;
		colorIndexes[TLInner3] = r9;
		colorIndexes[BRInner3] = r6;
		colorIndexes[CornerInner3] = r7;
		colorIndexes[Fill] = r7;
		}
	else if (isEnabled)
		{
		colorIndexes[TLFrame] = isOn ? rA2 : r9;
		colorIndexes[BRFrame] = isOn ? r11 : rA1;
		colorIndexes[CornerFrame] = isOn ? r12 : r10;
		colorIndexes[TLInner1] = isOn ? rA1 : r3;
		colorIndexes[BRInner1] = isOn ? r4 : r10;
		colorIndexes[CornerInner1] = isOn ? r11 : r5;
		colorIndexes[TLInner2] = isOn ? r11 : r1;
		colorIndexes[BRInner2] = isOn ? r5 : r8;
		colorIndexes[CornerInner2] = isOn ? r8 : r4;
		colorIndexes[TLInner3] = isOn ? r9 : rW;
		colorIndexes[BRInner3] = r6;
		colorIndexes[CornerInner3] = isOn ? r7 : r3;
		colorIndexes[Fill] = isOn ? r7 : r3;
		}
	else	// disabled
		{
		colorIndexes[TLFrame] = isOn ? r8 : r5;
		colorIndexes[BRFrame] = isOn ? r6 : r7;
		colorIndexes[CornerFrame] = isOn ? r7 : r6;
		colorIndexes[TLInner1] = isOn ? r4 : r2;
		colorIndexes[BRInner1] = isOn ? r4 : r2;
		colorIndexes[CornerInner1] = isOn ? r4 : r2;
		colorIndexes[TLInner2] = isOn ? r4 : r2;
		colorIndexes[BRInner2] = isOn ? r4 : r2;
		colorIndexes[CornerInner2] = isOn ? r4 : r2;
		colorIndexes[TLInner3] = isOn ? r4 : r2;
		colorIndexes[BRInner3] = isOn ? r4 : r2;
		colorIndexes[CornerInner3] = isOn ? r4 : r2;
		colorIndexes[Fill] = isOn ? r4 : r2;
		}
	
	// Draw the outer frame.

	::RGBForeColor(&gAGARamp[colorIndexes[TLFrame]]);
	::MoveTo(r.left, r.bottom - 2);
	::LineTo(r.left, r.top);
	::LineTo(r.right - 1, r.top);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRFrame]]);
	::Move(0, 1);
	::LineTo(r.right - 1, r.bottom - 1);
	::LineTo(r.left, r.bottom - 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	// Draw the 1st pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner1]]);
	::MoveTo(r.left + 1, r.bottom - 3);
	::LineTo(r.left + 1, r.top + 1);
	::LineTo(r.right - 2, r.top + 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner1]]);
	::Move(0, 1);
	::LineTo(r.right - 2, r.bottom - 2);
	::LineTo(r.left + 1, r.bottom - 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	// Draw the 2nd pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner2]]);
	::MoveTo(r.left + 2, r.bottom - 4);
	::LineTo(r.left + 2, r.top + 2);
	::LineTo(r.right - 3, r.top + 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner2]]);
	::Move(0, 1);
	::LineTo(r.right - 3, r.bottom - 3);
	::LineTo(r.left + 2, r.bottom - 3);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	// Draw the 3rd pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner3]]);
	::MoveTo(r.left + 3, r.bottom - 5);
	::LineTo(r.left + 3, r.top + 3);
	::LineTo(r.right - 4, r.top + 3);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner3]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner3]]);
	::Move(0, 1);
	::LineTo(r.right - 4, r.bottom - 4);
	::LineTo(r.left + 3, r.bottom - 4);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner3]]);
	::Line(0, 0);

	// Fill the interior.

	::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
	::InsetRect(&r, 4, 4);
	::PaintRect(&r);
	::InsetRect(&r, -4, -4);
	}

void AGADrawBevelButton(Rect* bounds, ButtonFrameType frameType, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		AGADrawBevelButtonAtDepth(deep, bounds, frameType, isOn, isPressed, isEnabled);
	}

void AGADrawBevelButtonAtDepth(Boolean deep, Rect* bounds, ButtonFrameType frameType, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	Rect	r = *bounds;
	
	if (frameType == kAutoFrame)
		frameType = AGAGetDefaultFrameType(bounds->right - bounds->left);

	if (deep)
		{
		if (frameType == kMiniFrame)
			DrawMiniBevelEdges(bounds, isOn, isPressed, isEnabled);
		else if (frameType == kSmallFrame)
			DrawSmallBevelEdges(bounds, isOn, isPressed, isEnabled);
		else
			DrawLargeBevelEdges(bounds, isOn, isPressed, isEnabled);
		}
	else
		{
		if (isPressed || isOn)
			::RGBForeColor(&gAGARamp[rB]);
		else
			::RGBForeColor(&gAGARamp[rW]);

		::PaintRect(&r);
		
		if (! isEnabled)
			PenPatGray();

		::RGBForeColor(&gAGARamp[rB]);
		::FrameRect(&r);
		::PenNormal();
		}
	}

ButtonFrameType AGAGetDefaultFrameType(SInt16 buttonWidth)
	{
	if (buttonWidth > 39)
		return kLargeFrame;
	else if (buttonWidth > 21)
		return kSmallFrame;
	else
		return kMiniFrame;
	}

//
// AGAInstallGrayWCTB ------------------------------------------------
//

OSErr AGAInstallGrayWCTB(WindowPtr aWindow)
	{
	RGBColor	dontCareColor;
	Boolean		dontCareBoolean;

	return AGAInstallWCTB(aWindow, &gAGARamp[r2], &dontCareBoolean, &dontCareColor);
	}

//
// AGAInstallWCTB ------------------------------------------------
//

OSErr AGAInstallWCTB(WindowPtr aWindow, const RGBColor* newColor, Boolean* hasOldColor, RGBColor* oldColor)
	{
	//
	// If it's a color window, make a copy of the default
	// color table, set the background to the specified gray,
	// and install it into the window.
	//

	OSErr			err = noErr;
    AuxWinHandle	auxWinHandle;
    CTabHandle		colorTableHandle;
    
    *hasOldColor = false;

    ::GetAuxWin(NULL, &auxWinHandle);

    if (auxWinHandle != NULL)
    	{
	    colorTableHandle = (**auxWinHandle).awCTable;
	    
	    *hasOldColor = true;
	    *oldColor = (**colorTableHandle).ctTable->rgb;
	    
	    err = ::HandToHand((Handle*) &colorTableHandle);

		if (err == noErr)
			{
		    (**colorTableHandle).ctTable->rgb = *newColor;    
		    ::SetWinColor(aWindow, (WCTabHandle) colorTableHandle);
		    }
		}
    
    return err;
	}

//
// AGADrawBorderFrame ------------------------------------------------
//

void AGADrawBorderFrame(const Rect* bounds, Boolean isEnabled, Boolean isNotched, Boolean drawFrame, const RGBColor* windowBackgroundColor)
	{
	// Draw the frame with shading outside.
	
	enum { Shadow, Frame, Light, kNumColors };

	Rect	r = *bounds;
	
	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		AGADrawingEnvironment	env;

		if (drawFrame)
			{
			if (deep)
				{
				if (isEnabled)
					::RGBForeColor(&gAGARamp[rB]);
				else
					::RGBForeColor(&gAGARamp[r10]);
				}
			else
				{
				::RGBForeColor(&gAGARamp[rB]);

				if (isEnabled)
					PenPatBlack();
				else
					PenPatGray();
				}

			::InsetRect(&r, 1, 1);
			
			if (isNotched)
				{
				::MoveTo(r.right - 1, r.top);
				::LineTo(r.left, r.top);
				::LineTo(r.left, r.bottom - 1);
				::LineTo(r.right - 16, r.bottom - 1);
				::Line(0, -15);
				::Line(15, 0);
				::LineTo(r.right - 1, r.top);
				}
			else
				::FrameRect(&r);

			::InsetRect(&r, -1, -1);
			}
		
		if (deep)
			{
			if (isEnabled)
				::RGBForeColor(&gAGARamp[r5]);
			else
				::RGBForeColor(windowBackgroundColor);

			::MoveTo(r.left, r.bottom - 2);
			::LineTo(r.left, r.top);
			::LineTo(r.right - 2, r.top);
			
			if (isEnabled)
				::RGBForeColor(&gAGARamp[rW]);
			else
				::RGBForeColor(windowBackgroundColor);

			::MoveTo(r.left + 1, r.bottom - 1);

			if (isNotched)
				{
				::LineTo(r.right - 16, r.bottom - 1);
				::Line(0, -15);
				::Line(15, 0);
				}
			else
				::LineTo(r.right - 1, r.bottom - 1);

			::LineTo(r.right - 1, r.top + 1);
			}
		}
	}

//
// AGAComputeTargetBorderRegion ------------------------------------------------
//

void AGAComputeTargetBorderRegion(RgnHandle borderRegion, const Rect* bounds, Boolean isNotched)
	{
	// Compute the border region to be painted.

	RgnHandle	innerRegion = ::NewRgn();
	Rect		r = *bounds;

	// First create a region framing the outer edge of the border.
	::OpenRgn();
	::FrameRoundRect(&r, 4, 4);
	::CloseRgn(borderRegion);

	// Now subtract away the hollow inside area.
	::InsetRect(&r, 2, 2);
	::RectRgn(innerRegion, &r);
	::DiffRgn(borderRegion, innerRegion, borderRegion);
	
	if (isNotched)
		{
		// OK, region madness as we add the top/left sides
		// of a new "notch" rect frame, and also remove the
		// bottom/right corner of our existing non-notched frame.
		
		RgnHandle	notchRegion = ::NewRgn();
		Rect		notchRect;
		
		// Create a new round rect for the corner area and
		// union it with the existing region.
		notchRect = *bounds;
		notchRect.left = notchRect.right - 17;
		notchRect.top = notchRect.bottom - 17;

		::RectRgn(notchRegion, &notchRect);
		::UnionRgn(borderRegion, notchRegion, borderRegion);
		
		// Subtract away the corner of the existing frame.
		notchRect.left += 2;
		notchRect.top += 2;

		::RectRgn(notchRegion, &notchRect);
		::DiffRgn(borderRegion, notchRegion, borderRegion);
		
		// Almost done. There's an extra pixel at each notch
		// corner that needs to be removed to make round corners.
		
		Rect	pixelRect;
		
		// The top/right corner.
		pixelRect = notchRect;
		pixelRect.left = pixelRect.right - 1;
		pixelRect.top--;
		pixelRect.bottom = pixelRect.top + 1;
		::RectRgn(notchRegion, &pixelRect);
		::DiffRgn(borderRegion, notchRegion, borderRegion);
		
		// The bottom/left corner.
		pixelRect = notchRect;
		pixelRect.top = pixelRect.bottom - 1;
		pixelRect.left--;
		pixelRect.right = pixelRect.left + 1;
		::RectRgn(notchRegion, &pixelRect);
		::DiffRgn(borderRegion, notchRegion, borderRegion);
		
		// The top/left corner.
		pixelRect = notchRect;
		pixelRect.top -= 2;
		pixelRect.left -= 2;
		pixelRect.bottom = pixelRect.top + 1;
		pixelRect.right = pixelRect.left + 1;
		::RectRgn(notchRegion, &pixelRect);
		::DiffRgn(borderRegion, notchRegion, borderRegion);
		
		::DisposeRgn(notchRegion);
		}
	
	::DisposeRgn(innerRegion);
	}

//
// AGAClipFurther ------------------------------------------------
//

void AGAClipFurther(Rect* clippingRect)
	{
	// Shrink the clip region further to prevent drawing outside
	// the specified rect.

	RgnHandle	oldClip = ::NewRgn();
	RgnHandle	newClip = ::NewRgn();
	
	if ((oldClip != NULL) && (newClip != NULL))
		{
		::GetClip(oldClip);
		::RectRgn(newClip, clippingRect);

		::SectRgn(oldClip, newClip, newClip);
		::SetClip(newClip);
		
		::DisposeRgn(oldClip);
		::DisposeRgn(newClip);
		}
	}

//
// AGA_PLstrcpy ------------------------------------------------
//

void AGA_PLstrcpy(StringPtr str1, ConstStr255Param str2)
	{
	// Copy pascal string str2 into pascal string str1.

	::BlockMoveData((Ptr) str2, (Ptr) str1, 1 + str2[0]);
	}

//
// AGA Defaults ------------------------------------------------
//

void SetGrayCouncilDefault(UInt32 itemMask, Boolean turnOn)
	{
	if (turnOn)
		gAGADefaults |= itemMask;
	else
		gAGADefaults &= ~itemMask;
	}

Boolean TestGrayCouncilDefault(UInt32 itemMask)
	{
	return ((gAGADefaults & itemMask) != 0);
	}

//
// Variable-color background support -------------------------------
//
// Tab panels provide us a more complicated background color situation
// than other situations. While the normal AGA background is gray ramp
// value r2, the tab panels are lighter, namely r1. However, when a tab
// panel is dimmed, it becomes r2. Thus the panel subpanes can't just
// blindly use the window background color, and if the tab panel becomes
// dimmed they can't just blindly use r1 either.
//
// So what we do is maintain static color values that are picked up by
// any adapter classes that need to erase their background. There are
// separate color values for enabled and disabled states. When a panel
// view hierarchy is installed into a tab panel pane, the tab panel class
// temporarily sets these color values so that any subpanes of the panel
// will pick them up as they are instantiated.
//

static Boolean	gUseSpecialBackgroundColors = false;	// toggled by tab panel installer
static RGBColor	gEnabledBackgroundColor;
static RGBColor	gDisabledBackgroundColor;

Boolean DetermineAGABackgroundColors(RGBColor* enabledColor, RGBColor* disabledColor, Boolean installIfNotCustom)
	{
	Boolean	wereColorsSet = false;

	if (gUseSpecialBackgroundColors)
		{
		*enabledColor = gEnabledBackgroundColor;
		*disabledColor = gDisabledBackgroundColor;
		wereColorsSet = true;
		}
	else if (installIfNotCustom)
		{
		// The calling view must have focused and set the port bg color.
		::GetBackColor(enabledColor);
		*disabledColor = *enabledColor;
		wereColorsSet = true;
		}

	return wereColorsSet;
	}
 
void InstallAGABackgroundColors(AGAObject* itsAGAObject, Boolean installIfNotCustom)
	{
	RGBColor	enabledColor;
	RGBColor	disabledColor;

	if (DetermineAGABackgroundColors(&enabledColor, &disabledColor, installIfNotCustom))
		itsAGAObject->SetBackgroundColors(&enabledColor, &disabledColor);
	}

void TurnOnCustomAGABackgroundColors(RGBColor* enabledColor, RGBColor* disabledColor)
	{
	gUseSpecialBackgroundColors = true;
	gEnabledBackgroundColor = *enabledColor;
	gDisabledBackgroundColor = *disabledColor;
	}

void TurnOffCustomAGABackgroundColors()
	{
	gUseSpecialBackgroundColors = false;
	}

//
// AGAStringOut --------------------------------------------------------
//
// This is the routine we use for all of our one-line text output.
//

void AGAStringOut(StringPtr aString,
					Rect* centeringFrame,
					SInt32 truncation,
					SInt32 justification,
					AGAStringOutColor outColor,
					Boolean deep,
					const AGATextStyle& textStyle)
	{
	Str255		outputString;
	FontInfo	fontInfo;
	SInt16		frameWidth = centeringFrame->right - centeringFrame->left;
	SInt16		frameHeight = centeringFrame->bottom - centeringFrame->top;
	Point		penOrigin;
	Boolean		setColor = true;
	UInt8		colorIndex;
	
	textStyle.PrepareForDrawing();
	::GetFontInfo(&fontInfo);

	switch (outColor)
		{
		case kNormalOutput:
			colorIndex = rB;
			break;
		case kDisabledOutput:
			if (deep)
				colorIndex = r7;
			else
				colorIndex = rB;
			break;
		case kInverseOutput:
			colorIndex = rW;
			break;
		case kUseCurrentColor:
			setColor = false;
			break;
		}

	if (setColor)
		::RGBForeColor(&gAGARamp[colorIndex]);
	
	AGA_PLstrcpy(outputString, aString);
	
	// Truncate the string if it does not fit the width of the frame.
	if (truncation != kNoTruncation)
		{
		// If the string would be truncated at the current
		// font settings, try again with condensed text.
		if (::TruncString(frameWidth, outputString, truncation) == truncated)
			{
			AGA_PLstrcpy(outputString, aString);
			::TextFace(condense);
			(void) ::TruncString(frameWidth, outputString, truncation);
			}
		}
	
	// Center the text vertically. Pen goes at baseline of
	// text, i.e. between the ascent and descent.
	penOrigin.v = centeringFrame->top + (frameHeight / 2) + ((fontInfo.ascent + fontInfo.descent) / 2) - fontInfo.descent;
	
	// Position the text horizontally left, centered, or right, based
	// on the specified justification and the system script direction.
	switch (RuntimeJustify(justification))
		{
		case teCenter:
			penOrigin.h = centeringFrame->left + (frameWidth / 2) - (::StringWidth(outputString) / 2);
			break;
		case teFlushRight:
			penOrigin.h = centeringFrame->right - ::StringWidth(outputString);
			break;
		default: //case teFlushLeft:
			penOrigin.h = centeringFrame->left;
			break;
		}

	::MoveTo(penOrigin.h, penOrigin.v);

	GrafPtr	currentPort;
	::GetPort(&currentPort);
	SInt16	savedTxMode = currentPort->txMode;
	::TextMode(srcOr);
	::DrawString(outputString);
	::TextMode(savedTxMode);
	
	if ((outColor == kDisabledOutput) && ! deep)
		{
		PenPatGray();
		::PenMode(patBic);
		::PaintRect(centeringFrame);

		::PenNormal();
		}
	}

//
// AGATextBox --------------------------------------------------------
//
// This is a replacement for TETextBox.
//

void AGATextBox(const void *textPtr,
				SInt32 textLength,
				const Rect *textFrame,
				SInt16 justification,
				AGAStringOutColor outColor,
				Boolean deep,
				const AGATextStyle& textStyle,
				const RGBColor* teBackgroundColor)
	{
	FontInfo	fontInfo;
	Boolean		setColor = true;
	UInt8		colorIndex;
	RGBColor	savedBackColor;
	
	::GetBackColor(&savedBackColor);
	::RGBBackColor(teBackgroundColor);
	
	textStyle.PrepareForDrawing();
	::GetFontInfo(&fontInfo);

	switch (outColor)
		{
		case kNormalOutput:
			colorIndex = rB;
			break;
		case kDisabledOutput:
			if (deep)
				colorIndex = r7;
			else
				colorIndex = rB;
			break;
		case kInverseOutput:
			colorIndex = rW;
			break;
		case kUseCurrentColor:
			setColor = false;
			break;
		}

	if (setColor)
		::RGBForeColor(&gAGARamp[colorIndex]);
	
	::TETextBox(textPtr, textLength, textFrame, justification);

	if ((outColor == kDisabledOutput) && ! deep)
		{
		PenPatGray();
		::PenMode(patBic);
		::PaintRect(textFrame);

		::PenNormal();
		}

	::RGBBackColor(&savedBackColor);
	}

//
// Previously static routines used only inside these files -------------------------
//

SInt16 MinSInt16(SInt16 a, SInt16 b) { return (a < b) ? a : b; }
SInt16 MaxSInt16(SInt16 a, SInt16 b) { return (a > b) ? a : b; }
SInt32 MinSInt32(SInt32 a, SInt32 b) { return (a < b) ? a : b; }
SInt32 MaxSInt32(SInt32 a, SInt32 b) { return (a > b) ? a : b; }

void DelayFuture(SInt32 delayTicks)
	{
	// Delay for the specified number of ticks.

	SInt32	dontCare;
	::Delay(delayTicks, &dontCare);
	}

void DelayFutureWhileStillDown(SInt32 delayTicks)
	{
	// Delay for the specified number of ticks, but bail
	// out if the user lets go of the mouse first.

	SInt32	future = delayTicks + ::TickCount();

	while (::StillDown() && (::TickCount() < future))
		{};
	}

SInt32 RuntimeJustify(SInt32 userJust)
	{
	// Return the actual runtime justification value for
	// the supplied value. The caller can use the result
	// alone to know determine drawing direction.
	// If the supplied value specifies default, we need
	// to check the current system script direction.
	// Otherwise, whatever was specified is OK as is.

	if (userJust == teFlushDefault)
		{
		if (::GetSysDirection() == 0)
			return teFlushLeft;
		else
			return teFlushRight;
		}
	else
		return userJust;
	}

//
// Global variables ---------------------------------------------------
//

const Str255 gGrayCouncilCopyright = kGrayCouncilCopyright;

Boolean gAGAHasColorQD;
UInt32	gAGADefaults = 0;
Boolean	gAGAStdControls = true;
Boolean	gAGAExtraControls = true;
Boolean	gAppearanceManagerSavvyInit = true;
Boolean	gAaronKaleidoscopeSavvyInit = true;
RGBColor	gFinder7EmptyProgressColor = { 0xCCCC, 0xCCCC, 0xFFFF };
RGBColor	gFinder7FilledProgressColor = { 0x4000, 0x4000, 0x4000 };
AGAGroupsContainer* gGroupsContainer;

// Standard system text style, used by most standard controls.
AGATextStyle gAGAStdSystemStyle = AGATextStyle();

// Standard small text style, used for slider labels.
AGATextStyle gAGAStdSmallStyle = AGATextStyle(::GetAppFont(), 10, normal);

// Standard small bold text style, used for small tab panels.
AGATextStyle gAGAStdBoldSmallStyle = AGATextStyle(::GetAppFont(), 10, bold);

// Extra small text style, a de facto standard.
AGATextStyle gAGAExtraSmallStyle = AGATextStyle(::GetAppFont(), 9, normal);

// These are the AGA-specified color values.
RGBColor gAGARamp[kNumRampColors] =
	{
	{65535, 65535, 65535},	// rW	white
	{61166, 61166, 61166},	// r1	lightest gray
	{56797, 56797, 56797},	// r2
	{52428, 52428, 52428},	// r3
	{48059, 48059, 48059},	// r4	light grays
	{43690, 43690, 43690},	// r5
	{39321, 39321, 39321},	// r6
	{34952, 34952, 34952},	// r7
	{30583, 30583, 30583},	// r8	dark grays
	{26214, 26214, 26214},	// r9
	{21845, 21845, 21845},	// r10
	{17476, 17476, 17476},	// r11
	{13107, 13107, 13107},	// r12	darkest "n" gray
	{8738, 8738, 8738},		// rA1	additional darker gray value 1
	{4369, 4369, 4369},		// rA2	additional darker gray value 2
	{0, 0, 0},				// rB	black
	{52428, 52428, 65535},	// rP1	lightest purple
	{39321, 39321, 65535},	// rP2	light purple
	{26214, 26214, 65535},	// rP3	dark purple
	{13107, 13107, 26214}	// rP4	darkest purple
	};

