//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// The Gray Council provides a set of standard C++ classes that implement
// the standard Apple Grayscale Appearance. The core classes do not
// require any other code such as a particular class framework.
// There are separate files provided that plug The Gray Council core
// into class frameworks.
//
// This file defines classes used by the other core controls.
//
// Classes defined below:
//	AGAObject -- abstract base class for most others
//	MExclusiveObject -- mixin class for mutual exclusivity (e.g. radio)
//	AGATextStyle -- text style attributes container
//	AGADrawingEnvironment -- stack-based save/restore class for drawing attributes
//	AGATrackingIndicator -- abstract superclass for scroll bar and slider
//	AGAOffscreenImage -- class for storing and drawing an offscreen pixmap
//	GDIterator -- stack-based class for multi-screen drawing
//	AGAGroupsContainer -- class for managing MExclusiveObject groups
//

#pragma once

#ifndef __GCCORE__
#define __GCCORE__

#define kGrayCouncilCopyright "\pThe Gray Council 1996-1997 by Trygve Isaacson. All Rights Reserved."

#include <Errors.h>
#include <Events.h>
#include <Fonts.h>
#include <Gestalt.h>
#include <Icons.h>
#include <Menus.h>
#include <OSUtils.h>
#include <Quickdraw.h>
#include <Resources.h>
#include <TextEdit.h>
#include <TextUtils.h>
#include <Types.h>
#include <Windows.h>

//
// You must call InitGrayCouncil after initializing the Toolbox
// and before instantiating any GrayCouncil objects. This routine
// is called automatically if you call the framework-specific
// initializer. If it returns an error, you should quit. Currently
// the only error situation is an out-of-memory error.
//
// There are two global variables you can set *before* calling
// InitGrayCouncil to affect things. These default to TRUE.
//
// gAppearanceManagerSavvyInit - if true, and the Appearance Manager
//	is present...
// gAaronKaleidoscopeSavvyInit - if true, and Aaron or Kaleidoscope
//	are present...
//
// If either of these situations is detected (AM & AM-savvy, or
// Aaron & Aaron-savvy), then:
// - the global variable gAGAStdControls will be set to false, so
//	that Control Manager controls are never drawn by The Gray Council.
//	The Control Manager CDEF, or the replacement, will draw controls.
// - the global variable gAGAExtraControls will be set to false, so
//	that The Gray Council will draw progress bars the way the System
//	7 Finder does, so that Aaron & Kaleidoscope will intercept it.
//
// You can also set these two global variables (gAGAStdControls &
//	gAGAExtraControls) yourself immediately after InitGrayCouncil if
//	want to control it yourself. You can't change these on-the-fly
//	later, except with certain constraints. Read the docs!
//

extern OSErr InitGrayCouncil();

#pragma mark AGAObject

//
// AGAObject ---------------------------------------------------------------
//
// All of the visual GrayCouncil objects inherit from AGAObject.
//
// In general, all you have to do with any AGAObject is instantiate it,
// and then call its DrawObject() and TrackMouse() functions when needed.
// When calling the drawing and mouse tracking functions (and this includes
// functions where you set an attribute with redraw set to true), you must
// ensure that the GrafPort and coordinate system are correct prior to the
// call.
//
// Public functions:
// - AGAObject. You supply its coordinate boundary.
// - GetObjectBounds. Returns its coordinate boundary.
// - SetObjectBounds. Call this to change its coordinate boundary.
// - DrawObject. Draws the object. Assumes current port, pen normal.
// - ContainsMouse. Returns true if specified mouse would hit it.
// - TrackMouse. Handles a mouse click. Returns true if hit did something.
// - SetEnable. Enables or disables it.
// - SetBackgroundColors. Sets the colors the object will use when erasing its background.
//
// You can #define AGA_SUPERCLASS to be TObject or whatever
// if you prefer to have the AGA stuff not be its own root
// class hierarchy.
//

class AGAObject

#ifdef AGA_SUPERCLASS
	: public AGA_SUPERCLASS
#endif // AGA_SUPERCLASS

	{
	public:

		AGAObject(Rect* bounds);
		virtual ~AGAObject();
		
		virtual void	GetObjectBounds(Rect* bounds);
		virtual void	SetObjectBounds(Rect* bounds, Boolean redraw);
		virtual void	DrawObject();
		virtual Boolean	ContainsMouse(Point mouseLocation);
		virtual Boolean	TrackMouse(Point mouseLocation);
		virtual void	SetEnable(Boolean isEnabled, Boolean redraw);
		virtual void	SetActive(Boolean isActive, Boolean redraw);
		virtual void	SetBackgroundColors(RGBColor* enabledColor, RGBColor* disabledColor);
		virtual SInt32	GetValue();
		virtual void	SetValue(SInt32 newValue, Boolean redraw);

		enum { kRedraw = true, kDontRedraw = false };
		enum { kEnabled = true, kDisabled = false };
		enum { kActive = true, kInactive = false };
		enum { kPressed = true, kNotPressed = false };
		enum { kFrameInside = true, kFrameOutside = false };
		enum { kIsDefault = true, kIsNotDefault = false };
		enum { kAutomaticState = true, kNoAutomaticState = false };
		enum { kLiveTracking = true, kNoLiveTracking = false };
		enum { kProportional = true, kNotProportional = false };
		enum { kHorizontal = true, kVertical = false };
		enum { kTrackingIn = true, kTrackingOut = false };

	protected:

		// HitTest and SetTrackingState are called repeatedly by the default
		// mouse tracker.
		virtual Boolean	HitTest(Point mouseLocation);
		virtual void	SetTrackingState(Boolean isIn);
		virtual void	ApplyBackgroundColor();
		
		Rect		mBounds;
		Boolean		mEnabled;	// if false, ignores mouse clicks etc.
		Boolean		mActive;	// if false, draws with dim appearance
		RGBColor	mEnabledBackgroundColor;
		RGBColor	mDisabledBackgroundColor;
	};

#pragma mark MExclusiveObject

//
// MExclusiveObject ---------------------------------------------------------------
//
// MExclusiveObject is a mixin class for automatically keeping several
// objects in a group mutually exclusive. The radio button class below
// uses this to allow automatic radio button group behavior. MExclusiveObject
// maintains knowledge of "groups" of objects. When one member of the group is
// "hit", the other members of the group are sent the TurnOff() message. To
// identify each object uniquely, the object has a pair of 32-bit identifiers.
// Typically these can set to the group's ID and GrafPtr. This allows having
// multiple windows that use the same group IDs. If you set the IDs to kNoGroupID,
// then this automatic maintenance is bypassed.
// The MacApp and PowerPlant adapter classes do not use this facility, because
// these frameworks provide their own radio button group management.
//

class MExclusiveObject
	{
	public:

		MExclusiveObject(UInt32 groupIDPart1, UInt32 groupIDPart2);
		virtual ~MExclusiveObject();
		
		virtual void	SetGroupID(UInt32 groupIDPart1, UInt32 groupIDPart2);

		virtual void	TurnOff();

		UInt32	mGroupIDPart1;
		UInt32	mGroupIDPart2;
	};

const UInt32 kNoGroupID = 0xFFFFFFFF;

#pragma mark AGATextStyle

//
// AGATextStyle ---------------------------------------------------------------
//
// Many of the controls can have their text style specified. AGATextStyle
// is used for this. It defines the font, size, and style to be used for
// drawing text.
//

class AGATextStyle
	{
	public:
	
		static void NormalizeTextStyle(TextStyle& aTextStyle, SInt16 systemFontNum, SInt16 applFontNum);

		AGATextStyle();
		AGATextStyle(SInt16 fontNum, SInt16 fontSize, Style fontStyle);
		AGATextStyle(StringPtr fontName, SInt16 fontSize, Style fontStyle);
		AGATextStyle(const TextStyle& textStyle);

#ifdef __PowerPlant__
		AGATextStyle(ResIDT inTextTraitsID);	// Note: implementation is in GCCorePP.cpp
#endif // __PowerPlant__
		
		void	PrepareForDrawing() const;

		SInt16	mFontNum;
		SInt16	mFontSize;
		Style	mFontStyle;
	};

// These globals should be used in most cases. You can also roll your own.
extern AGATextStyle gAGAStdSystemStyle;		// Typically Chicago 12. Standard.
extern AGATextStyle gAGAStdSmallStyle;		// Typically Geneva 10. Slider labels.
extern AGATextStyle gAGAStdBoldSmallStyle;	// Typically Geneva 10 bold. Small tab panels.
extern AGATextStyle gAGAExtraSmallStyle;	// Typically Geneva 9. Use where you need a small control.

#pragma mark AGADrawingEnvironment

//
// AGADrawingEnvironment ---------------------------------------------------------------
//
// AGADrawingEnvironment is a stack-based object that saves the existing drawing
// settings upon construction, and restores them upon destruction. It also
// cleanses the pen after saving the existing settings, so that our drawing
// code can assume basic settings are in place.
//

class AGADrawingEnvironment
	{
	public:
	
		AGADrawingEnvironment();
		virtual ~AGADrawingEnvironment();
		
		void	SaveEnvironment();		// called by constructor
		void	RestoreEnvironment();	// called by destructor
	
		void	Cleanse();				// called by constructor after saving environment
										// sets drawing state to a clean set of values
	
	protected:
	
		PenState	mSavedPenState;
		RGBColor	mSavedForeColor;
		RGBColor	mSavedBackColor;
		SInt16		mSavedFontNum;
		SInt16		mSavedFontSize;
		Style		mSavedFontStyle;
	};

#pragma mark AGATrackingIndicator

//
// AGATrackingIndicator ---------------------------------------------------------------
//
// AGATrackingIndicator is the abstract superclass for things like
// sliders and scroll bars that have a moving indicator to reflect
// the object's position. It has a minimum/maximum range, a current
// value, and optionally a page size for proportional indicator
// display. You can turn live tracking on or off; if live tracking
// is off, a "ghost" indicator is dragged and no notification occurs
// until the tracking completes.
//

class AGATrackingIndicator;

// This is the live tracking notification prototype for AGATrackingIndicator.
typedef void (*AGATrackingIndicatorNotifyPtr)(AGATrackingIndicator* theIndicator, SInt32 itsNewValue, void* userData);

class AGATrackingIndicator : public AGAObject
	{
	public:

		AGATrackingIndicator(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue);
		AGATrackingIndicator(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, SInt32 pageSize, Boolean liveTracking, Boolean proportional);
		virtual ~AGATrackingIndicator();
		
		virtual void	InstallNotificationRoutine(AGATrackingIndicatorNotifyPtr notificationRoutine, void* userData);
		
		virtual void	SetAttributes(Boolean isHorizontal, Boolean liveTracking, Boolean isProportional);
		virtual void	SetLiveTracking(Boolean liveTracking);

		virtual void	DrawObject();
		virtual Boolean	TrackMouse(Point mouseLocation);
		
		virtual SInt32	GetValue();
		virtual void	SetValue(SInt32 newValue, Boolean redraw);

		virtual void	GetRange(SInt32* minimum, SInt32* maximum);
		virtual void	SetRange(SInt32 newMinimum, SInt32 newMaximum, Boolean redraw);
		
		virtual SInt32	GetPageSize();
		virtual void	SetPageSize(SInt32 newPageSize, Boolean redraw);

	protected:

		enum { kNoTrackPress, kPageMinusPress, kArrowMinusPress, kIndicatorPress, kArrowPlusPress, kPagePlusPress };

		virtual Boolean	TrackPart(Point mouseLocation, SInt32 partHit);
		virtual SInt32	TrackTestPart(Point mouseLocation);
		virtual Boolean	IsValidIndicatorTrackMouse(Point mouseLocation);
		virtual SInt32	GetValueFromMouseDelta(SInt32 originalValue, Point originalMouseLocation, Point newMouseLocation);
		virtual void	SetGhostValue(SInt32 newValue, Boolean redraw);

		virtual void	DrawBackground();
		virtual void	DrawIndicatorTrack();
		virtual void	DrawTrackEnds();
		virtual void	DrawIndicator();
		virtual void	RemoveIndicator(SInt32 newValue);
		virtual void	DrawGhost();
		virtual void	RemoveGhost(SInt32 newValue);
		
		virtual void	SetRawValue(SInt32 newValue, Boolean redraw);
		virtual void	NotifyValue();
		
		SInt32	mMinimum;
		SInt32	mMaximum;
		SInt32	mValue;
		SInt32	mPageSize;
		SInt32	mGhostValue;
		Boolean	mLiveTracking;
		Boolean	mIsProportional;
		Boolean	mIsHorizontal;
		Boolean	mIsPressed;
		
		AGATrackingIndicatorNotifyPtr	mNotificationRoutine;
		void*							mUserData;
	};

#pragma mark AGAOffscreenImage

//
// AGAOffscreenImage ------------------------------------------------------------
//
// AGAOffscreenImage holds the data needed to blast a clipped region
// of pixel values onto the current port. This is used by the AGARadioButton
// class to avoid the ultra-tedious pixel-by-pixel drawing of the grayscale
// radio button images, and by the tab panel for the tab corners.
//

class AGAOffscreenImage

#ifdef AGA_SUPERCLASS
	: public AGA_SUPERCLASS
#endif // AGA_SUPERCLASS
	{
	public:
	
		AGAOffscreenImage();
		virtual ~AGAOffscreenImage();
		
		virtual OSErr	CreateImageData(Point imageSize,
								SInt8* imageColorIndexes,
								RgnHandle clippingRegion,
								const RGBColor* outsideColor);
		
		void	DrawImage(SInt16 imageHOrigin, SInt16 imageVOrigin);	// splats 1 copy of the image
		void	DrawPattern(Rect* imageRect);							// splats multiple copies of the image

	protected:

		Rect			mImageSourceRect;	// 0, 0, x, y
		PixMapHandle	mImagePixMap;
		RgnHandle		mClippingRegion;
	};

#pragma mark GDIterator

//
// GDIterator -----------------------------------------------------
//
// To properly handle drawing in grayscale with multiple monitors
// and differing color depths, we must draw while iterating over
// the GD handles. This iterator makes a for loop easy while also
// reducing the clipping area to the appropriate device and setting
// a boolean to indicate whether the device is "deep" or not. "Deep"
// meaning suitable for grayscale drawing as opposed to black & white.
//

class GDIterator
	{
	public:
	
		GDIterator();
		virtual ~GDIterator();
		
		void	Setup();			// called by constructor; can be called again
		Boolean	NextDevice(Boolean& deep);// returns true while you should draw again
		void	Cleanup();			// called by destructor; can be called earlier
		
		void	ClipFurtherToCurrentDevice();	// can be called for tricky clipping situations (see slider labels)

	protected:
	
		// To look good in gray we need 4 bits in monochrome, or 8 bits in color.
		// This is because 2 bits monochrome is only black, white, plus 2 shades;
		// and 4 bits in color is just that plus 12 irrelevant colors.
		enum { kDeepGrayscaleBits = 4, kDeepColorBits = 8 };
	
		GDHandle	FindNextValidDevice(GDHandle currentDevice, Boolean& deep);
		void		ClipFurtherToDevice(GDHandle aDevice);

		Boolean		mIteratedYet;
		GDHandle	mCurrentGDHandle;	// inited by Setup, changed by More

		static RgnHandle	mSavedClipping;		// saved by Setup, restored by Cleanup
	};

#pragma mark AGAGroupsContainer

//
// AGAGroupsContainer --------------------------------------------------------
//
// This class is used internally to maintain the MExclusiveObject
// behavior that allows radio buttons to automatically maintain
// standard radio button group XOR behavior.
// It just maintains an array of AGAGroupMember structs, each of
// which identifies the group ID and contains a pointer to a
// group member.
// If you're using the adapter classes, this is not used.
//

typedef struct
	{
	UInt32						mGroupIDPart1;
	UInt32						mGroupIDPart2;
	MExclusiveObject*			mGroupMember;
	} AGAGroupMember;

class AGAGroupsContainer

#ifdef AGA_SUPERCLASS
	: public AGA_SUPERCLASS
#endif // AGA_SUPERCLASS

	{
	public:
	
		AGAGroupsContainer();
		virtual ~AGAGroupsContainer();

		void	AGAAddGroupMember(MExclusiveObject* groupMember);
		void	AGARemoveGroupMember(MExclusiveObject* groupMember);
		void	AGAHitGroupMember(MExclusiveObject* groupMember);

	protected:
	
		void	AddGroupMember(MExclusiveObject* groupMember, Boolean allowExpand);
		void	ExpandArray();

		enum { kGroupArrayChunkSize = 16 };

		UInt32				mGroupsArraySize;
		AGAGroupMember**	mGroups;	// Handle to array of AGAGroupMember
	};

#pragma mark AGA_Globals

//
// Global functions and variables -------------------------------------------
//

//
// These are the gAGARamp RGB color value indexes.
// They match the numbers specified in the AGA doc.
// rW is white.
// 1 thru 12 go from light gray to dark gray, matching AGA doc indexes.
// rA1 and rA2 are the additional dark grays referenced in the AGA doc.
// rB is black.
// rP1 thru rP4 are the purple shades for the disclosure triangle.
//

typedef enum
	{
	rW = 0,
	r1 = 1,
	r2 = 2,
	r3 = 3,
	r4 = 4,
	r5 = 5,
	r6 = 6,
	r7 = 7,
	r8 = 8,
	r9 = 9,
	r10 = 10,
	r11 = 11,
	r12 = 12,
	rA1 = 13,
	rA2 = 13,
	rB = 15,
	rP1 = 16,
	rP2 = 17,
	rP3 = 18,
	rP4 = 19,
	
	kNumRampColors,
	
	OUT = -1
	} RampIndex;

extern RGBColor				gAGARamp[kNumRampColors];
extern Boolean				gAGAHasColorQD;
extern const Str255 		gGrayCouncilCopyright;
extern AGAGroupsContainer*	gGroupsContainer;
extern UInt32				gAGADefaults;
extern Boolean				gAGAStdControls;
extern Boolean				gAGAExtraControls;
extern Boolean				gAppearanceManagerSavvyInit;	// if true AGA is turned off in presence of AM
extern Boolean				gAaronKaleidoscopeSavvyInit;	// if true AGA is turned off in presence of Aaron
extern RGBColor				gFinder7EmptyProgressColor;
extern RGBColor				gFinder7FilledProgressColor;

// Turn default behavioral attributes off or on with these calls.
// These affect what default attributes are used for certain objects
// at the time they are instantiated, in the case where the attribute
// is not explicitly set as a constructor parameter or adapter class
// function.
extern void SetGrayCouncilDefault(UInt32 itemMask, Boolean turnOn);
extern Boolean TestGrayCouncilDefault(UInt32 itemMask);

// These are the attributes you can turn off or on with the above calls.
// They only affect new objects, not existing ones.
#define kAGALiveScrolling			0x00000001	// causes scroll bars to init with live scrolling ON
#define kAGAProportionalScrolling	0x00000002	// causes scroll bars to init with proportional thumb ON
#define kAGALiveRSliders			0x00000004	// causes rectangle sliders to init with live tracking ON
#define kAGAProportionalRSliders	0x00000008	// causes rectangle sliders to init with proportional thumb ON
#define kAGALivePSliders			0x00000010	// causes pointy sliders to init with live tracking ON
#define kAGAProportionalPSliders	0x00000020	// causes pointy sliders to init with proportional thumb ON
#define kAGAInverseHSliderValues	0x00000040	// causes horizontal sliders to init as min=right max=left
#define kAGAInverseVSliderValues	0x00000080	// causes vertical sliders to init as min=bottom max=top
#define kAGADimDeactivation			0x00000100	// causes adapters to dim AGAObject when deactivated (PowerPlant-only for now)

// AGAStringOut is used by the AGA classes to draw their text.
#define kNoTruncation -1	// Toolbox defines smTruncEnd, smTruncMiddle, etc.
enum AGAStringOutColor { kNormalOutput, kDisabledOutput, kInverseOutput, kUseCurrentColor };
extern void AGAStringOut(StringPtr aString,
						Rect* centeringFrame,
						SInt32 truncation,
						SInt32 justification,
						AGAStringOutColor outColor,
						Boolean deep,
						const AGATextStyle& textStyle);

// AGATextBox is a replacement for TETextBox, with the same initial parameters, plus
// a couple of extra ones for AGA control over the text. This is used by AGAStaticText.
extern void AGATextBox(const void *textPtr,
						SInt32 textLength,
						const Rect *textFrame,
						SInt16 justification,
						AGAStringOutColor outColor,
						Boolean deep,
						const AGATextStyle& textStyle,
						const RGBColor* teBackgroundColor);

// AGABackgroundPaint is used to paint window backgrounds.
enum AGABackgroundKind { kRaisedModelessBackground, kRaisedModalBackground, kFlatWindowBackground, kNoBackgroundPaint };
extern void AGABackgroundPaint(const Rect* backgroundBounds,
								Boolean drawFill,
								AGABackgroundKind backgroundKind,
								Boolean isActive,
								Boolean hasGrowBox);

// AGADrawSeparator is used to draw separator lines.
enum AGASeparatorLayout { kSeparatorAutomatic, kSeparatorLeft, kSeparatorTop, kSeparatorRight, kSeparatorBottom };
extern void AGADrawSeparator(const Rect* bounds, AGASeparatorLayout layout = kSeparatorAutomatic);

// AGADrawBevelButton is used to draw an empty bevel button, such as an icon button.
enum ButtonFrameType { kAutoFrame, kLargeFrame, kSmallFrame, kMiniFrame };
extern ButtonFrameType AGAGetDefaultFrameType(SInt16 buttonWidth);
extern void AGADrawBevelButton(Rect* bounds, ButtonFrameType frameType, Boolean isOn, Boolean isPressed, Boolean isEnabled);
extern void AGADrawBevelButtonAtDepth(Boolean deep, Rect* bounds, ButtonFrameType frameType, Boolean isOn, Boolean isPressed, Boolean isEnabled);

// AGAInstallGrayWCTB is used to install gray into the window's 'wctb' background.				
extern OSErr AGAInstallGrayWCTB(WindowPtr aWindow);
extern OSErr AGAInstallWCTB(WindowPtr aWindow, const RGBColor* newColor, Boolean* hasOldColor, RGBColor* oldColor);

// AGADrawBorderFrame is used to draw the "3D" sunken frame around an object.
// It supports correct dimming and "notched" shape (contoured to scroll bars).
extern void AGADrawBorderFrame(const Rect* bounds, Boolean isEnabled, Boolean isNotched, Boolean drawFrame, const RGBColor* windowBackgroundColor);

// AGAComputeTargetBorderRegion is used to build a region that defines the
// target border, aka "focus" frame.
// It supports "notched" shape (contoured to scroll bars).
extern void AGAComputeTargetBorderRegion(RgnHandle borderRegion, const Rect* bounds, Boolean isNotched);

// AGAClipFurther shrinks the clip region further from its current area
// to prevent drawing outside the specified rect.
extern void AGAClipFurther(Rect* clippingRect);

// AGA_PLstrcpy is a replacement for PLstrcpy so the plstringfuncs library is not needed.
// It's just a BlockMoveData call.
extern void AGA_PLstrcpy(StringPtr str1, ConstStr255Param str2);

// These background color determination routines are used by the adapter
// classes during instantiation of certain view types that must erase
// their background, and because tab panels vary the background color.
// See more comments where these are implemented.
enum { kInstallOnlyIfCustom = false, kInstallAlways = true };	// MacApp & PowerPlant need different behavior here
extern Boolean DetermineAGABackgroundColors(RGBColor* enabledColor, RGBColor* disabledColor, Boolean installIfNotCustom);
extern void InstallAGABackgroundColors(AGAObject* itsAGAObject, Boolean installIfNotCustom);
extern void TurnOnCustomAGABackgroundColors(RGBColor* enabledColor, RGBColor* disabledColor);
extern void TurnOffCustomAGABackgroundColors();

// These inlines replace PenPat(xxx) where x is the qd.gray or qd.black global, so
// that a ConstPatternParam is used instead of a global. This makes it possible to
// use the code from a code resource.
inline void PenPatBlack() { ::PenPat((ConstPatternParam)"\xff\xff\xff\xff\xff\xff\xff\xff"); }
inline void PenPatGray() { ::PenPat((ConstPatternParam)"\xAA\x55\xAA\x55\xAA\x55\xAA\x55"); }

extern SInt16 MinSInt16(SInt16 a, SInt16 b);
extern SInt16 MaxSInt16(SInt16 a, SInt16 b);
extern SInt32 MinSInt32(SInt32 a, SInt32 b);
extern SInt32 MaxSInt32(SInt32 a, SInt32 b);
extern void DelayFuture(SInt32 delayTicks);
extern void DelayFutureWhileStillDown(SInt32 delayTicks);
extern SInt32 RuntimeJustify(SInt32 userJust);

#endif // __GCCORE__
