//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// The Gray Council provides a set of standard C++ classes that implement
// the standard Apple Grayscale Appearance. The core classes do not
// require any other code such as a particular class framework.
// There are separate files provided that plug The Gray Council core
// into class frameworks.
//
// This file defines core classes for the new controls defined
// in the AGA spec that do not exist in the Control Manager.
//
// Classes defined below:
//	MIconButtonObject -- mixin class for icon buttons
//	AGAIconPushButton -- AGAPushButton subclass for icon push buttons
//	AGAIconCheckBox -- AGACheckBox subclass for on/off icon buttons
//	AGAIconRadioButton -- AGARadioButton subclass for mutually exclusive
//		on/off icon buttons
//	AGASlider -- slider with pointy or rectangular indicator, optional
//		labels, live tracking, proportional indicator capabilities
//	AGALittleArrows -- little up/down arrows with notification
//	AGADisclosureTriangle -- disclosure triangle with animation
//	AGAProgressIndicator -- determinate or indeterminate progress gauge
//		with moving determinate origin capability
//	AGASeparator -- separator line with automatic directionality
//	AGAGroupBox -- primary or secondary group box with optional title or
//		room for external title or control
//	AGATabPanel -- folder tabs for switching panels
//

#pragma once

#ifndef __GCNEWCONTROLS__
#define __GCNEWCONTROLS__

#include "GCCore.h"
#include "GCStdControls.h"

#pragma mark MIconButtonObject

//
// MIconButtonObject ---------------------------------------------------------------
//
// MIconButtonObject is a mixin class for the classes that have an
// icon button appearance. There are two different basic visual
// attributes for a button: the frame type and the image type.
//
// The frame type determines what the button edges look like. There
// are 3 different frame types specified by the AGA. The size of
// the button determines which frame type is used, but you can
// force a given type to be used.
// - 40x40 or larger: kLargeFrame (4 pixels shading on each side)
//		This is room for a 32x32 'iclx' plus the shading.
// - 22x22 or larger: kSmallFrame (3 pixels shading on each side)
//		This is room for a 16x16 'icsx' plus the shading.
// - 20x16 or larger: kMiniFrame (2 pixels shading on each side)
//		This is room for a 16x12 'icmx' plus the shading.
// The constant kAutoFrame is used by default. This sets the frame type
// automatically based on the actual button bounds. The dimensions shown
// above are the ideal dimensions shown in the AGA spec.
//
// The image type determines how the actual picture image inside the
// button is drawn. By default the image type is an icon family of size
// based on the button size (according to the frame sizes listed above),
// but you can explicitly set the type, including non-icon-family types.
// In all cases, the image is centered inside the button frame.
// - kAutoIconFamily: 32x32, 16x16, or 12x12 icon rect, derived from
//		the button bounds, according to the ideal sizes list above,
//		using an icon family.
// - kLargeIcon: 32x32 icon rect ('iclx' size) using an icon family.
// - kSmallIcon: 16x16 icon rect ('icsx' size) using an icon family.
// - kMiniIcon: 16x12 icon rect ('icmx' size) using an icon family.
// - kColorIcon: arbitrary icon rect derived from / using a 'cicn' resource.
// - kPictureIcon: arbitrary icon rect derived from / using a 'PICT' resource.
//
// An icon button has two image resource IDs. Usually these are the same,
// but for some button types you can make them different. For example, if
// you want different icons for on/off when the button is used as a radio
// button or check box, you can set the IDs to different values.
//
// Note that mini icons are a rather rare resource type, and ResEdit 2.x
// doesn't have an editor for them. I think Resorcerer does.
//

class MIconButtonObject
	{
	public:
	
		enum { kNoIconID = -1 };
		enum ButtonImageType { kAutoIconFamily, kLargeIcon, kSmallIcon, kMiniIcon, kColorIcon, kPictureIcon };

		MIconButtonObject();
		MIconButtonObject(SInt16 offIconID, SInt16 onIconID, ButtonFrameType frameType, ButtonImageType imageType);
		virtual ~MIconButtonObject();
		
		virtual void	SetIconIDs(SInt16 offIconID, SInt16 onIconID);
		virtual void	SetFrameType(ButtonFrameType frameType);
		virtual void	SetImageType(ButtonImageType imageType);
		
		virtual void	DrawIconButton(Rect* bounds,
										Boolean isOn,
										Boolean isPressed,
										Boolean isEnabled,
										Boolean deep);

		static ButtonImageType	GetDefaultImageType(SInt16 buttonWidth);

	protected:

		virtual void	DrawMiniEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled);
		virtual void	DrawSmallEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled);
		virtual void	DrawLargeEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled);

		virtual PicHandle	GetPictureImage(Point* imageSize, SInt16 pictureResourceID);
		virtual CIconHandle	GetColorIconImage(Point* imageSize, SInt16 cicnResourceID);

		SInt16			mOffIconID;
		SInt16			mOnIconID;
		ButtonFrameType	mFrameType;
		ButtonImageType	mImageType;
	};

#pragma mark AGAIconPushButton

//
// AGAIconPushButton ---------------------------------------------------------------
//
// AGAIconPushButton implements the standard AGA icon button, with
// pushbutton behavior; that is to say, it has no state. This is
// done by subclassing AGAPushButton for tracking, and mixing in
// MIconButtonObject for drawing.
//

class AGAIconPushButton : public AGAPushButton, public MIconButtonObject
	{
	public:

		AGAIconPushButton(Rect* bounds);
		AGAIconPushButton(Rect* bounds, SInt16 iconID, ButtonFrameType frameType, ButtonImageType imageType);
		virtual ~AGAIconPushButton();
		
	protected:

		virtual void	DrawButtonNormal(Boolean deep);
		virtual void	DrawButtonPressed(Boolean deep);
		virtual void	DrawButtonDisabled(Boolean deep);
	};

#pragma mark AGAIconCheckBox

//
// AGAIconCheckBox ---------------------------------------------------------------
//
// AGAIconCheckBox implements the standard AGA icon button, with
// check box behavior; that is to say, it has an on or off state
// that toggles when hit. This is done by subclassing AGACheckBox
// for tracking and state, and mixing in MIconButtonObject for drawing.
//

class AGAIconCheckBox : public AGACheckBox, public MIconButtonObject
	{
	public:

		AGAIconCheckBox(Rect* bounds, Boolean automaticState);
		AGAIconCheckBox(Rect* bounds, Boolean automaticState, SInt16 offIconID, SInt16 onIconID, ButtonFrameType frameType, ButtonImageType imageType);
		virtual ~AGAIconCheckBox();
		
	protected:

		virtual void	DrawButtonNormal(Boolean deep);
		virtual void	DrawButtonPressed(Boolean deep);
		virtual void	DrawButtonDisabled(Boolean deep);
	};

#pragma mark AGAIconRadioButton

//
// AGAIconRadioButton ---------------------------------------------------------------
//
// AGAIconRadioButton implements the standard AGA icon button, with
// radio button behavior; that is to say, it has an on or off state
// that can be automatically made mutually exclusive with other
// buttons in its group. This is done by subclassing AGARadioButton
// for tracking and state, and mixing in MIconButtonObject for drawing.
//

class AGAIconRadioButton : public AGARadioButton, public MIconButtonObject
	{
	public:

		AGAIconRadioButton(Rect* bounds, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState);
		AGAIconRadioButton(Rect* bounds, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState, SInt16 offIconID, SInt16 onIconID, ButtonFrameType frameType, ButtonImageType imageType);
		virtual ~AGAIconRadioButton();
		
	protected:

		virtual void	DrawButtonNormal(Boolean deep);
		virtual void	DrawButtonPressed(Boolean deep);
		virtual void	DrawButtonDisabled(Boolean deep);
	};

#pragma mark AGASlider

//
// AGASlider ---------------------------------------------------------------
//
// AGASlider implements the standard AGA slider. You specify whether it
// is horizontal or vertical, labeled or unlabeled, rectangular or pointy.
// By default, if you specify a labeled slider, it is pointy; otherwise it
// is rectangular. You can change this by calling SetSliderKind, and you
// can set the labels up after construction by calling the label functions.
// It also supports live indicator tracking and a proportional indicator
// as options.
//

class AGASlider : public AGATrackingIndicator
	{
	public:

		enum SliderKind { kRectSlider, kPointerSlider };
		enum { kInverseDirection = true, kNormalDirection = false };

		AGASlider(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, const AGATextStyle& textStyle, SInt16 labelsID);
		AGASlider(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, const AGATextStyle& textStyle, SInt16 labelsID, SliderKind kind, Boolean liveTracking, Boolean proportional);
		virtual ~AGASlider();
		
		virtual void	SetSliderKind(SliderKind kind);
		virtual void	SetLabelsFromResource(SInt16 stringListResourceID);
		virtual void	SetNumLabels(UInt32 numLabels);
		virtual void	SetLabel(UInt32 labelIndex, StringPtr itsLabel);
		virtual void	GetLabel(UInt32 labelIndex, StringPtr itsLabel);
		virtual void	SetLabelsStyle(const AGATextStyle& textStyle);
		virtual void	SetJustification(SInt32 justification);
		virtual void	SetInverseDirection(Boolean inverseDirection);
		virtual SInt32	GetValue();
		virtual void	SetValue(SInt32 newValue, Boolean redraw);

	protected:

		virtual SInt32	TrackTestPart(Point mouseLocation);
		virtual Boolean	IsValidIndicatorTrackMouse(Point mouseLocation);
		virtual SInt32	GetValueFromMouseDelta(SInt32 originalValue, Point originalMouseLocation, Point newMouseLocation);

		virtual void	DrawBackground();
		virtual void	DrawIndicatorTrack();
		virtual void	DrawIndicator();
		virtual void	RemoveIndicator(SInt32 newValue);
		virtual void	DrawGhost();
		virtual void	RemoveGhost(SInt32 newValue);
		
		enum { kRectIndicatorPixelInset = 6 };
		enum { kPointerIndicatorPixelInset = 7 };
		enum { kIndicatorWidth = 16 };
		enum { kEndGapSize = 5 };
		enum { kLabelTickEndInset = 11 };
		enum { kLabelTickOffset = 8 };
		enum { kLabelPosTextOffset = 14 };
		enum { kLabelNegTextOffset = 8 };
		enum { kLabelTickLength = 6 };
		enum { kGhostIndicator = true, kNormalIndicator = false };
		enum { kIncludeBackground = true, kTrackOnly = false };

		virtual void	GetTrackFrame(Rect* trackFrame, Boolean includeBackground);
		virtual void	GetIndicatorBox(SInt32 value, Rect* indicatorBox);
		virtual SInt32	GetIndicatorPixelRange(SInt32* proportionalIndicatorPixels);
		virtual void	GetIndicatorPixelEnds(SInt32* startPt, SInt32* stopPt);
		virtual void	GetIndicatorSpan(SInt32 value, SInt32* startPt, SInt32* valuePt, SInt32* stopPt);
		
		virtual void	DrawLabels();
		virtual void	DrawRectIndicator(Rect* indicatorRect, Boolean isGhost);
		virtual void	DrawPointerIndicator(Rect* indicatorRect, Boolean isGhost);
		virtual void	BuildPointerIndicator(Rect* indicatorRect);
		
		virtual void	DisposeLabels();
		
		Handle			mLabelStringHandles;	// an array of StringHandle, may be NULL
		SInt32			mJustification;
		SliderKind		mSliderKind;
		AGATextStyle	mTextStyle;
		Boolean			mInverseDirection;
		
		// These regions are used by RemoveIndicator and RemoveGhost.
		// They are tiny but we allocate them statically to avoid
		// repeated allocation, which could conceivably fail.
		static RgnHandle	mgSavedIndicatorClip;
		static RgnHandle	mgOldIndicatorClip;
		static RgnHandle	mgNewIndicatorClip;
	};

#pragma mark AGALittleArrows

//
// AGALittleArrows ---------------------------------------------------------------
//
// AGALittleArrows implements the standard AGA little arrows. To respond
// to continuous messages during tracking, you should either install a
// notification routine or subclass and override NotifyDelta. The supplied
// MacApp and PowerPlant classes do this for you.
//

class AGALittleArrows;

// This is the live tracking notification prototype for AGALittleArrows.
typedef void (*AGALittleArrowsNotifyPtr)(AGALittleArrows* theIndicator, SInt32 pressedPart, void* userData);

class AGALittleArrows : public AGAObject
	{
	public:

		enum { kDownArrowPressed = -1, kNoArrowPressed = 0, kUpArrowPressed = 1 };

		AGALittleArrows(Rect* bounds);
		virtual ~AGALittleArrows();
		
		virtual void	InstallNotificationRoutine(AGALittleArrowsNotifyPtr notificationRoutine, void* userData);
		
		virtual void	DrawObject();
		virtual Boolean	TrackMouse(Point mouseLocation);

		virtual void	DrawButton(SInt32 pressedPart);
		
	protected:

		enum { kLittleArrowWidth = 13, kLittleArrowHeight = 12, kLittleArrowTotalHeight = 23 };

		virtual Boolean	TrackPart(Point mouseLocation, SInt32 partHit);
		virtual SInt32	TrackTestPart(Point mouseLocation);

		virtual void	NotifyDelta(SInt32 partHit);
		
		AGALittleArrowsNotifyPtr	mNotificationRoutine;
		void*						mUserData;
	};

#pragma mark AGADisclosureTriangle

//
// AGADisclosureTriangle ---------------------------------------------------------------
//
// AGADisclosureTriangle implements the standard AGA disclosure triangle.
// You can think of it as a check box that is either off (closed) or on
// (disclosed).
//

class AGADisclosureTriangle : public AGAObject
	{
	public:

		enum { kDisclosedState = true, kClosedState = false };

		AGADisclosureTriangle(Rect* bounds, Boolean automaticState);
		virtual ~AGADisclosureTriangle();
		
		virtual void	DrawObject();
		virtual Boolean	TrackMouse(Point mouseLocation);
		
		virtual Boolean	GetState();
		virtual void	SetState(Boolean isDisclosed, Boolean redraw);
		virtual void	SetStateAnimate(Boolean isDisclosed);

	protected:

		virtual void	SetTrackingState(Boolean isIn);

		enum DTAnimationState { kDTClosed, kDTPressedClosed, kDTIntermediate, kDTPressedDisclosed, kDTDisclosed };

		virtual void	DrawTriangle(DTAnimationState state);
		virtual void	EraseBackground();

		Boolean		mAutomaticState;	// if true, triangle is toggled on click w/o caller needing to call SetValue
		Boolean		mIsDisclosed;
	};

#pragma mark AGAProgressIndicator

//
// AGAProgressIndicator ---------------------------------------------------------------
//
// AGAProgressIndicator implements the standard AGA progress indicator.
// It can be determinate (a moving gauge) or indeterminate (a rotating
// barber pole). In addition, it supports a "moving origin", which is
// where, in addition to moving the gauge end point, the gauge start
// point also moves.
//

class AGAProgressIndicator : public AGAObject
	{
	public:

		static Boolean mImagesAllocated;
		static OSErr AllocateProgressImages();
		static void DisposeProgressImages();

		AGAProgressIndicator(Rect* bounds, SInt32 minimum, SInt32 maximum);
		virtual ~AGAProgressIndicator();
		
		virtual void	DrawObject();
		
		virtual SInt32	GetValue();
		virtual void	SetValue(SInt32 newValue, Boolean redraw);
		virtual void	Increment(SInt32 delta, Boolean redraw);
		
		virtual void	Animate();			// moves indeterminate indicator by 1
		virtual void	QuantizedAnimate();	// moves it only if kAnimationDelay has elapsed
		
		virtual void	GetRange(SInt32* minimum, SInt32* maximum);
		virtual void	SetRange(SInt32 newMinimum, SInt32 newMaximum, Boolean redraw);
		
		virtual SInt32	GetOriginValue();
		virtual void	SetOriginValue(SInt32 newValue, Boolean redraw);

	protected:

		virtual void	DrawFrame();
		virtual void	DrawAnimationStep();
		virtual void	DrawGauge();
		virtual void	DrawOriginArea();
		virtual void	DrawProgressArea();
		virtual void	DrawRemainderArea();
		
		virtual Boolean	GetOriginFrame(Rect* originFrame);
		virtual Boolean	GetProgressFrame(Rect* progressFrame);
		virtual Boolean	GetRemainderFrame(Rect* remainderFrame);
		virtual void	GetRangeFrame(SInt32 startValue, SInt32 endValue, Rect* rangeFrame);

		enum { kProgressHeight = 14 };
		enum { kProgressImageHeight = 10, kProgressImageWidth = 16 };
		enum { k1BitPattern, kDeepPattern, kNumProgressPatterns };
		enum { kNumProgressAnimationSteps = 4 };	// 2, 4, 8, or 16 will work
		enum { kAnimationDelay = 5 };

		SInt32		mMinimum;
		SInt32		mMaximum;
		SInt32		mValue;
		SInt32		mOriginValue;		// usually == mMinimum
		UInt32		mAnimationIndex;
		SInt32		mLastAnimationTime;	// used for quantized animation

		static SInt8				kIndeterminateProgressImage[kNumProgressPatterns]
																[kProgressImageHeight]
																[kProgressImageWidth];

		static AGAOffscreenImage*	mgOffscreenIndeterminateProgressImage[kNumProgressPatterns];
	};

#pragma mark AGASeparator

//
// AGASeparator ---------------------------------------------------------------
//
// AGASeparator implements the standard AGA separator line.
// It automatically determines its orientation based on the coordinate
// bounding box dimensions.
//

class AGASeparator : public AGAObject
	{
	public:

		AGASeparator(Rect* bounds);
		virtual ~AGASeparator();
		
		virtual void	DrawObject();
	};

#pragma mark AGAGroupBox

//
// AGAGroupBox ---------------------------------------------------------------
//
// AGAGroupBox implements the standard AGA group box. A group box can
// either be primary or secondary. Secondary group boxes should be used
// inside primary group boxes, sez the AGA spec. You can supply a title
// string for the group box, or you can specify a "gap" width that is to
// be left blank to make room for a control or title string that will
// be drawn elsewhere.
//

class AGAGroupBox : public AGAObject
	{
	public:

		enum { kPrimaryGroupBox = true, kSecondaryGroupBox = false };

		AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType);
		AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType, SInt16 titleGap);
		AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType, StringPtr titleString);
		AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType, SInt16 titleStringListResourceID, SInt16 titleStringIndex);
		virtual ~AGAGroupBox();
		
		virtual void	SetTitleGap(SInt16 titleGap);
		virtual void	SetTitle(StringPtr newTitle, Boolean redraw);
		
		virtual void	DrawObject();

	protected:

		virtual void	DrawPrimaryFrame(SInt16 topIndent, SInt16 gapSize);
		virtual void	DrawSecondaryFrame(SInt16 topIndent, SInt16 gapSize);

		Boolean			mIsPrimaryBox;
		SInt16			mTitleGap;	// horizontal gap for other object in title, no top indent
		Str255			mTitle;		// zero-length string indicates no title, no top indent
		AGATextStyle	mTextStyle;
	};

#pragma mark AGATabPanel

//
// AGATabPanel ---------------------------------------------------------------
//
// AGATabPanel implements an AGA tab panels control.
//
// You can specify the location of the tabs (top or bottom, left or right).
// (Only "top" tabs are currently implemented!!! Other values do NOT work yet.)
// You can specify whether the whole tab panel frame is drawn, or whether just
// the tabs themselves are drawn, which allows you to attach tabs to something
// that you are drawing yourself. The simple constructor decides this based on
// the object bounds: if it's too small, there's no frame.
//

typedef struct
	{
	StringHandle	mLabel;
	SInt32			mTabWidth;
	Boolean			mDisabled;
	void*			mUserData;	// used by adapter classes for pane/view pointer
	} TabInfo, *TabInfoPtr, **TabInfoHandle;

class AGATabPanel : public AGAObject
	{
	public:

		enum TabSize { kSmallTabs, kLargeTabs };
		enum TabsLocation { kTopTabs, kBottomTabs, kLeftTabs, kRightTabs };
		enum TabWidthKind { kVariableTabWidth, kCommonTabWidth };	// AGA spec is kVariableTabWidth
		enum { kDontShrink = false, kShrinkToFit = true };			// for SetNumTabs shrinkToFit

		static Boolean mImagesAllocated;
		static OSErr AllocateTabImages();
		static void DisposeTabImages();

		AGATabPanel(Rect* bounds, SInt32 initialTabIndex, const AGATextStyle& textStyle, SInt16 labelsID);
		AGATabPanel(Rect* bounds, SInt32 initialTabIndex, const AGATextStyle& textStyle, SInt16 labelsID, TabWidthKind widthKind, TabsLocation location, Boolean drawFrame);
		virtual ~AGATabPanel();
		
		virtual void	DrawObject();
		virtual Boolean	TrackMouse(Point mouseLocation);
		virtual void	SetObjectBounds(Rect* bounds, Boolean redraw);
		
		virtual SInt32	GetCurrentTab();
		virtual void	SetCurrentTab(SInt32 tabIndex, Boolean redraw);
		virtual SInt32	GetLastTabClicked();
		
		virtual void	SetTabSize(TabSize size);							// specify large or small tabs
		virtual void	SetTabsLocation(TabsLocation location);				// specify tabs location
		virtual void	SetTabsWidth(TabWidthKind widthKind);				// specify tab vs. label width
		virtual SInt16	GetTabHeight();										// get "height" of the tabs
		virtual Boolean	HasFrame();											// true if we will draw a frame
		virtual void	SetLabelsFromResource(SInt16 stringListResourceID);	// creates 1 tab per string in the 'STR#'
		virtual void	SetNumTabs(SInt32 numLabels, Boolean shrinkToFit);	// ensures space for n tabs, optionally removes excess
		virtual SInt32	GetNumTabs();										// get number of tabs
		virtual void	RemoveTab(SInt32 tabIndex);							// removes specified tab from list
		virtual void	SetTabLabel(SInt32 tabIndex, StringPtr itsLabel);	// set one tab's label
		virtual void	GetTabLabel(SInt32 tabIndex, StringPtr itsLabel);	// get one tab's label
		virtual void	SetTabEnable(SInt32 tabIndex, Boolean enable);		// enable/disable one tab
		virtual Boolean	IsTabEnabled(SInt32 tabIndex);						// get enable/disable state of one tab
		virtual void	SetTabUserData(SInt32 tabIndex, void* userData);	// set extra data value for one tab; used by adapter classes for panel ptr
		virtual void*	GetTabUserData(SInt32 tabIndex);					// get extra data value for one tab; used by adapter classes for panel ptr
		virtual void	CalculateTabWidths();								// recalc all tab widths using current labels
		virtual void	SetLabelsStyle(const AGATextStyle& textStyle);		// set the text style for the labels

	protected:

		enum { kSmallTabImageHeight = 19, kSmallTabImageWidth = 12 };
		enum { kLargeTabImageHeight = 24, kLargeTabImageWidth = 12 };
		enum { kPanelEndMargin = 6 };

		// AGAObject mouse tracking overrides.
		virtual Boolean	HitTest(Point mouseLocation);
		virtual void	SetTrackingState(Boolean isIn);

		virtual void	DrawBackground();
		virtual void	DrawFrame();
		virtual void	DrawTabs();
		
		virtual void	DrawFrameBox();
		virtual void	DrawFrameTop();
		virtual void	DrawOneTab(SInt32 tabIndex, Boolean pressed);
		virtual void	GetTabBounds(SInt32 tabIndex, Rect* tabBounds);
		virtual void	CalculateTabWidthsByKind(TabWidthKind widthKind);
		
		virtual void	CalculateTabSizeInfo();
		
		virtual void	AllocateTabInfo(SInt32 numTabs, Boolean shrinkToFit);
		virtual void	DisposeTabInfo();
		
		virtual void	CopyTabImage(UInt32 imageIndex, Point imageDestination, Boolean deep);
		
		SInt32			mNumTabs;
		SInt32			mCurrentTabIndex;
		SInt32			mLastTabClicked;

		TabInfoHandle	mTabInfo;				// handle to array of TabInfo
		TabSize			mTabSize;				// small or large
		TabsLocation	mTabsLocation;			// top, bottom, left, or right
		TabWidthKind	mTabWidthKind;			// variable or same
		Point			mTabImageSize;			// depends on mTabSize & mTabsLocation
		Boolean			mDrawFrame;
		AGATextStyle	mTextStyle;

		// Drawing the tab contours manually is too tedious.
		// So we use offscreen PixMap structures and blast them onto the screen.
		// Current implementation doesn't use clip, so must be on an r2 background.

		enum { kSmallTabImages, kLargeTabImages, kNumImageTypes };
		enum { kTabNormalDeselectedL, kTabNormalDeselectedR,
				kTabNormalSelectedL, kTabNormalSelectedR,
				kTabPressedL, kTabPressedR,
				kTabDisabledDeselectedL, kTabDisabledDeselectedR,
				kTabDisabledSelectedL, kTabDisabledSelectedR,
				kNumTabImages };
		
		static SInt8	kSmallTabImageValues [kNumTabImages] [kSmallTabImageHeight] [kSmallTabImageWidth];
		static SInt8	kLargeTabImageValues [kNumTabImages] [kLargeTabImageHeight] [kLargeTabImageWidth];

		static SInt8	k1BitSmallTabImageValues [kNumTabImages] [kSmallTabImageHeight] [kSmallTabImageWidth];
		static SInt8	k1BitLargeTabImageValues [kNumTabImages] [kLargeTabImageHeight] [kLargeTabImageWidth];

		static AGAOffscreenImage*	mgOffscreenTabImages [kNumTabImages] [kNumImageTypes];
		static AGAOffscreenImage*	mg1BitOffscreenTabImages [kNumTabImages] [kNumImageTypes];
	};

#endif // __GCNEWCONTROLS__

