//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// See the header file "GCStdControls.h" for usage.
//

#include "GCStdControls.h"

#include <LowMem.h>

//
// AGAStaticText ---------------------------------------------------------------
//
// This class draws a static text label.
//

AGAStaticText::AGAStaticText(Rect* bounds, const AGATextStyle& textStyle, SInt32 justification)
: AGAObject(bounds)
	{
	// Construct with empty string.

	mJustification = justification;
	mTitle[0];
	mTextStyle = textStyle;
	}

AGAStaticText::AGAStaticText(Rect* bounds, const AGATextStyle& textStyle, SInt32 justification, StringPtr title)
: AGAObject(bounds)
	{
	// Construct with supplied string.

	mJustification = justification;
	AGA_PLstrcpy(mTitle, title);
	mTextStyle = textStyle;
	}

AGAStaticText::AGAStaticText(Rect* bounds, const AGATextStyle& textStyle, SInt32 justification, SInt16 stringListResourceID, SInt16 stringIndex)
: AGAObject(bounds)
	{
	// Construct by getting string from specified resource.

	mJustification = justification;
	::GetIndString(mTitle, stringListResourceID, stringIndex);
	mTextStyle = textStyle;
	}

AGAStaticText::~AGAStaticText()
	{
	}

void AGAStaticText::DrawObject()
	{
	AGADrawingEnvironment	env;

	// Paint over possible obsolete text, draw current text.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			this->ApplyBackgroundColor();
		else
			::RGBForeColor(&gAGARamp[rW]);

		::PaintRect(&mBounds);

		RGBColor	teBackgroundColor;
		::GetForeColor(&teBackgroundColor);	// fg color right now is what the TE must leave in the bg

		AGATextBox(&mTitle[1], mTitle[0], &mBounds, mJustification, mEnabled ? kNormalOutput : kDisabledOutput, deep, mTextStyle, &teBackgroundColor);
		}
	}

void AGAStaticText::SetTitle(StringPtr title, Boolean redraw)
	{
	// Change text, redraw if specified.

	AGA_PLstrcpy(mTitle, title);
	
	if (redraw)
		this->DrawObject();
	}

void AGAStaticText::SetTextStyle(const AGATextStyle& textStyle, Boolean redraw)
	{
	// Change text, redraw if specified.

	mTextStyle = textStyle;
	
	if (redraw)
		this->DrawObject();
	}

void AGAStaticText::SetJustification(SInt32 justification, Boolean redraw)
	{
	// Change text, redraw if specified.

	mJustification = justification;
	
	if (redraw)
		this->DrawObject();
	}

//
// AGAPushButton ---------------------------------------------------------------
//
// This class implements a standard text pushbutton object.
//

AGAPushButton::AGAPushButton(Rect* bounds, const AGATextStyle& textStyle)
: AGAObject(bounds)
	{
	// Construct with initial empty button title.

	mIsDefault = false;
	mFrameInside = true;
	mTitle[0] = 0;
	mTextStyle = textStyle;
	}

AGAPushButton::AGAPushButton(Rect* bounds, const AGATextStyle& textStyle, StringPtr title)
: AGAObject(bounds)
	{
	// Construct with specified button title.

	mIsDefault = false;
	mFrameInside = true;
	AGA_PLstrcpy(mTitle, title);
	mTextStyle = textStyle;
	}

AGAPushButton::AGAPushButton(Rect* bounds, const AGATextStyle& textStyle, SInt16 stringListResourceID, SInt16 stringIndex)
: AGAObject(bounds)
	{
	// Construct by getting button title from specified resource.

	mIsDefault = false;
	mFrameInside = true;
	::GetIndString(mTitle, stringListResourceID, stringIndex);
	mTextStyle = textStyle;
	}

AGAPushButton::~AGAPushButton()
	{
	}

void AGAPushButton::DrawObject()
	{
	// Draw button in normal unpressed state.

	this->DrawButton(kNotPressed);
	}

void AGAPushButton::SetTitle(StringPtr title, Boolean redraw)
	{
	// Change button title, redraw if specified.

	AGA_PLstrcpy(mTitle, title);
	
	if (redraw)
		this->DrawObject();
	}

void AGAPushButton::SetDefault(Boolean isDefault, Boolean frameInside)
	{
	// Set default state. If mIsDefault is true, the button will
	// be drawn with a default button outline. If mFrameInside
	// is true, the outline will be inside mBounds and the button
	// will be drawn inset from there; if mFrameInside is false,
	// the outline will be outside mBounds and the button will
	// be drawn at mBounds.

	mIsDefault = isDefault;
	mFrameInside = frameInside;
	}

Boolean AGAPushButton::IsDefault()
	{
	// Return true if this button has a default outline.

	return mIsDefault;
	}

void AGAPushButton::DrawButton(Boolean pressed)
	{
	AGADrawingEnvironment	env;

	// Draw the button in the appropriate state.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (! mEnabled || ! mActive)
			this->DrawButtonDisabled(deep);
		else if (pressed)
			this->DrawButtonPressed(deep);
		else
			this->DrawButtonNormal(deep);
		}
	}

Boolean AGAPushButton::HitTest(Point mouseLocation)
	{
	// Return true if the mouse is in the button.
	// Account for case where default button outline
	// is inside mBounds.

	Rect	r = mBounds;
	
	if (mIsDefault && mFrameInside)
		::InsetRect(&r, 3, 3);

	return ::PtInRect(mouseLocation, &r);
	}

void AGAPushButton::SetTrackingState(Boolean isIn)
	{
	// Draw the button pressed or unpressed as specified.

	this->DrawButton(isIn);
	}

void AGAPushButton::DrawButtonNormal(Boolean deep)
	{
	// Draw the button in enabled/unpressed state.

	Rect	r = mBounds;
	
	if (mIsDefault && mFrameInside)
		::InsetRect(&r, 3, 3);
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r2]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(r.right - r.left - 6, 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, r.bottom - r.top - 6);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(- (r.right - r.left - 6), 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, -(r.bottom - r.top - 6));
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		
		// Add 1st inner pixels.
		::MoveTo(r.left + 1, r.top + 2);
		::RGBForeColor(&gAGARamp[r4]);
		::Line(0, 0);
		::Move(1, -1);
		::Line(0, 0);
		::Move(r.right - r.left - 5, 0);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r5]);
		::Line(0, 0);
		::Move(0, 1);
		::RGBForeColor(&gAGARamp[r8]);
		::Line(0, r.bottom - r.top - 6);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(- (r.right - r.left - 6), 0);
		::RGBForeColor(&gAGARamp[r4]);
		::Move(-1, 0);
		::Line(-1, -1);

		// Add 2nd inner pixels.
		::MoveTo(r.left + 2, r.bottom - 4);
		::RGBForeColor(&gAGARamp[rW]);
		::Line(0, -(r.bottom - r.top - 6));
		::Line(1, 1);
		::Line(0, -1);
		::Line(r.right - r.left - 7, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r5]);
		::Line(0, r.bottom - r.top - 7);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(- (r.right - r.left - 7), 0);
		}
	else	// 1-bit
		{
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRoundRect(&r, 7, 7);
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRoundRect(&r, 7, 7);
		}

	// Draw the title.
	AGAStringOut(mTitle, &r, truncMiddle, teJustCenter, kNormalOutput, deep, mTextStyle);
	
	// Draw the default button outline.
	this->DrawOutlineNormal(deep);
	}

void AGAPushButton::DrawButtonPressed(Boolean deep)
	{
	// Draw the button in enabled/pressed state.

	Rect	r = mBounds;
	
	if (mIsDefault && mFrameInside)
		::InsetRect(&r, 3, 3);
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r9]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(r.right - r.left - 6, 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, r.bottom - r.top - 6);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(- (r.right - r.left - 6), 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, -(r.bottom - r.top - 6));
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		
		// Add 1st inner pixels.
		::MoveTo(r.left + 1, r.bottom - 3);
		::RGBForeColor(&gAGARamp[r11]);
		::Line(0, -(r.bottom - r.top - 5));
		::Line(1, 0);
		::Line(0, -1);
		::Line(r.right - r.left - 5, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r8]);
		::Line(0, 0);
		::Move(0, 1);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(0, r.bottom - r.top - 6);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(- (r.right - r.left - 5), 0);
		::RGBForeColor(&gAGARamp[r8]);
		::Line(0, 0);

		// Add 2nd inner pixels.
		::MoveTo(r.left + 2, r.bottom - 4);
		::RGBForeColor(&gAGARamp[r10]);
		::Line(0, -(r.bottom - r.top - 7));
		::Line(1, 0);
		::Line(0, -1);
		::Line(r.right - r.left - 7, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r8]);
		::Line(0, r.bottom - r.top - 7);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(- (r.right - r.left - 7), 0);
		}
	else	// 1-bit
		{
		::RGBForeColor(&gAGARamp[rB]);
		::PaintRoundRect(&r, 7, 7);
		}

	// Draw the title.
	AGAStringOut(mTitle, &r, truncMiddle, teJustCenter, kInverseOutput, deep, mTextStyle);
	
	// Draw the default button outline.
	this->DrawOutlineNormal(deep);
	}

void AGAPushButton::DrawButtonDisabled(Boolean deep)
	{
	// Draw the button in disabled/unpressed state.

	Rect	r = mBounds;
	
	if (mIsDefault && mFrameInside)
		::InsetRect(&r, 3, 3);
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r2]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 1);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(0, 0);
		::Move(1, -1);
		::Line(0, 0);
		::Move(1, 0);
		::Line(r.right - r.left - 6, 0);
		::Line(0, 0);
		::Move(1, 1);
		::Line(0, 0);
		::Move(1, 1);
		::Line(0, 0);
		::Move(0, 1);
		::Line(0, r.bottom - r.top - 6);
		::Line(0, 0);
		::Move(-1, 1);
		::Line(0, 0);
		::Move(-1, 1);
		::Line(0, 0);
		::Move(-1, 0);
		::Line(- (r.right - r.left - 6), 0);
		::Line(0, 0);
		::Move(-1, -1);
		::Line(0, 0);
		::Move(-1, -1);
		::Line(0, 0);
		::Move(0, -1);
		::Line(0, -(r.bottom - r.top - 6));
		::Line(0, 0);
		}
	else	// 1-bit
		{
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRoundRect(&r, 7, 7);
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRoundRect(&r, 7, 7);
		}

	// Draw the title.
	AGAStringOut(mTitle, &r, truncMiddle, teJustCenter, kDisabledOutput, deep, mTextStyle);
	
	// Draw the default button outline.
	this->DrawOutlineDisabled(deep);
	
	if (! deep)
		{
		if (mIsDefault)
			::InsetRect(&r, -3, -3);

		::RGBForeColor(&gAGARamp[rB]);
		PenPatGray();
		::PenMode(patBic);
		::PaintRect(&r);

		::PenNormal();
		}
	}

void AGAPushButton::DrawOutlineNormal(Boolean deep)
	{
	// Draw the default button outline in normal
	// state if the button has one. It may be either
	// inside or outside mBounds.
	
	// If we are drawing the outline outside the normal
	// bounds, we should force the clipping to include
	// that outside area, because the framework or whoever
	// may have reduced clipping to our bounds.

	if (! mIsDefault)
		return;

	Rect	r = mBounds;
	
	if (! mFrameInside)
		::InsetRect(&r, -3, -3);

	if (deep)
		{
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 2);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(1, -1);
		::Move(1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(r.right - r.left - 8, 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(1, 1);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, r.bottom - r.top - 8);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(-1, 1);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(- (r.right - r.left - 8), 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(-1, -1);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, -(r.bottom - r.top - 8));
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		
		// Draw the interior.
		::MoveTo(r.left + 1, r.bottom - 4);
		::RGBForeColor(&gAGARamp[r2]);
		::Line(0, -(r.bottom - r.top - 7));
		::Line(1, 0);
		::Line(0, -1);
		::Line(1, 0);
		::Line(0, -1);
		::Line(r.right - r.left - 8, 0);
		::Move(1, 0);
		::RGBForeColor(&gAGARamp[r3]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r4]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(0, 1);
		::RGBForeColor(&gAGARamp[r8]);
		::Line(0, r.bottom - r.top - 8);
		::Line(-2, 2);
		::Line(-(r.right - r.left - 8), 0);
		::Move(-1, 0);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[r4]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[r3]);
		::Line(0, 0);
		::Move(1, 0);
		::RGBForeColor(&gAGARamp[r5]);
		::Line(0, -(r.bottom - r.top - 8));
		::Line(2, -2);
		::Line(r.right - r.left - 8, 0);
		::Line(0, 1);
		::Line(1, 0);
		::Line(0, r.bottom - r.top - 8);
		::Line(-2, 2);
		::Line(-(r.right - r.left - 8), 0);
		::Line(0, -1);
		::Move(0, -1);
		::RGBForeColor(&gAGARamp[r8]);
		::Line(1, 1);
		::MoveTo(r.left + 3, r.top + 4);
		::Line(1, -1);
		::MoveTo(r.right -5, r.top + 3);
		::Line(1, 1);
		::MoveTo(r.right -5, r.bottom - 4);
		::Line(1, -1);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(-1, 1);
		}
	else
		{
		::RGBForeColor(&gAGARamp[rB]);
		::PenSize(2, 2);
		::FrameRoundRect(&r, 11, 11);
		::PenSize(1, 1);
		}
	}

void AGAPushButton::DrawOutlineDisabled(Boolean deep)
	{
	// Draw the default button outline in disabled
	// state if the button has one. It may be either
	// inside or outside mBounds.

	if (! mIsDefault)
		return;

	Rect	r = mBounds;
	
	if (! mFrameInside)
		::InsetRect(&r, -3, -3);

	if (deep)
		{
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 2);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(1, -1);
		::Move(1, -1);
		::Line(0, 0);
		::Move(1, 0);
		::Line(r.right - r.left - 8, 0);
		::Line(0, 0);
		::Move(1, 1);
		::Line(1, 1);
		::Move(1, 1);
		::Line(0, 0);
		::Move(0, 1);
		::Line(0, r.bottom - r.top - 8);
		::Line(0, 0);
		::Move(-1, 1);
		::Line(-1, 1);
		::Move(-1, 1);
		::Line(0, 0);
		::Move(-1, 0);
		::Line(- (r.right - r.left - 8), 0);
		::Line(0, 0);
		::Move(-1, -1);
		::Line(-1, -1);
		::Move(-1, -1);
		::Line(0, 0);
		::Move(0, -1);
		::Line(0, -(r.bottom - r.top - 8));
		::Line(0, 0);
		
		// Draw the interior.
		::MoveTo(r.left + 1, r.bottom - 4);
		::RGBForeColor(&gAGARamp[r4]);
		::Line(0, -(r.bottom - r.top - 7));
		::Line(1, 0);
		::Line(0, -1);
		::Line(1, 0);
		::Line(0, -1);
		::Line(r.right - r.left - 8, 0);
		::Move(3, 2);
		::Line(0, r.bottom - r.top - 7);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(-(r.right - r.left - 7), 0);
		::Move(-2, -2);
		::Line(1, 1);
		::Line(0, -(r.bottom - r.top - 7));
		::Line(1, 0);
		::Line(0, -1);
		::Line(1, 0);
		::Line(0, -1);
		::Line(r.right - r.left - 7, 0);
		::Line(-1, -1);
		::Move(-1, 2);
		::Line(2, 0);
		::Move(-1, 1);
		::Line(1, 0);
		::Line(0, r.bottom - r.top - 9);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(-1, 0);
		::Line(0, 1);
		::Line(-(r.right - r.left - 8), 0);
		::Line(0, -2);
		::Line(1, 1);
		}
	else
		{
		::RGBForeColor(&gAGARamp[rB]);
		::PenSize(2, 2);
		::FrameRoundRect(&r, 11, 11);
		::PenSize(1, 1);
		}
	}

//
// AGACheckBox ---------------------------------------------------------------
//
// This class implements a standard checkbox object with the additional
// capability of supporting mixed state.
//

AGACheckBox::AGACheckBox(Rect* bounds, const AGATextStyle& textStyle, Boolean automaticState)
: AGAObject(bounds)
	{
	// Construct with initial empty button title.

	mValue = kCheckBoxOff;
	mAutomaticState = automaticState;
	mTitle[0];
	mTextStyle = textStyle;
	}

AGACheckBox::AGACheckBox(Rect* bounds, const AGATextStyle& textStyle, StringPtr title, Boolean automaticState)
: AGAObject(bounds)
	{
	// Construct with specified button title.

	mValue = kCheckBoxOff;
	mAutomaticState = automaticState;
	AGA_PLstrcpy(mTitle, title);
	mTextStyle = textStyle;
	}

AGACheckBox::AGACheckBox(Rect* bounds, const AGATextStyle& textStyle, SInt16 stringListResourceID, SInt16 stringIndex, Boolean automaticState)
: AGAObject(bounds)
	{
	// Construct by getting button title from specified resource.

	mValue = kCheckBoxOff;
	mAutomaticState = automaticState;
	::GetIndString(mTitle, stringListResourceID, stringIndex);
	mTextStyle = textStyle;
	}

AGACheckBox::~AGACheckBox()
	{
	}

void AGACheckBox::DrawObject()
	{
	// Draw button in normal unpressed state.

	this->DrawButton(kNotPressed);
	}

Boolean AGACheckBox::TrackMouse(Point mouseLocation)
	{
	// If inherited tracking succeeded and we are set to
	// automatically change our state, toggle it.

	Boolean	wasItHit = AGAObject::TrackMouse(mouseLocation);

	if (wasItHit && mAutomaticState)
		this->SetValue((mValue == kCheckBoxOn) ? kCheckBoxOff : kCheckBoxOn, kRedraw);

	return wasItHit;
	}

void AGACheckBox::SetTitle(StringPtr title, Boolean redraw)
	{
	// Change button title, redraw if specified.

	AGA_PLstrcpy(mTitle, title);
	
	if (redraw)
		this->DrawObject();
	}

SInt32 AGACheckBox::GetValue()
	{
	// Return current state.

	return mValue;
	}

void AGACheckBox::SetValue(SInt32 newValue, Boolean redraw)
	{
	// Set current state, redraw if specified and necessary.

	if (newValue != mValue)
		{
		mValue = newValue;
		
		if (redraw)
			{
			this->DrawButton(kNotPressed);
			}
		}
	}

void AGACheckBox::DrawButton(Boolean pressed)
	{
	AGADrawingEnvironment	env;

	// Draw the button in the appropriate state.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (! mEnabled || ! mActive)
			this->DrawButtonDisabled(deep);
		else if (pressed)
			this->DrawButtonPressed(deep);
		else
			this->DrawButtonNormal(deep);
		}
	}

void AGACheckBox::SetTrackingState(Boolean isIn)
	{
	// Draw the button pressed or unpressed as specified.

	this->DrawButton(isIn);
	}

void AGACheckBox::DrawButtonNormal(Boolean deep)
	{
	// Draw the button in enabled/unpressed state.

	Rect	r = mBounds;
	
	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.bottom = r.top + kCheckBoxSize;
	r.right = r.left + kCheckBoxSize;
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r2]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRect(&r);
		
		// Draw the edges.
		::MoveTo(r.left + 1, r.bottom - 3);
		::RGBForeColor(&gAGARamp[rW]);
		::Line(0, -8);
		::Line(8, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(0, 8);
		::Line(-8, 0);
		
		if (mValue == kCheckBoxOn)
			{
			// Draw the X and its shadow.
			::MoveTo(r.left + 3, r.top + 2);
			::RGBForeColor(&gAGARamp[rB]);
			::Line(5, 5);
			::Move(0, 1);
			::Line(-5, -5);
			::Move(0, 4);
			::Line(5, -5);
			::Move(0, 1);
			::Line(-5, 5);
			::Move(1, 0);
			::RGBForeColor(&gAGARamp[r8]);
			::Line(1, -1);
			::Move(2, -2);
			::Line(2, -2);
			::Move(0, 1);
			::RGBForeColor(&gAGARamp[r5]);
			::Line(-1, 1);
			::Move(-2, 2);
			::Line(-2, 2);
			::Move(5, 0);
			::Line(0, 0);
			::Move(0, -1);
			::RGBForeColor(&gAGARamp[r8]);
			::Line(0, 0);
			}
		else if (mValue == kCheckBoxMixed)
			{
			// Draw the dash and its shadow.
			::MoveTo(r.left + 3, r.top + 5);
			::PenSize(2, 2);
			::RGBForeColor(&gAGARamp[rB]);
			::Line(4, 0);
			
			::MoveTo(r.left + 4, r.top + 7);
			::PenSize(1, 1);
			::RGBForeColor(&gAGARamp[r5]);
			::Line(5, 0);
			::Line(-2, 0);
			}
		}
	else
		{
		// Paint white over previous junk, if any.
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRect(&r);
		// Paint frame.
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRect(&r);

		// Draw check or dash.
		if (mValue == kCheckBoxOn)
			{
			::PenSize(1, 1);
			::MoveTo(r.left, r.top);
			::Line(kCheckBoxSize-1, kCheckBoxSize-1);
			::Move(-(kCheckBoxSize-1), 0);
			::Line(kCheckBoxSize-1, -(kCheckBoxSize-1));
			}
		else if (mValue == kCheckBoxMixed)
			{
			::PenSize(1, 2);
			::MoveTo(r.left + 3, r.top + 5);
			::Line(5, 0);
			}
		
		::PenNormal();
		}

	// Draw the title.
	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.left += (kCheckBoxSize + 4);
	r.bottom = r.top + kCheckBoxSize;
	AGAStringOut(mTitle, &r, truncMiddle, teJustLeft, kNormalOutput, deep, mTextStyle);
	}

void AGACheckBox::DrawButtonPressed(Boolean deep)
	{
	// Draw the button in enabled/pressed state.

	Rect	r = mBounds;
	
	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.bottom = r.top + kCheckBoxSize;
	r.right = r.left + kCheckBoxSize;
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r8]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRect(&r);
		
		// Draw the edges.
		::MoveTo(r.left + 1, r.bottom - 3);
		::RGBForeColor(&gAGARamp[r10]);
		::Line(0, -8);
		::Line(8, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r6]);
		::Line(0, 8);
		::Line(-8, 0);
		
		if (mValue == kCheckBoxOn)
			{
			// Draw the X and its shadow.
			::MoveTo(r.left + 3, r.top + 2);
			::RGBForeColor(&gAGARamp[rB]);
			::Line(5, 5);
			::Move(0, 1);
			::Line(-5, -5);
			::Move(0, 4);
			::Line(5, -5);
			::Move(0, 1);
			::Line(-5, 5);
			::Move(1, 0);
			::RGBForeColor(&gAGARamp[r11]);
			::Line(1, -1);
			::Move(2, -2);
			::Line(2, -2);
			::Move(0, 1);
			::RGBForeColor(&gAGARamp[r10]);
			::Line(-1, 1);
			::Move(-2, 2);
			::Line(-2, 2);
			::Move(5, 0);
			::Line(0, 0);
			::Move(0, -1);
			::RGBForeColor(&gAGARamp[r11]);
			::Line(0, 0);
			}
		else if (mValue == kCheckBoxMixed)
			{
			// Draw the dash and its shadow.
			::MoveTo(r.left + 3, r.top + 5);
			::PenSize(2, 2);
			::RGBForeColor(&gAGARamp[rB]);
			::Line(4, 0);
			
			::MoveTo(r.left + 4, r.top + 7);
			::PenSize(1, 1);
			::RGBForeColor(&gAGARamp[r10]);
			::Line(5, 0);
			::Line(-2, 0);
			}
		}
	else
		{
		// Paint white over previous junk, if any.
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRect(&r);
		// Paint frame.
		::RGBForeColor(&gAGARamp[rB]);
		::PenSize(2, 2);
		::FrameRect(&r);

		// Draw check or dash.
		if (mValue == kCheckBoxOn)
			{
			::PenSize(1, 1);
			::MoveTo(r.left, r.top);
			::Line(kCheckBoxSize-1, kCheckBoxSize-1);
			::Move(-(kCheckBoxSize-1), 0);
			::Line(kCheckBoxSize-1, -(kCheckBoxSize-1));
			}
		else if (mValue == kCheckBoxMixed)
			{
			::PenSize(1, 2);
			::MoveTo(r.left + 3, r.top + 5);
			::Line(5, 0);
			}
		
		::PenNormal();
		}

	// Draw the title.
	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.left += (kCheckBoxSize + 4);
	r.bottom = r.top + kCheckBoxSize;
	AGAStringOut(mTitle, &r, truncMiddle, teJustLeft, kNormalOutput, deep, mTextStyle);
	}

void AGACheckBox::DrawButtonDisabled(Boolean deep)
	{
	// Draw the button in disabled/unpressed state.

	Rect	r = mBounds;
	
	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.bottom = r.top + kCheckBoxSize;
	r.right = r.left + kCheckBoxSize;
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r2]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::RGBForeColor(&gAGARamp[r7]);
		::FrameRect(&r);

		if (mValue == kCheckBoxOn)
			{
			// Draw the X.
			::PenSize(1, 2);
			::MoveTo(r.left + 3, r.top + 2);
			::RGBForeColor(&gAGARamp[r7]);
			::Line(5, 5);
			::Move(-5, 0);
			::Line(5, -5);
			::Move(0, 1);
			}
		else if (mValue == kCheckBoxMixed)
			{
			// Draw the dash.
			::MoveTo(r.left + 3, r.top + 5);
			::PenSize(2, 2);
			::RGBForeColor(&gAGARamp[r7]);
			::Line(4, 0);
			}
		}
	else
		{
		// Paint white over previous junk, if any.
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRect(&r);

		// Paint frame.
		PenPatGray();
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRect(&r);

		// Draw check or dash.
		if (mValue == kCheckBoxOn)
			{
			PenPatBlack();	// 1-pixel diagonals can disappear when qd.gray
			::MoveTo(r.left + 2, r.top + 2);
			::Line(kCheckBoxSize-5, kCheckBoxSize-5);
			::Move(-(kCheckBoxSize-5), 0);
			::Line(kCheckBoxSize-5, -(kCheckBoxSize-5));
			}
		else if (mValue == kCheckBoxMixed)
			{
			::PenSize(1, 2);
			::MoveTo(r.left + 3, r.top + 5);
			::Line(5, 0);
			}
		
		::PenNormal();
		}

	// Draw the title.
	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.left += (kCheckBoxSize + 4);
	r.bottom = r.top + kCheckBoxSize;
	AGAStringOut(mTitle, &r, truncMiddle, teJustLeft, kDisabledOutput, deep, mTextStyle);
	}

//
// AGARadioButton ---------------------------------------------------------------
//
// This class implements a standard radio button object,
// with the additional capabilities of automatic handling of
// radio group mutual exclusivity, and mixed state.
//

Boolean AGARadioButton::mImagesAllocated = false;

// Each radio button drawing state has an offscreen image.
AGAOffscreenImage*	AGARadioButton::mgOffscreenRadioImages[kNumRadioImages];

// These are the color ramp indexes that are used to build the offscreen images.
SInt8				AGARadioButton::kRadioImageValues[kNumRadioImages][kRadioImageHeight][kRadioImageWidth] =
	{
		{	// kRadioNormalOff
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		{ OUT, OUT,  rB, r10,  r2,  r2,  r2,  r4, r11,  rB, OUT, OUT },
		{ OUT,  rB,  r4,  r2,  r1,  rW,  rW,  rW,  r2,  r7,  rB, OUT },
		{  r5, r10,  r2,  r1,  rW,  rW,  r1,  r1,  r2,  r4, r11,  r5 },
		{ r11,  r2,  r1,  rW,  rW,  r1,  r1,  r2,  r2,  r4,  r7, r11 },
		{  rB,  r2,  rW,  rW,  r1,  r1,  r2,  r2,  r4,  r4,  r7,  rB },
		{  rB,  r2,  rW,  r1,  r1,  r2,  r2,  r4,  r4,  r5,  r7,  rB },
		{ r11,  r4,  rW,  r1,  r2,  r2,  r4,  r4,  r5,  r5,  r7, r11 },
		{  r5, r10,  r2,  r2,  r2,  r4,  r4,  r5,  r5,  r7, r11,  r5 },
		{ OUT,  rB,  r7,  r4,  r4,  r4,  r5,  r5,  r7,  r7,  rB, OUT },
		{ OUT, OUT,  rB, r10,  r7,  r7,  r7,  r7, r11,  rB, OUT, OUT },
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		},

		{	// kRadioNormalOn
		{ OUT, OUT, OUT,  r5, r12,  rB,  rB, r12,  r5, OUT, OUT, OUT },
		{ OUT, OUT, r12,  rB, r11, r10, r10, r10, r11,  rB, OUT, OUT },
		{ OUT, r12, r11, r10,  r8,  r8,  r8,  r7,  r7,  r6,  rB, OUT },
		{  r5,  rB, r10,  r8,  rB,  rB,  rB,  rB,  r6,  r6, r11,  r5 },
		{ r12, r11,  r8,  rB,  rB,  rB,  rB,  rB,  rB,  r6,  r4, r11 },
		{  rB, r10,  r8,  rB,  rB,  rB,  rB,  rB,  rB,  r4,  r4,  rB },
		{  rB, r10,  r8,  rB,  rB,  rB,  rB,  rB,  rB,  r4,  r2,  rB },
		{ r12, r10,  r7,  rB,  rB,  rB,  rB,  rB,  rB,  r2,  rW, r11 },
		{  r5, r12,  r7,  r6,  rB,  rB,  rB,  rB,  r2,  rW, r11,  r5 },
		{ OUT, r12,  r6,  r6,  r6,  r4,  r4,  r2,  rW,  rW,  rB, OUT },
		{ OUT, OUT, r12, r11,  r4,  r4,  r2,  rW, r11,  rB, OUT, OUT },
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		},

		{	// kRadioNormalMixed
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		{ OUT, OUT,  rB, r10,  r2,  r2,  r2,  r4, r11,  rB, OUT, OUT },
		{ OUT,  rB,  r4,  r2,  r1,  rW,  rW,  rW,  r2,  r7,  rB, OUT },
		{  r5, r10,  r2,  r1,  rW,  rW,  r1,  r1,  r2,  r4, r11,  r5 },
		{ r11,  r2,  r1,  rW,  rW,  r1,  r1,  r2,  r2,  r4,  r7, r11 },
		{  rB,  r2,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  r4,  r7,  rB },
		{  rB,  r2,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  r5,  r7,  rB },
		{ r11,  r4,  rW,  r1,  r2,  r2,  r4,  r4,  r5,  r5,  r7, r11 },
		{  r5, r10,  r2,  r2,  r2,  r4,  r4,  r5,  r5,  r7, r11,  r5 },
		{ OUT,  rB,  r7,  r4,  r4,  r4,  r5,  r5,  r7,  r7,  rB, OUT },
		{ OUT, OUT,  rB, r10,  r7,  r7,  r7,  r7, r11,  rB, OUT, OUT },
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		},

		{	// kRadioPressedOff
		{ OUT, OUT, OUT,  r5, r12,  rB,  rB, r12,  r5, OUT, OUT, OUT },
		{ OUT, OUT, r12,  rB, r11, r11, r11, r11, r11,  rB, OUT, OUT },
		{ OUT, r12, r11, r11, r10, r10,  r9,  r9,  r9,  r8,  rB, OUT },
		{  r5,  rB, r11, r10, r10,  r9,  r9,  r8,  r8,  r8, r11,  r5 },
		{ r12, r11, r10, r10,  r9,  r9,  r8,  r8,  r8,  r7,  r6, r11 },
		{  rB, r11, r10,  r9,  r9,  r8,  r8,  r8,  r7,  r7,  r6,  rB },
		{  rB, r11,  r9,  r9,  r8,  r8,  r8,  r7,  r7,  r6,  r6,  rB },
		{ r12, r11,  r9,  r8,  r8,  r8,  r7,  r7,  r6,  r6,  r4, r11 },
		{  r5, r12,  r9,  r8,  r8,  r7,  r7,  r6,  r6,  r4, r10,  r5 },
		{ OUT, r11,  r8,  r8,  r7,  r7,  r6,  r6,  r4,  r4,  rB, OUT },
		{ OUT, OUT, r12, r11,  r6,  r6,  r6,  r4, r10,  rB, OUT, OUT },
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		},

		{	// kRadioPressedOn
		{ OUT, OUT, OUT,  r5, r12,  rB,  rB, r12,  r5, OUT, OUT, OUT },
		{ OUT, OUT, r12,  rB, r11, r11, r11, r11, r11,  rB, OUT, OUT },
		{ OUT, r12, r11, r11, r10, r10,  r9,  r9,  r9,  r8,  rB, OUT },
		{  r5,  rB, r11, r10,  rB,  rB,  rB,  rB,  r8,  r8, r11,  r5 },
		{ r12, r11, r10,  rB,  rB,  rB,  rB,  rB,  rB,  r7,  r6, r11 },
		{  rB, r11, r10,  rB,  rB,  rB,  rB,  rB,  rB,  r7,  r6,  rB },
		{  rB, r11,  r9,  rB,  rB,  rB,  rB,  rB,  rB,  r6,  r6,  rB },
		{ r12, r11,  r9,  rB,  rB,  rB,  rB,  rB,  rB,  r6,  r4, r11 },
		{  r5, r12,  r9,  r8,  rB,  rB,  rB,  rB,  r6,  r4, r10,  r5 },
		{ OUT, r11,  r8,  r8,  r7,  r7,  r6,  r6,  r4,  r4,  rB, OUT },
		{ OUT, OUT, r12, r11,  r6,  r6,  r6,  r4, r10,  rB, OUT, OUT },
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		},

		{	// kRadioPressedMixed
		{ OUT, OUT, OUT,  r5, r12,  rB,  rB, r12,  r5, OUT, OUT, OUT },
		{ OUT, OUT, r12,  rB, r11, r11, r11, r11, r11,  rB, OUT, OUT },
		{ OUT, r12, r11, r11, r10, r10,  r9,  r9,  r9,  r8,  rB, OUT },
		{  r5,  rB, r11, r10, r10,  r9,  r9,  r8,  r8,  r8, r11,  r5 },
		{ r12, r11, r10, r10,  r9,  r9,  r8,  r8,  r8,  r7,  r6, r11 },
		{  rB, r11, r10,  rB,  rB,  rB,  rB,  rB,  rB,  r7,  r6,  rB },
		{  rB, r11,  r9,  rB,  rB,  rB,  rB,  rB,  rB,  r6,  r6,  rB },
		{ r12, r11,  r9,  r8,  r8,  r8,  r7,  r7,  r6,  r6,  r4, r11 },
		{  r5, r12,  r9,  r8,  r8,  r7,  r7,  r6,  r6,  r4, r10,  r5 },
		{ OUT, r11,  r8,  r8,  r7,  r7,  r6,  r6,  r4,  r4,  rB, OUT },
		{ OUT, OUT, r12, r11,  r6,  r6,  r6,  r4, r10,  rB, OUT, OUT },
		{ OUT, OUT, OUT,  r5, r11,  rB,  rB, r11,  r5, OUT, OUT, OUT },
		},

		{	// kRadioDisabledOff
		{ OUT, OUT, OUT,  r4,  r7,  r7,  r7,  r7,  r4, OUT, OUT, OUT },
		{ OUT, OUT,  r7,  r7,  r3,  r2,  r2,  r3,  r7,  r7, OUT, OUT },
		{ OUT,  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7, OUT },
		{  r4,  r7,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r7,  r4 },
		{  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7 },
		{  r7,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r7 },
		{  r7,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r7 },
		{  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7 },
		{  r4,  r7,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r7,  r4 },
		{ OUT,  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7, OUT },
		{ OUT, OUT,  r7,  r7,  r3,  r2,  r2,  r3,  r7,  r7, OUT, OUT },
		{ OUT, OUT, OUT,  r4,  r7,  r7,  r7,  r7,  r4, OUT, OUT, OUT },
		},

		{	// kRadioDisabledOn
		{ OUT, OUT, OUT,  r4,  r7,  r7,  r7,  r7,  r4, OUT, OUT, OUT },
		{ OUT, OUT,  r7,  r7,  r4,  r4,  r4,  r4,  r7,  r7, OUT, OUT },
		{ OUT,  r7,  r5,  r4,  r4,  r4,  r4,  r4,  r4,  r5,  r7, OUT },
		{  r4,  r7,  r4,  r4,  r8,  r8,  r8,  r8,  r4,  r4,  r7,  r4 },
		{  r7,  r4,  r4,  r8,  r8,  r8,  r8,  r8,  r8,  r4,  r4,  r7 },
		{  r7,  r4,  r4,  r8,  r8,  r8,  r8,  r8,  r8,  r4,  r4,  r7 },
		{  r7,  r4,  r4,  r8,  r8,  r8,  r8,  r8,  r8,  r4,  r4,  r7 },
		{  r7,  r4,  r4,  r8,  r8,  r8,  r8,  r8,  r8,  r4,  r4,  r7 },
		{  r4,  r7,  r4,  r4,  r8,  r8,  r8,  r8,  r4,  r4,  r7,  r4 },
		{ OUT,  r7,  r5,  r4,  r4,  r4,  r4,  r4,  r4,  r5,  r7, OUT },
		{ OUT, OUT,  r7,  r7,  r4,  r4,  r4,  r4,  r7,  r7, OUT, OUT },
		{ OUT, OUT, OUT,  r4,  r7,  r7,  r7,  r7,  r4, OUT, OUT, OUT },
		},

		{	// kRadioDisabledMixed
		{ OUT, OUT, OUT,  r4,  r7,  r7,  r7,  r7,  r4, OUT, OUT, OUT },
		{ OUT, OUT,  r7,  r7,  r3,  r2,  r2,  r3,  r7,  r7, OUT, OUT },
		{ OUT,  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7, OUT },
		{  r4,  r7,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r7,  r4 },
		{  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7 },
		{  r7,  r2,  r2,  r8,  r8,  r8,  r8,  r8,  r8,  r2,  r2,  r7 },
		{  r7,  r2,  r2,  r8,  r8,  r8,  r8,  r8,  r8,  r2,  r2,  r7 },
		{  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7 },
		{  r4,  r7,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r7,  r4 },
		{ OUT,  r7,  r3,  r2,  r2,  r2,  r2,  r2,  r2,  r3,  r7, OUT },
		{ OUT, OUT,  r7,  r7,  r3,  r2,  r2,  r3,  r7,  r7, OUT, OUT },
		{ OUT, OUT, OUT,  r4,  r7,  r7,  r7,  r7,  r4, OUT, OUT, OUT },
		}

	};

OSErr AGARadioButton::AllocateRadioImages()
	{
	// Allocate the offscreen image for each radio button state.

	OSErr	result = noErr;
	Point	imageSize;
	
	imageSize.h = kRadioImageWidth;
	imageSize.v = kRadioImageHeight;
	
	// A simple round rect defines the radio image clipping region.

	RgnHandle	radioImageClippingRegion = ::NewRgn();
	
	if (radioImageClippingRegion == NULL)
		result = memFullErr;
	else
		{
		Rect	imageClippingRect;
		
		::SetRect(&imageClippingRect, 0, 0, kRadioImageWidth, kRadioImageHeight);
		::OpenRgn();
		::FrameRoundRect(&imageClippingRect, 11, 11);
		//::FrameOval(&imageClippingRect);	ugly AGA 7/96 specs use just the chunky circle -- we'll stick with earlier specs
		::CloseRgn(radioImageClippingRegion);
		}

	for (UInt32 imageIndex = 0; imageIndex < kNumRadioImages; imageIndex++)
		{
		if (result == noErr)
			{
			mgOffscreenRadioImages[imageIndex] = new AGAOffscreenImage;
			
			result = mgOffscreenRadioImages[imageIndex]->CreateImageData(
						imageSize,
						&kRadioImageValues[imageIndex][0][0],
						radioImageClippingRegion, 
						&gAGARamp[rB]);
			}
		}
		
	if (radioImageClippingRegion != NULL)
		::DisposeRgn(radioImageClippingRegion);
	
	mImagesAllocated = true;

	return result;
	}

void AGARadioButton::DisposeRadioImages()
	{
	for (UInt32 imageIndex = 0; imageIndex < kNumRadioImages; imageIndex++)
		if (mgOffscreenRadioImages[imageIndex] != NULL)
			delete mgOffscreenRadioImages[imageIndex];
	}

AGARadioButton::AGARadioButton(Rect* bounds, const AGATextStyle& textStyle, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState)
: AGAObject(bounds), MExclusiveObject(groupIDPart1, groupIDPart2)
	{
	// Construct with initial empty button title.

	mValue = kRadioButtonOff;
	mAutomaticState = automaticState;
	mTitle[0];
	mTextStyle = textStyle;
	
	if (! mImagesAllocated)
		(void) AllocateRadioImages();
	}

AGARadioButton::AGARadioButton(Rect* bounds, const AGATextStyle& textStyle, StringPtr title, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState)
: AGAObject(bounds), MExclusiveObject(groupIDPart1, groupIDPart2)
	{
	// Construct with specified button title.

	mValue = kRadioButtonOff;
	mAutomaticState = automaticState;
	AGA_PLstrcpy(mTitle, title);
	mTextStyle = textStyle;
	
	if (! mImagesAllocated)
		(void) AllocateRadioImages();
	}

AGARadioButton::AGARadioButton(Rect* bounds, const AGATextStyle& textStyle, SInt16 stringListResourceID, SInt16 stringIndex, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState)
: AGAObject(bounds), MExclusiveObject(groupIDPart1, groupIDPart2)
	{
	// Construct by getting button title from specified resource.

	mValue = kRadioButtonOff;
	mAutomaticState = automaticState;
	::GetIndString(mTitle, stringListResourceID, stringIndex);
	mTextStyle = textStyle;
	
	if (! mImagesAllocated)
		(void) AllocateRadioImages();
	}

AGARadioButton::~AGARadioButton()
	{
	}

void AGARadioButton::DrawObject()
	{
	// Draw button in normal unpressed state.

	this->DrawButton(kNotPressed);
	}

Boolean AGARadioButton::TrackMouse(Point mouseLocation)
	{
	// If inherited tracking succeeded and we are set to
	// automatically change our state, toggle it.

	Boolean	wasItHit = AGAObject::TrackMouse(mouseLocation);

	if (wasItHit && mAutomaticState)
		this->SetValue(kRadioButtonOn, kRedraw);

	return wasItHit;
	}

void AGARadioButton::SetTitle(StringPtr title, Boolean redraw)
	{
	// Change button title, redraw if specified.

	AGA_PLstrcpy(mTitle, title);
	
	if (redraw)
		this->DrawObject();
	}

SInt32 AGARadioButton::GetValue()
	{
	// Return current state.

	return mValue;
	}

void AGARadioButton::SetValue(SInt32 newValue, Boolean redraw)
	{
	// Set current state, redraw if specified and necessary.

	if (newValue == kRadioButtonOn)
		gGroupsContainer->AGAHitGroupMember(this);

	if (newValue != mValue)
		{
		mValue = newValue;
		
		if (redraw)
			{
			this->DrawButton(kNotPressed);
			}
		}
	}

void AGARadioButton::SetGroupID(UInt32 groupIDPart1, UInt32 groupIDPart2)
	{
	MExclusiveObject::SetGroupID(groupIDPart1, groupIDPart2);

	// If we are being installed into a group, we should turn automatic
	// mode on.
	
	if ((groupIDPart1 != kNoGroupID) && (groupIDPart2 != kNoGroupID))
		mAutomaticState = true;
	}

void AGARadioButton::TurnOff()
	{
	// Turn off the radio button, presumably in response
	// to another button in the group being turned on.

	this->SetValue(kRadioButtonOff, kRedraw);
	}

void AGARadioButton::DrawButton(Boolean pressed)
	{
	AGADrawingEnvironment	env;

	// Draw the button in the appropriate state.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (! mEnabled || ! mActive)
			this->DrawButtonDisabled(deep);
		else if (pressed)
			this->DrawButtonPressed(deep);
		else
			this->DrawButtonNormal(deep);
		}
	}

void AGARadioButton::SetTrackingState(Boolean isIn)
	{
	// Draw the button pressed or unpressed as specified.

	this->DrawButton(isIn);
	}

void AGARadioButton::DrawButtonNormal(Boolean deep)
	{
	Rect	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	// Draw the button in enabled/unpressed state.

	if (deep)
		{
		if (mValue == kRadioButtonMixed)
			this->CopyImage(kRadioNormalMixed);
		else if (mValue == kRadioButtonOn)
			this->CopyImage(kRadioNormalOn);
		else
			this->CopyImage(kRadioNormalOff);
		}
	else
		{
		r.right = r.left + kRadioImageWidth;
		r.bottom = r.top + kRadioImageHeight;

		// Paint white over previous junk, if any.
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRect(&r);
		// Paint frame.
		::RGBForeColor(&gAGARamp[rB]);
		::FrameOval(&r);

		// Draw dot or dash.
		if (mValue == kRadioButtonOn)
			{
			::InsetRect(&r, 3, 3);
			::PaintRoundRect(&r, 4, 4);
			}
		else if (mValue == kRadioButtonMixed)
			{
			::PenSize(1, 2);
			::MoveTo(r.left + 3, r.top + 5);
			::Line(5, 0);
			}
		
		::PenNormal();
		}

	// Draw the title.
	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.left += (kRadioImageWidth + 4);
	r.bottom = r.top + kRadioImageHeight;
	AGAStringOut(mTitle, &r, truncMiddle, teJustLeft, kNormalOutput, deep, mTextStyle);
	}

void AGARadioButton::DrawButtonPressed(Boolean deep)
	{
	Rect	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	// Draw the button in enabled/pressed state.

	if (deep)
		{
		if (mValue == kRadioButtonMixed)
			this->CopyImage(kRadioPressedMixed);
		else if (mValue == kRadioButtonOn)
			this->CopyImage(kRadioPressedOn);
		else
			this->CopyImage(kRadioPressedOff);
		}
	else
		{
		r.right = r.left + kRadioImageWidth;
		r.bottom = r.top + kRadioImageHeight;

		// Paint white over previous junk, if any.
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRect(&r);
		// Paint frame.
		::RGBForeColor(&gAGARamp[rB]);
		::PenSize(2, 2);
		::FrameOval(&r);

		// Draw dot or dash.
		if (mValue == kRadioButtonOn)
			{
			::InsetRect(&r, 3, 3);
			::PaintRoundRect(&r, 4, 4);
			}
		else if (mValue == kRadioButtonMixed)
			{
			::PenSize(1, 2);
			::MoveTo(r.left + 3, r.top + 5);
			::Line(5, 0);
			}
		
		::PenNormal();
		}

	// Draw the title.
	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.left += (kRadioImageWidth + 4);
	r.bottom = r.top + kRadioImageHeight;
	AGAStringOut(mTitle, &r, truncMiddle, teJustLeft, kNormalOutput, deep, mTextStyle);
	}

void AGARadioButton::DrawButtonDisabled(Boolean deep)
	{
	Rect	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	// Draw the button in disabled/unpressed state.

	if (deep)
		{
		if (mValue == kRadioButtonMixed)
			this->CopyImage(kRadioDisabledMixed);
		else if (mValue == kRadioButtonOn)
			this->CopyImage(kRadioDisabledOn);
		else
			this->CopyImage(kRadioDisabledOff);
		}
	else
		{
		r.right = r.left + kRadioImageWidth;
		r.bottom = r.top + kRadioImageHeight;

		// Paint white over previous junk, if any.
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRect(&r);

		// Paint frame.
		PenPatGray();
		::RGBForeColor(&gAGARamp[rB]);
		::FrameOval(&r);

		// Draw check or dash.
		if (mValue == kRadioButtonOn)
			{
			::InsetRect(&r, 3, 3);
			::PaintRoundRect(&r, 4, 4);
			}
		else if (mValue == kRadioButtonMixed)
			{
			::PenSize(1, 2);
			::MoveTo(r.left + 3, r.top + 5);
			::Line(5, 0);
			}
		
		::PenNormal();
		}

	// Draw the title.
	r = mBounds;

	::OffsetRect(&r, 2, 2);	// visually match the system CDEF

	r.left += (kRadioImageWidth + 4);
	r.bottom = r.top + kRadioImageHeight;
	AGAStringOut(mTitle, &r, truncMiddle, teJustLeft, kDisabledOutput, deep, mTextStyle);
	}

void AGARadioButton::CopyImage(UInt32 index)
	{
	// Blast the specified offscreen image into the
	// radio button's button area. We offset by (2, 2)
	// to match the positioning of the system CDEF.

	mgOffscreenRadioImages[index]->DrawImage(mBounds.left + 2, mBounds.top + 2);
	}

//
// AGAScrollBar ---------------------------------------------------------------
//
// This class implements a standard AGA scroll bar, with the additional
// ability to perform live tracking and a proportional thumb.
//

// These regions are used momentarily while tracking the scroll
// bar indicator to clip and optimize redrawing.
RgnHandle AGAScrollBar::mgSavedIndicatorClip = ::NewRgn();
RgnHandle AGAScrollBar::mgOldIndicatorClip = ::NewRgn();
RgnHandle AGAScrollBar::mgNewIndicatorClip = ::NewRgn();

AGAScrollBar::AGAScrollBar(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue)
	: AGATrackingIndicator(bounds, minimum, maximum, initialValue)
	{
	// Construct with default options.
	mLiveTracking = TestGrayCouncilDefault(kAGALiveScrolling);
	mIsProportional = TestGrayCouncilDefault(kAGAProportionalScrolling);

	mSingleStepSize = 1;
	mPageStepSize = 1;
	}

AGAScrollBar::AGAScrollBar(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, SInt32 singleStepSize, SInt32 pageStepSize, SInt32 pageSize, Boolean liveTracking, Boolean proportional)
	: AGATrackingIndicator(bounds, minimum, maximum, initialValue, pageSize, liveTracking, proportional)
	{
	// Construct with specified options.

	mSingleStepSize = singleStepSize;
	mPageStepSize = pageStepSize;
	}

AGAScrollBar::~AGAScrollBar()
	{
	}

void AGAScrollBar::SetRange(SInt32 newMinimum, SInt32 newMaximum, Boolean redraw)
	{
	// For scroll bars, min==max means the scroll bar is drawn disabled.
	// So we need to catch a range change that changes this implicit state
	// that is peculiar to this subclass. When this happens, we don't want
	// the normal indicator redraw to happen; rather, we want to redraw the
	// whole object in its new state.
	
	Boolean	wasEnabled = (mMaximum > mMinimum);
	Boolean	isEnabled = (newMaximum > newMinimum);
	Boolean	changedState = (isEnabled != wasEnabled);
	
	AGATrackingIndicator::SetRange(newMinimum, newMaximum, redraw && ! changedState);

	if (redraw && changedState)
		this->DrawObject();
	}

void AGAScrollBar::SetStepSizes(SInt32 singleStepSize, SInt32 pageStepSize)
	{
	// Set the scrolling step delta amounts.

	mSingleStepSize = singleStepSize;
	mPageStepSize = pageStepSize;
	}

Boolean AGAScrollBar::TrackPart(Point mouseLocation, SInt32 partHit)
	{
	// Track the mouse and return true if tracking changed the
	// indicator value. Thumb tracking is handled by superclass.
	// Scroll bar just handles arrow and page tracking.

	if ((partHit != kArrowMinusPress) && (partHit != kArrowPlusPress) &&
		(partHit != kPageMinusPress) && (partHit != kPagePlusPress))
		return AGATrackingIndicator::TrackPart(mouseLocation, partHit);

	if (mMaximum <= mMinimum)	// implicitly disabled (no range)
		return false;

	Boolean	wasIn = false;
	Boolean	isIn = true;
	Boolean	isButtonDown = true;	// ensure at least one time through
	Point	newMouseLocation;
	UInt32	accelerationDelay = 18;	// will divide by 2 each time
	
	while (isButtonDown)
		{
		if (isIn)
			this->TrackDelta(partHit);

		::GetMouse(&newMouseLocation);	// gives local coordinates
		
		isIn = (this->TrackTestPart(newMouseLocation) == partHit);
		
		if (isIn != wasIn)
			this->DrawArrow(partHit, isIn);
		
		wasIn = isIn;

		if (isButtonDown && (accelerationDelay != 0))
			{
			DelayFutureWhileStillDown(accelerationDelay);
			accelerationDelay = accelerationDelay >> 1;	// divide by 2
			}

		isButtonDown = ::StillDown();
		}
	
	if (isIn)	// if finished IN, restore normal appearance
		this->DrawArrow(partHit, kNotPressed);
	
	return true;
	}

SInt32 AGAScrollBar::TrackTestPart(Point mouseLocation)
	{
	// Return the part value that the specified mouse
	// position would hit.

	Rect	arrowMinusRect = mBounds;
	Rect	arrowPlusRect = mBounds;
	Rect	pageMinusRect = mBounds;
	Rect	pagePlusRect = mBounds;
	Rect	indicatorRect;
	
	this->GetIndicatorBox(mValue, &indicatorRect);
	
	if (mIsHorizontal)
		{
		arrowPlusRect.left = arrowPlusRect.right - kArrowSize;
		arrowMinusRect.right = arrowMinusRect.left + kArrowSize;
		
		pageMinusRect.left = arrowMinusRect.right;
		pageMinusRect.right = indicatorRect.left;
		
		pagePlusRect.left = indicatorRect.right;
		pagePlusRect.right = arrowPlusRect.left;
		}
	else
		{
		arrowPlusRect.top = arrowPlusRect.bottom - kArrowSize;
		arrowMinusRect.bottom = arrowMinusRect.top + kArrowSize;
		
		pageMinusRect.top = arrowMinusRect.bottom;
		pageMinusRect.bottom = indicatorRect.top;
		
		pagePlusRect.top = indicatorRect.bottom;
		pagePlusRect.bottom = arrowPlusRect.top;
		}
	
	if (::PtInRect(mouseLocation, &arrowMinusRect))
		return kArrowMinusPress;
	else if (::PtInRect(mouseLocation, &arrowPlusRect))
		return kArrowPlusPress;
	else if (::PtInRect(mouseLocation, &pageMinusRect))
		return kPageMinusPress;
	else if (::PtInRect(mouseLocation, &pagePlusRect))
		return kPagePlusPress;
	else if (::PtInRect(mouseLocation, &indicatorRect))
		return kIndicatorPress;
	else
		return kNoTrackPress;
	}

Boolean AGAScrollBar::IsValidIndicatorTrackMouse(Point mouseLocation)
	{
	// Return true if the mouse is reasonably near
	// the track.

	Rect	okRect = mBounds;
	
	::InsetRect(&okRect, -32, -32);	// slop for mouse tracking

	return ::PtInRect(mouseLocation, &okRect);
	}

SInt32 AGAScrollBar::GetValueFromMouseDelta(SInt32 originalValue, Point originalMouseLocation, Point newMouseLocation)
	{
	// Return the new indicator value based on the mouse delta
	// from the original mouse location and indicator value.

	SInt32	dontCare;
	SInt32	valueRange = mMaximum - mMinimum;
	SInt32	pixelRange = this->GetIndicatorPixelRange(&dontCare);
	SInt32	pixelDelta;
	float	pixelFraction;
	
	if (mIsHorizontal)
		pixelDelta = newMouseLocation.h - originalMouseLocation.h;
	else
		pixelDelta = newMouseLocation.v - originalMouseLocation.v;

	pixelFraction = ((float) pixelDelta) / (float) pixelRange;
	
	float delta = pixelFraction * ((float) valueRange);
	delta += (float) 0.5;	// round to nearest integer, don't just truncate fraction
	
	return originalValue + delta;
	}

void AGAScrollBar::DrawIndicatorTrack()
	{
	AGADrawingEnvironment	env;

	// Draw the empty scroll bar track.

	GDIterator	iter;
	Boolean		deep;

	Boolean	enabled = mEnabled && (mMaximum > mMinimum);

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			Rect	r = mBounds;

			if (mActive)
				::RGBForeColor(&gAGARamp[rB]);
			else
				::RGBForeColor(&gAGARamp[r10]);

			::FrameRect(&r);
			
			if (mIsHorizontal)
				::InsetRect(&r, kArrowSize - 1, 0);
			else
				::InsetRect(&r, 0, kArrowSize - 1);
			
			if ((! mActive) || (! enabled))
				{
				::RGBForeColor(&gAGARamp[r1]);
				::InsetRect(&r, 1, 1);
				::PaintRect(&r);
				::InsetRect(&r, -1, -1);
				
				if (mActive)
					::RGBForeColor(&gAGARamp[r8]);
				
				if (mIsHorizontal)
					{
					::MoveTo(r.left, r.top + 1);
					::LineTo(r.left, r.bottom - 2);
					::MoveTo(r.right - 1, r.top + 1);
					::LineTo(r.right - 1, r.bottom - 2);
					}
				else
					{
					::MoveTo(r.left + 1, r.top);
					::LineTo(r.right - 2, r.top);
					::MoveTo(r.left + 1, r.bottom - 1);
					::LineTo(r.right - 2, r.bottom - 1);
					}
				}
			else	// enabled
				{
				::RGBForeColor(&gAGARamp[r5]);
				::InsetRect(&r, 1, 1);
				::PaintRect(&r);
				::InsetRect(&r, -1, -1);
				
				::RGBForeColor(&gAGARamp[rB]);
				
				if (mIsHorizontal)
					{
					::MoveTo(r.left, r.top + 1);
					::LineTo(r.left, r.bottom - 2);
					::MoveTo(r.right - 1, r.top + 1);
					::LineTo(r.right - 1, r.bottom - 2);
					
					// Draw the shadowing.

					::RGBForeColor(&gAGARamp[r8]);
					::MoveTo(r.left + 1, r.bottom - 3);
					::LineTo(r.left + 1, r.top + 1);
					::LineTo(r.right - 2, r.top + 1);

					::RGBForeColor(&gAGARamp[r7]);
					::MoveTo(r.left + 2, r.bottom - 4);
					::LineTo(r.left + 2, r.top + 2);
					::LineTo(r.right - 3, r.top + 2);

					::RGBForeColor(&gAGARamp[r3]);
					::MoveTo(r.left + 2, r.bottom - 2);
					::LineTo(r.right - 2, r.bottom - 2);

					::RGBForeColor(&gAGARamp[r4]);
					::MoveTo(r.left + 3, r.bottom - 3);
					::LineTo(r.right - 2, r.bottom - 3);
					}
				else
					{
					::MoveTo(r.left + 1, r.top);
					::LineTo(r.right - 2, r.top);
					::MoveTo(r.left + 1, r.bottom - 1);
					::LineTo(r.right - 2, r.bottom - 1);
					
					// Draw the shadowing.

					::RGBForeColor(&gAGARamp[r8]);
					::MoveTo(r.left + 1, r.bottom - 2);
					::LineTo(r.left + 1, r.top + 1);
					::LineTo(r.right - 3, r.top + 1);

					::RGBForeColor(&gAGARamp[r7]);
					::MoveTo(r.left + 2, r.bottom - 2);
					::LineTo(r.left + 2, r.top + 2);
					::LineTo(r.right - 4, r.top + 2);

					::RGBForeColor(&gAGARamp[r3]);
					::MoveTo(r.right - 2, r.top + 2);
					::LineTo(r.right - 2, r.bottom - 2);

					::RGBForeColor(&gAGARamp[r4]);
					::MoveTo(r.right - 3, r.top + 3);
					::LineTo(r.right - 3, r.bottom - 2);
					}
				}
			}
		else	// 1-bit
			{
			Rect	r = mBounds;

			::InsetRect(&r, 1, 1);
			::RGBForeColor(&gAGARamp[rW]);
			::PaintRect(&r);
			::InsetRect(&r, -1, -1);

			::RGBForeColor(&gAGARamp[rB]);
			::FrameRect(&mBounds);

			if (enabled && mActive)
				{
				if (mIsHorizontal)
					{
					::InsetRect(&r, kArrowSize - 1, 0);
					::MoveTo(r.left, r.top + 1);
					::LineTo(r.left, r.bottom - 2);
					::MoveTo(r.right - 1, r.top + 1);
					::LineTo(r.right - 1, r.bottom - 2);
					}
				else
					{
					::InsetRect(&r, 0, kArrowSize - 1);
					::MoveTo(r.left + 1, r.top);
					::LineTo(r.right - 2, r.top);
					::MoveTo(r.left + 1, r.bottom - 1);
					::LineTo(r.right - 2, r.bottom - 1);
					}
				}
			}
		}
	}

void AGAScrollBar::DrawTrackEnds()
	{
	AGADrawingEnvironment	env;

	// Draw the scroll arrows.

	this->DrawArrow(kArrowMinusPress, kNotPressed);
	this->DrawArrow(kArrowPlusPress, kNotPressed);
	}

void AGAScrollBar::DrawIndicator()
	{
	Boolean	enabled = mEnabled && (mMaximum > mMinimum);

	if ((! enabled) || (! mActive))
		return;

	AGADrawingEnvironment	env;

	// Draw the thumb indicator.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		enum { Fill, Light, TL, BR, Dark, kIndicatorColors };

		UInt8	colorIndexes[kIndicatorColors];
		Rect	r = mBounds;
		SInt32	startPt;
		SInt32	valuePt;
		SInt32	stopPt;
		Point	drawDirection;
		Point	moveDirection;
		SInt16	lineLength;
		
		if (deep)
			{
			if (mIsPressed)
				{
				colorIndexes[Fill] = r8;
				colorIndexes[Light] = r3;
				colorIndexes[TL] = r5;
				colorIndexes[BR] = r10;
				colorIndexes[Dark] = r12;
				}
			else
				{
				colorIndexes[Fill] = r5;
				colorIndexes[Light] = r1;
				colorIndexes[TL] = r3;
				colorIndexes[BR] = r8;
				colorIndexes[Dark] = r10;
				}
			}
		else
			{
			if (mIsPressed)
				{
				colorIndexes[Fill] = rB;
				colorIndexes[Light] = rB;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = rB;
				colorIndexes[Dark] = rB;
				}
			else
				{
				colorIndexes[Fill] = rW;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rB;
				colorIndexes[BR] = rW;
				colorIndexes[Dark] = rW;
				}
			}
		
		this->GetIndicatorSpan(mValue, &startPt, &valuePt, &stopPt);
		this->GetIndicatorBox(mValue, &r);
		
		if (mIsHorizontal)
			{
			::SetPt(&drawDirection, 0, 1);
			::SetPt(&moveDirection, 1, 0);
			lineLength = r.bottom - r.top - 11;
			}
		else
			{
			::SetPt(&drawDirection, 1, 0);
			::SetPt(&moveDirection, 0, 1);
			lineLength = r.right - r.left - 11;
			}

		::RGBForeColor(&gAGARamp[rB]);
		::FrameRect(&r);
		
		::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
		::InsetRect(&r, 1, 1);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		if (deep)
			{
			::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
			::MoveTo(r.left + 1, r.bottom - 3);
			::LineTo(r.left + 1, r.top + 2);
			::Move(1, -1);
			::LineTo(r.right - 3, r.top + 1);
			
			::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
			::MoveTo(r.right - 2, r.top + 2);
			::LineTo(r.right - 2, r.bottom - 2);
			::LineTo(r.left + 2, r.bottom - 2);

			::RGBForeColor(&gAGARamp[colorIndexes[Light]]);
			::MoveTo(r.left + 1, r.top + 1);
			::Line(0, 0);
			}
		
		::MoveTo(r.left + 4 + (moveDirection.h * (((r.right - r.left) / 2) - 8)),
				r.top + 4 + (moveDirection.v * (((r.bottom - r.top) / 2) - 8)));

		if (! deep)
			::Move(moveDirection.h, moveDirection.v);

		for (SInt32 i = 0; i < 4; i++)
			{
			::RGBForeColor(&gAGARamp[colorIndexes[Light]]);
			::Line(0, 0);
			
			::Move(drawDirection.h, drawDirection.v);
			::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
			::Line(drawDirection.h * lineLength, drawDirection.v * lineLength);
			
			::Move(moveDirection.h, moveDirection.v);
			::Move(drawDirection.h, drawDirection.v);
			::RGBForeColor(&gAGARamp[colorIndexes[Dark]]);
			::Line(- (drawDirection.h * (lineLength+1)), - (drawDirection.v * (lineLength+1)));

			::Move(moveDirection.h, moveDirection.v);
			::Move(- drawDirection.h, - drawDirection.v);
			}
		}
	}

void AGAScrollBar::RemoveIndicator(SInt32 newValue)
	{
	AGADrawingEnvironment	env;

	// Draw the track/background to remove the current
	// indicator.

	//
	// Set clipping so we only draw the track to
	// wipe out the old indicator, but not where
	// the new indicator will get drawn anyway.
	// This reduces flicker.
	//

	::GetClip(mgSavedIndicatorClip);

	if (newValue != mValue)
		{
		Rect	oldIndicatorBox;
		Rect	newIndicatorBox;

		this->GetIndicatorBox(mValue, &oldIndicatorBox);
		this->GetIndicatorBox(newValue, &newIndicatorBox);
		
		::RectRgn(mgOldIndicatorClip, &oldIndicatorBox);
		::RectRgn(mgNewIndicatorClip, &newIndicatorBox);
		::DiffRgn(mgOldIndicatorClip, mgNewIndicatorClip, mgOldIndicatorClip);

		::SetClip(mgOldIndicatorClip);
		}

	this->DrawIndicatorTrack();

	::SetClip(mgSavedIndicatorClip);
	}

void AGAScrollBar::DrawGhost()
	{
	AGADrawingEnvironment	env;

	// Draw the indicator as a ghost at the current mGhostValue
	// position.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		Rect	r = mBounds;
		SInt32	startPt;
		SInt32	valuePt;
		SInt32	stopPt;
		Point	drawDirection;
		Point	moveDirection;
		SInt16	lineLength;
		
		this->GetIndicatorSpan(mGhostValue, &startPt, &valuePt, &stopPt);
		this->GetIndicatorBox(mGhostValue, &r);
		
		if (deep)
			{
			if (mIsHorizontal)
				{
				::SetPt(&drawDirection, 0, 1);
				::SetPt(&moveDirection, 1, 0);
				lineLength = r.bottom - r.top - 11;
				
				::RGBForeColor(&gAGARamp[r4]);
				::InsetRect(&r, 1, 3);
				::PaintRect(&r);
				::InsetRect(&r, -1, -3);
				
				::RGBForeColor(&gAGARamp[r11]);
				::MoveTo(r.left, r.top + 1);
				::Line(0, 0);
				::MoveTo(r.right - 1, r.top + 1);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r10]);
				::MoveTo(r.left, r.top + 2);
				::Line(0, 0);
				::MoveTo(r.right - 1, r.top + 2);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r8]);
				::MoveTo(r.left, r.top + 3);
				::LineTo(r.left, r.bottom - 4);
				::MoveTo(r.right - 1, r.top + 3);
				::LineTo(r.right - 1, r.bottom - 4);
				
				::RGBForeColor(&gAGARamp[r6]);
				::MoveTo(r.left, r.bottom - 3);
				::Line(0, 0);
				::MoveTo(r.right - 1, r.bottom - 3);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r4]);
				::MoveTo(r.left, r.bottom - 2);
				::Line(0, 0);
				::MoveTo(r.right - 1, r.bottom - 2);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r6]);
				::MoveTo(r.left + 1, r.top + 1);
				::LineTo(r.right - 2, r.top + 1);
				
				::RGBForeColor(&gAGARamp[r5]);
				::MoveTo(r.left + 1, r.top + 2);
				::LineTo(r.right - 2, r.top + 2);
				
				::RGBForeColor(&gAGARamp[r3]);
				::MoveTo(r.left + 1, r.bottom - 3);
				::LineTo(r.right - 2, r.bottom - 3);
				
				::RGBForeColor(&gAGARamp[r2]);
				::MoveTo(r.left + 1, r.bottom - 2);
				::LineTo(r.right - 2, r.bottom - 2);
				}
			else
				{
				::SetPt(&drawDirection, 1, 0);
				::SetPt(&moveDirection, 0, 1);
				lineLength = r.right - r.left - 11;
				
				::RGBForeColor(&gAGARamp[r4]);
				::InsetRect(&r, 3, 1);
				::PaintRect(&r);
				::InsetRect(&r, -3, -1);
				
				::RGBForeColor(&gAGARamp[r11]);
				::MoveTo(r.left + 1, r.top);
				::Line(0, 0);
				::MoveTo(r.left + 1, r.bottom - 1);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r10]);
				::MoveTo(r.left + 2, r.top);
				::Line(0, 0);
				::MoveTo(r.left + 2, r.bottom - 1);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r8]);
				::MoveTo(r.left + 3, r.top);
				::LineTo(r.right - 4, r.top);
				::MoveTo(r.left + 3, r.bottom - 1);
				::LineTo(r.right - 4, r.bottom - 1);
				
				::RGBForeColor(&gAGARamp[r6]);
				::MoveTo(r.right - 3, r.top);
				::Line(0, 0);
				::MoveTo(r.right - 3, r.bottom - 1);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r4]);
				::MoveTo(r.right - 2, r.top);
				::Line(0, 0);
				::MoveTo(r.right - 2, r.bottom - 1);
				::Line(0, 0);
				
				::RGBForeColor(&gAGARamp[r6]);
				::MoveTo(r.left + 1, r.top + 1);
				::LineTo(r.left + 1, r.bottom - 2);
				
				::RGBForeColor(&gAGARamp[r5]);
				::MoveTo(r.left + 2, r.top + 1);
				::LineTo(r.left + 2, r.bottom - 2);
				
				::RGBForeColor(&gAGARamp[r3]);
				::MoveTo(r.right - 3, r.top + 1);
				::LineTo(r.right - 3, r.bottom - 2);
				
				::RGBForeColor(&gAGARamp[r2]);
				::MoveTo(r.right - 2, r.top + 1);
				::LineTo(r.right - 2, r.bottom - 2);
				}

			::MoveTo(r.left + 4 + (moveDirection.h * (((r.right - r.left) / 2) - 8)),
					r.top + 4 + (moveDirection.v * (((r.bottom - r.top) / 2) - 8)));

			for (SInt32 i = 0; i < 4; i++)
				{
				::RGBForeColor(&gAGARamp[r2]);
				::Line(drawDirection.h * (lineLength+1), drawDirection.v * (lineLength+1));
				
				::Move(moveDirection.h, moveDirection.v);
				::Move(drawDirection.h, drawDirection.v);
				::RGBForeColor(&gAGARamp[r6]);
				::Line(- (drawDirection.h * (lineLength+1)), - (drawDirection.v * (lineLength+1)));

				::Move(moveDirection.h, moveDirection.v);
				::Move(- drawDirection.h, - drawDirection.v);
				}
			}
		else
			{
			::RGBForeColor(&gAGARamp[rB]);
			::FrameRect(&r);
			}
		}
	}

void AGAScrollBar::RemoveGhost(SInt32 /*newValue*/)
	{
	AGADrawingEnvironment	env;

	// Draw the track/background to remove the current
	// ghost indicator.

	Rect	r;
	RgnHandle	savedClip = ::NewRgn();
	
	::GetClip(savedClip);

	this->GetIndicatorBox(mGhostValue, &r);
	::ClipRect(&r);
	this->DrawIndicatorTrack();
	this->DrawIndicator();
	
	::SetClip(savedClip);
	}

void AGAScrollBar::TrackDelta(SInt32 partHit)
	{
	// Change our value based on the delta amount
	// for the part that was hit. Do notification
	// of the value change.

	SInt32	oldValue = mValue;
	SInt32	newValue = mValue;
	
	switch (partHit)
		{
		case kPageMinusPress:
			newValue -= mPageStepSize;
			break;
		case kArrowMinusPress:
			newValue -= mSingleStepSize;
			break;
		case kArrowPlusPress:
			newValue += mSingleStepSize;
			break;
		case kPagePlusPress:
			newValue += mPageStepSize;
			break;
		}
	
	this->SetValue(newValue, kRedraw);
	
	if (mValue != oldValue)
		this->NotifyValue();
	}

void AGAScrollBar::DrawArrow(SInt32 arrowPart, Boolean pressed)
	{
	AGADrawingEnvironment	env;

	// Draw the specified arrow in the specified mode.

	if ((arrowPart != kArrowMinusPress) &&
		(arrowPart != kArrowPlusPress))
		return;

	enum { Fill, TL, BR, Arrow, kNumArrowColors };

	Boolean	enabled = mEnabled && (mMaximum > mMinimum);
	UInt8	colorIndexes[kNumArrowColors];
	Rect	r = mBounds;
	
	if (mIsHorizontal)
		{
		if (arrowPart == kArrowPlusPress)
			r.left = r.right - kArrowSize;
		else
			r.right = r.left + kArrowSize;
		}
	else
		{
		if (arrowPart == kArrowPlusPress)
			r.top = r.bottom - kArrowSize;
		else
			r.bottom = r.top + kArrowSize;
		}
	
	// Note: we don't draw the outer frame.
	// That's already been drawn.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			if (! mActive)
				{
				colorIndexes[Fill] = r1;
				colorIndexes[TL] = r1;
				colorIndexes[BR] = r1;
				colorIndexes[Arrow] = r1;
				}
			else if (pressed)
				{
				colorIndexes[Fill] = r8;
				colorIndexes[TL] = r10;
				colorIndexes[BR] = r6;
				colorIndexes[Arrow] = rB;
				}
			else if (enabled)
				{
				colorIndexes[Fill] = r2;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = r5;
				colorIndexes[Arrow] = rB;
				}
			else	// disabled
				{
				colorIndexes[Fill] = r2;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = r4;
				colorIndexes[Arrow] = r8;
				}
			}
		else
			{
			if ((! mActive) || ! enabled)
				{
				colorIndexes[Fill] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = rW;
				colorIndexes[Arrow] = rW;
				}
			else if (pressed)
				{
				colorIndexes[Fill] = rB;
				colorIndexes[TL] = rB;
				colorIndexes[BR] = rB;
				colorIndexes[Arrow] = rW;
				}
			else // unpressed + enabled
				{
				colorIndexes[Fill] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = rW;
				colorIndexes[Arrow] = rB;
				}
			}

		::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
		::InsetRect(&r, 1, 1);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
		::MoveTo(r.left + 1, r.bottom - 3);
		::LineTo(r.left + 1, r.top + 1);
		::LineTo(r.right - 3, r.top + 1);

		::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
		::MoveTo(r.right - 2, r.top + 2);
		::LineTo(r.right - 2, r.bottom - 2);
		::LineTo(r.left + 2, r.bottom - 2);

		::RGBForeColor(&gAGARamp[colorIndexes[Arrow]]);
		SInt16	direction = (arrowPart == kArrowPlusPress) ? 1 : -1;
		SInt16	offset = (arrowPart == kArrowPlusPress) ? 6 : 9;
		
		if (mIsHorizontal)
			{
			::MoveTo(r.left + offset, ((r.top + r.bottom) / 2) - 4);
			::Line(0, 7);
			::Move(direction, -6);
			::Line(0, 5);
			::Move(direction, -4);
			::Line(0, 3);
			::Move(direction, -2);
			::Line(0, 1);
			}
		else
			{
			::MoveTo(((r.left + r.right) / 2) - 4, r.top + offset);
			::Line(7, 0);
			::Move(-6, direction);
			::Line(5, 0);
			::Move(-4, direction);
			::Line(3, 0);
			::Move(-2, direction);
			::Line(1, 0);
			}
		}
	}

void AGAScrollBar::GetIndicatorBox(SInt32 indicatorValue, Rect* indicatorBox)
	{
	// Return the bounds of the indicator box assuming that
	// it has specified indicator value.

	Rect	r = mBounds;
	SInt32	startPt;
	SInt32	valuePt;
	SInt32	stopPt;

	this->GetIndicatorSpan(indicatorValue, &startPt, &valuePt, &stopPt);
	
	if (mIsHorizontal)
		{
		r.left = startPt;
		r.right = stopPt;
		}
	else
		{
		r.top = startPt;
		r.bottom = stopPt;
		}
	
	*indicatorBox = r;
	}

SInt32 AGAScrollBar::GetIndicatorPixelRange(SInt32* proportionalIndicatorPixels)
	{
	// Return the pixel range that the center point of the indicator
	// can travel along the track, and also return the amount of
	// range reduction that is caused by proportional indicator growth.
	// Non-proportional indicators use the full track range, minus
	// the size of the indicator itself.

	SInt32	range;
	
	*proportionalIndicatorPixels = 0;

	if (mIsHorizontal)
		range = mBounds.right - mBounds.left - (2 * (kArrowSize + kIndicatorPixelInset)) + 1;
	else
		range = mBounds.bottom - mBounds.top - (2 * (kArrowSize + kIndicatorPixelInset)) + 1;

	if (mIsProportional)
		{
		SInt32	pagePixels;
		
		pagePixels =
			(range * (((float) mPageSize) / ((float) (mMaximum - mMinimum))));
		
		if (pagePixels > 0)
			{
			range -= pagePixels;
			*proportionalIndicatorPixels = pagePixels;
			}
		}
	
	return range;
	}

void AGAScrollBar::GetIndicatorPixelEnds(SInt32* startPt, SInt32* stopPt)
	{
	// Return the start and end pixel locations of the indicator
	// track traveled by the center point of the indicator.

	SInt32	dontCare;
	SInt32	range = this->GetIndicatorPixelRange(&dontCare);

	if (mIsHorizontal)
		{
		*startPt = ((mBounds.left + mBounds.right) / 2) - (range / 2) - 1;
		
		// Bump by 1 if width is odd but range isn't.
		if ((((mBounds.right - mBounds.left) & 1) != 0) &&
			((range & 1) == 0))
			(*startPt)++;
		}
	else
		{
		*startPt = ((mBounds.top + mBounds.bottom) / 2) - (range / 2) - 1;
		
		// Bump by 1 if height is odd but range isn't.
		if ((((mBounds.bottom - mBounds.top) & 1) != 0) &&
			((range & 1) == 0))
			(*startPt)++;
		}

	*stopPt = *startPt + range;	// avoid rounding of 1/2 pixel twice
	}

void AGAScrollBar::GetIndicatorSpan(SInt32 value, SInt32* startPt, SInt32* valuePt, SInt32* stopPt)
	{
	// Return the pixel locations of the indicator's ends and value.

	SInt32	proportionalIndicatorPixels;
	SInt32	trackStart;
	SInt32	trackStop;
	SInt32	trackRange = this->GetIndicatorPixelRange(&proportionalIndicatorPixels);
	
	this->GetIndicatorPixelEnds(&trackStart, &trackStop);
	
	*valuePt = trackStart +
		(trackRange * (((float) (value - mMinimum)) / ((float) (mMaximum - mMinimum))));

	*startPt = *valuePt - kIndicatorPixelInset - (proportionalIndicatorPixels / 2);
	*stopPt = *startPt + (2 * kIndicatorPixelInset) + proportionalIndicatorPixels + 1;
	}

//
// AGAPopupMenu ---------------------------------------------------------------
//
// This class implements a standard popup menu look, and handles the
// mouse by calling PopUpMenuSelect such that the visible menu is
// located where the AGA says it should.
//

AGAPopupMenu::AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment)
: AGAObject(bounds)
	{
	// Construct without a menu installed.

	mMenuID = 0;
	mMenuRef = NULL;
	mMenuRefDisposalKind = kDontDispose;
	mCurrentItemNo = 1;
	mWidthAdjust = adjustment;
	mTitleStyle = titleStyle;
	mTextStyle = textStyle;
	AGA_PLstrcpy(mTitle, title);
	mTitleOffset = titleOffset;
	mTitleJustification = titleJustification;
	}

AGAPopupMenu::AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment, MenuRef itsMenuRef, DisposalKind menuRefDisposalKind)
: AGAObject(bounds)
	{
	// Construct with specified MenuRef.

	if (itsMenuRef == NULL)
		mMenuID = 0;
	else
		mMenuID = (**itsMenuRef).menuID;

	mMenuRef = itsMenuRef;
	mMenuRefDisposalKind = menuRefDisposalKind;

	mCurrentItemNo = 1;
	mWidthAdjust = adjustment;
	mTitleStyle = titleStyle;
	mTextStyle = textStyle;
	AGA_PLstrcpy(mTitle, title);
	mTitleOffset = titleOffset;
	mTitleJustification = titleJustification;
	}

AGAPopupMenu::AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment, SInt16 itemsStringListID, SInt16 newMenuID)
: AGAObject(bounds)
	{
	// Construct with menu built from strings in a 'STR#' resource.

	mMenuID = newMenuID;
	mMenuRef = ::NewMenu(newMenuID, "\pAGA");
	mMenuRefDisposalKind = kDispose;
	mCurrentItemNo = 1;
	mWidthAdjust = adjustment;
	mTitleStyle = titleStyle;
	mTextStyle = textStyle;
	AGA_PLstrcpy(mTitle, title);
	mTitleOffset = titleOffset;
	mTitleJustification = titleJustification;
	
	Handle	stringListHandle = ::GetResource('STR#', itemsStringListID);
	if (stringListHandle != NULL)
		{
		SInt16	numStrings = **((SInt16 **) stringListHandle);
		
		for (SInt16 stringIndex = 1; stringIndex <= numStrings; stringIndex++)
			{
			Str255	aString;
			
			::GetIndString(aString, itemsStringListID, stringIndex);
			
			// Use SetMenuItemText to avoid metacharacter interpretation.
			::AppendMenu(mMenuRef, "\pXXX");
			::SetMenuItemText(mMenuRef, stringIndex, aString);
			}
		}
	}

AGAPopupMenu::AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment, ResType appendResourceType, SInt16 newMenuID)
: AGAObject(bounds)
	{
	// Construct with menu built from specified appended resource type.
	// Typical use is for a Fonts menu, using resource type 'FOND'.

	mMenuID = newMenuID;
	mMenuRef = ::NewMenu(newMenuID, "\pAGA");
	mMenuRefDisposalKind = kDispose;
	mCurrentItemNo = 1;
	mWidthAdjust = adjustment;
	mTitleStyle = titleStyle;
	mTextStyle = textStyle;
	AGA_PLstrcpy(mTitle, title);
	mTitleOffset = titleOffset;
	mTitleJustification = titleJustification;
	
	::AppendResMenu(mMenuRef, appendResourceType);
	}

AGAPopupMenu::~AGAPopupMenu()
	{
	// Destruct by cleaning up the menu data.
	// This depends on how the menu was created.

	this->DisposeExistingMenuRef();
	}

void AGAPopupMenu::DisposeExistingMenuRef()
	{
	// Clean up the menu data based on how it was created
	// or how the client code specified.

	if (mMenuRef != NULL)
		{
		if (mMenuRefDisposalKind == kDispose)
			::DisposeMenu(mMenuRef);
		else if (mMenuRefDisposalKind == kRelease)
			::ReleaseResource((Handle) mMenuRef);
		}

	mMenuRefDisposalKind = kDontDispose;
	}

void AGAPopupMenu::SetMenuRef(MenuRef itsMenuRef, DisposalKind menuRefDisposalKind)
	{
	// Install the specified MenuRef.

	this->DisposeExistingMenuRef();

	if (itsMenuRef == NULL)
		mMenuID = 0;
	else
		mMenuID = (**itsMenuRef).menuID;

	mMenuRef = itsMenuRef;
	mMenuRefDisposalKind = menuRefDisposalKind;

	mCurrentItemNo = 1;
	}

void AGAPopupMenu::SetWidthAdjustment(WidthAdjust adjustment)
	{
	// Set the width adjustment, which determines how wide
	// the closed menu appears.

	mWidthAdjust = adjustment;
	}

void AGAPopupMenu::DrawObject()
	{
	// Draw the unpressed popup menu "button".

	this->DrawButton(kNotPressed);
	}

Boolean AGAPopupMenu::TrackMouse(Point /*mouseLocation*/)
	{
	if (! mEnabled)
		return false;

	// Track the mouse in the popup, and return true
	// if it caused the current popup selection to change.

	Boolean	selectionChanged;
	SInt32	menuResult;
	Point	globalOrigin;
	
	this->DrawButton(kPressed);
	
	::SetPt(&globalOrigin, mBounds.left + mTitleOffset, mBounds.bottom);
	::LocalToGlobal(&globalOrigin);

	::InsertMenu(mMenuRef, -1);			// put it in the hierarchical list
	
	if ((mTextStyle.mFontNum != 0) || ((mTextStyle.mFontSize != 0) && (mTextStyle.mFontSize != 12)))//sizer.mIsCustom)
		::SetItemMark(mMenuRef, mCurrentItemNo, '');
	else
		::CheckItem(mMenuRef, mCurrentItemNo, true);

	{	// create scope so AGACustomPopupMenuSizer is destructed right away
	AGACustomPopupMenuSizer	sizer(mTextStyle.mFontNum, mTextStyle.mFontSize);
	
	menuResult = ::PopUpMenuSelect(mMenuRef, globalOrigin.v, globalOrigin.h + 1, 1);
	}

	::CheckItem(mMenuRef, mCurrentItemNo, false);

	::DeleteMenu((**mMenuRef).menuID);	// get it out of the hierarchical list

	selectionChanged = (menuResult != 0);

	if (selectionChanged)
		this->SetCurrentItemNo(menuResult & 0x0000FFFF, kDontRedraw);
	
	this->DrawButton(kNotPressed);

	return selectionChanged;
	}

SInt16 AGAPopupMenu::GetCurrentItemNo()
	{
	// Return the currently selected item index (1-based).

	return mCurrentItemNo;
	}

void AGAPopupMenu::SetCurrentItemNo(SInt16 newItemNo, Boolean redraw)
	{
	// Set the current item index (1-based), optionally redraw.

	mCurrentItemNo = newItemNo;
	
	if (redraw)
		this->DrawButton(kNotPressed);
	}

void AGAPopupMenu::GetCurrentItemText(StringPtr itemText)
	{
	// Return the menu item text of the currently selected item.

	itemText[0] = 0;
	
	if (mMenuRef != NULL)
		::GetMenuItemText(mMenuRef, this->GetCurrentItemNo(), itemText);
	}

Boolean AGAPopupMenu::SetCurrentItemText(StringPtr itemTextToMatch, Boolean redraw)
	{
	// Set the current item index to the item whose text
	// matches the specified string. Optionally redraw.

	Boolean	wasMatchFound = false;
	SInt16	numItems = 0;

	if (mMenuRef != NULL)
		numItems = ::CountMenuItems(mMenuRef);

	for (SInt16 itemIndex = 1; itemIndex <= numItems; itemIndex++)
		{
		Str255	itemText;

		::GetMenuItemText(mMenuRef, itemIndex, itemText);

		if (::EqualString(itemText, itemTextToMatch, true, true))
			{
			wasMatchFound = true;
			this->SetCurrentItemNo(itemIndex, redraw);
			break;
			}
		}

	return wasMatchFound;
	}

void AGAPopupMenu::SetTitle(StringPtr newTitle, Boolean redraw)
	{
	// Install the new title. Optionally redraw.

	AGA_PLstrcpy(mTitle, newTitle);
	
	if (redraw)
		this->DrawButton(kNotPressed);
	}

void AGAPopupMenu::DrawButton(Boolean pressed)
	{
	// Draw the popup button in the specified state.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (! mEnabled || ! mActive)
			this->DrawButtonDisabled(deep);
		else if (pressed)
			this->DrawButtonPressed(deep);
		else
			this->DrawButtonNormal(deep);
		}
	}

void AGAPopupMenu::DrawButtonNormal(Boolean deep)
	{
	AGADrawingEnvironment	env;

	// Draw the popup button in enabled/unpressed state.

	Rect	r;
	Str255	currentItemText;
	
	this->GetCurrentItemText(currentItemText);
	this->GetButtonBounds(&r);
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r2]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(r.right - r.left - 6, 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, r.bottom - r.top - 6);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(- (r.right - r.left - 6), 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, -(r.bottom - r.top - 6));
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		
		if (r.right - r.left > kPopupArrowSectionWidth)
			{
			// Add flat section inner pixels.
			::RGBForeColor(&gAGARamp[rW]);
			::MoveTo(r.left + 1, r.bottom - 3);
			::LineTo(r.left + 1, r.top + 2);
			::Move(1, -1);
			::LineTo(r.right - kPopupArrowSectionWidth, r.top + 1);
			::RGBForeColor(&gAGARamp[r1]);
			::Line(0, 0);
			::RGBForeColor(&gAGARamp[r5]);
			::MoveTo(r.left + 2, r.bottom - 2);
			::LineTo(r.right - kPopupArrowSectionWidth, r.bottom - 2);
			::LineTo(r.right - kPopupArrowSectionWidth, r.top + 2);
			
			// Add arrow section left corners
			::RGBForeColor(&gAGARamp[r4]);
			::MoveTo(r.right - kPopupArrowSectionWidth + 1, r.bottom - 2);
			::Line(0, 0);
			::RGBForeColor(&gAGARamp[r2]);
			::MoveTo(r.right - kPopupArrowSectionWidth + 1, r.top + 1);
			::Line(0, 0);
			}

		// Add arrow section shading.
		::RGBForeColor(&gAGARamp[r8]);
		::MoveTo(r.right - kPopupArrowSectionWidth + 2, r.bottom - 2);
		::LineTo(r.right - 3, r.bottom - 2);
		::Line(0, -1);
		::Line(1, 0);
		::LineTo(r.right - 2, r.top + 3);
		::Move(-1, -2);
		::RGBForeColor(&gAGARamp[r4]);
		::Line(1, 1);
		::RGBForeColor(&gAGARamp[r5]);
		::Line(-1, 1);
		::LineTo(r.right - 3, r.bottom - 4);
		::Move(-1, 1);
		::LineTo(r.right - kPopupArrowSectionWidth + 3, r.bottom - 3);
		::RGBForeColor(&gAGARamp[rW]);
		::MoveTo(r.right - kPopupArrowSectionWidth + 2, r.bottom - 4);
		::LineTo(r.right - kPopupArrowSectionWidth + 2, r.top + 2);
		::LineTo(r.right - 4, r.top + 2);
		}
	else	// 1-bit
		{
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRoundRect(&r, 7, 7);
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRoundRect(&r, 7, 7);
		}
	
	// Add arrow.
	::RGBForeColor(&gAGARamp[rB]);
	::MoveTo(r.right - kPopupArrowSectionWidth + 6, r.top + ((r.bottom - r.top) / 2) - 2);
	::Line(8, 0);
	::Move(-1, 1);
	::Line(-6, 0);
	::Move(1, 1);
	::Line(4, 0);
	::Move(-1, 1);
	::Line(-2, 0);
	::Line(1, 1);

	if (r.right - r.left > kPopupArrowSectionWidth)
		{
		// Draw the current item text.
		r.left += 7;
		r.right -= kPopupArrowSectionWidth;
		AGAStringOut(currentItemText, &r, truncEnd, teFlushDefault, kNormalOutput, deep, mTextStyle);
		}

	// If the popup has a title, draw it next to the button.	
	if (mTitle[0] != 0)
		{
		this->GetTitleBounds(&r);
		AGAStringOut(mTitle, &r, truncEnd, mTitleJustification, kNormalOutput, deep, mTitleStyle);
		}
	}

void AGAPopupMenu::DrawButtonPressed(Boolean deep)
	{
	AGADrawingEnvironment	env;

	// Draw the popup button in enabled/pressed state.

	Rect	r;
	Str255	currentItemText;
	
	this->GetCurrentItemText(currentItemText);
	this->GetButtonBounds(&r);
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r9]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(r.right - r.left - 6, 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, r.bottom - r.top - 6);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, 0);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(- (r.right - r.left - 6), 0);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, 0);
		::Move(-1, -1);
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		::Move(0, -1);
		::RGBForeColor(&gAGARamp[rB]);
		::Line(0, -(r.bottom - r.top - 6));
		::RGBForeColor(&gAGARamp[r12]);
		::Line(0, 0);
		
		if (r.right - r.left > kPopupArrowSectionWidth)
			{
			// Add flat section inner pixels.
			::RGBForeColor(&gAGARamp[r10]);
			::MoveTo(r.left + 1, r.bottom - 3);
			::LineTo(r.left + 1, r.top + 2);
			::Line(1, 0);
			::Line(0, -1);
			::LineTo(r.right - kPopupArrowSectionWidth, r.top + 1);
			::RGBForeColor(&gAGARamp[r8]);
			::MoveTo(r.left + 2, r.bottom - 2);
			::LineTo(r.right - kPopupArrowSectionWidth, r.bottom - 2);
			::LineTo(r.right - kPopupArrowSectionWidth, r.top + 2);
			
			// Add arrow section left corners
			::RGBForeColor(&gAGARamp[r11]);
			::MoveTo(r.right - kPopupArrowSectionWidth + 1, r.bottom - 2);
			::Line(0, 0);
			::MoveTo(r.right - kPopupArrowSectionWidth + 1, r.top + 1);
			::Line(0, 0);
			}

		// Add arrow section shading.
		::RGBForeColor(&gAGARamp[r7]);
		::MoveTo(r.right - kPopupArrowSectionWidth + 2, r.bottom - 2);
		::LineTo(r.right - 3, r.bottom - 2);
		::Line(0, -1);
		::Line(1, 0);
		::LineTo(r.right - 2, r.top + 2);
		::Move(-1, 1);
		::RGBForeColor(&gAGARamp[r8]);
		::LineTo(r.right - 3, r.bottom - 4);
		::Move(-1, 1);
		::LineTo(r.right - kPopupArrowSectionWidth + 3, r.bottom - 3);
		::RGBForeColor(&gAGARamp[r10]);
		::MoveTo(r.right - kPopupArrowSectionWidth + 2, r.bottom - 4);
		::LineTo(r.right - kPopupArrowSectionWidth + 2, r.top + 2);
		::LineTo(r.right - 4, r.top + 2);
		::RGBForeColor(&gAGARamp[r11]);
		::MoveTo(r.right - kPopupArrowSectionWidth + 1, r.bottom - 3);
		::LineTo(r.right - kPopupArrowSectionWidth + 1, r.top + 2);
		::Move(1, -1);
		::LineTo(r.right - 3, r.top + 1);
		}
	else	// 1-bit
		{
		::RGBForeColor(&gAGARamp[rB]);
		::PaintRoundRect(&r, 7, 7);
		}

	// Add arrow.
	
	::RGBForeColor(&gAGARamp[rW]);
	::MoveTo(r.right - kPopupArrowSectionWidth + 6, r.top + ((r.bottom - r.top) / 2) - 2);
	::Line(8, 0);
	::Move(-1, 1);
	::Line(-6, 0);
	::Move(1, 1);
	::Line(4, 0);
	::Move(-1, 1);
	::Line(-2, 0);
	::Line(1, 1);

	if (r.right - r.left > kPopupArrowSectionWidth)
		{
		// Draw the current item text.
		r.left += 7;
		r.right -= kPopupArrowSectionWidth;
		AGAStringOut(currentItemText, &r, truncEnd, teFlushDefault, kInverseOutput, deep, mTextStyle);
		}

	// (No need to draw the title, since it does not change
	//  upon pressing the button.)
	}

void AGAPopupMenu::DrawButtonDisabled(Boolean deep)
	{
	AGADrawingEnvironment	env;

	// Draw the popup button in disabled/unpressed state.

	Rect	r;
	Str255	currentItemText;
	
	this->GetCurrentItemText(currentItemText);
	this->GetButtonBounds(&r);
	
	::PenSize(1, 1);

	if (deep)
		{
		// Fill the interior.
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[r2]);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		// Draw the frame.
		::MoveTo(r.left + 1, r.top + 1);
		::RGBForeColor(&gAGARamp[r7]);
		::Line(0, 0);
		::Move(1, -1);
		::Line(0, 0);
		::Move(1, 0);
		::Line(r.right - r.left - 6, 0);
		::Line(0, 0);
		::Move(1, 1);
		::Line(0, 0);
		::Move(1, 1);
		::Line(0, 0);
		::Move(0, 1);
		::Line(0, r.bottom - r.top - 6);
		::Line(0, 0);
		::Move(-1, 1);
		::Line(0, 0);
		::Move(-1, 1);
		::Line(0, 0);
		::Move(-1, 0);
		::Line(- (r.right - r.left - 6), 0);
		::Line(0, 0);
		::Move(-1, -1);
		::Line(0, 0);
		::Move(-1, -1);
		::Line(0, 0);
		::Move(0, -1);
		::Line(0, -(r.bottom - r.top - 6));
		::Line(0, 0);

		if (r.right - r.left > kPopupArrowSectionWidth)
			{
			::RGBForeColor(&gAGARamp[r5]);
			::MoveTo(r.right - kPopupArrowSectionWidth, r.bottom - 2);
			::LineTo(r.right - kPopupArrowSectionWidth, r.top + 2);
			}
		}
	else	// 1-bit
		{
		::RGBForeColor(&gAGARamp[rW]);
		::PaintRoundRect(&r, 7, 7);
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRoundRect(&r, 7, 7);
		}

	// Add arrow.
	::RGBForeColor(&gAGARamp[r7]);
	::MoveTo(r.right - kPopupArrowSectionWidth + 6, r.top + ((r.bottom - r.top) / 2) - 2);
	::Line(8, 0);
	::Move(-1, 1);
	::Line(-6, 0);
	::Move(1, 1);
	::Line(4, 0);
	::Move(-1, 1);
	::Line(-2, 0);
	::Line(1, 1);

	if (r.right - r.left > kPopupArrowSectionWidth)
		{
		// Draw the current item text.
		r.left += 7;
		r.right -= kPopupArrowSectionWidth;
		AGAStringOut(currentItemText, &r, truncEnd, teFlushDefault, kDisabledOutput, deep, mTextStyle);
		}
	
	// If the popup has a title, draw it next to the button.	
	if (mTitle[0] != 0)
		{
		this->GetTitleBounds(&r);
		AGAStringOut(mTitle, &r, truncEnd, mTitleJustification, kDisabledOutput, deep, mTitleStyle);
		}
	
	if (! deep)
		{
		::RGBForeColor(&gAGARamp[rB]);
		PenPatGray();
		::PenMode(patBic);
		::PaintRect(&mBounds);

		::PenNormal();
		}
	}

void AGAPopupMenu::GetButtonBounds(Rect* buttonBounds)
	{
	// Return the bounds of the popup button. We may need
	// to check the actual menu's width, depending on the
	// width adjustment specifier.

	*buttonBounds = mBounds;
	buttonBounds->left += mTitleOffset;
	
	if ((mMenuRef != NULL) && (mWidthAdjust != kFixedWidth))
		{
		AGACustomPopupMenuSizer	sizer(mTextStyle.mFontNum, mTextStyle.mFontSize);

		::CalcMenuSize(mMenuRef);
		buttonBounds->right = MinSInt16(mBounds.right, buttonBounds->left + 3 + (**mMenuRef).menuWidth + mWidthAdjust);
		}
	}

void AGAPopupMenu::GetTitleBounds(Rect* titleBounds)
	{
	// Return the bounds of the title area.

	*titleBounds = mBounds;
	
	titleBounds->right = titleBounds->left + mTitleOffset - 3;
	}

//
// AGACustomPopupMenuSizer ---------------------------------------------------------------
//
// Stack-based object to hack in a custom popup menu font/size
// around a call to PopUpMenuSelect or CalcMenuSize.
//

AGACustomPopupMenuSizer::AGACustomPopupMenuSizer(SInt16 fontNum, SInt16 fontSize)
	{
	mCustomFontNum = fontNum;
	mCustomFontSize = fontSize;
	mIsCustom = (fontNum != 0) || ((fontSize != 0) && (fontSize != 12));

	this->SaveEnvironment();
	this->InstallCustomEnvironment();
	}

AGACustomPopupMenuSizer::~AGACustomPopupMenuSizer()
	{
	this->RestoreEnvironment();
	}

void AGACustomPopupMenuSizer::SaveEnvironment()
	{
	if (mIsCustom)
		{
		mSavedFontNum = ::LMGetSysFontFam();
		mSavedFontSize = ::LMGetSysFontSize();
		}
	}

void AGACustomPopupMenuSizer::RestoreEnvironment()
	{
	if (mIsCustom)
		{
		::LMSetSysFontFam(mSavedFontNum);
		::LMSetSysFontSize(mSavedFontSize);
		::LMSetLastSPExtra(-1);
		}
	}

void AGACustomPopupMenuSizer::InstallCustomEnvironment()
	{
	if (mIsCustom)
		{
		::LMSetSysFontFam(mCustomFontNum);
		::LMSetSysFontSize(mCustomFontSize);
		::LMSetLastSPExtra(-1);
		}
	}

