//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// The Gray Council provides a set of standard C++ classes that implement
// the standard Apple Grayscale Appearance. The core classes do not
// require any other code such as a particular class framework.
// There are separate files provided that plug The Gray Council core
// into class frameworks.
//
// This file defines core classes that implement the standard Control Manager
// set of controls.
//
// Classes defined below:
//	AGAStaticText -- simple string object
//	AGAPushButton -- push button with auto default outline capability
//	AGACheckBox -- check box with mixed state capability
//	AGARadioButton -- radio button with mixed state capability
//	AGAScrollBar -- scroll bar with 32-bit values, live tracking, and
//		proportional indicator capabilities
//	AGAPopupMenu -- popup menu with optional title
//	AGACustomPopupMenuSizer -- stack-based class to make PopupMenuSelect use
//		custom font/size
//

#pragma once

#ifndef __GCSTDCONTROLS__
#define __GCSTDCONTROLS__

#include "GCCore.h"

#pragma mark AGAStaticText

//
// AGAStaticText ---------------------------------------------------------------
//
// AGAStaticText draws a static text string on a gray background.
//

class AGAStaticText : public AGAObject
	{
	public:

		AGAStaticText(Rect* bounds, const AGATextStyle& textStyle, SInt32 justification);
		AGAStaticText(Rect* bounds, const AGATextStyle& textStyle, SInt32 justification, StringPtr title);
		AGAStaticText(Rect* bounds, const AGATextStyle& textStyle, SInt32 justification, SInt16 stringListResourceID, SInt16 stringIndex);
		virtual ~AGAStaticText();
		
		virtual void	DrawObject();

		virtual void	SetTitle(StringPtr buttonTitle, Boolean redraw);
		virtual void	SetTextStyle(const AGATextStyle& textStyle, Boolean redraw);
		virtual void	SetJustification(SInt32 justification, Boolean redraw);

	protected:

		SInt32			mJustification;
		Str255			mTitle;
		AGATextStyle	mTextStyle;
	};

#pragma mark AGAPushButton

//
// AGAPushButton ---------------------------------------------------------------
//
// AGAPushButton implements the standard AGA pushbutton. It automatically
// draws the default button outline if you set its mIsDefault property
// by calling SetDefault; you specify whether the outline is inside or
// outside the object's mBounds.
//

class AGAPushButton : public AGAObject
	{
	public:

		AGAPushButton(Rect* bounds, const AGATextStyle& textStyle);
		AGAPushButton(Rect* bounds, const AGATextStyle& textStyle, StringPtr buttonTitle);
		AGAPushButton(Rect* bounds, const AGATextStyle& textStyle, SInt16 stringListResourceID, SInt16 stringIndex);
		virtual ~AGAPushButton();
		
		virtual void	DrawObject();

		virtual void	SetTitle(StringPtr buttonTitle, Boolean redraw);
		
		virtual void	SetDefault(Boolean isDefault, Boolean frameInside);
		virtual Boolean	IsDefault();

		virtual void	DrawButton(Boolean pressed);
		
	protected:

		virtual Boolean	HitTest(Point mouseLocation);
		virtual void	SetTrackingState(Boolean isIn);

		virtual void	DrawButtonNormal(Boolean deep);
		virtual void	DrawButtonPressed(Boolean deep);
		virtual void	DrawButtonDisabled(Boolean deep);
		
		virtual void	DrawOutlineNormal(Boolean deep);
		virtual void	DrawOutlineDisabled(Boolean deep);

		Boolean			mIsDefault;		// if true this button gets a default frame
		Boolean			mFrameInside;	// if true the frame is inside mBounds; else outside
		Str255			mTitle;
		AGATextStyle	mTextStyle;
	};

#pragma mark AGACheckBox

//
// AGACheckBox ---------------------------------------------------------------
//
// AGACheckBox implements the standard AGA check box. It also supports
// a mixed state that is neither on nor off, which is drawn as a dash.
//

class AGACheckBox : public AGAObject
	{
	public:

		enum { kCheckBoxSize = 12 };
		enum { kCheckBoxOff = 0, kCheckBoxOn = 1, kCheckBoxMixed = 2 };

		AGACheckBox(Rect* bounds, const AGATextStyle& textStyle, Boolean automaticState);
		AGACheckBox(Rect* bounds, const AGATextStyle& textStyle, StringPtr buttonTitle, Boolean automaticState);
		AGACheckBox(Rect* bounds, const AGATextStyle& textStyle, SInt16 stringListResourceID, SInt16 stringIndex, Boolean automaticState);
		virtual ~AGACheckBox();
		
		virtual void	DrawObject();
		virtual Boolean	TrackMouse(Point mouseLocation);

		virtual void	SetTitle(StringPtr buttonTitle, Boolean redraw);
		
		virtual SInt32	GetValue();
		virtual void	SetValue(SInt32 newValue, Boolean redraw);

		virtual void	DrawButton(Boolean pressed);

	protected:

		virtual void	SetTrackingState(Boolean isIn);

		virtual void	DrawButtonNormal(Boolean deep);
		virtual void	DrawButtonPressed(Boolean deep);
		virtual void	DrawButtonDisabled(Boolean deep);
		
		SInt32			mValue;
		Boolean			mAutomaticState;	// if true, check is toggled on click w/o caller needing to call SetValue
		Str255			mTitle;
		AGATextStyle	mTextStyle;
	};

#pragma mark AGARadioButton

//
// AGARadioButton ---------------------------------------------------------------
//
// AGARadioButton implements the standard AGA radio button. It also supports
// a mixed state that is neither on nor off, which is drawn as a dash.
// You can also assign an ID pair that will be used to automatically keep
// the radio button mutually exclusive with others with the same ID pair.
// Pass kNoGroupID for the IDs to bypass this. The adapter classes bypass it
// since the frameworks manage radio button groups.
//

class AGARadioButton : public AGAObject, public MExclusiveObject
	{
	public:

		enum { kRadioButtonOff = 0, kRadioButtonOn = 1, kRadioButtonMixed = 2 };

		static Boolean mImagesAllocated;
		static OSErr AllocateRadioImages();
		static void DisposeRadioImages();

		AGARadioButton(Rect* bounds, const AGATextStyle& textStyle, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState);
		AGARadioButton(Rect* bounds, const AGATextStyle& textStyle, StringPtr buttonTitle, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState);
		AGARadioButton(Rect* bounds, const AGATextStyle& textStyle, SInt16 stringListResourceID, SInt16 stringIndex, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState);
		virtual ~AGARadioButton();
		
		virtual void	DrawObject();
		virtual Boolean	TrackMouse(Point mouseLocation);

		virtual void	SetTitle(StringPtr buttonTitle, Boolean redraw);
		
		virtual SInt32	GetValue();
		virtual void	SetValue(SInt32 newValue, Boolean redraw);

		virtual void	SetGroupID(UInt32 groupIDPart1, UInt32 groupIDPart2);	// MExclusiveObject mixin override
		virtual void	TurnOff();	// MExclusiveObject mixin override

		virtual void	DrawButton(Boolean pressed);

	protected:

		virtual void	SetTrackingState(Boolean isIn);

		virtual void	DrawButtonNormal(Boolean deep);
		virtual void	DrawButtonPressed(Boolean deep);
		virtual void	DrawButtonDisabled(Boolean deep);
		
		virtual void	CopyImage(UInt32 index);
		
		SInt32			mValue;
		Boolean			mAutomaticState;	// if true, radio is set on click w/o caller needing to call SetValue
		UInt32			mGroupID;
		Str255			mTitle;
		AGATextStyle	mTextStyle;
		
		// Drawing the 3D radio button manually is just too damn tedious.
		// So we use offscreen PixMap structures and blast them onto the screen.
		// The mask (clip) region for the images is a round rectangle.

		enum { kRadioImageWidth = 12, kRadioImageHeight = 12 };
		enum { kRadioNormalOff, kRadioNormalOn, kRadioNormalMixed,
				kRadioPressedOff, kRadioPressedOn, kRadioPressedMixed,
				kRadioDisabledOff, kRadioDisabledOn, kRadioDisabledMixed,
				kNumRadioImages };
		
		static SInt8				kRadioImageValues[kNumRadioImages]
													[kRadioImageHeight]
													[kRadioImageWidth];

		static AGAOffscreenImage*	mgOffscreenRadioImages[kNumRadioImages];
	};

#pragma mark AGAScrollBar

//
// AGAScrollBar ---------------------------------------------------------------
//
// AGAScrollBar implements the standard AGA scroll bar. It also supports
// live indicator tracking and a proportional indicator as options.
//

class AGAScrollBar : public AGATrackingIndicator
	{
	public:

		AGAScrollBar(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue);
		AGAScrollBar(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, SInt32 singleStepSize, SInt32 pageStepSize, SInt32 pageSize, Boolean liveTracking, Boolean proportional);
		virtual ~AGAScrollBar();

		virtual void	SetRange(SInt32 newMinimum, SInt32 newMaximum, Boolean redraw);

		virtual void	SetStepSizes(SInt32 singleStepSize, SInt32 pageStepSize);

	protected:

		virtual Boolean	TrackPart(Point mouseLocation, SInt32 partHit);
		virtual SInt32	TrackTestPart(Point mouseLocation);
		virtual Boolean	IsValidIndicatorTrackMouse(Point mouseLocation);
		virtual SInt32	GetValueFromMouseDelta(SInt32 originalValue, Point originalMouseLocation, Point newMouseLocation);

		virtual void	DrawIndicatorTrack();
		virtual void	DrawTrackEnds();
		virtual void	DrawIndicator();
		virtual void	RemoveIndicator(SInt32 newValue);
		virtual void	DrawGhost();
		virtual void	RemoveGhost(SInt32 newValue);
		
		virtual void	TrackDelta(SInt32 partHit);

		virtual void	DrawArrow(SInt32 arrowPart, Boolean pressed);
		
		enum { kIndicatorPixelInset = 8 };
		enum { kArrowSize = 16 };

		virtual void	GetIndicatorBox(SInt32 value, Rect* indicatorBox);
		virtual SInt32	GetIndicatorPixelRange(SInt32* proportionalIndicatorPixels);
		virtual void	GetIndicatorPixelEnds(SInt32* startPt, SInt32* stopPt);
		virtual void	GetIndicatorSpan(SInt32 value, SInt32* startPt, SInt32* valuePt, SInt32* stopPt);
		
		SInt32	mSingleStepSize;	// amount to delta for arrow tracking
		SInt32	mPageStepSize;		// amount to delta for page tracking
		
		// These regions are used by RemoveIndicator and RemoveGhost.
		// They are tiny but we allocate them statically to avoid
		// repeated allocation, which could conceivably fail.
		static RgnHandle	mgSavedIndicatorClip;
		static RgnHandle	mgOldIndicatorClip;
		static RgnHandle	mgNewIndicatorClip;
	};

#pragma mark AGAPopupMenu

//
// AGAPopupMenu ---------------------------------------------------------------
//
// AGAPopupMenu implements the standard AGA pop-up menu. You can specify
// whether the supplied boundary is fixed or whether the pop-up button
// should adapt its width to the width of the associated menu. You can
// also supply a title to be drawn next to the pop-up button.
//

class AGAPopupMenu : public AGAObject
	{
	public:

		enum WidthAdjust { kFixedWidth = -1, kUseMenuWidth = 0, kSystemMDEFAdjustment = 19 };
		enum DisposalKind { kDontDispose, kDispose, kRelease };

		AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment);
		AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment, MenuRef itsMenuRef, DisposalKind menuRefDisposalKind);
		AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment, SInt16 itemsStringListID, SInt16 newMenuID);
		AGAPopupMenu(Rect* bounds, SInt16 titleOffset, StringPtr title, SInt32 titleJustification, const AGATextStyle& titleStyle, const AGATextStyle& textStyle, WidthAdjust adjustment, ResType appendResourceType, SInt16 newMenuID);
		virtual ~AGAPopupMenu();
		
		virtual void	SetMenuRef(MenuRef itsMenuRef, DisposalKind menuRefDisposalKind);
		virtual void	SetWidthAdjustment(WidthAdjust adjustment);
		
		virtual void	DrawObject();
		virtual Boolean	TrackMouse(Point mouseLocation);

		virtual SInt16	GetCurrentItemNo();
		virtual void	SetCurrentItemNo(SInt16 newItemNo, Boolean redraw);
		
		virtual void	GetCurrentItemText(StringPtr itemText);
		virtual Boolean	SetCurrentItemText(StringPtr itemTextToMatch, Boolean redraw);	// returns false if no match found
		virtual void	SetTitle(StringPtr newTitle, Boolean redraw);

		virtual void	DrawButton(Boolean pressed);

	protected:

		virtual void	DrawButtonNormal(Boolean deep);
		virtual void	DrawButtonPressed(Boolean deep);
		virtual void	DrawButtonDisabled(Boolean deep);
		virtual void	GetButtonBounds(Rect* buttonBounds);
		virtual void	GetTitleBounds(Rect* titleBounds);
		
		virtual void	DisposeExistingMenuRef();
		
		enum { kPopupArrowSectionWidth = 22 };

		SInt16			mMenuID;
		MenuRef			mMenuRef;
		DisposalKind	mMenuRefDisposalKind;
		SInt16			mCurrentItemNo;
		WidthAdjust		mWidthAdjust;
		AGATextStyle	mTitleStyle;
		AGATextStyle	mTextStyle;
		Str255			mTitle;
		SInt16			mTitleOffset;
		SInt32			mTitleJustification;
	};

#pragma mark AGACustomPopupMenuSizer

//
// AGACustomPopupMenuSizer -------------------------------------------------------
//
// AGACustomPopupMenuSizer is a stack-based object that will prepare the
// drawing environment for a custom popup menu font/size. This is a
// low-level toolbox hack capability.
// This is used to wrap calls to PopupMenuSelect and CalcMenuSize.
//

class AGACustomPopupMenuSizer
	{
	public:
	
		AGACustomPopupMenuSizer(SInt16 fontNum, SInt16 fontSize);
		virtual ~AGACustomPopupMenuSizer();
		
		void	SaveEnvironment();			// called by constructor
		void	RestoreEnvironment();		// called by destructor
		
		void	InstallCustomEnvironment();	// called by constructor after saving

		Boolean	mIsCustom;

	protected:
	
		SInt16	mSavedFontNum;
		SInt16	mSavedFontSize;
		SInt16	mCustomFontNum;
		SInt16	mCustomFontSize;
	};

#endif // __GCSTDCONTROLS__
