//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// See header file "GCDialogManager.h" for usage.
//

#include "GCDialogManager.h"

#include <LowMem.h>

static UserItemUPP gUserItemUPP = NewUserItemProc(AGADialog::UserItem);
static ModalFilterUPP gModalFilterUPP = NewModalFilterProc(AGADialog::ModalFilter);

static SInt16 AGADoAlert(SInt16 alertID, ModalFilterUPP modalFilter, SInt16 iconID)
	{
	// Run alert using the specified icon ID.

	AGADialog	agaDialog;

	return agaDialog.Alert(alertID, modalFilter, iconID);
	}

SInt16 AGAAlert(SInt16 alertID, ModalFilterUPP modalFilter)
	{
	// Run the alert without an icon.

	return AGADoAlert(alertID, modalFilter, -1);
	}

SInt16 AGAStopAlert(SInt16 alertID, ModalFilterUPP modalFilter)
	{
	// Run the alert using the Stop icon.

	return AGADoAlert(alertID, modalFilter, stopIcon);
	}

SInt16 AGANoteAlert(SInt16 alertID, ModalFilterUPP modalFilter)
	{
	// Run the alert using the Note icon.

	return AGADoAlert(alertID, modalFilter, noteIcon);
	}

SInt16 AGACautionAlert(SInt16 alertID, ModalFilterUPP modalFilter)
	{
	// Run the alert using the Caution icon.

	return AGADoAlert(alertID, modalFilter, cautionIcon);
	}

//
// AGADialog --------------------------------------------------------------
//

AGADialog::AGADialog()
	{
	// Construct with defaults.

	mMarker = kAGADialogMarker;
	mIsAlert = false;
	mDialogPtr = NULL;
	mNumAGAObjects = 0;
	mAGAObjects = (CustomObjectInfoHandle) ::NewHandle(0);
	mUserFilter = NULL;
	mDefaultItemNo = -1;
	mCancelItemNo = -1;
	mDefaultButton = NULL;
	mCancelButton = NULL;
	}

AGADialog::~AGADialog()
	{
	// Destruct by freeing extra memory we allocated.

	this->DestroyAllCustomObjects();
	
	if (mAGAObjects != NULL)
		::DisposeHandle((Handle) mAGAObjects);
	}

DialogPtr AGADialog::GetNewDialog(SInt16 dialogID, Ptr dStorage, WindowPtr behind)
	{
	// Wrapper for GetNewDialog.
	//
	// We get the DLOG resource to get the DITL ID and positioning info.
	// We make our own copy of the DITL data to give to the dialog,
	// create a color dialog, position it, and then install our custom
	// AGA objects.
	
	UInt8		savedHandleState;
	DialogTHndl	dlogHandle;
	Handle		ditlHandle = NULL;
	OSErr		err = memFullErr;	// assume the worst
	Rect		bounds;
	
	dlogHandle = (DialogTHndl) ::GetResource('DLOG', dialogID);

	if (dlogHandle != NULL)
		{
		// Release the DLOG and re-load it, to make sure we get a fresh copy.
		// If previously loaded by the DM, the bounds may have been wacked.
		::ReleaseResource((Handle) dlogHandle);
		dlogHandle = (DialogTHndl) ::GetResource('DLOG', dialogID);
		}

	if (dlogHandle != NULL)
		{
		savedHandleState = ::HGetState((Handle) dlogHandle);
		::HNoPurge((Handle) dlogHandle);
		ditlHandle = ::GetResource('DITL', (**dlogHandle).itemsID);
		}

	if (ditlHandle != NULL)
		{
		::HandToHand(&ditlHandle);
		err = ::MemError();
		}

	if (err != noErr)	// punt if DLOG or DITL not gotten or DITL not duplicated
		mDialogPtr = ::GetNewDialog(dialogID, dStorage, behind);
	else
		{
		Str255	title;
		Boolean	visible = (**dlogHandle).visible;
		SInt16	procID = (**dlogHandle).procID;
		Boolean	goAwayFlag = (**dlogHandle).goAwayFlag;
		SInt32	refCon = (**dlogHandle).refCon;

		AGA_PLstrcpy(title, (**dlogHandle).title);
		
		bounds = (**dlogHandle).boundsRect;
		mDialogPtr = ::NewColorDialog(dStorage, &bounds, title, visible, procID, behind, goAwayFlag, refCon, ditlHandle);
		}

	if (dlogHandle != NULL)
		::HSetState((Handle) dlogHandle, savedHandleState);

	this->AdjustDialogPosition(dlogHandle, &bounds);
	this->BeginAGADialog();
	
	return mDialogPtr;
	}

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

typedef struct
	{
	UInt32	reserved1;
	Rect	itemRect;
	UInt8	itemType;
	UInt8	reserved2;
	SInt16	iconID;
	} IconItemData, *IconItemPtr;

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=reset
#endif

SInt16 AGADialog::Alert(SInt16 alertID, ModalFilterUPP modalFilter, SInt16 iconID)
	{
	// Wrapper for Alert and its variants.
	//
	// We get the ALRT resource to get the DITL ID and positioning info.
	// We make our own copy of the DITL data to give to the dialog,
	// create a color dialog, position it, and then install our custom
	// AGA objects.
	// Then we run the modal dialog right here before returning.
	
	SInt16		itemHit = -1;
	UInt8		savedHandleState;
	AlertTHndl	alrtHandle;
	Handle		ditlHandle = NULL;
	OSErr		err = memFullErr;	// assume the worst
	
	mIsAlert = true;

	alrtHandle = (AlertTHndl) ::GetResource('ALRT', alertID);

	if (alrtHandle != NULL)
		{
		// Release the ALRT and re-load it, to make sure we get a fresh copy.
		// If previously loaded by the DM, the bounds may have been wacked.
		::ReleaseResource((Handle) alrtHandle);
		alrtHandle = (AlertTHndl) ::GetResource('ALRT', alertID);
		}

	if (alrtHandle != NULL)
		{
		savedHandleState = ::HGetState((Handle) alrtHandle);
		::HNoPurge((Handle) alrtHandle);
		
		// !!! On my machine the Dialog Manager seems to ignore the CANCEL
		// flag in the alert stage info. To get identical behavior, I guess
		// I'll do the same. Need to investigate further, this is weird...
		
		// Peek into the ALRT stage info, which is 4 bytes of bit settings.
		//if (((**alrtHandle).stages & 0x0008) == 0)
			mDefaultItemNo = ok;
		//else
		//	mDefaultItemNo = cancel;
	
		// Read the DITL resource.
		ditlHandle = ::GetResource('DITL', (**alrtHandle).itemsID);
		}

	if (ditlHandle != NULL)
		{
		::HandToHand(&ditlHandle);
		err = ::MemError();
		}

	if ((err == noErr) && (iconID != -1))
		{
		// We need to expand the DITL handle and append a
		// disabled icon item to it.
		
		SInt32	oldHandleSize = ::GetHandleSize(ditlHandle);
		SInt32	newHandleSize = oldHandleSize + sizeof(IconItemData);
		
		::SetHandleSize(ditlHandle, newHandleSize);
		
		err = ::MemError();
		
		if (err == noErr)
			{
			(*((SInt16*) *ditlHandle))++;

			IconItemPtr	iconItemPtr = (IconItemPtr) ((*ditlHandle) + oldHandleSize);
			
			iconItemPtr->reserved1 = 0;
			iconItemPtr->itemRect.left = 18;
			iconItemPtr->itemRect.top = 9;
			iconItemPtr->itemRect.right = 50;
			iconItemPtr->itemRect.bottom = 41;
			iconItemPtr->itemType = kIconDialogItem | itemDisable;
			iconItemPtr->reserved2 = 2;
			iconItemPtr->iconID = iconID;
			}
		}

	if (err != noErr)	// punt if ALRT or DITL not gotten or DITL not duplicated
		{
		switch (iconID)
			{
			case -1:
				itemHit = ::Alert(alertID, modalFilter); 
				break;
			case stopIcon:
				itemHit = ::StopAlert(alertID, modalFilter); 
				break;
			case noteIcon:
				itemHit = ::NoteAlert(alertID, modalFilter); 
				break;
			case cautionIcon:
				itemHit = ::CautionAlert(alertID, modalFilter); 
				break;
			}
		}
	else
		{
		Rect	bounds = (**alrtHandle).boundsRect;

		mDialogPtr = ::NewColorDialog(NULL, &bounds, "\p", false /*visible*/, dBoxProc, (WindowPtr) -1, false /*goAwayFlag*/, 0 /*refCon*/, ditlHandle);

		this->AdjustAlertPosition(alrtHandle, &bounds);
		this->BeginAGADialog();
		
		::ShowWindow(mDialogPtr);
		
		while (itemHit == -1)
			{
			this->ModalDialog(modalFilter, &itemHit);
			
			if (! this->IsDITLItemEnabled(itemHit))
				itemHit = -1;
			}
		
		this->DisposeDialog();
		}

	if (alrtHandle != NULL)
		::HSetState((Handle) alrtHandle, savedHandleState);
	
	return itemHit;
	}

void AGADialog::BeginAGADialog()
	{
	// Called after the dialog and the standard items have been created
	// in memory, but before it has been altered by us and run.
	// We install a gray background and then create our objects.

	if (mDialogPtr == NULL)
		return;

	::SetWRefCon(mDialogPtr, (SInt32) this);

	(void) AGAInstallGrayWCTB(mDialogPtr);

	this->CreateAllCustomObjects();
	}

void AGADialog::CloseDialog()
	{
	// Clean up after ourselves before closing the dialog.

	this->DestroyAllCustomObjects();

	::CloseDialog(mDialogPtr);
	
	mDialogPtr = NULL;
	}

void AGADialog::DisposeDialog()
	{
	// Clean up after ourselves before disposing the dialog.

	this->DestroyAllCustomObjects();

	::DisposeDialog(mDialogPtr);
	
	mDialogPtr = NULL;
	}

void AGADialog::GetDialogItem(SInt16 itemNo, SInt16* itemType, Handle* item, Rect* box)
	{
	// This wrapper is not strictly needed. It just calls through.

	::GetDialogItem(mDialogPtr, itemNo, itemType, item, box);
	}

void AGADialog::SetDialogItem(SInt16 itemNo, SInt16 itemType, Handle item, Rect* box)
	{
	// Need to find out how people are using this. About all we can
	// do is set the corresponding custom object's bounds to match.
	// Not sure whether we should call though for replaced items --
	// might this un-hide the original item?

	::SetDialogItem(mDialogPtr, itemNo, itemType, item, box);

	AGAObject*	anAGAObject = this->GetAGAObjectPtr(itemNo);
	
	if (anAGAObject != NULL)
		anAGAObject->SetObjectBounds(box, AGAObject::kDontRedraw);
	}

void AGADialog::HideDialogItem(SInt16 itemNo)
	{
	// The base item is already hidden. Hide the custom item.

	if (this->IsCustomObject(itemNo))
		::HideDialogItem(mDialogPtr, itemNo + mNumAGAObjects);
	else
		::HideDialogItem(mDialogPtr, itemNo);
	}

void AGADialog::ShowDialogItem(SInt16 itemNo)
	{
	// The base item must remain hidden. Show the custom item.

	if (this->IsCustomObject(itemNo))
		::ShowDialogItem(mDialogPtr, itemNo + mNumAGAObjects);
	else
		::ShowDialogItem(mDialogPtr, itemNo);
	}

SInt16 AGADialog::FindDialogItem(Point thePt)
	{
	// Note: FindDialogItem is zero-based, unlike all other dialog itemNo usage.
	SInt16	result = ::FindDialogItem(mDialogPtr, thePt);
	
	if (result > mNumAGAObjects)
		result -= mNumAGAObjects;

	return result;
	}

void AGADialog::AppendDITL(Handle theDITL, DITLMethod theMethod)
	{
	// If the client is using AppendDITL to mess with the item list,
	// we'll need to respect their order. All we can do is try to
	// pretend we're not here, and then rebuild ourself to fit the
	// new items. If this doesn't work for you, let me know so we
	// can see if there are ways to make this work transparently.

	this->DestroyAllCustomObjects();
	
	::AppendDITL(mDialogPtr, theDITL, theMethod);
	
	this->CreateAllCustomObjects();
	}

void AGADialog::ShortenDITL(SInt16 numberOfItems)
	{
	// If the client is using ShortenDITL to mess with the item list,
	// we'll need to respect their order. All we can do is try to
	// pretend we're not here, and then rebuild ourself to fit the
	// new items. If this doesn't work for you, let me know so we
	// can see if there are ways to make this work transparently.

	this->DestroyAllCustomObjects();
	
	::ShortenDITL(mDialogPtr, numberOfItems);
	
	this->CreateAllCustomObjects();
	}

SInt16 AGADialog::CountDITL()
	{
	// Adjust the count so we are invisible.

	SInt16	result = ::CountDITL(mDialogPtr);
	
	if (result > mNumAGAObjects)
		result -= mNumAGAObjects;

	return result;
	}

OSErr AGADialog::SetDialogDefaultItem(SInt16 newItem)
	{
	// Remove default outline from current default button, if any.
	
	if (mDefaultButton != NULL)
		{
		mDefaultButton->SetDefault(AGAObject::kIsNotDefault, AGAObject::kFrameOutside);
		mDefaultButton = NULL;
		}

	// Add default outline to new default button, if any.

	mDefaultItemNo = newItem;
	
	mDefaultButton = this->GetVerifiedPushButton(mDefaultItemNo);
	
	if (mDefaultButton != NULL)
		mDefaultButton->SetDefault(AGAObject::kIsDefault, AGAObject::kFrameOutside);

	return ::SetDialogDefaultItem(mDialogPtr, newItem);
	}

OSErr AGADialog::SetDialogCancelItem(SInt16 newItem)
	{
	// Set the cancel item so we can catch the escape key in our filter.

	mCancelItemNo = newItem;
	
	mCancelButton = this->GetVerifiedPushButton(mCancelItemNo);

	return ::SetDialogCancelItem(mDialogPtr, newItem);
	}

void AGADialog::GetDialogItemText(Handle item, Str255 text)
	{
	// This wrapper is not strictly needed. It just calls through.

	::GetDialogItemText(item, text);
	}

void AGADialog::SetDialogItemText(Handle item, ConstStr255Param text)
	{
	// If it's an editText item, the original object handles it.
	// If it's a statText item, the original is hidden (but should
	// get the assigned text, so that GetDialogItemText will just
	// work) and ours must handle it.

	::SetDialogItemText(item, text);
	
	if (this->GetDITLItemType(this->GetDITLItemNoFromHandle(item)) == statText)
		{
		AGAStaticText*	theStaticText = (AGAStaticText*) this->GetAGAObjectPtrFromHandle(item);
	
		if (theStaticText != NULL)
			theStaticText->SetTitle((StringPtr) text, AGAObject::kRedraw);
		}
	}

void AGADialog::SelectDialogItemText(SInt16 itemNo, SInt16 strtSel, SInt16 endSel)
	{
	// This wrapper is not strictly needed. It just calls through.

	::SelectDialogItemText(mDialogPtr, itemNo, strtSel, endSel);
	}

void AGADialog::ModalDialog(ModalFilterUPP filterProc, SInt16* itemHit)
	{
	// Wrap ModalDialog by saving aside the client filterProc
	// (which will be used by our filter) and afterwards adjusting
	// the itemHit if one of our custom objects was hit.

	mUserFilter = filterProc;

	::ModalDialog(gModalFilterUPP, itemHit);

	mUserFilter = NULL;
	
	if (*itemHit > mNumAGAObjects)
		(*itemHit) -= mNumAGAObjects;
	}

Boolean AGADialog::DialogSelect(EventRecord* event, SInt16* itemHit)
	{
	// Wrapper for modeless dialog DialogSelect.
	// Unlike modal dialogs and alerts, we have to handle update
	// events by calling DrawDialog.
	// Note that the case of multiple modeless dialogs has not been
	// totally worked out. Normally DialogSelect is called without
	// regard to which dialog is going to get the event; in fact, you
	// don't know which window is going to get the event! But this
	// is a method of an object that wraps a particular dialog.
	//

	Boolean	wasEventHandled = false;

	switch (event->what)
		{
		case updateEvt:
		
			if (AGADialog::ObjectFromDialogPtr((WindowPtr) (event->message)) == this)
				{
				::BeginUpdate(mDialogPtr);
				this->DrawDialog();
				::EndUpdate(mDialogPtr);

				wasEventHandled = true;
				*itemHit = -1;
				}

			break;
		}

	if (! wasEventHandled)
		wasEventHandled = this->HandleModalFilter(event, itemHit);
	
	return wasEventHandled;
	}

void AGADialog::DrawDialog()
	{
	// This wrapper is not strictly needed. It just calls through.

	::DrawDialog(mDialogPtr);
	}

void AGADialog::UpdateDialog(RgnHandle updateRgn)
	{
	// This wrapper is not strictly needed. It just calls through.

	::UpdateDialog(mDialogPtr, updateRgn);
	}

void AGADialog::SetControlValue(ControlHandle theControl, SInt16 theValue)
	{
	// We set the real object's value, but also have to see if we have
	// an AGA replacement object to set.

	::SetControlValue(theControl, theValue);

	AGAObject*		theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	AGAPortSetter	port(mDialogPtr);

	if (theObject != NULL)
		theObject->SetValue(theValue, AGAObject::kRedraw);
	}

void AGADialog::SetControlMinimum(ControlHandle theControl, SInt16 minValue)
	{
	// We don't yet support AGA scroll bars in the AGADialog.

	::SetControlMinimum(theControl, minValue);

	//AGAObject*	theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	
	//	if (theObject != NULL)
	//		theObject->SetMinValue(minValue, AGAObject::kRedraw);
	}

void AGADialog::SetControlMaximum(ControlHandle theControl, SInt16 maxValue)
	{
	// We don't yet support AGA scroll bars in the AGADialog.

	::SetControlMaximum(theControl, maxValue);

	AGAObject*	theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	
	//	if (theObject != NULL)
	//		theObject->SetMaxValue(maxValue, AGAObject::kRedraw);
	}

void AGADialog::SetControlTitle(ControlHandle theControl, ConstStr255Param title)
	{
	// We don't yet support changing AGA button titles in the AGADialog.
	// In the meantime, you can do it yourself if you know the itemNo
	// and the object type.

	::SetControlTitle(theControl, title);

	//AGAObject*	theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	
	//	if (theObject != NULL)
	//		theObject->SetTitle(title, AGAObject::kRedraw);
	}

void AGADialog::HideControl(ControlHandle theControl)
	{
	// Need to add a visibility flag to the core AGAObject.

	::HideControl(theControl);

	//AGAObject*	theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	
	//	if (theObject != NULL)
	//		::HideDialogItem(itemNo + mNumAGAObjects);
	}

void AGADialog::MoveControl(ControlHandle theControl, SInt16 h, SInt16 v)
	{
	// Move the real control, and our replacement, if any.

	::MoveControl(theControl, h, v);

	AGAObject*	theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	
	if (theObject != NULL)
		{
		Rect	bounds;
		
		theObject->GetObjectBounds(&bounds);
		
		::OffsetRect(&bounds, h, v);
		
		theObject->SetObjectBounds(&bounds, AGAObject::kDontRedraw);
		}
	}

void AGADialog::SizeControl(ControlHandle theControl, SInt16 width, SInt16 height)
	{
	// Resize the real control, and our replacement, if any.

	::SizeControl(theControl, width, height);

	AGAObject*	theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	
	if (theObject != NULL)
		{
		Rect	bounds;
		
		theObject->GetObjectBounds(&bounds);
		
		bounds.right = bounds.left + width;
		bounds.bottom = bounds.top + height;
		
		theObject->SetObjectBounds(&bounds, AGAObject::kDontRedraw);
		}
	}

void AGADialog::HiliteControl(ControlHandle theControl, SInt16 hiliteState)
	{
	// If the control being highlighted is one we are replacing,
	// this routine is problematic. Ideally we would just highlight
	// the replacement, but we need to find out the object class
	// and do the appropriate action. Usually HiliteControl is just
	// called as part of trying to flash a button, and we have a
	// FlashButton method.

	::HiliteControl(theControl, hiliteState);

	//AGAObject*	theObject = this->GetAGAObjectPtrFromControlHandle(theControl);
	
	//	if (theObject != NULL)
	//		theObject->Enable/Dim/Flash/Etc.(...);
	}

// These constants are not defined in the Universal Headers!
// Had to swipe from Types.r and make an enum.

enum { noAutoCenter = 0x0000,
		centerMainScreen = 0x280a,
		alertPositionMainScreen = 0x300a,
		staggerMainScreen = 0x380a,
		centerParentWindow = 0xa80a,
		alertPositionParentWindow = 0xb00a,
		staggerParentWindow = 0xb80a,
		centerParentWindowScreen = 0x680a,
		alertPositionParentWindowScreen = 0x700a,
		staggerParentWindowScreen = 0x780a };

void AGADialog::AdjustAlertPosition(AlertTHndl alrtHandle, const Rect* origBounds)
	{
	SInt16	positionSpecifier = noAutoCenter;
	SInt32	alrtSize = ::GetHandleSize((Handle) alrtHandle);
	
	if (alrtSize > sizeof(AlertTemplate))
		{
		// This alert has the optional position specifier at
		// the end of the normal ALRT structure. We do some
		// skanky pointer arithmetic and typecasting to get
		// a pointer to that extra short past the standard
		// struct, and then dereference it.
		
		positionSpecifier = *((SInt16*) ((*alrtHandle) + 1));
		}

	this->AdjustWindowPosition(positionSpecifier, origBounds);
	}

void AGADialog::AdjustDialogPosition(DialogTHndl dlogHandle, const Rect* origBounds)
	{
	SInt16	positionSpecifier = noAutoCenter;
	SInt32	dlogActualSize = ::GetHandleSize((Handle) dlogHandle);
	SInt32	dlogStdDataSize;
	
	// Calculate how big this DLOG would really be if didn't have
	// the positioner at the end. This depends on the title string
	// length, and an optional alignment byte.
	dlogStdDataSize = sizeof(DialogTemplate) - sizeof(Str255) + 1;
	dlogStdDataSize += (**dlogHandle).title[0];
	
	if (dlogStdDataSize & 0x00000001)	// odd size, add alignment byte
		dlogStdDataSize++;
	
	if (dlogActualSize > dlogStdDataSize)
		{
		// This dialog has the optional position specifier at
		// the end of the normal DLOG data. We do some
		// skanky pointer arithmetic and typecasting to get
		// a pointer to that extra short past the standard
		// data, and then dereference it.
		
		positionSpecifier = *((SInt16*) (((Ptr) (*dlogHandle)) + dlogStdDataSize));
		}

	this->AdjustWindowPosition(positionSpecifier, origBounds);
	}

static void CenterRectOnPoint(Rect* rect, Point itsCenter)
	{
	Point	rectSize;

	::SetPt(&rectSize, rect->right - rect->left, rect->bottom - rect->top);

	rect->left = itsCenter.h - (rectSize.h / 2);
	rect->top = itsCenter.v - (rectSize.v / 2);
	rect->right = rect->left + rectSize.h;
	rect->bottom = rect->top + rectSize.v;
	}

void AGADialog::AdjustWindowPosition(SInt16 positionSpecifier, const Rect* origBounds)
	{
	// Man, why doesn't the Dialog Manager have a function call to do this
	// so we don't have to try and approximate its behavior?
	
	Boolean		adjustIt = true;
	Rect		targetRect;	// the bounds in which we must center the window
	Rect		windowRect = *origBounds;
	GDHandle	targetDevice;
	WindowPeek	parentWindow;
	Point		targetCenter;
	SInt16		targetExcessV;
	
	switch (positionSpecifier)
		{
		case centerMainScreen:
		case alertPositionMainScreen:
		case staggerMainScreen:
			// Use the main monitor as the general target area.

			targetDevice = ::GetMainDevice();
			targetRect = (**targetDevice).gdRect;
			break;
		case centerParentWindow:
		case alertPositionParentWindow:
		case staggerParentWindow:
			// Use the front window as the general target area.

			parentWindow = (WindowPeek) ::FrontWindow();
			if (parentWindow == NULL)
				adjustIt = false;
			else
				targetRect = (**(parentWindow->contRgn)).rgnBBox;
			break;
		case centerParentWindowScreen:
		case alertPositionParentWindowScreen:
		case staggerParentWindowScreen:
			// Use the front window's monitor as the general target area.

			parentWindow = (WindowPeek) ::FrontWindow();
			if (parentWindow == NULL)
				adjustIt = false;
			else
				{
				Boolean	found = false;
				Rect	parentWindowBounds = (**(parentWindow->contRgn)).rgnBBox;
				Point	parentWindowCenter;
				
				targetRect = (**(parentWindow->contRgn)).rgnBBox;
			
				::SetPt(&parentWindowCenter,
					(parentWindowBounds.left + parentWindowBounds.right) / 2,
					(parentWindowBounds.top + parentWindowBounds.bottom) / 2);

				// Find the device that contains the window's center.
				// Not as precise as calculating the max intersected device,
				// but good enough.
				
				targetDevice = ::GetDeviceList();
				
				while (targetDevice != NULL)
					{
					if (::PtInRect(parentWindowCenter, &(**targetDevice).gdRect))
						{
						targetRect = (**targetDevice).gdRect;
						
						if (targetDevice == ::GetMainDevice())
							targetRect.top += LMGetMBarHeight();

						break;
						}
					
					targetDevice = ::GetNextDevice(targetDevice);
					}
				}

			break;
		}

	if (adjustIt)
		{
		::SetPt(&targetCenter,
			(targetRect.left + targetRect.right) / 2,
			(targetRect.top + targetRect.bottom) / 2);

		switch (positionSpecifier)
			{
			case centerMainScreen:
			case centerParentWindow:
			case centerParentWindowScreen:
			
				CenterRectOnPoint(&windowRect, targetCenter);

				break;

			case alertPositionMainScreen:
			case alertPositionParentWindow:
			case alertPositionParentWindowScreen:
			
				// If target is smaller than window (no "excess"),
				// just center. Otherwise, center on upper area.

				targetExcessV = (targetRect.bottom - targetRect.top) -
								(windowRect.bottom - windowRect.top);

				if (targetExcessV > 0)
					targetCenter.v -= (targetExcessV / 4);	// center at top 1/4

				CenterRectOnPoint(&windowRect, targetCenter);

				break;

			case staggerMainScreen:
			case staggerParentWindow:
			case staggerParentWindowScreen:
				adjustIt = false;
				// Hmmm... the system's staggering info is not documented.
				break;
			}
		}

	if (adjustIt)
		{
		// Before moving the window, make sure all 4 corners are somewhere
		// on screen. If not, center it on the main screen.

		RgnHandle	grayRegion = ::GetGrayRgn();
		Point		wTopLeft;
		Point		wTopRight;
		Point		wBotLeft;
		Point		wBotRight;

		::SetPt(&wTopLeft, windowRect.left, windowRect.top);
		::SetPt(&wTopRight, windowRect.right, windowRect.top);
		::SetPt(&wBotLeft, windowRect.left, windowRect.bottom);
		::SetPt(&wBotRight, windowRect.right, windowRect.bottom);

		if ((! ::PtInRgn(wTopLeft, grayRegion)) || (! ::PtInRgn(wTopRight, grayRegion)) ||
			(! ::PtInRgn(wBotLeft, grayRegion)) || (! ::PtInRgn(wBotRight, grayRegion)))
			{
			targetRect = (**::GetMainDevice()).gdRect;

			::SetPt(&targetCenter,
				(targetRect.left + targetRect.right) / 2,
				(targetRect.top + targetRect.bottom) / 2);

			CenterRectOnPoint(&windowRect, targetCenter);
			}

		::MoveWindow(mDialogPtr, windowRect.left, windowRect.top, false);
		}
	}

void AGADialog::CreateAllCustomObjects()
	{
	// This and BuildCustomObject are the two core functions that
	// perform most of our Patented Dialog Manager Grayscale Magic.
	//
	// We need to construct an n item DITL manually in memory.
	// It starts with a 2-byte value that is the n-1.
	// It is followed by the n variable-length items.
	// In our case every item is in fact a userItem (14 bytes).
	// See IM:TBE p. 6-152 and p. 6-155 for details.
	//
	// Once we've built this DITL and the AGAObject objects that
	// will replace certain of the original items (the originals
	// are simply hidden), we append it to the user's standard DITL
	// that came from the DITL resource.

	OSErr	err = noErr;
	Handle	customItemsDITL;
	Ptr		ditlDataPtr;
	SInt16	numOriginalItems = ::CountDITL(mDialogPtr);
	
	//
	// Allocate a handle to contain our duplicate DITL items.
	// Each one is a userItem (14 bytes).
	//

	customItemsDITL = ::NewHandleClear(2 + (numOriginalItems * 14));

	if (customItemsDITL == NULL)
		err = memFullErr;

	//
	// Size our mAGAObjects handle to make room for n AGAObject pointers.
	//

	if (err == noErr)
		{
		::SetHandleSize((Handle) mAGAObjects, numOriginalItems * sizeof(CustomObjectInfo));
		err = ::MemError();
		}
	
	if (err == noErr)
		mNumAGAObjects = numOriginalItems;
	else
		mNumAGAObjects = 0;

	//
	// Create the replacement objects and store their pointers in the handle.
	//

	{
	AGAHandleLocker	locker(((DialogPeek) mDialogPtr)->items);
	Ptr				ditlItemPtr = *(((DialogPeek) mDialogPtr)->items);
	
	ditlItemPtr += 2;	// skip past "count minus 1" integer

	for (SInt16 i = 1; i <= mNumAGAObjects; i++)
		this->CreateCustomObject(i, &ditlItemPtr);
	}
	
	if (customItemsDITL != NULL)
		{
		::HLock(customItemsDITL);

		ditlDataPtr = *customItemsDITL;
		*((SInt16*) ditlDataPtr) = mNumAGAObjects - 1;	// set the item count field (n-1 actually)
		ditlDataPtr += 2;								// move past item count field

		for (SInt16 i = 1; i <= mNumAGAObjects; i++)
			{
			SInt16		itemType;
			Handle		item;
			Rect		box;
			UInt8		ditlItemType = userItem;

			::GetDialogItem(mDialogPtr, i, &itemType, &item, &box);
			
			if ((itemType & itemDisable) != 0)
				ditlItemType |= itemDisable;
			
			// Since these items are all "hidden", and since the Dialog Manager
			// hides items by moving them, we need to adjust to get the real
			// box where our item is to be located. See IM:TBE p. 6-123.
			if (box.left > 8192)
				::OffsetRect(&box, -16384, 0);
			
			// OK, we're touching the DITL byte layout directly here!
			// We know this is not nice, but it's the only way!

			ditlDataPtr += 4;						// move past reserved field
			
			*((Rect*) ditlDataPtr) = box;			// set the display rectangle field
			ditlDataPtr += 8;						// move past display rectangle field
			
			*((UInt8*) ditlDataPtr) = ditlItemType;	// set the item type / enable field
			ditlDataPtr += 2;						// move past item type and reserved fields
			}

		::HUnlock(customItemsDITL);
		
		::AppendDITL(mDialogPtr, customItemsDITL, overlayDITL);
		::DisposeHandle(customItemsDITL);

		// Now install our userItem proc for each of our objects.
		for (SInt16 i = 1; i <= mNumAGAObjects; i++)
			{
			SInt16		itemType;
			Handle		item;
			Rect		box;

			::GetDialogItem(mDialogPtr, i + mNumAGAObjects, &itemType, &item, &box);
			::SetDialogItem(mDialogPtr, i + mNumAGAObjects, itemType, (Handle) gUserItemUPP, &box);
			}
		}
	}

void AGADialog::DestroyAllCustomObjects()
	{
	// Free the AGAObject associated with each custom userItem.

	for (SInt16 i = 1; i <= mNumAGAObjects; i++)
		this->DestroyCustomObject(i);

	mNumAGAObjects = 0;
	}

void AGADialog::CreateCustomObject(SInt16 itemIndex, Ptr* ditlItemPtr)
	{
	// Create an AGAObject of the appropriate class, if any, for the
	// specified original DITL item. Hide the original if we created
	// an AGAObject.

	Boolean				hideOriginal;
	AGAObject*			theAGAObject = this->BuildCustomObject(itemIndex, &hideOriginal, ditlItemPtr);
	CustomObjectInfoPtr	infoPtr = this->GetCustomObjectInfoPtr(itemIndex);
	
	if (infoPtr != NULL)
		{
		infoPtr->mAGAObject = theAGAObject;
		infoPtr->mTracksMouse = hideOriginal;
		}

	if (hideOriginal)
		::HideDialogItem(mDialogPtr, itemIndex);
	}

void AGADialog::DestroyCustomObject(SInt16 itemIndex)
	{
	// Just delete the AGAObject if there is one for this custom userItem.

	CustomObjectInfoPtr	infoPtr = this->GetCustomObjectInfoPtr(itemIndex);
	
	if (infoPtr->mAGAObject != NULL)
		{
		delete infoPtr->mAGAObject;
		infoPtr->mAGAObject = NULL;
		}
	}

Boolean AGADialog::IsCustomObject(SInt16 itemIndex)
	{
	// Return true if there's an AGAObject for this custom userItem.

	CustomObjectInfoPtr	infoPtr = this->GetCustomObjectInfoPtr(itemIndex);
	
	return (infoPtr->mAGAObject != NULL);
	}

CustomObjectInfoPtr AGADialog::GetCustomObjectInfoPtr(SInt16 itemIndex)
	{
	// Return a pointer to the CustomObjectInfo entry for this userItem.
	// Warning: this is a pointer into a relocatable block; use it briefly.

	if ((itemIndex > 0) && (itemIndex <= mNumAGAObjects))
		return &((*mAGAObjects)[itemIndex - 1]);
	else
		return NULL;
	}

AGAObject* AGADialog::GetAGAObjectPtr(SInt16 itemIndex)
	{
	// Return the AGAObject pointer for this custom item, or NULL if
	// it has no custom object.

	CustomObjectInfoPtr	infoPtr = this->GetCustomObjectInfoPtr(itemIndex);
	
	if (infoPtr == NULL)
		return NULL;
	else
		return infoPtr->mAGAObject;
	}

AGAObject* AGADialog::GetAGAObjectPtrFromHandle(Handle itemHandle)
	{
	// Get the AGAObject, if any, that corresponds to the supplied
	// original item handle.

	AGAObject*	result = NULL;
	SInt16		itemType;
	Handle		item;
	Rect		box;

	for (SInt16 i = 1; i <= mNumAGAObjects; i++)
		{
		::GetDialogItem(mDialogPtr, i, &itemType, &item, &box);
		
		if (item == itemHandle)
			{
			result = this->GetAGAObjectPtr(i);
			break;
			}
		}

	return result;
	}

SInt16 AGADialog::GetDITLItemNoFromHandle(Handle itemHandle)
	{
	// Get the original DITL item number for the supplied item handle.

	SInt16	result = 0;
	SInt16	itemType;
	Handle	item;
	Rect	box;

	for (SInt16 i = 1; i <= mNumAGAObjects; i++)
		{
		::GetDialogItem(mDialogPtr, i, &itemType, &item, &box);
		
		if (item == itemHandle)
			{
			result = i;
			break;
			}
		}

	return result;
	}

AGAObject* AGADialog::GetAGAObjectPtrFromControlHandle(ControlHandle theControl)
	{
	// Get the AGAObject, if any, that corresponds to the supplied
	// control handle.

	return this->GetAGAObjectPtrFromHandle((Handle) theControl);
	}

AGAObject* AGADialog::BuildCustomObject(SInt16 itemIndex, Boolean* hideOriginal, Ptr* ditlItemPtr)
	{
	// This and CreateAllCustomObjects are the two core functions that
	// perform most of our Patented Dialog Manager Grayscale Magic.
	//
	// We use the original item's itemType to determine whether and which
	// type of AGAObject will replace the original. In certain cases
	// (such as editText items) we keep the original but add an object
	// to enhance it (an editText gets a framing object).

	AGAObject*	anAGAObject = NULL;
	SInt16		itemType;
	Handle		item;
	Rect		box;
	Str255		title;
	Boolean		enabled = true;
	Ptr			aPtr;
	
	*hideOriginal = true;

	::GetDialogItem(mDialogPtr, itemIndex, &itemType, &item, &box);
	
	if ((itemType & itemDisable) != 0)
		{
		enabled = false;
		itemType &= ~itemDisable;
		}
	
	// Get any item-type-specific information ready.
	switch (itemType)
		{
		case kButtonDialogItem:
		case kCheckBoxDialogItem:
		case kRadioButtonDialogItem:
			::GetControlTitle((ControlHandle) item, title);
			break;
		case kStaticTextDialogItem:
		case kEditTextDialogItem:
			::GetDialogItemText(item, title);
			break;
		}

	// Create the appropriate custom object.
	switch (itemType)
		{
		case kButtonDialogItem:

			if (gAGAStdControls)
				{
				anAGAObject = new AGAPushButton(&box, gAGAStdSystemStyle, title);
				
				if (itemIndex == mDefaultItemNo)
					{
					((AGAPushButton*) anAGAObject)->SetDefault(AGAObject::kIsDefault, AGAObject::kFrameOutside);
					mDefaultButton = (AGAPushButton*) anAGAObject;
					}
				else if (itemIndex == mCancelItemNo)
					mCancelButton = (AGAPushButton*) anAGAObject;
				}
			else
				*hideOriginal = false;

			break;
		case kCheckBoxDialogItem:

			if (gAGAStdControls)
				anAGAObject = new AGACheckBox(&box, gAGAStdSystemStyle, title, AGAObject::kAutomaticState);
			else
				*hideOriginal = false;

			break;
		case kRadioButtonDialogItem:

			if (gAGAStdControls)
				anAGAObject = new AGARadioButton(&box, gAGAStdSystemStyle, title, kNoGroupID, kNoGroupID, AGAObject::kNoAutomaticState);
			else
				*hideOriginal = false;

			break;
		case kStaticTextDialogItem:
			this->DoParamTextSubstitution(title);
			anAGAObject = new AGAStaticText(&box, gAGAStdSystemStyle, teFlushDefault, title);
			break;
		case kEditTextDialogItem:
			::InsetRect(&box, -4, -4);
			anAGAObject = new AGAEditTextFrame(&box);
			*hideOriginal = false;
			break;
		case kIconDialogItem:
			aPtr = (*ditlItemPtr) + 14;	// point to the icon id field of the item
			anAGAObject = new AGADialogIcon(&box, *((SInt16*) aPtr));
			break;
		default:
			*hideOriginal = false;
			break;
		}

	// Increment the ditlItemPtr to the next item, based on the size of
	// this item element.
	this->GetNextDITLItemPtr(itemType, ditlItemPtr);

	return anAGAObject;
	}

void AGADialog::GetNextDITLItemPtr(SInt16 itemType, Ptr* ditlItemPtr)
	{
	UInt32	offset;
	UInt8	titleLength;

	// Increment the ditl item pointer past the current item data.
	// See Inside Macintosh: Toolbox Essentials pp. 6-153 thru 6-155
	// for the ditl item layouts used here.

	switch (itemType)
		{
		case kButtonDialogItem:
		case kCheckBoxDialogItem:
		case kRadioButtonDialogItem:
		case kStaticTextDialogItem:
		case kEditTextDialogItem:
		
			titleLength = *((*ditlItemPtr) + 13);
		
			offset = 14 + titleLength;
			
			if ((titleLength % 2) != 0)	// odd string length
				offset++;		// adjust for alignment byte in DITL

			(*ditlItemPtr) += offset;

			break;
		case kResourceControlDialogItem:
		case kIconDialogItem:
		case kPictureDialogItem:
			(*ditlItemPtr) += 16;
			break;
		case kUserDialogItem:
			(*ditlItemPtr) += 14;
			break;
		case 1:	// no constant available such as kHelpDialogItem
			(*ditlItemPtr) += 20;
			break;
		}
	}

void AGADialog::HandleUserItem(SInt16 itemNo)
	{
	// Handle a userItem proc call for one of our custom items.
	// We just tell the AGAObject to draw if there is one.

	AGAObject*	anAGAObject = this->GetAGAObjectPtr(itemNo - mNumAGAObjects);
	
	if (anAGAObject != NULL)
		anAGAObject->DrawObject();
	}

Boolean AGADialog::IsDITLItemEnabled(SInt16 itemNo)
	{
	// Convenience function to test the enable flag of the specified item.

	SInt16	itemType;
	Handle	itemHandle;
	Rect	itemRect;

	::GetDialogItem(mDialogPtr, itemNo, &itemType, &itemHandle, &itemRect);
	
	return ((itemType & itemDisable) == 0);
	}

SInt16 AGADialog::GetDITLItemType(SInt16 itemNo)
	{
	// Convenience function to get the specified item's type, without
	// regard to its enable flag.

	SInt16	itemType;
	Handle	itemHandle;
	Rect	itemRect;

	::GetDialogItem(mDialogPtr, itemNo, &itemType, &itemHandle, &itemRect);
	
	return itemType & ~itemDisable;	// clear the disable bit; the caller doesn't want it
	}

AGAPushButton* AGADialog::GetVerifiedPushButton(SInt16 itemNo)
	{
	// Return the pointer to the AGAPushButton after checking the
	// original item's type to make sure it's really a Control Manager
	// pushbutton that we are replacing. By doing the verification
	// via the itemType, we don't need to do RTTI on our AGAObject.

	if ((this->GetDITLItemType(itemNo) & (ctrlItem + btnCtrl)) == 0)
		return NULL;
	else
		return (AGAPushButton*) this->GetAGAObjectPtr(itemNo);
	}

static void ParamSubst(StringPtr destString,
						const StringPtr param1,
						const StringPtr param2,
						const StringPtr param3,
						const StringPtr param4)
	{
	StringHandle	h = NULL;
	SInt32			result;
	UInt8			targetText[2];
	Str255			tempString;

	targetText[0] = '^';

	h = ::NewString(destString);
	
	if (h != NULL)
		{
		targetText[1] = '0';
		AGA_PLstrcpy(tempString, param1);
		result = Munger((Handle) h, 1, &targetText[0], 2, &tempString[1], tempString[0]);
		
		targetText[1] = '1';
		AGA_PLstrcpy(tempString, param2);
		result = Munger((Handle) h, 1, &targetText[0], 2, &tempString[1], tempString[0]);
		
		targetText[1] = '2';
		AGA_PLstrcpy(tempString, param3);
		result = Munger((Handle) h, 1, &targetText[0], 2, &tempString[1], tempString[0]);
		
		targetText[1] = '3';
		AGA_PLstrcpy(tempString, param4);
		result = Munger((Handle) h, 1, &targetText[0], 2, &tempString[1], tempString[0]);
		
		//
		// Update the length byte of the string in the handle, because Munger just
		// munged the string data but didnt change the length byte.
		//
		**((UInt8 **) h) = (UInt8) (GetHandleSize((Handle) h) - 1);
		
		AGA_PLstrcpy(destString, (UInt8 *) *h);
		
		::DisposeHandle((Handle) h);
		}
	}

void AGADialog::DoParamTextSubstitution(StringPtr dialogString)
	{
	Str255			s1;
	Str255			s2;
	Str255			s3;
	Str255			s4;
	StringHandle	aStringHandle;
	
	aStringHandle = LMGetDAStrings(0);
	if ((aStringHandle == NULL) || (*aStringHandle == NULL))
		s1[0] = 0;
	else
		AGA_PLstrcpy(s1, *aStringHandle);

	aStringHandle = LMGetDAStrings(1);
	if ((aStringHandle == NULL) || (*aStringHandle == NULL))
		s2[0] = 0;
	else
		AGA_PLstrcpy(s2, *aStringHandle);

	aStringHandle = LMGetDAStrings(2);
	if ((aStringHandle == NULL) || (*aStringHandle == NULL))
		s3[0] = 0;
	else
		AGA_PLstrcpy(s3, *aStringHandle);

	aStringHandle = LMGetDAStrings(3);
	if ((aStringHandle == NULL) || (*aStringHandle == NULL))
		s4[0] = 0;
	else
		AGA_PLstrcpy(s4, *aStringHandle);

	ParamSubst(dialogString, s1, s2, s3, s4);
	}

Boolean AGADialog::HandleModalFilter(EventRecord* theEvent, SInt16* itemHit)
	{
	// This is the filterProc that we always pass to ModalDialog.
	// If the caller supplied a filterProc, we saved it in mUserFilter,
	// and we call it first.

	Boolean	wasEventHandled = false;
	GrafPtr	savedPort;
	
	::GetPort(&savedPort);
	::SetPort(mDialogPtr);
	
	::RGBForeColor(&gAGARamp[rB]);
	::RGBBackColor(&gAGARamp[rW]);	// make any editText items paint a white bg

	if (mUserFilter != NULL)
		wasEventHandled = CallModalFilterProc(mUserFilter, mDialogPtr, theEvent, itemHit);

	if (! wasEventHandled)
		{
		switch (theEvent->what)
			{
			case mouseDown:
				wasEventHandled = this->FilterMouseDown(theEvent, itemHit);
				break;

			case keyDown:
			case autoKey:
				wasEventHandled = this->FilterKeyDown(theEvent, itemHit);
				break;

			case updateEvt:
			
				if (mIsAlert &&
					(! gAGAStdControls) &&
					(mDefaultItemNo != -1) &&
					(AGADialog::ObjectFromDialogPtr((WindowPtr) (theEvent->message)) == this))
					{
					// If we're an Alert, not a ModalDialog, and the AGA
					// controls are turned off, we need to draw the default
					// button's outline, just like the standard Alert()
					// call does. (If AGA controls are on, the default
					// button draws its own AGA outline.)
					
					SInt16	itemType;
					Handle	itemHandle;
					Rect	itemRect;

					::GetDialogItem(mDialogPtr, mDefaultItemNo, &itemType, &itemHandle, &itemRect);
					
					if ((itemType & (ctrlItem + btnCtrl)) != 0)
						{
						::InsetRect(&itemRect, -4, -4);

						::PenNormal();
						::PenSize(3, 3);
						::FrameRoundRect(&itemRect, 16, 16);
						::PenNormal();
						}
					}

				break;
			}
		}

	::SetPort(savedPort);

	return wasEventHandled;
	}

Boolean AGADialog::FilterMouseDown(EventRecord* theEvent, SInt16* itemHit)
	{
	Boolean	wasEventHandled = false;
	Point	localMouse = theEvent->where;
	
	::GlobalToLocal(&localMouse);
	
	//
	// FindDialogItem is unwieldy in cases where things overlap, because
	// our custom items are always appended later in the item list, so
	// non-clickable things in the original list will always be found by
	// FindDialogItem. So we have to check the item locations ourself,
	// and walk the list backwards so that our items get first crack.
	//

	SInt16	foundItemNo = -1;
	SInt16	numItems = ::CountDITL(mDialogPtr);
	
	for (SInt16 i = numItems; (i > mNumAGAObjects) && ! wasEventHandled; i--)
		{
		SInt16		itemType;
		Handle		itemHandle;
		Rect		itemRect;

		::GetDialogItem(mDialogPtr, i, &itemType, &itemHandle, &itemRect);
		
		if (::PtInRect(localMouse, &itemRect))
			{
			CustomObjectInfoPtr	infoPtr = this->GetCustomObjectInfoPtr(i - mNumAGAObjects);

			if ((infoPtr != NULL) && (infoPtr->mAGAObject != NULL) && infoPtr->mTracksMouse)
				{
				// It's an AGAObject that wants to track the mouse.
				wasEventHandled = true;

				if (infoPtr->mAGAObject->TrackMouse(localMouse))
					*itemHit = i - mNumAGAObjects;
				}
			else
				{
				// See if the original object can track the mouse.
				// This may be a placeholder for a trackable control. For
				// example, we may have had AGA controls turned off when
				// the dialog was initialized.

				::GetDialogItem(mDialogPtr, i - mNumAGAObjects, &itemType, &itemHandle, &itemRect);
		
				if (::PtInRect(localMouse, &itemRect) &&
					((itemType & itemDisable) == 0) &&
					((itemType & ctrlItem) != 0))
					{
					wasEventHandled = true;
					
					SInt16	variant = ::GetControlVariant((ControlHandle) itemHandle);

//					if ((variant & popupMenuProc) == popupMenuProc)

					if ((itemType & (ctrlItem + resCtrl)) == (ctrlItem + resCtrl))
						{
						// Once again, the goofball popup CDEF is non-standard.
						// Pass -1 for the UPP, and always return the itemNo.
						(void) ::TrackControl((ControlHandle) itemHandle, localMouse, (ControlActionUPP) -1);
						*itemHit = i - mNumAGAObjects;
						}
					else
						{
						if (::TrackControl((ControlHandle) itemHandle, localMouse, NULL) != kControlNoPart)
							*itemHit = i - mNumAGAObjects;
						}
					}
				}
			}
		}

	return wasEventHandled;
	}

Boolean AGADialog::FilterKeyDown(EventRecord* theEvent, SInt16* itemHit)
	{
	// We catch command-period, escape, enter, and return.
	// If you're concerned about multi-byte keys, international
	// cancelling, etc., your existing filter will probably
	// be handling this already.

	Boolean	wasEventHandled = false;
	UInt16	key = (theEvent->message) & charCodeMask;
	Boolean	isCommandKey = (((theEvent->modifiers) & cmdKey) != 0);

	switch (key)
		{
		case '.':
			if (isCommandKey && (mCancelItemNo != -1))
				{
				this->FlashButton(mCancelItemNo);
				wasEventHandled = true;
				*itemHit = mCancelItemNo;
				}
			break;

		case 0x1B:	// escape
			if (mCancelItemNo != -1)
				{
				this->FlashButton(mCancelItemNo);
				wasEventHandled = true;
				*itemHit = mCancelItemNo;
				}
			break;

		case 0x03:	// enter
		case 0x0D:	// return
			if (mDefaultItemNo != -1)
				{
				this->FlashButton(mDefaultItemNo);
				wasEventHandled = true;
				*itemHit = mDefaultItemNo;
				}
			break;
		}

	return wasEventHandled;
	}

void AGADialog::InstallAGAUserItem(AGAObject* anObject, SInt16 itemNo, Boolean tracksMouse)
	{
	// This handy routine lets you install any AGAObject that you've instantiated
	// as a userItem in the dialog. You just make a userItem for it in your
	// original DITL, and then use this routine to install the AGAObject.
	// The sample app shows how to do this. It has a Big Dialog Box o Stuff that
	// uses a whole bunch of such userItem AGAObject controls.

	SInt16	itemType;
	Handle	itemHandle;
	Rect	itemRect;
	
	this->GetDialogItem(itemNo + mNumAGAObjects, &itemType, &itemHandle, &itemRect);
	
	anObject->SetObjectBounds(&itemRect, AGAObject::kDontRedraw);

	CustomObjectInfoPtr	infoPtr = this->GetCustomObjectInfoPtr(itemNo);
	
	if (infoPtr != NULL)
		{
		infoPtr->mAGAObject = anObject;
		infoPtr->mVisible = true;
		infoPtr->mTracksMouse = tracksMouse;
		}
	}

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

typedef struct
	{
	Rect	controlRect;
	SInt16	titlePosition;	// was CNTL value
	Boolean	visible;
	Boolean	filler;
	SInt16	titleWidth;		// was CNTL max
	SInt16	menuID;			// was CNTL min
	SInt16	procID;			// variants matter!
	SInt32	refCon;
	Str255	title;			// only used string bytes are in the handle
	} PopupControlData, *PopupControlDataPtr, **PopupControlDataHandle;

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=reset
#endif

void AGADialog::HintPopupControl(SInt16 itemNo, SInt16 cntlRsrcID)
	{
	if (gAGAStdControls)
		{
		PopupControlDataHandle	cntlHandle = (PopupControlDataHandle) ::GetResource('CNTL', cntlRsrcID);
		
		if (cntlHandle != NULL)
			{
			SInt16			itemType;
			Handle			itemHandle;
			Rect			itemRect;
			Boolean			enabled = true;
			AGAHandleLocker	locker((Handle) cntlHandle);

			::GetDialogItem(mDialogPtr, itemNo, &itemType, &itemHandle, &itemRect);
	
			AGAPopupMenu*	aPopupMenu = new AGAPopupMenu(&itemRect,//&((**cntlHandle).controlRect),
														(**cntlHandle).titleWidth,
														(**cntlHandle).title,
														(**cntlHandle).titlePosition,
														gAGAStdSystemStyle /*check procID*/,
														gAGAStdSystemStyle /*check procID*/,
														AGAPopupMenu::kSystemMDEFAdjustment /*check procID*/,
														(**((PopupPrivateDataHandle) (**((ControlHandle) itemHandle)).contrlData)).mHandle,
														AGAPopupMenu::kDontDispose);

			if ((itemType & itemDisable) != 0)
				aPopupMenu->SetEnable(false, AGAObject::kDontRedraw);

			CustomObjectInfoPtr	infoPtr = this->GetCustomObjectInfoPtr(itemNo);
			
			if (infoPtr != NULL)
				{
				infoPtr->mAGAObject = aPopupMenu;
				infoPtr->mVisible = (**cntlHandle).visible;
				infoPtr->mTracksMouse = AGADialog::kUserItemTracksMouse;

				::HideDialogItem(mDialogPtr, itemNo);
				}
			}
		}
	}

void AGADialog::FlashButton(SInt16 itemNo)
	{
	// Flash the specified button. If it's an AGAPushButton, use the
	// AGAPushButton methods. Otherwise, use the Control Manager
	// highlighting functions, assuming it's a button.

	SInt32			dontCare;
	AGAPushButton*	buttonToFlash = GetVerifiedPushButton(itemNo);
	
	if (buttonToFlash != NULL)
		{
		buttonToFlash->DrawButton(AGAObject::kPressed);
		::Delay(8, &dontCare);
		buttonToFlash->DrawButton(AGAObject::kNotPressed);
		}
	else
		{
		SInt16	itemType;
		Handle	itemHandle;
		Rect	itemRect;

		::GetDialogItem(mDialogPtr, itemNo, &itemType, &itemHandle, &itemRect);
		
		if ((itemType & (ctrlItem + btnCtrl)) != 0)
			{
			// It is indeed a Control Manager button. It is safe to
			// call HiliteControl.
			
			::HiliteControl((ControlHandle) itemHandle, kControlButtonPart);
			::Delay(8, &dontCare);
			::HiliteControl((ControlHandle) itemHandle, kControlNoPart);
			}
		}
	}

void AGADialog::MarkDefaultOutlineUserItem(SInt16 itemNo)
	{
	if (gAGAStdControls)
		{
		// Remove the userItem proc that is intended to draw the default
		// button outline. The AGA pushbutton control will draw its own.
		
		SInt16	itemType;
		Handle	itemHandle;
		Rect	itemRect;

		::GetDialogItem(mDialogPtr, itemNo, &itemType, &itemHandle, &itemRect);
		::SetDialogItem(mDialogPtr, itemNo, itemType, NULL, &itemRect);
		}
	}

pascal void AGADialog::UserItem(WindowPtr theWindow, SInt16 itemNo)
	{
	// Static class function.
	// Call the dialog object's user item handling method.

	AGADialog*	theAGADialog = AGADialog::ObjectFromDialogPtr(theWindow);
	
	if (theAGADialog != NULL)
		theAGADialog->HandleUserItem(itemNo);
	}

pascal Boolean AGADialog::ModalFilter(DialogPtr theDialog, EventRecord* theEvent, SInt16* itemHit)
	{
	// Static class function.
	// Call the dialog object's filter proc method.

	AGADialog*	theAGADialog = AGADialog::ObjectFromDialogPtr(theDialog);
	
	if (theAGADialog != NULL)
		return theAGADialog->HandleModalFilter(theEvent, itemHit);
	else
		return false;	// we didn't handle the event
	}

AGADialog* AGADialog::ObjectFromDialogPtr(WindowPtr theWindow)
	{
	// Static class function.
	// Return the AGADialog pointer, if any, that is wrapping the
	// specified window.

	AGADialog*	agaDialog = (AGADialog*) ::GetWRefCon(theWindow);
	
	if (agaDialog != NULL)
		if (agaDialog->mMarker != kAGADialogMarker)
			agaDialog = NULL;	// It doesn't seem to actually be an AGADialog!
	
	return agaDialog;
	}

//
// AGAHandleLocker ---------------------------------------------------------------
//
// AGAHandleLocker is a stack-based object that locks a handle, and
// restores its original handle state upon destruction.
//

AGAHandleLocker::AGAHandleLocker(Handle itsHandle)
	{
	mHandle = itsHandle;
	
	this->Save();
	this->Lock();
	}

AGAHandleLocker::~AGAHandleLocker()
	{
	this->Restore();
	}

void AGAHandleLocker::Save()
	{
	mSavedHandleState = ::HGetState(mHandle);
	}

void AGAHandleLocker::Restore()
	{
	::HSetState(mHandle, mSavedHandleState);
	}

void AGAHandleLocker::Lock()
	{
	::HLock(mHandle);
	}

void AGAHandleLocker::Unlock()
	{
	::HUnlock(mHandle);
	}

//
// AGAPortSetter ---------------------------------------------------------------
//
// AGAPortSetter is a stack-based object that sets the port, and
// restores it upon destruction.
//

AGAPortSetter::AGAPortSetter(GrafPtr itsPort)
	{
	this->Save();

	::SetPort(itsPort);
	}

AGAPortSetter::~AGAPortSetter()
	{
	this->Restore();
	}

void AGAPortSetter::Save()
	{
	::GetPort(&mSavedPort);
	}

void AGAPortSetter::Restore()
	{
	::SetPort(mSavedPort);
	}

//
// AGAEditTextFrame ---------------------------------------------------------------
//
// AGAEditTextFrame draws a frame around an editText item in a dialog box.
//

AGAEditTextFrame::AGAEditTextFrame(Rect* bounds)
: AGAObject(bounds)
	{
	}

AGAEditTextFrame::~AGAEditTextFrame()
	{
	}

void AGAEditTextFrame::DrawObject()
	{
	// Draw the frame and erase the interior to white.

	AGADrawingEnvironment	env;
	Rect					r = mBounds;
	
	AGADrawBorderFrame(&r, true, false, true, &gAGARamp[r2]);
	
	::RGBForeColor(&gAGARamp[rW]);
	::InsetRect(&r, 2, 2);
	::PenSize(2, 2);
	::FrameRect(&r);
	}

//
// AGADialogIcon ---------------------------------------------------------------
//
// AGADialogIcon draws a DITL Icon object (a 1-bit 'ICON').
// If an 'icl8' color icon family resource is available, it
// is used instead of the black & white ICON.
//

AGADialogIcon::AGADialogIcon(Rect* bounds, SInt16 iconID)
: AGAObject(bounds)
	{
	mIconID = iconID;
	}

AGADialogIcon::~AGADialogIcon()
	{
	}

void AGADialogIcon::DrawObject()
	{
	// If there's an icon family for this ID around, use it.
	// Otherwise, use the black & white ICON.

	Handle	icl8Handle;

	::SetResLoad(false);
	icl8Handle = GetResource('icl8', mIconID);
	::SetResLoad(true);

	::RGBForeColor(&gAGARamp[rB]);
	::RGBBackColor(&gAGARamp[r2]);
	
	if (icl8Handle != NULL)
		(void) ::PlotIconID(&mBounds, atNone, ttNone, mIconID);
	else
		::PlotIcon(&mBounds, ::GetIcon(mIconID));
	}

