//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// The Gray Council provides a set of standard C++ classes that implement
// the standard Apple Grayscale Appearance. The core classes do not
// require any other code such as a particular class framework.
//
// This file defines a set of helper functions and classes that wrap the
// standard MacOS Dialog Manager functionality, and interface with the core
// Gray Council code. To give a standard dialog the AGA look, you must
// instantiate an AGADialog object to interface with Dialog Manager for
// that dialog.
//
// There are several ways that you normally access the Dialog Manager's
// alerts and dialogs, and this determines how you use the interfaces
// provided here.
//
// To wrap simple alert calls, you just replace the standard calls with
// the corresponding wrappers shown here. The parameters and function
// result are exactly the same. However, if you access dialog items from
// inside your filterProc, see the notes below for how you must also wrap
// those accesses.
//
// - Alert() becomes AGAAlert()
// - StopAlert() becomes AGAStopAlert()
// - NoteAlert() becomes AGANoteAlert()
// - CautionAlert() becomes AGACautionAlert()
//
// Dialogs are more complicated than Alerts, with or without the AGA. This
// is because you often set up various dialog items before displaying the
// dialog, and then run a more complicated filter that plays around with
// the dialog items as the user clicks and types. Fortunately, The Gray
// Council makes wrapping all of this easy -- you just have to be thorough
// about wrapping your accesses to dialog items.
//
// Running a simple modal dialog normally looks something like this:
//
/*
	SInt16		itemHit = -1;
	DialogPtr	aDialog = ::GetNewDialog(kDialogID, NULL, (WindowPtr) -1);
	
	::ShowWindow(aDialog);
	
	while (itemHit != ok)
		::ModalDialog(myFilterUPP, &itemHit);
	
	::DisposeDialog(aDialog);
*/
//
// Wrapping this with The Gray Council looks like this:
//
/*
	AGADialog	agaDialog;
	SInt16		itemHit = -1;
	DialogPtr	aDialog = agaDialog.GetNewDialog(kDialogID, NULL, (WindowPtr) -1);
	
	::ShowWindow(aDialog);
	
	while (itemHit != ok)
		agaDialog.ModalDialog(myFilterUPP, &itemHit);
	
	agaDialog.DisposeDialog(aDialog);
*/
//
// That's the basics. You create an AGADialog object, and let it wrap the
// GetNewDialog, ModalDialog, and DisposeDialog calls rather than calling
// those Dialog Manager functions directly.
//
// But you almost certainly will need to access dialog items either before
// you show the dialog or inside your filter proc. When you do this, you
// need to wrap the Dialog Manager or Control Manager call so that The Gray
// Council can do its work on the AGA object where appropriate. There are
// functions provided for all necessary calls:
//
// Dialog Manager wrappers:
// - AGADialog::SetDialogItem()
// - AGADialog::HideDialogItem()
// - AGADialog::ShowDialogItem()
// - AGADialog::FindDialogItem()
// - AGADialog::AppendDITL()
// - AGADialog::ShortenDITL()
// - AGADialog::CountDITL()
// - AGADialog::SetDialogDefaultItem()
// - AGADialog::SetDialogCancelItem()
// - AGADialog::SetDialogItemText()
// - AGADialog::SelectDialogItemText()
//
// Control Manager wrappers:
// - AGADialog::SetControlValue()
// - AGADialog::SetControlMinimum()
// - AGADialog::SetControlMaximum()
// - AGADialog::SetControlTitle()
// - AGADialog::HideControl()
// - AGADialog::MoveControl()
// - AGADialog::SizeControl()
// - AGADialog::HiliteControl()
//
// If you call one of these Toolbox functions directly, avoiding the wrapper,
// the desired operation may fail to appear on screen, or in some cases may cause
// all hell to break loose (mainly with the xxxDITL calls). So just remember
// that pretty much all Dialog Manager and Control Manager calls should be
// made via wrapper functions of the AGADialog class. Some are not strictly
// necessary but are provided for completeness. The ones list above are the
// ones that are required.
//
// Please also read the comments below for the other individual functions
// that you may need to use.
//

#pragma once

#ifndef __GCDIALOGMANAGER__
#define __GCDIALOGMANAGER__

#include "GCCore.h"
#include "GCStdControls.h"

#include <dialogs.h>

extern InitGrayCouncilDM();

extern SInt16 AGAAlert(SInt16 alertID, ModalFilterUPP modalFilter);
extern SInt16 AGAStopAlert(SInt16 alertID, ModalFilterUPP modalFilter);
extern SInt16 AGANoteAlert(SInt16 alertID, ModalFilterUPP modalFilter);
extern SInt16 AGACautionAlert(SInt16 alertID, ModalFilterUPP modalFilter);

//
// AGADialog --------------------------------------------------------------
//

typedef struct
	{
	AGAObject*	mAGAObject;
	Boolean		mVisible;
	Boolean		mTracksMouse;
	} CustomObjectInfo, *CustomObjectInfoPtr, **CustomObjectInfoHandle;

class AGADialog
	{
	public:
	
		AGADialog();
		virtual ~AGADialog();
		
		DialogPtr	GetNewDialog(SInt16 dialogID, Ptr dStorage, WindowPtr behind);
		SInt16		Alert(SInt16 alertID, ModalFilterUPP modalFilter, SInt16 iconID);

		//		
		// Dialog Manager wrapper functions. In order to properly wrap a
		// dialog with the AGA look and feel, you must use these methods
		// instead of their Dialog Manager counterparts when dealing with
		// an AGADialog object. If you call the Dialog Manager calls
		// directly, we can't do the right thing with the replacement
		// AGA objects.
		//

		void	CloseDialog();
		void	DisposeDialog();
		void	GetDialogItem(SInt16 itemNo, SInt16* itemType, Handle* item, Rect* box);	// wrapper not strictly needed
		void	SetDialogItem(SInt16 itemNo, SInt16 itemType, Handle item, Rect* box);
		void	HideDialogItem(SInt16 itemNo);
		void	ShowDialogItem(SInt16 itemNo);
		SInt16	FindDialogItem(Point thePt);
		void	AppendDITL(Handle theDITL, DITLMethod theMethod);
		void	ShortenDITL(SInt16 numberOfItems);
		SInt16	CountDITL();
		OSErr	SetDialogDefaultItem(SInt16 newItem);
		OSErr	SetDialogCancelItem(SInt16 newItem);
		void	GetDialogItemText(Handle item, Str255 text);	// wrapper not strictly needed
		void	SetDialogItemText(Handle item, ConstStr255Param text);
		void	SelectDialogItemText(SInt16 itemNo, SInt16 strtSel, SInt16 endSel);
		void	ModalDialog(ModalFilterUPP filterProc, SInt16* itemHit);
		Boolean	DialogSelect(EventRecord* event, SInt16* itemHit);
		void	DrawDialog();
		void	UpdateDialog(RgnHandle updateRgn);
		
		//
		// Control Manager wrapper functions. In order to properly wrap a
		// dialog with AGA look and feel, you must also use these methods
		// instead of their Control Manager counterparts when dealing a
		// control that is part of an AGADialog. If you call the Control
		// Manager calls directly, we can't do the right thing with the
		// replacement AGA objects.
		//
		
		void	SetControlValue(ControlHandle theControl, SInt16 theValue);
		void	SetControlMinimum(ControlHandle theControl, SInt16 minValue);
		void	SetControlMaximum(ControlHandle theControl, SInt16 maxValue);
		void	SetControlTitle(ControlHandle theControl, ConstStr255Param title);
		void	HideControl(ControlHandle theControl);
		void	MoveControl(ControlHandle theControl, SInt16 h, SInt16 v);
		void	SizeControl(ControlHandle theControl, SInt16 w, SInt16 h);
		void	HiliteControl(ControlHandle theControl, SInt16 hiliteState);

		// This is the actual UPP-based UserItem static function.
		static pascal void UserItem(WindowPtr theWindow, SInt16 itemNo);
		// This is our UserItem callback called by the static function.
		void	HandleUserItem(SInt16 itemNo);

		// This is the actual UPP-based ModalFilter static function.
		static pascal Boolean ModalFilter(DialogPtr theDialog, EventRecord* theEvent, SInt16* itemHit);
		// This is our ModalFilter callback called by the static function.
		Boolean	HandleModalFilter(EventRecord* theEvent, SInt16* itemHit);
		
		//
		// Useful utilities.
		//

		AGAObject*			GetAGAObjectPtr(SInt16 itemIndex);
		
		// InstallAGAUserItem will take the supplied AGA object and use it
		// as the userItem for the specified itemNo. To make things easier,
		// you can construct the object with any bounds; this function will
		// set the bounds to match the item. Note that it will actually
		// reside at the (itemNo + mNumAGAItems) item. Pass true for objects
		// that should eat mouse clicks. (e.g., false for group boxes)
		void	InstallAGAUserItem(AGAObject* anObject, SInt16 itemNo, Boolean tracksMouse);
		enum { kUserItemTracksMouse = true, kUserItemDoesNotTrackMouse = false };
		
		// HintPopupControl is necessary to get AGA popup menu CTRL items
		// in a dialog. This is because the popup menu CDEF is such a hack
		// that it is too unwieldy to track the necessary information to
		// replace them automatically. Anyway, this method will create
		// the AGAPopupMenu object and install it the same way it does for the
		// other controls that can be installed through the normal route.
		void	HintPopupControl(SInt16 itemNo, SInt16 cntlRsrcID);
		
		// FlashButton will flash a pushbutton object. If the specified itemNo
		// has been turned into an AGAPushButton, or if it is a Control Mgr control,
		// it will work.
		void	FlashButton(SInt16 itemNo);
		
		// MarkDefaultOutlineUserItem is what you should use to inform the
		// AGA dialog object which item (if any) is your userItem proc for
		// drawing the default button outline. It needs this so that in
		// AGA mode it can remove the proc (since the AGA pushbutton
		// object draws its own AGA outline), and in AGA-suppressed mode
		// it will do nothing (so the userItem proc will draw the outline).
		void	MarkDefaultOutlineUserItem(SInt16 itemNo);
		
		// ObjectFromDialogPtr is a static function that looks at the
		// supplied dialog's refcon, and if it appears to be owned by
		// an AGADialog, the AGADialog* is returned. Otherwise, NULL
		// is returned, indicating that the dialog is not wrapped by us.
		static AGADialog*	ObjectFromDialogPtr(WindowPtr theWindow);

		enum { kAGADialogMarker = 'GCdm'} ;
		OSType	mMarker;	// a sanity check so we can see if a dialog's refcon really points to an AGADialog

	protected:
	
		//
		// After constructing this object and giving it its DialogPtr (either
		// in the constructor or via SetDialog), we call BeginAGADialog
		// to let the object prepare all of its custom AGA stuff. Basically,
		// the rule is that this must be called before the dialog is made visible.
		//

		void	BeginAGADialog();

		//
		// Data usage:
		// When we install ourself into the dialog, suppose there are n items
		// in the item list.
		// We make room for n AGA objects in our array.
		// We append n userItems to the item list, one for each AGA object.
		// If an item is one we can replace, we hide the original, make
		// the appropriate AGA object take its coordinate position, and
		// install the user item callback for the AGA object.
		// If an item is not one we can replace, we store NULL in our array
		// and hide the useritem.
		// We always keep the original dialog items at the front of the DITL,
		// so that existing user code can continue to use the same item numbers
		// that it used to.
		//
		
		void				AdjustAlertPosition(AlertTHndl alrtHandle, const Rect* origBounds);
		void				AdjustDialogPosition(DialogTHndl dlogHandle, const Rect* origBounds);
		void				AdjustWindowPosition(SInt16 positionSpecifier, const Rect* origBounds);
		void				CreateAllCustomObjects();
		void				DestroyAllCustomObjects();
		void				CreateCustomObject(SInt16 itemIndex, Ptr* ditlItemPtr);
		void				DestroyCustomObject(SInt16 itemIndex);
		Boolean				IsCustomObject(SInt16 itemIndex);
		CustomObjectInfoPtr	GetCustomObjectInfoPtr(SInt16 itemIndex);
		AGAObject*			GetAGAObjectPtrFromHandle(Handle itemHandle);
		AGAObject*			GetAGAObjectPtrFromControlHandle(ControlHandle theControl);
		SInt16				GetDITLItemNoFromHandle(Handle itemHandle);
		
		AGAObject*			BuildCustomObject(SInt16 itemIndex, Boolean* hideOriginal, Ptr* ditlItemPtr);
		
		Boolean				FilterMouseDown(EventRecord* theEvent, SInt16* itemHit);
		Boolean				FilterKeyDown(EventRecord* theEvent, SInt16* itemHit);

		void				GetNextDITLItemPtr(SInt16 itemType, Ptr* ditlItemPtr);
		Boolean				IsDITLItemEnabled(SInt16 itemNo);
		SInt16				GetDITLItemType(SInt16 itemNo);
		AGAPushButton*		GetVerifiedPushButton(SInt16 itemNo);
		void				DoParamTextSubstitution(StringPtr dialogString);

		Boolean					mIsAlert;		// true if using Alert(), not ModalDialog()
		DialogPtr				mDialogPtr;		// the Dialog Manager object we are wrapping
		SInt16					mNumAGAObjects;	// num elements in our array 
		CustomObjectInfoHandle	mAGAObjects;	// handle to array of CustomObjectInfo
		ModalFilterUPP			mUserFilter;	// the filter passed by client code to our ModalDialog member function
		SInt16					mDefaultItemNo;	// itemNo of default button -- gets a button outline
		SInt16					mCancelItemNo;	// itemNo of cancel button -- cmd-. and esc keys
		AGAPushButton*			mDefaultButton;	// null if default is not a push button
		AGAPushButton*			mCancelButton;	// null if cancel is not a push button
	};

//
// AGAHandleLocker ----------------------------------------------------------
//
// AGAHandleLocker is a stack-based object that locks a handle, and
// restores its original handle state upon destruction.
//

class AGAHandleLocker
	{
	public:
	
		AGAHandleLocker(Handle itsHandle);
		virtual ~AGAHandleLocker();
		
		void	Save();
		void	Restore();
		void	Lock();
		void	Unlock();

	protected:
	
		Handle	mHandle;
		UInt8	mSavedHandleState;
	};

//
// AGAPortSetter ----------------------------------------------------------
//
// AGAPortSetter is a stack-based object that sets the port, and
// restores it upon destruction.
//

class AGAPortSetter
	{
	public:
	
		AGAPortSetter(GrafPtr itsPort);
		virtual ~AGAPortSetter();
		
		void	Save();
		void	Restore();

	protected:
	
		GrafPtr	mSavedPort;
	};

//
// AGAEditTextFrame ---------------------------------------------------------------
//
// AGAEditTextFrame draws a frame around an editText item in a dialog box.
//

class AGAEditTextFrame : public AGAObject
	{
	public:

		AGAEditTextFrame(Rect* bounds);
		virtual ~AGAEditTextFrame();
		
		virtual void	DrawObject();
	};

//
// AGADialogIcon ---------------------------------------------------------------
//
// AGADialogIcon draws a DITL Icon object (a 1-bit 'ICON').
// If an 'icl8' color icon family resource is available, it
// is used instead of the black & white ICON.
//

class AGADialogIcon : public AGAObject
	{
	public:

		AGADialogIcon(Rect* bounds, SInt16 iconID);
		virtual ~AGADialogIcon();
		
		virtual void	DrawObject();

	protected:
	
		SInt16	mIconID;
	};

#endif // __GCDIALOGMANAGER__
