//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// The Gray Council provides a set of standard C++ classes that implement
// the standard Apple Grayscale Appearance. The core classes do not
// require any other code such as a particular class framework.
// There are separate files provided that plug The Gray Council core
// into class frameworks.
//
// This file defines MacApp adapter classes needed for core stuff.
//
// Classes defined below:
//	AGABackgroundAdornerMA -- TAdorner subclass that you should attach to
//		typical modeless windows; gives it the proper modeless gray background
//	AGAModalBackgroundAdornerMA -- AGABackgroundAdornerMA (TAdorner) subclass
//		that you should attach to modal windows (i.e., things you PoseModally);
//		gives it the proper modal gray background; however, for dBoxProc and
//		moveableDBoxProc windows, leaves window structure region white, so
//		you may prefer to use AGAFlatBackgroundBehaviorMA in these cases
//	AGAFlatBackgroundBehaviorMA -- TBehavior subclass that you should attach
//		to modal windows (i.e., things you PoseModally); gives it the proper
//		gray background, including the window structure region, but without
//		raised edges
//	AGAWhiteBackgroundAdornerMA -- TAdorner subclass you should attach to any
//		view that needs a white background but lives in a gray background window
//		(e.g., TEditText, TScroller containing grid, etc.)
//	AGANotchedWhiteBackgroundAdornerMA -- AGAWhiteBackgroundAdornerMA (TAdorner)
//		subclass that leaves a notch in the bottom right corner, such as the
//		notch left by a pair of scroll bars
//	AGABorderFrameAdornerMA -- TAdorner subclass that you can attach to any
//		view that needs a "3D" sunken frame around it (e.g., the same things
//		that need a white background adorner)
//	AGANotchedBorderFrameAdornerMA -- AGABorderFrameAdornerMA (TAdorner) subclass
//		that leaves a notch in the bottom right corner, such as the notch left by
//		a pair of scroll bars
//	AGATargetBorderViewMA -- TTargetBorderView subclass that draws AGA-style
//		target border frame
//	AGANotchedTargetBorderViewMA -- AGATargetBorderViewMA (TTargetBorderView) subclass
//		that leaves a notch in the bottom right corner, such as the notch left by a
//		pair of scroll bars
//	AGATargetBorderFrameViewMA -- AGATargetBorderViewMA (TTargetBorderView)
//		subclass that draws a "3D" sunken frame just like an AGABorderFrameAdornerMA
//		in addition to the AGA-compatible TTargetBorderView.
//	AGATargetBorderFrameViewMA -- AGATargetBorderFrameViewMA (TTargetBorderView)
//		subclass that leaves a notch in the bottom right corner, such as the notch
//		left by a pair of scroll bars
//

#pragma once

#ifndef __GCCOREMA__
#define __GCCOREMA__

#include "GCCore.h"

extern void RegisterGCCoreMA();

extern void SetAGAObjectFrame(TView* itsView, AGAObject* theAGAObject, const VRect& newViewFrame, Boolean redraw);

#pragma mark AGABackgroundAdornerMA

//
// AGABackgroundAdornerMA ----------------------------------------------------
//
// An adorner that you should attach to typical modeless windows if you
// want the standard AGA gray background. It should be placed BEFORE the
// "Draw Adorner".
//

class AGABackgroundAdornerMA : public TAdorner
	{
	MA_DECLARE_CLASS;

	public:

		AGABackgroundAdornerMA();

		// MacApp overrides
		virtual void DoHighlightSelection(TView* itsView, const VRect& area, HLState fromHL, HLState toHL);
		virtual void Draw(TView* itsView, const VRect& area);
		virtual void ViewChangedFrame(TView* itsView, const VRect& oldFrame, const VRect& newFrame, Boolean	invalidate);
		virtual void AddedToView(TView* itsView);

	protected:

		virtual void DrawBackground(TView* itsView, const VRect& area, Boolean fill, Boolean active);
		virtual Boolean HasGrowBox(TView* itsView);

		AGABackgroundKind	mBackgroundKind;
	};

#pragma mark AGAModalBackgroundAdornerMA

//
// AGAModalBackgroundAdornerMA ----------------------------------------------------
//
// An adorner that you should attach to typical modal windows (dialogs)
// in order to get the standard AGA modal gray background. It should be placed
// BEFORE the "Draw Adorner".
//
// Usage note: This adorner draws the raised gray background in the window
// content region. Some window kinds (notably dBoxProc and movableDBoxProc
// for modal dialogs) pad the window frame with a few pixels around the edge
// of the content area. This adorner will leave a white ring around the usable
// content area, because you can't draw outside it.
// If you want a solid gray background with these kinds of windows, use an
// AGAFlatBackgroundBehaviorMA attached to the window. It will install the
// gray background into the window's color table, so that gray is drawn in
// the area outside the content region.
//

class AGAModalBackgroundAdornerMA : public AGABackgroundAdornerMA
	{
	MA_DECLARE_CLASS;

	public:

		AGAModalBackgroundAdornerMA();
	};

#pragma mark AGAFlatBackgroundBehaviorMA

//
// AGAFlatBackgroundBehaviorMA ----------------------------------------------------
//
// TBehavior subclass that handles DoPostCreate by installing a gray
// background color in the owner's window color table.
// This MUST be only attached to a TWindow or TView in a window.
// The reason we use a behavior rather than an adorner, is that adorners
// don't get notified in a timely fashion when they are attached to the
// host view, while behaviors get a DoPostCreate call, which is at the
// right time to install the color.
//

class AGAFlatBackgroundBehaviorMA : public TBehavior
	{
	MA_DECLARE_CLASS;

	public:

		AGAFlatBackgroundBehaviorMA();
		virtual ~AGAFlatBackgroundBehaviorMA();

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
	};

#pragma mark AGAWhiteBackgroundAdornerMA

//
// AGAWhiteBackgroundAdornerMA ----------------------------------------------------
//
// An adorner that you should attach to any subview where the window has
// a gray background but the particular subview needs a white background.
// Editable text fields and list boxes are the most common examples. It should
// be placed BEFORE the "Draw Adorner".
//

class AGAWhiteBackgroundAdornerMA : public TAdorner
	{
	MA_DECLARE_CLASS;

	public:

		// MacApp overrides
		virtual void Draw(TView* itsView, const VRect& area);
	};

#pragma mark AGANotchedWhiteBackgroundAdornerMA

//
// AGANotchedWhiteBackgroundAdornerMA ----------------------------------------------------
//
// Same as AGAWhiteBackgroundAdornerMA, but leaves a "notch" in bottom
// right corner for area created by a pair of scroll bars.
//

class AGANotchedWhiteBackgroundAdornerMA : public TAdorner
	{
	MA_DECLARE_CLASS;

	public:

		// MacApp overrides
		virtual void Draw(TView* itsView, const VRect& area);
	};

#pragma mark AGABorderFrameAdornerMA

//
// AGABorderFrameAdornerMA ----------------------------------------------------
//
// An adorner that you should attach to any subview where the subview
// needs a "3D" sunken frame around it. Editable text fields and list
// boxes are the most common examples. It should be placed AFTER the
// "Draw Adorner".
//

class AGABorderFrameAdornerMA : public TAdorner
	{
	MA_DECLARE_CLASS;

	public:

		// MacApp overrides
		virtual void Draw(TView* itsView, const VRect& area);
	};

#pragma mark AGANotchedBorderFrameAdornerMA

//
// AGANotchedBorderFrameAdornerMA ----------------------------------------------------
//
// An adorner that you should attach to any subview where you want a
// "3D" sunken frame around it, contoured to fit the "notch" at the
// bottom right of a pair of scroll bars.
//

class AGANotchedBorderFrameAdornerMA : public TAdorner
	{
	MA_DECLARE_CLASS;

	public:

		// MacApp overrides
		virtual void Draw(TView* itsView, const VRect& area);
	};

#pragma mark AGATargetBorderViewMA

//
// AGATargetBorderViewMA ----------------------------------------------------
//
// TTargetBorderView replacement subclass.
//

class AGATargetBorderViewMA : public TTargetBorderView
	{
	MA_DECLARE_CLASS;

	public:

		// TTargetBorderView-compatible I method
		void IAGATargetBorderViewMA(TDocument* itsDocument, TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, IDType itsTargetView);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void ComputeBorderRegion(RgnHandle borderRegion);

	protected:

		RGBColor	mEnabledBackgroundColor;
		RGBColor	mDisabledBackgroundColor;
	};

#pragma mark AGANotchedTargetBorderViewMA

//
// AGANotchedTargetBorderViewMA ----------------------------------------------------
//
// Same as AGATargetBorderViewMA, but with a "notched" bottom right corner.
//

class AGANotchedTargetBorderViewMA : public AGATargetBorderViewMA
	{
	MA_DECLARE_CLASS;

	public:

		// TTargetBorderView-compatible I method
		void IAGANotchedTargetBorderViewMA(TDocument* itsDocument, TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, IDType itsTargetView);

		// MacApp overrides
		virtual void ComputeBorderRegion(RgnHandle borderRegion);
	};

#pragma mark AGATargetBorderFrameViewMA

//
// AGATargetBorderFrameViewMA ----------------------------------------------------
//
// TTargetBorderView replacement subclass that also draws a
// AGABorderFrameAdornerMA-type "3D" sunken frame. Provided
// for convenience: makes it unnecessary to attach an
// AGABorderFrameAdornerMA to the subview that this target
// border view surrounds.
//

class AGATargetBorderFrameViewMA : public AGATargetBorderViewMA
	{
	MA_DECLARE_CLASS;

	public:

		// TTargetBorderView-compatible I method
		void IAGATargetBorderFrameViewMA(TDocument* itsDocument, TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, IDType itsTargetView);

		// MacApp overrides
		virtual void Draw(const VRect& area);
		virtual void Dim();
		virtual void SetEnable(Boolean state);
	};

#pragma mark AGANotchedTargetBorderFrameViewMA

//
// AGANotchedTargetBorderFrameViewMA ----------------------------------------------------
//
// Same as AGATargetBorderFrameViewMA, but with a "notched" bottom right corner.
//

class AGANotchedTargetBorderFrameViewMA : public AGATargetBorderViewMA
	{
	MA_DECLARE_CLASS;

	public:

		// TTargetBorderView-compatible I method
		void IAGANotchedTargetBorderFrameViewMA(TDocument* itsDocument, TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, IDType itsTargetView);

		// MacApp overrides
		virtual void Draw(const VRect& area);
		virtual void Dim();
		virtual void ComputeBorderRegion(RgnHandle borderRegion);
	};

#endif // __GCCOREMA__
