//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// The Gray Council provides a set of standard C++ classes that implement
// the standard Apple Grayscale Appearance. The core classes do not
// require any other code such as a particular class framework.
// There are separate files provided that plug The Gray Council core
// into class frameworks.
//
// This file defines MacApp adapter classes for the standard Control
// Manager controls.
//
// Classes defined below:
//	AGAIconPushButtonMA -- AGAPushButtonMA (TButton) subclass for AGAIconPushButton
//	AGAIconCheckBoxMA -- AGACheckBoxMA (TCheckBox) subclass for AGAIconCheckBox
//	AGAIconRadioButtonMA -- AGARadioButtonMA (TRadio) subclass for AGAIconRadioButton
//	AGASliderMA -- TControl subclass for AGASlider
//	AGALittleArrowsMA -- TControl subclass for AGALittleArrows
//	AGADisclosureTriangleMA -- TControl subclass for AGADisclosureTriangle
//	AGAProgressIndicatorMA -- TControl subclass for AGAProgressIndicator
//	AGASeparatorMA -- TView subclass for AGASeparator
//	AGAGroupBoxMA -- TCluster subclass for AGAGroupBox
//	AGASecondaryGroupBoxMA -- AGAClusterMA (TCluster) subclass for AGAGroupBox
//		with secondary box style set
//	AGATabPanelMA -- TControl subclass for AGATabPanel set for large tabs.
//	AGASmallTabPanelMA -- AGATabPanelMA (TControl) subclass for AGATabPanel
//		set for small tabs.
//

#pragma once

#ifndef __GCNEWCONTROLSMA__
#define __GCNEWCONTROLSMA__

#include "GCNewControls.h"
#include "GCStdControlsMA.h"

// qPart seems to be a good flag for determining 3.3.x or earlier vs. r11 or later
#if defined (qPart)
#include "UCluster.h"
#endif

extern void RegisterGCNewControlsMA();

#pragma mark AGAIconPushButtonMA

//
// AGAIconPushButtonMA ----------------------------------------------------
//
// TButton replacement subclass that implements an icon button that behaves
// like a TButton -- you click it and it sends its fEventNumber. When
// instantiated from a view resource, the fUserArea is used to obtain the
// button's icon ID. You can also set the icon ID of the AGAIconPushButton
// directly (cast the mAGAObject to AGAIconPushButton*).
//

class AGAIconPushButtonMA : public AGAPushButtonMA
	{
	MA_DECLARE_CLASS;

		AGAIconPushButtonMA();
		virtual ~AGAIconPushButtonMA();

		// TButton-compatible I method
		void IAGAIconPushButtonMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const CStr255& itsLabel);

	protected:

		// AGAPushButtonMA override
		virtual void CreateAGAObject();
	};

#pragma mark AGAIconCheckBoxMA

//
// AGAIconCheckBoxMA ----------------------------------------------------
//
// TCheckBox replacement subclass that implements an icon button that behaves
// like a check box -- you click it and it toggles its on/off state. When
// instantiated from a view resource, the fUserArea is used to obtain the
// button's icon ID. You can also set the icon ID of the AGAIconCheckBox
// directly (cast the mAGAObject to AGAIconCheckBox*). If you want the on/off
// states to have different icons, you can specify different icon IDs.
//

class AGAIconCheckBoxMA : public AGACheckBoxMA
	{
	MA_DECLARE_CLASS;

		AGAIconCheckBoxMA();
		virtual ~AGAIconCheckBoxMA();

		// TCheckBox-compatible I method
		void IAGAIconCheckBoxMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const CStr255& itsLabel, Boolean isTurnedOn);

	protected:

		// AGACheckBoxMA override
		virtual void CreateAGAObject();
	};

#pragma mark AGAIconRadioButtonMA

//
// AGAIconRadioButtonMA ----------------------------------------------------
//
// TRadio replacement subclass that implements an icon button that behaves
// like a radio button -- you click it and it turns on, and notifies its
// cluster so that the others in its group are turned off. When
// instantiated from a view resource, the fUserArea is used to obtain the
// button's icon ID. You can also set the icon ID of the AGAIconRadioButton
// directly (cast the mAGAObject to AGAIconRadioButton*). If you want the on/off
// states to have different icons, you can specify different icon IDs.
//

class AGAIconRadioButtonMA : public AGARadioButtonMA
	{
	MA_DECLARE_CLASS;

		AGAIconRadioButtonMA();
		virtual ~AGAIconRadioButtonMA();

		// TRadio-compatible I method
		void IAGAIconRadioButtonMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const CStr255& itsLabel, Boolean isTurnedOn);

	protected:

		// AGARadioButtonMA override
		virtual void CreateAGAObject();
	};

#pragma mark AGASliderMA

//
// AGASliderMA ----------------------------------------------------
//
// TControl subclass. To set/get the value, access the AGASlider
// directly. When instantiated from a view resource, the fUserArea,
// if non-zero, is used as the resource ID of an 'STR#' resource
// that contains the slider labels; in this case, the range is
// automatically set to match the number of labels, and the slider
// will have a pointy indicator. If the ID is zero or invalid, the
// slider will be rectangular.
//

class AGASliderMA : public TControl
	{
	MA_DECLARE_CLASS;

	public:

		AGASliderMA();
		virtual ~AGASliderMA();

		// TControl-compatible I method
		void IAGASliderMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const TextStyle& itsTextStyle = gSystemStyle);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void SetFrame(const VRect& newFrame, Boolean invalidate); 
		virtual void DoMouseCommand(VPoint& theMouse, TToolboxEvent* event, CPoint hysteresis);
		virtual void DimState(Boolean state, Boolean redraw);
		virtual void Dim();
		virtual void SetEnable(Boolean state);

		virtual void	InstallNotificationRoutine(AGANotifyMAPtr notificationRoutine, void* userData);
		virtual void	HandleNotification(SInt32 dataValue);
		
		AGANotifyMAPtr	mNotificationRoutine;
		void*			mUserData;

		AGASlider*	mAGAObject;

	protected:

		virtual void CreateAGAObject();
		
		static void RealAGANotifier(AGATrackingIndicator* theIndicator, SInt32 dataValue, void* userData);
	};

#pragma mark AGALittleArrowsMA

//
// AGALittleArrowsMA ----------------------------------------------------
//
// TControl subclass. You will need to either install a notification
// routine to respond to arrow tracking, or subclass and override
// the HandleNotification member function. However:
//
// If you simply want to link the little arrows to a TNumberText
// field, just put the TNumberText's view ID the little arrows view's
// fUserArea, and the arrows will automatically increment/decrement
// the TNumberText view's value.
//

class AGALittleArrowsMA : public TControl
	{
	MA_DECLARE_CLASS;

	public:

		AGALittleArrowsMA();
		virtual ~AGALittleArrowsMA();

		// TControl-compatible I method
		void IAGALittleArrowsMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const TextStyle& itsTextStyle = gSystemStyle);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void SetFrame(const VRect& newFrame, Boolean invalidate); 
		virtual void DoMouseCommand(VPoint& theMouse, TToolboxEvent* event, CPoint hysteresis);
		virtual void DimState(Boolean state, Boolean redraw);
		virtual void Dim();
		virtual void SetEnable(Boolean state);
		
		virtual void	InstallNotificationRoutine(AGANotifyMAPtr notificationRoutine, void* userData);
		virtual void	HandleNotification(SInt32 deltaValue);
		
		AGANotifyMAPtr	mNotificationRoutine;
		void*			mUserData;

		AGALittleArrows*	mAGAObject;

	protected:

		virtual void CreateAGAObject();
		
		static void RealAGANotifier(AGALittleArrows* theAGAObject, SInt32 deltaValue, void* userData);

		TNumberText*	mLinkedNumberText;
	};

#pragma mark AGADisclosureTriangleMA

//
// AGADisclosureTriangleMA ----------------------------------------------------
//
// TControl subclass. Set/get the AGADisclosureTriangle state
// directly. Think of it as a check box -- it's either on or off.
//

class AGADisclosureTriangleMA : public TControl
	{
	MA_DECLARE_CLASS;

	public:

		AGADisclosureTriangleMA();
		virtual ~AGADisclosureTriangleMA();

		// TControl-compatible I method
		void IAGADisclosureTriangleMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const TextStyle& itsTextStyle = gSystemStyle);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void SetFrame(const VRect& newFrame, Boolean invalidate); 
		virtual void DoMouseCommand(VPoint& theMouse, TToolboxEvent* event, CPoint hysteresis);
		virtual void DimState(Boolean state, Boolean redraw);
		virtual void Dim();
		virtual void SetEnable(Boolean state);

		AGADisclosureTriangle*	mAGAObject;

	protected:

		virtual void CreateAGAObject();
	};

#pragma mark AGAProgressIndicatorMA

//
// AGAProgressIndicatorMA ----------------------------------------------------
//
// TControl subclass. Access the AGAProgressIndicator directly. For
// determinate progress, set the min/max range, and then call SetValue
// or Increment repeatedly; you can also update the origin value if
// desired. For indeterminate progress, set min=max and then call Animate
// repeatedly.
//

class AGAProgressIndicatorMA : public TControl
	{
	MA_DECLARE_CLASS;

	public:

		AGAProgressIndicatorMA();
		virtual ~AGAProgressIndicatorMA();

		// TControl-compatible I method
		void IAGAProgressIndicatorMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const TextStyle& itsTextStyle = gSystemStyle);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void SetFrame(const VRect& newFrame, Boolean invalidate);
		virtual Boolean DoIdle(IdlePhase phase);
		
		// For an indeterminate progress indicator, you can use these functions
		// to set it running or stop it. It will place itself on the cohandler
		// chain and do a quantized animate at idle time.
		void	StartAutoAnimate();
		void	StopAutoAnimate();

		AGAProgressIndicator*	mAGAObject;

	protected:

		virtual void CreateAGAObject();
	};

#pragma mark AGASeparatorMA

//
// AGASeparatorMA ----------------------------------------------------
//
// TView subclass. The precise width or height is irrelevant; the
// longer of the two dimensions determines the separator direction,
// and the separator is drawn along the top or left edge.
//

class AGASeparatorMA : public TView
	{
	MA_DECLARE_CLASS;

	public:

		AGASeparatorMA();
		virtual ~AGASeparatorMA();

		// TView-compatible I method
		void IAGASeparatorMA(TDocument* itsDocument, TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void SetFrame(const VRect& newFrame, Boolean invalidate); 

		AGASeparator*	mAGAObject;

	protected:

		virtual void CreateAGAObject();
	};

#pragma mark AGAGroupBoxMA

//
// AGAGroupBoxMA ----------------------------------------------------
//
// TCluster replacement subclass. Defaults to primary group box
// type.
//

class AGAGroupBoxMA : public TCluster
	{
	MA_DECLARE_CLASS;

	public:

		AGAGroupBoxMA();
		virtual ~AGAGroupBoxMA();

		// TCluster-compatible I method
		void IAGAGroupBoxMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, ResNumber itsRsrcID, short itsIndex);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void SetFrame(const VRect& newFrame, Boolean invalidate);
		virtual void DoMouseCommand(VPoint& theMouse, TToolboxEvent* event, CPoint hysteresis);
		virtual void DimState(Boolean state, Boolean redraw);
		virtual void Dim();
		virtual void SetLabel(const CStr255& theLabel, Boolean redraw);

		AGAGroupBox*	mAGAObject;

	protected:

		virtual void CreateAGAObject();
		
		Boolean	mIsPrimaryType;	// used by CreateAGAObject to instantiate desired type
	};

#pragma mark AGASecondaryGroupBoxMA

//
// AGASecondaryGroupBoxMA ----------------------------------------------------
//
// TCluster replacement subclass that overrides AGAGroupBoxMA to
// use secondary group box appearance.
//

class AGASecondaryGroupBoxMA : public AGAGroupBoxMA
	{
	MA_DECLARE_CLASS;

		AGASecondaryGroupBoxMA();
		virtual ~AGASecondaryGroupBoxMA();

		// TCluster-compatible I method
		void IAGASecondaryGroupBoxMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, ResNumber itsRsrcID, short itsIndex);
	};

#pragma mark AGATabPanelMA

//
// AGATabPanelMA ----------------------------------------------------
//
// TControl subclass for AGATabPanel set for large tabs. When instantiated
// from a view resource, the fUserArea, if non-zero, is used as the resource
// ID of an 'STR#' resource that contains the tab labels; the number of
// tabs will be the number of strings in the 'STR#' resource. If you have
// existing tabs, such as via the 'STR#', use InstallPanel to assign a
// panel to a tab. If you have no tabs yet, use AddPanel to create a new
// tab for a panel. Note that the panel views must be non-window 'View'
// resources, i.e. view hierarchies with a TView at the root, not a TWindow.
//

class AGATabPanelMA : public TControl
	{
	MA_DECLARE_CLASS;

	public:

		AGATabPanelMA();
		virtual ~AGATabPanelMA();

		// TControl-compatible I method
		void IAGATabPanelMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const TextStyle& itsTextStyle = gSystemStyle);

		// MacApp overrides
		virtual void DoPostCreate(TDocument* itsDocument);
		virtual void Draw(const VRect& area);
		virtual void SetFrame(const VRect& newFrame, Boolean invalidate); 
		virtual void DoMouseCommand(VPoint& theMouse, TToolboxEvent* event, CPoint hysteresis);
		virtual void DimState(Boolean state, Boolean redraw);
		virtual void Dim();
		virtual void SetEnable(Boolean state);
		
		// HandleTabSwitch can be used to switch to the specified panel,
		// but it will call ValidatePanel before allowing the switch.
		// This is what ClickSelf does to handle a click in a tab.
		// To switch directly without validation, just call SetValue.
		virtual void	HandleTabSwitch(SInt32 newTabIndex);
		
		// AddPanel adds another tab, and assigns it the specified 'View' view.
		// If the specified label StringPtr is NULL, the 'View' rsrc name is used.
		virtual void	AddPanel(SInt16 panelResourceID,
								StringPtr tabLabel);

		// InstallPanel assigns the specified 'View' view to the specified
		// existing tab. If the specified label StringPtr is NULL, the existing
		// tab label is left untouched.
		virtual void	InstallPanel(SInt32 tabIndex,
								SInt16 panelResourceID,
								StringPtr tabLabel);

		// RemoveTab removes the specified tab. It does *not* destroy the
		// panel view associated with the tab. If you want to destroy the
		// panel view, call GetPanelView first to access it.
		// You are responsible for switching tabs *beforehand* if you are
		// removing the currently selected tab.
		virtual void	RemoveTab(SInt32 tabIndex, Boolean redraw);

		// GetPanelView returns the LView pointer to the root of the
		// pane hierarchy of the specified tab.
		virtual TView*	GetPanelView(SInt32 tabIndex);
		
		// SetSubviewsEnable is called by the SetEnable override
		// to walk all subviews of the currently installed panel
		// and set their enable state. Made public in case you
		// need to customize panel enabling behavior.
		virtual void	SetSubviewsEnable(Boolean state);

		AGATabPanel*	mAGAObject;

	protected:

		virtual void	CreateAGAObject();
		virtual void	CreateContainerView();

		virtual Boolean	ValidatePanel(SInt32 currentTabIndex);

		virtual void	SwitchPanels(SInt32 newTabIndex);
		virtual Boolean	SwitchPanelOut();	// returns true if panel contained target
		virtual void	SwitchPanelIn(SInt32 newTabIndex, Boolean targetInPanel);
		
		virtual void	DoSubviewsEnable(TView* aSuperView, Boolean state);
		
		TView*	mPanelContainerView;
		TView*	mActivePanelView;
	};

#pragma mark AGASmallTabPanelMA

//
// AGASmallTabPanelMA ----------------------------------------------------
//
// AGATabPanelMA (TControl) subclass for AGATabPanel set for small tabs.
//

class AGASmallTabPanelMA : public AGATabPanelMA
	{
	MA_DECLARE_CLASS;

	public:

		AGASmallTabPanelMA();
		virtual ~AGASmallTabPanelMA();

		// TControl-compatible I method
		void IAGASmallTabPanelMA(TView* itsSuperView, const VPoint& itsLocation, const VPoint& itsSize, SizeDeterminer itsHSizeDet, SizeDeterminer itsVSizeDet, const TextStyle& itsTextStyle = gSystemStyle);

	protected:

		virtual void CreateAGAObject();
	};

#endif // __GCNEWCONTROLSMA__
