//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// The Gray Council provides a set of standard C++ classes that implement
// the standard Apple Grayscale Appearance. The core classes do not
// require any other code such as a particular class framework.
// There are separate files provided that plug The Gray Council core
// into class frameworks.
//
// This file defines PowerPlant adapter classes needed for core stuff.
//
// Classes defined below:
//	AGAWindowPP -- LWindow subclass that draws a gray background, with the
//		background type depending on the LWindow windAttr_Modal flag.
//	AGAWindowPPX -- AGAWindowPP (LWindow) subclass, that uses extended
//		CPPb data to determine the background type.
//	AGADialogBoxPP -- LDialogBox subclass that draws a gray background, with
//		the background type depending on the WDEF procID and the LWindow
//		windAttr_Modal flag, and replaces the PowerPlant default button outline
//		by altering the style of the default button AGAPushButton object.
//	AGAWindowPPX -- AGADialogBoxPP (LDialogBox) subclass, that uses extended
//		CPPb data to determine the background type.
//	AGAWhiteBackgroundAttachmentPP -- LAttachment subclass that draws a
//		white background for the pane to which it is attached.
//	AGANotchedWhiteBackgroundAttachmentPP -- AGAWhiteBackgroundAttachmentPP (LAttachment)
//		subclass that avoids painting the "notch" at the bottom right.
//	AGABorderFrameAttachmentPP -- LAttachment subclass that draws a
//		"3D" sunken frame around the pane to which it is attached.
//	AGANotchedBorderFrameAttachmentPP -- AGABorderFrameAttachmentPP (LAttachment)
//		subclass that accounts for the "notch" at the bottom right.
//	AGAFocusBoxPP -- LFocusBox subclass used by AGAListBox to draw the
//		list box focus frame.
//	AGANotchedFocusBoxPP -- AGAFocusBoxPP (LFocusBox) subclass that draws the
//		focus box fitted to the "notch" in the bottom right corner.
//

#pragma once

#ifndef __GCCOREPP__
#define __GCCOREPP__

#include <LAttachment.h>
#include <LCaption.h>
#include <LControl.h>
#include <LDialogBox.h>
#include <LEditField.h>
#include <LFocusBox.h>
#include <LGroupBox.h>
#include <LListBox.h>
#include <LScroller.h>
#include <LStdControl.h>
#include <LStream.h>
#include <LTextEdit.h>
#include <LWindow.h>
#include <UDrawingState.h>
#include <URegistrar.h>
#include <UTextTraits.h>

#include "GCCore.h"

extern void RegisterGCCorePP();

extern SInt32 GetTextTraitsJustification(ResIDT inTextTraitsID);
extern void ResetAGAObjectFrame(LPane* itsPane, AGAObject* theAGAObject);
extern UInt8 ReadStreamEnum(LStream* inStream);

#pragma mark AGAWindowPP

//
// AGAWindowPP ----------------------------------------------------
//
// LWindow replacement subclass.
// Makes sure the background is painted gray.
// Knows how to shade around the grow box.
//

class AGAWindowPP : public LWindow
	{
	public:

		AGAWindowPP(LStream* inStream);
		virtual ~AGAWindowPP();
		
		enum { class_ID = 'GCwn' };
		static AGAWindowPP*	CreateStream(LStream* inStream);

		// PowerPlant overrides
		virtual void	FinishCreate();
		virtual void	DrawSelf();
		virtual void	ResizeFrameBy(Int16 inWidthDelta, Int16 inHeightDelta, Boolean inRefresh);

	protected:

		// PowerPlant overrides
		virtual void	FinishCreateSelf();
		virtual void	ActivateSelf();
		virtual void	DeactivateSelf();
		
		virtual void	DrawBackground(const Rect* area, Boolean fill, Boolean active);
		virtual void	InvalidateEdges(Boolean before, Int16 inWidthDelta, Int16 inHeightDelta);

		AGABackgroundKind	mBackgroundKind;
	};

#pragma mark AGAWindowPPX

//
// AGAWindowPPX ----------------------------------------------------
//
// Extended AGAWindowPP class for use with custom pane template.
// The background kind is read from the stream rather than determined
// by checking the window's "modal" attribute.
//

class AGAWindowPPX : public AGAWindowPP
	{
	public:

		AGAWindowPPX(LStream* inStream);
		virtual ~AGAWindowPPX();
		
		enum { class_ID = 'GXwn' };
		static AGAWindowPPX*	CreateStream(LStream* inStream);
	};

#pragma mark AGADialogBoxPP

//
// AGADialogBoxPP ----------------------------------------------------
//
// LDialogBox replacement subclass.
// Makes sure the background is painted gray and the default
// button is drawn with an AGA-style default outline.
// Knows how to shade around the grow box.
//

class AGADialogBoxPP : public LDialogBox
	{
	public:

		AGADialogBoxPP(LStream* inStream);
		virtual ~AGADialogBoxPP();
		
		enum { class_ID = 'GCdb' };
		static AGADialogBoxPP*	CreateStream(LStream* inStream);

		// PowerPlant overrides
		virtual void	FinishCreate();
		virtual void	SetDefaultButton(PaneIDT inButtonID);
		virtual void	DrawSelf();
		virtual void	ResizeFrameBy(Int16 inWidthDelta, Int16 inHeightDelta, Boolean inRefresh);

	protected:

		// PowerPlant overrides
		virtual void	FinishCreateSelf();
		virtual void	ActivateSelf();
		virtual void	DeactivateSelf();
		
		virtual void	DrawBackground(const Rect* area, Boolean fill, Boolean active);
		virtual void	InvalidateEdges(Boolean before, Int16 inWidthDelta, Int16 inHeightDelta);

		Boolean				mDeleteDefaultOutline;
		AGABackgroundKind	mBackgroundKind;
	};

#pragma mark AGADialogBoxPPX

//
// AGADialogBoxPPX ----------------------------------------------------
//
// Extended AGADialogBoxPP class for use with custom pane template.
// The background kind is read from the stream rather than determined
// by peeking at the WIND resource.
//

class AGADialogBoxPPX : public AGADialogBoxPP
	{
	public:

		AGADialogBoxPPX(LStream* inStream);
		virtual ~AGADialogBoxPPX();
		
		enum { class_ID = 'GXdb' };
		static AGADialogBoxPPX*	CreateStream(LStream* inStream);
	};

#pragma mark AGAWhiteBackgroundAttachmentPP

//
// AGAWhiteBackgroundAttachmentPP ----------------------------------------------------
//
// An attachment that you should attach to any subview where the window has
// a gray background but the particular subview needs a white background.
// Editable text fields and list boxes are the most common examples.
// AGAEditFieldPP and AGAListBoxPP both add one of these automatically.
//

class AGAWhiteBackgroundAttachmentPP : public LAttachment
	{
	public:

		AGAWhiteBackgroundAttachmentPP(MessageT inMessage = msg_DrawOrPrint, Boolean inExecuteHost = true);
		AGAWhiteBackgroundAttachmentPP(LStream *inStream);

		enum { class_ID = 'GCwa' };
		static AGAWhiteBackgroundAttachmentPP*	CreateStream(LStream* inStream);
		
		virtual	Boolean	Execute(MessageT inMessage, void *ioParam);

	protected:
	
		Boolean	mIsNotched;
	};

#pragma mark AGANotchedWhiteBackgroundAttachmentPP

//
// AGANotchedWhiteBackgroundAttachmentPP ----------------------------------------------------
//
// An AGAWhiteBackgroundAttachmentPP that accounts for a "notch" in
// the white background at the bottom right corner such as left by
// a pair of scroll bars.
//

class AGANotchedWhiteBackgroundAttachmentPP : public AGAWhiteBackgroundAttachmentPP
	{
	public:

		AGANotchedWhiteBackgroundAttachmentPP(MessageT inMessage = msg_DrawOrPrint, Boolean inExecuteHost = true);
		AGANotchedWhiteBackgroundAttachmentPP(LStream *inStream);

		enum { class_ID = 'GCnw' };
		static AGANotchedWhiteBackgroundAttachmentPP*	CreateStream(LStream* inStream);
	};

#pragma mark AGABorderFrameAttachmentPP

//
// AGABorderFrameAttachmentPP ----------------------------------------------------
//
// An attachment that you should attach to any subview where the subview
// needs a "3D" sunken frame around it. Editable text fields and list
// boxes are the most common examples.
// AGAEditFieldPP and AGAListBoxPP both add one of these automatically.
//

class AGABorderFrameAttachmentPP : public LAttachment
	{
	public:

		AGABorderFrameAttachmentPP(MessageT inMessage = msg_DrawOrPrint, Boolean inExecuteHost = true);
		AGABorderFrameAttachmentPP(LStream *inStream);

		enum { class_ID = 'GCba' };
		static AGABorderFrameAttachmentPP*	CreateStream(LStream* inStream);
		
		virtual	Boolean	Execute(MessageT inMessage, void *ioParam);

	protected:
	
		Boolean	mIsNotched;
	};

#pragma mark AGANotchedBorderFrameAttachmentPP

//
// AGANotchedBorderFrameAttachmentPP ----------------------------------------------------
//
// An AGABorderFrameAttachmentPP that accounts for a "notch" in
// the contour at the bottom right corner such as left by
// a pair of scroll bars.
//

class AGANotchedBorderFrameAttachmentPP : public AGABorderFrameAttachmentPP
	{
	public:

		AGANotchedBorderFrameAttachmentPP(MessageT inMessage = msg_DrawOrPrint, Boolean inExecuteHost = true);
		AGANotchedBorderFrameAttachmentPP(LStream *inStream);

		enum { class_ID = 'GCnb' };
		static AGANotchedBorderFrameAttachmentPP*	CreateStream(LStream* inStream);
	};

#pragma mark AGAFocusBoxPP

//
// AGAFocusBoxPP -------------------------------------------------------
//
// LFocusBox replacement subclass that implements an AGA focus frame.
// The AGAListBoxPP class replaces the standard LFocusBox with one of
// these.
//

class AGAFocusBoxPP : public LFocusBox
	{
	public:

		AGAFocusBoxPP();
		AGAFocusBoxPP(const LFocusBox &inOriginal);
		AGAFocusBoxPP(LStream* inStream);
		virtual ~AGAFocusBoxPP();
		
		enum { class_ID = 'GCfb' };
		static AGAFocusBoxPP*	CreateStream(LStream* inStream);
		
	protected:
	
		// PowerPlant overrides
		virtual void		DrawSelf();
		virtual RgnHandle	GetBoxRegion(const Rect &inFrame, const Rect &inRevealed) const;
		
		Boolean mIsNotched;
	};

#pragma mark AGANotchedFocusBoxPP

//
// AGANotchedFocusBoxPP -------------------------------------------------------
//
// An AGAFocusBoxPP that accounts for a "notch" in
// the contour at the bottom right corner such as left by
// a pair of scroll bars.
//

class AGANotchedFocusBoxPP : public AGAFocusBoxPP
	{
	public:

		AGANotchedFocusBoxPP();
		AGANotchedFocusBoxPP(const LFocusBox &inOriginal);
		AGANotchedFocusBoxPP(LStream* inStream);
		virtual ~AGANotchedFocusBoxPP();
		
		enum { class_ID = 'GCnf' };
		static AGANotchedFocusBoxPP*	CreateStream(LStream* inStream);
	};

#endif // __GCCOREPP__
