// =================================================================================
//	C256Document.cp				1997 eli curtz. All rights reserved.
// =================================================================================

#include <LFile.h>
#include <LPlaceHolder.h>
#include <LPrintout.h>
#include <LString.h>
#include <LWindow.h>
#include <PP_Messages.h>
#include <UMemoryMgr.h>
#include <UWindows.h>

#include "Constants.h"
#include "C256View.h"

#include "C256Document.h"

// ---------------------------------------------------------------------------------
//		 C256Document(LCommander*, FSSpec*)
// ---------------------------------------------------------------------------------


C256Document::C256Document(
	LCommander	*inSuper,
	FSSpec		*inFileSpec )
		: LSingleDoc( inSuper ), LListener()
{
	mCurrentBitmap = 0;

	// Create window for our document.
	mWindow = LWindow::CreateWindow( rPPob_Window, this );
	Assert_( mWindow != nil );
	
	// Find the active panes
	m256View = (C256View *) mWindow->FindPaneByID( kMainView );
	Assert_( m256View != nil );
	mNextButton = (LStdButton *) mWindow->FindPaneByID( kNextButton );
	Assert_( mNextButton != nil );
	mPrevButton = (LStdButton *) mWindow->FindPaneByID( kPrevButton );
	Assert_( mPrevButton != nil );
	
	// Listen to the buttons
	mNextButton->AddListener(this);
	mPrevButton->AddListener(this);
	
	if ( inFileSpec == nil ) {

		// Set the name of new window.
		NameNewDoc();
		
	} else {
	
		// Display contents of file in window.
		OpenFile( *inFileSpec );
	
	}
	
	// Make the window visible.
	mWindow->Show();
}


// ---------------------------------------------------------------------------------
//		 NameNewDoc
// ---------------------------------------------------------------------------------

void
C256Document::NameNewDoc()
{
	// Setup the window title. Start with the default title.
	LStr255	theTitle( "\pUntitled" );

	// Find the first available title. We could also check the window
	// pane id if we wanted to make sure we didn't collide with other
	// window types.
	Int32	theNumber = 0;
	while ( UWindows::FindNamedWindow( theTitle ) != nil ) {

		// An existing window has the current name
		// Increment counter and try again.
		++theNumber;
		theTitle = "\pUntitled ";
		theTitle += (LStr255) theNumber;

	}		
	
	// Finally, set window title.
	mWindow->SetDescriptor( theTitle );
}


// ---------------------------------------------------------------------------------
//		 OpenFile
// ---------------------------------------------------------------------------------

void
C256Document::OpenFile(
	FSSpec	&inFileSpec )
{
	Try_ {

		// Create a new file object.
		mFileStream = new LFileStream( inFileSpec );
		mFile = mFileStream;	// In case the superclass is using it
		
		// Open the data fork.
		mFileStream->OpenDataFork( fsRdPerm );
		
		// Set the window title to the name of the file and
		// flag that the document has an associated file.
		mWindow->SetDescriptor( inFileSpec.name );
		mIsSpecified = true;

		// Skip the file header and read the collection header
		dir_Header theFileHeader;
		Int32 byteCount = sizeof(dir_Header);
		mFileStream->GetBytes(&theFileHeader, byteCount);
		
		byteCount = sizeof(collection_header);
		mFileStream->GetBytes(&mCollectionHeader, byteCount);
		mHeaderOffset = mFileStream->GetMarker();

// Just for fun, read the sequence headers
sequence_reference * theSeqs = new sequence_reference[mCollectionHeader.sequence_count];
mFileStream->SetMarker(mHeaderOffset + mCollectionHeader.sequence_references_offset, streamFrom_Start);
byteCount = mCollectionHeader.sequence_references_size;
mFileStream->GetBytes(theSeqs, byteCount);

Ptr *theFrames = new Ptr[mCollectionHeader.sequence_count];

for (int seqCount = 0; seqCount < mCollectionHeader.sequence_count; seqCount++) {
	byteCount = theSeqs[seqCount].size;
	theSeqs[seqCount].sequence = (sequence_data *) ::NewPtrClear(byteCount);
	mFileStream->SetMarker(mHeaderOffset + theSeqs[seqCount].offset, streamFrom_Start);
	mFileStream->GetBytes(theSeqs[seqCount].sequence, byteCount);
	
	theFrames[seqCount] = (Ptr) get_sequence_data_frames(theSeqs[seqCount].sequence);
}

// Just for fun, read the bitmap instances
/*
bitmap_instance_data * theInstances = new bitmap_instance_data[mCollectionHeader.bitmap_instance_count];
mFileStream->SetMarker(mHeaderOffset + mCollectionHeader.bitmap_instances_offset, streamFrom_Start);
byteCount = mCollectionHeader.bitmap_instances_size;
mFileStream->GetBytes(theInstances, byteCount);
*/
	
		// Read a color table
		Assert_(mCollectionHeader.color_table_count >= 1);
		if(mCollectionHeader.color_table_count > 1)
			SignalPStr_("\pMore than one color table in this file.");
		Ptr theColorData = ::NewPtr(mCollectionHeader.color_tables_size);
		ThrowIfNil_(theColorData);
		mFileStream->SetMarker(mHeaderOffset + mCollectionHeader.color_tables_offset, streamFrom_Start);
		byteCount = mCollectionHeader.color_tables_size;
		mFileStream->GetBytes(theColorData, byteCount);
		
		mColorTable = ::GetCTable(40);		// default 8-bit grey
		::HLock((Handle) mColorTable);
		Ptr colorSourcePtr = theColorData + 32;		// skip header
		for (Int16 colorCount = 0; colorCount <= (*mColorTable)->ctSize; colorCount++) {
			(*mColorTable)->ctTable[colorCount].rgb = *((RGBColor*) colorSourcePtr);
			colorSourcePtr += 8;
		}
		::HUnlock((Handle) mColorTable);
		
		::CTabChanged(mColorTable);
		::DisposePtr(theColorData);
	
		GoToBitmap(0);
		
	} Catch_( inErr ) {

		// Cleanup and rethrow the error.
		delete this;
		Throw_( inErr );
	
	} EndCatch_
}


// ---------------------------------------------------------------------------------
//		 IsModified
// ---------------------------------------------------------------------------------

Boolean
C256Document::IsModified()
{
	return false;
}


// ---------------------------------------------------------------------------------
//		 DoAESave
// ---------------------------------------------------------------------------------

void
C256Document::DoAESave(
	FSSpec	&/*inFileSpec*/,
	OSType	/*inFileType*/ )
{
/*
	// Delete the existing file object.
	// Note: this does nothing to the actual file on disk.
	delete mFileStream;
	
	// Make a new file object.
	mFileStream = new LFile( inFileSpec );
	
	// Get the proper file type.
	OSType	theFileType = 'TEXT';
	if ( inFileType != fileType_Default ) theFileType = inFileType;

	// Make new file on disk (we'll use
	// SimpleText's creator for this example).
	mFileStream->CreateNewDataFile( 'ttxt', theFileType );
	
	// Write out the data.
	DoSave();

	// Change window title to reflect the new name.
	mWindow->SetDescriptor( inFileSpec.name );

	// Document now has a specified file.
	mIsSpecified = true;
*/
}


// ---------------------------------------------------------------------------------
//		 DoSave
// ---------------------------------------------------------------------------------

void
C256Document::DoSave()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Get the text from the text view.
	Handle	theTextH = mTextView->GetTextHandle();
	
	// Lock the text handle.
	StHandleLocker	theLock( theTextH );
	
	// Write the text to the file.
	mFileStream->WriteDataFork( *theTextH, ::GetHandleSize( theTextH ) );

	// Close the data fork.
	mFileStream->CloseDataFork();

	// Saving makes doc un-dirty.
	mTextView->SetDirty( false );
*/
}


// ---------------------------------------------------------------------------------
//		 DoRevert
// ---------------------------------------------------------------------------------

void
C256Document::DoRevert()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Read the entire file contents and close the file.
	Handle	theTextH = mFileStream->ReadDataFork();
	mFileStream->CloseDataFork();
	
	// Put the contents in the text view
	// and clear the dirty flag.
	mTextView->SetTextHandle( theTextH );
	mTextView->SetDirty( false );
	
	// Dispose of the text.
	::DisposeHandle( theTextH );

	// Refresh the text view.
	mTextView->Refresh();
*/
}


// ---------------------------------------------------------------------------------
//		 DoPrint
// ---------------------------------------------------------------------------------

void
C256Document::DoPrint()
{
/*
	// Create the printout.
	LPrintout	*thePrintout = LPrintout::CreatePrintout( rPPob_TextPrintout );
	Assert_( thePrintout != nil );
	
	// Set the print record.
	
	if (mPrintRecordH)
	{
		thePrintout->SetPrintRecord( mPrintRecordH );
	}
	
	// Get the text placeholder.
	LPlaceHolder	*thePlaceholder;
	thePlaceholder = (LPlaceHolder*) thePrintout->FindPaneByID( kTextPlaceholder );
	Assert_( thePlaceholder != nil );
	
	// Install the text view in the placeholder.
	thePlaceholder->InstallOccupant( mTextView, atNone );
	
	// Set the frame size.
	SetPrintFrameSize();
	
	// Print.
	thePrintout->DoPrintJob();
	
	// Delete the printout. The text view is returned
	// to the window when the placeholder is destroyed.
	delete thePrintout;
*/
}


// ---------------------------------------------------------------------------------
//		 SetPrintFrameSize
// ---------------------------------------------------------------------------------

void
C256Document::SetPrintFrameSize()
{
/*
	// Get the frame size.
	SDimension16	theFrameSize;
	mTextView->GetFrameSize( theFrameSize );
	
	// Get the text edit record handle.
	TEHandle	theTextEditH = mTextView->GetMacTEH();
	
	// Calculate the number of lines per page.
	Int16	theLinesPerPage;
	theLinesPerPage = theFrameSize.height / (**theTextEditH).lineHeight;

	// Resize the frame to an integral number of lines.
	mTextView->ResizeFrameTo( theFrameSize.width,
		(**theTextEditH).lineHeight * theLinesPerPage, false );

*/
}


void
C256Document::ListenToMessage(
	MessageT inMessage,
	void * /* ioParam */)
{
	switch (inMessage) {
		case msg_Next:
			GoToBitmap(mCurrentBitmap + 1);
			break;
			
		case msg_Previous:
			GoToBitmap(mCurrentBitmap - 1);
			break;
	}
}

void
C256Document::GoToBitmap(
	UInt16 inBitmapIndex)
{
	// Make sure index is valid
	if ((inBitmapIndex < 0) || (inBitmapIndex >= mCollectionHeader.bitmap_count)) {
		SysBeep(1);
		return;
	}
	
	// Read the bitmap "reference"
	Int32 bitmapRefMarker = mHeaderOffset + mCollectionHeader.bitmap_references_offset;
	bitmapRefMarker += inBitmapIndex * sizeof(bitmap_reference);
	
	bitmap_reference theBitmapRef;
	mFileStream->SetMarker(bitmapRefMarker, streamFrom_Start);
	Int32 byteCount = sizeof(bitmap_reference);
	mFileStream->GetBytes(&theBitmapRef, byteCount);
	
	// Read the bitmap data proper
	mFileStream->SetMarker(mHeaderOffset + theBitmapRef.offset, streamFrom_Start);

	Ptr theBitmapData = ::NewPtr(theBitmapRef.size);
	ThrowIfNil_(theBitmapData);
	byteCount = theBitmapRef.size;
	mFileStream->GetBytes(theBitmapData, byteCount);
		
	bitmapData_Header *bitmapHeaderPtr = (bitmapData_Header*) theBitmapData;
	Ptr bitmapDataPtr = theBitmapData + sizeof(bitmapData_Header);
	if (mCollectionHeader.user_data_flags & _collection_is_color_map_flag)
		bitmapDataPtr += 4;
	else
		bitmapDataPtr += bitmapHeaderPtr->height * sizeof(Ptr);	// jump over the row ptrs (?)
	
	Rect bitBounds = { 0, 0, 0, 0 };
	bitBounds.bottom = bitmapHeaderPtr->height;
	bitBounds.right = bitmapHeaderPtr->width;
	
	// Delete the old GWorld to save memory
	m256View->SetGWorld(nil);
	LGWorld* theNewGWorld = nil;

	// We can have a 16 bit transparent picture that actually contains
	// 8-bit data as seen in the map overlays, these can only be caught
	// accessing the instances	
	if (bitmapHeaderPtr->bitDepth > 8)
		theNewGWorld = new LGWorld(bitBounds, bitmapHeaderPtr->bitDepth);
	else
		theNewGWorld = new LGWorld(bitBounds, 8, 0, mColorTable);
	Assert_(theNewGWorld != nil);
	Assert_(bitmapHeaderPtr->bitDepth >= 8);
	UInt8 bytesPerColor = bitmapHeaderPtr->bitDepth / 8;
	
	// Write data into the new GWorld		
	if (theNewGWorld->BeginDrawing()) {
		PixMapHandle thePixH = ::GetGWorldPixMap(theNewGWorld->GetMacGWorld());
		UInt32 realPixRowBytes = (*thePixH)->rowBytes & 0x3FFF;
		Ptr realPixPtr = ::GetPixBaseAddr(thePixH);

		if (bitmapHeaderPtr->rowBytes == -1) {	// Encoded sprite
			for (UInt16 rowCount = 1; rowCount <= bitmapHeaderPtr->height; rowCount++) {
				UInt16 *encodingHeader = (UInt16 *) bitmapDataPtr;

				bitmapDataPtr += (2 * (encodingHeader[0] + 1) * sizeof(UInt16));
				for (UInt16 runCount = 1; runCount <= encodingHeader[0]; runCount++) {
					UInt16 columnStartLoc = encodingHeader[runCount * 2];
					UInt16 columnStopLoc = encodingHeader[runCount * 2 + 1];
					UInt16 byteCount = (columnStopLoc - columnStartLoc) * bytesPerColor;
					Ptr rowPixPtr = realPixPtr + (columnStartLoc * bytesPerColor);
// hack
/*
Ptr rowPixPtr = realPixPtr + (columnStartLoc);
// UInt32 pixelCount = columnStopLoc - columnStartLoc;
for (int countPix = 0; countPix < pixelCount; countPix++) {

UInt16 currentData = *((UInt16*) bitmapDataPtr);

*rowPixPtr = currentData & 0xFF;

bitmapDataPtr += 2;
rowPixPtr += 1;
}
*/
					::BlockMoveData(bitmapDataPtr, rowPixPtr, byteCount);
					bitmapDataPtr += byteCount;
				}

				if ((Int32) bitmapDataPtr & 0x01)	// Check for uneven row
					bitmapDataPtr++;
				realPixPtr += realPixRowBytes;
			}
		}
		else {	// Normal byte runs
			for (UInt16 rowCount = 1; rowCount <= bitmapHeaderPtr->height; rowCount++) {
				::BlockMoveData(bitmapDataPtr, realPixPtr, bitmapHeaderPtr->rowBytes);
				bitmapDataPtr += bitmapHeaderPtr->rowBytes;
				realPixPtr += realPixRowBytes;
			}
		}
	}
	theNewGWorld->EndDrawing();
	::DisposePtr(theBitmapData);

	// Set the new GWorld
	m256View->SetGWorld(theNewGWorld);

	mCurrentBitmap = inBitmapIndex;
}

