/*
MAP_ACTIONS.H
*/

#ifndef __MAP_ACTIONS_H__
#define __MAP_ACTIONS_H__

#include "data.h"
#include "tag_files.h"
#include "world.h"

/* ---------- types */

typedef unsigned long field_tag;
typedef unsigned long action_tag;

/* ---------- map action procs */

typedef unsigned long (*action_proc_pointer)(int action_index);

enum // map action result flags
{
	_map_action_proc_executed_bit,
	_map_action_proc_succeeded_bit, // never set if the action was not executed
	_map_action_proc_deactivate_bit, // deactivate us
	
	NUMBER_OF_MAP_ACTION_PROC_RESULT_FLAGS,
	
	_map_action_proc_executed_flag= FLAG(_map_action_proc_executed_bit),
	_map_action_proc_succeeded_flag= FLAG(_map_action_proc_succeeded_bit),
	_map_action_proc_deactivate_flag= FLAG(_map_action_proc_deactivate_bit)
};

/* ---------- map actions */

enum
{
	MAXIMUM_MAP_ACTION_NAME_LENGTH= 31,
	
	MAP_ACTION_COOKIE= 0xf0cacc1a,
	
	UNUSED_MAP_ACTION_IDENTIFIER= -2,
	MAXIMUM_MAP_ACTION_VISUAL_INDENT= 5
};

enum // map action parameter types
{
	// 1-byte
	_parameter_flag,
	_parameter_character,
	
	// 2-byte
	_parameter_monster_identifier, // monster identifier
	_parameter_action_identifier, // map action identifier
	_parameter_angle, // [0,FULL_CIRCLE)

	// 4-byte
	_parameter_integer, // [LONG_MIN,LONG_MAX]
	_parameter_world_distance,
	_parameter_field_name, // 4-byte, 'xxxx'
	_parameter_fixed,
	_parameter_projectile_group_tag, // resolved when the map is loaded
	_parameter_string_list_tag, // resolved when the map is loaded
	_parameter_sound_tag, // resolved when the map is loaded
	
	// 8-byte
	_parameter_world_point2d, // 2d point on the mesh
	
	// 16-byte
	_parameter_world_rectangle2d, // 2d rectangle on the mesh

	// 2-byte
	_parameter_object_identifier, // object identifier
	_parameter_model_identifier,
	_parameter_sound_source_identifier,
	
	// 12-byte
	_parameter_world_point3d, // 3d world point
		
	// 2-byte
	_parameter_local_projectile_group_identifier,
	
	NUMBER_OF_MAP_ACTION_PARAMETER_TYPES
};

enum // reserved map action parameter uses
{
	_field_name_link= 'link', // alias to another action (only valid with action_identifier types)
	_field_name_inhibitions= 'inhi', // will fail to trigger if all of the given actions have succeeded (not valid if linked to)
	_field_name_prerequisites= 'prer', // will fail to trigger unless one of the given actions succeeded (not valid if linked to)
	_field_name_deactivates_on_activation= 'deoa', // will deactivate all actions in this list when activated
	_field_name_deactivates_on_success= 'deos', // will deactivate all actions in this list when success occurs
	_field_name_activates_on_activation= 'acoa', // will activate all actions in this list when activated
	_field_name_activates_on_execution= 'acoe', // will activate all actions in this list when execution (success or failure) ocurrs
	_field_name_activates_on_failure= 'acof', // will activate all actions in this list when failure ocurrs
	_field_name_activates_on_success= 'acos', // will activate all actions in this list when success ocurrs
	_field_name_activates_on_trigger= 'acot', // will activate all actions in this list when triggered

	_field_name_activates_on_deactivation= 'acod', // will activate all actions in this list when deactivated
	_field_name_deactivates_on_deactivation= 'deod', // will deactivate all actions in this list when deactivated
	
	_field_name_activates_on_error= 'acer', // will activate all actions in this list when an error occurs
	_field_name_deactivates_on_error= 'deer', // will deactivate all actions in this list when an error ocurrs

	_field_name_triggers_simultaneously= 'trsi', // triggering this action simultaneously triggers all these
	_field_name_action_name= 'name',
	_field_name_debug= 'debg' // flag, if set, does map action debugging
};

enum // map action flags
{
	// user flags
	_map_action_initially_active_bit,
	_map_action_activates_only_once_bit,
	_map_action_no_initial_delay_bit,
	_map_action_only_initial_delay_bit,
	
	// runtime flags
	_map_action_is_active_bit,
	_map_action_succeeded_bit, // last return value
	_map_action_has_been_activated_bit,
	_map_action_was_just_activated_bit, // set on transition inactive->active, cleared after first execution
	
	NUMBER_OF_MAP_ACTION_FLAGS,
	
	_map_action_initially_active_flag= FLAG(_map_action_initially_active_bit),
	_map_action_activates_only_once_flag= FLAG(_map_action_activates_only_once_bit),
	_map_action_no_initial_delay_flag= FLAG(_map_action_no_initial_delay_bit),
	_map_action_only_initial_delay_flag= FLAG(_map_action_only_initial_delay_bit),
	_map_action_is_active_flag= FLAG(_map_action_is_active_bit),
	_map_action_succeeded_flag= FLAG(_map_action_succeeded_bit),
	_map_action_has_been_activated_flag= FLAG(_map_action_has_been_activated_bit),
	_map_action_was_just_activated_flag= FLAG(_map_action_was_just_activated_bit)
};

enum // map action deactivation modes
{
	_map_action_deactivates_on_trigger, // always deactivates on trigger
	_map_action_deactivates_on_execution, // only deactivates if it actually got executed
	_map_action_deactivates_on_successful_execution, // only deactivates if action proc returned success
	_map_action_deactivates_never,
	_map_action_deactivates_on_failed_execution,
	NUMBER_OF_MAP_ACTION_DEACTIVATION_MODES
};

#define SIZEOF_STRUCT_MAP_ACTION 64
struct map_action
{
	DATUM_HEADER;
	short deactivation_mode;
	
	action_tag tag; // the proc which this action will trigger, NONE means none
	
	unsigned long flags;
	
	long trigger_time_lower_bound, trigger_time_delta; // could be absolute time or activation-relative delay
	
	short parameter_count;
	short parameters_size; // in bytes
	
	long parameters_offset;

	short visual_indent;

	short unused[15];

	// internal

	long trigger_time; // only valid if active; absolute time the action will trigger 
};

/* ---------- iterating over a map action's parameter list */

enum
{
	MAXIMUM_FIND_MAP_ACTION_PARAMETER_STACK_SIZE= 32
};

struct find_map_action_parameter_stack_entry
{
	short action_index;
	short parameter_index;
};

struct find_map_action_parameter_state
{
	unsigned long flags;
	
	unsigned long field_name_filter; // NONE matches all uses
	short type_filter; // NONE matches all types

	short stack_index, stack_size;
	struct find_map_action_parameter_stack_entry stack[MAXIMUM_FIND_MAP_ACTION_PARAMETER_STACK_SIZE];
};

struct map_action_parameter
{
	unsigned long field_name;
	short type, count;
	
	short action_index, parameter_index;
};

enum // find_next_map_action_parameter() flags
{
	_dont_resolve_links_bit,
	_return_empty_arrays_bit,
	_only_return_one_element_arrays_bit,
	
	NUMBER_OF_FIND_MAP_ACTION_PARAMETER_FLAGS,
	
	_dont_resolve_links_flag= FLAG(_dont_resolve_links_bit),
	_return_empty_arrays_flag= FLAG(_return_empty_arrays_bit),
	_only_return_one_element_arrays_flag= FLAG(_only_return_one_element_arrays_bit)
};

/* ---------- globals */

extern struct datum_globals map_actions;

/* ---------- prototypes/MAP_ACTIONS.C */

void initialize_map_actions(void);
void initialize_map_actions_for_new_map(void);
void dispose_map_actions_from_old_map(void);

enum
{
	_map_action_debug_on,
	_map_action_debug_off,
	_map_action_debug_toggle
};
void set_map_action_debug_status(int new_status);

void update_map_actions(void);

int action_identifier_to_action_index(int action_identifier);

void set_map_action_success(int action_index);
void set_map_action_failure(int action_index);
void set_map_action_error(int action_index);

void deactivate_map_action(int action_index);
void activate_map_action(int action_index);
unsigned long trigger_map_action(int action_index);

int new_map_action(int parent_action_index);
void delete_map_action(int action_index);

boolean new_map_action_parameter(struct map_action_parameter *parameter, int action_index, field_tag field_name, int type);
void delete_map_action_parameter(struct map_action_parameter *parameter);
boolean set_map_action_parameter(struct map_action_parameter *parameter, int new_count, void *new_data);
boolean append_to_map_action_parameter(struct map_action_parameter *parameter, int count, void *data);
boolean copy_map_action_parameter(struct map_action_parameter *source, struct map_action_parameter *destination);
boolean set_unique_map_action_parameter(int action_index, field_tag field_name, int type, int count, void *data);

void build_map_action_parameter(struct map_action_parameter *parameter, int action_index, int parameter_index);
void set_map_action_parameter_type(struct map_action_parameter *parameter, int type);
void set_map_action_parameter_field_name(struct map_action_parameter *parameter, field_tag field_name);

void new_find_map_action_parameter_state(struct find_map_action_parameter_state *state,
	int action_index, unsigned long flags, unsigned long field_name_filter, int type_filter);
boolean find_next_map_action_parameter(struct find_map_action_parameter_state *state,
	struct map_action_parameter *result);
boolean get_map_action_parameter(struct map_action_parameter *result, int action_index,
	unsigned long flags, unsigned long field_name, int type);
int get_combined_map_action_parameter(int action_index, field_tag field_name, int type,
	void *data, int maximum_count);

boolean test_actions_for_success(short *action_identifiers, int count, boolean all);
void call_function_with_action_identifiers(void (*function)(int action_index), short *action_identifiers, int count);
#define trigger_action_identifiers(i, c) call_function_with_action_identifiers((void (*)(int))trigger_map_action, i, c)
#define activate_action_identifiers(i, c) call_function_with_action_identifiers(activate_map_action, i, c)
#define deactivate_action_identifiers(i, c) call_function_with_action_identifiers(deactivate_map_action, i, c)
#define delete_action_identifiers(i, c) call_function_with_action_identifiers(delete_map_action, i, c)

boolean call_functions_with_actions_in_parameter(void (*function)(int action_index), int action_index, field_tag field_name);
#define activate_actions_in_parameter(i, c) call_functions_with_actions_in_parameter(activate_map_action, i, c)
#define deactivate_actions_in_parameter(i, c) call_functions_with_actions_in_parameter(deactivate_map_action, i, c)

#define test_flag_field(action_index, field_name) get_map_action_parameter((struct map_action_parameter *) NULL, \
	action_index, _return_empty_arrays_flag, field_name, _parameter_flag)

void *extract_parameter_array_address(struct map_action_parameter *parameter);

#define extract_flag_array(p) ((short *)extract_parameter_array_address(p))
#define extract_monster_identifier_array(p) ((short *)extract_parameter_array_address(p))
#define extract_action_identifier_array(p) ((short *)extract_parameter_array_address(p))
#define extract_object_identifier_array(p) ((short *)extract_parameter_array_address(p))
#define extract_angle_array(p) ((angle *)extract_parameter_array_address(p))

#define extract_integer_array(p) ((int *)extract_parameter_array_address(p))
#define extract_world_distance_array(p) ((world_distance *)extract_parameter_array_address(p))
#define extract_field_name_array(p) ((field_tag *)extract_parameter_array_address(p))
#define extract_fixed_array(p) ((fixed *)extract_parameter_array_address(p))
#define extract_projectile_group_tag_array(p) ((file_tag *)extract_parameter_array_address(p))
#define extract_string_list_tag_array(p) ((file_tag *)extract_parameter_array_address(p))
#define extract_sound_tag_array(p) ((file_tag *)extract_parameter_array_address(p))

#define extract_world_point2d_array(p) ((world_point2d *)extract_parameter_array_address(p))
#define extract_world_rectangle2d_from_parameter(p) ((world_rectangle2d *)extract_parameter_array_address(p))
#define extract_world_point3d_array(p) ((world_point3d *)extract_parameter_array_address(p))

boolean get_map_action_structure(int action_index, field_tag field_name,
	int size, void *buffer);
#define set_map_action_structure(action_index, field_name, size, buffer) \
	set_unique_map_action_parameter(action_index, field_name, _parameter_character, size, buffer)

boolean get_indexed_byte_parameter(int action_index, field_tag field_name, int type, byte *result, int index);
boolean get_indexed_short_integer_parameter(int action_index, field_tag field_name, int type, short *short_integer, int index);
boolean get_indexed_long_integer_parameter(int action_index, field_tag field_name, int type, long *long_integer, int index);

boolean get_indexed_world_point2d_parameter(int action_index, field_tag field_name, world_point2d *point, int index);
boolean get_indexed_world_rectangle2d_parameter(int action_index, field_tag field_name, world_rectangle2d *rectangle, int index);
boolean get_indexed_world_point3d_parameter(int action_index, field_tag field_name, world_point3d *point, int index);

#define get_single_flag_parameter(action_index, field_name, pointer) get_indexed_byte_parameter(action_index, field_name, _parameter_flag, pointer, 0)
#define get_single_monster_identifier_parameter(action_index, field_name, pointer) get_indexed_short_integer_parameter(action_index, field_name, _parameter_monster_identifier, pointer, 0)
#define get_single_action_identifier_parameter(action_index, field_name, pointer) get_indexed_short_integer_parameter(action_index, field_name, _parameter_action_identifier, pointer, 0)
#define get_single_angle_parameter(action_index, field_name, pointer) ((angle)get_indexed_short_integer_parameter(action_index, field_name, _parameter_angle, (short *)pointer, 0))
#define get_single_object_identifier_parameter(action_index, field_name, pointer) get_indexed_short_integer_parameter(action_index, field_name, _parameter_object_identifier, pointer, 0)
#define get_single_model_identifier_parameter(action_index, field_name, pointer) get_indexed_short_integer_parameter(action_index, field_name, _parameter_model_identifier, pointer, 0)
#define get_single_sound_source_identifier_parameter(action_index, field_name, pointer) get_indexed_short_integer_parameter(action_index, field_name, _parameter_sound_source_identifier, pointer, 0)
#define get_single_local_projectile_group_identifier_parameter(action_index, field_name, pointer) get_indexed_short_integer_parameter(action_index, field_name, _parameter_local_projectile_group_identifier, pointer, 0)

#define get_single_integer_parameter(action_index, field_name, pointer) get_indexed_long_integer_parameter(action_index, field_name, _parameter_integer, (long *) pointer, 0)
#define get_single_world_distance_parameter(action_index, field_name, pointer) ((world_distance)get_indexed_long_integer_parameter(action_index, field_name, _parameter_world_distance, pointer, 0))
#define get_single_field_name_parameter(action_index, field_name, pointer) ((field_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_field_name, (long *) pointer, 0))
#define get_single_fixed_parameter(action_index, field_name, pointer) ((fixed)get_indexed_long_integer_parameter(action_index, field_name, _parameter_fixed, pointer, 0))
#define get_single_projectile_group_tag_parameter(action_index, field_name, pointer) ((file_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_projectile_group_tag, pointer, 0))
#define get_single_string_list_tag_parameter(action_index, field_name, pointer) ((file_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_string_list_tag, pointer, 0))
#define get_single_sound_tag_parameter(action_index, field_name, pointer) ((file_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_sound_tag, pointer, 0))
#define get_single_world_point3d_parameter(action_index, field_name, pointer) get_indexed_world_point3d_parameter(action_index, field_name, pointer, 0)
#define get_single_world_point2d_parameter(action_index, field_name, pointer) get_indexed_world_point2d_parameter(action_index, field_name, pointer, 0)
#define get_single_world_rectangle2d_parameter(action_index, field_name, pointer) get_indexed_world_world_rectangle2d_parameter(action_index, field_name, pointer, 0)

#define get_indexed_flag_parameter(action_index, field_name, pointer, index) get_indexed_byte_parameter(action_index, field_name, _parameter_flag, pointer, index)
#define get_indexed_monster_identifier_parameter(action_index, field_name, pointer, index) get_indexed_short_integer_parameter(action_index, field_name, _parameter_monster_identifier, pointer, index)
#define get_indexed_action_identifier_parameter(action_index, field_name, pointer, index) get_indexed_short_integer_parameter(action_index, field_name, _parameter_action_identifier, pointer, index)
#define get_indexed_angle_parameter(action_index, field_name, pointer, index) ((angle)get_indexed_short_integer_parameter(action_index, field_name, _parameter_angle, (short *)pointer, index))
#define get_indexed_object_identifier_parameter(action_index, field_name, pointer, index) get_indexed_short_integer_parameter(action_index, field_name, _parameter_object_identifier, pointer, index)
#define get_indexed_model_identifier_parameter(action_index, field_name, pointer, index) get_indexed_short_integer_parameter(action_index, field_name, _parameter_model_identifier, pointer, index)
#define get_indexed_sound_source_identifier_parameter(action_index, field_name, pointer, index) get_indexed_short_integer_parameter(action_index, field_name, _parameter_sound_source_identifier, pointer, index)
#define get_indexed_local_projectile_group_identifier_parameter(action_index, field_name, pointer, index) get_indexed_short_integer_parameter(action_index, field_name, _parameter_local_projectile_group_identifier, pointer, index)

#define get_indexed_integer_parameter(action_index, field_name, pointer, index) get_indexed_long_integer_parameter(action_index, field_name, _parameter_integer, (long *) pointer, index)
#define get_indexed_world_distance_parameter(action_index, field_name, pointer, index) ((world_distance)get_indexed_long_integer_parameter(action_index, field_name, _parameter_world_distance, pointer, index))
#define get_indexed_field_name_parameter(action_index, field_name, pointer, index) ((field_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_field_name, (long *) pointer, index))
#define get_indexed_fixed_parameter(action_index, field_name, pointer, index) ((fixed)get_indexed_long_integer_parameter(action_index, field_name, _parameter_fixed, pointer, index))
#define get_indexed_projectile_group_tag_parameter(action_index, field_name, pointer, index) ((file_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_projectile_group_tag, pointer, index))
#define get_indexed_string_list_tag_parameter(action_index, field_name, pointer, index) ((file_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_string_list_tag, pointer, index))
#define get_indexed_sound_tag_parameter(action_index, field_name, pointer, index) ((file_tag)get_indexed_long_integer_parameter(action_index, field_name, _parameter_sound_tag, pointer, index))

int get_parameter_size(int type, int count, boolean align);
#define get_unaligned_parameter_size(type, count) get_parameter_size(type, count, FALSE)
#define get_aligned_parameter_size(type, count) get_parameter_size(type, count, TRUE)

void map_action_debug(int action_index, const char *format, ...);
void map_action_error(int action_index, const char *format, ...);

#ifdef DEBUG
#define get_map_action(index) ((struct map_action *)get_datum(&map_actions, index))
#else
#define get_map_action(index) ((struct map_action *)map_actions.data + (index))
#endif

struct map_action *get_map_action_by_identifier(int identifier);

void import_map_actions(int map_action_count, int map_action_buffer_size, void *buffer, boolean byte_swap);
void *export_map_actions(long *map_action_count, long *map_action_buffer_size);

void offset_map_actions(world_distance dx, world_distance dy);

void preload_map_action_tags(void);

/* ---------- prototypes/MAP_ACTION_ACCESSORS.C */

void set_map_action_tag(int action_index, action_tag tag);
action_tag get_map_action_tag(int action_index);

int get_map_action_identifier(int action_index);

void set_map_action_name(int action_index, char *name);
char *get_map_action_name(int action_index);

void set_map_action_trigger_time(int action_index, int trigger_time_lower_bound, int trigger_time_delta);

#endif // __MAP_ACTIONS_H__