#ifndef __MESH_H__
#define __MESH_H__

/*
MESH.H
Wednesday, March 1, 1995 1:58:45 AM  (Jason')
*/

#include "world.h"
#include "mesh_file.h"
#include "definitions.h"

/* ---------- constants */

// the mesh is an array of submeshes
enum
{
	MAXIMUM_MESH_WIDTH= 256,
	MAXIMUM_MESH_HEIGHT= 256,

	// submeshes are 16x16 cells
 	SUBMESH_WIDTH_BITS= 4,
	SUBMESH_WIDTH= 1<<SUBMESH_WIDTH_BITS,
	MAXIMUM_SUBMESH_WIDTH= MAXIMUM_MESH_WIDTH>>SUBMESH_WIDTH_BITS,
	MAXIMUM_SUBMESH_HEIGHT= MAXIMUM_MESH_HEIGHT>>SUBMESH_WIDTH_BITS,

	// submesh textures are 256x256 pixels
	SUBMESH_TEXTURE_WIDTH_BITS= 8,
	SUBMESH_TEXTURE_WIDTH= 1<<SUBMESH_TEXTURE_WIDTH_BITS,
	
	// cells on the submesh are 16x16 pixels
	SUBMESH_CELL_TEXTURE_WIDTH_BITS= SUBMESH_TEXTURE_WIDTH_BITS-SUBMESH_WIDTH_BITS,
	SUBMESH_CELL_TEXTURE_WIDTH= 1<<SUBMESH_CELL_TEXTURE_WIDTH_BITS,
	
	// submeshes are 16x16 cells
	SUBMESH_CELL_WIDTH= SUBMESH_TEXTURE_WIDTH/SUBMESH_CELL_TEXTURE_WIDTH,

	// we'd rather the volatile mesh only be 26x26 but we also need to shift to read addresses
	VOLATILE_MESH_WIDTH_BITS= 5,
	VOLATILE_MESH_WIDTH= 1<<VOLATILE_MESH_WIDTH_BITS,
	VOLATILE_MESH_HEIGHT= 32,
	
	// the volatile texture maps are 512x512
	VOLATILE_TEXTURE_WIDTH_BITS= 9,
	VOLATILE_TEXTURE_WIDTH= VOLATILE_MESH_WIDTH*SUBMESH_CELL_TEXTURE_WIDTH,
	VOLATILE_TEXTURE_HEIGHT= VOLATILE_MESH_HEIGHT*SUBMESH_CELL_TEXTURE_WIDTH,

	VOLATILE_TEXTURE_WIDTH_UPSHIFT= LONG_BITS-2-VOLATILE_MESH_WIDTH_BITS,
	VOLATILE_TEXTURE_HEIGHT_UPSHIFT= LONG_BITS-2-VOLATILE_MESH_WIDTH_BITS,
	VOLATILE_TEXTURE_WIDTH_DOWNSHIFT= LONG_BITS-2-VOLATILE_TEXTURE_WIDTH_BITS,
	VOLATILE_TEXTURE_HEIGHT_DOWNSHIFT= LONG_BITS-2-2*VOLATILE_TEXTURE_WIDTH_BITS,
	VOLATILE_TEXTURE_WIDTH_MASK= VOLATILE_TEXTURE_WIDTH-1,
	VOLATILE_TEXTURE_HEIGHT_MASK= (VOLATILE_TEXTURE_WIDTH-1)<<VOLATILE_TEXTURE_WIDTH_BITS,

	VOLATILE_SUBTEXTURE_CELL_WIDTH= 3,
	VOLATILE_SUBTEXTURE_SLOP_PIXELS= 4,
	
	VOLATILE_SUBTEXTURE_WIDTH_BITS= 6,
	VOLATILE_SUBTEXTURE_WIDTH= 1<<VOLATILE_SUBTEXTURE_WIDTH_BITS,

	// media textures are 128x128
	MEDIA_TEXTURE_WIDTH_BITS= 7,
	MEDIA_TEXTURE_HEIGHT_BITS= MEDIA_TEXTURE_WIDTH_BITS,
	MEDIA_TEXTURE_WIDTH= 1<<MEDIA_TEXTURE_WIDTH_BITS,
	MEDIA_TEXTURE_HEIGHT= 1<<MEDIA_TEXTURE_HEIGHT_BITS,
	MEDIA_MESH_WIDTH= MEDIA_TEXTURE_WIDTH/(2*SUBMESH_CELL_TEXTURE_WIDTH),

	// upshifts are identical to SUBMESH_TEXTURE_WIDTH_UPSHIFT, SUBMESH_TEXTURE_HEIGHT_UPSHIFT
	MEDIA_TEXTURE_WIDTH_DOWNSHIFT= LONG_BITS-3-VOLATILE_TEXTURE_WIDTH_BITS,
	MEDIA_TEXTURE_HEIGHT_DOWNSHIFT= LONG_BITS-3-MEDIA_TEXTURE_WIDTH_BITS-VOLATILE_TEXTURE_WIDTH_BITS,
	MEDIA_TEXTURE_WIDTH_MASK= MEDIA_TEXTURE_WIDTH-1,
	MEDIA_TEXTURE_HEIGHT_MASK= (MEDIA_TEXTURE_HEIGHT-1)<<MEDIA_TEXTURE_WIDTH_BITS,

	VERTICES_PER_MESH_CELL= 4,
	
	MESH_MEDIA_HEIGHT= 0
};

#define SUBMESH_INDEX(x, y) (((y)>>SUBMESH_WIDTH_BITS)*submesh_width + ((x)>>SUBMESH_WIDTH_BITS))

/* ---------- macros */

/* ---------- structures */

// itd be nice to steal some bits for light intensity; sprites would interpolate these numbers

// we need bits for impassable-to-walking and impassable-to-flying; is it possible to walk
// under stuff you can't fly over?  it is certainly possible to fly over stuff you can't walk
// over; these are usually the results of polygonal scenery but could be precomputed from heights

// media height is not user-editable, but is calculated based on the deepest vertex

enum // terrain type (5 bits)
{
	_terrain_media_dwarf_depth, // dwarf-sized waist height
	_terrain_media_human_depth, // man-sized waist height
	_terrain_media_giant_depth, // giant-sized waist height
	_terrain_media_deep, // deeper than giant height
	_terrain_sloped,
	_terrain_steep,
	_terrain_grass,
	_terrain_desert,
	_terrain_rocky,
	_terrain_marsh,
	_terrain_snow,
	_terrain_forest, // identical to grass but prevents long-range sighting
	_terrain_unused1,
	_terrain_unused2,
	_terrain_walking_impassable,
	_terrain_flying_impassable,
	NUMBER_OF_TERRAIN_TYPES, // <=MAXIMUM_NUMBER_OF_TERRAIN_TYPES
	
	DEFAULT_TERRAIN_TYPE= _terrain_grass,
	FIRST_MEDIA_TERRAIN_TYPE= _terrain_media_dwarf_depth,
	NUMBER_OF_MEDIA_TERRAIN_TYPES= _terrain_media_deep-FIRST_MEDIA_TERRAIN_TYPE+1,

	FIRST_ELEVATION_TERRAIN_TYPE= _terrain_sloped,
	LAST_ELEVATION_TERRAIN_TYPE= _terrain_steep,
	NUMBER_OF_ELEVATION_TERRAIN_TYPES= LAST_ELEVATION_TERRAIN_TYPE-FIRST_ELEVATION_TERRAIN_TYPE,

	TERRAIN_TYPE_BITS= 4,
	MAXIMUM_NUMBER_OF_TERRAIN_TYPES= 1<<TERRAIN_TYPE_BITS,
	TERRAIN_TYPE_MASK= MAXIMUM_NUMBER_OF_TERRAIN_TYPES-1
};
#define GET_CELL_TERRAIN_TYPE(c,w) ((w) ? (((c)->flags)&TERRAIN_TYPE_MASK) : (((c)->flags>>TERRAIN_TYPE_BITS)&TERRAIN_TYPE_MASK))
#define GET_TERRAIN_TYPE(cell,x,y) GET_CELL_TERRAIN_TYPE(cell, (((x)^(y))&WORLD_ONE) ? ((y)&(WORLD_ONE-1))>((x)&(WORLD_ONE-1)) : ((y)&(WORLD_ONE-1))<(WORLD_ONE-((x)&(WORLD_ONE-1))))

#define TERRAIN_TYPE_IS_MEDIA(t) ((t)<FIRST_MEDIA_TERRAIN_TYPE+NUMBER_OF_MEDIA_TERRAIN_TYPES)

enum // mesh flags
{
	_mesh_vertex_is_media_bit= 2*TERRAIN_TYPE_BITS,
	_mesh_cell_is_on_fire_bit,
	_mesh_vertex_is_animated_media_bit,
	_mesh_cell_is_partially_media_bit, // some media pixels exist in the cells texture
	_mesh_cell_is_wholly_media_bit, // this cell is all media
	_mesh_cell_has_reflection_bit, // build a reflection for this cell and it's objects
	_mesh_cell_is_not_rendered_bit,
	_mesh_cell_has_model_bit,
	
	NUMBER_OF_MESH_FLAGS, // <=16
	
	_mesh_vertex_is_media_flag= FLAG(_mesh_vertex_is_media_bit),
	_mesh_cell_is_on_fire_flag= FLAG(_mesh_cell_is_on_fire_bit),
	_mesh_vertex_is_animated_media_flag= FLAG(_mesh_vertex_is_animated_media_bit),
	_mesh_cell_is_partially_media_flag= FLAG(_mesh_cell_is_partially_media_bit),
	_mesh_cell_is_wholly_media_flag= FLAG(_mesh_cell_is_wholly_media_bit),
	_mesh_cell_has_reflection_flag= FLAG(_mesh_cell_has_reflection_bit),
	_mesh_cell_is_not_rendered_flag= FLAG(_mesh_cell_is_not_rendered_bit),
	_mesh_cell_has_model_flag= FLAG(_mesh_cell_has_model_bit),

	_mesh_cell_was_scorched_bit= _mesh_vertex_is_animated_media_bit,
	_mesh_cell_was_scorched_flag= FLAG(_mesh_cell_was_scorched_bit)
};

#define SIZEOF_STRUCT_MESH_CELL 8
struct mesh_cell
{
	short_world_distance height; // (might have a few free bits here)
	word normal; // two 8-bit indexes into the precalculated normal table
	word flags; // includes terrain type of each triangle

	short first_object_index;
};

/* ---------- globals */

extern short mesh_width, mesh_height;
extern short submesh_width, submesh_height;
extern struct mesh_cell *mesh_cells;

extern struct definition_globals mesh_header;

/* ---------- prototypes/MESH.C */

void initialize_mesh(void);
boolean initialize_mesh_for_new_map(file_tag tag);
void dispose_mesh_from_old_map(void);

#define get_mesh_tag_header() get_definition_tag_header(&mesh_header, 0)
#define get_mesh_name() get_definition_name(&mesh_header, 0)
#ifdef DEBUG
#define get_mesh_file_header() ((struct mesh_header *)get_definition(&mesh_header, 0))
#else
#define get_mesh_file_header() ((struct mesh_header *)(mesh_header.definitions))
#endif

// safe_get_mesh_file_header() can return NULL, but it can be called anytime
#define safe_get_mesh_file_header() ((struct mesh_header *)(mesh_header.definitions))


#define get_first_mesh_cell() get_mesh_cell(0, 0)
struct mesh_cell *get_mesh_cell(int x, int y);

void *get_reflection_map_base_address(void);
void *get_reflection_map_shading_table(void);

//void *get_media_map_shading_table(void);
void *get_color_map_shading_table(void);
void *get_color_map_blend_table(void);

//void *get_media_map_base_address(void);
void *get_submesh_color_map_base_address(short submesh_index, boolean write_enable);
void *get_submesh_shadow_map_base_address(short submesh_index, boolean write_enable);

void stain_mesh(int collection_index, int sequence_index, world_point2d *p);
void bloody_mesh(world_point2d *p);

byte get_mesh_shadow_map_pixel(int x, int y);
byte get_mesh_color_map_pixel(int x, int y);

short_world_distance get_mesh_height(world_point2d *p);
world_vector3d *get_mesh_normal(world_point2d *p);
int get_mesh_terrain_type(world_point2d *p);

int get_mesh_triangle_index(int x, int y);

boolean test_cell_elevation(int x, int y, int maximum_elevation);

void precalculate_mesh_normals(int x0, int y0, int x1, int y1);

void postprocess_mesh(int x0, int y0, int x1, int y1);
void reset_mesh_terrain_types(int x0, int y0, int x1, int y1);

int get_mesh_media_projectile_group_type(int projectile_group_index);

boolean read_mesh_header(file_tag subgroup, struct mesh_header *mesh_header);

void build_media_local_projectile_effects(struct view_data *view, struct mesh_region *region);

/* ---------- prototypes/MESH_FILE.C */

void assert_mesh_file_constants(void);

boolean build_new_mesh(struct tag_header *mesh_tag_header, file_tag media_tag, int horizontal_submesh_count, int vertical_submesh_count);
boolean write_mesh_to_tag(void);

/* ---------- prototypes/MESH_UTILITIES.C */

void get_mesh_triangle_verticies(world_point2d *p, point2d *a, point2d *b, point2d *c);
void get_mesh_cell_verticies(world_point2d *p, point2d *a, point2d *b, point2d *c, point2d *d);

void get_vertex_from_cell(struct mesh_cell *cell, point2d *p);

/* ---------- prototypes/MESH_BLOOD.C */

void build_blood_remapping_table(void);

#endif // __MESH_H__
