// =================================================================================
//	C256Tag.cp					 1998 Hans Hvidsten Birkeland. All rights reserved.
// =================================================================================

#include "C256Tag.h"

#include "CTagHeader.h"
#include "CCollectionHeader.h"
#include "CColorTable.h"
#include "CHueChange.h"
#include "CBitmapInstance.h"
#include "CSequence.h"
#include "CShadowMap.h"
#include "CBlendTable.h"
#include "CRemapTable.h"
#include "CBitmap.h"

#include "CTagElementArray.h"

#include "AmberMessages.h"
#include "256Messages.h"

#include "StNewBlock.h"

// =================================================================================
//	Constants																 [local]
// =================================================================================

const char C256Tag::sCopyrightNotice[] = "This tag was saved by Amber 2.0";


// =================================================================================
//	 C256Tag																[public]
// =================================================================================
// Default constructor
					
C256Tag::C256Tag( void )
{
	mTagHeader->SetTagGroup( '.256' );
	
	mCollectionHeader = new CCollectionHeader;
	mColorTables = new CTagElementArray;
	mHueChanges = new CTagElementArray;
	mSequences = new CTagElementArray;
	mBlendTables = new CTagElementArray;
	mRemapTables = new CTagElementArray;
	mColorBitmaps = new CTagElementArray;
	mShadowBitmaps = new CTagElementArray;
}


// =================================================================================
//	 ~C256Tag																[public]
// =================================================================================

C256Tag::~C256Tag( void )
{
	delete mCollectionHeader;
	delete mColorTables;
	delete mHueChanges;
	delete mSequences;
	delete mBlendTables;
	delete mRemapTables;
	delete mColorBitmaps;
	delete mShadowBitmaps;
}


#pragma mark -

// =================================================================================
//	 ReadTagData															[public]
// =================================================================================
// inStream is a stream containing the tag data as they are stored in the .gor
// files and plugin files.

void
C256Tag::ReadStream( LStream &inStream, Int32 inOffset )
{
	// Offset, count and size of element array
	Int32 theOffset;
	Int32 theCount;
	Int32 theSize;
	
	// Don't listen to all the elements getting initialized
	StopListening();
	
	// The Collection Header
	const Int32 theHeaderSize = mCollectionHeader->GetSize();
	StNewBlock theCollectionHeader( theHeaderSize );
	
	inStream.SetMarker( inOffset, streamFrom_Start );
	inStream.ReadBlock( theCollectionHeader, theHeaderSize );
	
	mCollectionHeader->ReadData( theCollectionHeader );
	
	// Offset to tag data
	const Int32 theDataOffset = mCollectionHeader->GetDataOffset() + inOffset;
	
	// Color Tables
	theOffset = mCollectionHeader->GetColorTablesOffset();
	theCount = mCollectionHeader->GetColorTableCount();
	theSize = mCollectionHeader->GetColorTablesSize();
	
	if ( theCount > 0 ) {
		if ( theSize <= 0 )
			Throw_(err_WrongTagFormat);
		
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		inStream.ReadBlock( theBuffer, theSize );
		
		for ( Int32 i = 0; i < theCount; i++ ) {
			CColorTable *theColorTable = new CColorTable;
			mColorTables->InsertItemsAt( 1, LArray::index_Last, &theColorTable );
			theColorTable->AddListener( this );
			
			theColorTable->ReadData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theColorTable->GetSize();
			
			if ( theElementOffset > theSize )
				Throw_(err_WrongTagFormat);
		}
	}
	
	// Hue Changes
	theOffset = mCollectionHeader->GetHueChangesOffset();
	theCount = mCollectionHeader->GetHueChangeCount();
	theSize = mCollectionHeader->GetHueChangesSize();
	
	if ( theCount*CHueChange::kElementSize != theSize )
		Throw_(err_WrongTagFormat);
	
	if ( theCount > 0 ) {
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		inStream.ReadBlock( theBuffer, theSize );
		
		for ( Int32 i = 0; i < theCount; i++ ) {
			CHueChange *theHueChange = new CHueChange;
			mHueChanges->InsertItemsAt( 1, LArray::index_Last, &theHueChange );
			theHueChange->AddListener( this );
			
			theHueChange->ReadData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theHueChange->GetSize();
			
			if ( theElementOffset > theSize )
				Throw_(err_WrongTagFormat);
		}
	}
	
	// Bitmap Instances
	theOffset = mCollectionHeader->GetBitmapInstancesOffset();
	theCount = mCollectionHeader->GetBitmapInstanceCount();
	theSize = mCollectionHeader->GetBitmapInstancesSize();
	
	if ( theCount*CBitmapInstance::kElementSize != theSize )
		Throw_(err_WrongTagFormat);
	
	StNewBlock theBitmapInstanceBlock( theSize );
	
	inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
	inStream.ReadBlock( theBitmapInstanceBlock, theSize );
	
	// Shadow Maps
	theOffset = mCollectionHeader->GetShadowMapsOffset();
	theCount = mCollectionHeader->GetShadowMapCount();
	theSize = mCollectionHeader->GetShadowMapsSize();
	
	if ( theCount*CShadowMap::kElementSize != theSize )
		Throw_(err_WrongTagFormat);
	
	StNewBlock theShadowMapBlock( theSize );
	
	inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
	inStream.ReadBlock( theShadowMapBlock, theSize );
	
	// Blend Tables
	theOffset = mCollectionHeader->GetBlendTablesOffset();
	theCount = mCollectionHeader->GetBlendTableCount();
	theSize = mCollectionHeader->GetBlendTablesSize();
	
	if ( theCount > 0 ) {
		if ( theSize <= 0 )
			Throw_(err_WrongTagFormat);
		
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		inStream.ReadBlock( theBuffer, theSize );
		
		for ( Int32 i = 0; i < theCount; i++ ) {
			CBlendTable *theBlendTable = new CBlendTable;
			mBlendTables->InsertItemsAt( 1, LArray::index_Last, &theBlendTable );
			theBlendTable->AddListener( this );
			
			theBlendTable->ReadData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theBlendTable->GetSize();
			
			if ( theElementOffset > theSize )
				Throw_(err_WrongTagFormat);
		}
	}
	
	// Remap Tables
	theOffset = mCollectionHeader->GetRemapTablesOffset();
	theCount = mCollectionHeader->GetRemapTableCount();
	theSize = mCollectionHeader->GetRemapTablesSize();
	
	if ( theCount > 0 ) {
		if ( theSize <= 0 )
			Throw_(err_WrongTagFormat);
		
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		inStream.ReadBlock( theBuffer, theSize );
		
		for ( Int32 i = 0; i < theCount; i++ ) {
			CRemapTable *theRemapTable = new CRemapTable;
			mRemapTables->InsertItemsAt( 1, LArray::index_Last, &theRemapTable );
			theRemapTable->AddListener( this );
			
			theRemapTable->ReadData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theRemapTable->GetSize();
			
			if ( theElementOffset > theSize )
				Throw_(err_WrongTagFormat);
		}
	}
	
	// Bitmap References and Data (will be sorted later)
	theOffset = mCollectionHeader->GetBitmapsOffset();
	theCount = mCollectionHeader->GetBitmapCount();
	theSize = mCollectionHeader->GetBitmapsSize();
	
	CTagElementArray theBitmaps;
	
	if ( theCount > 0 ) {
		if ( theSize <= 0 )
			Throw_(err_WrongTagFormat);
		
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		inStream.ReadBlock( theBuffer, theSize );
		
		for ( Int32 i = 0; i < theCount; i++ ) {
			CBitmap *theBitmap = new CBitmap;
			theBitmaps.InsertItemsAt( 1, LArray::index_Last, &theBitmap );
			theBitmap->AddListener( this );
			
			theBitmap->ReadReference( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theBitmap->GetReferenceSize();
			
			if ( theElementOffset > theSize )
				Throw_(err_WrongTagFormat);
			
			// Read Bitmap Data too
			Int32 theBitmapDataOffset = theBitmap->GetDataOffset();
			Int32 theBitmapDataSize = theBitmap->GetDataSize();
			
			if ( theBitmapDataOffset <= 0 || theBitmapDataSize <= 0 )
				Throw_(err_WrongTagFormat);
			
			StNewBlock theDataBuffer( theBitmapDataSize );
			
			inStream.SetMarker( theBitmapDataOffset + theDataOffset, streamFrom_Start );
			inStream.ReadBlock( theDataBuffer, theBitmapDataSize );
			
			theBitmap->ReadData( theDataBuffer );
			
			if ( theBitmap->GetDataSize() > theBitmapDataSize )
				Throw_(err_WrongTagFormat);
			
		}
	}
	
	// Sequences and Sequence Data (last because they depend on the others)
	theOffset = mCollectionHeader->GetSequencesOffset();
	theCount = mCollectionHeader->GetSequenceCount();
	theSize = mCollectionHeader->GetSequencesSize();
	
	if ( theCount > 0 ) {
		if ( theSize <= 0 )
			Throw_(err_WrongTagFormat);
		
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		inStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		inStream.ReadBlock( theBuffer, theSize );
		
		for ( Int32 i = 0; i < theCount; i++ ) {
			CSequence *theSequence = new CSequence;
			mSequences->InsertItemsAt( 1, LArray::index_Last, &theSequence );
			theSequence->AddListener( this );
			
			theSequence->ReadReference( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theSequence->GetReferenceSize();
			
			if ( theElementOffset > theSize )
				Throw_(err_WrongTagFormat);
			
			// Read Sequence Data too
			Int32 theSeqDataOffset = theSequence->GetDataOffset();
			Int32 theSeqDataSize = theSequence->GetRefDataSize();
			
			if ( theSeqDataOffset <= 0 || theSeqDataSize <= 0 )
				Throw_(err_WrongTagFormat);
			
			StNewBlock theDataBuffer( theSeqDataSize );
			
			inStream.SetMarker( theSeqDataOffset + theDataOffset, streamFrom_Start );
			inStream.ReadBlock( theDataBuffer, theSeqDataSize );
			
			theSequence->ReadData( theDataBuffer );
			
			Int32 theInstanceCount = mCollectionHeader->GetBitmapInstanceCount();
			Int32 theShadowMapCount = mCollectionHeader->GetShadowMapCount();
			
			if ( theSequence->GetDataSize() > theSeqDataSize
			||   !theSequence->AssignBitmapInstances( theBitmapInstanceBlock, theInstanceCount, theBitmaps )
			||   !theSequence->AssignShadowMaps( theShadowMapBlock, theShadowMapCount, theBitmaps ) )
				Throw_(err_WrongTagFormat);
			
		}
	}
	
	// Sort the bitmaps into color and shadow bitmaps
	LArrayIterator theIterator( theBitmaps );
	CBitmap *theBitmap;
	
	while ( theIterator.Next( &theBitmap ) ) {
		theBitmaps.Remove( &theBitmap );
		
		if ( theBitmap->GetFlags() & CBitmap::flag_GreyscaleBitmap )
			mShadowBitmaps->InsertItemsAt( 1, LArray::index_Last, &theBitmap );
		else
			mColorBitmaps->InsertItemsAt( 1, LArray::index_Last, &theBitmap );
	}
	
	// Update the hue changed color table
	UpdateHueChangedColorTable();
	
	// Restore order
	SetDirty( false );
	StartListening();
	BroadcastMessage( msg_256UpdateTag, nil );
}
	

// =================================================================================
//	 WriteStream															[public]
// =================================================================================
// outStream is filled with the tag data as they are stored on file.
					
void
C256Tag::WriteStream( LStream &outStream, Int32 inOffset )
{
	// We're too busy to listen to tag elements now :)
	StopListening();
	
	// An array containing all the bitmaps (color + shadow)
	const Int32 theBitmapCount = mColorBitmaps->GetCount() + mShadowBitmaps->GetCount();
	LArray theBitmaps( sizeof(void *) );
	
	LArrayIterator theColorIterator( *mColorBitmaps );
	LArrayIterator theShadowIterator( *mShadowBitmaps );
	CBitmap *theBitmap;
	
	while ( theColorIterator.Next( &theBitmap ) )
		theBitmaps.InsertItemsAt( 1, LArray::index_Last, &theBitmap );
	
	while ( theShadowIterator.Next( &theBitmap ) )
		theBitmaps.InsertItemsAt( 1, LArray::index_Last, &theBitmap );
	
	// Gather arrays of Bitmap Instances and Shadow Maps, assign indices
	LArray theBitmapInstances( sizeof(void *) );
	LArray theShadowMaps( sizeof(void *) );
	
	if ( mSequences->GetCount() > 0 ) {
		LArrayIterator theIterator( *mSequences );
		CSequence *theSequence;
		
		// Append element pointers to the arrays, and note indices
		while ( theIterator.Next( &theSequence ) )
			theSequence->DeclareElements( theBitmapInstances, theShadowMaps, theBitmaps );
	}
	
	// Offset, count and size of element array
	const Int32 theDataOffset = mCollectionHeader->GetSize() + inOffset;
	Int32 theOffset;
	Int32 theCount;
	Int32 theSize;
	
	theOffset = 0;
	
	// Amber copyright notice
	theSize = sizeof(sCopyrightNotice);
	
	outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
	outStream.WriteBlock( sCopyrightNotice, theSize );
	
	theOffset += theSize;
	
	// Color Tables
	theCount = mColorTables->GetCount();
	theSize = 0;
	
	if ( theCount > 0 ) {
		LArrayIterator theIterator( *mColorTables );
		CColorTable *theColorTable;
		
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		
		while ( theIterator.Next( &theColorTable ) ) {
			Int32 theLength = theColorTable->GetSize();
			StNewBlock theBuffer( theLength );
			
			theColorTable->WriteData( theBuffer );
			outStream.WriteBlock( theBuffer, theLength );
			
			theSize += theLength;
		}
		
	}
	
	mCollectionHeader->SetColorTableCount( theCount );
	mCollectionHeader->SetColorTablesOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetColorTablesSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize;
	
	// Hue Changes
	theCount = mHueChanges->GetCount();
	theSize = (CHueChange::kElementSize)*theCount;
	
	if ( theCount > 0 ) {
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		LArrayIterator theIterator( *mHueChanges );
		CHueChange *theHueChange;
		
		while ( theIterator.Next( &theHueChange ) ) {
			theHueChange->WriteData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theHueChange->GetSize();
		}
		
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		outStream.WriteBlock( theBuffer, theSize );
	}
	
	mCollectionHeader->SetHueChangeCount( theCount );
	mCollectionHeader->SetHueChangesOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetHueChangesSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize;
	
	// Bitmap Instances
	theCount = theBitmapInstances.GetCount();
	theSize = (CBitmapInstance::kElementSize)*theCount;
	
	if ( theCount > 0 ) {
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		LArrayIterator theIterator( theBitmapInstances );
		CBitmapInstance *theInstance;
		
		while ( theIterator.Next( &theInstance ) ) {
			theInstance->WriteData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theInstance->GetSize();
		}
		
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		outStream.WriteBlock( theBuffer, theSize );
	}
	
	mCollectionHeader->SetBitmapInstanceCount( theCount );
	mCollectionHeader->SetBitmapInstancesOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetBitmapInstancesSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize;
	
	// Sequences and Sequence Data
	theCount = mSequences->GetCount();
	theSize = (CSequence::kElementSize)*theCount;
	
	Int32 theSequenceDataSize = 0;
	
	if ( theCount > 0 ) {
		StNewBlock theBuffer( theSize );
		Int32 theSeqDataOffset = theOffset + theSize;
		Int32 theElementOffset = 0;
		
		LArrayIterator theIterator( *mSequences );
		CSequence *theSequence;
		
		while ( theIterator.Next( &theSequence ) ) {
		
			// Buffer the Sequence References
			theSequence->SetDataOffset( theSeqDataOffset );
			theSequence->WriteReference( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theSequence->GetReferenceSize();
			
			// Write Sequence Data
			Int32 theSeqDataSize = theSequence->GetDataSize();
			StNewBlock theDataBuffer( theSeqDataSize );
				
			theSequence->WriteData( theDataBuffer );
			outStream.SetMarker( theSeqDataOffset + theDataOffset, streamFrom_Start );
			outStream.WriteBlock( theDataBuffer, theSeqDataSize );
			
			theSequenceDataSize += theSeqDataSize;
			theSeqDataOffset += theSeqDataSize;
		}
		
		// Write the Sequence References
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		outStream.WriteBlock( theBuffer, theSize );
	}
	
	mCollectionHeader->SetSequenceCount( theCount );
	mCollectionHeader->SetSequencesOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetSequencesSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize + theSequenceDataSize;
	
	// Shadow Maps
	theCount = theShadowMaps.GetCount();
	theSize = (CShadowMap::kElementSize)*theCount;
	
	if ( theCount > 0 ) {
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		LArrayIterator theIterator( theShadowMaps );
		CShadowMap *theShadow;
		
		while ( theIterator.Next( &theShadow ) ) {
			theShadow->WriteData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theShadow->GetSize();
		}
		
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		outStream.WriteBlock( theBuffer, theSize );
	}
	
	mCollectionHeader->SetShadowMapCount( theCount );
	mCollectionHeader->SetShadowMapsOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetShadowMapsSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize;
	
	// Blend Tables
	theCount = mBlendTables->GetCount();
	theSize = (CBlendTable::kElementSize)*theCount;
	
	if ( theCount > 0 ) {
		if ( theSize <= 0 )
			Throw_(err_WrongTagFormat);
		
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		LArrayIterator theIterator( *mBlendTables );
		CBlendTable *theBlendTable;
		
		while ( theIterator.Next( &theBlendTable ) ) {
			theBlendTable->WriteData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theBlendTable->GetSize();
		}
		
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		outStream.WriteBlock( theBuffer, theSize );
	}
	
	mCollectionHeader->SetBlendTableCount( theCount );
	mCollectionHeader->SetBlendTablesOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetBlendTablesSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize;
	
	// Remap Tables
	theCount = mRemapTables->GetCount();
	theSize = (CRemapTable::kElementSize)*theCount;
	
	if ( theCount > 0 ) {
		StNewBlock theBuffer( theSize );
		Int32 theElementOffset = 0;
		
		LArrayIterator theIterator( *mRemapTables );
		CRemapTable *theRemapTable;
		
		while ( theIterator.Next( &theRemapTable ) ) {
			theRemapTable->WriteData( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theRemapTable->GetSize();
		}
		
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		outStream.WriteBlock( theBuffer, theSize );
	}
	
	mCollectionHeader->SetRemapTableCount( theCount );
	mCollectionHeader->SetRemapTablesOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetRemapTablesSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize;
	
	// Bitmap References and Data
	theCount = theBitmaps.GetCount();
	theSize = (CBitmap::kElementSize)*theCount;
	
	Int32 theBitmapDataSize = 0;
	
	if ( theCount > 0 ) {
		Int32 theBitmapDataOffset = theOffset + theSize;
		Int32 theElementOffset = 0;
		StNewBlock theBuffer( theSize );
		
		LArrayIterator theIterator( theBitmaps );
		CBitmap *theBitmap;
		Int32 theIndex = 0;
		
		
		while ( theIterator.Next( &theBitmap ) ) {
			
			// Buffer the Bitmap Reference
			theBitmap->SetDataOffset( theBitmapDataOffset );
			theBitmap->WriteReference( (Byte *)theBuffer + theElementOffset );
			theElementOffset += theBitmap->GetReferenceSize();
			
			// Write the Bitmap Data
			Int32 theDataSize = theBitmap->GetDataSize();
			StNewBlock theDataBuffer( theDataSize );
			
			theBitmap->WriteData( theDataBuffer );
			outStream.SetMarker( theBitmapDataOffset + theDataOffset, streamFrom_Start );
			outStream.WriteBlock( theDataBuffer, theDataSize );
			
			theBitmapDataSize += theDataSize;
			theBitmapDataOffset += theDataSize;
		}
		
		// Write the Bitmap References
		outStream.SetMarker( theOffset + theDataOffset, streamFrom_Start );
		outStream.WriteBlock( theBuffer, theSize );
	}
	
	mCollectionHeader->SetBitmapCount( theCount );
	mCollectionHeader->SetBitmapsOffset( theCount == 0 ? 0 : theOffset );
	mCollectionHeader->SetBitmapsSize( theCount == 0 ? 0 : theSize );
	
	theOffset += theSize + theBitmapDataSize;
	
	// The Collection Header
	StNewBlock theBuffer( mCollectionHeader->GetSize() );
	mCollectionHeader->SetDataSize( theOffset );
	mCollectionHeader->SetDataOffset( mCollectionHeader->GetSize() );
	mCollectionHeader->WriteData( theBuffer );
	
	outStream.SetMarker( inOffset, streamFrom_Start );
	outStream.WriteBlock( theBuffer, mCollectionHeader->GetSize() );
	
	StartListening();
}


// =================================================================================
//	 GetSize														 [const, public]
// =================================================================================
// Return the tag data length in bytes

Int32
C256Tag::GetSize( void ) const
{
	Int32 theSize = 0;
	
	// Collection header
	theSize += mCollectionHeader->GetSize();
	
	// Copyright notice
	theSize += sizeof( sCopyrightNotice );
	
	// Color tables
	LArrayIterator theColorTableIterator( *mColorTables );
	CColorTable *theColorTable;
	
	while ( theColorTableIterator.Next( &theColorTable ) )
		theSize += theColorTable->GetSize();
	
	// Hue changes
	theSize += (CHueChange::kElementSize)*mHueChanges->GetCount();
	
	// Sequences
	LArrayIterator theSequenceIterator( *mSequences );
	CSequence *theSequence;
	
	while ( theSequenceIterator.Next( &theSequence ) )
		theSize += theSequence->GetSize();
	
	// Blend tables
	theSize += (CBlendTable::kElementSize)*mBlendTables->GetCount();
	
	// Remap tables
	theSize += (CRemapTable::kElementSize)*mRemapTables->GetCount();
	
	// Bitmaps
	LArrayIterator theColorIterator( *mColorBitmaps );
	LArrayIterator theShadowIterator( *mShadowBitmaps );
	CBitmap *theBitmap;
	
	while ( theColorIterator.Next( &theBitmap ) )
		theSize += theBitmap->GetSize();
	
	while ( theShadowIterator.Next( &theBitmap ) )
		theSize += theBitmap->GetSize();
	
	return theSize;
}


#pragma mark -

// =================================================================================
//	 ListenToMessage														[public]
// =================================================================================

void
C256Tag::ListenToMessage( MessageT inMessage, void *ioData )
{
	switch ( inMessage ) {
		default: // Pass it on
			BroadcastMessage( inMessage, ioData );
			break;
		
		case msg_TagDeleteElement:
			#pragma remove it from the list
			break;
		
		case msg_256UpdateColorTable:
			UpdateHueChangedColorTable();
			BroadcastMessage( inMessage, ioData );
			break;
			
		case msg_256UpdateHueChange:
			if ( ioData != nil )
				UpdateHueChangedColorTable();
			
			BroadcastMessage( inMessage, ioData );
			break;
	}
}


#pragma mark -

// =================================================================================
//	 GetUserDataFlags												 [const, public]
// =================================================================================

Uint32
C256Tag::GetUserDataFlags( void ) const
{
	return mCollectionHeader->GetUserDataFlags();
}


#pragma mark -

// =================================================================================
//	 GetColorTableCount											 [const, public]
// =================================================================================
// Return the number of color tables in the tag.

Int32
C256Tag::GetColorTableCount( void ) const
{
	return mColorTables->GetCount();
}


// =================================================================================
//	 GetHueChangeCount												 [const, public]
// =================================================================================
// Return the number of hue changes in the tag.

Int32
C256Tag::GetHueChangeCount( void ) const
{
	return mHueChanges->GetCount();
}


// =================================================================================
//	 GetSequenceCount												 [const, public]
// =================================================================================
// Return the number of sequences in the tag.

Int32
C256Tag::GetSequenceCount( void ) const
{
	return mSequences->GetCount();
}


// =================================================================================
//	 GetBlendTableCount											 [const, public]
// =================================================================================
// Return the number of blending tables in the tag.

Int32
C256Tag::GetBlendTableCount( void ) const
{
	return mBlendTables->GetCount();
}


// =================================================================================
//	 GetRemapTableCount											 [const, public]
// =================================================================================
// Return the number of remap tables in the tag.

Int32
C256Tag::GetRemapTableCount( void ) const
{
	return mRemapTables->GetCount();
}


// =================================================================================
//	 GetBitmapCount												 [const, public]
// =================================================================================
// Return the number of bitmaps in the tag.

Int32
C256Tag::GetBitmapCount( void ) const
{
	return mColorBitmaps->GetCount() + mShadowBitmaps->GetCount();
}


// =================================================================================
//	 GetColorBitmapCount											 [const, public]
// =================================================================================
// Return the number of color bitmaps in the tag.

Int32
C256Tag::GetColorBitmapCount( void ) const
{
	return mColorBitmaps->GetCount();
}


// =================================================================================
//	 GetShadowBitmapCount											 [const, public]
// =================================================================================
// Return the number of shadow bitmaps in the tag.

Int32
C256Tag::GetShadowBitmapCount( void ) const
{
	return mShadowBitmaps->GetCount();
}


#pragma mark -

// =================================================================================
//	 GetColorTable													 [const, public]
// =================================================================================
// Return a const pointer to the tag element with the given index

const CColorTable *
C256Tag::GetColorTable( Int32 inIndex ) const
{
	CColorTable *theColorTable;
	
	if ( mColorTables->FetchItemAt( inIndex+1, &theColorTable ) )
		return theColorTable;
	
	return nil;
}


// =================================================================================
//	 GetColorTable															[public]
// =================================================================================
// Return a pointer to the tag element with the given index

CColorTable *
C256Tag::GetColorTable( Int32 inIndex )
{
	CColorTable *theColorTable;
	
	if ( mColorTables->FetchItemAt( inIndex+1, &theColorTable ) )
		return theColorTable;
	
	return nil;
}


// =================================================================================
//	 MoveColorTable														[public]
// =================================================================================
// Move the tag element to a different index.  Nothing is changed if the new index
// is invalid.  Returns true if the element was successfully moved.

Boolean
C256Tag::MoveColorTable( CColorTable *inColorTable, Int32 inNewIndex )
{
	// LArray is 1-based while the tag is 0-based
	inNewIndex += 1;
	
	ArrayIndexT theIndex = mColorTables->FetchIndexOf( &inColorTable );
	
	if ( theIndex == LArray::index_Bad
	||   !mColorTables->ValidIndex( inNewIndex ) )
		return false;
	
	mColorTables->MoveItem( theIndex, inNewIndex );
	return true;
}


// =================================================================================
//	 InsertColorTable														[public]
// =================================================================================
// Insert a tag element into the array.  Returns true if the element was
// successfully inserted.  Array elements will be DELETED when the array is
// deleted.

Boolean
C256Tag::InsertColorTable( CColorTable *inColorTable, Int32 inAtIndex )
{
	if ( inColorTable == nil )
		return false;
	
	mColorTables->InsertItemsAt( 1, inAtIndex+1, &inColorTable );
	return true;
}


// =================================================================================
//	 RemoveColorTable														[public]
// =================================================================================
// Remove the tag element from the array.  Returns true if the element was
// successfully removed.  Does NOT delete the element. 

Boolean
C256Tag::RemoveColorTable( CColorTable *inColorTable )
{
	ArrayIndexT theIndex = mColorTables->FetchIndexOf( &inColorTable );
	
	if ( theIndex == LArray::index_Bad )
		return false;
	
	mColorTables->Remove( &inColorTable );
	return true;
}


// =================================================================================
//	 GetColorTable													 [const, public]
// =================================================================================
// Copy the colors of the given color table into outColorTable

Boolean
C256Tag::GetColorTable( RGBColor *outColorTable, Int32 inIndex ) const
{
	Assert_(outColorTable != nil);
	
	// Find the color table
	const CColorTable *theColorTable = GetColorTable( inIndex );
	
	if ( theColorTable == nil )
		return false;
	
	// Copy the colors
	theColorTable->GetColorTable( outColorTable );
	
	return true;
}


// =================================================================================
//	 SetColorTable															[public]
// =================================================================================
// Copy the colors of inColorTable into the given color table

Boolean
C256Tag::SetColorTable( const RGBColor *inColorTable, Int32 inIndex )
{
	Assert_(inColorTable != nil);
	
	// Find the color table
	CColorTable *theColorTable = GetColorTable( inIndex );
	
	if ( theColorTable == nil )
		return false;
	
	// Copy the colors
	theColorTable->SetColorTable( inColorTable );
	
	return true;
}


#pragma mark -

// =================================================================================
//	 GetHueChange													 [const, public]
// =================================================================================
// Return a const pointer to the tag element with the given index

const CHueChange *
C256Tag::GetHueChange( Int32 inIndex ) const
{
	CHueChange *theHueChange;
	
	if ( mHueChanges->FetchItemAt( inIndex+1, &theHueChange ) )
		return theHueChange;
	
	return nil;
}


// =================================================================================
//	 GetHueChange															[public]
// =================================================================================
// Return a pointer to the tag element with the given index

CHueChange *
C256Tag::GetHueChange( Int32 inIndex )
{
	CHueChange *theHueChange;
	
	if ( mHueChanges->FetchItemAt( inIndex+1, &theHueChange ) )
		return theHueChange;
	
	return nil;
}


// =================================================================================
//	 MoveHueChange															[public]
// =================================================================================
// Move the tag element to a different index.  Nothing is changed if the new index
// is invalid.  Returns true if the element was successfully moved.

Boolean
C256Tag::MoveHueChange( CHueChange *inHueChange, Int32 inNewIndex )
{
	// LArray is 1-based while the tag is 0-based
	inNewIndex += 1;
	
	ArrayIndexT theIndex = mHueChanges->FetchIndexOf( &inHueChange );
	
	if ( theIndex == LArray::index_Bad
	||   !mHueChanges->ValidIndex( inNewIndex ) )
		return false;
	
	mHueChanges->MoveItem( theIndex, inNewIndex );
	return true;
}


// =================================================================================
//	 InsertHueChange														[public]
// =================================================================================
// Insert a tag element into the array.  Returns true if the element was
// successfully inserted.  Array elements will be DELETED when the array is
// deleted.

Boolean
C256Tag::InsertHueChange( CHueChange *inHueChange, Int32 inAtIndex )
{
	if ( inHueChange == nil )
		return false;
	
	mHueChanges->InsertItemsAt( 1, inAtIndex+1, &inHueChange );
	return true;
}


// =================================================================================
//	 RemoveHueChange														[public]
// =================================================================================
// Remove the tag element from the array.  Returns true if the element was
// successfully removed.  Does NOT delete the element. 

Boolean
C256Tag::RemoveHueChange( CHueChange *inHueChange )
{
	ArrayIndexT theIndex = mHueChanges->FetchIndexOf( &inHueChange );
	
	if ( theIndex == LArray::index_Bad )
		return false;
	
	mHueChanges->Remove( &inHueChange );
	return true;
}


// =================================================================================
//	 GetHueChangedColorTable										 [const, public]
// =================================================================================
// Copy the colors of the given color table into outColorTable after applying
// the active hue changes.

void
C256Tag::GetHueChangedColorTable( RGBColor *outColorTable ) const
{
	Assert_(outColorTable != nil);
	
	for ( Int32 i = 0; i < 256; i++ )
		outColorTable[i] = mHueChangedColorTable[i];
}


// =================================================================================
//	 GetHueChangedColorIndex										 [const, public]
// =================================================================================
// Sets the bits in outIndex according to whether the corresponding colors are
// affected by the active hue changes.

void
C256Tag::GetHueChangedColorIndex( Byte *outIndex ) const
{
	Assert_(outIndex != nil);
	
	LArrayIterator theIterator( *mHueChanges );
	CHueChange *theHueChange;
	
	// Clear the indices
	for ( Int32 i = 0; i < 32; i++ )
		outIndex[i] = 0;
	
	while ( theIterator.Next( &theHueChange ) ) {
		if ( theHueChange->GetUsedFlag() ) {
			Byte theBits[32];
			theHueChange->GetAffectedColorBits( theBits );
			
			for ( Int32 i = 0; i < 32; i++ )
				outIndex[i] |= theBits[i];
		}
	}
}


#pragma mark -

// =================================================================================
//	 GetSequence													 [const, public]
// =================================================================================
// Return a const pointer to the tag element with the given index

const CSequence *
C256Tag::GetSequence( Int32 inIndex ) const
{
	CSequence *theSequence;
	
	if ( mSequences->FetchItemAt( inIndex+1, &theSequence ) )
		return theSequence;
	
	return nil;
}


// =================================================================================
//	 GetSequence															[public]
// =================================================================================
// Return a pointer to the tag element with the given index

CSequence *
C256Tag::GetSequence( Int32 inIndex )
{
	CSequence *theSequence;
	
	if ( mSequences->FetchItemAt( inIndex+1, &theSequence ) )
		return theSequence;
	
	return nil;
}


// =================================================================================
//	 MoveSequence															[public]
// =================================================================================
// Move the tag element to a different index.  Nothing is changed if the new index
// is invalid.  Returns true if the element was successfully moved.

Boolean
C256Tag::MoveSequence( CSequence *inSequence, Int32 inNewIndex )
{
	// LArray is 1-based while the tag is 0-based
	inNewIndex += 1;
	
	ArrayIndexT theIndex = mSequences->FetchIndexOf( &inSequence );
	
	if ( theIndex == LArray::index_Bad
	||   !mSequences->ValidIndex( inNewIndex ) )
		return false;
	
	mSequences->MoveItem( theIndex, inNewIndex );
	return true;
}


// =================================================================================
//	 InsertSequence														[public]
// =================================================================================
// Insert a tag element into the array.  Returns true if the element was
// successfully inserted.  Array elements will be DELETED when the array is
// deleted.

Boolean
C256Tag::InsertSequence( CSequence *inSequence, Int32 inAtIndex )
{
	if ( inSequence == nil )
		return false;
	
	mSequences->InsertItemsAt( 1, inAtIndex+1, &inSequence );
	return true;
}


// =================================================================================
//	 RemoveSequence														[public]
// =================================================================================
// Remove the tag element from the array.  Returns true if the element was
// successfully removed.  Does NOT delete the element. 

Boolean
C256Tag::RemoveSequence( CSequence *inSequence )
{
	ArrayIndexT theIndex = mSequences->FetchIndexOf( &inSequence );
	
	if ( theIndex == LArray::index_Bad )
		return false;
	
	mSequences->Remove( &inSequence );
	return true;
}


#pragma mark -

// =================================================================================
//	 GetBlendTable													 [const, public]
// =================================================================================
// Return a const pointer to the tag element with the given index

const CBlendTable *
C256Tag::GetBlendTable( Int32 inIndex ) const
{
	CBlendTable *theBlendTable;
	
	if ( mBlendTables->FetchItemAt( inIndex+1, &theBlendTable ) )
		return theBlendTable;
	
	return nil;
}


// =================================================================================
//	 GetBlendTable															[public]
// =================================================================================
// Return a pointer to the tag element with the given index

CBlendTable *
C256Tag::GetBlendTable( Int32 inIndex )
{
	CBlendTable *theBlendTable;
	
	if ( mBlendTables->FetchItemAt( inIndex+1, &theBlendTable ) )
		return theBlendTable;
	
	return nil;
}


// =================================================================================
//	 MoveBlendTable														[public]
// =================================================================================
// Move the tag element to a different index.  Nothing is changed if the new index
// is invalid.  Returns true if the element was successfully moved.

Boolean
C256Tag::MoveBlendTable( CBlendTable *inBlendTable, Int32 inNewIndex )
{
	// LArray is 1-based while the tag is 0-based
	inNewIndex += 1;
	
	ArrayIndexT theIndex = mBlendTables->FetchIndexOf( &inBlendTable );
	
	if ( theIndex == LArray::index_Bad
	||   !mBlendTables->ValidIndex( inNewIndex ) )
		return false;
	
	mBlendTables->MoveItem( theIndex, inNewIndex );
	return true;
}


// =================================================================================
//	 InsertBlendTable														[public]
// =================================================================================
// Insert a tag element into the array.  Returns true if the element was
// successfully inserted.  Array elements will be DELETED when the array is
// deleted.

Boolean
C256Tag::InsertBlendTable( CBlendTable *inBlendTable, Int32 inAtIndex )
{
	if ( inBlendTable == nil )
		return false;
	
	mBlendTables->InsertItemsAt( 1, inAtIndex+1, &inBlendTable );
	return true;
}


// =================================================================================
//	 RemoveBlendTable														[public]
// =================================================================================
// Remove the tag element from the array.  Returns true if the element was
// successfully removed.  Does NOT delete the element. 

Boolean
C256Tag::RemoveBlendTable( CBlendTable *inBlendTable )
{
	ArrayIndexT theIndex = mBlendTables->FetchIndexOf( &inBlendTable );
	
	if ( theIndex == LArray::index_Bad )
		return false;
	
	mBlendTables->Remove( &inBlendTable );
	return true;
}


#pragma mark -

// =================================================================================
//	 GetRemapTable													 [const, public]
// =================================================================================
// Return a const pointer to the tag element with the given index

const CRemapTable *
C256Tag::GetRemapTable( Int32 inIndex ) const
{
	CRemapTable *theRemapTable;
	
	if ( mRemapTables->FetchItemAt( inIndex+1, &theRemapTable ) )
		return theRemapTable;
	
	return nil;
}


// =================================================================================
//	 GetRemapTable															[public]
// =================================================================================
// Return a pointer to the tag element with the given index

CRemapTable *
C256Tag::GetRemapTable( Int32 inIndex )
{
	CRemapTable *theRemapTable;
	
	if ( mRemapTables->FetchItemAt( inIndex+1, &theRemapTable ) )
		return theRemapTable;
	
	return nil;
}


// =================================================================================
//	 MoveRemapTable														[public]
// =================================================================================
// Move the tag element to a different index.  Nothing is changed if the new index
// is invalid.  Returns true if the element was successfully moved.

Boolean
C256Tag::MoveRemapTable( CRemapTable *inRemapTable, Int32 inNewIndex )
{
	// LArray is 1-based while the tag is 0-based
	inNewIndex += 1;
	
	ArrayIndexT theIndex = mRemapTables->FetchIndexOf( &inRemapTable );
	
	if ( theIndex == LArray::index_Bad
	||   !mRemapTables->ValidIndex( inNewIndex ) )
		return false;
	
	mRemapTables->MoveItem( theIndex, inNewIndex );
	return true;
}


// =================================================================================
//	 InsertRemapTable														[public]
// =================================================================================
// Insert a tag element into the array.  Returns true if the element was
// successfully inserted.  Array elements will be DELETED when the array is
// deleted.

Boolean
C256Tag::InsertRemapTable( CRemapTable *inRemapTable, Int32 inAtIndex )
{
	if ( inRemapTable == nil )
		return false;
	
	mRemapTables->InsertItemsAt( 1, inAtIndex+1, &inRemapTable );
	return true;
}


// =================================================================================
//	 RemoveRemapTable														[public]
// =================================================================================
// Remove the tag element from the array.  Returns true if the element was
// successfully removed.  Does NOT delete the element. 

Boolean
C256Tag::RemoveRemapTable( CRemapTable *inRemapTable )
{
	ArrayIndexT theIndex = mRemapTables->FetchIndexOf( &inRemapTable );
	
	if ( theIndex == LArray::index_Bad )
		return false;
	
	mRemapTables->Remove( &inRemapTable );
	return true;
}


#pragma mark -

// =================================================================================
//	 GetColorBitmap												 [const, public]
// =================================================================================
// Return a const pointer to the tag element with the given index

const CBitmap *
C256Tag::GetColorBitmap( Int32 inIndex ) const
{
	CBitmap *theBitmap;
	
	if ( mColorBitmaps->FetchItemAt( inIndex+1, &theBitmap ) )
		return theBitmap;
	
	return nil;
}


// =================================================================================
//	 GetColorBitmap														[public]
// =================================================================================
// Return a pointer to the tag element with the given index

CBitmap *
C256Tag::GetColorBitmap( Int32 inIndex )
{
	CBitmap *theBitmap;
	
	if ( mColorBitmaps->FetchItemAt( inIndex+1, &theBitmap ) )
		return theBitmap;
	
	return nil;
}


// =================================================================================
//	 GetNextColorBitmap											 [const, public]
// =================================================================================
// Return the next bitmap in the list (nil if inBitmap is the last one)

const CBitmap *
C256Tag::GetNextColorBitmap( const CBitmap *inBitmap ) const
{
	Int32 theIndex = mColorBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mColorBitmaps->FetchItemAt( theIndex+1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 GetNextColorBitmap													[public]
// =================================================================================
// Return the next bitmap in the list (nil if inBitmap is the last one)

CBitmap *
C256Tag::GetNextColorBitmap( const CBitmap *inBitmap )
{
	Int32 theIndex = mColorBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mColorBitmaps->FetchItemAt( theIndex+1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 GetPrevColorBitmap											 [const, public]
// =================================================================================
// Return the previous bitmap in the list (nil if inBitmap is the first one)

const CBitmap *
C256Tag::GetPrevColorBitmap( const CBitmap *inBitmap ) const
{
	Int32 theIndex = mColorBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mColorBitmaps->FetchItemAt( theIndex-1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 GetPrevColorBitmap													[public]
// =================================================================================
// Return the previous bitmap in the list (nil if inBitmap is the first one)

CBitmap *
C256Tag::GetPrevColorBitmap( const CBitmap *inBitmap )
{
	Int32 theIndex = mColorBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mColorBitmaps->FetchItemAt( theIndex-1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 MoveColorBitmap														[public]
// =================================================================================
// Move the tag element to a different index.  Nothing is changed if the new index
// is invalid.  Returns true if the element was successfully moved.

Boolean
C256Tag::MoveColorBitmap( CBitmap *inBitmap, Int32 inNewIndex )
{
	// LArray is 1-based while the tag is 0-based
	inNewIndex += 1;
	
	ArrayIndexT theIndex = mColorBitmaps->FetchIndexOf( &inBitmap );
	
	if ( theIndex == LArray::index_Bad
	||   !mColorBitmaps->ValidIndex( inNewIndex ) )
		return false;
	
	mColorBitmaps->MoveItem( theIndex, inNewIndex );
	return true;
}


// =================================================================================
//	 InsertColorBitmap														[public]
// =================================================================================
// Insert a tag element into the array.  Returns true if the element was
// successfully inserted.  Array elements will be DELETED when the array is
// deleted.

Boolean
C256Tag::InsertColorBitmap( CBitmap *inBitmap, Int32 inAtIndex )
{
	if ( inBitmap == nil )
		return false;
	
	mColorBitmaps->InsertItemsAt( 1, inAtIndex+1, &inBitmap );
	return true;
}


// =================================================================================
//	 RemoveColorBitmap														[public]
// =================================================================================
// Remove the tag element from the array.  Returns true if the element was
// successfully removed.  Does NOT delete the element. 

Boolean
C256Tag::RemoveColorBitmap( CBitmap *inBitmap )
{
	ArrayIndexT theIndex = mColorBitmaps->FetchIndexOf( &inBitmap );
	
	if ( theIndex == LArray::index_Bad )
		return false;
	
	mColorBitmaps->Remove( &inBitmap );
	return true;
}


#pragma mark -

// =================================================================================
//	 GetShadowBitmap												 [const, public]
// =================================================================================
// Return a const pointer to the tag element with the given index

const CBitmap *
C256Tag::GetShadowBitmap( Int32 inIndex ) const
{
	CBitmap *theBitmap;
	
	if ( mShadowBitmaps->FetchItemAt( inIndex+1, &theBitmap ) )
		return theBitmap;
	
	return nil;
}


// =================================================================================
//	 GetShadowBitmap														[public]
// =================================================================================
// Return a pointer to the tag element with the given index

CBitmap *
C256Tag::GetShadowBitmap( Int32 inIndex )
{
	CBitmap *theBitmap;
	
	if ( mShadowBitmaps->FetchItemAt( inIndex+1, &theBitmap ) )
		return theBitmap;
	
	return nil;
}


// =================================================================================
//	 GetNextShadowBitmap											 [const, public]
// =================================================================================
// Return the next bitmap in the list (nil if inBitmap is the last one)

const CBitmap *
C256Tag::GetNextShadowBitmap( const CBitmap *inBitmap ) const
{
	Int32 theIndex = mShadowBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mShadowBitmaps->FetchItemAt( theIndex+1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 GetNextShadowBitmap													[public]
// =================================================================================
// Return the next bitmap in the list (nil if inBitmap is the last one)

CBitmap *
C256Tag::GetNextShadowBitmap( const CBitmap *inBitmap )
{
	Int32 theIndex = mShadowBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mShadowBitmaps->FetchItemAt( theIndex+1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 GetPrevShadowBitmap											 [const, public]
// =================================================================================
// Return the previous bitmap in the list (nil if inBitmap is the first one)

const CBitmap *
C256Tag::GetPrevShadowBitmap( const CBitmap *inBitmap ) const
{
	Int32 theIndex = mShadowBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mShadowBitmaps->FetchItemAt( theIndex-1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 GetPrevShadowBitmap													[public]
// =================================================================================
// Return the previous bitmap in the list (nil if inBitmap is the first one)

CBitmap *
C256Tag::GetPrevShadowBitmap( const CBitmap *inBitmap )
{
	Int32 theIndex = mShadowBitmaps->FetchIndexOf( &inBitmap );
	CBitmap *theBitmap = nil;
	
	mShadowBitmaps->FetchItemAt( theIndex-1, &theBitmap );
	return theBitmap;
}


// =================================================================================
//	 MoveShadowBitmap														[public]
// =================================================================================
// Move the tag element to a different index.  Nothing is changed if the new index
// is invalid.  Returns true if the element was successfully moved.

Boolean
C256Tag::MoveShadowBitmap( CBitmap *inBitmap, Int32 inNewIndex )
{
	// LArray is 1-based while the tag is 0-based
	inNewIndex += 1;
	
	ArrayIndexT theIndex = mShadowBitmaps->FetchIndexOf( &inBitmap );
	
	if ( theIndex == LArray::index_Bad
	||   !mShadowBitmaps->ValidIndex( inNewIndex ) )
		return false;
	
	mShadowBitmaps->MoveItem( theIndex, inNewIndex );
	return true;
}


// =================================================================================
//	 InsertShadowBitmap													[public]
// =================================================================================
// Insert a tag element into the array.  Returns true if the element was
// successfully inserted.  Array elements will be DELETED when the array is
// deleted.

Boolean
C256Tag::InsertShadowBitmap( CBitmap *inBitmap, Int32 inAtIndex )
{
	if ( inBitmap == nil )
		return false;
	
	mShadowBitmaps->InsertItemsAt( 1, inAtIndex+1, &inBitmap );
	return true;
}


// =================================================================================
//	 RemoveShadowBitmap													[public]
// =================================================================================
// Remove the tag element from the array.  Returns true if the element was
// successfully removed.  Does NOT delete the element. 

Boolean
C256Tag::RemoveShadowBitmap( CBitmap *inBitmap )
{
	ArrayIndexT theIndex = mShadowBitmaps->FetchIndexOf( &inBitmap );
	
	if ( theIndex == LArray::index_Bad )
		return false;
	
	mShadowBitmaps->Remove( &inBitmap );
	return true;
}


#pragma mark -

// =================================================================================
//	 UpdateTagHeader													   [private]
// =================================================================================

void
C256Tag::UpdateTagHeader( void )
{
	Boolean isDirty = false;
	
	if ( mTagHeader->GetTagOffset() != mTagHeader->GetSize() ) {
		mTagHeader->SetTagOffset( mTagHeader->GetSize() );
		isDirty = true;
	}
	
	#pragma Find out what mFlags is!
	Int32 theTagLength = GetSize();
	Int32 theUserData = mCollectionHeader->GetUserDataFlags();
	
	if ( theTagLength != mTagHeader->GetTagSize()
	||   theUserData != mTagHeader->GetUserData() ) {
	//	mTagHeader->mFlags = ???;	
		mTagHeader->SetTagSize( theTagLength );
		mTagHeader->SetUserData( theUserData );
		isDirty = true;
	}
	
	if ( isDirty ) {
		SetDirty();
		BroadcastMessage( msg_UpdateTagHeader, nil );
	}
	
	return;
}

// =================================================================================
//	 UpdateHueChangedColorTable										   [private]
// =================================================================================
// Updates all hue changes and the precalculated changed color table, and then
// spawns an update hue changed color table event

void
C256Tag::UpdateHueChangedColorTable( void )
{
	RGBColor theColorTable[256];
	GetColorTable( theColorTable );
	GetColorTable( mHueChangedColorTable );
	
	LArrayIterator theIterator( *mHueChanges );
	CHueChange *theHueChange;
	
	while ( theIterator.Next( &theHueChange ) ) {
		if ( theHueChange->UpdateAffectedColorBits( theColorTable ) )
			theHueChange->AffectColorTable( mHueChangedColorTable );
	}
	
	BroadcastMessage( msg_256UpdateHueChangedColorTable, nil );
}

