// =================================================================================
//	CBitmapData.cp				 1998 Hans Hvidsten Birkeland. All rights reserved.
// =================================================================================

#include "CBitmapData.h"
#include "CopyBytes.h"

#include "256Messages.h"


// =================================================================================
//	Constants
// =================================================================================

// These two MUST correspond with the defs in UPlates.h
const Byte		kBackgroundIndex		= 2;
const Byte		kTranspIndex			= 0;


// =================================================================================
//	 CBitmapData															[public]
// =================================================================================

CBitmapData::CBitmapData( void )
{
	mWidth 			= 0;
	mHeight			= 0;
	mRowBytes		= 0;
	mFlags			= 0;
	mBitDepth		= 0;
	mWidthBits		= 0;
	mHeightBits		= 0;
	mByteDepth		= 0;
	
	mAlphas = NULL;
	mPixels = NULL;
}


// =================================================================================
//	 ~CBitmapData															[public]
// =================================================================================
					
CBitmapData::~CBitmapData( void )
{
	delete [] mAlphas;
	delete [] mPixels;
}


#pragma mark -

// =================================================================================
//	 ReadData																[public]
// =================================================================================
// inBitmapDataData is a buffer containing the bitmap data as they are stored in
// the .gor files and plugin files.

void
CBitmapData::ReadData( const Byte *inBitmapDataData )
{
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mWidth, 2 );
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mHeight, 2 );
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mRowBytes, 2 );
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mFlags, 2 );
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mBitDepth, 2 );
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mWidthBits, 2 );
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mHeightBits, 2 );
	inBitmapDataData += ::CopyBytes( inBitmapDataData, &mByteDepth, 2 );
	
	// Skip runtime stuff
	inBitmapDataData += 32;
	
	// Skip the row info table
	if ( mFlags & flag_NoRowAddressTable )
		inBitmapDataData += 4;
	
	else
		inBitmapDataData += 4*mHeight;
	
	if ( mRowBytes == -1 ) {
		switch ( mBitDepth ) {
			default:	Throw_(err_BadBitDepth);				break;	
			case 8:		Read8bitSprite( inBitmapDataData );		break;
			case 16:	Read16bitSprite( inBitmapDataData );	break;
		}
	} else
		ReadRaw( inBitmapDataData );
	
}
	

// =================================================================================
//	 ReadRaw															   [private]
// =================================================================================
// inBitmapDataData is a buffer containing the bitmap data in "raw" format

void
CBitmapData::ReadRaw( const Byte *inBitmapDataData )
{
	Int32 size = mWidth*mHeight;
	mPixels = new Byte[size];
	
	if ( mBitDepth == 8 ) {
		::CopyBytes( inBitmapDataData, mPixels, size );
	} else {
		mAlphas = new Byte[size];
		
		Byte *thePixels = mPixels;
		Byte *theAlphas = mAlphas;
		
		for ( int i = 0; i < size; i++ ) {
			*theAlphas++ = *inBitmapDataData++;
			*thePixels++ = *inBitmapDataData++;
		}
	}
}
	

// =================================================================================
//	 Read8bitSprite													   [private]
// =================================================================================
// inBitmapDataData is a buffer containing the bitmap data in 8 bit encoded format

void
CBitmapData::Read8bitSprite( const Byte *inBitmapDataData )
{
	mPixels = new Byte[mWidth*mHeight];
	Byte *thePixels = mPixels;
	
	for ( int i = 0; i < mWidth*mHeight; i++ )
		thePixels[i] = kBackgroundIndex;
	
	for ( int i = 0; i < mHeight; i++ ) {
		Int16 segments;
		Int16 totalPixels;
		
		inBitmapDataData += ::CopyBytes( inBitmapDataData, &segments, 2 );
		inBitmapDataData += ::CopyBytes( inBitmapDataData, &totalPixels, 2 );
		
		segmentEntry *segData = (segmentEntry *)inBitmapDataData;
		inBitmapDataData += segments*sizeof(segmentEntry);
		
		// Use offset to be able to be able to parse Remus' early tags
		Int32 theOffset = 0;
		
		for ( int j = 0; j < segments; j++ ) {
			
			for ( int k = segData[j].segmentStart; k < segData[j].segmentEnd; k++ )
				thePixels[k] = *(inBitmapDataData + theOffset++);
			
		}
		
		inBitmapDataData += totalPixels;
		thePixels += mWidth;
		
		if ( totalPixels & 0x01 )	// Skip filler?
			++inBitmapDataData;
	}
}

	
// =================================================================================
//	 Read16bitSprite													   [private]
// =================================================================================
// inBitmapDataData is a buffer containing the bitmap data in 16 bit encoded format

void
CBitmapData::Read16bitSprite( const Byte *inBitmapDataData )
{
	Int32 size = mWidth*mHeight;
	Byte *theAlphas = new Byte[size];
	Byte *thePixels = new Byte[size];
	
	mAlphas = theAlphas;
	mPixels = thePixels;
	
	for ( int i = 0; i < size; i++ ) {
		#pragma should alphas be initialized to 255 now?
		theAlphas[i] = 0;
		thePixels[i] = kBackgroundIndex;
	}
	
	for ( int i = 0; i < mHeight; i++ ) {
		Int16 segments;
		Int16 totalPixels;
		
		inBitmapDataData += ::CopyBytes( inBitmapDataData, &segments, 2 );
		inBitmapDataData += ::CopyBytes( inBitmapDataData, &totalPixels, 2 );
		
		segmentEntry *segData = (segmentEntry *)inBitmapDataData;
		inBitmapDataData += segments*sizeof(segmentEntry);
		
		// Use offset to be able to be able to parse Remus' early tags
		Int32 theOffset = 0;
		
		for ( int j = 0; j < segments; j++ ) {
			
			for ( int k = segData[j].segmentStart; k < segData[j].segmentEnd; k++ ) {
				theAlphas[k] = (inBitmapDataData[theOffset++] << 4);
				thePixels[k] = inBitmapDataData[theOffset++];
			}
			
		}
		
		(Int16 *)inBitmapDataData += totalPixels;
		theAlphas += mWidth;
		thePixels += mWidth;
		
	}
}

	
#pragma mark -

// =================================================================================
//	 WriteData														 [const, public]
// =================================================================================
// outBitmapDataData is filled with the bitmap data as they are stored on file
					
void
CBitmapData::WriteData( Byte *outBitmapDataData ) const
{
	outBitmapDataData += ::CopyBytes( &mWidth, outBitmapDataData, 2 );
	outBitmapDataData += ::CopyBytes( &mHeight, outBitmapDataData, 2 );
	outBitmapDataData += ::CopyBytes( &mRowBytes, outBitmapDataData, 2 );
	outBitmapDataData += ::CopyBytes( &mFlags, outBitmapDataData, 2 );
	outBitmapDataData += ::CopyBytes( &mBitDepth, outBitmapDataData, 2 );
	outBitmapDataData += ::CopyBytes( &mWidthBits, outBitmapDataData, 2 );
	outBitmapDataData += ::CopyBytes( &mHeightBits, outBitmapDataData, 2 );
	outBitmapDataData += ::CopyBytes( &mByteDepth, outBitmapDataData, 2 );
	
	// Zero out the runtime stuff
	for ( Int32 i = 0; i < 32; i++ )
		*(outBitmapDataData++) = 0;
	
	// Zeros for the row address table
	if ( mFlags & flag_NoRowAddressTable )
		for ( int i = 0; i < 4; i++ )
			*(outBitmapDataData++) = 0;
	
	else
		for ( int i = 0; i < 4*mHeight; i++ )
			*(outBitmapDataData++) = 0;
	
	if ( mRowBytes == -1 ) {
		switch ( mBitDepth ) {
			case 8:		Write8bitSprite( outBitmapDataData );	break;
			case 16:	Write16bitSprite( outBitmapDataData );	break;
		}
	} else
		WriteRaw( outBitmapDataData );

}


// =================================================================================
//	 WriteRaw														[const, private]
// =================================================================================
// out_rawData is filled with the bitmap data in raw format
					
void
CBitmapData::WriteRaw( Byte *out_rawData ) const
{
	Int32 size = mWidth*mHeight;
	
	if ( mBitDepth == 8 )
		::CopyBytes( mPixels, out_rawData, size );
	else {
		Byte *thePixels = mPixels;
		Byte *theAlphas = mAlphas;
		
		for ( int i = 0; i < size; i++ ) {
			*out_rawData++ = *theAlphas++;
			*out_rawData++ = *thePixels++;
		}
	}
}


// =================================================================================
//	 Write8bitSprite												[const, private]
// =================================================================================
// out_spriteData is filled with the bitmap data in sprite (compressed) format
					
void
CBitmapData::Write8bitSprite( Byte *out_spriteData ) const
{
	Byte *theRow = mPixels;
	
	for ( int i = 0; i < mHeight; i++ ) {
	
		Boolean wasTransparent = true;
		Int16 &segments = *(Int16 *)out_spriteData;
		Int16 &totalPixels = *(Int16 *)(out_spriteData + 2);
		
		segments = 0;
		totalPixels = 0;
		
		out_spriteData += 4;
		
		// Build the segment list
		
		for ( int j = 0; j < mWidth; j++ ) {
			
			if ( theRow[j] != kBackgroundIndex && wasTransparent ) {
				*(Int16 *)out_spriteData = j;
				out_spriteData += 2;
				wasTransparent = false;
				++segments;
			} else if ( theRow[j] == kBackgroundIndex && !wasTransparent ) {
				*(Int16 *)out_spriteData = j;
				out_spriteData += 2;
				wasTransparent = true;
			}
		}
		
		// Close the last segment if necessary
		
		if ( !wasTransparent ) {
			*(Int16 *)out_spriteData = mWidth;
			out_spriteData += 2;
		}
		
		// Copy the pixels
		
		for ( int j = 0; j < mWidth; j++ ) {
			
			if ( theRow[j] != kBackgroundIndex ) {
				*out_spriteData++ = theRow[j];
				++totalPixels;
			}
		}
		
		// Filler ?
		
		if ( totalPixels & 0x01 )
			*(out_spriteData++) = 0;
		
		// Update the row pointer
		
		theRow += mWidth;
	}
}


// =================================================================================
//	 Write16bitSprite												[const, private]
// =================================================================================
// out_spriteData is filled with the bitmap data in sprite (compressed) format
					
void
CBitmapData::Write16bitSprite( Byte *out_spriteData ) const
{
	Byte *theAlphaRow = mAlphas;
	Byte *theRow = mPixels;
	
	for ( int i = 0; i < mHeight; i++ ) {
	
		Boolean wasTransparent = true;
		Int16 &segments = *(Int16 *)out_spriteData;
		Int16 &totalPixels = *(Int16 *)(out_spriteData + 2);
		
		segments = 0;
		totalPixels = 0;
		
		out_spriteData += 4;
		
		// Build the segment list
		
		for ( int j = 0; j < mWidth; j++ ) {
			
			if ( theRow[j] != kBackgroundIndex && wasTransparent ) {
				*(Int16 *)out_spriteData = j;
				out_spriteData += 2;
				wasTransparent = false;
				++segments;
			} else if ( theRow[j] == kBackgroundIndex && !wasTransparent ) {
				*(Int16 *)out_spriteData = j;
				out_spriteData += 2;
				wasTransparent = true;
			}
		}
		
		// Close the last segment if necessary
		
		if ( !wasTransparent ) {
			*(Int16 *)out_spriteData = mWidth;
			out_spriteData += 2;
		}
		
		// Copy the pixels
		
		for ( int j = 0; j < mWidth; j++ ) {
			
			if ( theRow[j] != kBackgroundIndex ) {
				*out_spriteData++ = (theAlphaRow[j] >> 4) & 0x0F;
				*out_spriteData++ = theRow[j];
				++totalPixels;
			}
		}
		
		// Update the row pointers
		
		theAlphaRow += mWidth;
		theRow += mWidth;
	}
}


#pragma mark -

// =================================================================================
//	 GetSize														 [const, public]
// =================================================================================
// Return the bitmap data raw data length in bytes

Int32
CBitmapData::GetSize( void ) const
{
	Int16 pixelSize = mBitDepth/8;
	Int32 byteCount = 48;
	
	if ( mFlags & flag_NoRowAddressTable )
		byteCount += 4;
	else
		byteCount += 4*mHeight;
	
	if ( mRowBytes != -1  )
		byteCount += mWidth*mHeight*pixelSize;
	else {

		Byte *theRow = mPixels;
		
		for ( int i = 0; i < mHeight; i++ ) {
		
			Boolean wasTransparent = true;
			Int16 segments = 0;
			Int16 totalPixels = 0;
			
			byteCount += 4;
			
			// Count segments
			
			for ( int j = 0; j < mWidth; j++ ) {
				
				if ( theRow[j] != kBackgroundIndex && wasTransparent ) {
					wasTransparent = false;
					++segments;
				} else if ( theRow[j] == kBackgroundIndex && !wasTransparent ) {
					wasTransparent = true;
				}
			}
			
			byteCount += 4*segments;
			
			// Count the pixels
			
			for ( int j = 0; j < mWidth; j++ )
				if ( theRow[j] != kBackgroundIndex )
					totalPixels += pixelSize;
			
			byteCount += totalPixels;
			
			// Filler ?
			
			if ( (mFlags & flag_WordAlignedEncoding) && totalPixels & 0x01 )
				byteCount += 1;
			
			// Update the row pointer
			
			theRow += mWidth;
		}
	}
	
	return byteCount;
}


#pragma mark -

// =================================================================================
//	 SetPixels																[public]
// =================================================================================
// Copies from inPixels

void
CBitmapData::SetPixels(
	const Byte *inPixels,
	Int32 inRowBytes,
	Int16 inWidth,
	Int16 inHeight )
{
	#pragma also make SetShadowPixels
	
	StElementUpdater theUpdater( this );
	
	// Allocate new pixel buffer
	delete [] mPixels;
	mPixels = new Byte[inWidth*inHeight];
	
	// Copy the pixels, and count # of background pixels
	Byte *thePixels = mPixels;
	Int32 thePixelCount = 0;
	Int32 theBackgroundCount = 0;
	Int32 theTranspCount = 0;
	
	for ( Int32 i = 0; i < inHeight; i++ ) {
		for ( Int32 j = 0; j < inWidth; j++ ) {
			*thePixels++ = inPixels[j];
			
			theBackgroundCount += (inPixels[j] == kBackgroundIndex);
			theTranspCount += (inPixels[j] == kTranspIndex);
		}
		
		inPixels += inRowBytes;
	}
	
	// Update width and height
	mWidth = inWidth;
	mHeight = inHeight;
	
	// Set bitmap flags according to pixel setup
	if ( theTranspCount > 0 )
		mFlags |= flag_ColorKeyTransparency;
	else
		mFlags &= ~flag_ColorKeyTransparency;
	
	if ( theBackgroundCount > 0 )
		mFlags |= flag_1BitTranspEncoded;
	else
		mFlags &= ~flag_1BitTranspEncoded;
	
	// mRowBytes depends on the encoding
	if ( mFlags & flag_1BitTranspEncoded )
		mRowBytes = -1;
	else
		mRowBytes = mWidth;
	
	// Compute width and height bits for power of two bitmaps
	for ( mWidthBits = 0; (mWidth >> mWidthBits+1) != 0; mWidthBits++ )
		;
	
	for ( mHeightBits = 0; (mHeight >> mHeightBits+1) != 0; mHeightBits++ )
		;
	
	// The power of two flag
	if ( mWidth == mRowBytes
	&&   mWidth == (1 << mWidthBits)
	&&   mHeight == (1 << mHeightBits) )
		mFlags |= flag_PowerOfTwo;
	else
		mFlags &= ~flag_PowerOfTwo;
	
	// Inserting a bitmap makes it an 8 bit image
	if ( mBitDepth != 8 )
		RemoveAlphas();
	
	BroadcastMessage( msg_256UpdateBitmapData, nil );
}


// =================================================================================
//	 SetAlphas																[public]
// =================================================================================
// Insert a new alpha map (copies from inAlphas)

void
CBitmapData::SetAlphas( const Byte *inAlphas, Int32 inRowBytes )
{
	StElementUpdater theUpdater( this );
	
	if ( inAlphas == NULL )
		RemoveAlphas();
		
	else {
		// Inserting alpha values makes it a 16 bit image
		if ( mBitDepth != 16 ) {
			mBitDepth = 16;
			mFlags |= flag_4BitTranspEncoded;
			mByteDepth = 1;
			
			if ( mRowBytes != -1 )
				mRowBytes = mWidth*2;
		}
		
		delete [] mAlphas;
		mAlphas = new Byte[mWidth*mHeight];
		
		// Copy the alphas
		Byte *theAlphas = mAlphas;
		
		for ( Int32 i = 0; i < mHeight; i++ ) {
			::BlockMove( inAlphas, theAlphas, mWidth );
			inAlphas += inRowBytes;
			theAlphas += mWidth;
		}

	}
	
	BroadcastMessage( msg_256UpdateBitmapData, nil );
}


// =================================================================================
//	 RemoveAlphas															[public]
// =================================================================================
// Remove the alpha map

void
CBitmapData::RemoveAlphas( void )
{
	// Inserting a bitmap makes it an 8 bit image
	delete [] mAlphas;
	mAlphas = NULL;
	
	mBitDepth = 8;
	mFlags &= ~flag_4BitTranspEncoded;
	mByteDepth = 0;
	
	BroadcastMessage( msg_256UpdateBitmapData, nil );
}


#pragma mark -

// =================================================================================
//	 DrawBitmap															[public]
// =================================================================================

void
CBitmapData::DrawBitmap(
	Byte *outPixels,
	Int32 inRowBytes,
	Boolean inDrawAlphas )
{
	Byte *thePixels;
	
	if ( inDrawAlphas )
		thePixels = mAlphas;
	else
		thePixels = mPixels;
	
	for ( Uint32 i = 0; i < mHeight; i++ ) {
		::BlockMoveData( thePixels, outPixels, mWidth );
		thePixels += mWidth;
		outPixels += inRowBytes;
	}
}


// =================================================================================
//	 DrawBitmap															[public]
// =================================================================================

void
CBitmapData::DrawBitmap(
	Byte *outPixels,
	Int32 inRowBytes,
	Boolean inFlipHorizontal,
	Boolean inFlipVertical,
	Boolean inDrawAlphas )
{
	Byte *thePixels;
	Int32 theRowStep = 1;
	Int32 theColStep = 0;
	
	if ( inDrawAlphas )
		thePixels = mAlphas;
	else
		thePixels = mPixels;
	
	if ( inFlipHorizontal ) {
		theRowStep = -1;
		theColStep = 2*mWidth;
		thePixels += mWidth - 1;
	}
	
	if ( inFlipVertical ) {
		thePixels += (mHeight - 1)*mWidth;
		theColStep -= 2*mWidth;
	}
	
	for ( Int32 i = 0; i < mHeight; i++ ) {
		for ( Int32 j = 0; j < mWidth; j++ ) {
			if ( *thePixels != 0 )
				outPixels[j] = *thePixels;
				
			thePixels += theRowStep;
		}
		
		thePixels += theColStep;
		outPixels += inRowBytes;
	}
	
}


// =================================================================================
//	 DrawAlphas															[public]
// =================================================================================

void
CBitmapData::DrawAlphas(
	Byte *outPixels,
	Int32 inRowBytes )
{
	DrawBitmap( outPixels, inRowBytes, true );
}


// =================================================================================
//	 DrawAlphas															[public]
// =================================================================================

void
CBitmapData::DrawAlphas(
	Byte *outPixels,
	Int32 inRowBytes,
	Boolean inFlipHorizontal,
	Boolean inFlipVertical )
{
	DrawBitmap( outPixels, inRowBytes, inFlipHorizontal, inFlipVertical, true );
}


#pragma mark -

// =================================================================================
//	 GetWidth														 [const, public]
// =================================================================================

Int16
CBitmapData::GetWidth( void ) const
{
	return mWidth;
}


// =================================================================================
//	 GetHeight														 [const, public]
// =================================================================================

Int16
CBitmapData::GetHeight( void ) const
{
	return mHeight;
}


// =================================================================================
//	 GetBitDepth													 [const, public]
// =================================================================================

Int16
CBitmapData::GetBitDepth( void ) const
{
	return mBitDepth;
}


// =================================================================================
//	 GetFlags														 [const, public]
// =================================================================================

Uint16
CBitmapData::GetFlags( void ) const
{
	return mFlags;
}


