// =================================================================================
//	CBitmapData.h				 1998 Hans Hvidsten Birkeland. All rights reserved.
// =================================================================================

#pragma once

#include "CTagElement.h"


// =================================================================================
//	CBitmapData															 [48+ bytes]
// =================================================================================
// The actual bitmap data with a 48 byte header giving dimensions, bit depth.
// flags etc.
//
// Flags and meanings:
//	flag_PowerOfTwo				- Must be set for textures.  Means that
//								  mWidht == mRowBytes == 1 << mWidthBits, and
//								  mHeight == 1 << mHeightBits
//	flag_ColorKeyTransparency	- pixel index zero is transparent (unused it seems)
//	flag_1BitTranspEncoded		- cannot be set for textures (run length encoding)
//	flag_4BitLightMaskEncoded	- must be 16bpp	(Not used in any bungie .256)
//	flag_4BitTranspEncoded		- must be 16bpp ("normal" 16 bit bitmaps)
//	flag_NoRowAddressTable		- dont compute row address table (it doesnt exist)
//								  (Set for texture map bitmaps)
//	flag_WordAlignedEncoding	- encoded span headers are on word boundaries
//								  (always set when 4BitTranspEncoded?)
//	flag_GreyscaleBitmap		- 8bpp, [0,256) [black,white)
//								  (Set for shadow bitmaps, and for texture map
//								  texture bitmaps - not the shadow map bitmaps
//								  for some reason).
# pragma test pixel doubling at some time
//	flag_PixelDoubledBitmap		- will be pixel-doubled when drawn (not used?)
//	flag_PostprocessedBitmap	- implementation dependant (not used?)
//	flag_ProtectedBitmap		- used on disk; means "can't be recomputed"
//								  (Set for all bitmaps in most texture map tags,
//								  removing the flag doesn't seem to affect anything)
//	flag_BitmapIsOverlay		- uses intensity as transparency (correlated with
//								  CBitmapInstance::flag_Transparent ?)
//	flag_BitmapChanged			- must be reuploaded to hardware (runtime?)
//	flag_BitmapInCollection		- this bitmap is part of a loaded shape collection
//								  (runtime?)

class CBitmapData : public CTagElement {
friend 				class CBitmap;
friend				class CBitmapReference;

protected:
	enum			{	// Exceptions
						err_BadBitDepth				= 'BABI',
	
						// Flags
						flag_PowerOfTwo				= 0x0001,
						flag_ColorKeyTransparency	= 0x0002,	// Not used
						flag_1BitTranspEncoded		= 0x0004,
						flag_4BitLightMaskEncoded	= 0x0008,	// Not used
						flag_4BitTranspEncoded		= 0x0010,
						flag_NoRowAddressTable		= 0x0020,
						flag_WordAlignedEncoding	= 0x0040,
						flag_GreyscaleBitmap		= 0x0080,
						flag_PixelDoubledBitmap		= 0x0100,	// Not used
						flag_PostprocessedBitmap	= 0x0200,	// Not used
						flag_ProtectedBitmap		= 0x0400,	// Not really used
						flag_BitmapIsOverlay		= 0x0800,
						flag_BitmapChanged			= 0x1000,	// Runtime(?)
						flag_BitmapInCollection		= 0x2000	// Runtime(?)
					};
					
					// Default constructor and destructor
					
					CBitmapData();
					~CBitmapData();
					
					// inBitmapDataData is a buffer containing the bitmap
					// data as they are stored in the .gor files and plugin
					// files.
					
	void			ReadData( const Byte *inBitmapDataData );
	
					// outBitmapDataData is filled with the bitmap
					// data as they are stored on file
					
	void			WriteData( Byte *outBitmapDataData ) const;
	
					// Return the bitmap data size on disk
					
	Int32			GetSize() const;
	
					// Drawing the bitmap and alphas
	
	void			DrawBitmap(
						Byte *outPixels,
						Int32 inRowBytes,
						Boolean inDrawAlphas = false );
	void			DrawBitmap(
						Byte *outPixels,
						Int32 inRowBytes,
						Boolean inFlipHorizontal,
						Boolean inFlipVertical,
						Boolean inDrawAlphas = false );
	void			DrawAlphas(
						Byte *outPixels,
						Int32 inRowBytes );
	void			DrawAlphas(
						Byte *outPixels,
						Int32 inRowBytes,
						Boolean inFlipHorizontal,
						Boolean inFlipVertical );
					
					// Setting pixels and alphas
	
	void			SetPixels(
						const Byte *inPixels,
						Int32 inRowBytes,
						Int16 inWidth,
						Int16 inHeight );
	void			SetAlphas( const Byte *inAlphas, Int32 inRowBytes );
	void			RemoveAlphas();
	
					// Accessors
	
	Int16			GetWidth() const;
	Int16			GetHeight() const;
	Int16			GetBitDepth() const;
	
	Uint16			GetFlags() const;
	
private:
					// Defensive programming
					CBitmapData( const CBitmapData & );
	CBitmapData		&operator=( const CBitmapData & );
	
	void			ReadRaw( const Byte *inBitmapDataData );
	void			Read8bitSprite( const Byte *inBitmapDataData );
	void			Read16bitSprite( const Byte *inBitmapDataData );
					
	void			WriteRaw( Byte *out_rawData ) const;
	void			Write8bitSprite( Byte *out_rawData ) const;
	void			Write16bitSprite( Byte *out_rawData ) const;
	
	Int16 mWidth;				// width of the bitmap in pixels
	Int16 mHeight;				// height of the bitmap in pixels
	Int16 mRowBytes;			// FF FF for encoded bitmap
	Uint16 mFlags;
	Int16 mBitDepth;			// # of bits per pixel (0x08 or 0x10)
	Int16 mWidthBits;			// log(2) width
	Int16 mHeightBits;			// log(2) height
	Int16 mByteDepth;			// Physical byte depth - 1
//	Int16 mEncoding;			// Runtime
//	Int16 mHardwareTextureCacheIndex;
//	Uint32 mHardwareAddress;
//	void *mHardwareFormat;
//	Int16 mHardwareWidthBits;
//	Int16 mHardwareHeightBits;
//	Int16 mHardwareHeight;
//	Int16 mHardwareWidth;
//	Int16 x0;
//	Int16 y0;
//	Int16 mCollectionIndex;
//	Int16 mBitmapIndex;
//	Int16 mShadedHardwareAddresses[2];
//	void *rowAddresses;			// [height] see notes [1]
	
	Byte *mAlphas;
	Byte *mPixels;
	
	struct segmentEntry {
		Int16 segmentStart;
		Int16 segmentEnd;
	};
	
/*	struct rowEntry {
		Int16 segments;
		Int16 totalPixels;
		struct segmentEntry {
		
			Int16 segmentStart;
			Int16 segmentEnd;
		} *segData;				// [segments] see notes [2]

		Byte *pixel;			// [totalPixels]
//		Byte filler[??];		// see notes [3]
	} *rowData;					// [height]
*/
};

