// =================================================================================
//	CSequenceData.h				 1998 Hans Hvidsten Birkeland. All rights reserved.
// =================================================================================

#pragma once

#include "CTagElement.h"
#include "LListener.h"

class CBitmapInstance;
class CShadowMap;
class CBitmap;
class CFrameView;


// =================================================================================
//	CSequenceData														 [64+ bytes]
// =================================================================================
// The Sequence data hold more information about the sequence, and stores the
// shadow and bitmaps for each frame of animation.
// When flag_UsesMirroring is set, only the first mViewCount/2+1 instances in
// the view table are really used.  The rest are dummy instances (all zeros,
// except for max light intensity 0xFFFF).

// The "has no rotation" flag has this note:  Instead of being different rotations
// of the same shape each view here is a different permutation of a class
// (e.g., flying pieces of a body)


class CSequenceData : public CTagElement, public LListener {
friend				class CSequence;
friend				class CSequenceReference;

protected:
	enum			{	// Flags
						flag_UsesMirroring		= 0x01,
						flag_AlwaysSet			= 0x02,
						flag_HasNoRotation		= 0x04,
	
						// Transfer modes
						transferMode_Normal		= 0,
						transferMode_FadeOut	= 1,
						transferMode_FadeIn		= 2,
						transferMode_ShrinkOut	= 3,	// Not used?
						transferMode_ShrinkIn	= 4		// Not used?
					};
					
					// Default constructor and destructor
					CSequenceData();
					~CSequenceData();
					
					// inSequenceData is a buffer containing the
					// CSequenceData data as they are stored in the
					// .gor files and plugin files.
					
	void			ReadData( const Byte *inSequenceData );
	
					// outSequenceData is filled with the
					// CSequenceData data as they are stored on file
					
	void			WriteData( Byte *outSequenceData ) const;
	
					// GetSize returns the sequence data size on file,
					// GetTotalSize returns the size of the data and its
					// associated bitmap instances and shadow map entries.
					
	Int32			GetSize() const;
	Int32			GetTotalSize() const;
	
				// LListener
	
	virtual void	ListenToMessage( MessageT inMessage, void *ioData );
	
				// Accessors
	
	Uint32			GetFlags() const;
	void			SetFlags( Uint32 inFlags );

	Int16			GetViewCount() const;
	void			SetViewCount( Int16 inCount );

	Int16			GetFrameCount() const;
	void			SetFrameCount( Int16 inCount );

	Int16			GetFrameTicks() const;
	void			SetFrameTicks( Int16 inTicks );

	Int16			GetKeyFrame() const;
	void			SetKeyFrame( Int16 inFrame );

	Int16			GetLoopFrame() const;
	void			SetLoopFrame( Int16 inFrame );

	Int16			GetTransferPeriod() const;
	void			SetTransferPeriod( Int16 inPeriod );
	
	Uint16			GetTransferMode() const;
	void			SetTransferMode( Uint16 inMode );
	
	Int32			GetFirstSound() const;
	Int16			GetFirstSoundIndex() const;
	void			SetFirstSound( Int32 inSound = -1, Int16 inIndex = -1 );
	
	Int32			GetKeySound() const;
	Int16			GetKeySoundIndex() const;
	void			SetKeySound( Int32 inSound = -1, Int16 inIndex = -1 );
	
	Int32			GetLastSound() const;
	Int16			GetLastSoundIndex() const;
	void			SetLastSound( Int32 inSound = -1, Int16 inIndex = -1 );
	
	Int32			GetBBoxRadius() const;
	void			SetBBoxRadius( Int32 inRadius );
	
	Int32			GetBBoxBottom() const;
	void			SetBBoxBottom( Int32 inBottom );
	
	Int32			GetBBoxTop() const;
	void			SetBBoxTop( Int32 inTop );
	
					// Instance accessors
	
	const CBitmapInstance *GetInstance( Int16 inFrame, Int16 inView ) const;
	CBitmapInstance		*GetInstance( Int16 inFrame, Int16 inView );
	
	const CBitmap	*GetBitmap( Int16 inFrame, Int16 inView ) const;
	CBitmap			*GetBitmap( Int16 inFrame, Int16 inView );
	
					// Shadow map accessors

	const CShadowMap *GetShadowMap( Int16 inFrame ) const;
	CShadowMap		*GetShadowMap( Int16 inFrame );
	
	const CBitmap	*GetBitmap( Int16 inFrame ) const;
	CBitmap			*GetBitmap( Int16 inFrame );
	
					// Misc accessors
					
	Boolean			GetFrameView(
						const CBitmapInstance *inInstance,
						Int16 &outFrame,
						Int16 &outView ) const;

	Boolean			GetFrameView(
						const CShadowMap *inShadowMap,
						Int16 &outFrame ) const;

	Boolean			GetFrameView(
						const CBitmap *inBitmap,
						Int16 &outFrame,
						Int16 &outView ) const;
	
					// Initializing and declaring instances/shadows
					
	Boolean			AssignBitmapInstances(
						Byte *inBytes,
						Int32 inCount,
						LArray &inBitmaps );
	Boolean			AssignShadowMaps(
						Byte *inBytes,
						Int32 inCount,
						LArray &inBitmaps );
	
	void			DeclareElements(
						LArray &inInstances,
						LArray &inShadowMaps,
						LArray &inBitmaps );
					
private:
					// Defensive programming
					CSequenceData( const CSequenceData & );
	CSequenceData	&operator=( const CSequenceData & );
	
	Uint32	mFlags;
	#pragma Can Pixels to world be zeroed out?
	Int32	mPixelsToWorld;		// Scaling (0 -> 1:1, as does 1.0 .. comp. from mons?)
//	Int16	mViewCount;			// # of views
//	Int16	mFrameCount;		// # of frames per view
	Int16	mFrameTicks;		// # of ticks (1/30secs) per frame
	Int16	mKeyFrame;			// For attaching sounds(?)
	Int16	mLoopFrame;			// Loops from this frame(?)
	
	Int16	mFirstSoundIndex;	// Index into soun tag for first frame sound
	Int16	mKeySoundIndex;		// Index into soun tag for key frame sound
	Int16	mLastSoundIndex;	// Index into soun tag for last frame sound
	
	Int32	mFirstFrameSound;	// soun tag to be played at first frame
	Int32	mKeyFrameSound;		// soun tag to be played at key frame
	Int32	mLastFrameSound;	// soun tag to be played at last frame
	
	Uint16	mTransferMode;		// What happens when the sequence is played out
	Int16	mTransferPeriod;	// Timing for the mode (how long it takes to fade for inst.)
	
	Int32	mBBoxRadius;		// Radius of bounding rectangle (centered on ref point)	Int16	mUnknown6;
	Int32	mBBoxBottom;		// Top of bounding rectangle (up from ref point)
	Int32	mBBoxTop;			// Bottom of bounding rectangle (up from ref point)
	
	// These are mentioned in the bungie header, but seem unused
//	Int16	mWorldRadius;
//	Int16	mWorldBottom;
//	Int16	mWorldTop;
	
//	Byte	mUnused[6];
	
	CFrameView *mFrameView;
};


