// =================================================================================
//	CTagElement.cp				 1998 Hans Hvidsten Birkeland. All rights reserved.
// =================================================================================
// Implements basic behavior for a part of a Myth tag
// - Support for broadcasting a message when the element is deleted to allow objects
//   that rely on the element to take appropriate action.
// - Supports an "update" mode where outbound messages are stored
//   and sent as a batch when the element leaves update mode.
//
// This class also marks a definitive end to the illusion of keeping the tag
// structure classes platform independent.
//
// File created November 14, 1998

#include "CTagElement.h"
#include "AmberMessages.h"
#include <LArray.h>


// =================================================================================
//	 SMessageInfo															 [local]
// =================================================================================

struct SMessageInfo {
	MessageT	sMessage;
	void		*sParam;
};

typedef struct SMessageInfo SMessageInfo;


// =================================================================================
//	 CTagElement															[public]
// =================================================================================
// Default constructor

CTagElement::CTagElement( void )
{
	mUpdating = 0;
	mMessages = new LArray( sizeof(SMessageInfo) );
}


// =================================================================================
//	 ~CTagElement															[public]
// =================================================================================
// Buffered messages are not sent, because the subclass destructor is called first
// so the element isn't valid by the time we reach this destructor.  (This also
// means that objects receiving a delete message cannot refer to the element
// pointer in ioData other than to compare it to other pointers.)

CTagElement::~CTagElement( void )
{
	LBroadcaster::BroadcastMessage( msg_TagDeleteElement, this );
	delete mMessages;
}


#pragma mark -

// =================================================================================
//	 BroadcastMessage														[public]
// =================================================================================

void
CTagElement::BroadcastMessage( MessageT inMessage, void *ioParam )
{
	// Pass it along if we're not updating
	if ( mUpdating == 0 ) {
		LBroadcaster::BroadcastMessage( inMessage, ioParam );
		return;
	}
	
	// Store the message
	SMessageInfo theInfo = { inMessage, ioParam };
	
	if ( mMessages->FetchIndexOf( &inMessage ) == LArray::index_Bad )
		mMessages->InsertItemsAt( 1, LArray::index_Last, &theInfo );
}


// =================================================================================
//	 BeginUpdate															[public]
// =================================================================================

void
CTagElement::BeginUpdate( void )
{
	// Increase the nesting level
	++mUpdating;
}


// =================================================================================
//	 EndUpdate																[public]
// =================================================================================

void
CTagElement::EndUpdate( void )
{
	// Decrease the nesting level
	--mUpdating;
	
	// Sanity check
	ThrowIf_(mUpdating < 0);
	
	// Send all stored messages if this was the outermost nesting level
	if ( mUpdating == 0 ) {
		LArrayIterator theIterator( *mMessages );
		SMessageInfo theInfo;
		
		while ( theIterator.Next( &theInfo ) ) {
			LBroadcaster::BroadcastMessage( theInfo.sMessage, theInfo.sParam );
			mMessages->Remove( &theInfo );
		}
	}
}


#pragma mark -

// =================================================================================
//	 StElementUpdater														[public]
// =================================================================================
// Default constructor

StElementUpdater::StElementUpdater( CTagElement *inElement )
{
	if ( inElement != nil )
		inElement->BeginUpdate();
	
	mElement = inElement;
}


// =================================================================================
//	 ~StElementUpdater														[public]
// =================================================================================
// Default destructor

StElementUpdater::~StElementUpdater( void )
{
	if ( mElement != nil )
		mElement->EndUpdate();
}


