// =================================================================================
//	DetermineFileType.cp		 1998 Hans Hvidsten Birkeland. All rights reserved.
// =================================================================================
// Returns the tag type for Myth tag plugins, '.gor' for Myth .gor files, and
// '????' for any other file.
//
// May throw file errors

#include "DetermineFileType.h"

#include "CTagHeader.h"
#include "StNewBlock.h"
#include <LFileStream.h>


// =================================================================================
//	Constants
// =================================================================================

const OSType	OSType_UnkownFile		= '????';


// =================================================================================
//	All known tag types, and the .gor file type
// =================================================================================

OSType kKnownFileTypes[] = {
	'.gor',		// Myth .gor file
	'.256',		// Graphics and sprites
	'ARTI',		// Artifact descriptions
	'BINA',		// Interface
	'CORE',		// Color references
	'FONT',		// Myth's internal fonts
	'GEOM',		// Geometry objects
	'INTE',		// Only has one object: the standard interface
	'LIGH',		// Lightnings
	'LPGR',		// Light programs or local projectile groups, depending on who you ask
	'MEDI',		// Water descriptions
	'MEEF',		// Mesh effects; cause the mesh to get distorted when certain things happen
	'MELI',		// Affect how the shadows look on a mesh
 	'MESH',		// Describe heights, water, units, and scripts on a mesh
 	'MODE',		// 3D objects
 	'MONS',		// Monster (unit abilities)
 	'OBJE',		// General objects
 	'PART',		// Weather conditions
 	'PHYS',		// Physics
 	'PRGR',		// Projectile groups
 	'PROJ',		// Projectiles
 	'REVA',		// Reflection values for water
 	'RECO',		// Recording of a game
 	'SAVE',		// Saved game
 	'SCEN',		// Scenery objects
  	'SOLI',		// Internal sounds
  	'SOUN',		// Sounds in IMA 4:1 format
 	'STLI',		// String lists
 	'TEMP',		// Types of map actions
 	'TEXT',		// Texts; used for the story
	'UNIT',		// Unit descriptions
	'WIND'		// Winds
};

OSType kKnownTagTypes[] = {
	'.gor',		// Myth .gor file
	'.256',		// Graphics and sprites
	'arti',		// Artifact descriptions
	'bina',		// Interface
	'core',		// Color references
	'font',		// Myth's internal fonts
	'geom',		// Geometry objects
	'inte',		// Only has one object: the standard interface
	'ligh',		// Lightnings
	'lpgr',		// Light programs or local projectile groups, depending on who you ask
	'medi',		// Water descriptions
	'meef',		// Mesh effects; cause the mesh to get distorted when certain things happen
	'meli',		// Affect how the shadows look on a mesh
 	'mesh',		// Describe heights, water, units, and scripts on a mesh
 	'mode',		// 3D objects
 	'mons',		// Monster (unit abilities)
 	'obje',		// General objects
 	'part',		// Weather conditions
 	'phys',		// Physics
 	'prgr',		// Projectile groups
 	'proj',		// Projectiles
 	'reva',		// Reflection values for water
 	'reco',		// Recording of a game
 	'save',		// Saved game
 	'scen',		// Scenery objects
  	'soli',		// Internal sounds
  	'soun',		// Sounds in IMA 4:1 format
 	'stli',		// String lists
 	'temp',		// Types of map actions
 	'text',		// Texts; used for the story
	'unit',		// Unit descriptions
	'wind'		// Winds
};

// =================================================================================
//	 DetermineFileType
// =================================================================================

OSType DetermineFileType( FSSpec *inFSSpec )
{
	FInfo theFInfo;
	OSErr theErr;
	
	theErr = ::FSpGetFInfo( inFSSpec, &theFInfo );
	ThrowIfOSErr_(theErr);
	
	// First check the file type
	for ( int i = 0; i < sizeof(kKnownFileTypes)/sizeof(OSType); i++ )
		if ( theFInfo.fdType == kKnownFileTypes[i]
		||	 theFInfo.fdType == kKnownTagTypes[i] )
			// Can't simply return 'TEXT' files, they could be anything :)
			if ( theFInfo.fdType != 'TEXT' )
				return kKnownTagTypes[i];
	
	// It wasn't that easy.. now look inside the file
	LFileStream theFile( *inFSSpec );
	CTagHeader theHeader;
	
	theFile.OpenDataFork( fsRdPerm );
	
	if ( theFile.GetLength() < theHeader.GetSize() )
		return OSType_UnkownFile;
	
	Try_ {
		StNewBlock theHeaderData( theHeader.GetSize() );
		theFile.ReadBlock( theHeaderData, theHeader.GetSize() );
		
		LDataStream theStream( theHeaderData, theHeader.GetSize() );
		
		theHeader.ReadStream( theStream );
		
		if ( theHeader.IsValidTag() )
			return theHeader.GetTagGroup();
		
	} Catch_( inErr ) {
		// Do nothing
	}
	
	// It's not a tag plugin, check if it's a .gor/monolithic tag
	#pragma not implemented
	
	return OSType_UnkownFile;
}
