// =================================================================================
//	C3dmfDocument.cp				1997 eli curtz. All rights reserved.
// =================================================================================

#include <LFile.h>
#include <LPlaceHolder.h>
#include <LPrintout.h>
#include <LString.h>
#include <LWindow.h>
#include <PP_Messages.h>
#include <UMemoryMgr.h>
#include <UWindows.h>

#include <QD3DGroup.h>
#include <QD3DStorage.h>
#include <QD3DShader.h>
#include <QD3DView.h>
#include <QD3DLight.h>
#include <QD3DIO.h>
#include <QD3DMath.h>

#include "Constants.h"

#include "bitmaps.h"
#include "shape_file.h"
#include "model_definitions.h"
#include "collection_references.h"

#include "C3dmfDocument.h"
#include "UQuickdraw3D.h"
#include "CgeomView.h"

// Debugging
// #include "CGWorldView.h"

#define _FLIP_X_AXIS true

#ifdef _FLIP_X_AXIS
const TQ3Matrix4x4 QD3DToMythMatrix = {{{-1.0,  0.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  1.0,  0.0 }, \
										{ 0.0,  1.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  0.0,  1.0 } }};
const TQ3Matrix4x4 MythToQD3DMatrix = {{{-1.0,  0.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  1.0,  0.0 }, \
										{ 0.0,  1.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  0.0,  1.0 } }};
#else
const TQ3Matrix4x4 QD3DToMythMatrix = {{{ 1.0,  0.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  1.0,  0.0 }, \
										{ 0.0,  1.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  0.0,  1.0 } }};
const TQ3Matrix4x4 MythToQD3DMatrix = {{{ 1.0,  0.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  1.0,  0.0 }, \
										{ 0.0,  1.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  0.0,  1.0 } }};
#endif

const TQ3Point3D zeroPoint = { 0.0, 0.0, 0.0 };

const Str255 numberFormat = "\p###,###,##0.0#;-###,###,###.0#;0.0";

const PaneIDT	scaleXID 	= 'SizX';
const PaneIDT	scaleYID 	= 'SizY';
const PaneIDT	scaleZID 	= 'SizZ';
const PaneIDT	alignXID 	= 'MidX';
const PaneIDT	alignYID 	= 'MidY';
const PaneIDT	alignZID 	= 'MidZ';
const PaneIDT	cellID 		= 'Celz';
const PaneIDT	tagID 		= 'TagF';
const PaneIDT	nameID 		= 'NamF';

const PaneIDT	exportID 		= 'ExpB';

const MessageT message_DoExport = 'Xprt';


// ---------------------------------------------------------------------------------
//		 CNotifyField
// ---------------------------------------------------------------------------------


CNotifyField::CNotifyField() : LEditField(), LBroadcaster()
{
	m_bChanged = false;
};

CNotifyField::CNotifyField( LStream *inStream ) :  LEditField(inStream), LBroadcaster()
{
	m_bChanged = false;
};

void CNotifyField::UserChangedText()
{
	 m_bChanged = true;
};

void CNotifyField::DontBeTarget()
{
	LEditField::DontBeTarget();
	
	if ( m_bChanged )
	{
		this->BroadcastMessage(message_TextChanged, &mPaneID);
		m_bChanged = false;
	}
};


// ---------------------------------------------------------------------------------
//		 CgeomDocument(LCommander*, FSSpec*)
// ---------------------------------------------------------------------------------


C3dmfDocument::C3dmfDocument(
	LCommander	*inSuper,
	FSSpec		*inFileSpec )
		: LSingleDoc( inSuper ), LListener()
{
	mGeometry.flags 					= 0L;
	mGeometry.collection_reference_tag 	= 'test';
	mGeometry.material_count 			= 0;
	mGeometry.vertex_count 				= 0;
	mGeometry.surface_count 			= 0;
	mGeometry.dependency_count 			= 0;
	
	mModel.flags						= 0L;
	mModel.geometry_tag 				= 'test';
	mModel.geometry_index				= 0;
	mModel.model_geometry_vertex_count	= 0;
	mModel.model_permutation_count		= 0;
	mModel.model_cell_count				= 0;
	
	::Q3BoundingBox_Set(&mBoundingBox, &zeroPoint, &zeroPoint, kQ3True);
	
	// Create window for our document.
	mWindow = LWindow::CreateWindow( 1001, this );
	Assert_( mWindow != nil );
	
	// Find the active panes
	mGeomView = (CgeomView *) mWindow->FindPaneByID( kMainView );
	Assert_( mGeomView != nil );
	mScaleXField = (CNotifyField *) mWindow->FindPaneByID( scaleXID );
	Assert_( mScaleXField != nil );
	mScaleYField = (CNotifyField *) mWindow->FindPaneByID( scaleYID );
	Assert_( mScaleYField != nil );
	mScaleZField = (CNotifyField *) mWindow->FindPaneByID( scaleZID );
	Assert_( mScaleZField != nil );
	mAlignXField = (CNotifyField *) mWindow->FindPaneByID( alignXID );
	Assert_( mScaleXField != nil );
	mAlignYField = (CNotifyField *) mWindow->FindPaneByID( alignYID );
	Assert_( mScaleYField != nil );
	mAlignZField = (CNotifyField *) mWindow->FindPaneByID( alignZID );
	Assert_( mScaleZField != nil );
	mTagField = (CNotifyField *) mWindow->FindPaneByID( tagID );
	Assert_( mTagField != nil );
	mTagField->SetMaxChars(4);
	mNameField = (CNotifyField *) mWindow->FindPaneByID( nameID );
	Assert_( mNameField != nil );
	
	mScaleXField->AddListener(this);
	mScaleYField->AddListener(this);
	mScaleZField->AddListener(this);
	mAlignXField->AddListener(this);
	mAlignYField->AddListener(this);
	mAlignZField->AddListener(this);
	
	mMythScale = 1.0;
	mMeshOffsetX = 0;
	mMeshOffsetY = 0;
	mMeshOffsetZ = 0;
	mAlignmentX = 0.0;
	mAlignmentY = 0.0;
	mAlignmentZ = 0.0;
	
	mTransparency = 0;
	mDependencySurface = -1;
	
	mDependencyArrays = new LArray( sizeof( LArray *) );
	
	LButton* theButton = (LButton*) mWindow->FindPaneByID( exportID ); 
	Assert_( theButton != nil );
	theButton->AddListener(this);
	
	Try_ {
		if ( inFileSpec == nil ) {

			// Set the name of new window.
			NameNewDoc();
			
		} else {
			ThrowIfOSErr_(::NewPictInfo(&mPictInfoID, returnPalette + suppressBlackAndWhite, 254, medianMethod, 0));
			
			// Display contents of file in window.
			OpenFile( *inFileSpec );
		}
		
		// Make the window visible.
		mWindow->Show();
		
	} Catch_( inErr ) {
		Throw_Err( inErr );
	} EndCatch_
}

C3dmfDocument::~C3dmfDocument()
{
}


// ---------------------------------------------------------------------------------
//		 NameNewDoc
// ---------------------------------------------------------------------------------

void
C3dmfDocument::NameNewDoc()
{
	// Setup the window title. Start with the default title.
	LStr255	theTitle( "\pUntitled" );

	// Find the first available title. We could also check the window
	// pane id if we wanted to make sure we didn't collide with other
	// window types.
	Int32	theNumber = 0;
	while ( UWindows::FindNamedWindow( theTitle ) != nil ) {

		// An existing window has the current name
		// Increment counter and try again.
		++theNumber;
		theTitle = "\pUntitled ";
		theTitle += (LStr255) theNumber;

	}		
	
	// Finally, set window title.
	mWindow->SetDescriptor( theTitle );
}


// ---------------------------------------------------------------------------------
//		 OpenFile
// ---------------------------------------------------------------------------------

void
C3dmfDocument::OpenFile(
	FSSpec	&inFileSpec )
{
	Try_ {
		// Initialize the normal table
		Handle theNormalResourceH = ::GetResource('Nrml', 128);
		ThrowIfNULL_(theNormalResourceH);
		{
			StHandleLocker lock(theNormalResourceH);
			::BlockMoveData(*theNormalResourceH, mVertexNormals, NUMBER_OF_PRECALCULATED_MODEL_NORMALS * sizeof(model_normal));
		}
		::ReleaseResource(theNormalResourceH);

		// Create a new file object.
		mFileStream = new LFileStream( inFileSpec );
		mFile = mFileStream;	// In case the superclass is using it
		
		// Read the file to check the bounds
		{
			TQ3StorageObject theStorage = ::Q3FSSpecStorage_New(&inFileSpec);
			ThrowIfNULL_(theStorage);
			TQ3FileObject theQD3DFile = ::Q3File_New();
			ThrowIfNULL_(theQD3DFile);
			
			TQ3Status theStatus = ::Q3File_SetStorage(theQD3DFile, theStorage);
			::Q3Object_Dispose(theStorage);
			ThrowIf_(theStatus != kQ3Success);
			
			theStatus = ::Q3File_OpenRead(theQD3DFile, NULL);
			ThrowIf_(theStatus != kQ3Success);
			
			theStatus = ::Q3File_SetReadInGroup(theQD3DFile, kQ3FileReadObjectsInGroup);
			ThrowIf_(theStatus != kQ3Success);

			while(::Q3File_IsEndOfFile(theQD3DFile) == kQ3False)
			{
				TQ3Object theObject = NULL;
				theObject = ::Q3File_ReadObject(theQD3DFile);
				
				if(::Q3Object_IsType(theObject, kQ3GeometryTypeTriMesh))
				{
					TQ3TriMeshData theData;
					theStatus = ::Q3TriMesh_GetData(theObject, &theData);
					ThrowIf_(theStatus != kQ3Success);
					
					::Q3BoundingBox_UnionPoint3D(&mBoundingBox, &theData.bBox.min, &mBoundingBox);
					::Q3BoundingBox_UnionPoint3D(&mBoundingBox, &theData.bBox.max, &mBoundingBox);
				}

				if (theObject != NULL)
				{
					::Q3Object_Dispose(theObject);
				}
			}
			theStatus = Q3File_Close(theQD3DFile);
			ThrowIf_(theStatus != kQ3Success);
		}
		
		// Create the QD3D group.
		TQ3GroupObject qd3DNewDisplayGroup = ::Q3OrderedDisplayGroup_New();
		ThrowIfNULL_(qd3DNewDisplayGroup);

		mGeomView->SetGroup(qd3DNewDisplayGroup);
		mGeomView->SetDocument(this);
		
		// Set the bounding box
		mGeomView->SetBoundingBox(&mBoundingBox);

		// Set the transform
		::Q3Matrix4x4_Copy(&QD3DToMythMatrix, &mTransformMatrix);
		
		// Scale the points to Myth coordinates
		TQ3Matrix4x4 theScalingMatrix;
		::Q3Matrix4x4_SetScale(&theScalingMatrix,
			mMythScale * (float) GEOMETRY_NORMAL_SCALE,
			mMythScale * (float) GEOMETRY_NORMAL_SCALE,
			mMythScale * (float) GEOMETRY_NORMAL_SCALE);
			
		::Q3Matrix4x4_Multiply(&mTransformMatrix, &theScalingMatrix, &mTransformMatrix);	
		
		// Build the Myth data
		{
			FSSpec theSourceFileSpec;
			mFile->GetSpecifier(theSourceFileSpec);
			TQ3StorageObject theStorage = ::Q3FSSpecStorage_New(&theSourceFileSpec);
			ThrowIfNULL_(theStorage);
			TQ3FileObject theQD3DFile = ::Q3File_New();
			ThrowIfNULL_(theQD3DFile);
			
			TQ3Status theStatus = ::Q3File_SetStorage(theQD3DFile, theStorage);
			::Q3Object_Dispose(theStorage);
			ThrowIf_(theStatus != kQ3Success);
			
			theStatus = ::Q3File_OpenRead(theQD3DFile, NULL);
			ThrowIf_(theStatus != kQ3Success);
			
			theStatus = ::Q3File_SetReadInGroup(theQD3DFile, kQ3FileReadObjectsInGroup);
			ThrowIf_(theStatus != kQ3Success);

			while(::Q3File_IsEndOfFile(theQD3DFile) == kQ3False)
			{
				TQ3Object theObject = NULL;
				theObject = ::Q3File_ReadObject(theQD3DFile);
				
				if(::Q3Object_IsType(theObject, kQ3GeometryTypeTriMesh))
				{
					this->AddTriMeshObject(theObject);
				}

				if (theObject != NULL)
				{
					::Q3Object_Dispose(theObject);
				}
			}
			theStatus = Q3File_Close(theQD3DFile);
			ThrowIf_(theStatus != kQ3Success);
		}
		
		this->AdjustMythVertices();
		
		// Adjust the camera
		mGeomView->ResetHomeCamera();

		// Set the fields
		this->AdjustEditFields();

		mTagField->SetDescriptor("\ptest");
		mNameField->SetDescriptor(LStr255(inFileSpec.name));
			
	} Catch_( inErr ) {
		// Tell the user what happened
		::ParamText( LStr255( rSTRl_Alerts, eAlertBadFile ), nil, nil, nil );
		UModalAlerts::StopAlert( ALRT_Generic );

		// Cleanup and rethrow the error.
		Throw_Err( inErr );
	
	} EndCatch_
}

// ---------------------------------------------------------------------------------
//		 AddTriMeshObject
// ---------------------------------------------------------------------------------

void
C3dmfDocument::AddTriMeshObject(TQ3Object& inTriMesh)
{
	Assert_(Q3Object_IsType(inTriMesh, kQ3GeometryTypeTriMesh) == kQ3True);
	
	// Add the new material for the trimesh
#ifdef DEBUG
	Assert_(mGeometry.material_count <= MAXIMUM_MATERIALS_PER_GEOMETRY);
#else
	if (mGeometry.material_count > MAXIMUM_MATERIALS_PER_GEOMETRY)
	{
		::ParamText( LStr255( rSTRl_Alerts, eAlertExtraTextures ), nil, nil, nil );
		UModalAlerts::StopAlert( ALRT_Generic );
		return;
	}
#endif
	
	TQ3TriMeshData theData;
	TQ3Status theStatus = ::Q3TriMesh_GetData(inTriMesh, &theData);
	ThrowIf_(theStatus != kQ3Success);
	
	// Locate the texture for the surface
	TQ3ShaderObject theShader;
	if (::Q3AttributeSet_Contains(theData.triMeshAttributeSet, kQ3AttributeTypeSurfaceShader) == kQ3True)
	{
		theStatus = ::Q3AttributeSet_Get(theData.triMeshAttributeSet, kQ3AttributeTypeSurfaceShader, &theShader);
		ThrowIf_(theStatus != kQ3Success);
		
		if (::Q3Object_IsType(theShader, kQ3SurfaceShaderTypeTexture))
		{
			TQ3TextureObject theTexture;
			theStatus = ::Q3TextureShader_GetTexture(theShader, &theTexture);
			ThrowIf_(theStatus != kQ3Success);
			
			TQ3ObjectType textureType = ::Q3Texture_GetType(theTexture);
			if (textureType == kQ3TextureTypePixmap)
			{
				TQ3StoragePixmap theTexturePixmap;
				theStatus = ::Q3PixmapTexture_GetPixmap(theTexture, &theTexturePixmap);
				ThrowIf_(theStatus != kQ3Success);

				Rect textureBounds = { 0, 0, 0, 0 };
				Assert_(theTexturePixmap.height <= 256);
				Assert_(theTexturePixmap.width <= 256);
				textureBounds.bottom = theTexturePixmap.height;
				textureBounds.right = theTexturePixmap.width;
				
				LGWorld* theGWorld = new LGWorld(textureBounds, theTexturePixmap.pixelSize);
				PixMapHandle thePixMapH = ::GetGWorldPixMap(theGWorld->GetMacGWorld());

				UInt32 texturemapOffset = 0;
				Ptr   pixmapAddr = ::GetPixBaseAddr(thePixMapH);
				UInt32 pixmapRowBytes = (*thePixMapH)->rowBytes & 0x3FFF;
				
				::LockPixels(thePixMapH);
				for (Uint16 rowCount = 0; rowCount < theTexturePixmap.height; rowCount++)
				{
					UInt32 copySize = 0;
					theStatus = ::Q3Storage_GetData(theTexturePixmap.image, texturemapOffset, theTexturePixmap.rowBytes, (UInt8*) pixmapAddr, &copySize);
					ThrowIf_(theStatus != kQ3Success);
					
					texturemapOffset += theTexturePixmap.rowBytes;
					pixmapAddr += pixmapRowBytes;
				}
				
				// Gather the information for making CLUT
				ThrowIfOSErr_(::RecordPixMapInfo(mPictInfoID, thePixMapH));
				
				::UnlockPixels(thePixMapH);
				mGWorlds[mGeometry.material_count] = theGWorld;
			}
			else if (textureType == kQ3TextureTypeMipmap)
			{
				TQ3Mipmap theMipmap;
				theStatus = ::Q3MipmapTexture_GetMipmap(theTexture, &theMipmap);
				ThrowIf_(theStatus != kQ3Success);
				
				Rect textureBounds = { 0, 0, 0, 0 };
				Assert_(theMipmap.mipmaps[0].height <= 256);
				Assert_(theMipmap.mipmaps[0].width <= 256);
				textureBounds.bottom = theMipmap.mipmaps[0].height;
				textureBounds.right = theMipmap.mipmaps[0].width;
				SInt16 pixelDepth;
				switch(theMipmap.pixelType)
				{
					case kQ3PixelTypeRGB32:
					case kQ3PixelTypeARGB32:
						pixelDepth = 32;
						break;
					
					case kQ3PixelTypeRGB16:
					case kQ3PixelTypeARGB16:
						pixelDepth = 16;
						break;
					
					default:
						SignalPStr_("\pBad pixel depth in mipmap.");
				}
				
				LGWorld* theGWorld = new LGWorld(textureBounds, pixelDepth);
				PixMapHandle thePixMapH = ::GetGWorldPixMap(theGWorld->GetMacGWorld());

				UInt32 mipmapOffset = theMipmap.mipmaps[0].offset;
				UInt32 mipmapRowBytes = theMipmap.mipmaps[0].rowBytes;
				Ptr   pixmapAddr = ::GetPixBaseAddr(thePixMapH);
				UInt32 pixmapRowBytes = (*thePixMapH)->rowBytes & 0x3FFF;
				
				::LockPixels(thePixMapH);
				for (Uint16 rowCount = 0; rowCount < theMipmap.mipmaps[0].height; rowCount++)
				{
					UInt32 copySize = 0;
					theStatus = ::Q3Storage_GetData(theMipmap.image, mipmapOffset, mipmapRowBytes, (UInt8*) pixmapAddr, &copySize);
					ThrowIf_(theStatus != kQ3Success);
					
					mipmapOffset += mipmapRowBytes;
					pixmapAddr += pixmapRowBytes;
				}
				
				// Gather the information for making CLUT
				ThrowIfOSErr_(::RecordPixMapInfo(mPictInfoID, thePixMapH));
				
				::UnlockPixels(thePixMapH);
				mGWorlds[mGeometry.material_count] = theGWorld;
			}
			
			// Create a new shader for rendering
			mShaders[mGeometry.material_count] = ::Q3TextureShader_New(theTexture);
			::Q3Object_Dispose(theTexture);
		}

		::Q3Object_Dispose(theShader);
	}
	else
	{
		SignalPStr_("\pNo texture found on geometry - that's bad.");
	}
	
	// Add the triangles
	for (UInt16 surfaceCount = 0; surfaceCount < theData.numTriangles; surfaceCount++)
	{
		mSurfaces[mGeometry.surface_count].flags = 0 /* _geometry_surface_no_backface_cull_flag */;

		mSurfaces[mGeometry.surface_count].normal.i = 0;
		mSurfaces[mGeometry.surface_count].normal.j = 0;
		mSurfaces[mGeometry.surface_count].normal.k = 0;

		mSurfaces[mGeometry.surface_count].d = 0;

		mSurfaces[mGeometry.surface_count].transparency = 0;

		mSurfaces[mGeometry.surface_count].material_index = mGeometry.material_count;
	
#ifdef _FLIP_X_AXIS
		// change vertex 1 and 3 to maintain proper handedness
		SetCorner(theData, theData.triangles[surfaceCount].pointIndices[0], mSurfaces[mGeometry.surface_count].corners[2]);
		SetCorner(theData, theData.triangles[surfaceCount].pointIndices[1], mSurfaces[mGeometry.surface_count].corners[1]);
		SetCorner(theData, theData.triangles[surfaceCount].pointIndices[2], mSurfaces[mGeometry.surface_count].corners[0]);
#else
		SetCorner(theData, theData.triangles[surfaceCount].pointIndices[0], mSurfaces[mGeometry.surface_count].corners[0]);
		SetCorner(theData, theData.triangles[surfaceCount].pointIndices[1], mSurfaces[mGeometry.surface_count].corners[1]);
		SetCorner(theData, theData.triangles[surfaceCount].pointIndices[2], mSurfaces[mGeometry.surface_count].corners[2]);
#endif
		
		geometry_surface& currentSurface = mSurfaces[mGeometry.surface_count];
		short vertex0 = currentSurface.corners[0].vertex_index;
		short vertex1 = currentSurface.corners[1].vertex_index;
		short vertex2 = currentSurface.corners[2].vertex_index;

		// Make an attempt to recover from slightly misaligned textures
		// 512 == 2 pixels in a 256x256 texture or 1 pixel in a 128x128
		for (UInt16 interpolantCount = 0; interpolantCount < NUMBER_OF_GEOMETRY_CORNER_INTERPOLANTS; interpolantCount++)
		{
			if (abs((float) currentSurface.corners[0].interpolants[interpolantCount] - currentSurface.corners[1].interpolants[interpolantCount]) <= 512.0)
				{
				 	if (currentSurface.corners[0].interpolants[interpolantCount] > currentSurface.corners[1].interpolants[interpolantCount])
				 		currentSurface.corners[1].interpolants[interpolantCount] = currentSurface.corners[0].interpolants[interpolantCount];
				 	else
				 		currentSurface.corners[0].interpolants[interpolantCount] = currentSurface.corners[1].interpolants[interpolantCount];
				}
			if (abs((float) currentSurface.corners[0].interpolants[interpolantCount] - currentSurface.corners[2].interpolants[interpolantCount]) <= 512.0)
				{
				 	if (currentSurface.corners[0].interpolants[interpolantCount] > currentSurface.corners[2].interpolants[interpolantCount])
				 		currentSurface.corners[2].interpolants[interpolantCount] = currentSurface.corners[0].interpolants[interpolantCount];
				 	else
				 		currentSurface.corners[0].interpolants[interpolantCount] = currentSurface.corners[2].interpolants[interpolantCount];
				}
			if (abs((float) currentSurface.corners[1].interpolants[interpolantCount] - currentSurface.corners[2].interpolants[interpolantCount]) <= 512.0)
				{
				 	if (currentSurface.corners[1].interpolants[interpolantCount] > currentSurface.corners[2].interpolants[interpolantCount])
				 		currentSurface.corners[2].interpolants[interpolantCount] = currentSurface.corners[1].interpolants[interpolantCount];
				 	else
				 		currentSurface.corners[1].interpolants[interpolantCount] = currentSurface.corners[2].interpolants[interpolantCount];
				}
		}
		
		for (UInt16 attributeIndex = 0; attributeIndex < theData.numTriangleAttributeTypes; attributeIndex++)
		{
			if (theData.triangleAttributeTypes[attributeIndex].attributeType == kQ3AttributeTypeNormal)
			{
				TQ3Vector3D* theNormalDataPtr = (TQ3Vector3D*) theData.triangleAttributeTypes[attributeIndex].data;
				TQ3Vector3D theNormalizedNormal; // heh
				::Q3Vector3D_Transform(&theNormalDataPtr[surfaceCount], &QD3DToMythMatrix, &theNormalizedNormal);
				::Q3Vector3D_Normalize(&theNormalizedNormal, &theNormalizedNormal);
				
				currentSurface.normal.i = theNormalizedNormal.x * 0x4000;
				currentSurface.normal.j = theNormalizedNormal.y * 0x4000;
				currentSurface.normal.k = theNormalizedNormal.z * 0x4000;
				
				currentSurface.d = 0 -
					((currentSurface.normal.i * mVertices[vertex0].p.x) +
					(currentSurface.normal.j * mVertices[vertex0].p.y) +
					(currentSurface.normal.k * mVertices[vertex0].p.z));
			}
		}

		currentSurface.next_negative_surface_index = -1;
		currentSurface.next_parallel_surface_index = -1;
		currentSurface.next_positive_surface_index = -1;

		currentSurface.mesh_cell_contained = 0;
		currentSurface.dependency_table_index = 0xFFFF;
		currentSurface.unused[0] = 0;
		
		// Save the triangle data for rendering
		TQ3TriangleData& currentTriangle = mTriangles[mGeometry.surface_count];
		currentTriangle.vertices[0].point = theData.points[theData.triangles[surfaceCount].pointIndices[0]];
		currentTriangle.vertices[0].attributeSet = ::Q3AttributeSet_New();
		currentTriangle.vertices[1].point = theData.points[theData.triangles[surfaceCount].pointIndices[1]];
		currentTriangle.vertices[1].attributeSet = ::Q3AttributeSet_New();
		currentTriangle.vertices[2].point = theData.points[theData.triangles[surfaceCount].pointIndices[2]];
		currentTriangle.vertices[2].attributeSet = ::Q3AttributeSet_New();
		
		currentTriangle.triangleAttributeSet = ::Q3AttributeSet_New();
		if ( ::Q3AttributeSet_Contains(theData.triMeshAttributeSet, kQ3AttributeTypeSurfaceShader) == kQ3True )
		{
//			theStatus = ::Q3AttributeSet_Add(currentTriangle.triangleAttributeSet, kQ3AttributeTypeSurfaceShader, mShaders[mGeometry.material_count]);
		}
		for ( int iAttribute = 0; iAttribute < theData.numTriangleAttributeTypes; iAttribute++ )
		{
			TQ3AttributeType attributeType = theData.triangleAttributeTypes[iAttribute].attributeType;
			switch(attributeType)
			{
				case kQ3AttributeTypeNormal:
					theStatus = ::Q3AttributeSet_Add(currentTriangle.triangleAttributeSet, attributeType, &((TQ3Vector3D*)theData.triangleAttributeTypes[iAttribute].data)[surfaceCount]);
					break;
			    case kQ3AttributeTypeTransparencyColor:
//					theStatus = ::Q3AttributeSet_Add(currentTriangle.triangleAttributeSet, attributeType, &((TQ3ColorRGB*)theData.triangleAttributeTypes[iAttribute].data)[surfaceCount]);
					break;
			}
		}
		for ( int iAttribute = 0; iAttribute < theData.numVertexAttributeTypes; iAttribute++ )
		{
			TQ3AttributeType attributeType = theData.vertexAttributeTypes[iAttribute].attributeType;
			switch(attributeType)
			{
				case kQ3AttributeTypeNormal:
					theStatus = ::Q3AttributeSet_Add(currentTriangle.vertices[0].attributeSet, attributeType, &((TQ3Vector3D*)theData.vertexAttributeTypes[iAttribute].data)[theData.triangles[surfaceCount].pointIndices[0]]);
					theStatus = ::Q3AttributeSet_Add(currentTriangle.vertices[1].attributeSet, attributeType, &((TQ3Vector3D*)theData.vertexAttributeTypes[iAttribute].data)[theData.triangles[surfaceCount].pointIndices[1]]);
					theStatus = ::Q3AttributeSet_Add(currentTriangle.vertices[2].attributeSet, attributeType, &((TQ3Vector3D*)theData.vertexAttributeTypes[iAttribute].data)[theData.triangles[surfaceCount].pointIndices[2]]);
					break;
			    case kQ3AttributeTypeSurfaceUV:
			    case kQ3AttributeTypeShadingUV:
					theStatus = ::Q3AttributeSet_Add(currentTriangle.vertices[0].attributeSet, attributeType, &((TQ3Param2D*)theData.vertexAttributeTypes[iAttribute].data)[theData.triangles[surfaceCount].pointIndices[0]]);
					theStatus = ::Q3AttributeSet_Add(currentTriangle.vertices[1].attributeSet, attributeType, &((TQ3Param2D*)theData.vertexAttributeTypes[iAttribute].data)[theData.triangles[surfaceCount].pointIndices[1]]);
					theStatus = ::Q3AttributeSet_Add(currentTriangle.vertices[2].attributeSet, attributeType, &((TQ3Param2D*)theData.vertexAttributeTypes[iAttribute].data)[theData.triangles[surfaceCount].pointIndices[2]]);
					break;
			}
		}

		mGeometry.surface_count++;
	}

	LStr255 theName("Material #");
	theName += mGeometry.material_count;
	theName += '\0';
	::BlockMoveData(&theName[1], mMaterials[mGeometry.material_count].name, MAXIMUM_GEOMETRY_MATERIAL_NAME_LENGTH);
	mMaterials[mGeometry.material_count].sequence_index = mGeometry.material_count;
	mMaterials[mGeometry.material_count].collection_index = 0;
	mMaterials[mGeometry.material_count].color_table_index = 0;
	mMaterials[mGeometry.material_count].bitmap_index = mGeometry.material_count;

	mGeometry.material_count++;

	theStatus = ::Q3TriMesh_EmptyData(&theData);
	ThrowIf_(theStatus != kQ3Success);
}

// ---------------------------------------------------------------------------------
//		 SetCorner
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SetCorner(TQ3TriMeshData& inData, UInt16 inVertex, geometry_corner& ioCorner)
{
	ioCorner.interpolants[0] = 0;
	ioCorner.interpolants[1] = 0;
	ioCorner.vertex_index = mGeometry.vertex_count;
	ioCorner.encoded_normal = 0;
	
	for (UInt16 vertexIndex = 0; vertexIndex < mGeometry.vertex_count; vertexIndex++)
	{
		if ((inData.points[inVertex].x == mPoints[vertexIndex].x) &&
			(inData.points[inVertex].y == mPoints[vertexIndex].y) &&
			(inData.points[inVertex].z == mPoints[vertexIndex].z))
		{
			ioCorner.vertex_index = vertexIndex;
			break;
		}
	}
	
	if (ioCorner.vertex_index == mGeometry.vertex_count)
	{
		TQ3Point3D mythPoint;
		::Q3Point3D_Transform(&inData.points[inVertex], &mTransformMatrix, &mythPoint);
		geometry_vertex mythVertex;
		mythVertex.p.x = mythPoint.x;
		mythVertex.p.y = mythPoint.y;
		mythVertex.p.z = mythPoint.z;

		mVertices[mGeometry.vertex_count] = mythVertex;
		mVertexData[mGeometry.vertex_count].flags = 0;
		mPoints[mGeometry.vertex_count] = inData.points[inVertex];
		mGeometry.vertex_count++;
	}
	
	for (UInt16 attributeIndex = 0; attributeIndex < inData.numVertexAttributeTypes; attributeIndex++)
	{
		if ((inData.vertexAttributeTypes[attributeIndex].attributeType == kQ3AttributeTypeSurfaceUV)
		|| (inData.vertexAttributeTypes[attributeIndex].attributeType == kQ3AttributeTypeShadingUV))
		{
			TQ3Param2D* theUVMapDataPtr = (TQ3Param2D*) inData.vertexAttributeTypes[attributeIndex].data;
			float temp = (1 - theUVMapDataPtr[inVertex].v) * 0xFFFF;
			ioCorner.interpolants[0] = temp;
			temp = theUVMapDataPtr[inVertex].u * 0xFFFF;
			ioCorner.interpolants[1] = temp;
		}
		else if (inData.vertexAttributeTypes[attributeIndex].attributeType == kQ3AttributeTypeNormal)
		{
			TQ3Vector3D* theNormalDataPtr = (TQ3Vector3D*) inData.vertexAttributeTypes[attributeIndex].data;
			TQ3Vector3D theNormalizedNormal; // heh
			::Q3Vector3D_Transform(&theNormalDataPtr[inVertex], &QD3DToMythMatrix, &theNormalizedNormal);
			::Q3Vector3D_Normalize(&theNormalizedNormal, &theNormalizedNormal);
			
			world_vector3d testNormal;
			testNormal.i = theNormalizedNormal.x * 0x4000;
			testNormal.j = theNormalizedNormal.y * 0x4000;
			testNormal.k = theNormalizedNormal.z * 0x4000;

        	Int16 precalculated_normal_index;
        	Int16 closest_index;
       		UInt32 closest_distance = 0xFFFFFFFF;
        
	        for (precalculated_normal_index = 0; precalculated_normal_index < NUMBER_OF_PRECALCULATED_MODEL_NORMALS; ++precalculated_normal_index)
	        {
                Int32 di = testNormal.i - mVertexNormals[precalculated_normal_index].normal.i;
                Int32 dj = testNormal.j - mVertexNormals[precalculated_normal_index].normal.j;
                Int32 dk = testNormal.k - mVertexNormals[precalculated_normal_index].normal.k;
                UInt32 distance= di*di + dj*dj + dk*dk;
                
                if (distance < closest_distance)
                {
                    closest_index = precalculated_normal_index;
                    closest_distance = distance;
                }
	        }
	        
	        ioCorner.encoded_normal = closest_index;
		}
	}
}


// ---------------------------------------------------------------------------------
//		 SubmitPickSurfaces
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SubmitPickSurfaces( TQ3ViewObject inView, UInt32 inInitialID )
{
	for ( int iTriangle = 0; iTriangle < mGeometry.surface_count; iTriangle++)
	{	
		::Q3PickIDStyle_Submit( inInitialID + iTriangle, inView );
		::Q3Triangle_Submit( &mTriangles[iTriangle], inView );
	}
}


// ---------------------------------------------------------------------------------
//		 SubmitDrawSurfaces ( with SetTransparency utility )
// ---------------------------------------------------------------------------------

inline void
C3dmfDocument::SetTransparency( TQ3ViewObject inView, UInt16 inTriangle )
{
	TQ3ColorRGB theColorValue;
	
	if ( ( inTriangle < mGeometry.surface_count ) && ( mSurfaces[inTriangle].flags & _geometry_surface_is_transparent_flag ) )
	{																									
		float transparencyFraction = 1.0 - ((float) mSurfaces[inTriangle].transparency / (float) FIXED_ONE);
		theColorValue.r = transparencyFraction;
		theColorValue.g = transparencyFraction;
		theColorValue.b = transparencyFraction;
	}
	else
	{
		theColorValue.r = 1.0;
		theColorValue.g = 1.0;
		theColorValue.b = 1.0;
	}
	
	::Q3Attribute_Submit(kQ3AttributeTypeTransparencyColor, &theColorValue, inView);
}

void
C3dmfDocument::SubmitDrawSurfaces( TQ3ViewObject inView, UInt32 inFlag )
{
	TQ3ColorRGB theColorValue;

	if ( ( inFlag == _GEOMETRY_DEPENDENCY_FAKE_FLAG ) && ( mDependencySurface >= 0 ) )
	{
		Boolean bDrawnArray[MAXIMUM_SURFACES_PER_GEOMETRY];
		for (int i = 0; i < MAXIMUM_SURFACES_PER_GEOMETRY; i++ )
		{
			bDrawnArray[i] = false;
		}
		bDrawnArray[mDependencySurface] = true;
		
		// Draw the dependent surfaces
		theColorValue.r = 0.0;
		theColorValue.g = 0.0;
		theColorValue.b = 1.0;
		::Q3Attribute_Submit( kQ3AttributeTypeDiffuseColor, &theColorValue, inView );

		LArray* dependencyArray = nil;
		if ( mSurfaces[mDependencySurface].dependency_table_index >= 0 )
		{
			mDependencyArrays->FetchItemAt( mSurfaces[mDependencySurface].dependency_table_index, &dependencyArray );
		}
			
		if ( dependencyArray )
		{
			for ( int iSurface = 1; iSurface <= dependencyArray->GetCount(); iSurface++ )
			{
				short dependency = -1;
				if ( ( dependencyArray->FetchItemAt( iSurface, &dependency ) ) && ( dependency >= 0 ) )
				{
					TQ3Status theStatus = ::Q3Shader_Submit( mShaders[mSurfaces[dependency].material_index], inView );
					
					SetTransparency( inView, dependency );
					theStatus = ::Q3Triangle_Submit( &mTriangles[dependency], inView );
					
					bDrawnArray[dependency] = true;
				}
			}
		}
		
		// Draw the non-dependent surfaces
		theColorValue.r = 1.0;
		theColorValue.g = 1.0;
		theColorValue.b = 1.0;
		::Q3Attribute_Submit( kQ3AttributeTypeDiffuseColor, &theColorValue, inView );

		for ( int iTriangle = 0; iTriangle < mGeometry.surface_count; iTriangle++)
		{
			TQ3Status theStatus = ::Q3Shader_Submit( mShaders[mSurfaces[iTriangle].material_index], inView );

			if ( !bDrawnArray[iTriangle] )
			{
				SetTransparency( inView, iTriangle );
				theStatus = ::Q3Triangle_Submit( &mTriangles[iTriangle], inView );
				if ( theStatus != kQ3Success )
				{
					break;
				}
			}
		}
		
		// Draw main surface last to help visibility
		theColorValue.r = 1.0;
		theColorValue.g = 0.0;
		theColorValue.b = 0.0;
		::Q3Attribute_Submit( kQ3AttributeTypeDiffuseColor, &theColorValue, inView );
		
		TQ3Status theStatus = ::Q3Shader_Submit( mShaders[mSurfaces[mDependencySurface].material_index], inView );

		SetTransparency( inView, mDependencySurface );
		theStatus = ::Q3Triangle_Submit( &mTriangles[mDependencySurface], inView );
	}
	else
	for ( int iTriangle = 0; iTriangle < mGeometry.surface_count; iTriangle++)
	{
		if ( inFlag )	// Don't mark if we're just drawing the triangles
		{
			if ( mSurfaces[iTriangle].flags & inFlag )
			{
				theColorValue.r = 1.0;
				theColorValue.g = 0.0;
				theColorValue.b = 0.0;
			}
			else
			{
				theColorValue.r = 1.0;
				theColorValue.g = 1.0;
				theColorValue.b = 1.0;
			}
			::Q3Attribute_Submit( kQ3AttributeTypeDiffuseColor, &theColorValue, inView );
		}
			
		TQ3Status theStatus = ::Q3Shader_Submit( mShaders[mSurfaces[iTriangle].material_index], inView );
		
		SetTransparency( inView, iTriangle );
		theStatus = ::Q3Triangle_Submit( &mTriangles[iTriangle], inView );
		if ( theStatus != kQ3Success )
		{
			break;
		}
	}
}


// ---------------------------------------------------------------------------------
//		 SetSurfaceFlag
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SetSurfaceFlag( UInt32 inSurface, UInt32 inFlag )
{
	if ( inFlag == _geometry_surface_is_transparent_flag )
	{
		if ( mSurfaces[inSurface].flags & _geometry_surface_is_transparent_flag )
		{
			if ( mSurfaces[inSurface].transparency != mTransparency )
			{
				mSurfaces[inSurface].transparency = mTransparency;
			}
			else
			{
				mSurfaces[inSurface].flags ^= inFlag;
				mSurfaces[inSurface].transparency = 0;
			}
		}
		else
		{
			mSurfaces[inSurface].flags |= inFlag;
			mSurfaces[inSurface].transparency = mTransparency;
		}
	}
	else if ( inFlag == _GEOMETRY_DEPENDENCY_FAKE_FLAG )
	{
		if ( mDependencySurface < 0 )
		{
			mDependencySurface = inSurface;
		}
		else
		{
			LArray* dependencyArray = nil;
			if ( mSurfaces[mDependencySurface].dependency_table_index < 0 )
			{
				dependencyArray = new LArray( sizeof( short ) );
				mSurfaces[mDependencySurface].dependency_table_index = mDependencyArrays->AddItem( &dependencyArray );
			}
			else
			{
				mDependencyArrays->FetchItemAt( mSurfaces[mDependencySurface].dependency_table_index, &dependencyArray );
			}
			
			if ( dependencyArray )
			{
				short dependency = inSurface;
				if ( dependencyArray->ContainsItem( &dependency ) )
				{
					dependencyArray->Remove( &dependency );
				}
				else
				{
					dependencyArray->AddItem( &dependency );
				}
			}
		}
	}
	else
	{
		mSurfaces[inSurface].flags ^= inFlag;
	}
}


// ---------------------------------------------------------------------------------
//		 SubmitPickMesh
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SubmitPickMesh( TQ3ViewObject inView, UInt32 inInitialID )
{
	TQ3PolygonData thePolygon;
	thePolygon.numVertices = 4;
	thePolygon.vertices = new TQ3Vertex3D[4];
	for ( int iVertex = 0; iVertex < 4; iVertex++ )
	{
		thePolygon.vertices[iVertex].attributeSet = NULL;
	}
	thePolygon.polygonAttributeSet = NULL;
	
	// Set the transform
	TQ3Matrix4x4 theTransformMatrix;
	::Q3Matrix4x4_Copy(&MythToQD3DMatrix, &theTransformMatrix);
	
	// Offset the origin
	TQ3Matrix4x4 theTranslationMatrix;
	::Q3Matrix4x4_SetTranslate(&theTranslationMatrix,
		(float) -mMeshOffsetX / (float) GEOMETRY_NORMAL_SCALE,
		(float) -mMeshOffsetY / (float) GEOMETRY_NORMAL_SCALE,
		(float) -mMeshOffsetZ / (float) GEOMETRY_NORMAL_SCALE);

	// Scale the points to QD3D coordinates
	float fMeshScale = 1.0 / mMythScale;
	TQ3Matrix4x4 theScalingMatrix;
	::Q3Matrix4x4_SetScale(&theScalingMatrix,
		fMeshScale,
		fMeshScale,
		fMeshScale);	
		
	::Q3Matrix4x4_Multiply(&theTranslationMatrix, &theTransformMatrix, &theTransformMatrix);	
	::Q3Matrix4x4_Multiply(&theTransformMatrix, &theScalingMatrix, &theTransformMatrix);	

	for (UInt16 cellIndex = 0; cellIndex < mModel.model_cell_count; cellIndex++)
	{
		float fX = (float) mCells[cellIndex].x;
		float fY = (float) mCells[cellIndex].y;
		
		// Once again, we have some coordinate fun.
		// Y & Z are swapping and the mesh moves "opposite" of the alignment adjustments.
		thePolygon.vertices[0].point.x = fX - mAlignmentX;
		thePolygon.vertices[0].point.y = fY - mAlignmentZ;
		thePolygon.vertices[0].point.z = mAlignmentY;
		thePolygon.vertices[1].point.x = fX - mAlignmentX + 1.0;
		thePolygon.vertices[1].point.y = fY - mAlignmentZ;
		thePolygon.vertices[1].point.z = mAlignmentY;
		thePolygon.vertices[2].point.x = fX - mAlignmentX + 1.0;
		thePolygon.vertices[2].point.y = fY - mAlignmentZ + 1.0;
		thePolygon.vertices[2].point.z = mAlignmentY;
		thePolygon.vertices[3].point.x = fX - mAlignmentX;
		thePolygon.vertices[3].point.y = fY - mAlignmentZ + 1.0;
		thePolygon.vertices[3].point.z = mAlignmentY;
			
		::Q3Point3D_Transform(&thePolygon.vertices[0].point, &theTransformMatrix, &thePolygon.vertices[0].point);
		::Q3Point3D_Transform(&thePolygon.vertices[1].point, &theTransformMatrix, &thePolygon.vertices[1].point);
		::Q3Point3D_Transform(&thePolygon.vertices[2].point, &theTransformMatrix, &thePolygon.vertices[2].point);
		::Q3Point3D_Transform(&thePolygon.vertices[3].point, &theTransformMatrix, &thePolygon.vertices[3].point);
			
		::Q3PickIDStyle_Submit( inInitialID + cellIndex, inView );

		TQ3Status theStatus = ::Q3Polygon_Submit( &thePolygon, inView );
		if ( theStatus != kQ3Success )
		{
			break;
		}
	}

	delete[] thePolygon.vertices;
}


// ---------------------------------------------------------------------------------
//		 SubmitDrawMesh
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SubmitDrawMesh( TQ3ViewObject inView, UInt32 inFlag )
{
	TQ3PolygonData thePolygon;
	thePolygon.numVertices = 4;
	thePolygon.vertices = new TQ3Vertex3D[4];
	for ( int iVertex = 0; iVertex < 4; iVertex++ )
	{
		thePolygon.vertices[iVertex].attributeSet = NULL;
	}
	thePolygon.polygonAttributeSet = NULL;
	
	// Set the transform
	TQ3Matrix4x4 theTransformMatrix;
	::Q3Matrix4x4_Copy(&MythToQD3DMatrix, &theTransformMatrix);
	
	// Offset the origin
	TQ3Matrix4x4 theTranslationMatrix;
	::Q3Matrix4x4_SetTranslate(&theTranslationMatrix,
		(float) -mMeshOffsetX / (float) GEOMETRY_NORMAL_SCALE,
		(float) -mMeshOffsetY / (float) GEOMETRY_NORMAL_SCALE,
		(float) -mMeshOffsetZ / (float) GEOMETRY_NORMAL_SCALE);

	// Scale the points to QD3D coordinates
	float fMeshScale = 1.0 / mMythScale;
	TQ3Matrix4x4 theScalingMatrix;
	::Q3Matrix4x4_SetScale(&theScalingMatrix,
		fMeshScale,
		fMeshScale,
		fMeshScale);	
		
	::Q3Matrix4x4_Multiply(&theTranslationMatrix, &theTransformMatrix, &theTransformMatrix);	
	::Q3Matrix4x4_Multiply(&theTransformMatrix, &theScalingMatrix, &theTransformMatrix);	
	
	TQ3ColorRGB theFlagColor;
	switch ( inFlag )
	{
		case _model_mesh_cell_replaces_mesh_flag:
			theFlagColor.r = 0.8;
			theFlagColor.g = 0.94;
			theFlagColor.b = 0.8;
			break;
		case _mesh_cell_is_impassable_walking_flag:
			theFlagColor.r = 0.05;
			theFlagColor.g = 0.4;
			theFlagColor.b = 0.05;
			break;
		case _mesh_cell_is_impassable_flying_flag:
			theFlagColor.r = 0.4;
			theFlagColor.g = 0.99;
			theFlagColor.b = 0.1;
			break;
		case _mesh_cell_overlapped_by_model_flag:
		default:
			theFlagColor.r = 0.25;
			theFlagColor.g = 0.6;
			theFlagColor.b = 0.25;
	}

	const float kMeshInset = 0.035;
	for (UInt16 cellCount = 0; cellCount < mModel.model_cell_count; cellCount++)
	{
		float fX = (float) mCells[cellCount].x;
		float fY = (float) mCells[cellCount].y;
		
		// Once again, we have some coordinate fun.
		// Y & Z are swapping and the mesh moves "opposite" of the alignment adjustments.
		thePolygon.vertices[0].point.x = fX - mAlignmentX + kMeshInset;
		thePolygon.vertices[0].point.y = fY - mAlignmentZ + kMeshInset;
		thePolygon.vertices[0].point.z = mAlignmentY;
		thePolygon.vertices[1].point.x = fX - mAlignmentX + 1.0 - kMeshInset;
		thePolygon.vertices[1].point.y = fY - mAlignmentZ + kMeshInset;
		thePolygon.vertices[1].point.z = mAlignmentY;
		thePolygon.vertices[2].point.x = fX - mAlignmentX + 1.0 - kMeshInset;
		thePolygon.vertices[2].point.y = fY - mAlignmentZ + 1.0 - kMeshInset;
		thePolygon.vertices[2].point.z = mAlignmentY;
		thePolygon.vertices[3].point.x = fX - mAlignmentX + kMeshInset;
		thePolygon.vertices[3].point.y = fY - mAlignmentZ + 1.0 - kMeshInset;
		thePolygon.vertices[3].point.z = mAlignmentY;
			
		::Q3Point3D_Transform(&thePolygon.vertices[0].point, &theTransformMatrix, &thePolygon.vertices[0].point);
		::Q3Point3D_Transform(&thePolygon.vertices[1].point, &theTransformMatrix, &thePolygon.vertices[1].point);
		::Q3Point3D_Transform(&thePolygon.vertices[2].point, &theTransformMatrix, &thePolygon.vertices[2].point);
		::Q3Point3D_Transform(&thePolygon.vertices[3].point, &theTransformMatrix, &thePolygon.vertices[3].point);
			
		TQ3ColorRGB theColorValue;
		if (  mCells[cellCount].flags & inFlag )
		{
			theColorValue = theFlagColor;
		}
		else
		{
			theColorValue.r = 0.4;
			theColorValue.g = 0.7;
			theColorValue.b = 0.4;
		}
		::Q3Attribute_Submit( kQ3AttributeTypeDiffuseColor, &theColorValue, inView );

		TQ3Status theStatus = ::Q3Polygon_Submit( &thePolygon, inView );
		if ( theStatus != kQ3Success )
		{
			break;
		}
	}

	delete[] thePolygon.vertices;
}


// ---------------------------------------------------------------------------------
//		 SetMeshFlag
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SetMeshFlag( UInt32 inCell, UInt32 inFlag )
{
	if ( inCell < MAXIMUM_CELLS_PER_MODEL )
	{
		mCells[inCell].flags ^= inFlag;
	}
}


// ---------------------------------------------------------------------------------
//		 SubmitPickVertices
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SubmitPickVertices( TQ3ViewObject inView, UInt32 inInitialID )
{
	TQ3PointData thePoint;
	thePoint.pointAttributeSet = NULL;
	
	for ( int iVertex = 0; iVertex < mGeometry.vertex_count; iVertex++)
	{	
		thePoint.point = mPoints[iVertex];
		
		::Q3PickIDStyle_Submit( inInitialID + iVertex, inView );
		::Q3Point_Submit( &thePoint, inView );
	}
}


// ---------------------------------------------------------------------------------
//		 SubmitDrawVertices
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SubmitDrawVertices( TQ3ViewObject inView, UInt32 inFlag )
{
	TQ3ColorRGB theColorValue;
	theColorValue.r = 1.0;
	theColorValue.g = 0.0;
	theColorValue.b = 0.0;

	::Q3Attribute_Submit( kQ3AttributeTypeDiffuseColor, &theColorValue, inView );
	
	float fRadius = ::Q3Point3D_Distance( &mBoundingBox.min, &mBoundingBox.max ) / 100.0;

	TQ3EllipsoidData theEllipsoid;
    ::Q3Vector3D_Set( &theEllipsoid.orientation, 0.0, 0.0, fRadius );
    ::Q3Vector3D_Set( &theEllipsoid.majorRadius, fRadius, 0.0, 0.0 );
    ::Q3Vector3D_Set( &theEllipsoid.minorRadius, 0.0, fRadius, 0.0 );
    theEllipsoid.uMin = 0.0;
    theEllipsoid.uMax = 1.0;
    theEllipsoid.vMin = 0.0;
    theEllipsoid.vMax = 1.0;
    theEllipsoid.caps = kQ3EndCapNone;
    theEllipsoid.interiorAttributeSet = NULL;
    theEllipsoid.ellipsoidAttributeSet = NULL;
	
	for ( int iVertex = 0; iVertex < mGeometry.vertex_count; iVertex++)
	{
		if ( mVertexData[iVertex].flags & inFlag )
		{
			theEllipsoid.origin = mPoints[iVertex];
		
			TQ3Status theStatus = ::Q3Ellipsoid_Submit( &theEllipsoid, inView );
			if ( theStatus != kQ3Success )
			{
				break;
			}
		}
	}
}


// ---------------------------------------------------------------------------------
//		 SetVertexFlag
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SetVertexFlag( UInt32 inVertex, UInt32 inFlag )
{
	mVertexData[inVertex].flags ^= inFlag;
}


// ---------------------------------------------------------------------------------
//		 IsModified
// ---------------------------------------------------------------------------------

Boolean
C3dmfDocument::IsModified()
{
	return false;
}


// ---------------------------------------------------------------------------------
//		 DoAESave
// ---------------------------------------------------------------------------------

void
C3dmfDocument::DoAESave(
	FSSpec	& /*inFileSpec*/,
	OSType	/*inFileType*/ )
{
/*
	// Delete the existing file object.
	// Note: this does nothing to the actual file on disk.
	delete mFileStream;
	
	// Make a new file object.
	mFileStream = new LFile( inFileSpec );
	
	// Get the proper file type.
	OSType	theFileType = 'TEXT';
	if ( inFileType != fileType_Default ) theFileType = inFileType;

	// Make new file on disk (we'll use
	// SimpleText's creator for this example).
	mFileStream->CreateNewDataFile( 'ttxt', theFileType );
	
	// Write out the data.
	DoSave();

	// Document now has a specified file.
	mIsSpecified = true;
*/
}


// ---------------------------------------------------------------------------------
//		 DoSave
// ---------------------------------------------------------------------------------

void
C3dmfDocument::DoSave()
{
}


// ---------------------------------------------------------------------------------
//		 DoExport
// ---------------------------------------------------------------------------------

void
C3dmfDocument::DoExport()
{
	if ( !ValidMeshData() )
	{
		::ParamText( LStr255( rSTRl_Alerts, eAlertModelTooLargeToSave ), nil, nil, nil );
		UModalAlerts::StopAlert( ALRT_Generic );
		return;
	}
	
	Try_
	{
// Clean up the Myth data
	AdjustMythVertices();
	
// Set the file info
	LStr255 theTagIDString;
	LStr255 theNameBase;
	mTagField->GetDescriptor(theTagIDString);
	mNameField->GetDescriptor(theNameBase);
	FourCharCode theTagID = theTagIDString;

// Write the graphic file
// Create a new file with the .256 extension
	FSSpec theGraphicFSSpec;
	mFileStream->GetSpecifier(theGraphicFSSpec);
	LString::CopyPStr(theNameBase + "\p.256", theGraphicFSSpec.name, sizeof(StrFileName));

	LFileStream theGraphicFile(theGraphicFSSpec);
	Try_
	{
		theGraphicFile.CreateNewDataFile('M2SB', 'File');
	}
	Catch_( inErr )
	{
		if (inErr.GetErrorCode() == dupFNErr)
		{
			SignalPStr_(LStr255("\pDuplicate file, tell eli to add the damn dialog."));
			Throw_Err(inErr);
		}
		else
		{
			SignalPStr_(LStr255("\pFile Error: ") + inErr.GetErrorString());
			Throw_Err(inErr);
		}
	} EndCatch_
// Open the data fork.
	theGraphicFile.OpenDataFork( fsRdWrPerm );

// Build the tag header
	myth2dir_Header theFileHeader;
	theFileHeader.mIdentifierM2 = 0;
  	theFileHeader.mFlagsM2 = 0;
  	theFileHeader.mTypeM2 = 0;
  	::BlockMoveData("eli made this graphic\0", theFileHeader.mName, 30);
  	theFileHeader.mTagGroup = COLLECTION_GROUP_TAG;
  	theFileHeader.mTagID = theTagID;
  	theFileHeader.mTagOffset = sizeof(myth2dir_Header);
// We'll go and write over this when we know the size
	theFileHeader.mTagSize = 0;
  	theFileHeader.mUserData = 0;
  	theFileHeader.mVersionM2 = 1;
  	theFileHeader.mFoundationM2 = 2;
		// Foundation types
		//  foundation_SmallInstall		= 0,
		//  foundation_MediumInstall	= 1,
		//  foundation_LargeInstall		= 2,
		//  foundation_SmallInterface	= 3,
		//  foundation_LargeInterface	= 4
  	theFileHeader.mOwnerIndexM2 = 0xFF;
  	theFileHeader.mMythCode = 'mth2';

// Write the tag header (rewrite it later)
	SInt32 byteCount = sizeof(myth2dir_Header);
	ThrowIfOSErr_(theGraphicFile.PutBytes(&theFileHeader, byteCount));

// Write the data file - was changed to match Amber's file order on
// 4/1/99 to aid in mututal debugging - blame Vodi, not me.

// Heh, turns out that Vodi's a genius - the Bungie bastards assume the
// sequence data immediately follows the sequence reference

// Build the bitmap header
	collection_header theBitmapHeader;
	UInt32 currentOffset = 0;
	
	theBitmapHeader.flags = 0;
	theBitmapHeader.user_data_flags = _collection_can_be_used_for_textures_flag;

	for (UInt16 unusedCount = 0; unusedCount < 28; unusedCount++)
		theBitmapHeader.unused1[unusedCount] = 0;

    theBitmapHeader.shading_table_count = 0;
    theBitmapHeader.shading_table_pointers = NULL;

	for (UInt16 unusedCount = 0; unusedCount < 24; unusedCount++)
		theBitmapHeader.unused3[unusedCount] = 0;

    theBitmapHeader.data_offset = SIZEOF_STRUCT_COLLECTION_HEADER;
    theBitmapHeader.data_size = currentOffset;
    theBitmapHeader.data_delta = 0;
    theBitmapHeader.data = NULL;

	for (UInt16 unusedCount = 0; unusedCount < 20; unusedCount++)
    	theBitmapHeader.unused2[unusedCount] = 0;

	for (UInt16 unusedCount = 0; unusedCount < EXTRACTOR_DATA_SIZE; unusedCount++)
    	theBitmapHeader.extractor_data[unusedCount] = 0;

// Write the bitmap header (rewrite it later)
	byteCount = SIZEOF_STRUCT_COLLECTION_HEADER;
	ThrowIfOSErr_(theGraphicFile.PutBytes(&theBitmapHeader, byteCount));

// Write the color table
	theBitmapHeader.color_table_count = 1;
	theBitmapHeader.color_tables_offset = currentOffset;
	theBitmapHeader.color_tables_size = sizeof(color_table);
    theBitmapHeader.color_tables = NULL;
    currentOffset += theBitmapHeader.color_tables_size;

	PictInfo thePictInfo;
	ThrowIfOSErr_(::RetrievePictInfo(mPictInfoID, &thePictInfo, 254));
	PaletteHandle theRealPalette = ::NewPalette(256, NULL, pmTolerant, pmTolerant);
	::CopyPalette(thePictInfo.thePalette, theRealPalette, 0, 1, 254);
	::SetEntryColor(theRealPalette, 0, &Color_White);
	CTabHandle theMacColorTableH = ::GetCTable(72);
	Palette2CTab(theRealPalette, theMacColorTableH);
	::CTabChanged(theMacColorTableH);
	ThrowIfOSErr_(::DisposePictInfo(mPictInfoID));
	{
		StHandleLocker lock((Handle) theMacColorTableH);
		color_table theColorTable;
		theColorTable.color_count = PIXEL8_MAXIMUM_COLORS;
		theColorTable.seed = 0;
		theColorTable.acceleration_cookie = 0;
		for (UInt16 unusedCount = 0; unusedCount < 5; unusedCount++)
	    	theColorTable.unused[unusedCount] = 0;
	    for (UInt16 colorCount = 0; colorCount < PIXEL8_MAXIMUM_COLORS; colorCount++)
	    {
	    	theColorTable.colors[colorCount].red = (*theMacColorTableH)->ctTable[colorCount].rgb.red;
	    	theColorTable.colors[colorCount].green = (*theMacColorTableH)->ctTable[colorCount].rgb.green;
	    	theColorTable.colors[colorCount].blue = (*theMacColorTableH)->ctTable[colorCount].rgb.blue;
	    	theColorTable.colors[colorCount].flags = 0;
	    }
		byteCount = sizeof(color_table);
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theColorTable, byteCount));
	}
	::DisposePalette(thePictInfo.thePalette);
	::DisposePalette(theRealPalette);

// Don't have any hue changes
    theBitmapHeader.hue_change_count = 0;
    theBitmapHeader.hue_changes_offset = currentOffset;
    theBitmapHeader.hue_changes_size = theBitmapHeader.hue_change_count * SIZEOF_STRUCT_HUE_CHANGE;
    theBitmapHeader.hue_changes = NULL;
    currentOffset += theBitmapHeader.hue_changes_size;

// Write the bitmap instances
    theBitmapHeader.bitmap_instance_count = mGeometry.material_count;
    theBitmapHeader.bitmap_instances_offset = currentOffset;
    theBitmapHeader.bitmap_instances_size = theBitmapHeader.bitmap_instance_count * SIZEOF_STRUCT_BITMAP_INSTANCE_DATA;
    theBitmapHeader.bitmap_instances = NULL;
    currentOffset += theBitmapHeader.bitmap_instances_size;

	bitmap_instance_data theBitmapInstance;
	theBitmapInstance.flags = _bitmap_instance_texture_flag;
	theBitmapInstance.minimum_light_intensity = 0x0000;
	theBitmapInstance.maximum_light_intensity = 0xFFFF;
    theBitmapInstance.transparency = 0;
    theBitmapInstance.unused1[1] = 0;

    theBitmapInstance.registration_point.v = 0;
    theBitmapInstance.registration_point.h = 0;

    theBitmapInstance.world_top = 0;
    theBitmapInstance.world_left = 0;
    theBitmapInstance.world_right = 0;
	theBitmapInstance.world_bottom = 0;

    theBitmapInstance.key_point.v = 0;
    theBitmapInstance.key_point.h = 0;

	for (UInt16 unusedCount = 0; unusedCount < 9; unusedCount++)
    	theBitmapInstance.unused2[unusedCount] = 0;
	for (UInt16 unusedCount = 0; unusedCount < EXTRACTOR_DATA_SIZE; unusedCount++)
    	theBitmapInstance.extractor_data[unusedCount] = 0;

    for (UInt16 bitmapCount = 0; bitmapCount < theBitmapHeader.bitmap_instance_count; bitmapCount++)
    {
        theBitmapInstance.registration_point.v = 0; // set to center of bitmap?
        theBitmapInstance.registration_point.h = 0; 
        theBitmapInstance.bitmap_index = bitmapCount;

		byteCount = SIZEOF_STRUCT_BITMAP_INSTANCE_DATA;
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theBitmapInstance, byteCount));
	}
	
// Write the sequence references
    theBitmapHeader.sequence_count = mGeometry.material_count;
    theBitmapHeader.sequence_references_offset = currentOffset;
    theBitmapHeader.sequence_references_size = theBitmapHeader.sequence_count * SIZEOF_STRUCT_SEQUENCE_REFERENCE;
    theBitmapHeader.sequence_references = NULL;
    currentOffset += theBitmapHeader.sequence_references_size;

	sequence_reference theSequenceRef;
    theSequenceRef.size = SIZEOF_STRUCT_SEQUENCE_DATA + sizeof(sequence_frame_data) + sizeof(UInt16); // 1 frame, 1 index
	theSequenceRef.sequence = NULL;
	for (UInt16 unusedCount = 0; unusedCount < 18; unusedCount++)
    	theSequenceRef.unused[unusedCount] = 0;
	for (UInt16 unusedCount = 0; unusedCount < EXTRACTOR_DATA_SIZE; unusedCount++)
    	theSequenceRef.extractor_data[unusedCount] = 0;
    	
    for (UInt16 sequenceCount = 0; sequenceCount < theBitmapHeader.sequence_count; sequenceCount++)
    {
		LStr255 theName("Sequence #");
		theName += (Int32) sequenceCount;
		theName += '\0';
		::BlockMoveData(&theName[1], theSequenceRef.name, SHAPE_NAME_LENGTH);

        theSequenceRef.offset = currentOffset;
        
        currentOffset += theSequenceRef.size;

		byteCount = SIZEOF_STRUCT_SEQUENCE_REFERENCE;
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theSequenceRef, byteCount));
	}
	
// Write the sequences
	sequence_data theSequenceData;
    theSequenceData.flags = _sequence_has_frame_data_flag;
	theSequenceData.pixels_to_world = 0x00010000;
	theSequenceData.number_of_views = 1;
	theSequenceData.frames_per_view = 1;
	theSequenceData.ticks_per_frame = 4;
    theSequenceData.key_frame = 0;
    theSequenceData.loop_frame = 0;

    for (UInt16 soundCount = 0; soundCount < NUMBER_OF_SEQUENCE_SOUNDS; soundCount++)
    {
        theSequenceData.sound_indexes[soundCount] = 0xFFFF;
        theSequenceData.sound_tags[soundCount] = 0xFFFFFFFF;
	}
	
    theSequenceData.transfer_mode = _transfer_mode_none;
    theSequenceData.transfer_period = 4;

    theSequenceData.radius = 0xFFFFFFFF;
    theSequenceData.height0 = 0;
    theSequenceData.height1 = 0;
    theSequenceData.world_radius = 0;
    theSequenceData.world_height0 = 0;
    theSequenceData.world_height1 = 0;

	for (UInt16 unusedCount = 0; unusedCount < 3; unusedCount++)
    	theSequenceData.unused[unusedCount] = 0;

//
	sequence_frame_data theFrameData;
    theFrameData.shadow_map_index = 0x0000; // was set to 0xFFFF, Vodi has it as 0x0000
	theFrameData.key_point.x = 0;
	theFrameData.key_point.y = 0;
	theFrameData.key_point.z = 0;

	for (UInt16 unusedCount = 0; unusedCount < 19; unusedCount++)
    	theFrameData.unused[unusedCount] = 0;

    for (UInt16 sequenceCount = 0; sequenceCount < theBitmapHeader.sequence_count; sequenceCount++)
    {
		byteCount = SIZEOF_STRUCT_SEQUENCE_DATA;
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theSequenceData, byteCount));
		byteCount = sizeof(sequence_frame_data);
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theFrameData, byteCount));

    	UInt16 theBitmapIndex = sequenceCount;
		byteCount = sizeof(UInt16);
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theBitmapIndex, byteCount));
	}
	
// Write the shadow maps
    theBitmapHeader.shadow_map_count = mGeometry.material_count;
    theBitmapHeader.shadow_maps_offset = currentOffset;
    theBitmapHeader.shadow_maps_size = theBitmapHeader.shadow_map_count * SIZEOF_STRUCT_SHADOW_MAP_DATA;
    theBitmapHeader.shadow_maps = NULL;
    currentOffset += theBitmapHeader.shadow_maps_size;

	shadow_map_data theShadowMap;
    theShadowMap.flags = 0;
	theShadowMap.registration_point.x = 0;
	theShadowMap.registration_point.y = 0;
    theShadowMap.bitmap_index = 0xFFFF;
    theShadowMap.unused[0] = 0;
	for (UInt16 unusedCount = 0; unusedCount < EXTRACTOR_DATA_SIZE; unusedCount++)
    	theShadowMap.extractor_data[unusedCount] = 0;

    for (UInt16 shadowCount = 0; shadowCount < theBitmapHeader.shadow_map_count; shadowCount++)
    {
		byteCount = SIZEOF_STRUCT_SHADOW_MAP_DATA;
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theShadowMap, byteCount));
	}

// Don't have any blend tables
    theBitmapHeader.blend_table_count = 0;
    theBitmapHeader.blend_table_offset = currentOffset;
    theBitmapHeader.blend_table_size = /* theBitmapHeader.blend_table_count * */ 0;
    theBitmapHeader.blend_table = NULL;
    currentOffset += theBitmapHeader.blend_table_size;

// Don't have any remapping tables
    theBitmapHeader.remapping_table_count = 0;
    theBitmapHeader.remapping_table_offset = currentOffset;
    theBitmapHeader.remapping_table_size = /* theBitmapHeader.remapping_table_count * */ 0;
    theBitmapHeader.remapping_table = NULL;
    currentOffset += theBitmapHeader.remapping_table_size;

// Write the bitmap references
    theBitmapHeader.bitmap_count = mGeometry.material_count;
    theBitmapHeader.bitmap_references_offset = currentOffset;
    theBitmapHeader.bitmap_references_size = theBitmapHeader.bitmap_count * SIZEOF_STRUCT_BITMAP_REFERENCE;
    theBitmapHeader.bitmap_references = NULL;
    currentOffset += theBitmapHeader.bitmap_references_size;

    for (UInt16 bitmapCount = 0; bitmapCount < theBitmapHeader.bitmap_count; bitmapCount++)
    {
    	bitmap_reference theBitmapReference;
    	
		LStr255 theName("Bitmap #");
		theName += (Int32) bitmapCount;
		theName += '\0';
		::BlockMoveData(&theName[1], theBitmapReference.name, SHAPE_NAME_LENGTH);

        theBitmapReference.offset = currentOffset;

        Rect textureBounds;
        (mGWorlds[bitmapCount])->GetBounds(textureBounds);
        
        theBitmapReference.size = sizeof(bitmap_data) - sizeof(void*);
        theBitmapReference.size += (textureBounds.bottom - textureBounds.top) * ((textureBounds.right - textureBounds.left) + sizeof(void*));
        
        currentOffset += theBitmapReference.size;
        
        theBitmapReference.bitmap = NULL;

		for (UInt16 unusedCount = 0; unusedCount < 17; unusedCount++)
        	theBitmapReference.unused[unusedCount] = 0;
		for (UInt16 unusedCount = 0; unusedCount < EXTRACTOR_DATA_SIZE; unusedCount++)
        	theBitmapReference.extractor_data[unusedCount] = 0;

        theBitmapReference.virtual_id = 0;
    	
		byteCount = SIZEOF_STRUCT_BITMAP_REFERENCE;
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theBitmapReference, byteCount));
	}
	
// Write the bitmaps
	bitmap_data theBitmapData;
    theBitmapData.flags = _power_of_two_flag;
    theBitmapData.logical_bit_depth = _8bit;
    theBitmapData.physical_byte_depth_minus_one = 0;
    theBitmapData.encoding = 0;
        
    theBitmapData.hardware_texture_cache_index = 0;
    theBitmapData.hardware_address = 0;
    theBitmapData.hardware_format = NULL;

    theBitmapData.hardware_width_bits = 0;
    theBitmapData.hardware_height_bits = 0;
    theBitmapData.hardware_height = 0;
    theBitmapData.hardware_width = 0;
    theBitmapData.x0 = 0;
    theBitmapData.y0 = 0;
        
    theBitmapData.collection_index = 0;
    theBitmapData.bitmap_index = 0;

	for (UInt16 count = 0; count < MAXIMUM_NUMBER_OF_SHADED_BITMAPS; count++)
        theBitmapData.shaded_hardware_addresses[count] = 0;
        
    theBitmapData.row_addresses[0] = 0;

//	
	Rect theBounds = { 0, 0, 256, 256 };
	LGWorld theGWorld(theBounds, 8, 0, theMacColorTableH);

	for (UInt16 bitmapCount = 0; bitmapCount < theBitmapHeader.bitmap_instance_count; bitmapCount++)
	{
		mGWorlds[bitmapCount]->GetBounds(theBounds);
		mGWorlds[bitmapCount]->CopyImage((GrafPtr) theGWorld.GetMacGWorld(), theBounds);
		
	    theBitmapData.width = theBounds.right - theBounds.left;
	    theBitmapData.height = theBounds.bottom - theBounds.top;
	    theBitmapData.bytes_per_row = theBitmapData.width;
	    theBitmapData.width_bits = log2(theBitmapData.width);
	    theBitmapData.height_bits = log2(theBitmapData.height);

		byteCount = sizeof(bitmap_data);
		ThrowIfOSErr_(theGraphicFile.PutBytes(&theBitmapData, byteCount));
		void* rowPtr = NULL;
		for (int rowCount = 1 /*already wrote first */; rowCount < theBitmapData.height; rowCount++)
		{
			byteCount = sizeof(void*);
			ThrowIfOSErr_(theGraphicFile.PutBytes(&rowPtr, byteCount));
		}

		PixMapHandle thePixMapH = ::GetGWorldPixMap(theGWorld.GetMacGWorld());
		Ptr pixmapAddr = ::GetPixBaseAddr(thePixMapH);
		UInt32 pixmapRowBytes = (*thePixMapH)->rowBytes & 0x3FFF;
		
		::LockPixels(thePixMapH);
		for (Uint16 rowCount = 0; rowCount < theBitmapData.height; rowCount++)
		{
			byteCount = theBitmapData.bytes_per_row;
			ThrowIfOSErr_(theGraphicFile.PutBytes(pixmapAddr, byteCount));

			pixmapAddr += pixmapRowBytes;
		}
		::UnlockPixels(thePixMapH);
	}

	::DisposeCTable(theMacColorTableH);

// Rewrite the headers
	theGraphicFile.SetMarker(0, streamFrom_Start);
	
	theFileHeader.mTagSize = currentOffset + SIZEOF_STRUCT_COLLECTION_HEADER;
	byteCount = sizeof(myth2dir_Header);
	ThrowIfOSErr_(theGraphicFile.PutBytes(&theFileHeader, byteCount));

    theBitmapHeader.data_size = currentOffset;
	byteCount = SIZEOF_STRUCT_COLLECTION_HEADER;
	ThrowIfOSErr_(theGraphicFile.PutBytes(&theBitmapHeader, byteCount));

// Close the data fork.
	theGraphicFile.CloseDataFork();


// ------------------	
// Write the reference file
	FSSpec theReferenceFSSpec;
	mFileStream->GetSpecifier(theReferenceFSSpec);
	LString::CopyPStr(theNameBase + "\p.core", theReferenceFSSpec.name, sizeof(StrFileName));
	
	LFileStream theCoreFile(theReferenceFSSpec);
	Try_
	{
		theCoreFile.CreateNewDataFile('M2SB', 'File');
	}
	Catch_( inErr )
	{
		if (inErr.GetErrorCode() == dupFNErr)
		{
			SignalPStr_(LStr255("\pDuplicate file, tell eli to add the damn dialog."));
			Throw_Err(inErr);
		}
		else
		{
			SignalPStr_(LStr255("\pFile Error: ") + inErr.GetErrorString());
			Throw_Err(inErr);
		}
	} EndCatch_
// Open the data fork.
	theCoreFile.OpenDataFork( fsRdWrPerm );

// Reuse the tag header
  	::BlockMoveData("eli made this core\0", theFileHeader.mName, 30);
  	theFileHeader.mTagGroup = COLLECTION_REFERENCE_GROUP_TAG;
  	theFileHeader.mTagID = theTagID;
  	theFileHeader.mTagOffset = sizeof(myth2dir_Header);
	theFileHeader.mTagSize = SIZEOF_STRUCT_COLLECTION_REFERENCE;

	collection_reference theCore;
	theCore.collection_tag = theTagID;
	theCore.number_of_permutations = 1;
	theCore.tint_fraction = 0;

	for (UInt16 permutationCount = 0; permutationCount < MAXIMUM_PERMUTATIONS_PER_COLLECTION_REFERENCE; permutationCount++)
		for (UInt16 hueCount = 0; hueCount < MAXIMUM_HUE_CHANGES_PER_COLLECTION; hueCount++)
		{
			theCore.colors[permutationCount][hueCount].red = 0;
			theCore.colors[permutationCount][hueCount].green = 0;
			theCore.colors[permutationCount][hueCount].blue = 0;
			theCore.colors[permutationCount][hueCount].flags = 0;
		}
	theCore.tint_color.red = 0;
	theCore.tint_color.green = 0;
	theCore.tint_color.blue = 0;
	theCore.tint_color.flags = 0;

	for (UInt16 unusedCount = 0; unusedCount < 18; unusedCount++)
    	theCore.unused[unusedCount] = 0;
	
	theCore.collection_index = 0;
	for (UInt16 colorCount = 0; colorCount < MAXIMUM_PERMUTATIONS_PER_COLLECTION_REFERENCE; colorCount++)
		theCore.color_table_indexes[colorCount] = 0;
	
// Write the tag header
	byteCount = sizeof(myth2dir_Header);
	ThrowIfOSErr_(theCoreFile.PutBytes(&theFileHeader, byteCount));

// Write the core
	byteCount = SIZEOF_STRUCT_COLLECTION_REFERENCE;
	ThrowIfOSErr_(theCoreFile.PutBytes(&theCore, byteCount));

// Close the data fork.
	theCoreFile.CloseDataFork();


// ------------------	
// Write the mode file
	LStdCheckBox * theCellCheckbox = (LStdCheckBox *) mWindow->FindPaneByID( kMainView );
	Assert_( theCellCheckbox != nil );
	Boolean includeCells = (theCellCheckbox->GetValue() > 0);

	FSSpec theModelFSSpec;
	mFileStream->GetSpecifier(theModelFSSpec);
	LString::CopyPStr(theNameBase + "\p.mode", theModelFSSpec.name, sizeof(StrFileName));
	
	LFileStream theModelFile(theModelFSSpec);
	Try_
	{
		theModelFile.CreateNewDataFile('M2SB', 'File');
	}
	Catch_( inErr )
	{
		if (inErr.GetErrorCode() == dupFNErr)
		{
			SignalPStr_(LStr255("\pDuplicate file, tell eli to add the damn dialog."));
			Throw_Err(inErr);
		}
		else
		{
			SignalPStr_(LStr255("\pFile Error: ") + inErr.GetErrorString());
			Throw_Err(inErr);
		}
	} EndCatch_
// Open the data fork.
	theModelFile.OpenDataFork( fsRdWrPerm );

// Reuse the tag header
  	::BlockMoveData("eli made this model\0", theFileHeader.mName, 30);
  	theFileHeader.mTagGroup = MODEL_GROUP_TAG;
  	theFileHeader.mTagID = theTagID;
  	theFileHeader.mTagOffset = sizeof(myth2dir_Header);
// We'll go and write over this when we know the size
	theFileHeader.mTagSize = 0;
	theFileHeader.mVersionM2 = 4;	// the 'mode' version got updated... whatever

	currentOffset = 0;
	mModel.flags = 0;
	mModel.geometry_tag = theTagID;
	mModel.geometry_index = 0;

	mModel.model_geometry_vertex_count = mGeometry.vertex_count;
	mModel.model_permutation_count = 1;

	mModel.model_geometry_vertex_offset = currentOffset;
	mModel.model_geometry_vertex_size = mModel.model_geometry_vertex_count * SIZEOF_STRUCT_GEOMETRY_INDEX_DATA;
	mModel.vertex_flags = NULL;
	currentOffset += mModel.model_geometry_vertex_size;

	mModel.model_permutations_offset = currentOffset;
	mModel.model_permutations_size = mModel.model_permutation_count * SIZEOF_STRUCT_MODEL_PERMUTATION_DATA;
	mModel.model_permutations = NULL;
	currentOffset += mModel.model_permutations_size;

	mModel.model_cell_offset = currentOffset;
	mModel.model_cell_size = mModel.model_cell_count * SIZEOF_STRUCT_MODEL_MESH_DATA;
	mModel.model_cells = NULL;
	currentOffset += mModel.model_cell_size;

	mModel.data_offset = SIZEOF_STRUCT_MODEL_DEFINITION;
	mModel.data_size = currentOffset;
	mModel.data = NULL;

// No weird references so we know the actual tag size. woohoo!
// Write the tag header
	theFileHeader.mTagSize = SIZEOF_STRUCT_MODEL_DEFINITION + mModel.data_size;
	byteCount = sizeof(myth2dir_Header);
	ThrowIfOSErr_(theModelFile.PutBytes(&theFileHeader, byteCount));

// Write the model header
	byteCount = SIZEOF_STRUCT_MODEL_DEFINITION;
	ThrowIfOSErr_(theModelFile.PutBytes(&mModel, byteCount));
	
// Write the vertices
	geometry_vertex_data theVertex;
	theVertex.flags = 0;
	for (UInt16 vertexCount = 0; vertexCount < mModel.model_geometry_vertex_count; vertexCount++)
	{
		byteCount = SIZEOF_STRUCT_GEOMETRY_INDEX_DATA;
		ThrowIfOSErr_(theModelFile.PutBytes(&theVertex, byteCount));
	}

// Write the permutation
	model_permutation_data thePermutation;
	thePermutation.collection_reference_permutation = 0;
	for (UInt16 materialCount = 0; materialCount < MAXIMUM_MATERIALS_PER_GEOMETRY; materialCount++)
		thePermutation.permutations[materialCount] = 0;
	for (UInt16 unusedCount = 0; unusedCount < 13; unusedCount++)
    	thePermutation.unused[unusedCount] = 0;
	thePermutation.collection_index = 0;
	thePermutation.color_table_index = 0;
	byteCount = SIZEOF_STRUCT_MODEL_PERMUTATION_DATA;
	ThrowIfOSErr_(theModelFile.PutBytes(&thePermutation, byteCount));

// Write the cells
	for (UInt16 cellCount = 0; cellCount < mModel.model_cell_count; cellCount++)
	{
		byteCount = SIZEOF_STRUCT_MODEL_MESH_DATA;
		ThrowIfOSErr_(theModelFile.PutBytes(&mCells[cellCount], byteCount));
	}

// Close the data fork.
	theModelFile.CloseDataFork();


// ------------------	
// Write the geom file
	FSSpec theGeometryFSSpec;
	mFileStream->GetSpecifier(theGeometryFSSpec);
	LString::CopyPStr(theNameBase + "\p.geom", theGeometryFSSpec.name, sizeof(StrFileName));
	
	LFileStream theGeometryFile(theGeometryFSSpec);
	Try_
	{
		theGeometryFile.CreateNewDataFile('M2SB', 'File');
	}
	Catch_( inErr )
	{
		if (inErr.GetErrorCode() == dupFNErr)
		{
			SignalPStr_(LStr255("\pDuplicate file, tell eli to add the damn dialog."));
			Throw_Err(inErr);
		}
		else
		{
			SignalPStr_(LStr255("\pFile Error: ") + inErr.GetErrorString());
			Throw_Err(inErr);
		}
	} EndCatch_
// Open the data fork.
	theGeometryFile.OpenDataFork( fsRdWrPerm );

// Reuse the tag header
  	::BlockMoveData("eli made this geometry\0", theFileHeader.mName, 30);
  	theFileHeader.mTagGroup = GEOMETRY_GROUP_TAG;
  	theFileHeader.mTagID = theTagID;
  	theFileHeader.mTagOffset = sizeof(myth2dir_Header);
// We'll go and write over this when we know the size
	theFileHeader.mTagSize = 0;
	theFileHeader.mVersionM2 = 1;

	currentOffset = 0;

	mGeometry.flags = 0;
	mGeometry.collection_reference_tag = theTagID;
	mGeometry.dependency_count = 0;
	mGeometry.radius_squared = 0;	// wth is this radius they're squaring?
	
// Clean up the centerpoint, remembering to swap yet another set of axis
	mGeometry.center.x = ((mGeometry.bounds.minimum_corner.x + mGeometry.bounds.maximum_corner.x) / 2) + (mAlignmentX * GEOMETRY_NORMAL_SCALE);
	mGeometry.center.y = ((mGeometry.bounds.minimum_corner.y + mGeometry.bounds.maximum_corner.y) / 2) + (mAlignmentZ * GEOMETRY_NORMAL_SCALE);
	mGeometry.center.z = mAlignmentY * GEOMETRY_NORMAL_SCALE;
	mGeometry.height = mGeometry.bounds.maximum_corner.z - mGeometry.bounds.minimum_corner.z;

	mGeometry.materials_offset = currentOffset;
	mGeometry.materials_size = mGeometry.material_count * SIZEOF_STRUCT_GEOMETRY_MATERIAL;
	mGeometry.materials = NULL;
	currentOffset += mGeometry.materials_size;
	
	mGeometry.vertices_offset = currentOffset;
	mGeometry.vertices_size = mGeometry.vertex_count * SIZEOF_STRUCT_GEOMETRY_VERTEX;
	mGeometry.vertices = NULL;
	currentOffset += mGeometry.vertices_size;
	
	mGeometry.surfaces_offset = currentOffset;
	mGeometry.surfaces_size = mGeometry.surface_count * SIZEOF_STRUCT_GEOMETRY_SURFACE;
	mGeometry.surfaces = NULL;
	currentOffset += mGeometry.surfaces_size;

	mGeometry.dependency_count = 0;
	for ( int iArray = 1; iArray <= mDependencyArrays->GetCount(); iArray++ )
	{
		LArray* dependencyArray = nil;
		if ( ( mDependencyArrays->FetchItemAt( iArray, &dependencyArray ) ) && ( dependencyArray ) )
		{
			mGeometry.dependency_count += dependencyArray->GetCount();
			mGeometry.dependency_count += 1;	// termination 0xFFFF
		}
	}
	mGeometry.dependency_list_offset = currentOffset;
	mGeometry.dependency_size = mGeometry.dependency_count * sizeof(short);
	mGeometry.dependency_list = NULL;
	currentOffset += mGeometry.dependency_size;
	
	mGeometry.data_offset = SIZEOF_STRUCT_GEOMETRY_DEFINITION_MYTH2;
	mGeometry.data_size = currentOffset;
	mGeometry.data = NULL;
	
/*
	// replaced by the bounding box in later builds
	for (UInt16 unusedCount = 0; unusedCount < 3; unusedCount++)
    	mGeometry.unused[unusedCount] = 0;

	mGeometry.collection_reference_index = 0;

	for (UInt16 dunnoCount = 0; dunnoCount < 16; dunnoCount++)
		mGeometry.myth2_dunno_what[dunnoCount];
*/
	for (UInt16 unusedCount = 0; unusedCount < 14; unusedCount++)
		mGeometry.hopefully_unused[unusedCount];
	
// Write the tag header
	theFileHeader.mTagSize = SIZEOF_STRUCT_GEOMETRY_DEFINITION_MYTH2 + mGeometry.data_size;
	byteCount = sizeof(myth2dir_Header);
	ThrowIfOSErr_(theGeometryFile.PutBytes(&theFileHeader, byteCount));

// Write the geometry header
	byteCount = SIZEOF_STRUCT_GEOMETRY_DEFINITION_MYTH2;
	ThrowIfOSErr_(theGeometryFile.PutBytes(&mGeometry, byteCount));

// Write the materials
	byteCount = mGeometry.materials_size;
	ThrowIfOSErr_(theGeometryFile.PutBytes(mMaterials, byteCount));

// Write the vertices
	byteCount = mGeometry.vertices_size;
	ThrowIfOSErr_(theGeometryFile.PutBytes(mVertices, byteCount));

// Calculate the whole dependeny list so we can clean
// up the surfaces with the correct indices
	short* dependencies = new short[mGeometry.dependency_count];
	short* offsets = new short[mDependencyArrays->GetCount() + 1]; // 1-based LArray
	UInt16 dependencyIndex = 0;
	for ( int iArray = 1; iArray <= mDependencyArrays->GetCount(); iArray++ )
	{
		LArray* dependencyArray = nil;
		if ( ( mDependencyArrays->FetchItemAt( iArray, &dependencyArray ) ) && ( dependencyArray ) )
		{
			offsets[iArray] = dependencyIndex;
			
			for ( int iDependency = 1; iDependency <= dependencyArray->GetCount(); iDependency++ )
			{
				short dependency = -1;
				if ( ( dependencyArray->FetchItemAt( iDependency, &dependency ) ) && ( dependency >= 0 ) )
				{
					dependencies[dependencyIndex] = dependency;
					dependencyIndex++;
				}
			}
			
			dependencies[dependencyIndex] = 0xFFFF;
			dependencyIndex++;
		}
	}
	
	for ( int iSurface = 0; iSurface < mGeometry.surface_count; iSurface++)
	{
		if ( mSurfaces[iSurface].dependency_table_index >= 0 )
		{
			mSurfaces[iSurface].dependency_table_index = offsets[mSurfaces[iSurface].dependency_table_index];
		}
	}
	
	delete [] offsets;
	
// Write the surfaces
	byteCount = mGeometry.surfaces_size;
	ThrowIfOSErr_(theGeometryFile.PutBytes(mSurfaces, byteCount));
	
// Write the dependencies
	byteCount = mGeometry.dependency_size;
	ThrowIfOSErr_(theGeometryFile.PutBytes(dependencies, byteCount));
	delete [] dependencies;

// Close the data fork.
	theGeometryFile.CloseDataFork();
	
	}
	Catch_( inErr )
	{
		SignalPStr_(LStr255("\pError saving Myth tags ") + inErr.GetErrorString());
	} EndCatch_
}


// ---------------------------------------------------------------------------------
//		 AdjustMythVertices
// ---------------------------------------------------------------------------------

void
C3dmfDocument::AdjustEditFields()
{
	mScaleXField->SetDescriptor(LStr255((mBoundingBox.max.x - mBoundingBox.min.x) * mMythScale, numberFormat));
	mScaleYField->SetDescriptor(LStr255((mBoundingBox.max.y - mBoundingBox.min.y) * mMythScale, numberFormat));
	mScaleZField->SetDescriptor(LStr255((mBoundingBox.max.z - mBoundingBox.min.z) * mMythScale, numberFormat));

	float fCenterX = ( (mBoundingBox.max.x - mBoundingBox.min.x) * mMythScale ) / 2.0;
	mAlignXField->SetDescriptor(LStr255(fCenterX + mAlignmentX, numberFormat));
	// Y axis (height) defaults to 0.0 rather than the true center
	mAlignYField->SetDescriptor(LStr255(mAlignmentY, numberFormat));
	float fCenterZ = ( (mBoundingBox.max.z - mBoundingBox.min.z) * mMythScale ) / 2.0;
	mAlignZField->SetDescriptor(LStr255(fCenterZ + mAlignmentZ, numberFormat));
}


// ---------------------------------------------------------------------------------
//		 AdjustMythVertices
// ---------------------------------------------------------------------------------

void
C3dmfDocument::AdjustMythVertices()
{
	// Set the transform
	::Q3Matrix4x4_Copy(&QD3DToMythMatrix, &mTransformMatrix);

	// Scale the points to Myth coordinates
	TQ3Matrix4x4 theScalingMatrix;
	::Q3Matrix4x4_SetScale(&theScalingMatrix,
		mMythScale * (float) GEOMETRY_NORMAL_SCALE,
		mMythScale * (float) GEOMETRY_NORMAL_SCALE,
		mMythScale * (float) GEOMETRY_NORMAL_SCALE);
		
	::Q3Matrix4x4_Multiply(&mTransformMatrix, &theScalingMatrix, &mTransformMatrix);
	
	// Use full world_rectangle3d here to try and avoid short overflows in the math
	world_point3d minimum_corner;
	world_point3d maximum_corner;
	minimum_corner.x = WORLD_ONE;	
	minimum_corner.y = WORLD_ONE;	
	minimum_corner.z = WORLD_ONE;	
	maximum_corner.x = 0;	
	maximum_corner.y = 0;	
	maximum_corner.z = 0;	

	for (UInt16 vertexIndex = 0; vertexIndex < mGeometry.vertex_count; vertexIndex++)
	{
		TQ3Point3D testPoint;
		::Q3Point3D_Transform(&mPoints[vertexIndex], &mTransformMatrix, &testPoint);
		
		mVertices[vertexIndex].p.x = testPoint.x;
		mVertices[vertexIndex].p.y = testPoint.y;
		mVertices[vertexIndex].p.z = testPoint.z;

		if (minimum_corner.x > mVertices[vertexIndex].p.x)
			minimum_corner.x = mVertices[vertexIndex].p.x;
		if (minimum_corner.y > mVertices[vertexIndex].p.y)
			minimum_corner.y = mVertices[vertexIndex].p.y;
		if (minimum_corner.z > mVertices[vertexIndex].p.z)
			minimum_corner.z = mVertices[vertexIndex].p.z;
		if (maximum_corner.x < mVertices[vertexIndex].p.x)
			maximum_corner.x = mVertices[vertexIndex].p.x;
		if (maximum_corner.y < mVertices[vertexIndex].p.y)
			maximum_corner.y = mVertices[vertexIndex].p.y;
		if (maximum_corner.z < mVertices[vertexIndex].p.z)
			maximum_corner.z = mVertices[vertexIndex].p.z;
	}
	
	int iAdjustX = 0;
	int iCellSize = (int) GEOMETRY_NORMAL_SCALE;
	if ( minimum_corner.x < 0 ) iAdjustX = -minimum_corner.x;
	if ( minimum_corner.x >= iCellSize ) iAdjustX = ( minimum_corner.x / iCellSize ) * -iCellSize;
	int iAdjustY = 0;
	if ( minimum_corner.y < 0 ) iAdjustY = -minimum_corner.y;
	if ( minimum_corner.y >= iCellSize ) iAdjustY = ( minimum_corner.y / iCellSize ) * -iCellSize;
	int iAdjustZ = 0;
	if ( minimum_corner.z < 0 ) iAdjustZ = -minimum_corner.z;
	if ( minimum_corner.z >= iCellSize ) iAdjustZ = ( minimum_corner.z / iCellSize ) * -iCellSize;
	
	if ( iAdjustX || iAdjustY || iAdjustZ )
	{
		for (UInt16 vertexIndex = 0; vertexIndex < mGeometry.vertex_count; vertexIndex++)
		{
			mVertices[vertexIndex].p.x += iAdjustX;
			mVertices[vertexIndex].p.y += iAdjustY;
			mVertices[vertexIndex].p.z += iAdjustZ;
		}
		
		mGeometry.bounds.minimum_corner.x = minimum_corner.x + iAdjustX;
		mGeometry.bounds.maximum_corner.x = maximum_corner.x + iAdjustX;
		mGeometry.bounds.minimum_corner.y = minimum_corner.y + iAdjustY;
		mGeometry.bounds.maximum_corner.y = maximum_corner.y + iAdjustY;
		mGeometry.bounds.minimum_corner.z = minimum_corner.z + iAdjustZ;
		mGeometry.bounds.maximum_corner.z = maximum_corner.z + iAdjustZ;
	
		mMeshOffsetX = iAdjustX;
		mMeshOffsetY = iAdjustY;
		mMeshOffsetZ = iAdjustZ;
	}
	
	for (UInt16 surfaceIndex = 0; surfaceIndex < mGeometry.surface_count; surfaceIndex++)
	{
		geometry_surface& currentSurface = mSurfaces[surfaceIndex];
		short vertex0 = currentSurface.corners[0].vertex_index;
		short vertex1 = currentSurface.corners[1].vertex_index;
		short vertex2 = currentSurface.corners[2].vertex_index;

		currentSurface.bounds.minimum_corner.x = mVertices[vertex0].p.x;
		if (currentSurface.bounds.minimum_corner.x > mVertices[vertex1].p.x)
			currentSurface.bounds.minimum_corner.x = mVertices[vertex1].p.x;
		if (currentSurface.bounds.minimum_corner.x > mVertices[vertex2].p.x)
			currentSurface.bounds.minimum_corner.x = mVertices[vertex2].p.x;
		currentSurface.bounds.minimum_corner.y = mVertices[vertex0].p.y;
		if (currentSurface.bounds.minimum_corner.y > mVertices[vertex1].p.y)
			currentSurface.bounds.minimum_corner.y = mVertices[vertex1].p.y;
		if (currentSurface.bounds.minimum_corner.y > mVertices[vertex2].p.y)
			currentSurface.bounds.minimum_corner.y = mVertices[vertex2].p.y;
		currentSurface.bounds.minimum_corner.z = mVertices[vertex0].p.z;
		if (currentSurface.bounds.minimum_corner.z > mVertices[vertex1].p.z)
			currentSurface.bounds.minimum_corner.z = mVertices[vertex1].p.z;
		if (currentSurface.bounds.minimum_corner.z > mVertices[vertex2].p.z)
			currentSurface.bounds.minimum_corner.z = mVertices[vertex2].p.z;
		currentSurface.bounds.maximum_corner.x = mVertices[vertex0].p.x;
		if (currentSurface.bounds.maximum_corner.x < mVertices[vertex1].p.x)
			currentSurface.bounds.maximum_corner.x = mVertices[vertex1].p.x;
		if (currentSurface.bounds.maximum_corner.x < mVertices[vertex2].p.x)
			currentSurface.bounds.maximum_corner.x = mVertices[vertex2].p.x;
		currentSurface.bounds.maximum_corner.y = mVertices[vertex0].p.y;
		if (currentSurface.bounds.maximum_corner.y < mVertices[vertex1].p.y)
			currentSurface.bounds.maximum_corner.y = mVertices[vertex1].p.y;
		if (currentSurface.bounds.maximum_corner.y < mVertices[vertex2].p.y)
			currentSurface.bounds.maximum_corner.y = mVertices[vertex2].p.y;
		currentSurface.bounds.maximum_corner.z = mVertices[vertex0].p.z;
		if (currentSurface.bounds.maximum_corner.z < mVertices[vertex1].p.z)
			currentSurface.bounds.maximum_corner.z = mVertices[vertex1].p.z;
		if (currentSurface.bounds.maximum_corner.z < mVertices[vertex2].p.z)
			currentSurface.bounds.maximum_corner.z = mVertices[vertex2].p.z;

		currentSurface.d = 0 -
			((currentSurface.normal.i * mVertices[vertex0].p.x) +
			(currentSurface.normal.j * mVertices[vertex0].p.y) +
			(currentSurface.normal.k * mVertices[vertex0].p.z));
	}
	
	int iCellsX = ceil(((float)(mGeometry.bounds.minimum_corner.x + mGeometry.bounds.maximum_corner.x) / iCellSize) + mAlignmentX);
	int iCellsY = ceil(((float)(mGeometry.bounds.minimum_corner.y + mGeometry.bounds.maximum_corner.y) / iCellSize) + mAlignmentZ);	// Damn coordinates
	if ( ( iCellsX > 0 ) && ( iCellsY > 0 ) && ( (iCellsX * iCellsY) <= MAXIMUM_CELLS_PER_MODEL ) )
	{
		// Only reset the mesh if the scale has changed
		if ( (iCellsX * iCellsY) != mModel.model_cell_count )
		{
			mModel.model_cell_count = iCellsX * iCellsY;
			int iCellIndex = 0;
		
			for ( int iY = 0; iY < iCellsY; iY++ )
			{
				for ( int iX = 0; iX < iCellsX; iX++ )
				{
					mCells[iCellIndex].x = iX;
					mCells[iCellIndex].y = iY;
					mCells[iCellIndex].flags = 0;
					mCells[iCellIndex].unused = 0;
				
					iCellIndex++;
				}
			}
		}
	}
	else
	{
		mModel.model_cell_count = 0;
	}

}


// ---------------------------------------------------------------------------------
//		 DoRevert
// ---------------------------------------------------------------------------------

void
C3dmfDocument::DoRevert()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Read the entire file contents and close the file.
	Handle	theTextH = mFileStream->ReadDataFork();
	mFileStream->CloseDataFork();
	
	// Put the contents in the text view
	// and clear the dirty flag.
	mTextView->SetTextHandle( theTextH );
	mTextView->SetDirty( false );
	
	// Dispose of the text.
	::DisposeHandle( theTextH );

	// Refresh the text view.
	mTextView->Refresh();
*/
}


// ---------------------------------------------------------------------------------
//		 DoPrint
// ---------------------------------------------------------------------------------

void
C3dmfDocument::DoPrint()
{
/*
	// Create the printout.
	LPrintout	*thePrintout = LPrintout::CreatePrintout( rPPob_TextPrintout );
	Assert_( thePrintout != nil );
	
	// Set the print record.
	
	if (mPrintRecordH)
	{
		thePrintout->SetPrintRecord( mPrintRecordH );
	}
	
	// Get the text placeholder.
	LPlaceHolder	*thePlaceholder;
	thePlaceholder = (LPlaceHolder*) thePrintout->FindPaneByID( kTextPlaceholder );
	Assert_( thePlaceholder != nil );
	
	// Install the text view in the placeholder.
	thePlaceholder->InstallOccupant( mTextView, atNone );
	
	// Set the frame size.
	SetPrintFrameSize();
	
	// Print.
	thePrintout->DoPrintJob();
	
	// Delete the printout. The text view is returned
	// to the window when the placeholder is destroyed.
	delete thePrintout;
*/
}


// ---------------------------------------------------------------------------------
//		 SetPrintFrameSize
// ---------------------------------------------------------------------------------

void
C3dmfDocument::SetPrintFrameSize()
{
/*
	// Get the frame size.
	SDimension16	theFrameSize;
	mTextView->GetFrameSize( theFrameSize );
	
	// Get the text edit record handle.
	TEHandle	theTextEditH = mTextView->GetMacTEH();
	
	// Calculate the number of lines per page.
	Int16	theLinesPerPage;
	theLinesPerPage = theFrameSize.height / (**theTextEditH).lineHeight;

	// Resize the frame to an integral number of lines.
	mTextView->ResizeFrameTo( theFrameSize.width,
		(**theTextEditH).lineHeight * theLinesPerPage, false );

*/
}


void
C3dmfDocument::ListenToMessage(
	MessageT inMessage,
	void * ioParam)
{
	switch (inMessage)
	{
		case message_TextChanged:
			switch(*((PaneIDT *) ioParam))
			{
				case scaleXID:
				{
					LStr255 theValueString;
					mScaleXField->GetDescriptor(theValueString);
					
					float theValue = LString::StringToLongDouble(theValueString);
					mMythScale = theValue / (mBoundingBox.max.x - mBoundingBox.min.x);
					AdjustEditFields();
					AdjustMythVertices();
					if ( mGeomView )
					{
						mGeomView->Refresh();
					}
				}	break;
				
				case scaleYID:
				{
					LStr255 theValueString;
					mScaleYField->GetDescriptor(theValueString);
					
					float theValue = LString::StringToLongDouble(theValueString);
					mMythScale = theValue / (mBoundingBox.max.y - mBoundingBox.min.y);
					AdjustEditFields();
					AdjustMythVertices();
					if ( mGeomView )
					{
						mGeomView->Refresh();
					}
				}	break;
				
				case scaleZID:
				{
					LStr255 theValueString;
					mScaleZField->GetDescriptor(theValueString);
					
					float theValue = LString::StringToLongDouble(theValueString);
					mMythScale = theValue / (mBoundingBox.max.z - mBoundingBox.min.z);
					AdjustEditFields();
					AdjustMythVertices();
					if ( mGeomView )
					{
						mGeomView->Refresh();
					}
				}	break;
				
				case alignXID:
				{
					LStr255 theValueString;
					mAlignXField->GetDescriptor(theValueString);
					
					float fCenterX = LString::StringToLongDouble(theValueString);
					
					float fMaxX = (mBoundingBox.max.x - mBoundingBox.min.x) * mMythScale;
					if ( ( fCenterX >= 0.0 ) && ( fCenterX < fMaxX ) )
					{
						mAlignmentX = fCenterX - ( fMaxX / 2.0 );
						while ( mAlignmentX < 0.0 )
						{
							mAlignmentX += 1.0;
						}
						while ( mAlignmentX >= 1.0 )
						{
							mAlignmentX -= 1.0;
						}
						
						AdjustEditFields();
						AdjustMythVertices();
						if ( mGeomView )
						{
							mGeomView->Refresh();
						}
					}
					else	// Invalid value, reset the field
					{
						SysBeep( 1 );
						AdjustEditFields();
					}
				}	break;
				
				case alignYID:
				{
					LStr255 theValueString;
					mAlignYField->GetDescriptor(theValueString);
					
					float fCenterY = LString::StringToLongDouble(theValueString);
					
					float fMaxY = (mBoundingBox.max.y - mBoundingBox.min.y) * mMythScale;
					if ( ( fCenterY >= 0.0 ) && ( fCenterY < fMaxY ) )
					{
						// Y can vary across the entire height, so don't normalize
						mAlignmentY = fCenterY;
						
						AdjustEditFields();
						AdjustMythVertices();
						if ( mGeomView )
						{
							mGeomView->Refresh();
						}
					}
					else	// Invalid value, reset the field
					{
						SysBeep( 1 );
						AdjustEditFields();
					}
				}	break;
				
				case alignZID:
				{
					LStr255 theValueString;
					mAlignZField->GetDescriptor(theValueString);
					
					float fCenterZ = LString::StringToLongDouble(theValueString);
					
					float fMaxZ = (mBoundingBox.max.z - mBoundingBox.min.z) * mMythScale;
					if ( ( fCenterZ >= 0.0 ) && ( fCenterZ < fMaxZ ) )
					{
						mAlignmentZ = fCenterZ - ( fMaxZ / 2.0 );
						while ( mAlignmentZ < 0.0 )
						{
							mAlignmentZ += 1.0;
						}
						while ( mAlignmentZ >= 1.0 )
						{
							mAlignmentZ -= 1.0;
						}
						
						AdjustEditFields();
						AdjustMythVertices();
						if ( mGeomView )
						{
							mGeomView->Refresh();
						}
					}
					else	// Invalid value, reset the field
					{
						SysBeep( 1 );
						AdjustEditFields();
					}
				}	break;
			}
			break;	// case message_TextChanged:
			
		case message_DoExport:
			this->DoExport();
			break;
	}
}
