// =================================================================================
//	CgeomDocument.cp				1997 eli curtz. All rights reserved.
// =================================================================================

#include <LFile.h>
#include <LPlaceHolder.h>
#include <LPrintout.h>
#include <LString.h>
#include <LWindow.h>
#include <PP_Messages.h>
#include <UMemoryMgr.h>
#include <UWindows.h>
#include <UStandardDialogs.h>

#include <QD3DGroup.h>
#include <QD3DGeometry.h>
#include <QD3DStorage.h>
#include <QD3DShader.h>
#include <QD3DView.h>
#include <QD3DLight.h>
#include <QD3DIO.h>
#include <QD3DMath.h>
#include <QuickTimeComponents.h>

#include "Constants.h"
#include "CgeomView.h"

#include "CgeomDocument.h"

#define _FLIP_X_AXIS true

#ifdef _FLIP_X_AXIS
const TQ3Matrix4x4 MythToQD3DMatrix = {{{-1.0,  0.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  1.0,  0.0 }, \
										{ 0.0,  1.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  0.0,  1.0 } }};
#else
const TQ3Matrix4x4 MythToQD3DMatrix = {{{ 1.0,  0.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  1.0,  0.0 }, \
										{ 0.0,  1.0,  0.0,  0.0 }, \
										{ 0.0,  0.0,  0.0,  1.0 } }};
#endif

const TQ3Point3D zeroPoint = { 0.0, 0.0, 0.0 };

const PaneIDT	exportID 	= 'ExpB';

const MessageT message_DoExport = 'Xprt';

// ---------------------------------------------------------------------------------
//		 CgeomDocument(LCommander*, FSSpec*)
// ---------------------------------------------------------------------------------


CgeomDocument::CgeomDocument(
	LCommander	*inSuper,
	FSSpec		*inFileSpec )
		: LSingleDoc( inSuper ), LListener()
{
	mTextureGWorlds = new LArray(sizeof(LGWorld*));
	
	// Create window for our document.
	mWindow = LWindow::CreateWindow( rPPob_Window, this );
	Assert_( mWindow != nil );
	
	// Find the active panes
	mGeomView = (CgeomView *) mWindow->FindPaneByID( kMainView );
	Assert_( mGeomView != nil );
	
	LButton* theButton = (LButton*) mWindow->FindPaneByID( exportID ); 
	Assert_( theButton != nil );
	theButton->AddListener(this);
	
	if ( inFileSpec == nil ) {

		// Set the name of new window.
		NameNewDoc();
		
	} else {
	
		// Display contents of file in window.
		OpenFile( *inFileSpec );
	
	}
	
	// Make the window visible.
	mWindow->Show();
}

CgeomDocument::~CgeomDocument()
{
	// Kill the GWorlds which may be around from texturing
	for (Int32 textureCount = mTextureGWorlds->GetCount(); textureCount > 0; textureCount--) {
		LGWorld * theTexture = nil;
		if (mTextureGWorlds->FetchItemAt(textureCount, &theTexture)) {
			// delete theTexture;	// QD3D is deleteing the GWorld, live with the struct leak for now
		}
	}
	
	delete mTextureGWorlds;
}


// ---------------------------------------------------------------------------------
//		 NameNewDoc
// ---------------------------------------------------------------------------------

void
CgeomDocument::NameNewDoc()
{
	// Setup the window title. Start with the default title.
	LStr255	theTitle( "\pUntitled" );

	// Find the first available title. We could also check the window
	// pane id if we wanted to make sure we didn't collide with other
	// window types.
	Int32	theNumber = 0;
	while ( UWindows::FindNamedWindow( theTitle ) != nil ) {

		// An existing window has the current name
		// Increment counter and try again.
		++theNumber;
		theTitle = "\pUntitled ";
		theTitle += (LStr255) theNumber;

	}		
	
	// Finally, set window title.
	mWindow->SetDescriptor( theTitle );
}


// ---------------------------------------------------------------------------------
//		 OpenFile
// ---------------------------------------------------------------------------------

void
CgeomDocument::OpenFile(
	FSSpec	&inFileSpec )
{
	Try_ {

		// Create a new file object.
		mFileStream = new LFileStream( inFileSpec );
		mFile = mFileStream;	// In case the superclass is using it
		
		// Open the data fork.
		mFileStream->OpenDataFork( fsRdPerm );
		
		// Set the window title to the name of the file and
		// flag that the document has an associated file.
		mWindow->SetDescriptor( inFileSpec.name );
		mIsSpecified = true;

		// Set the transform matrix
		TQ3Matrix4x4 theTransformMatrix;
		::Q3Matrix4x4_Copy(&MythToQD3DMatrix, &theTransformMatrix);
		// Scale the points to QD3D coordinates
		TQ3Matrix4x4 theScalingMatrix;
		float theScale = 1.0 / (float) GEOMETRY_NORMAL_SCALE;
		::Q3Matrix4x4_SetScale(&theScalingMatrix,
			theScale,
			theScale,
			theScale);
			
		::Q3Matrix4x4_Multiply(&theTransformMatrix, &theScalingMatrix, &theTransformMatrix);	

		// Create the QD3D group.
		TQ3GroupObject qd3DNewDisplayGroup = ::Q3OrderedDisplayGroup_New();
		ThrowIfNULL_(qd3DNewDisplayGroup);

		// Skip the file header and read the geometry header
		myth2dir_Header theFileHeader;
		Int32 byteCount = sizeof(myth2dir_Header);
		mFileStream->GetBytes(&theFileHeader, byteCount);
		Assert_(theFileHeader.mTagGroup == GEOMETRY_GROUP_TAG);
		
		byteCount = SIZEOF_STRUCT_GEOMETRY_DEFINITION_MYTH2;
		mFileStream->GetBytes(&mGeometry, byteCount);
		mHeaderOffset = sizeof(myth2dir_Header) + mGeometry.data_offset;
	
		// Read the materials
		mFileStream->SetMarker(mHeaderOffset + mGeometry.materials_offset, streamFrom_Start);
		geometry_material* theMaterials = new geometry_material[mGeometry.material_count];
		ThrowIfNULL_(theMaterials);
		byteCount = mGeometry.materials_size;
		mFileStream->GetBytes(theMaterials, byteCount);
		TQ3TriMeshData* theQD3DTriMeshes = new TQ3TriMeshData[mGeometry.material_count];
		ThrowIfNULL_(theQD3DTriMeshes);

		for (Int32 materialCount = 0; materialCount < mGeometry.material_count; materialCount++) {
			theQD3DTriMeshes[materialCount].triMeshAttributeSet = NULL;	// Add the texture here

			theQD3DTriMeshes[materialCount].numTriangles = 0;
			// Reserve memory for all the surfaces, they won't be saved
			theQD3DTriMeshes[materialCount].triangles = new TQ3TriMeshTriangleData[mGeometry.surface_count +1];
			ThrowIfNULL_(theQD3DTriMeshes[materialCount].triangles);
			theQD3DTriMeshes[materialCount].numTriangleAttributeTypes = 2;
			theQD3DTriMeshes[materialCount].triangleAttributeTypes = new TQ3TriMeshAttributeData[2];
			ThrowIfNULL_(theQD3DTriMeshes[materialCount].triangleAttributeTypes);
			theQD3DTriMeshes[materialCount].triangleAttributeTypes[0].attributeType = kQ3AttributeTypeNormal;
			theQD3DTriMeshes[materialCount].triangleAttributeTypes[0].data = new TQ3Vector3D[mGeometry.surface_count +1];
			ThrowIfNULL_(theQD3DTriMeshes[materialCount].triangleAttributeTypes[0].data);
			theQD3DTriMeshes[materialCount].triangleAttributeTypes[0].attributeUseArray = NULL;
			theQD3DTriMeshes[materialCount].triangleAttributeTypes[1].attributeType = kQ3AttributeTypeDiffuseColor;
			theQD3DTriMeshes[materialCount].triangleAttributeTypes[1].data = new TQ3ColorRGB[mGeometry.surface_count +1];
			ThrowIfNULL_(theQD3DTriMeshes[materialCount].triangleAttributeTypes[1].data);
			theQD3DTriMeshes[materialCount].triangleAttributeTypes[1].attributeUseArray = new char[mGeometry.surface_count +1];
			
			theQD3DTriMeshes[materialCount].numEdges = 0;
			theQD3DTriMeshes[materialCount].edges = NULL;
			theQD3DTriMeshes[materialCount].numEdgeAttributeTypes = 0;
			theQD3DTriMeshes[materialCount].edgeAttributeTypes = NULL;

			theQD3DTriMeshes[materialCount].numPoints = 0;
			theQD3DTriMeshes[materialCount].points = NULL;

			theQD3DTriMeshes[materialCount].numVertexAttributeTypes = 1;
			theQD3DTriMeshes[materialCount].vertexAttributeTypes = new TQ3TriMeshAttributeData;
			ThrowIfNULL_(theQD3DTriMeshes[materialCount].vertexAttributeTypes);

			::Q3BoundingBox_Set(&theQD3DTriMeshes[materialCount].bBox, &zeroPoint, &zeroPoint, kQ3True);
		}
		
		delete [] theMaterials;

		// Read the vertices
		mFileStream->SetMarker(mHeaderOffset + mGeometry.vertices_offset, streamFrom_Start);
		geometry_vertex* theVertices = new geometry_vertex[mGeometry.vertex_count];
		ThrowIfNULL_(theVertices);
		byteCount = mGeometry.vertices_size;
		mFileStream->GetBytes(theVertices, byteCount);
		TQ3Point3D* theQD3DVertices = new TQ3Point3D[mGeometry.vertex_count];
		ThrowIfNULL_(theQD3DVertices);

		for (Int32 vertexCount = 0; vertexCount < mGeometry.vertex_count; vertexCount++) {
			::Q3Point3D_Set(&theQD3DVertices[vertexCount],
							theVertices[vertexCount].p.x,
							theVertices[vertexCount].p.y,
							theVertices[vertexCount].p.z);
			
			::Q3Point3D_Transform(&theQD3DVertices[vertexCount], &theTransformMatrix, &theQD3DVertices[vertexCount]);
		}
		
		// Add vertices to the TriMeshes
		for (Int32 materialCount = 0; materialCount < mGeometry.material_count; materialCount++) {
			theQD3DTriMeshes[materialCount].numPoints = mGeometry.vertex_count;
			theQD3DTriMeshes[materialCount].points = theQD3DVertices;

			theQD3DTriMeshes[materialCount].vertexAttributeTypes[0].attributeType = kQ3AttributeTypeSurfaceUV;
			theQD3DTriMeshes[materialCount].vertexAttributeTypes[0].data = new TQ3Param2D[mGeometry.vertex_count];
			ThrowIfNULL_(theQD3DTriMeshes[materialCount].vertexAttributeTypes[0].data);
			theQD3DTriMeshes[materialCount].vertexAttributeTypes[0].attributeUseArray = NULL;
		}
		
		delete [] theVertices;

		// Read the surfaces
		mFileStream->SetMarker(mHeaderOffset + mGeometry.surfaces_offset, streamFrom_Start);
		geometry_surface* theSurfaces = new geometry_surface[mGeometry.surface_count];;
		ThrowIfNULL_(theSurfaces);
		byteCount = mGeometry.surfaces_size;
		mFileStream->GetBytes(theSurfaces, byteCount);
		for (Int32 surfaceCount = 0; surfaceCount < mGeometry.surface_count; surfaceCount++)
		{
			Assert_( theSurfaces[surfaceCount].material_index >= 0 );
			Assert_( theSurfaces[surfaceCount].material_index < mGeometry.material_count );
			TQ3TriMeshData& currentMesh = theQD3DTriMeshes[theSurfaces[surfaceCount].material_index];
			for (Int32 vertexCount = 0; vertexCount < 3; vertexCount++)
			{
				Int32 vertexIndex = theSurfaces[surfaceCount].corners[vertexCount].vertex_index;
				Assert_( vertexIndex < mGeometry.vertex_count );
				// Switch the vertex order to correct handedness
				currentMesh.triangles[currentMesh.numTriangles].pointIndices[2 - vertexCount] = vertexIndex;
				
				TQ3Vector3D* currentNormalPtr = &((TQ3Vector3D *) currentMesh.triangleAttributeTypes[0].data)[currentMesh.numTriangles];			
				::Q3Vector3D_Set(currentNormalPtr,
								theSurfaces[surfaceCount].normal.i,
								theSurfaces[surfaceCount].normal.j,
								theSurfaces[surfaceCount].normal.k);
				::Q3Vector3D_Transform(currentNormalPtr, &theTransformMatrix, currentNormalPtr);

				float scratchU = theSurfaces[surfaceCount].corners[vertexCount].interpolants[0];
				float scratchV = theSurfaces[surfaceCount].corners[vertexCount].interpolants[1];
				((TQ3Param2D *) currentMesh.vertexAttributeTypes[0].data)[vertexIndex].v = 1.0 - (scratchU / 0xFFFF);
				((TQ3Param2D *) currentMesh.vertexAttributeTypes[0].data)[vertexIndex].u = scratchV / 0xFFFF;
			}
			
			TQ3ColorRGB* currentColorPtr = &((TQ3ColorRGB *) currentMesh.triangleAttributeTypes[1].data)[currentMesh.numTriangles];			
			::Q3ColorRGB_Set( currentColorPtr, 0.2 + (0.1 * theSurfaces[surfaceCount].material_index), 0.0, 0.0 );

			currentMesh.numTriangles++;
		}
		// There seems to be some sort of off by one error, so we'll hack it
		for (Int32 materialCount = 0; materialCount < mGeometry.material_count; materialCount++)
		{
			TQ3TriMeshData& currentMesh = theQD3DTriMeshes[materialCount];
			for (Int32 vertexCount = 0; vertexCount < 3; vertexCount++)
			{
				currentMesh.triangles[currentMesh.numTriangles].pointIndices[vertexCount] = 0;
			}
			TQ3ColorRGB* currentColorPtr = &((TQ3ColorRGB *) currentMesh.triangleAttributeTypes[1].data)[currentMesh.numTriangles];			
			::Q3ColorRGB_Set( currentColorPtr, 1.0, 0.0, 0.0 );
			
			currentMesh.numTriangles++;
		}
		
		delete [] theSurfaces;

		// Put together the group
		for (Int32 materialCount = 0; materialCount < mGeometry.material_count; materialCount++) {
			TQ3GeometryObject theNewTriMesh = ::Q3TriMesh_New(&theQD3DTriMeshes[materialCount]);
			ThrowIfNULL_(theNewTriMesh);

			::Q3Group_AddObject(qd3DNewDisplayGroup, theNewTriMesh);
			::Q3Object_Dispose(theNewTriMesh);
		}

		mGeomView->SetGroup(qd3DNewDisplayGroup);
		
	} Catch_( inErr ) {

		// Cleanup and rethrow the error.
		delete this;
		Throw_Err( inErr );
	
	} EndCatch_
}


// ---------------------------------------------------------------------------------
//		 FindRelatedFile
// ---------------------------------------------------------------------------------
// Most of this code was ripped out of DropShell
// I Believe that code reuse is the sincerest form of slack.
// ALL HAIL SLACK!

LFileStream*
CgeomDocument::FindRelatedFile(
	FSSpec& inFileSpec,
	FourCharCode inGroup,
	FourCharCode inSubGroup )
{
	LFileStream* theFilePtr = nil;

	OSErr		err = noErr;
	short		index, oldIndex, localIndex;
	FSSpec		localFSSpec, curFSSpec;
	CInfoPBRec	cipb;
	Str255		fName, vFName;
	long		dirID, origDirID;
	Boolean		foundPosition;
 
	// We want the group folder inside the grandparent of our file
 	// parent folder FSSpec
 	ThrowIfOSErr_(::FSMakeFSSpec(inFileSpec.vRefNum, inFileSpec.parID, nil, &curFSSpec));
 	// group folder FSSpec
 	LStr255 groupName(inGroup);
 	ThrowIfOSErr_(::FSMakeFSSpec(curFSSpec.vRefNum, curFSSpec.parID, groupName, &localFSSpec));
 	
	//	get the dirID for group folder, not it's parent!
	BlockMoveData(localFSSpec.name, fName, 32);

	cipb.hFileInfo.ioCompletion	= 0L;
	cipb.hFileInfo.ioNamePtr	= fName;
	cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
	cipb.hFileInfo.ioFDirIndex	= 0;	// use the dir & vRefNum;
	cipb.hFileInfo.ioDirID		= localFSSpec.parID;
	ThrowIfOSErr_(PBGetCatInfoSync(&cipb));
	
	origDirID = cipb.dirInfo.ioDrDirID; // copy the sucker
	index = 1;
			
	// index through all contents of this folder
	Boolean foundItem = false;
	while ((err == noErr) && !foundItem) {
		dirID = origDirID;
		localIndex = index;
		fName [0] = 0;
		cipb.hFileInfo.ioCompletion	= 0L;
		cipb.hFileInfo.ioNamePtr	= fName;
		cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
		cipb.hFileInfo.ioFDirIndex	= localIndex;	// use a real index
		cipb.hFileInfo.ioDirID		= dirID;
		err = PBGetCatInfoSync(&cipb);

		if (!err) {
			BlockMoveData(fName, curFSSpec.name, 32);
			curFSSpec.vRefNum	= cipb.hFileInfo.ioVRefNum;
			curFSSpec.parID		= dirID;
		
			// Check to see if this entry is a folder.
			if (cipb.hFileInfo.ioFlAttrib & ioDirMask) {
				SignalPStr_("\pFolder found inside data folder!");
			}
			// Is it the file we're looking for?
			else {
				theFilePtr = new LFileStream(curFSSpec);
				theFilePtr->OpenDataFork(fsRdPerm);
				dir_Header theFileHeader;
				Int32 dataSize = sizeof(dir_Header);
				theFilePtr->GetBytes(&theFileHeader, dataSize);
				theFilePtr->CloseDataFork();
				
				if (theFileHeader.subGroup == inSubGroup) {
					foundItem = true;
				}
				else {
					delete theFilePtr;
					theFilePtr = nil;
				}
			}
			
			/*	If we've had an error, get out! */
			if (err)	break;

			// dirID = origDirID;	
			localIndex = index;	

			/*	
				Now take into account new files being created
				in the current directory & messing up our index.
				See Dev.CD Vol. XI:Tools & Apps (Moof!):Misc Utilities:
				Disinfectant & Source 2.5.1:Sample:Notes:Scan Alg	
			*/
			vFName [0] = 0;
			cipb.hFileInfo.ioCompletion	= 0L;
			cipb.hFileInfo.ioNamePtr	= vFName;
			cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
			cipb.hFileInfo.ioFDirIndex	= localIndex;	// use a real index
			cipb.hFileInfo.ioDirID		= dirID;
			err = PBGetCatInfoSync(&cipb);
			oldIndex = index;
			if (!err) {
				/*	If they're equal - same place, go to next */
				if (EqualString (vFName, fName, false, false))
					index++;
			}
			
			/*	If we didn't advance, then perhaps a file was created or deleted */
			if (oldIndex == index) {
				oldIndex		= index;	/* save off the old */
				index			= 0;		/* and start at the beginning */
				err				= noErr;
				vFName [0]		= 0;
				foundPosition	= false;
				
				while (!foundPosition) {
					index++;
					vFName [0] = 0;
					cipb.hFileInfo.ioCompletion	= 0L;
					cipb.hFileInfo.ioNamePtr	= vFName;
					cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
					cipb.hFileInfo.ioFDirIndex	= index;	/* now use a real index */
					cipb.hFileInfo.ioDirID		= dirID;
					err = PBGetCatInfoSync(&cipb);
					
					if (err == fnfErr) {  // we've just been deleted
						index = oldIndex;
						foundPosition = true;
						err = noErr;	// have to remember to reset this!
					}
					
				/*	found same file & same index position */
				/*	so try the next item */
					if ((!foundPosition) && EqualString(fName, vFName, false, false)) {
						index++;
						foundPosition = true;
					}
				}
			}
		}
	}
	
	ThrowIfOSErr_(err);
	return theFilePtr;
}


// ---------------------------------------------------------------------------
//	 AddTexture									 [private]
// ---------------------------------------------------------------------------

void
CgeomDocument::AddTexture(TQ3GeometryObject& ioObject, LFileStream* inFile, UInt16 inTextureIndex)
{
	inFile->OpenDataFork(fsRdPerm);
	
	// Skip the file header and read the collection header
	dir_Header theFileHeader;
	Int32 byteCount = sizeof(dir_Header);
	inFile->GetBytes(&theFileHeader, byteCount);
	Assert_(theFileHeader.group == '.256');
	
	collection_header theCollectionHeader;
	byteCount = sizeof(collection_header);
	inFile->GetBytes(&theCollectionHeader, byteCount);
	Int32 textureHeaderOffset = inFile->GetMarker();

	// Read a color table
	Assert_(theCollectionHeader.color_table_count >= 1);
	if(theCollectionHeader.color_table_count > 1)
		SignalPStr_("\pMore than one color table in this file.");
	Ptr theColorData = ::NewPtr(theCollectionHeader.color_tables_size);
	ThrowIfNil_(theColorData);
	inFile->SetMarker(textureHeaderOffset + theCollectionHeader.color_tables_offset, streamFrom_Start);
	byteCount = theCollectionHeader.color_tables_size;
	inFile->GetBytes(theColorData, byteCount);
	
	CTabHandle theColorTable = ::GetCTable(40);				// default 8-bit grey
	::HLock((Handle) theColorTable);
	Ptr colorSourcePtr = theColorData + 32;		// skip header
	for (Int16 colorCount = 0; colorCount <= (*theColorTable)->ctSize; colorCount++) {
		(*theColorTable)->ctTable[colorCount].rgb = *((RGBColor*) colorSourcePtr);
		colorSourcePtr += 8;
	}
	::HUnlock((Handle) theColorTable);
	
	::CTabChanged(theColorTable);
	::DisposePtr(theColorData);
	
	// Read in the bitmap data header
	Int32 bitmapRefMarker = textureHeaderOffset + theCollectionHeader.bitmap_references_offset;
	bitmapRefMarker += inTextureIndex * sizeof(bitmap_reference);
	
	bitmap_reference theBitmapRef;
	inFile->SetMarker(bitmapRefMarker, streamFrom_Start);
	byteCount = sizeof(bitmap_reference);
	inFile->GetBytes(&theBitmapRef, byteCount);

	// Read the bitmap data proper
	inFile->SetMarker(textureHeaderOffset + theBitmapRef.offset, streamFrom_Start);

	Ptr theBitmapData = ::NewPtr(theBitmapRef.size);
	ThrowIfNil_(theBitmapData);
	byteCount = theBitmapRef.size;
	inFile->GetBytes(theBitmapData, byteCount);
		
	bitmapData_Header *bitmapHeaderPtr = (bitmapData_Header*) theBitmapData;
	Ptr bitmapDataPtr = theBitmapData + sizeof(bitmapData_Header);
	if (theCollectionHeader.user_data_flags & _collection_is_color_map_flag)
		bitmapDataPtr += 4;
	else
		bitmapDataPtr += bitmapHeaderPtr->height * sizeof(Ptr);	// jump over the row ptrs (?)
	
	// Build the new GWorld
	Rect bitBounds = { 0, 0, 0, 0 };
	bitBounds.bottom = bitmapHeaderPtr->height;
	bitBounds.right = bitmapHeaderPtr->width;
	
	LGWorld* theNewGWorld = nil;

	if (bitmapHeaderPtr->bitDepth > 8)
		theNewGWorld = new LGWorld(bitBounds, bitmapHeaderPtr->bitDepth);
	else
		theNewGWorld = new LGWorld(bitBounds, 8, 0, theColorTable);
	Assert_(theNewGWorld != nil);
	Assert_(bitmapHeaderPtr->bitDepth >= 8);
	UInt8 bytesPerColor = bitmapHeaderPtr->bitDepth / 8;
	
	// Write data into the new GWorld		
	if (theNewGWorld->BeginDrawing()) {
		PixMapHandle thePixH = ::GetGWorldPixMap(theNewGWorld->GetMacGWorld());
		UInt32 realPixRowBytes = (*thePixH)->rowBytes & 0x3FFF;
		Ptr realPixPtr = ::GetPixBaseAddr(thePixH);

		if (bitmapHeaderPtr->rowBytes == -1) {	// Encoded sprite
			for (UInt16 rowCount = 1; rowCount <= bitmapHeaderPtr->height; rowCount++) {
				UInt16 *encodingHeader = (UInt16 *) bitmapDataPtr;

				bitmapDataPtr += (2 * (encodingHeader[0] + 1) * sizeof(UInt16));
				for (UInt16 runCount = 1; runCount <= encodingHeader[0]; runCount++) {
					UInt16 columnStartLoc = encodingHeader[runCount * 2];
					UInt16 columnStopLoc = encodingHeader[runCount * 2 + 1];
					UInt16 byteCount = (columnStopLoc - columnStartLoc) * bytesPerColor;
					Ptr rowPixPtr = realPixPtr + (columnStartLoc * bytesPerColor);
					::BlockMoveData(bitmapDataPtr, rowPixPtr, byteCount);
					bitmapDataPtr += byteCount;
				}

				if ((Int32) bitmapDataPtr & 0x01)	// Check for uneven row
					bitmapDataPtr++;
				realPixPtr += realPixRowBytes;
			}
		}
		else {	// Normal byte runs
			for (UInt16 rowCount = 1; rowCount <= bitmapHeaderPtr->height; rowCount++) {
				::BlockMoveData(bitmapDataPtr, realPixPtr, bitmapHeaderPtr->rowBytes);
				bitmapDataPtr += bitmapHeaderPtr->rowBytes;
				realPixPtr += realPixRowBytes;
			}
		}
	}
	theNewGWorld->EndDrawing();
	::DisposePtr(theBitmapData);
	
	// Force the GWorld to 16 bit for QD3D
	GWorldPtr macGWorld = theNewGWorld->GetMacGWorld();	
	::UpdateGWorld(&macGWorld, 16, &bitBounds, nil, nil, clipPix);

	PixMapHandle theGWorldPixMap = ::GetGWorldPixMap(macGWorld);
	::LockPixels(theGWorldPixMap);
	unsigned long pixMapWidth = (*theGWorldPixMap)->bounds.right - (*theGWorldPixMap)->bounds.left;
	unsigned long pixMapHeight = (*theGWorldPixMap)->bounds.bottom - (*theGWorldPixMap)->bounds.top;
	unsigned long realRowBytes = (*theGWorldPixMap)->rowBytes & 0x3FFF;
	unsigned long dataSize = realRowBytes * pixMapHeight;
	
	TQ3StorageObject theNewStorage = ::Q3MemoryStorage_NewBuffer((unsigned char *) ::GetPixBaseAddr(theGWorldPixMap), 0, dataSize);
	TQ3StoragePixmap theNewPixmap;

	theNewPixmap.image = theNewStorage;

	theNewPixmap.width = pixMapWidth;
	theNewPixmap.height = pixMapHeight;
	theNewPixmap.rowBytes = (*theGWorldPixMap)->rowBytes  & 0x3FFF;

	Assert_((*theGWorldPixMap)->pixelSize == 16);
	theNewPixmap.pixelType = kQ3PixelTypeRGB16;

	theNewPixmap.bitOrder = kQ3EndianBig;
	theNewPixmap.byteOrder = kQ3EndianBig;
	
	TQ3TextureObject theNewTexture = ::Q3PixmapTexture_New(&theNewPixmap);
	::Q3Object_Dispose(theNewStorage);
	
	TQ3ShaderObject theNewShader = Q3TextureShader_New(theNewTexture);
	::Q3Object_Dispose(theNewTexture);

	TQ3AttributeSet objectAttributes = nil;
	::Q3Geometry_GetAttributeSet(ioObject, &objectAttributes);
	Assert_(objectAttributes);
	::Q3AttributeSet_Add(objectAttributes, kQ3AttributeTypeSurfaceShader, &theNewShader);
	::Q3Object_Dispose(theNewShader);
	::Q3Object_Dispose(objectAttributes);
	
	// Save the GWorld so we don't leak as much
	mTextureGWorlds->InsertItemsAt(1, LArray::index_Last, &theNewGWorld);
}


// ---------------------------------------------------------------------------------
//		 IsModified
// ---------------------------------------------------------------------------------

Boolean
CgeomDocument::IsModified()
{
	return false;
}


// ---------------------------------------------------------------------------------
//		 DoAESave
// ---------------------------------------------------------------------------------

void
CgeomDocument::DoAESave(
	FSSpec	& /*inFileSpec*/,
	OSType	/*inFileType*/ )
{
/*
	// Delete the existing file object.
	// Note: this does nothing to the actual file on disk.
	delete mFileStream;
	
	// Make a new file object.
	mFileStream = new LFile( inFileSpec );
	
	// Get the proper file type.
	OSType	theFileType = 'TEXT';
	if ( inFileType != fileType_Default ) theFileType = inFileType;

	// Make new file on disk (we'll use
	// SimpleText's creator for this example).
	mFileStream->CreateNewDataFile( 'ttxt', theFileType );
	
	// Write out the data.
	DoSave();

	// Change window title to reflect the new name.
	mWindow->SetDescriptor( inFileSpec.name );

	// Document now has a specified file.
	mIsSpecified = true;
*/
}


// ---------------------------------------------------------------------------------
//		 DoSave
// ---------------------------------------------------------------------------------

void
CgeomDocument::DoSave()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Get the text from the text view.
	Handle	theTextH = mTextView->GetTextHandle();
	
	// Lock the text handle.
	StHandleLocker	theLock( theTextH );
	
	// Write the text to the file.
	mFileStream->WriteDataFork( *theTextH, ::GetHandleSize( theTextH ) );

	// Close the data fork.
	mFileStream->CloseDataFork();

	// Saving makes doc un-dirty.
	mTextView->SetDirty( false );
*/
}


// ---------------------------------------------------------------------------------
//		 DoExport
// ---------------------------------------------------------------------------------

void
CgeomDocument::DoExport()
{
	PP_StandardDialogs::LFileDesignator theSaveDesignator;
	theSaveDesignator.SetFileCreator('ttxt');
	theSaveDesignator.SetFileType(kQTFileType3DMF);
	
	FSSpec theFSSpec;
	mFileStream->GetSpecifier(theFSSpec);
	
	theSaveDesignator.SetDefaultLocation(theFSSpec, true);
	
	LStr255 defaultName;
	mWindow->GetDescriptor( defaultName );
	defaultName += "\p.3dmf";
	
	if (theSaveDesignator.AskDesignateFile(defaultName))
	{
		theSaveDesignator.GetFileSpec(theFSSpec);
	
		LFile theFile(theFSSpec);
		if (!theSaveDesignator.IsReplacing())
		{
			theFile.CreateNewDataFile('ttxt', kQTFileType3DMF);
		}
		
//		mGeomView->SaveToFile(&theFile);
	}
}


// ---------------------------------------------------------------------------------
//		 DoRevert
// ---------------------------------------------------------------------------------

void
CgeomDocument::DoRevert()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Read the entire file contents and close the file.
	Handle	theTextH = mFileStream->ReadDataFork();
	mFileStream->CloseDataFork();
	
	// Put the contents in the text view
	// and clear the dirty flag.
	mTextView->SetTextHandle( theTextH );
	mTextView->SetDirty( false );
	
	// Dispose of the text.
	::DisposeHandle( theTextH );

	// Refresh the text view.
	mTextView->Refresh();
*/
}


// ---------------------------------------------------------------------------------
//		 DoPrint
// ---------------------------------------------------------------------------------

void
CgeomDocument::DoPrint()
{
/*
	// Create the printout.
	LPrintout	*thePrintout = LPrintout::CreatePrintout( rPPob_TextPrintout );
	Assert_( thePrintout != nil );
	
	// Set the print record.
	
	if (mPrintRecordH)
	{
		thePrintout->SetPrintRecord( mPrintRecordH );
	}
	
	// Get the text placeholder.
	LPlaceHolder	*thePlaceholder;
	thePlaceholder = (LPlaceHolder*) thePrintout->FindPaneByID( kTextPlaceholder );
	Assert_( thePlaceholder != nil );
	
	// Install the text view in the placeholder.
	thePlaceholder->InstallOccupant( mTextView, atNone );
	
	// Set the frame size.
	SetPrintFrameSize();
	
	// Print.
	thePrintout->DoPrintJob();
	
	// Delete the printout. The text view is returned
	// to the window when the placeholder is destroyed.
	delete thePrintout;
*/
}


// ---------------------------------------------------------------------------------
//		 SetPrintFrameSize
// ---------------------------------------------------------------------------------

void
CgeomDocument::SetPrintFrameSize()
{
/*
	// Get the frame size.
	SDimension16	theFrameSize;
	mTextView->GetFrameSize( theFrameSize );
	
	// Get the text edit record handle.
	TEHandle	theTextEditH = mTextView->GetMacTEH();
	
	// Calculate the number of lines per page.
	Int16	theLinesPerPage;
	theLinesPerPage = theFrameSize.height / (**theTextEditH).lineHeight;

	// Resize the frame to an integral number of lines.
	mTextView->ResizeFrameTo( theFrameSize.width,
		(**theTextEditH).lineHeight * theLinesPerPage, false );

*/
}


void
CgeomDocument::ListenToMessage(
	MessageT inMessage,
	void * /* ioParam */)
{
	switch (inMessage) {
		case message_DoExport:
			this->DoExport();
			break;
	}
}
