/*******************************************************************************\
|																				|
|  CVersCaption.cp 1997-1998 John C. Daub.  All rights reserved				|
|																				|
|  See the file "CVersCaption README" for full details, instructions, changes,	|
|  licensing agreement, etc.  Due to the important information included in that	|
|  file, if you did not receive a copy of it, please contact the author for		|
|  a copy.																		|
|																				|
|  John C. Daub <mailto:hsoi@eden.com>											|
|  <http://www.eden.com/~hsoi/>  <http://www.eden.com/~hsoi/prog.html>			|
|																				|
\*******************************************************************************/


#ifdef PowerPlant_PCH
#include PowerPlant_PCH
#endif

#include "CVersCaption.h"
	
#include <LString.h>
#include <UException.h>
#include <LStream.h>
#include <UDrawingUtils.h>
#include <UTextTraits.h>
#include <UMemoryMgr.h>
 

//========================================================================
//	 CVersCaption								[public]
//========================================================================
// Default constructor

CVersCaption::CVersCaption()
{
	mVersion = 0x000;
	mUseShortString = false;
	mUseCountryCode = false;
	mText = "\p";
	
	InitVersCaption();
}


//========================================================================
//	 CVersCaption								[public]
//========================================================================
// Copy constructor

CVersCaption::CVersCaption(
	const CVersCaption &inOriginal )
		: LCaption( inOriginal )
{
	mVersion = inOriginal.mVersion;
	mUseShortString = inOriginal.mUseShortString;
	mUseCountryCode = inOriginal.mUseCountryCode;

	// there's no need to call InitVersCaption() here since we
	// should already have this information
}


//========================================================================
//	 operator=									[public]
//========================================================================
//	Assignment operator

CVersCaption&	
CVersCaption::operator=(
	const CVersCaption &inOriginal )
{
	// check for self-assignment
	if ( this != &inOriginal ) {
	
		LCaption::operator=(inOriginal);
		mVersion = inOriginal.mVersion;
		mUseShortString = inOriginal.mUseShortString;
		mUseCountryCode = inOriginal.mUseCountryCode;
		
		// LCaption doesn't explictly provide its own assignment operator,
		// so to not risk losing vital data, we'll (re)assign here.
		
		SetDescriptor( inOriginal.mText );
		SetTextTraitsID( inOriginal.mTxtrID );
	}
		
	return *this;
}


//========================================================================
//	 CVersCaption								[public]
//========================================================================
// Parameterized constructor

CVersCaption::CVersCaption(
	const SPaneInfo		&inPaneInfo,
	ConstStringPtr		inString,
	ResIDT				inTextTraitsID,
	bool				inUseShortString,
	bool				inUseCountryCode )
		: LCaption( inPaneInfo, inString, inTextTraitsID )
{
	mVersion = 0x000;
	mUseShortString = inUseShortString;
	mUseCountryCode = inUseCountryCode;
	mText = "\p";
	
	InitVersCaption();
}


//========================================================================
//	 CVersCaption								[public]
//========================================================================
// LStream constructor

CVersCaption::CVersCaption(
	LStream *inStream )
		: LCaption( inStream )
{
	mVersion = 0x0000;
	mText = "\p";
	
	inStream->ReadData( &mUseShortString, sizeof(Boolean) );
	inStream->ReadData( &mUseCountryCode, sizeof(Boolean) );
	
	InitVersCaption();
}


//========================================================================
//	 ~CVersCaption								[public, virtual]
//========================================================================
// Destructor

CVersCaption::~CVersCaption()
{
	// nothing
}


//========================================================================
//	 InitVersCaption							[private]
//========================================================================
//	Private initializer of the object
//
//	Gets the 'vers' #1 resource resource.  Return 0 if resource missing.
//	Convert resource's NumVersion (see Types.h) to the version "string"
//
//	This does assume that the application's resource fork is the current
//	resource file.  If it is not and/or a 'vers' #1 resource doesn't
//	exist, unpredictable results can occur.

void
CVersCaption::InitVersCaption()
{
	// get the resource, 'vers' 1 in the current resource fork

	StResource	versH( 'vers', 1, true, true );
	
	// if they wish to use the short string (VersRec.shortVersion), we'll just
	// extract that and return (1.2 addition)

	if ( mUseShortString ) {
		mText += (*(VersRecHndl)((Handle)versH))->shortVersion;
		return;
	}

	// copy out the version information
		
	mVersion = *((Uint32 *)(*((Handle)versH)));
	
	
	// set ver1-3, relStatus, prerelNum from the version info.
	// Note that the first two bytes are in an unusual format.
	
	Uchar	ver1,
			ver2,
			ver3,
			relStatus, 
			prerelNum;

	ver1 = ((Uchar *)&mVersion)[0];
	ver1 = (((ver1 & 0xF0 ) >> 4) * 10) + (ver1 & 0x0F);
	ver2 = (((Uchar *)&mVersion)[1] & 0xF0) >> 4;
	ver3 = (((Uchar *)&mVersion)[1] & 0x0F);
	relStatus = ((Uchar *)&mVersion)[2];
	prerelNum = ((Uchar *)&mVersion)[3];
	prerelNum = (((prerelNum & 0xF0 ) >> 4) * 10) + (prerelNum & 0x0F); // 1.1 fix
	
	// Insert v1 and v2 into our version string
	
	mText += static_cast<Int32>(ver1);
	
	StringHandle dotH = ::GetString( STR_dot );
	Uchar theDot = (dotH) ? (*dotH)[1] : '.';
	::ReleaseResource((Handle)dotH);
	
	mText += theDot;
	mText += static_cast<Int32>(ver2);
	
	// to make it nicer, we only print the third number if it's non-zero
	if ( ver3 ) {
		mText += theDot;
		mText += static_cast<Int32>(ver3);
	}
		
	
	// if the release status is development, alpha or beta, add
	// a 'd', 'a', or 'b' to our version string.
	// if you do something like "GM" or "FC1/FC2/etc" before you
	// ship a product, you could add in another case here for that
	// and then just conditionally wrap it with preprocessor macros
	// to be compiled out when you #define __DEBUG__ to false (or
	// whatever debugger macros are turned off).
		
	ResIDT	stringID = 0;
	
	switch( relStatus ) {
		case 0x20: {			
			// development
			stringID = STR_devel;
			break;
		}
		
		case 0x40: {
			// alpha
			stringID = STR_alpha;
			break;
		}
		
		case 0x60: {
			// beta
			stringID = STR_beta;
			break;
		}
	}

	
	if ( relStatus != 0x80 ) {	
		StringHandle relH = ::GetString(stringID);
		ThrowIfResFail_(relH);
		
		// if we've added a 'd', 'a', or 'b', print the
		// pre-release number at the end.
	
		mText += (Uchar)((*relH)[1]);
		mText += static_cast<Int32>(prerelNum);

		::ReleaseResource((Handle)relH);
	}
	
	// lastly, if we want to use the country code, let's stick that
	// on.
	
	if ( mUseCountryCode ) {
		Str255	countryStr;
		::GetIndString( countryStr, STRx_countries,
						(*(VersRecHndl)((Handle)versH))->countryCode + 1 );
		mText += "\p ";
		LString::AppendPStr( mText, countryStr );
	}
}


//========================================================================
//	 ActivateSelf							[protected, virtual]
//========================================================================
//
//	We need to force a redraw when the pane becomes active so that it
//	will draw properly.

void
CVersCaption::ActivateSelf()
{
	if ( mActive == triState_On ) {
		Refresh();
	}
}


//========================================================================
//	 DeactivateSelf						[protected, virtual]
//========================================================================
//
//	We need to force a redraw when the pane becomes deactivated

void
CVersCaption::DeactivateSelf()
{
	if ( mActive == triState_Off || mActive == triState_Latent ) {
		Refresh();
	}
}


//========================================================================
//	 DrawSelf								[protected, virtual]
//========================================================================
//
//	Similar to LCaption::DrawSelf, but if the pane isn't active, the
//	text is drawn in a "dimmed" state.

void
CVersCaption::DrawSelf()
{
	Rect	frame;
	CalcLocalFrameRect(frame);
	
	Int16	just = UTextTraits::SetPortTextTraits(mTxtrID);
	
	RGBColor	textColor;
	::GetForeColor(&textColor);
	
	ApplyForeAndBackColors();
	
	if ( !IsActive() ) {

		// lighten the text color a bit
		
		textColor.red = textColor.red + 65535 >> 1;
		textColor.green = textColor.green + 65535 >> 1;
		textColor.blue = textColor.blue + 65535 >> 1;
	}	
	
	::RGBForeColor(&textColor);
	
	UTextDrawing::DrawWithJustification((Ptr)&mText[1], mText[0], frame, just);
}


// the following functions are declared inline in the header file. The #pragma
// mark let's them show up in the CW IDE's function popup for ease of navigation
// and reference. :-)

#pragma mark CVersCaption::GetVersion
#pragma mark CVersCaption::SetVersion
#pragma mark CVersCaption::GetUseShortString
#pragma mark CVersCaption::SetUseShortString
#pragma mark CVersCaption::GetUseCountryCode
#pragma mark CVersCaption::SetUseCountryCode