/*
Copyright  1997 Sailmaker Software Ltd. All rights reserved.
			<mailto:richardb@sailmaker.co.uk>
			<http://www.sailmaker.co.uk/>

File:		CDecimalDescription.cp
Author:		Richard Buckle
Purpose:	mixin class to manage formatted conversion to and from decimal numbers
Status:		Copyrighted freeware

See read me for instructions and conditions of use.

*/

#include "CDecimalDescription.h"
#include <fp.h>

// See Inside Mac:Text page 5-35 onwards.
const ResType FStrType = 'FStr';
	// 'FStr' resources are the same format as 'Str ' resources,
	// but contain a number format string as described in Inside Mac.
	// 'FStr' resources are localisation-dependent and so it's 
	// better to work with 'FMAT' resources.
const ResType FMATType = 'FMAT';
	// 'FMAT' resources are a binary format containing a
	// canonical (ie localisation-independent) number format
	// structure, NumFormatStringRec, as per <NumberFormatting.h> 
	// The easiest way to create them is Michael Hecht's
	// 'FMAT' editor ResEdit extension, which is on the Apple
	// developer Tool Chest CDs under Utilities:Localization. 
	// There's also a copy at my website
	// <http://www.sailmaker.co.uk/>


#pragma mark ----- Instance methods -----
// stream constructor
CDecimalDescription::CDecimalDescription( LStream *inStream )
	{
	// get our format string resource ID
	ResIDT	theFStrID;
	*inStream >> theFStrID;
	InitDecimalDescription(theFStrID);
	}

// ResID constructor
CDecimalDescription::CDecimalDescription( const ResIDT	inFStrID )
	{
	InitDecimalDescription(inFStrID);
	}

// destructor	
CDecimalDescription::~CDecimalDescription()
	{}
	
#pragma mark -
#pragma mark ----- Public accessors -----
// return the value as a double
double CDecimalDescription::StrToDoubleValue( const Str255 iStr )
	{
	// if the string is empty then return zero
	if( iStr[0] == 0 )
		return 0.0;
	
	// convert text to extended80
	extended80 theNumber;
	FormatResultType theStatus 
		= ::StringToExtended( iStr, &mFormat, &mNumberParts, &theNumber );
	
	// beep if the string was bad
	if( theStatus > fBestGuess )
		::SysBeep(0);
	
	// convert extended80 to double
	double result = ToDouble( theNumber );
	
	// convert NANs to zero
	if( isnan(result) )	result = 0.0;
	
	return result;
	}
	
// set the value to the given double
void CDecimalDescription::DoubleValueToStr( Str255 oStr, 
											const double inValue, 
											const Boolean inStripTrailingDecimal )
	{
	// convert double to extended80
	extended80 theNumber = ToExtended80( inValue );
	
	// convert extended80 to text
	::ExtendedToString( &theNumber, &mFormat, &mNumberParts, oStr );
	
	if( inStripTrailingDecimal )
		{
		DoStripTrailingDecimal( oStr, inValue );
		}
	}
	
	
#pragma mark -
#pragma mark ----- Utility functions for 68K -----
// convert double to extended80
extended80 CDecimalDescription::ToExtended80( const double inValue )
	{
	extended80 result;
	
	#if TARGET_CPU_68K
		#if TARGET_RT_MAC_68881
			long double theLongDouble = inValue;
			::x96tox80( &theLongDouble, &result );
		#else
			result = inValue;
		#endif
	#elif TARGET_CPU_PPC
		::dtox80( &inValue, &result );
	#else
		#error // floating point environment not recognised
	#endif
	
	return result;
	}

// convert extended80 to double	
double CDecimalDescription::ToDouble( const extended80 inValue )
	{
	double result;
	
	#if TARGET_CPU_68K
		#if TARGET_RT_MAC_68881
			long double theLongDouble;
			::x80tox96( &inValue, &theLongDouble );
			result = theLongDouble;
		#else
			result = inValue;
		#endif
	#elif TARGET_CPU_PPC
		result = ::x80tod( &inValue );
	#else
		#error // floating point environment not recognised
	#endif
	
	return result;
	}
	

#pragma mark -
#pragma mark ----- Implementation -----
// strip any trailing ".0" if we think the number's an integer
void CDecimalDescription::DoStripTrailingDecimal( Str255 ioStr, const double iValue )
	{
	// round iValue to our format
	Str255 tempStr;
	DoubleValueToStr( tempStr, iValue );
	double roundedValue = StrToDoubleValue( tempStr );
	
	// test whether roundedValue is integral
	if( roundedValue==::round(roundedValue) )
		{
		// check that theStr ends in ".0"
		unsigned char theLength = StrLength(ioStr);
		if( (theLength > 2)
			&& (ioStr[theLength-1] == '.') 
			&& (ioStr[theLength] == '0') )
			{
			// truncate
			ioStr[0] -= 2;
			}
		}
	
	}
	
	
// init or re-init ourselves from a 'FMAT' or 'FStr' resource
void CDecimalDescription::InitDecimalDescription( const ResIDT iResID )
	{
	// get the local number parts table
	Handle	intlHandle;
	long 	numpartsOffset, numpartsLength;
	::GetIntlResourceTable( smSystemScript, smNumberPartsTable, &intlHandle, 
												&numpartsOffset, &numpartsLength );
	ThrowIfNil_( intlHandle );
	::BlockMoveData( *intlHandle+numpartsOffset, &mNumberParts, sizeof(mNumberParts) );
	
	// first look for an FMAT resource
	NumFormatString **theFormatHdl 
		= ( NumFormatString ** )::GetResource( FMATType, iResID );
	// got it?
	if( theFormatHdl && *theFormatHdl )
		{
		Size s1 = ::GetHandleSize( Handle(theFormatHdl) );
		Size s2 = sizeof(mFormat);
		// store it in mFormat
		::BlockMoveData( *theFormatHdl, &mFormat, sizeof(mFormat) );
		
		// we don't release the resource - better just to mark it purgeable
		}
	else
		{
		// no FMAT with the given ID; fall back on FStr
		
		// set a sensible default
		StringPtr theStr = "\p###,##0.#";
		
		// look for an FStr resource
		StringHandle theFormatStrHdl = (StringHandle)::GetResource( FStrType, iResID );
		
		// got it?
		if( theFormatStrHdl && *theFormatStrHdl )
			{
			theStr = *theFormatStrHdl;
			}
		
		// convert the format string to a canonical number format
		FormatResultType theStatus 
			= ::StringToFormatRec( theStr, &mNumberParts, &mFormat );
		Assert_( theStatus == fFormatOK );

		// we don't release the resource - better just to mark it purgeable
		}
	
	}
	
