// ===========================================================================
//	C3DMapView.cp				   1997 eli curtz. All rights reserved.
// ===========================================================================
//

#ifdef PowerPlant_PCH
#include PowerPlant_PCH
#endif

#include <LPicture.h>
#include <LStream.h>
#include <UDrawingState.h>
#include <UGWorld.h>

#include <QD3DGroup.h>
#include <QD3DLight.h>
#include <QD3DCamera.h>
#include <QD3DShader.h>
#include <QD3DGeometry.h>
#include <QD3DDrawContext.h>
#include <QD3DRenderer.h>
#include <QD3DView.h>
#include <QD3DMath.h>
#include <QD3DStorage.h>
#include <QD3DPick.h>

#include "Constants.h"
#include "CMapObject.h"

#include "C3DMapView.h"

// ---------------------------------------------------------------------------
//	 C3DMapView						Default Constructor		  [public]
// ---------------------------------------------------------------------------

C3DMapView::C3DMapView()
	: LView(), LBroadcaster()
{
	mGWorld = nil;

	mQD3DView = nil;
	mQD3DMeshGroup = nil;
	mRenderTexture = false;
	
	InitializeView();
}


// ---------------------------------------------------------------------------
//	 C3DMapView(LStream*)			Stream Constructor	  [public]
// ---------------------------------------------------------------------------

C3DMapView::C3DMapView(
	LStream	*inStream)
		: LView(inStream), LBroadcaster(), LCommander()
{
	mGWorld = nil;

	mQD3DView = nil;
	mQD3DMeshGroup = nil;
	mRenderTexture = false;

	InitializeView();
}

// ---------------------------------------------------------------------------
//	 ~C3DMapView()			
// ---------------------------------------------------------------------------

C3DMapView::~C3DMapView()
{
	if (mQD3DMeshGroup)
		::Q3Object_Dispose(mQD3DMeshGroup);
}

// ---------------------------------------------------------------------------
//	 InitializeView()
// ---------------------------------------------------------------------------

void
C3DMapView::InitializeView()
{
	mQD3DView = ::Q3View_New();
	TQ3Status theStatus;
	Rect thePortRect;
	Assert_(CalcPortFrameRect(thePortRect));
	
// Draw Context
	// Build the generic draw context
	TQ3DrawContextObject theDrawContext = nil;
	TQ3DrawContextData theContextData;
	
	// Erase Color
	TQ3ColorARGB theWhiteColor;	
	::Q3ColorARGB_Set(&theWhiteColor, 1.0, 1.0, 1.0, 1.0);
	
	// Clip Rect
	TQ3Area thePaneArea;
	TQ3Point2D theMinPanePoint;
	TQ3Point2D theMaxPanePoint;
	theMinPanePoint.x = thePortRect.left; theMinPanePoint.y = thePortRect.top;
	theMaxPanePoint.x = thePortRect.right; theMaxPanePoint.y = thePortRect.bottom;
	thePaneArea.min = theMinPanePoint;
	thePaneArea.max = theMaxPanePoint;

	// Mask
	TQ3Bitmap theNullBitmap = { nil, 0, 0, 0, kQ3EndianBig};
	
	theContextData.clearImageMethod = kQ3ClearMethodWithColor;
	theContextData.clearImageColor = theWhiteColor;
	theContextData.pane = thePaneArea;
	theContextData.paneState = kQ3True;
	theContextData.mask = theNullBitmap;
	theContextData.maskState = kQ3False;
	theContextData.doubleBufferState = kQ3True;

	// Build the macintosh draw context
	TQ3MacDrawContextData theMacContextData;
	
	theMacContextData.drawContextData = theContextData;
	theMacContextData.window = (CWindowPtr) GetMacPort();
	theMacContextData.library = kQ3Mac2DLibraryNone;	// Do we need this for clipping?
	theMacContextData.viewPort = nil;
	theMacContextData.grafPort = nil;

	// Set the draw context in the view
	theDrawContext = ::Q3MacDrawContext_New(&theMacContextData);
	ThrowIfNil_(theDrawContext);
	theStatus = ::Q3View_SetDrawContext(mQD3DView, theDrawContext);
	ThrowIf_(theStatus == kQ3Failure);
	::Q3Object_Dispose(theDrawContext);
	

// Lights
	// Build the light group
	TQ3GroupObject theLightList;
	TQ3LightData theLightData;
	TQ3LightObject theAmbientLight;
	TQ3GroupPosition theGroupPosition;
	
	// Light color
	TQ3ColorRGB	theWhiteLightColor;
	::Q3ColorRGB_Set(&theWhiteLightColor, 1.0, 1.0, 1.0);
	
	// Create ambient light
	theLightData.isOn = kQ3True;
	theLightData.brightness = 0.2;
	theLightData.color = theWhiteLightColor;
	
	theAmbientLight = ::Q3AmbientLight_New(&theLightData);
	ThrowIfNil_(theAmbientLight);

	// Create light group and add lights to the view
	theLightList = ::Q3LightGroup_New();
	ThrowIfNil_(theLightList);
	
	theGroupPosition = ::Q3Group_AddObject(theLightList, theAmbientLight);
	ThrowIfNil_(theGroupPosition);
	::Q3Object_Dispose(theAmbientLight);
	
	theStatus = ::Q3View_SetLightGroup(mQD3DView, theLightList);
	ThrowIf_(theStatus == kQ3Failure);
	::Q3Object_Dispose(theLightList);
	

// Camera	
	// Build the camera
	TQ3CameraObject	theCamera;
	TQ3CameraData theCameraData;
	
	// Generic camera data
	TQ3Point3D cameraLocation = { 0.0, 15.0, 40.0 };
	TQ3Point3D cameraFocus = { 30.0, 0.0, 30.0 };
	TQ3Vector3D cameraUp = { 0.0, 1.0, 0.0 };

	float hitherValue = 0.001;
	float yonValue = 1000;

	theCameraData.placement.cameraLocation = cameraLocation;
	theCameraData.placement.pointOfInterest = cameraFocus;
	theCameraData.placement.upVector = cameraUp;

	theCameraData.range.hither = hitherValue;
	theCameraData.range.yon = yonValue;

	theCameraData.viewPort.origin.x = -1.0;
	theCameraData.viewPort.origin.y = 1.0;
	theCameraData.viewPort.width = 2.0;
	theCameraData.viewPort.height = 2.0;
	
	// Angle Aspect camera data
	TQ3ViewAngleAspectCameraData theAngleAspectCameraData;
	
	float fieldOfView = 1.0;

	theAngleAspectCameraData.cameraData = theCameraData;
	theAngleAspectCameraData.fov = fieldOfView;
	theAngleAspectCameraData.aspectRatioXToY = (float) (thePortRect.right - thePortRect.left) / 
		(float) (thePortRect.bottom - thePortRect.top);
	
	// Create the camera and add it to the view
	theCamera = ::Q3ViewAngleAspectCamera_New(&theAngleAspectCameraData);
	ThrowIfNil_(theCamera);
	
	theStatus = ::Q3View_SetCamera(mQD3DView, theCamera);
	ThrowIf_(theStatus == kQ3Failure);
	::Q3Object_Dispose(theCamera);
	

// Renderer	
	// Build the renderer and add it to view
	TQ3RendererObject theRenderer;
	if (mRenderTexture)
		theRenderer = Q3Renderer_NewFromType(kQ3RendererTypeInteractive);
	else
		theRenderer = Q3Renderer_NewFromType(kQ3RendererTypeWireFrame);
	
	ThrowIfNil_(theRenderer);	
	
	theStatus = ::Q3View_SetRenderer(mQD3DView, theRenderer);
	ThrowIf_(theStatus == kQ3Failure);
	::Q3Object_Dispose(theRenderer);

// Picker
	TQ3WindowPointPickData theData;
	theData.data.sort = kQ3PickSortNearToFar;
	theData.data.mask = kQ3PickDetailMaskPickID | kQ3PickDetailMaskXYZ | kQ3PickDetailMaskUV;
	theData.data.numHitsToReturn = kQ3ReturnAllHits;
	theData.point.x = 0;
	theData.point.y = 0;
	theData.vertexTolerance = 0.1;
	theData.edgeTolerance = 0.1;

	mQD3DPicker = ::Q3WindowPointPick_New(&theData);
	if (mQD3DPicker == nil)
		SignalPStr_("\pCouldn't create QucikDraw 3D Pick Object");
}


// ---------------------------------------------------------------------------
//	 GetGWorld													  [public]
// ---------------------------------------------------------------------------

LGWorld*
C3DMapView::GetGWorld()
{
	return mGWorld;
}


// ---------------------------------------------------------------------------
//	 SetGWorld													  [public]
// ---------------------------------------------------------------------------

void
C3DMapView::SetGWorld(
	LGWorld*	inGWorld)
{
	mGWorld = inGWorld;
	if (mQD3DMeshGroup) {
		GWorldPtr macGWorld = mGWorld->GetMacGWorld();	

		PixMapHandle theGWorldPixMap = ::GetGWorldPixMap(macGWorld);
		::LockPixels(theGWorldPixMap);
		unsigned long pixMapWidth = (*theGWorldPixMap)->bounds.right - (*theGWorldPixMap)->bounds.left;
		unsigned long pixMapHeight = (*theGWorldPixMap)->bounds.bottom - (*theGWorldPixMap)->bounds.top;
		unsigned long realRowBytes = (*theGWorldPixMap)->rowBytes & 0x3FFF;
		unsigned long dataSize = realRowBytes * pixMapHeight;
		
		TQ3StorageObject theNewStorage = ::Q3MemoryStorage_NewBuffer((unsigned char *) ::GetPixBaseAddr(theGWorldPixMap), 0, dataSize);
		TQ3StoragePixmap theNewPixmap;

		theNewPixmap.image = theNewStorage;

		theNewPixmap.width = pixMapWidth;
		theNewPixmap.height = pixMapHeight;
		theNewPixmap.rowBytes = (*theGWorldPixMap)->rowBytes  & 0x3FFF;

		Assert_((*theGWorldPixMap)->pixelSize == 16);
		theNewPixmap.pixelType = kQ3PixelTypeRGB16;

		theNewPixmap.bitOrder = kQ3EndianBig;
		theNewPixmap.byteOrder = kQ3EndianBig;
		
		TQ3TextureObject theNewTexture = ::Q3PixmapTexture_New(&theNewPixmap);
		::Q3Object_Dispose(theNewStorage);
		
		TQ3ShaderObject theNewShader = Q3TextureShader_New(theNewTexture);
		::Q3Object_Dispose(theNewTexture);

		::Q3Group_AddObject(mQD3DMeshGroup, theNewShader);
		::Q3Object_Dispose(theNewShader);
	}
}


// ---------------------------------------------------------------------------
//	 GetMapMesh													  [public]
// ---------------------------------------------------------------------------

TQ3GroupObject
C3DMapView::GetMapMesh()
{
	return mQD3DMeshGroup;
}


// ---------------------------------------------------------------------------
//	 SetMapMesh(TQ3GroupObject)													  [public]
// ---------------------------------------------------------------------------

void
C3DMapView::SetMapMesh(
	TQ3GroupObject	inMesh)
{
// Hack - we should probably ref count the mesh here and ditch the old mesh
	mQD3DMeshGroup = inMesh;
	
	// Force the build of QD3D texture map
	if (mGWorld)
		SetGWorld(mGWorld);
}

// ---------------------------------------------------------------------------
//	 SetUseTexture(Boolean)													  [public]
// ---------------------------------------------------------------------------

void
C3DMapView::SetUseTexture(
	Boolean inTexture)
{
	mRenderTexture = inTexture;
	
	TQ3RendererObject theRenderer;
	if (mRenderTexture)
		theRenderer = Q3Renderer_NewFromType(kQ3RendererTypeInteractive);
	else
		theRenderer = Q3Renderer_NewFromType(kQ3RendererTypeWireFrame);
	
	ThrowIfNil_(theRenderer);	
	
	TQ3Status theStatus = ::Q3View_SetRenderer(mQD3DView, theRenderer);
	ThrowIf_(theStatus == kQ3Failure);
	::Q3Object_Dispose(theRenderer);
	
	Refresh();
}

// ---------------------------------------------------------------------------
//	 DrawSelf													   [protected]
// ---------------------------------------------------------------------------

void
C3DMapView::DrawSelf()
{
	TQ3BoxData theBox;
	theBox.faceAttributeSet = nil;
	theBox.boxAttributeSet = nil;
	::Q3Vector3D_Set(&theBox.orientation, 0.0, 0.2, 0.0);
	::Q3Vector3D_Set(&theBox.minorAxis, 0.2, 0.0, 0.0);
	::Q3Vector3D_Set(&theBox.majorAxis, 0.0, 0.0, 0.2);

	if (mQD3DView && mQD3DMeshGroup) {
		TQ3Status status;
		TQ3ViewStatus viewStatus;
		TQ3GroupObject displayGroup;
		
		status = ::Q3View_StartRendering(mQD3DView);
		ThrowIf_(status == kQ3Failure);
		
		CMapObject* currentObject = nil;
		LArrayIterator objectIter(*mObjects);
		SPosition32 objectLoc;
		
		do {
			status = ::Q3DisplayGroup_Submit(mQD3DMeshGroup, mQD3DView);
			ThrowIf_(status == kQ3Failure);

			while (objectIter.Next(&currentObject)) {
				if (currentObject->IsVisible()) {
					if (displayGroup = currentObject->GetDisplayGroup()) {
						status = ::Q3DisplayGroup_Submit(displayGroup, mQD3DView);
						ThrowIf_(status == kQ3Failure);
					}
					else {
						objectLoc = currentObject->GetPositionXYZ();
						
						::Q3Point3D_Set(&theBox.origin, (float) (objectLoc.x / 1024) - 0.1, (float) objectLoc.y / 1024, (float) (objectLoc.z / 1024) - 0.1);
						theBox.boxAttributeSet = CMapObject::sColorAttributes[currentObject->GetTypeIndex()];
						
						status = ::Q3Box_Submit(&theBox, mQD3DView);
						ThrowIf_(status == kQ3Failure);
					}
				}
			}
			
			viewStatus = ::Q3View_EndRendering(mQD3DView);
		} while (viewStatus == kQ3ViewStatusRetraverse);
	}	
}
	
// ----------------------------------------------------------------------------
//	 ResizeFrameBy												[public]
// ----------------------------------------------------------------------------

void
C3DMapView::ResizeFrameBy(Int16 inWidthDelta, Int16 inHeightDelta, Boolean /* inRefresh */)
{
	LView::ResizeFrameBy(inWidthDelta, inHeightDelta, false);
	if (mQD3DView) {
		Rect thePortRect;
		CalcPortFrameRect(thePortRect);
		
		TQ3Status status;
		TQ3DrawContextObject theDrawContext;
		
		// Update the draw context
		status = ::Q3View_GetDrawContext(mQD3DView, &theDrawContext);
		ThrowIf_(status == kQ3Failure);
		
		TQ3Area thePaneArea;
		TQ3Point2D theMinPanePoint;
		TQ3Point2D theMaxPanePoint;
		theMinPanePoint.x = thePortRect.left; theMinPanePoint.y = thePortRect.top;
		theMaxPanePoint.x = thePortRect.right; theMaxPanePoint.y = thePortRect.bottom;
		thePaneArea.min = theMinPanePoint;
		thePaneArea.max = theMaxPanePoint;
		
		status = ::Q3DrawContext_SetPane(theDrawContext, &thePaneArea);
		ThrowIf_(status == kQ3Failure);
		
		// update the camera
		TQ3CameraObject theCamera;
		status = ::Q3View_GetCamera(mQD3DView, &theCamera);
		ThrowIf_(status == kQ3Failure);
		
		float aspectRatioXToY = (float) (thePortRect.right - thePortRect.left) / 
			(float) (thePortRect.bottom - thePortRect.top);
		
		status = ::Q3ViewAngleAspectCamera_SetAspectRatio(theCamera, aspectRatioXToY);
		ThrowIf_(status == kQ3Failure);
	}
}


// ---------------------------------------------------------------------------
//	 ClickSelf(const SMouseDownEvent&)								 [protected]
// ---------------------------------------------------------------------------

void
C3DMapView::ClickSelf(
	const SMouseDownEvent& inMouseDown)
{
//	if (GetClickCount() > 1) {
//	}
//	else {
// HACK - Total hack
HandlePick(inMouseDown);
/*
TQ3CameraObject theCamera;
TQ3Status status = ::Q3View_GetCamera(mQD3DView, &theCamera);
ThrowIf_(status == kQ3Failure);

TQ3CameraPlacement theCameraPlacement;
status = ::Q3Camera_GetPlacement(theCamera, &theCameraPlacement);
ThrowIf_(status == kQ3Failure);

float xAngle = 0.0;
float yAngle = 0.0;
float zAngle = 0.0;

Rect localRect;
ThrowIfNot_(CalcLocalFrameRect(localRect));
::InsetRect(&localRect, 15, 15);

if (inMouseDown.whereLocal.h < localRect.left)
	yAngle -= 0.2;
if (inMouseDown.whereLocal.h > localRect.right)
	yAngle += 0.2;
if (inMouseDown.whereLocal.v < localRect.top)
	zAngle -= 0.2;
if (inMouseDown.whereLocal.v > localRect.bottom)
	zAngle += 0.2;


if ((yAngle != 0) || (zAngle != 0)) {
TQ3Matrix4x4 cameraMatrix;
::Q3Matrix4x4_SetRotateAboutPoint(&cameraMatrix, &theCameraPlacement.pointOfInterest, xAngle, yAngle, zAngle);
TQ3Point3D newCameraLoc;
::Q3Point3D_Transform(&theCameraPlacement.cameraLocation, &cameraMatrix, &newCameraLoc);
theCameraPlacement.cameraLocation =	newCameraLoc;
}
else {
TQ3Vector3D theMovementVector;
theMovementVector.x = theCameraPlacement.pointOfInterest.x - theCameraPlacement.cameraLocation.x;
theMovementVector.z = theCameraPlacement.pointOfInterest.z - theCameraPlacement.cameraLocation.z;
::Q3Vector3D_Normalize(&theMovementVector, &theMovementVector);
::Q3Vector3D_Scale(&theMovementVector, 2.0, &theMovementVector);
::Q3Point3D_Vector3D_Add(&theCameraPlacement.cameraLocation, &theMovementVector, &theCameraPlacement.cameraLocation);
::Q3Point3D_Vector3D_Add(&theCameraPlacement.pointOfInterest, &theMovementVector, &theCameraPlacement.pointOfInterest);
}

status = ::Q3Camera_SetPlacement(theCamera, &theCameraPlacement);
ThrowIf_(status == kQ3Failure);

this->Refresh();
//}
*/
}


// ---------------------------------------------------------------------------
//	 HandlePick(const SMouseDownEvent&)								 [protected]
// ---------------------------------------------------------------------------

void
C3DMapView::HandlePick(
	const SMouseDownEvent& inMouseDown)
{
// HACK
	Boolean pickInObjectMode = false;

	TQ3Status status;
	TQ3Point2D qd3DWhere;
	qd3DWhere.x = inMouseDown.wherePort.h;
	qd3DWhere.y = inMouseDown.wherePort.v;
	
	TQ3Status qd3DStatus = ::Q3WindowPointPick_SetPoint(mQD3DPicker, &qd3DWhere);
	if (qd3DStatus != kQ3Success)
		SignalPStr_("\pCouldn't set QD3D picker point for Peter");
	
	::Q3View_StartPicking(mQD3DView, mQD3DPicker);
	if (qd3DStatus != kQ3Success)
		SignalPStr_("\pCouldn't start QD3D picking");

	do {
		if (pickInObjectMode) {
			// Try for a selection pick among visible objects
			TQ3BoxData theBox;
			theBox.faceAttributeSet = nil;
			theBox.boxAttributeSet = nil;
			::Q3Vector3D_Set(&theBox.orientation, 0.0, 0.2, 0.0);
			::Q3Vector3D_Set(&theBox.minorAxis, 0.2, 0.0, 0.0);
			::Q3Vector3D_Set(&theBox.majorAxis, 0.0, 0.0, 0.2);

			CMapObject* currentObject = nil;
			TQ3GroupObject displayGroup;
			SPosition32 objectLoc;

			LArrayIterator objectIter(*mObjects);
			while (objectIter.Next(&currentObject)) {
				if (currentObject->IsVisible()) {
					// Use the object pointer as the Pick ID
					status = ::Q3PickIDStyle_Submit ((unsigned long) currentObject, mQD3DView);
					ThrowIf_(status == kQ3Failure);
					
					if (displayGroup = currentObject->GetDisplayGroup()) {
						status = ::Q3DisplayGroup_Submit(displayGroup, mQD3DView);
						ThrowIf_(status == kQ3Failure);
					}
					else {
						objectLoc = currentObject->GetPositionXYZ();
						
						::Q3Point3D_Set(&theBox.origin, (float) (objectLoc.x / 1024) - 0.1, (float) objectLoc.y / 1024, (float) (objectLoc.z / 1024) - 0.1);
						status = ::Q3Box_Submit(&theBox, mQD3DView);
						ThrowIf_(status == kQ3Failure);
					}
				}
			}
		}
		else {
			// Pick the mesh for hit information
			::Q3DisplayGroup_Submit(mQD3DMeshGroup, mQD3DView);
			ThrowIf_(status == kQ3Failure);
		}
	} while (::Q3View_EndPicking(mQD3DView) == kQ3ViewStatusRetraverse);
	
	unsigned long hitCount = 0;
	qd3DStatus = ::Q3Pick_GetNumHits(mQD3DPicker, &hitCount);
	if (qd3DStatus != kQ3Success)
		SignalPStr_("\pCouldn't get QD3D picker hit list");
	
	TQ3PickDetail qd3DHitDetail;
	UInt32 qd3DPickID;
// TQ3Param2D qd3DUVParam;
	TQ3Point3D qd3DXYZPoint;
	// We're only going to deal with the frontmost hit
	if (hitCount >= 1) {
		qd3DStatus = ::Q3Pick_GetPickDetailValidMask(mQD3DPicker, 0, &qd3DHitDetail);
		if (qd3DStatus != kQ3Success)
			SignalPStr_("\pBad hit detail information from QD3D pick");
		
		if (pickInObjectMode) {
			if (qd3DHitDetail & kQ3PickDetailMaskPickID) {
				qd3DStatus = ::Q3Pick_GetPickDetailData(mQD3DPicker, 0, kQ3PickDetailMaskPickID, &qd3DPickID);
				if (qd3DStatus != kQ3Success)
					SignalPStr_("\pBad pick ID from QD3D");
				else {
					CMapObject* pickedObject = (CMapObject *) qd3DPickID;
					pickedObject->DisplayObjectWindow();
				}
			}
		}
		else {
			if (qd3DHitDetail & kQ3PickDetailMaskXYZ) {
				qd3DStatus = ::Q3Pick_GetPickDetailData(mQD3DPicker, 0, kQ3PickDetailMaskXYZ, &qd3DXYZPoint);
				if (qd3DStatus != kQ3Success)
					SignalPStr_("\pBad XYZ from QD3D pick");
				
				// Build the world and frustum transforms
				TQ3CameraObject theCamera;
				TQ3Matrix4x4 matrixWorldToFrustum;
				TQ3Matrix4x4 matrixFrustumToWorld;
				
				qd3DStatus = ::Q3View_GetCamera(mQD3DView, &theCamera);
				if (qd3DStatus != kQ3Success)
					SignalPStr_("\pCouldn't get camera.");
				qd3DStatus = ::Q3Camera_GetWorldToFrustum(theCamera, &matrixWorldToFrustum);
				if (qd3DStatus != kQ3Success)
					SignalPStr_("\pCouldn't get frustum transform.");
				
				::Q3Matrix4x4_Invert(&matrixWorldToFrustum, &matrixFrustumToWorld);
				
				// Get the mesh point corresponding to our hit
				Int16 theClickedRow = (qd3DXYZPoint.z + 0.5);
				Int16 theClickedColumn = (qd3DXYZPoint.x + 0.5);

				// Get the TriGrid
				TQ3GroupPosition thePosition;
				qd3DStatus = ::Q3Group_GetFirstPositionOfType(mQD3DMeshGroup, kQ3GeometryTypeTriGrid, &thePosition);
				ThrowIf_((qd3DStatus == kQ3Failure) || (thePosition == nil));
				
				TQ3GeometryObject theTriGrid;
				qd3DStatus = ::Q3Group_GetPositionObject(mQD3DMeshGroup, thePosition, &theTriGrid);
				ThrowIf_(qd3DStatus == kQ3Failure);

				TQ3TriGridData theTriGridData;
				qd3DStatus = ::Q3TriGrid_GetData(theTriGrid, &theTriGridData);
				ThrowIf_(qd3DStatus == kQ3Failure);

				Point mousePoint;
				TQ3Point3D qd3DFrustumPoint;
				Rect frameRect;
				Assert_(CalcLocalFrameRect(frameRect));

				while(::StillDown()) {
					::GetMouse(&mousePoint);	// Which is in PORT coordinates IM::Mac Toolbox Essentials 2-108
					
					if (PointIsInFrame(mousePoint.h, mousePoint.v)) {
						::Q3Point3D_Transform(&qd3DXYZPoint, &matrixWorldToFrustum, &qd3DFrustumPoint);

						PortToLocalPoint(mousePoint);
						qd3DFrustumPoint.y = ((float) mousePoint.v * -2.0 / frameRect.bottom) + 1.0;
						
						::Q3Point3D_Transform(&qd3DFrustumPoint, &matrixFrustumToWorld, &qd3DXYZPoint);
						qd3DXYZPoint.x = theClickedColumn;
						qd3DXYZPoint.z = theClickedRow;
						
						// Adjust for our strange but convenient mapping
						::Q3TriGrid_SetVertexPosition(theTriGrid, theClickedRow, theTriGridData.numColumns - theClickedColumn, &qd3DXYZPoint);
						DrawSelf();
					}
				}
				
				::Q3Object_Dispose(theTriGrid);	// Q3Group_GetPositionObject() ref counted it
			}
/*
			if (qd3DHitDetail & kQ3PickDetailMaskUV) {
				qd3DStatus = ::Q3Pick_GetPickDetailData(mQD3DPicker, 0, kQ3PickDetailMaskUV, &qd3DUVParam);
				if (qd3DStatus != kQ3Success)
					SignalPStr_("\pBad UV from QD3D pick");
			}
*/		}
	}
	
	::Q3Pick_EmptyHitList(mQD3DPicker);
}	


// ---------------------------------------------------------------------------
//	 SetObjects(LArray*)								
// ---------------------------------------------------------------------------

void
C3DMapView::SetObjects(LArray* inObjects)
{
	mObjects = inObjects;
}


// ---------------------------------------------------------------------------
//	 HandleKeyPress(EventRecord&)								
// ---------------------------------------------------------------------------

Boolean
TestKeyBit(KeyMap& inKeyMap, UInt32 inTestBit)
{
	UInt8* theRealArray = (UInt8*) inKeyMap;
	
	return (theRealArray[inTestBit / 8] & (0x01 << (inTestBit % 8)));	
}

Boolean
C3DMapView::HandleKeyPress(
	const EventRecord&	inKeyEvent)
{
	Boolean		keyHandled 	= false;
	
	if ( inKeyEvent.modifiers & cmdKey )
	{
		keyHandled = LCommander::HandleKeyPress( inKeyEvent );
	}
	else
	{
		KeyMap theKeyState;

		do
		{
			keyHandled 	= false;
			::GetKeys( theKeyState );

			if ( TestKeyBit(theKeyState, 0x00) ) {	// 'A'
				CameraPan(3.0);
				keyHandled = true;
			}
							
			if ( TestKeyBit(theKeyState, 0x02) ) {	// 'D'
				CameraPan(-3.0);
				keyHandled = true;
			}

			if ( TestKeyBit(theKeyState, 0x0C) ) {	// 'Q'
				CameraOrbit(-3.0);
				keyHandled = true;
			}
				
			if ( TestKeyBit(theKeyState, 0x0E) ) {	// 'E'
				CameraOrbit(3.0);
				keyHandled = true;
			}

			if ( TestKeyBit(theKeyState, 0x0F) ) {	// 'R'
				CameraOrbit(0.0, 3.0);
				keyHandled = true;
			}
				
			if ( TestKeyBit(theKeyState, 0x03) ) {	// 'F'
				CameraOrbit(0.0, -3.0);
				keyHandled = true;
			}

			if ( TestKeyBit(theKeyState, 0x08) ) {	// 'C'
				CameraZoom(0.3);
				keyHandled = true;
			}
				
			if ( TestKeyBit(theKeyState, 0x09) ) {	// 'V'
				CameraZoom(-0.3);
				keyHandled = true;
			}
				
			if ( TestKeyBit(theKeyState, 0x0D) ) {	// 'W'
				CameraMove(0.3, 0.0);
				keyHandled = true;
			}
				
			if ( TestKeyBit(theKeyState, 0x01) ) {	// 'S'
				CameraMove(-0.3, 0.0);
				keyHandled = true;
			}
				
			if ( TestKeyBit(theKeyState, 0x06) ) {	// 'Z'
				CameraMove(0.0, 0.3);
				keyHandled = true;
			}
				
			if ( TestKeyBit(theKeyState, 0x07) ) {	// 'X'
				CameraMove(0.0, -0.3);
				keyHandled = true;
			}
			 
			if ( keyHandled )
				this->DrawSelf();
		}
		while ( keyHandled );
	}
	
	return keyHandled;
}


void C3DMapView::CameraOrbit(float inDegrees, float inVerticalDegrees)
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera(mQD3DView, &theCamera);
	ThrowIf_(status == kQ3Failure);

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement(theCamera, &theCameraPlacement);
	ThrowIf_(status == kQ3Failure);

	float xAngle = 0.0;
	float yAngle = Q3Math_DegreesToRadians(inDegrees);
	float zAngle = 0.0;
	
	TQ3Matrix4x4 cameraMatrix;
	::Q3Matrix4x4_SetRotateAboutPoint(&cameraMatrix, &theCameraPlacement.pointOfInterest, xAngle, yAngle, zAngle);
	
	::Q3Point3D_Transform(&theCameraPlacement.cameraLocation, &cameraMatrix, &theCameraPlacement.cameraLocation);

	if (inVerticalDegrees != 0.0)
	{
		float verticalAngle = Q3Math_DegreesToRadians(inVerticalDegrees);
		TQ3Vector3D theCameraVector;
		::Q3Point3D_Subtract(&theCameraPlacement.pointOfInterest, &theCameraPlacement.cameraLocation, &theCameraVector);
		
		TQ3Vector3D theOrthogonalVector;
		theOrthogonalVector.x = theCameraVector.z;
		theOrthogonalVector.y = 0.0;
		theOrthogonalVector.z = 0.0 - theCameraVector.x;
		::Q3Vector3D_Normalize(&theOrthogonalVector, &theOrthogonalVector);

		TQ3Matrix4x4 verticalMatrix;
		::Q3Matrix4x4_SetRotateAboutAxis(&verticalMatrix, &theCameraPlacement.pointOfInterest, &theOrthogonalVector, verticalAngle);

		::Q3Point3D_Transform(&theCameraPlacement.cameraLocation, &verticalMatrix, &theCameraPlacement.cameraLocation);
	}

	status = ::Q3Camera_SetPlacement(theCamera, &theCameraPlacement);
	ThrowIf_(status == kQ3Failure);
	
	::Q3Object_Dispose(theCamera);
}


void C3DMapView::CameraPan(float inDegrees)
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera(mQD3DView, &theCamera);
	ThrowIf_(status == kQ3Failure);

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement(theCamera, &theCameraPlacement);
	ThrowIf_(status == kQ3Failure);

	float xAngle = 0.0;
	float yAngle = Q3Math_DegreesToRadians(inDegrees);
	float zAngle = 0.0;
	
	TQ3Matrix4x4 cameraMatrix;
	::Q3Matrix4x4_SetRotateAboutPoint(&cameraMatrix, &theCameraPlacement.cameraLocation, xAngle, yAngle, zAngle);
	::Q3Point3D_Transform(&theCameraPlacement.pointOfInterest, &cameraMatrix, &theCameraPlacement.pointOfInterest);

	status = ::Q3Camera_SetPlacement(theCamera, &theCameraPlacement);
	ThrowIf_(status == kQ3Failure);
	
	::Q3Object_Dispose(theCamera);
}


void C3DMapView::CameraZoom(float inDistance)
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera(mQD3DView, &theCamera);
	ThrowIf_(status == kQ3Failure);

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement(theCamera, &theCameraPlacement);
	ThrowIf_(status == kQ3Failure);

	TQ3Vector3D theCameraVector;
	::Q3Point3D_Subtract(&theCameraPlacement.pointOfInterest, &theCameraPlacement.cameraLocation, &theCameraVector);
	
	float theCameraDistance = ::Q3Vector3D_Length(&theCameraVector);
	
	if (inDistance < theCameraDistance)
	{
		::Q3Vector3D_Normalize(&theCameraVector, &theCameraVector);
		::Q3Vector3D_Scale(&theCameraVector, inDistance, &theCameraVector);
		
		::Q3Point3D_Vector3D_Add(&theCameraPlacement.cameraLocation, &theCameraVector, &theCameraPlacement.cameraLocation);
		
		status = ::Q3Camera_SetPlacement(theCamera, &theCameraPlacement);
		ThrowIf_(status == kQ3Failure);
	
		::Q3Object_Dispose(theCamera);
	}
}


void C3DMapView::CameraMove(float inDistance, float inOrthogonalDistance)
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera(mQD3DView, &theCamera);
	ThrowIf_(status == kQ3Failure);

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement(theCamera, &theCameraPlacement);
	ThrowIf_(status == kQ3Failure);

	TQ3Vector3D theCameraVector;
	::Q3Point3D_Subtract(&theCameraPlacement.pointOfInterest, &theCameraPlacement.cameraLocation, &theCameraVector);
	theCameraVector.y = 0.0;
	
	TQ3Vector3D theOrthogonalVector;
	::Q3Vector3D_Normalize(&theCameraVector, &theCameraVector);
	theOrthogonalVector.x = theCameraVector.z;
	theOrthogonalVector.y = 0.0;
	theOrthogonalVector.z = 0.0 - theCameraVector.x;
	
	::Q3Vector3D_Scale(&theCameraVector, inDistance, &theCameraVector);
	::Q3Vector3D_Scale(&theOrthogonalVector, inOrthogonalDistance, &theOrthogonalVector);
	::Q3Vector3D_Add(&theCameraVector, &theOrthogonalVector, &theCameraVector);
		
	::Q3Point3D_Vector3D_Add(&theCameraPlacement.pointOfInterest, &theCameraVector, &theCameraPlacement.pointOfInterest);
	::Q3Point3D_Vector3D_Add(&theCameraPlacement.cameraLocation, &theCameraVector, &theCameraPlacement.cameraLocation);
		
	status = ::Q3Camera_SetPlacement(theCamera, &theCameraPlacement);
	ThrowIf_(status == kQ3Failure);
	
	::Q3Object_Dispose(theCamera);
}