// =================================================================================
//	CGorFile.cp				1997 eli curtz. All rights reserved.
// =================================================================================

#include <LArray.h>

#include "CGorFile.h"


// LTagComparator

// ---------------------------------------------------------------------------
//		 Compare
// ---------------------------------------------------------------------------
//	Compare to group/subGroup tags

Int32
LTagComparator::Compare(
	const void*		inItemOne,
	const void*		inItemTwo,
	Uint32			/* inSizeOne */,
	Uint32			/* inSizeTwo */) const
{
	dir_Header * dirHeader1 = (dir_Header *) inItemOne;
	dir_Header * dirHeader2 = (dir_Header *) inItemTwo;

	if (dirHeader1->group != dirHeader2->group)
		return ( dirHeader1->group - dirHeader2->group );
	else
		return ( dirHeader1->subGroup - dirHeader2->subGroup );
}


// CGorFile

// ---------------------------------------------------------------------------
//		 CGorFile
// ---------------------------------------------------------------------------
//	Default Contructor

CGorFile::CGorFile()
{
	mSegmentList = nil;
	mInitialized = false;
}


// ---------------------------------------------------------------------------
//		 CGorFile(FSSpec&)
// ---------------------------------------------------------------------------
//	Contruct a GorFile from a Toolbox File System Specification

CGorFile::CGorFile(
	const FSSpec	&inFileSpec)
		: LFileStream(inFileSpec)
{
	mSegmentList = nil;
	mInitialized = false;
	
	Initialize();
}


// ---------------------------------------------------------------------------
//		 CGorFile(AliasHandle, Boolean, FSSpec*)
// ---------------------------------------------------------------------------
//	Contruct a GorFile from an Alias

CGorFile::CGorFile(
	AliasHandle	inAlias,
	Boolean		&outWasChanged,
	FSSpec		*inFromFile)
		: LFileStream(inAlias, outWasChanged, inFromFile)
{
	mSegmentList = nil;
	mInitialized = false;
	
	Initialize();
}


// ---------------------------------------------------------------------------
//		 Initialize
// ---------------------------------------------------------------------------
//	Set up the header and segment list

void
CGorFile::Initialize()
{
	Int32 byteCount = sizeof(gor_Header);
	
	OpenDataFork(fsRdPerm);
	
	ThrowIfOSErr_(GetBytes(&mHeader, byteCount));
	
	SetMarker(mHeader.tagHeaderOffset, streamFrom_Start);
	byteCount = mHeader.headerCount * sizeof(dir_Header);
	Handle theDirHeaders = ::NewHandle(byteCount);
	ThrowIfMemFail_(theDirHeaders);
	{
		StHandleLocker lock(theDirHeaders);
		ThrowIfOSErr_(GetBytes(*theDirHeaders, byteCount));
	}
	
	mSegmentList = new LArray(sizeof(dir_Header), theDirHeaders, new LTagComparator(), false, true);

	CloseDataFork();
	mInitialized = true;
}


// ---------------------------------------------------------------------------
//		 ~CGorFile
// ---------------------------------------------------------------------------
//	Destructor

CGorFile::~CGorFile()
{
	delete mSegmentList;
}

// ---------------------------------------------------------------------------
//		 SegmentExists
// ---------------------------------------------------------------------------
//	Confirm a particular segment is in the file

Boolean
CGorFile::SegmentExists(
	FourCharCode inGroup,
	FourCharCode inSubGroup)
{
	Boolean segmentFound = false;
	if (mInitialized) {
		// look for the item in the array
		dir_Header scratchHeader;
		scratchHeader.group = inGroup;
		scratchHeader.subGroup = inSubGroup;
		ArrayIndexT segmentIndex = mSegmentList->FetchIndexOf(&scratchHeader);
		segmentFound = mSegmentList->ValidIndex(segmentIndex);
	}
	
	return segmentFound;
}


// ---------------------------------------------------------------------------
//		 GetSegmentHeader
// ---------------------------------------------------------------------------
//	Copy the segment header to outHeader (unchanged if not found), return whether header was found

Boolean
CGorFile::GetSegmentHeader(
	FourCharCode inGroup,
	FourCharCode inSubGroup,
	dir_Header* outHeader)
{
	Boolean headerFound = false;
	if (mInitialized) {
		// look for the item in the array
		dir_Header scratchHeader;
		scratchHeader.group = inGroup;
		scratchHeader.subGroup = inSubGroup;
		ArrayIndexT segmentIndex = mSegmentList->FetchIndexOf(&scratchHeader);
		if (mSegmentList->ValidIndex(segmentIndex)) {
			headerFound = mSegmentList->FetchItemAt(segmentIndex, outHeader);
		}
	}
	
	return headerFound;
}


// ---------------------------------------------------------------------------
//		 GetSegment
// ---------------------------------------------------------------------------
//	Return a segment (nil if segment does not exist)

Handle
CGorFile::GetSegment(
	FourCharCode inGroup,
	FourCharCode inSubGroup)
{
	Handle segmentData = nil;

	if (mInitialized) {
		// look for the item in the array
		dir_Header scratchHeader;
		scratchHeader.group = inGroup;
		scratchHeader.subGroup = inSubGroup;
		ArrayIndexT segmentIndex = mSegmentList->FetchIndexOf(&scratchHeader);
		// build the handle and read the segment data
		if (mSegmentList->ValidIndex(segmentIndex)) {
			dir_Header segmentHeader;
			Assert_(mSegmentList->FetchItemAt(segmentIndex, &segmentHeader));
			
			Int32 byteCount = segmentHeader.dataSize;
			segmentData = ::NewHandle(sizeof(dir_Header) + byteCount);
			ThrowIfMemFail_(segmentData);
			StHandleLocker lock(segmentData);
			::BlockMoveData(&segmentHeader, *segmentData, sizeof(dir_Header));

			Ptr dataPtr = *segmentData + sizeof(dir_Header);
			OpenDataFork(fsRdPerm);
			GetBytes(dataPtr, byteCount);
			CloseDataFork();
		}
	}
	
	return segmentData;
}

