// ===========================================================================
//	UQuickDraw3D.cp			   1997 eli curtz. All rights reserved.
// ===========================================================================

#ifdef PowerPlant_PCH
#include PowerPlant_PCH
#endif

#include "UQuickDraw3D.h"
#include <UEnvironment.h>

#ifndef __GESTALT__
#include <Gestalt.h>
#endif


#pragma mark === UQuickDraw3D ===

Boolean UQuickDraw3D::sQuickDraw3DPresent = false;

// ---------------------------------------------------------------------------
//	 QuickDraw3DIsPresent									 [static] [public]
// ---------------------------------------------------------------------------
//	Is QuickDraw 3D present?  Initialize must have been called prior to this.

Boolean
UQuickDraw3D::QuickDraw3DIsPresent()
{
	return sQuickDraw3DPresent;	// rather than add a new flag to UEnvironment
}


// ---------------------------------------------------------------------------
//	 Initialize											 [static] [public]
// ---------------------------------------------------------------------------
//	Initialize QuickDraw 3D if it is present

void
UQuickDraw3D::Initialize()
{
	Int32	qd3DVersion;
	Boolean	qd3DPresent = (::Gestalt(gestaltQD3D, &qd3DVersion) == noErr)
							&& (qd3DVersion >= gestaltQD3DPresent)
							&& CFM_AddressIsResolved_(::Q3Initialize);
	
	if (qd3DPresent) {
		sQuickDraw3DPresent = true;
		TQ3Status qd3DErr = ::Q3Initialize();
		ThrowIf_(qd3DErr == kQ3Failure);
	}
}


// ---------------------------------------------------------------------------
//	 Finalize												 [static] [public]
// ---------------------------------------------------------------------------
//	Perform any necessary cleanup

void
UQuickDraw3D::Finalize()
{
	if (QuickDraw3DIsPresent()) {
		::Q3Exit();
		sQuickDraw3DPresent = false;
	}
}

#pragma mark === LQD3DViewer ===

// ---------------------------------------------------------------------------
//	 LQD3DViewer						Default Constructor		  [public]
// ---------------------------------------------------------------------------

LQD3DViewer::LQD3DViewer()
{
	mViewerObject = nil;

	mSourceFile = nil;
	mSourceData = nil;
}


// ---------------------------------------------------------------------------
//	 LQD3DViewer						Constructor				  [public]
// ---------------------------------------------------------------------------

LQD3DViewer::LQD3DViewer(LStream *inStream)
		: LPane(inStream)
{
	mViewerObject = nil;

	mSourceFile = nil;
	mSourceData = nil;
}


// ---------------------------------------------------------------------------
//	 ~LQD3DViewer						Destructor				  [public]
// ---------------------------------------------------------------------------
//	Destructor

LQD3DViewer::~LQD3DViewer()
{
	if (mViewerObject) {
		::Q3ViewerDispose(mViewerObject);
		mViewerObject = nil;
	}
	if (mSourceFile) {
		delete mSourceFile;
		mSourceFile = 0;
	}
	if (mSourceData) {
		::DisposeHandle(mSourceData);
		mSourceData = nil;
	}
}

// ----------------------------------------------------------------------------
//   FinishCreateSelf												[public]
// ----------------------------------------------------------------------------
// Create the viewer object

void
LQD3DViewer::FinishCreateSelf()	
{
	LPane::FinishCreateSelf();

	FocusDraw();
	Rect theRect;
	CalcPortFrameRect(theRect);

	mViewerObject = ::Q3ViewerNew((CGrafPtr) GetMacPort(), &theRect, kQ3ViewerDefault);
	if (mViewerObject == nil)	{
		SignalPStr_("\pCouldn't create viewer");
	}
	
	ResizeFrameBy(0, 0, false);	// Handle strange QD3D origin bug
}

// ---------------------------------------------------------------------------
//	 DrawSelf													   [public]
// ---------------------------------------------------------------------------
//	Draw the viewer

void
LQD3DViewer::DrawSelf()
{
	if (mViewerObject)
		::Q3ViewerDraw(mViewerObject);
}


// ----------------------------------------------------------------------------
//	 AdjustCursorSelf												[public]
// ----------------------------------------------------------------------------

void
LQD3DViewer::AdjustCursorSelf(Point inPortPt, const EventRecord& /*inMacEvent*/)
{
	if (mViewerObject)
		::Q3ViewerAdjustCursor(mViewerObject, &inPortPt);
}


// ----------------------------------------------------------------------------
//	 ResizeFrameBy												[public]
// ----------------------------------------------------------------------------

void
LQD3DViewer::ResizeFrameBy(Int16 inWidthDelta, Int16 inHeightDelta, Boolean inRefresh)
{
	LPane::ResizeFrameBy(inWidthDelta, inHeightDelta, inRefresh);
	if (mViewerObject) {
		Rect theRect;
		CalcPortFrameRect(theRect);
		
		::Q3ViewerSetBounds(mViewerObject, &theRect);
		::Q3ViewerDraw(mViewerObject);
	}
}


// ----------------------------------------------------------------------------
//	 ClickSelf												[public]
// ----------------------------------------------------------------------------

void
LQD3DViewer::ClickSelf(const SMouseDownEvent &inMouseDown)	
{
	HandleViewerEvent((EventRecord &) inMouseDown.macEvent);
}


// ----------------------------------------------------------------------------
//	 SetSourceFile											[public]
// ----------------------------------------------------------------------------

void
LQD3DViewer::SetSourceFile(LFile*  inSourceFile)
{
	if (mViewerObject && inSourceFile) {
		mSourceFile = new LFile(*inSourceFile);
		ThrowIfNil_(mSourceFile);
		mSourceFile->OpenDataFork(fsRdPerm);
		
		ThrowIfOSErr_(::Q3ViewerUseFile(mViewerObject, mSourceFile->GetDataForkRefNum()));
	}
}


// ----------------------------------------------------------------------------
//	 SetSourceData											[public]
// ----------------------------------------------------------------------------

void
LQD3DViewer::SetSourceData(Handle /*inSourceData*/)
{
}


// ----------------------------------------------------------------------------
//	 SetGroup											[public]
// ----------------------------------------------------------------------------

void
LQD3DViewer::SetGroup(TQ3GroupObject inGroup)
{
	if (mViewerObject && inGroup) {
		ThrowIfOSErr_(::Q3ViewerUseGroup(mViewerObject, inGroup));
	}
}


// ----------------------------------------------------------------------------
//	 GetView											[public]
// ----------------------------------------------------------------------------

TQ3ViewObject
LQD3DViewer::GetView()
{
	TQ3ViewObject theViewObject = nil;

	if (mViewerObject)
		theViewObject = ::Q3ViewerGetView(mViewerObject);
	
	return theViewObject;
}


// ----------------------------------------------------------------------------
//	 SaveToFile											[public]
// ----------------------------------------------------------------------------

void
LQD3DViewer::SaveToFile(LFile* /*inDestinationFile*/)
{
}


// ----------------------------------------------------------------------------
//	 HandleViewerEvent											[protected]
// ----------------------------------------------------------------------------

Boolean
LQD3DViewer::HandleViewerEvent(EventRecord	&inMacEvent)	
{
	if (mViewerObject)
		return ::Q3ViewerEvent(mViewerObject, &inMacEvent);
	else
		return false;
}

