// =================================================================================
// CTextureManager.cp
// =================================================================================


#include "CTextureManager.h"
#include "CTagManager.h"
#include "bitmaps.h"
#include "CProgressDialog.h"
#include "CPhotoshopFile.h"


const	Int32	kImageSize	= 65588;	// 256 * 256 image + header


// -----------------------------------------------------------------------------
// CTextureManager
// -----------------------------------------------------------------------------

CTextureManager::CTextureManager()
{
	fTextureTagID	= 0xffff;
	fSubmeshWidth	= 0;
	fSubmeshHeight	= 0;
	fColorTable		= NULL;
	fBlendTable		= NULL;
	fRemapTable		= NULL;
	
	// Listen for out-of-memory messages
	LGrowZone::GetGrowZone()->AddListener( this );
}


// -----------------------------------------------------------------------------
// ~CTextureManager
// -----------------------------------------------------------------------------

CTextureManager::~CTextureManager()
{
	this->UnloadAllImages();
	this->UnloadColorTable();
	this->UnloadBlendTable();
	this->UnloadRemapTable();
	
	CTagManager::DiscardTag( COLLECTION_GROUP_TAG, fTextureTagID );
}


// -----------------------------------------------------------------------------
// ListenToMessage
// -----------------------------------------------------------------------------

void CTextureManager::ListenToMessage( MessageT inMessage, void *ioParam )
{
	Int32 theFreedSize = 0;
		
	if ( inMessage == msg_GrowZone )
	{
		while ( theFreedSize < *((Int32*) ioParam) )
		{
			ArrayIndexT	theOldestIndex	= LArray::index_Bad;
			Int32		theOldestTime	= 0x7fffffff;
			
			// Scan the array of bitmaps to find the one used least recently
			for ( ArrayIndexT i = LArray::index_First; i <= fBitmapArray.GetCount(); i++ )
			{
				if ( fBitmapArray[i].fTimestamp < theOldestTime )
				{
					theOldestIndex	= i;
					theOldestTime	= fBitmapArray[i].fTimestamp;
				}
			}
			
			// Dispose of the old image I just found
			if ( theOldestIndex != LArray::index_Bad )
			{
				theFreedSize += ::GetHandleSize( (Handle) fBitmapArray[theOldestIndex].fImage );
				::DisposeHandle( (Handle) fBitmapArray[theOldestIndex].fImage );
				fBitmapArray[theOldestIndex].fImage = NULL;
				fBitmapArray[theOldestIndex].fTimestamp = 0x7fffffff;
			}
			else
			{
				// I've freed all I can, so let someone else try
				*((Int32*) ioParam) = theFreedSize;
				return;
			}
		}
	}
	
	// Report the amount of freed memory
	*((Int32*) ioParam) = theFreedSize;
}


// -----------------------------------------------------------------------------
// BringToFront
// -----------------------------------------------------------------------------

void CTextureManager::BringToFront()
{
	// A window displaying this texture has been activated, so make sure the image
	// chunks I have loaded will be the last to be considered for disposal in the
	// event of a low memory situation by setting myself as the last listener.
	LGrowZone::GetGrowZone()->RemoveListener( this );
	LGrowZone::GetGrowZone()->AddListener( this );
}


// -----------------------------------------------------------------------------
// Open
// -----------------------------------------------------------------------------

void CTextureManager::Open( FourCharCode inTextureTagID, Int16 inSubmeshWidth, Int16 inSubmeshHeight )
{
	fTextureTagID	= inTextureTagID;
	fSubmeshWidth	= inSubmeshWidth;
	fSubmeshHeight	= inSubmeshHeight;
	
	// Prepare to read the texture from an existing tag file
	this->ReadHeader();
	this->ReadBitmapArray();

	fIsDirty		= false;
}


// -----------------------------------------------------------------------------
// New
// -----------------------------------------------------------------------------

void CTextureManager::New( FourCharCode		inTextureTagID,
						   TString<Str31>	inTagName,
						   Int16			inSubmeshWidth,
						   Int16			inSubmeshHeight )
{
	StProgressDialog	theProgressDialog( 1010, NULL );
	theProgressDialog.SetProgressPrompt( "\pCreating Blank Landscape Texture Map" );
	theProgressDialog.SetProgressBarMax( inSubmeshWidth * inSubmeshHeight * 2 + 5 );
	theProgressDialog.Show();
	
	CTagManager::CreateTag( COLLECTION_GROUP_TAG, inTextureTagID, inTagName, _collection_is_color_map_flag );
	
	fTextureTagID	= inTextureTagID;
	fSubmeshWidth	= inSubmeshWidth;
	fSubmeshHeight	= inSubmeshHeight;
	
	// Write a blank header as a placeholder
	{
		StResource	theHeaderResource( 'defa', 2000 );
		fHeader		= **((collection_header**)(Handle) theHeaderResource);
		fHeader.user_data_flags				= _collection_is_color_map_flag;
		fHeader.data_offset					= SIZEOF_STRUCT_COLLECTION_HEADER;
		fHeader.data_size					= 0;
		
		CTagManager::WriteTagHeader( COLLECTION_GROUP_TAG, fTextureTagID, &fHeader );
		
		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
	}
	
	// Write the color table
	{
		// Load and write the default color table
		StResource	theColorTableResource( 'defa', 2001 );
		CTagManager::WriteTagData( COLLECTION_GROUP_TAG, fTextureTagID,
								   fHeader.data_size,
								   (Handle) theColorTableResource );
		
		// Update the header with the offset and size of the color table
		fHeader.color_table_count			=  1;
		fHeader.color_tables_offset			=  fHeader.data_size;
		fHeader.color_tables_size			=  ::GetHandleSize( (Handle) theColorTableResource );
		fHeader.data_size					+= fHeader.color_tables_size;

		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
	}
	
	// Write the blend table
	{
		// Load a fraction of the default blend table from a resource
		StResource		theBlendTableResource( 'defa', 2002 );
		StHandleLocker	lock( theBlendTableResource );
		
		// Make a handle stream to hold the blend table
		LHandleStream	theBlendTableHandle;
		theBlendTableHandle.SetLength( 256 * ::GetHandleSize( theBlendTableResource ) );
		theBlendTableHandle.SetMarker( 0, streamFrom_Start );
		
		// We've loaded 1/256th of the blend table, so write it 256 times
		for ( Int32 i = 0; i < 256; i++ )
		{
			theBlendTableHandle.WriteBlock( *((Handle) theBlendTableResource),
											::GetHandleSize( theBlendTableResource ) );
		}

		CTagManager::WriteTagData( COLLECTION_GROUP_TAG, inTextureTagID,
								   fHeader.data_size,
								   theBlendTableHandle.GetDataHandle() );

		fHeader.blend_table_count			=  1;
		fHeader.blend_table_offset			=  fHeader.data_size;
		fHeader.blend_table_size			=  ::GetHandleSize( theBlendTableHandle.GetDataHandle() );
		fHeader.data_size					+= fHeader.blend_table_size;

		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
	}
	
	// Write the remapping table	
	{
		// Load the default blood remapping table from a resource
		StResource	theRemappingTableResource( 'defa', 2003 );
		CTagManager::WriteTagData( COLLECTION_GROUP_TAG, inTextureTagID,
								   fHeader.data_size,
								   (Handle) theRemappingTableResource );
		
		// Update the header with the offset and size of the remapping table			   
		fHeader.remapping_table_count		=  1;
		fHeader.remapping_table_offset		=  fHeader.data_size;
		fHeader.remapping_table_size		=  ::GetHandleSize( (Handle) theRemappingTableResource );
		fHeader.data_size					+= fHeader.remapping_table_size;

		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
	}
	
	// Write the blank bitmap references
	{	
		fHeader.bitmap_count				= fSubmeshWidth * fSubmeshHeight * 2;
		fHeader.bitmap_references_offset	= fHeader.data_size;
		fHeader.bitmap_references_size		= 0;
		
		StResource	theBitmapRefResource( 'defa', 2004 );

		for ( Int32 i = 0; i < fSubmeshWidth * fSubmeshHeight * 2; i++ )
		{
			CTagManager::WriteTagData( COLLECTION_GROUP_TAG, inTextureTagID,
									   fHeader.data_size,
									   (Handle) theBitmapRefResource );
									   
			fHeader.bitmap_references_size	+= ::GetHandleSize( (Handle) theBitmapRefResource );
			fHeader.data_size				+= ::GetHandleSize( (Handle) theBitmapRefResource );
		}

		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
	}
	
	// Write the bitmaps
	{
		fHeader.bitmap_instance_count		= fSubmeshWidth * fSubmeshHeight * 2;
		fHeader.bitmap_instances_offset		= fHeader.data_size;
		fHeader.bitmap_instances_size		= 0;
		
		StResource	theBitmapResource( 'defa', 2005 );
		
		for ( Int32 i = 0; i < fHeader.bitmap_instance_count; i++ )
		{
			bitmap_reference	theBitmapRef;
			
			CTagManager::ReadTagData( COLLECTION_GROUP_TAG, inTextureTagID,
									  fHeader.bitmap_references_offset
									  + ( i * SIZEOF_STRUCT_BITMAP_REFERENCE ),
									  SIZEOF_STRUCT_BITMAP_REFERENCE,
									  &theBitmapRef );
									  
			theBitmapRef.offset	= fHeader.data_size;
			theBitmapRef.size	= ::GetHandleSize( (Handle) theBitmapResource );
									  
			CTagManager::WriteTagData( COLLECTION_GROUP_TAG, inTextureTagID,
									  fHeader.bitmap_references_offset
									  + ( i * SIZEOF_STRUCT_BITMAP_REFERENCE ),
									  SIZEOF_STRUCT_BITMAP_REFERENCE,
									  &theBitmapRef );
									  
			CTagManager::WriteTagData( COLLECTION_GROUP_TAG, inTextureTagID,
									  fHeader.data_size,
									  (Handle) theBitmapResource );
									  									  
			fHeader.bitmap_instances_size	+= theBitmapRef.size;
			fHeader.data_size				+= theBitmapRef.size;

		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
		}
	}
	
	CTagManager::WriteTagHeader( COLLECTION_GROUP_TAG, inTextureTagID, &fHeader );
	this->ReadBitmapArray();
	this->Changed();
}


// -----------------------------------------------------------------------------
// SaveAs
// -----------------------------------------------------------------------------

void CTextureManager::SaveAs( FourCharCode inTextureTagID, TString<Str31> inTextureName )
{
	CTagManager::CreateTag( COLLECTION_GROUP_TAG, fTextureTagID, inTextureTagID, inTextureName );
	
	// Update the storage tag ID and do a normal save
	fTextureTagID = inTextureTagID;
	this->Save();	
}


// -----------------------------------------------------------------------------
// Save
// -----------------------------------------------------------------------------

void CTextureManager::Save()
{
	if ( this->IsDirty() )
	{
		// Scan the bitmap array looking for any unsaved images
		for ( ArrayIndexT i = LArray::index_First; i <= fBitmapArray.GetCount(); i++ )
		{
			if ( fBitmapArray[i].fIsDirty )
			{
				// Found one, so write it out & mark it as saved
				this->WriteImage( fBitmapArray[i].fBitmapRef.offset, fBitmapArray[i].fImage );
				fBitmapArray[i].fIsDirty = false;
			}
		}
		
		// Now ask the tag manager to move the temp tag file to its permanent location
		CTagManager::IncrementTagVersion( COLLECTION_GROUP_TAG, fTextureTagID );
		CTagManager::SaveTag( COLLECTION_GROUP_TAG, fTextureTagID );
		fIsDirty = false;
	}
}


// -----------------------------------------------------------------------------
// IsDirty
// -----------------------------------------------------------------------------
/*
Boolean CTextureManager::IsDirty()
{
	// Scan the bitmap array looking for dirty images
	for ( ArrayIndexT i = LArray::index_First; i <= fBitmapArray.GetCount(); i++ )
	{
		if ( fBitmapArray[i].fIsDirty )
			return true;
	}
	
	// Couldn't find any (or we wouldn't have gotten here)
	return false;
}
*/


// -----------------------------------------------------------------------------
// ImportTextureMap
// -----------------------------------------------------------------------------

void CTextureManager::ImportTextureMap( FSSpec &inFileSpec )
{
	try
	{
		// Display a progress dialog
		StProgressDialog	theProgressDialog( 1010, NULL );
		theProgressDialog.SetProgressPrompt( "\pImporting Landscape Texture Map" );
		theProgressDialog.Show();

		this->UnloadAllImages();
		this->UnloadColorTable();
		this->UnloadBlendTable();
		this->UnloadRemapTable();
		this->ImportImages( inFileSpec, true, &theProgressDialog );
		this->BuildRemappingTable( this->GetColorTable(), &theProgressDialog );
		this->BuildBlendTable( this->GetColorTable(), &theProgressDialog );
		this->Changed();
	}
	catch ( ExceptionCode theErr )
	{
	}
}



// -----------------------------------------------------------------------------
// ImportShadowMap
// -----------------------------------------------------------------------------

void CTextureManager::ImportShadowMap( FSSpec &inFileSpec )
{
	try
	{
		StProgressDialog	theProgressDialog( 1010, NULL );
		theProgressDialog.SetProgressPrompt( "\pImporting Landscape Shadow Map" );
		theProgressDialog.Show();

		this->UnloadAllImages();
		this->UnloadColorTable();
		this->UnloadBlendTable();
		this->UnloadRemapTable();
		this->ImportImages( inFileSpec, false, &theProgressDialog );
		this->Changed();
	}
	catch ( ExceptionCode theErr )
	{
	}
}


// -----------------------------------------------------------------------------
// ImportImages
// -----------------------------------------------------------------------------

void CTextureManager::ImportImages( FSSpec				&inFileSpec,
									Boolean				inIsLandscape,
									StProgressDialog	*inProgress )
{
	Assert_( fBitmapArray.GetCount() >= fSubmeshWidth * fSubmeshHeight * 2 );
	
	// Are we displaying a progress dialog?
	if ( inProgress )
	{
		inProgress->SetProgressBarMax( fSubmeshWidth * fSubmeshHeight );
	}
	
	GraphicsImportComponent	theImporter		= NULL;
	ImageDescriptionHandle	theImageDesc	= NULL;
	GWorldPtr				theGWorld		= NULL;
		
	try
	{
		// Open a graphics importer for this file
		ThrowIfOSErr_( ::GetGraphicsImporterForFile( &inFileSpec, &theImporter ) );
		ThrowIfNULL_( theImporter );
		
		// Get some vital stats for this image
		ThrowIfOSErr_( ::GraphicsImportGetImageDescription( theImporter, &theImageDesc ) );
		ThrowIfNULL_( theImageDesc );
		
		// Sanity check to make sure the image is the correct size
		//  replace this with a nice error dialog
		if ( theImageDesc[0]->width != fSubmeshWidth * 256 || theImageDesc[0]->height != fSubmeshHeight * 256 )
		{
			SignalCStr_( "The image size is not the same size as the mesh." );
			Throw_( noErr );
		}
		
		if ( theImageDesc[0]->depth != 8 )
		{
			SignalCStr_( "The image is not indexed 8-bit depth." );
			Throw_( noErr );
		}
		
		theImageDesc[0]->width	= 256;
		theImageDesc[0]->height	= 256;
		
		ThrowIfOSErr_( ::NewImageGWorld( &theGWorld, theImageDesc, 0 ) );

		// Create a temporary GWorld and imported image space
		PixMapHandle	theGWorldPixMap	= ::GetGWorldPixMap( theGWorld );
		Ptr				theGWorldPixH	= theGWorldPixMap[0]->baseAddr;
		StHandleBlock	theRawImageH( 256 * 256 );
		
		// Set up to draw the image into my GWorld
		ThrowIfOSErr_( ::GraphicsImportSetGWorld( theImporter, theGWorld, NULL ) );
		ThrowIfOSErr_( ::GraphicsImportSetGraphicsMode( theImporter, srcCopy, NULL ) );

		// 
		Int32			theRowCounter	= 0;
		Int32			theColCounter	= fSubmeshWidth - 1;
		MatrixRecord	theMatrix;
		
		// Loop through all the submeshes (256 * 256 pixel chuncks)
		for ( Int32 i = 1; i <= fSubmeshWidth * fSubmeshHeight; i++ )
		{
			// Calculate rect of the chunk we're interested in
			Rect	theSourceRect = { theRowCounter * 256, theColCounter * 256,
									  theSourceRect.top + 256, theSourceRect.left + 256 };
			
			// Calculate the deltas to move the image to the port's origin	
			Fixed	deltaH = 0 - ( theSourceRect.left << 16 );
			Fixed	deltaV = 0 - ( theSourceRect.top << 16 );
			
			// Build the translation matrix
			::SetIdentityMatrix( &theMatrix );
			::TranslateMatrix( &theMatrix, deltaH, deltaV );
			
			// Set up the importer to draw a chunk of the image at the port's origin
			ThrowIfOSErr_( ::GraphicsImportSetSourceRect( theImporter, &theSourceRect ) );
			ThrowIfOSErr_( ::GraphicsImportSetMatrix( theImporter, &theMatrix ) );
			
			// Set up the GWorld to use our (correctly sized) pixel storage
			theGWorldPixMap[0]->baseAddr = (Ptr) (Handle) theRawImageH;
			theGWorldPixMap[0]->rowBytes = 0x8000 | 256;
			
			// All set; now draw that sucker!
			ThrowIfOSErr_( ::GraphicsImportDraw( theImporter ) );
			
			// All done, save the imported image chunk
			SBitmapRef	theBitmapRef	 = fBitmapArray[(i*2)-inIsLandscape];
			this->WriteImage( theBitmapRef.fBitmapRef.offset, theRawImageH );
			
			// Set up for the next chunk
			theColCounter--;
			if ( theColCounter < 0 )
			{
				theRowCounter++;
				theColCounter = fSubmeshWidth - 1;
			}
			
			// Update the progress dialog
			if ( inProgress ) inProgress->IncrementProgressBar();
		}

		// If we are importing a landscape texture, extract the color table
		if ( inIsLandscape )
		{
			this->WriteColorTable( theGWorld->portPixMap[0]->pmTable );
		}
		
		// Tidy up...
		theGWorldPixMap[0]->baseAddr = theGWorldPixH;
		::DisposeGWorld( theGWorld );
		::DisposeHandle( (Handle) theImageDesc );
		::CloseComponent( theImporter );
	}
	catch ( ExceptionCode theErr )
	{
		if ( theGWorld )
			::DisposeGWorld( theGWorld );
			
		if ( theImageDesc )
			::DisposeHandle( (Handle) theImageDesc );
			
		if ( theImporter )
			::CloseComponent( theImporter );
		
		Throw_( theErr );
	}
}


// -----------------------------------------------------------------------------
// ExportTextureMap
// -----------------------------------------------------------------------------

void CTextureManager::ExportTextureMap( FSSpec &inFileSpec )
{
	// Display a progress dialog
	StProgressDialog	theProgressDialog( 1010, NULL );
	theProgressDialog.SetProgressPrompt( "\pExporting Landscape Texture Map" );
	theProgressDialog.SetProgressBarMax( fSubmeshHeight );
	theProgressDialog.Show();

	CPhotoshopFile	theFile( inFileSpec, this->GetPixelWidth(), this->GetPixelHeight(),
							 2, this->GetColorTable() );
							 
	StPointerBlock	theDestBuffer( 256 * this->GetPixelWidth() );
	
	// Process each row of submesh
	for ( Int32 i = 0; i < fSubmeshHeight; i++ )
	{
		// Process each column in the row
		for ( Int32 j = 0; j < fSubmeshWidth; j++ )
		{
			Int32			theImageIndex	= ( i * fSubmeshWidth ) + j + 1;
			Handle			theImage		= this->GetTextureMap( theImageIndex );
			StHandleLocker	lock( theImage );
			char			*theSourcePtr	= *theImage;
			char			*theDestPtr		= (Ptr) theDestBuffer + this->GetPixelWidth() - ( ( j + 1 ) * 256 );
			
			for ( Int32 i = 0; i < 256; i++ )
			{
				::BlockMoveData( theSourcePtr, theDestPtr, 256 );
				theSourcePtr	+= 256;
				theDestPtr		+= this->GetPixelWidth();
			}
		}
		
		theFile.WriteImageData( 256, theDestBuffer );
		
		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
	}
}


// -----------------------------------------------------------------------------
// ExportShadowMap
// -----------------------------------------------------------------------------

void CTextureManager::ExportShadowMap( FSSpec &inFileSpec )
{
	// Display a progress dialog
	StProgressDialog	theProgressDialog( 1010, NULL );
	theProgressDialog.SetProgressPrompt( "\pExporting Landscape Shadow Map" );
	theProgressDialog.SetProgressBarMax( fSubmeshHeight );
	theProgressDialog.Show();

	CPhotoshopFile	theFile( inFileSpec, this->GetPixelWidth(), this->GetPixelHeight(),
							 1, NULL );
							 
	StPointerBlock	theDestBuffer( 256 * this->GetPixelWidth() );
	
	// Process each row of submesh
	for ( Int32 i = 0; i < fSubmeshHeight; i++ )
	{
		// Process each column in the row
		for ( Int32 j = 0; j < fSubmeshWidth; j++ )
		{
			Int32			theImageIndex	= ( i * fSubmeshWidth ) + j + 1;
			Handle			theImage		= this->GetShadowMap( theImageIndex );
			StHandleLocker	lock( theImage );
			char			*theSourcePtr	= *theImage;
			char			*theDestPtr		= (Ptr) theDestBuffer + this->GetPixelWidth() - ( ( j + 1 ) * 256 );
			
			for ( Int32 i = 0; i < 256; i++ )
			{
				::BlockMoveData( theSourcePtr, theDestPtr, 256 );
				theSourcePtr	+= 256;
				theDestPtr		+= this->GetPixelWidth();
			}
		}
		
		theFile.WriteImageData( 256, theDestBuffer );

		// Update the progress dialog
		theProgressDialog.IncrementProgressBar();
	}
}


// -----------------------------------------------------------------------------
// GetColorTable
// -----------------------------------------------------------------------------

CTabHandle CTextureManager::GetColorTable()
{
	if ( fColorTable == NULL )
	{
		this->ReadColorTable();
	}
	
	return fColorTable;
}


// -----------------------------------------------------------------------------
// GetBlendTable
// -----------------------------------------------------------------------------

Handle CTextureManager::GetBlendTable()
{
	if ( fBlendTable == NULL )
	{
		this->ReadBlendTable();
	}
	
	return fBlendTable;
}


// -----------------------------------------------------------------------------
// GetRemapTable
// -----------------------------------------------------------------------------

Handle CTextureManager::GetRemapTable()
{
	if ( fRemapTable == NULL )
	{
		this->ReadRemapTable();
	}
	
	return fRemapTable;
}


// -----------------------------------------------------------------------------
// GetTextureMap
// -----------------------------------------------------------------------------

Handle CTextureManager::GetTextureMap( Int32 inSubmeshIndex )
{
	ArrayIndexT theBitmapIndex = ( inSubmeshIndex * 2 ) - 1;
	Assert_( fBitmapArray.ValidIndex( theBitmapIndex ) );
	
	if ( fBitmapArray[theBitmapIndex].fImage == NULL )
	{
		this->ReadImage( theBitmapIndex );
	}
	
	fBitmapArray[theBitmapIndex].fTimestamp = ::LMGetTime();
	
	return fBitmapArray[theBitmapIndex].fImage;
}


// -----------------------------------------------------------------------------
// GetShadowMap
// -----------------------------------------------------------------------------

Handle CTextureManager::GetShadowMap( Int32 inSubmeshIndex )
{
	ArrayIndexT theBitmapIndex = inSubmeshIndex * 2;
	Assert_( fBitmapArray.ValidIndex( theBitmapIndex ) );
	
	if ( fBitmapArray[theBitmapIndex].fImage == NULL )
	{
		this->ReadImage( theBitmapIndex );
	}
	
	fBitmapArray[theBitmapIndex].fTimestamp = ::LMGetTime();

	return fBitmapArray[theBitmapIndex].fImage;
}


// -----------------------------------------------------------------------------
// GetCombinedMap
// -----------------------------------------------------------------------------

Handle CTextureManager::GetCombinedMap( Int32 inSubmeshIndex )
{
	// Hmmm... the blend table doesn't seem to work the way I thought it would
	Handle			theBlendTable	= this->GetBlendTable();
	Handle			theTextureMap	= this->GetTextureMap( inSubmeshIndex );
	Handle			theShadowMap	= this->GetShadowMap( inSubmeshIndex );
	StHandleBlock	theCombinedMap( 256 * 256 );
	StHandleLocker	lockB( theBlendTable );
	StHandleLocker	lockT( theTextureMap );
	StHandleLocker	lockS( theShadowMap );
	StHandleLocker	lockC( theCombinedMap );
	
	char			*theColorIndex	= *theBlendTable;
	char			*theTexturePtr	= *theTextureMap;
	char			*theShadowPtr	= *theShadowMap;
	char			*theCombinedPtr	= *theCombinedMap;
	
	for ( Int32 i = 0; i < 256 * 256; i++ )
	{
		char	theTextureVal	= *theTexturePtr;
		char	theShadowVal	= *theShadowPtr;
//		char	theCombinedVal	= *( theColorIndex + ( theShadowVal * 256 ) + theTextureVal );
		char	theCombinedVal	= *( theColorIndex + ( theTextureVal * 256 ) + theShadowVal );
	
		*theCombinedPtr	= theCombinedVal;
		theTexturePtr++;
		theShadowPtr++;
		theCombinedPtr++;
	}
	
	return theCombinedMap.Release();
}


// -----------------------------------------------------------------------------
// Changed
// -----------------------------------------------------------------------------

void CTextureManager::Changed()
{
	fIsDirty = true;
	this->BroadcastMessage( msg_TextureChanged, NULL );
}


// -----------------------------------------------------------------------------
// UnloadAllImages
// -----------------------------------------------------------------------------

void CTextureManager::UnloadAllImages()
{
	// Free all the loaded images before the array is auto-destructed
	for ( ArrayIndexT i = LArray::index_First; i <= fBitmapArray.GetCount(); i++ )
	{
		if ( fBitmapArray[i].fIsDirty )
		{
			this->WriteImage( fBitmapArray[i].fBitmapRef.offset, fBitmapArray[i].fImage );
			fBitmapArray[i].fIsDirty = false;
		}

		if ( fBitmapArray[i].fImage != NULL )
		{
			::DisposeHandle( fBitmapArray[i].fImage );
			fBitmapArray[i].fImage		= NULL;
			fBitmapArray[i].fTimestamp	= 0x7fffffff;
		}
	}
}


// -----------------------------------------------------------------------------
// UnloadColorTable
// -----------------------------------------------------------------------------

void CTextureManager::UnloadColorTable()
{
	if ( fColorTable )
	{
		::DisposeCTable( fColorTable );
		fColorTable = NULL;
	}
}


// -----------------------------------------------------------------------------
// UnloadBlendTable
// -----------------------------------------------------------------------------

void CTextureManager::UnloadBlendTable()
{
	if ( fBlendTable )
	{
		::DisposeHandle( fBlendTable );
		fBlendTable = NULL;
	}
}


// -----------------------------------------------------------------------------
// UnloadRemapTable
// -----------------------------------------------------------------------------

void CTextureManager::UnloadRemapTable()
{
	if ( fRemapTable )
	{
		::DisposeHandle( fRemapTable );
		fRemapTable = NULL;
	}
}


// -----------------------------------------------------------------------------
// ReadHeader
// -----------------------------------------------------------------------------

void CTextureManager::ReadHeader()
{
	CTagManager::ReadTagHeader( COLLECTION_GROUP_TAG, fTextureTagID, (char*) &fHeader );
	
	Assert_( fHeader.color_table_count == 1 );
	Assert_( fHeader.blend_table_count == 1 );
	Assert_( fHeader.remapping_table_count == 1 );
}


// -----------------------------------------------------------------------------
// ReadBitmapArray
// -----------------------------------------------------------------------------

void CTextureManager::ReadBitmapArray()
{
	TArray<bitmap_reference> theBitmapRefs( CTagManager::ReadTagData(
											COLLECTION_GROUP_TAG, fTextureTagID,
											fHeader.bitmap_references_offset,
											fHeader.bitmap_references_size ) );

	fBitmapArray.RemoveAllItemsAfter( LArray::index_Bad );
	
	for ( ArrayIndexT i = LArray::index_First; i <= theBitmapRefs.GetCount(); i++ )
	{
		SBitmapRef	myBitmapRef;
		
		myBitmapRef.fBitmapRef	= theBitmapRefs[i];
		myBitmapRef.fImage		= NULL;
		myBitmapRef.fTimestamp	= 0x7fffffff;
		myBitmapRef.fIsDirty	= false;
		
		fBitmapArray.AddItem( myBitmapRef );
	}
}


// -----------------------------------------------------------------------------
// ReadColorTable
// -----------------------------------------------------------------------------

void CTextureManager::ReadColorTable()
{
	Assert_( fColorTable == NULL );

	color_table_data	mythColorTable;
	CTabHandle			macColorTable	= ::GetCTable( 40 );
	StHandleLocker		lock( (Handle) macColorTable );
	
	CTagManager::ReadTagData( COLLECTION_GROUP_TAG, fTextureTagID,
							  fHeader.color_tables_offset,
							  sizeof( color_table_data ),
							  &mythColorTable );

	for ( Int32 i = 0; i < mythColorTable.count; i++ )
		macColorTable[0]->ctTable[i].rgb = mythColorTable.color[i].rgb;
		
	::CTabChanged( macColorTable );

	fColorTable = macColorTable;
}


// -----------------------------------------------------------------------------
// ReadBlendTable
// -----------------------------------------------------------------------------

void CTextureManager::ReadBlendTable()
{
	Assert_( fBlendTable == NULL );
	
	fBlendTable = CTagManager::ReadTagData( COLLECTION_GROUP_TAG, fTextureTagID,
											fHeader.blend_table_offset,
											fHeader.blend_table_size );
}


// -----------------------------------------------------------------------------
// ReadRemapTable
// -----------------------------------------------------------------------------

void CTextureManager::ReadRemapTable()
{
	Assert_( fRemapTable == NULL );
	
	fRemapTable = CTagManager::ReadTagData( COLLECTION_GROUP_TAG, fTextureTagID,
											fHeader.remapping_table_offset,
											fHeader.remapping_table_size );
}


// -----------------------------------------------------------------------------
// ReadImage
// -----------------------------------------------------------------------------

void CTextureManager::ReadImage( ArrayIndexT inBitmapIndex )
{
	Assert_( fBitmapArray.ValidIndex( inBitmapIndex ) );

	// Get info about the bitmap
	SBitmapRef		theBitmapRef	= fBitmapArray[inBitmapIndex];

	// Grab memory for the unprocessed image source
	StPointerBlock	theBuffer( kImageSize );
	bitmap_data		*theBitmapData	= (bitmap_data*) (Ptr) theBuffer;
	
	// Read the raw source image from the tag file
	CTagManager::ReadTagData( COLLECTION_GROUP_TAG, fTextureTagID,
							  theBitmapRef.fBitmapRef.offset,
							  theBitmapRef.fBitmapRef.size,
							  theBuffer );

	// Grab memory for the processed image destination
	StHandleBlock	theImage( 256 * 256 );
	StHandleLocker	lock( theImage );
	
	// Process the source image by flipping it horizontally
	this->FlipImage( &(theBitmapData->data[4]), *theImage, 256, 256 );
	
	// Finish by setting the image handle in the bitmap array
	fBitmapArray[inBitmapIndex].fImage = theImage.Release();
}


// -----------------------------------------------------------------------------
// WriteHeader
// -----------------------------------------------------------------------------

void CTextureManager::WriteHeader()
{
	CTagManager::WriteTagHeader( COLLECTION_GROUP_TAG, fTextureTagID, (char*) &fHeader );
}


// -----------------------------------------------------------------------------
// WriteBitmapArray
// -----------------------------------------------------------------------------

void CTextureManager::WriteBitmapArray()
{
}


// -----------------------------------------------------------------------------
// WriteColorTable
// -----------------------------------------------------------------------------

void CTextureManager::WriteColorTable( CTabHandle inColorTable )
{
	// Color table must exist before we attempt to write it
	ThrowIfNULL_( inColorTable );

	color_table_data	mythColorTable;
	
	// All texture map color tables should have 256 entries
	mythColorTable.count = 256;
	
	// Translate mac color table to myth color table format
	for ( Int32 i = 0; i < mythColorTable.count; i++ )
	{
		mythColorTable.color[i].rgb	  = inColorTable[0]->ctTable[i].rgb;
		mythColorTable.color[i].alpha = 0;
	}
		
	// Now write the translated color table to the tag file
	CTagManager::WriteTagData( COLLECTION_GROUP_TAG, fTextureTagID,
							  fHeader.color_tables_offset,
							  sizeof( color_table_data ),
							  (char*) &mythColorTable );
}


// -----------------------------------------------------------------------------
// WriteBlendTable
// -----------------------------------------------------------------------------

void CTextureManager::WriteBlendTable( Handle theBlendTable )
{
	ThrowIfNULL_( theBlendTable );
	
	CTagManager::WriteTagData( COLLECTION_GROUP_TAG, fTextureTagID,
							   fHeader.blend_table_offset,
							   theBlendTable );
}


// -----------------------------------------------------------------------------
// WriteRemapTable
// -----------------------------------------------------------------------------

void CTextureManager::WriteRemapTable( Handle theRemapTable )
{
	ThrowIfNULL_( theRemapTable );
	
	CTagManager::WriteTagData( COLLECTION_GROUP_TAG, fTextureTagID,
							   fHeader.remapping_table_offset,
							   theRemapTable );
}


// -----------------------------------------------------------------------------
// WriteImage
// -----------------------------------------------------------------------------

void CTextureManager::WriteImage( ArrayIndexT inBitmapIndex )
{
	// Sanity check; the image must exist before we attempt to write it
	Assert_( fBitmapArray.ValidIndex( inBitmapIndex ) );
	Assert_( fBitmapArray[inBitmapIndex].fImage != NULL );
	
	this->WriteImage( fBitmapArray[inBitmapIndex].fBitmapRef.offset,
					  fBitmapArray[inBitmapIndex].fImage );
}


// -----------------------------------------------------------------------------
// WriteImage
// -----------------------------------------------------------------------------

void CTextureManager::WriteImage( UInt32 inOffset, Handle inImage )
{
	// Sanity check; the image must exist before we attempt to write it
	Assert_( inImage != NULL );
	
	StPointerBlock	theBuffer( kImageSize );
	bitmap_data		*theBitmapData	= (bitmap_data*) (Ptr) theBuffer;
	StHandleLocker	lock( inImage );
	
	// Set up the bitmap_data struct
	theBitmapData->width		= 256;
	theBitmapData->height		= 256;
	theBitmapData->rowBytes		= 256;
	theBitmapData->flags		= 0x04a1;
	theBitmapData->bitDepth		= 8;
	theBitmapData->widthShift	= 8;
	theBitmapData->heightShift	= 8;
	theBitmapData->unused1[0]	= 0;
	theBitmapData->unused1[1]	= 0;
	theBitmapData->unused1[2]	= 0;
	theBitmapData->unused1[3]	= 0;
	theBitmapData->unused1[4]	= 0;
	theBitmapData->unused1[5]	= 0;
	theBitmapData->unused1[6]	= 0;
	theBitmapData->unknown5		= 8;
	theBitmapData->unknown6		= 8;
	theBitmapData->unknown7		= 256;
	theBitmapData->unknown8		= 256;
	theBitmapData->unused2[0]	= 0;
	theBitmapData->unused2[1]	= 0;
	theBitmapData->unused2[2]	= 0;
	theBitmapData->unused2[3]	= 0;
	theBitmapData->unused2[4]	= 0;
	theBitmapData->unused2[5]	= 0;
	theBitmapData->data[0]		= 0;
	theBitmapData->data[1]		= 0;
	theBitmapData->data[2]		= 0;
	theBitmapData->data[3]		= 0; 
	
	this->FlipImage( *inImage, &(theBitmapData->data[4]), 256, 256 );
					 
	CTagManager::WriteTagData( '.256', fTextureTagID, inOffset, kImageSize, theBuffer );
}


// -----------------------------------------------------------------------------
// FlipImage
// -----------------------------------------------------------------------------
// Myth images are stored backwards, so we need to flip the images horizontally
// after reading and before writing the data to the tag file.

void CTextureManager::FlipImage( char* inSource, char* inDest,
								 UInt16 inHeight, UInt16 inWidth )
{
	// Calculate the source and destination pointers
	Ptr	theSourcePtr	= inSource + inWidth - 1;
	Ptr	theDestPtr		= inDest;
	
	for ( UInt32 i = inHeight; i > 0; i-- )
	{
		for ( UInt32 j = inWidth; j > 0; j-- )
		{
			// Copy the pixel from the source buffer to the destination buffer
			// and set up the pointers for the next copy (destination just
			// proceeds through the buffer, but source is *decremented* through
			// each row to be copied resulting in reverse pixel order in dest)
			*theDestPtr++ = *theSourcePtr--;
		}
		
		// Set source pointer to end of *next* row
		theSourcePtr += 2 * inWidth;
	}
}


// -----------------------------------------------------------------------------
// BuildColorTable
// -----------------------------------------------------------------------------

Handle CTextureManager::BuildColorTable( CTabHandle &inMacColorTable )
{
#pragma unused( inMacColorTable )
	return NULL;
}


// -----------------------------------------------------------------------------
// ExportToPhotoshopFile
// -----------------------------------------------------------------------------

void CTextureManager::ExportToPhotoshopFile( const FSSpec	&inFileSpec,
											 Boolean		inIsLandscape )
// Hack, clean up later
{
	CPhotoshopFile	theFile( inFileSpec, this->GetPixelWidth(), this->GetPixelHeight(),
							 2, this->GetColorTable() );
							 
	StPointerBlock	theDestBuffer( 256 * this->GetPixelWidth() );
	
	// Process each row of submesh
	for ( Int32 i = 0; i < fSubmeshHeight; i++ )
	{
		// Process each column in the row
		for ( Int32 j = 0; j < fSubmeshWidth; j++ )
		{
			Int32			theImageIndex	= ( i * fSubmeshWidth ) + j + 1;
			Handle			theImage		= inIsLandscape ? this->GetTextureMap( theImageIndex )
															: this->GetShadowMap( theImageIndex );
			StHandleLocker	lock( theImage );
			char			*theSourcePtr	= *theImage;
			char			*theDestPtr		= (Ptr) theDestBuffer + this->GetPixelWidth() - ( ( j + 1 ) * 256 );
			
			for ( Int32 i = 0; i < 256; i++ )
			{
				::BlockMoveData( theSourcePtr, theDestPtr, 256 );
				theSourcePtr	+= 256;
				theDestPtr		+= this->GetPixelWidth();
			}
		}
		
		theFile.WriteImageData( 256, theDestBuffer );
	}
}


// =============================================================================
// Bungie-sponsored Blend Table and Blood Remapping Table code
// =============================================================================
// These functions were ported to C++/PowerPlant from Bungie-supplied C code


// -----------------------------------------------------------------------------
// BuildBlendTable
// -----------------------------------------------------------------------------
// Based on Bungie-supplied code

void CTextureManager::BuildBlendTable( CTabHandle		inMacColorTable,
									   StProgressDialog	*inProgress )
{
	ThrowIfNULL_( inMacColorTable );
	
	// Are we displaying a progress dialog?
	if ( inProgress )
	{
		inProgress->SetProgressPrompt( "\pBuilding Blend Table" );
		inProgress->SetProgressBarMax( 256 );
	}
	
	StHandleBlock	theBlendTable( 256 * 256 );
	StHandleLocker	lock( theBlendTable );
	
	for ( Int16 thePrimaryIndex = 0; thePrimaryIndex < ( inMacColorTable[0]->ctSize + 1 ); thePrimaryIndex++ )
	{
		for ( Int16 theSecondaryIndex = 0; theSecondaryIndex < ( inMacColorTable[0]->ctSize + 1 ); theSecondaryIndex++ )
		{
			Int16	theBlendIndex;
			
			if ( thePrimaryIndex < 1 )
			{
				theBlendIndex		= theSecondaryIndex;
			}
			else if ( theSecondaryIndex < 1 )
			{
				theBlendIndex		= thePrimaryIndex;
			}
			else
			{
				RGBColor	theBlendColor;
				
				theBlendColor.red	= ( inMacColorTable[0]->ctTable[thePrimaryIndex].rgb.red +
									    inMacColorTable[0]->ctTable[theSecondaryIndex].rgb.red )   / 2;
				theBlendColor.green	= ( inMacColorTable[0]->ctTable[thePrimaryIndex].rgb.green +
									    inMacColorTable[0]->ctTable[theSecondaryIndex].rgb.green ) / 2;
				theBlendColor.blue	= ( inMacColorTable[0]->ctTable[thePrimaryIndex].rgb.blue +
									    inMacColorTable[0]->ctTable[theSecondaryIndex].rgb.blue )  / 2;
				theBlendIndex		= FindClosestColorIndex( inMacColorTable, theBlendColor );
			}
			
			theBlendTable[0][256 * thePrimaryIndex + theSecondaryIndex] = theBlendIndex;
		}
		
		// Update the progress dialog
		if ( inProgress ) inProgress->IncrementProgressBar();
	}
	
	this->WriteBlendTable( theBlendTable );
}


// -----------------------------------------------------------------------------
// BuildRemappingTable
// -----------------------------------------------------------------------------
// Based on Bungie-supplied code

void CTextureManager::BuildRemappingTable( CTabHandle		inMacColorTable,
										   StProgressDialog	*inProgress )
{
	ThrowIfNULL_( inMacColorTable );
	
	// Are we displaying a progress dialog?
	if ( inProgress )
	{
		inProgress->SetProgressPrompt( "\pBuilding Blood Remapping Table" );
		inProgress->SetProgressBarMax( fSubmeshWidth * fSubmeshHeight );
	}
	
	TArray<histogram_entry>	theHistogram;
	TArray<short>			theUnusedIndicies;
	StHandleBlock			theRemappingTable( 256 );
	StHandleLocker			lock( (Handle) inMacColorTable );
	
	// Initialize the histogram
	for ( Int32 i = 0; i < 256; i++ )
	{
		histogram_entry	theEntry	= { 0, i, 0 };
		theHistogram.AddItem( theEntry );
	}

	// Process each texture map to build the histogram
	for ( ArrayIndexT i = LArray::index_First; i <= fSubmeshWidth * fSubmeshHeight; i++ )
	{
		Handle			theImage		= this->GetTextureMap( i );
		StHandleLocker	lock( theImage );
		Uchar			*theImageData	= (Uchar*) *theImage;

		for ( Int32 i = 0; i < 256 * 256; i++, theImageData++ )
		{
			theHistogram[*theImageData + 1].count++;
		}

		// Update the progress dialog
		if ( inProgress ) inProgress->IncrementProgressBar();
	}

	// Build the array of unused color indicies
	for ( ArrayIndexT i = LArray::index_First; i <= theHistogram.GetCount(); i++ )
	{
		if ( theHistogram[i].count == 0 )
			theUnusedIndicies.AddItem( theHistogram[i].index );
	}

	//  need to put up an alert dialog here!
	if ( theUnusedIndicies.GetCount() == 0 )
	{
		SignalCStr_( "No room for blood colors in this color table." );
	}
	
	// Calculate the delta to distribute blood colors across the histogram
	Int32	theHistogramDelta = ( this->GetPixelWidth() * this->GetPixelHeight()
								  - theHistogram[1].count ) / ( theUnusedIndicies.GetCount() + 1 );
	
	// Sort the histogram so that the most-used colors come first
	theHistogram[1].count = 0;
	theHistogram.SetComparator( new CHistogramComparator, true ); //  need comparator!
	theHistogram.Sort();
	
	// Generate the blood colors
	Int32		theHistogramThreshold	= 0;
	ArrayIndexT	theHistogramIndex		= 1;	
	for ( ArrayIndexT i = LArray::index_First; i <= theUnusedIndicies.GetCount(); i++ )
	{		
		RGBColor	theBloodyColor;
		
		// Calculate the bloodied color and save it in the mac color table
		this->GenerateBloodColor( inMacColorTable[0]->ctTable[theHistogram[theHistogramIndex].index].rgb,
								  theBloodyColor );
		inMacColorTable[0]->ctTable[theUnusedIndicies[i]].rgb = theBloodyColor;
		
		// Find the next original color to use, distributed across the histogram
		do
		{
			theHistogramThreshold	+= theHistogram[theHistogramIndex++].count;
			Assert_( theHistogramIndex < 256 );
		}
		while ( theHistogramThreshold < theHistogramDelta );
		theHistogramThreshold -= theHistogramDelta;
	}
	
	// We've munged the color table, so let MacOS know about it
//	::CTabChanged( inMacColorTable );
	
	// Build a table remapping an original color to the closest bloodied color
	for ( Int32 i = 0; i < inMacColorTable[0]->ctSize + 1; i++ )
	{
		RGBColor	theBloodyColor;
		
		this->GenerateBloodColor( inMacColorTable[0]->ctTable[i].rgb, theBloodyColor );
		theRemappingTable[0][i] = this->FindClosestColorIndex( inMacColorTable, theBloodyColor );
	}
	
	this->WriteColorTable( inMacColorTable );
	this->UnloadColorTable();
	this->WriteRemapTable( theRemappingTable );
}


// -----------------------------------------------------------------------------
// FindClosestColorIndex
// -----------------------------------------------------------------------------
// Based on Bungie-supplied code

Int16 CTextureManager::FindClosestColorIndex( CTabHandle inMacColorTable, RGBColor &inColor )
{
	Int16	theClosestIndex		= 0;
	Int32	theClosestDistance	= 0x7fffffff;
	
	for ( Int16 theColorIndex = 0; theColorIndex < ( inMacColorTable[0]->ctSize + 1 ); theColorIndex++ )
	{
		Int32	dr	= ((Int32) inColor.red   - (Int32) inMacColorTable[0]->ctTable[theColorIndex].rgb.red)   >> 4;
		Int32	dg	= ((Int32) inColor.green - (Int32) inMacColorTable[0]->ctTable[theColorIndex].rgb.green) >> 4;
		Int32	db	= ((Int32) inColor.blue  - (Int32) inMacColorTable[0]->ctTable[theColorIndex].rgb.blue)  >> 4;
		Int32	d	= 2*dr*dr + 5*dg*dg + db*db;
		
		if ( d < theClosestDistance )
		{
			theClosestIndex		= theColorIndex;
			theClosestDistance	= d;
		}
	}
	
	return theClosestIndex;
}


// -----------------------------------------------------------------------------
// GenerateBloodColor
// -----------------------------------------------------------------------------
// Based on Bungie-supplied code

void CTextureManager::GenerateBloodColor( RGBColor &inOriginalColor,
										  RGBColor &ioBloodyColor )
{
	ioBloodyColor		= inOriginalColor;
	ioBloodyColor.red	= MIN(( 0xffff + (long) ioBloodyColor.red ) >> 1, 0xffff );
	ioBloodyColor.green	>>= 1;
	ioBloodyColor.blue	>>= 1;
}


// =============================================================================
// CHistogramComparator
// =============================================================================

// -----------------------------------------------------------------------------
// Compare
// -----------------------------------------------------------------------------

Int32 CHistogramComparator::Compare( const void* inItemOne, const void* inItemTwo,
									 Uint32 inSizeOne, Uint32 inSizeTwo) const
{
#pragma unused( inSizeOne )
#pragma unused( inSizeTwo )

	return ((histogram_entry*) inItemTwo)->count - ((histogram_entry*) inItemOne)->count;
}
