// ===========================================================================
// C3DMapView.cp
// ===========================================================================


#include "C3DMapView.h"
#include "CMeshEditor.h"
#include "CToolManager.h"
#include "CTextureManager.h"
//#include "CModelManager.h"
#include "CPatManager.h"

#include "PiriReisCommands.h"

#include <QD3DGroup.h>
#include <QD3DLight.h>
#include <QD3DCamera.h>
#include <QD3DShader.h>
#include <QD3DGeometry.h>
#include <QD3DDrawContext.h>
#include <QD3DRenderer.h>
#include <QD3DView.h>
#include <QD3DMath.h>
#include <QD3DStorage.h>
#include <QD3DPick.h>
#include <QD3DTransform.h>

const Int16	kConvertToolOptionToTerrainType[12]	= { 0, 6, 0, 7, 8, 10, 0, 4, 5, 0, 14, 15 };

// ---------------------------------------------------------------------------
// C3DMapView
// ---------------------------------------------------------------------------

C3DMapView::C3DMapView( LStream *inStream )
		  : LView( inStream ), LListener(), LCommander()
{
	fMeshEditor			= NULL;
	fTextureManager 	= NULL;
	fMeshCellManager	= NULL;
	fMeshMarkerManager	= NULL;
	fModelManager 		= NULL;

	fRenderType			= kWireframeRender;

	fShowMarkers		= true;
	fShowModels			= true;
	
	fCameraKeys[kCamera_MoveForward]= 0x0D;	// 'W'
	fCameraKeys[kCamera_MoveBack]	= 0x01;	// 'S'
	fCameraKeys[kCamera_MoveLeft]	= 0x06;	// 'Z'
	fCameraKeys[kCamera_MoveRight]	= 0x07;	// 'X'
	fCameraKeys[kCamera_PanLeft]	= 0x00;	// 'A'
	fCameraKeys[kCamera_PanRight]	= 0x02;	// 'D'
	fCameraKeys[kCamera_OrbitLeft]	= 0x0C;	// 'Q'
	fCameraKeys[kCamera_OrbitRight]	= 0x0E;	// 'E'
	fCameraKeys[kCamera_OrbitUp]	= 0x0F;	// 'R'
	fCameraKeys[kCamera_OrbitDown]	= 0x03;	// 'F'
	fCameraKeys[kCamera_ZoomIn]		= 0x08;	// 'C'
	fCameraKeys[kCamera_ZoomOut]	= 0x09;	// 'V'
	
	fQD3DView			= NULL;
	fQD3DMeshGroup		= NULL;
	fQD3DPointPicker	= NULL;
	fQD3DRectPicker		= NULL;
	fQD3DTriGrid		= NULL;
	
	fTextureWorld		= NULL;
	fTextureDetail		= kFullResolution;
	fTextureAttribute	= NULL;
	fTextureBufferSize	= 0L;
	
	fMeshSelection		= ::NewRgn();
	
	// Listen for out-of-memory messages
	LGrowZone::GetGrowZone()->AddListener( this );
}


// ---------------------------------------------------------------------------
// ~C3DMapView
// ---------------------------------------------------------------------------

C3DMapView::~C3DMapView()
{
	::Q3Object_Dispose( fQD3DView );
	::Q3Object_Dispose( fQD3DMeshGroup );
	::Q3Object_Dispose( fQD3DPointPicker );
	::Q3Object_Dispose( fQD3DRectPicker );
	
	if ( fQD3DTriGrid ) {
		::Q3Object_Dispose( fQD3DTriGrid );
	}
	if ( fTextureWorld ) {
		delete fTextureWorld;
	}
	
	::DisposeRgn( fMeshSelection );
}


// ---------------------------------------------------------------------------
// FinishCreateSelf
// ---------------------------------------------------------------------------

void C3DMapView::FinishCreateSelf()
{
	LView::FinishCreateSelf();
	
	this->InitializeQD3DView();
}



// ---------------------------------------------------------------------------
// InitializeQD3DView
// ---------------------------------------------------------------------------

void C3DMapView::InitializeQD3DView()
{
	TQ3Status theStatus;
	Rect thePortRect;
	Assert_( CalcPortFrameRect( thePortRect ) );

	fQD3DView = ::Q3View_New();
	ThrowIfNULL_( fQD3DView );
	
// Draw Context
	// Build the generic draw context
	TQ3DrawContextData theContextData;
	TQ3DrawContextObject theDrawContext = NULL;
	
	// Erase Color
	TQ3ColorARGB theWhiteColor;	
	::Q3ColorARGB_Set( &theWhiteColor, 1.0, 1.0, 1.0, 1.0 );
	
	// Clip Rect
	TQ3Area thePaneArea;
	TQ3Point2D theMinPanePoint;
	TQ3Point2D theMaxPanePoint;
	theMinPanePoint.x 	= thePortRect.left;
	theMinPanePoint.y 	= thePortRect.top;
	theMaxPanePoint.x 	= thePortRect.right;
	theMaxPanePoint.y 	= thePortRect.bottom;
	thePaneArea.min 	= theMinPanePoint;
	thePaneArea.max 	= theMaxPanePoint;

	// Mask
	TQ3Bitmap theNullBitmap = { NULL, 0, 0, 0, kQ3EndianBig};
	
	theContextData.clearImageMethod 	= kQ3ClearMethodWithColor;
	theContextData.clearImageColor 		= theWhiteColor;
	theContextData.pane 				= thePaneArea;
	theContextData.paneState 			= kQ3True;
	theContextData.mask 				= theNullBitmap;
	theContextData.maskState 			= kQ3False;
	theContextData.doubleBufferState 	= kQ3True;

	// Build the macintosh draw context
	TQ3MacDrawContextData theMacContextData;
	
	theMacContextData.drawContextData 	= theContextData;
	theMacContextData.window 			= (CGrafPtr) GetMacPort();
	theMacContextData.library 			= kQ3Mac2DLibraryNone;	// Do we need this for clipping?
	theMacContextData.viewPort 			= NULL;
	theMacContextData.grafPort 			= NULL;

	// Set the draw context in the view
	theDrawContext = ::Q3MacDrawContext_New( &theMacContextData );
	ThrowIfNULL_( theDrawContext );
	theStatus = ::Q3View_SetDrawContext( fQD3DView, theDrawContext );
	ThrowIf_( theStatus == kQ3Failure );
	::Q3Object_Dispose( theDrawContext );
	

// Lights
	// Build the light group
	TQ3GroupObject theLightList;
	TQ3LightData theLightData;
	TQ3LightObject theAmbientLight;
	TQ3GroupPosition theGroupPosition;
	
	// Light color
	TQ3ColorRGB	theWhiteLightColor;
	::Q3ColorRGB_Set( &theWhiteLightColor, 1.0, 1.0, 1.0 );
	
	// Create ambient light
	theLightData.isOn 		= kQ3True;
	theLightData.brightness = 0.2;
	theLightData.color		= theWhiteLightColor;
	
	theAmbientLight = ::Q3AmbientLight_New( &theLightData );
	ThrowIfNULL_( theAmbientLight );

	// Create light group and add lights to the view
	theLightList = ::Q3LightGroup_New();
	ThrowIfNULL_( theLightList );
	
	theGroupPosition = ::Q3Group_AddObject( theLightList, theAmbientLight );
	ThrowIfNULL_( theGroupPosition );
	::Q3Object_Dispose( theAmbientLight );
	
	theStatus = ::Q3View_SetLightGroup( fQD3DView, theLightList );
	ThrowIf_( theStatus == kQ3Failure );
	::Q3Object_Dispose( theLightList );
	

// Camera	
	// Build the camera
	TQ3CameraObject	theCamera;
	TQ3CameraData theCameraData;
	
	// Generic camera data
	TQ3Point3D cameraLocation 	= { 0.0, 15.0, 40.0 };
	TQ3Point3D cameraFocus		= { 30.0, 0.0, 30.0 };
	TQ3Vector3D cameraUp 		= { 0.0, 1.0, 0.0 };

	float hitherValue 	= 0.001;
	float yonValue 		= 1000;

	theCameraData.placement.cameraLocation 	= cameraLocation;
	theCameraData.placement.pointOfInterest = cameraFocus;
	theCameraData.placement.upVector 		= cameraUp;

	theCameraData.range.hither 	= hitherValue;
	theCameraData.range.yon 	= yonValue;

	theCameraData.viewPort.origin.x 	= -1.0;
	theCameraData.viewPort.origin.y 	= 1.0;
	theCameraData.viewPort.width 		= 2.0;
	theCameraData.viewPort.height 		= 2.0;
	
	// Angle Aspect camera data
	TQ3ViewAngleAspectCameraData theAngleAspectCameraData;
	
	float fieldOfView = 1.0;

	theAngleAspectCameraData.cameraData 		= theCameraData;
	theAngleAspectCameraData.fov 				= fieldOfView;
	theAngleAspectCameraData.aspectRatioXToY 	= (float) ( thePortRect.right - thePortRect.left ) / 
													(float) ( thePortRect.bottom - thePortRect.top );
	
	// Create the camera and add it to the view
	theCamera = ::Q3ViewAngleAspectCamera_New( &theAngleAspectCameraData );
	ThrowIfNULL_( theCamera );
	
	theStatus = ::Q3View_SetCamera( fQD3DView, theCamera );
	ThrowIf_( theStatus == kQ3Failure );
	::Q3Object_Dispose( theCamera );
	

// Renderer	
	// Build the renderer and add it to view
	TQ3RendererObject theRenderer;
	if ( fRenderType == kWireframeRender )
		theRenderer = Q3Renderer_NewFromType( kQ3RendererTypeWireFrame );
	else
		theRenderer = Q3Renderer_NewFromType( kQ3RendererTypeInteractive );
	ThrowIfNULL_( theRenderer );	
	
	theStatus = ::Q3View_SetRenderer( fQD3DView, theRenderer );
	ThrowIf_( theStatus == kQ3Failure );
	::Q3Object_Dispose( theRenderer );

// Pickers
	TQ3WindowPointPickData thePointData;
	
	thePointData.data.sort 				= kQ3PickSortNearToFar;
	thePointData.data.mask 				= kQ3PickDetailMaskPickID | kQ3PickDetailMaskXYZ | kQ3PickDetailMaskUV;
	thePointData.data.numHitsToReturn 	= kQ3ReturnAllHits;
	thePointData.point.x 				= 0;
	thePointData.point.y 				= 0;
	thePointData.vertexTolerance 		= 0.1;
	thePointData.edgeTolerance			= 0.1;

	fQD3DPointPicker = ::Q3WindowPointPick_New( &thePointData );
	ThrowIfNULL_( fQD3DPointPicker );
	
	TQ3WindowRectPickData theRectData;
	
	theRectData.data.sort 				= kQ3PickSortNone;
	theRectData.data.mask 				= kQ3PickDetailMaskPickID;
	theRectData.data.numHitsToReturn 	= kQ3ReturnAllHits;
	theRectData.rect.min.x				= 0.0;
	theRectData.rect.min.y				= 0.0;
	theRectData.rect.max.x				= 0.0;
	theRectData.rect.max.y				= 0.0;

	fQD3DRectPicker = ::Q3WindowRectPick_New( &theRectData );
	ThrowIfNULL_( fQD3DRectPicker );
}


// ---------------------------------------------------------------------------
// SetMeshEditor
// ---------------------------------------------------------------------------

void C3DMapView::SetMeshEditor( CMeshEditor *inMeshEditor )
{
	fMeshEditor 		= inMeshEditor;
	fTextureManager		= inMeshEditor->GetTextureManager();
	fMeshCellManager	= inMeshEditor->GetMeshCellManager();
	fMeshMarkerManager	= inMeshEditor->GetMeshMarkerManager();
	fModelManager		= inMeshEditor->GetModelManager();

	fTextureManager	  ->AddListener( this );
	fMeshCellManager  ->AddListener( this );
	fMeshMarkerManager->AddListener( this );
	
	Int32 theMeshRows	= fMeshCellManager->GetCellHeight();
	Int32 theMeshCols	= fMeshCellManager->GetCellWidth();
	
	// Unfortuantely the Bungie data is incorrect in that it
	// lacks vertex positions for the final row and column.
	// We'll try to adjust for that here while we build
	// the TriGrid so things will be simpler later.
	
	// This is further complicated by the fact that the
	// "flipping" of the coordinates means the extra row
	// comes "last" while the extra column comes "first"
	Int32 theVertexRows	= theMeshRows + 1;
	Int32 theVertexCols	= theMeshCols + 1;
	
	if ( fQD3DMeshGroup )
		::Q3Object_Dispose( fQD3DMeshGroup );
		
	fQD3DMeshGroup = ::Q3OrderedDisplayGroup_New();
	ThrowIfNULL_( fQD3DMeshGroup );

	// Create the TriGrid
	TQ3Vertex3D* theVertices = new TQ3Vertex3D[theVertexRows * theVertexCols];
	
	CellArrayT theMeshArray = fMeshCellManager->GetCellArray();
	StArrayLocker lock( theMeshArray );
	
	ArrayIndexT meshIndex = LArray::index_First;
	
	for ( Int32 rowIndex = 0; rowIndex < theMeshRows; rowIndex++ )
	{
		for ( Int32 colIndex = theMeshCols; colIndex >= 1; colIndex-- )	// "Flip" the columns here
		{
			// Note the use of theVertexCols rather than theMeshCols
			UInt32	theVertexIndex	= colIndex + rowIndex * theVertexCols;
			
			theVertices[theVertexIndex].point.x		 = colIndex;
			theVertices[theVertexIndex].point.y		 = (float) WORLD_TO_FLOAT( theMeshArray[meshIndex].height );
			theVertices[theVertexIndex].point.z		 = rowIndex;
			theVertices[theVertexIndex].attributeSet = nil;
			
			meshIndex++;
		}
	}
	
	// Fix up the extra "edges" by duplicating the last row and the first column
	for ( Int32 rowIndex = 0; rowIndex < theVertexRows; rowIndex++ )
	{
		UInt32	theVertexIndex	= rowIndex * theVertexCols;

		theVertices[theVertexIndex].point.x		 = 0;
		theVertices[theVertexIndex].point.y		 = theVertices[theVertexIndex + 1].point.y;
		theVertices[theVertexIndex].point.z		 = rowIndex;
		theVertices[theVertexIndex].attributeSet = nil;
	}	
	for ( Int32 colIndex = 0; colIndex < theVertexCols; colIndex++ )
	{
		UInt32	theVertexIndex	= colIndex + theMeshRows * theVertexCols;

		theVertices[theVertexIndex].point.x		 = colIndex;
		theVertices[theVertexIndex].point.y		 = theVertices[theVertexIndex - theVertexCols].point.y;
		theVertices[theVertexIndex].point.z		 = theMeshRows;
		theVertices[theVertexIndex].attributeSet = nil;
	}	

	TQ3TriGridData 	theTriGridData;
	Int32			theFacetCount 	= 2 * ( theMeshRows * theMeshCols );

	theTriGridData.numRows 				= theVertexRows;
	theTriGridData.numColumns 			= theVertexCols;
	theTriGridData.facetAttributeSet 	= new TQ3AttributeSet[theFacetCount];
	theTriGridData.triGridAttributeSet 	= NULL;
	theTriGridData.vertices 			= theVertices;

	for ( Int32 facetIndex = 0; facetIndex < theFacetCount; facetIndex++ )
	{
		theTriGridData.facetAttributeSet[facetIndex] = NULL;
	}
	
	fQD3DTriGrid = ::Q3TriGrid_New( &theTriGridData );
	ThrowIfNULL_( fQD3DTriGrid );

	delete [] theVertices;

	// Parameterize the surface
	TQ3AttributeSet theVertexAttributes;
	TQ3Param2D theVertexUV;
	float paramUScale = 1.0 / ( theMeshCols );
	float paramVScale = 1.0 / ( theMeshRows );
	
	for ( UInt32 rowIndex = 0; rowIndex < theMeshRows; rowIndex++ )
	{
		for ( UInt32 colIndex = 1; colIndex <= theMeshCols; colIndex++ )
		{
			theVertexAttributes = ::Q3AttributeSet_New();
			ThrowIfNULL_( theVertexAttributes );
			theVertexUV.u = colIndex * paramUScale;
			theVertexUV.v = (( theMeshRows ) - rowIndex) * paramVScale;
			
			::Q3AttributeSet_Add(theVertexAttributes, kQ3AttributeTypeSurfaceUV, &theVertexUV);
			::Q3TriGrid_SetVertexAttributeSet(fQD3DTriGrid, rowIndex, colIndex, theVertexAttributes);
			::Q3Object_Dispose(theVertexAttributes);
		}
	}
	
	// Add the TriGrid to the group
	::Q3Group_AddObject( fQD3DMeshGroup, fQD3DTriGrid );
}


// ---------------------------------------------------------------------------
// SetRenderType
// ---------------------------------------------------------------------------

void C3DMapView::SetRenderType( ERenderType inRenderType )
{
	Assert_( fTextureManager != NULL );
	Assert_( fMeshCellManager != NULL );
	Assert_( fQD3DView != NULL );

	if ( fRenderType != inRenderType )
	{
		TQ3RendererObject theRenderer;
		if ( inRenderType == kWireframeRender )
			theRenderer = Q3Renderer_NewFromType( kQ3RendererTypeWireFrame );
		else
			theRenderer = Q3Renderer_NewFromType( kQ3RendererTypeInteractive );
		ThrowIfNULL_( theRenderer );	

		TQ3Status theStatus = ::Q3View_SetRenderer( fQD3DView, theRenderer );
		ThrowIf_( theStatus == kQ3Failure );
		::Q3Object_Dispose( theRenderer );
		
		// Add or remove the texture
		if ( inRenderType == kTextureRender )
		{
			if ( ( fTextureWorld == NULL ) && ( fTextureDetail > kNoTextureResolution ) )
			{
				BuildTexture();
			}

			// Add the mipmap
			theStatus = ::Q3Geometry_SetAttributeSet( fQD3DTriGrid, fTextureAttribute );
			ThrowIf_( theStatus == kQ3Failure );
		}
		if ( fRenderType == kTextureRender )
		{
			// Remove the mipmap
			theStatus = ::Q3Geometry_SetAttributeSet( fQD3DTriGrid, NULL );
			ThrowIf_( theStatus == kQ3Failure );
		}
		
		// Add or remove the terrain colors
		if ( inRenderType == kTerrainTypeRender )
		{
			// Get the mesh data
			Int32 theMeshRows	= fMeshCellManager->GetCellHeight();
			Int32 theMeshCols	= fMeshCellManager->GetCellWidth();
		
			CellArrayT theMeshArray = fMeshCellManager->GetCellArray();
			StArrayLocker lock( theMeshArray );
		
			// Bloody TArray<> braindead indices!
			ArrayIndexT meshIndex = LArray::index_First;
		
			for ( Int32 rowIndex = 0; rowIndex < theMeshRows; rowIndex++ )
			{
				for ( Int32 colIndex = theMeshCols - 1; colIndex >= 0; colIndex-- )	// "Flip" the columns here
				{
					UInt32	theFacetIndex	= 2 * ( colIndex + rowIndex * theMeshCols );
					
					TQ3AttributeSet	attribute1 = CPatManager::GetColorAttribute(kColorMediaDwarfDepth + GET_CELL_TERRAIN_TYPE( &theMeshArray[meshIndex], true ) );
					TQ3AttributeSet	attribute2 = CPatManager::GetColorAttribute(kColorMediaDwarfDepth + GET_CELL_TERRAIN_TYPE( &theMeshArray[meshIndex], false ) );
					theStatus = ::Q3TriGrid_SetFacetAttributeSet( fQD3DTriGrid, theFacetIndex, attribute1 );
					ThrowIf_( theStatus == kQ3Failure );
					theStatus = ::Q3TriGrid_SetFacetAttributeSet( fQD3DTriGrid, theFacetIndex + 1, attribute2 );
					ThrowIf_( theStatus == kQ3Failure );
					
					meshIndex++;
				}
			}
		}
		if ( fRenderType == kTerrainTypeRender )
		{
			// Get the mesh data
			Int32 theCellCount	= fMeshCellManager->GetCellHeight() * fMeshCellManager->GetCellWidth();
		
			for ( Int32 cellIndex = 0; cellIndex < theCellCount; cellIndex++ )
			{
				UInt32 theFacetIndex = 2 * cellIndex;
				
				theStatus = ::Q3TriGrid_SetFacetAttributeSet( fQD3DTriGrid, theFacetIndex, NULL );
				ThrowIf_( theStatus == kQ3Failure );
				theStatus = ::Q3TriGrid_SetFacetAttributeSet( fQD3DTriGrid, theFacetIndex + 1, NULL );
				ThrowIf_( theStatus == kQ3Failure );
			}
		}			


		fRenderType = inRenderType;
		this->Refresh();
	}
}


// ---------------------------------------------------------------------------
// BuildTexture
// ---------------------------------------------------------------------------

 void C3DMapView::BuildTexture( EDetailLevel inDetail )
 {
	Int16 cellWidth 	= fMeshCellManager->GetCellWidth();
	Int16 cellHeight	= fMeshCellManager->GetCellHeight();
	Rect worldBounds	= { 0, 0, 0, 0 };
	TQ3Status theStatus;

	// Clean up existing texture
	if ( fTextureAttribute )
	{
		theStatus = ::Q3Geometry_SetAttributeSet( fQD3DTriGrid, NULL );
		::Q3Object_Dispose( fTextureAttribute );
		fTextureAttribute = NULL;
		ThrowIf_( theStatus == kQ3Failure );
	}
	if ( fTextureWorld )
	{
		delete fTextureWorld;
		fTextureWorld = NULL;
	}
	fTextureBufferSize = 0L;
		
	// Reduce resolution until we successfully create the GWorld
	this->StopListening();
	if ( inDetail >= kFullResolution )
	{
		Try_
		{
			worldBounds.bottom = cellHeight * kFullResolution;
			worldBounds.right = cellWidth * kFullResolution;
			fTextureWorld = new LGWorld( worldBounds, 16 );
			fTextureDetail = kFullResolution;
		}
		Catch_( inErr )
		{
			delete fTextureWorld;
			fTextureWorld = NULL;
		}
	}
	if ( ( fTextureWorld == NULL ) && ( inDetail >= kHalfResolution ) )
	{
		Try_
		{
			worldBounds.bottom = cellHeight * kHalfResolution;
			worldBounds.right = cellWidth * kHalfResolution;
			fTextureWorld = new LGWorld( worldBounds, 16 );
			fTextureDetail = kHalfResolution;
		}
		Catch_( inErr )
		{
			delete fTextureWorld;
			fTextureWorld = NULL;
		}
	}
	if ( ( fTextureWorld == NULL ) && ( inDetail >= kQuarterResolution ) )
	{
		Try_
		{
			worldBounds.bottom = cellHeight * kQuarterResolution;
			worldBounds.right = cellWidth * kQuarterResolution;
			fTextureWorld = new LGWorld( worldBounds, 16 );
			fTextureDetail = kQuarterResolution;
		}
		Catch_( inErr )
		{
			delete fTextureWorld;
			fTextureWorld = NULL;
			fTextureDetail = kNoTextureResolution;
		}
	}
	this->StartListening();
	
	if ( fTextureWorld )
	{
		GWorldPtr 		theMacGWorld = fTextureWorld->GetMacGWorld();
		PixMapHandle 	theMacPixMap = ::GetGWorldPixMap( theMacGWorld );
		UInt32			theRowBytes = ( *theMacPixMap )->rowBytes & 0x3FFF;
		
		::LockPixels( theMacPixMap );

		// Draw the texture into our new GWorld
		Rect			scratchBounds		= { 0, 0, 256, 256 };
		LGWorld*		scratchGWorld		= new LGWorld( scratchBounds, 8, 0, fTextureManager->GetColorTable() );
		PixMapHandle	scratchMacPixMap	= ::GetGWorldPixMap( scratchGWorld->GetMacGWorld() );
		Ptr				scratchPixH			= ( *scratchMacPixMap )->baseAddr;

		( *scratchMacPixMap )->rowBytes = 0x8000 | 256;
		
		Int16 rowCount = fMeshCellManager->GetSubmeshHeight();
		Int16 colCount = fMeshCellManager->GetSubmeshWidth();
		Int16 submeshCount = rowCount * colCount;
		
		fTextureWorld->BeginDrawing();
		for (Int16 rowIndex = 0; rowIndex < rowCount; rowIndex++ )
		{
			for (Int16 colIndex = 0; colIndex < colCount; colIndex++ )
			{
				Rect 	theDestRect = { rowIndex * SUBMESH_CELL_WIDTH * fTextureDetail,
										colIndex * SUBMESH_CELL_WIDTH * fTextureDetail,
										theDestRect.top + SUBMESH_CELL_WIDTH * fTextureDetail,
										theDestRect.left + SUBMESH_CELL_WIDTH * fTextureDetail };

				Handle	theTexture	= fTextureManager->GetTextureMap( rowIndex * colCount + (colCount - colIndex) );
	
				( *scratchMacPixMap )->baseAddr = (Ptr) theTexture;
	
				scratchGWorld->CopyImage( (GrafPtr) theMacGWorld, theDestRect, srcCopy, NULL );
			}
		}
		fTextureWorld->EndDrawing();

		( *scratchMacPixMap )->baseAddr = scratchPixH;
		delete scratchGWorld;

		// We'll use a mipmap rather than a pixmap because Brian Greenstone says that
		// pixmaps automatically allocate mipmap memory. See "Cool 3D Apps" 2-13
		TQ3Mipmap theMipmap;
		fTextureBufferSize = cellHeight * fTextureDetail * theRowBytes;
		theMipmap.image = ::Q3MemoryStorage_NewBuffer( (UInt8*) ::GetPixBaseAddr( theMacPixMap ), fTextureBufferSize, fTextureBufferSize );
		ThrowIfNULL_( theMipmap.image );

		theMipmap.useMipmapping = kQ3False;
		theMipmap.pixelType 	= kQ3PixelTypeRGB16;
		theMipmap.bitOrder 		= kQ3EndianBig;
		theMipmap.byteOrder 	= kQ3EndianBig;
		theMipmap.reserved 		= 0L;
		
		theMipmap.mipmaps[0].width 		= cellWidth * fTextureDetail;
		theMipmap.mipmaps[0].height 	= cellHeight * fTextureDetail;
		theMipmap.mipmaps[0].rowBytes 	= theRowBytes;
		theMipmap.mipmaps[0].offset 	= 0;
		
		TQ3TextureObject theTexture = ::Q3MipmapTexture_New( &theMipmap );
		::Q3Object_Dispose( theMipmap.image );
		ThrowIfNULL_( theTexture );
		
		TQ3ShaderObject theShader = Q3TextureShader_New( theTexture );
		::Q3Object_Dispose( theTexture );
		ThrowIfNULL_( theShader );

		fTextureAttribute = ::Q3AttributeSet_New();
		ThrowIfNULL_( fTextureAttribute );

		::Q3AttributeSet_Add( fTextureAttribute, kQ3AttributeTypeSurfaceShader, &theShader );
		::Q3Object_Dispose( theShader );
	}
 }


// ---------------------------------------------------------------------------
// SetShowMarkers
// ---------------------------------------------------------------------------

void C3DMapView::SetShowMarkers( Boolean inShowMarkers )
{
	if ( fShowMarkers != inShowMarkers )
	{
		fShowMarkers = inShowMarkers;
		this->Refresh();
	}
}


// ---------------------------------------------------------------------------
// SetShowModels
// ---------------------------------------------------------------------------

void C3DMapView::SetShowModels( Boolean inShowModels )
{
	if ( fShowModels != inShowModels )
	{
		fShowModels = inShowModels;
		this->Refresh();
	}
}


// ----------------------------------------------------------------------------
// ListenToMessage									
// ----------------------------------------------------------------------------

void C3DMapView::ListenToMessage( MessageT inMessage, void *ioParam )
{
#pragma unused( ioParam )

	switch(inMessage)
	{
		case msg_MeshSelectionChanged:
		case msg_MeshDataChanged:
		case msg_MarkerSelectionChanged:
		case msg_MarkerDataChanged:
		case msg_MediaDataChanged:
			this->Refresh();
			break;
		
		case cmd_ImportHeightMap:
// Hack
			this->SetMeshEditor( fMeshEditor );
			this->DrawSelf();
			break;
		
		case msg_GrowZone:
			if ( fTextureAttribute != NULL )
			{
				Int32 originalMemory	= fTextureBufferSize;
				EDetailLevel newDetail 	= kNoTextureResolution;
				switch ( fTextureDetail )
				{
					case kFullResolution:
						newDetail = kHalfResolution;
						break;

					case kHalfResolution:
						newDetail = kQuarterResolution;
						break;

					case kQuarterResolution:
					default:
						newDetail = kNoTextureResolution;
						break;
				}
				this->BuildTexture( newDetail );
				// Force the mipmap to reappear
				TQ3Status theStatus = ::Q3Geometry_SetAttributeSet( fQD3DTriGrid, fTextureAttribute );
				this->Refresh();

				*((Int32*) ioParam) = originalMemory - fTextureBufferSize;
				ThrowIf_( theStatus == kQ3Failure );
			}
			break;
	}
}

// ----------------------------------------------------------------------------
// ResizeFrameBy									
// ----------------------------------------------------------------------------

void C3DMapView::ResizeFrameBy(Int16 inWidthDelta, Int16 inHeightDelta, Boolean inRefresh)
{
	#pragma unused( inRefresh )
	
	LView::ResizeFrameBy( inWidthDelta, inHeightDelta, false );

	if ( fQD3DView )
	{
		Rect thePortRect;
		CalcPortFrameRect( thePortRect );
		
		TQ3Status theStatus;
		TQ3DrawContextObject theDrawContext;
		
		// Update the draw context
		theStatus = ::Q3View_GetDrawContext( fQD3DView, &theDrawContext );
		ThrowIf_( theStatus == kQ3Failure );
		
		TQ3Area thePaneArea;
		TQ3Point2D theMinPanePoint;
		TQ3Point2D theMaxPanePoint;
		
		theMinPanePoint.x 	= thePortRect.left;
		theMinPanePoint.y 	= thePortRect.top;
		theMaxPanePoint.x 	= thePortRect.right;
		theMaxPanePoint.y 	= thePortRect.bottom;
		thePaneArea.min 	= theMinPanePoint;
		thePaneArea.max 	= theMaxPanePoint;
		
		theStatus = ::Q3DrawContext_SetPane( theDrawContext, &thePaneArea );
		ThrowIf_( theStatus == kQ3Failure );
		::Q3Object_Dispose( theDrawContext );
		
		// update the camera
		TQ3CameraObject theCamera;
		theStatus = ::Q3View_GetCamera( fQD3DView, &theCamera );
		ThrowIf_( theStatus == kQ3Failure );
		
		float aspectRatioXToY = (float) ( thePortRect.right - thePortRect.left ) / 
								(float)  (thePortRect.bottom - thePortRect.top );
		
		theStatus = ::Q3ViewAngleAspectCamera_SetAspectRatio( theCamera, aspectRatioXToY );
		ThrowIf_( theStatus == kQ3Failure );
		::Q3Object_Dispose( theCamera );
	}
}


// ---------------------------------------------------------------------------
// FocusDraw
// ---------------------------------------------------------------------------
// Restore the port origin to reduce confusion for QuickDraw3D drawing and clicking

Boolean C3DMapView::FocusDraw( LPane *inSubPane )
{
	Boolean superFocused = LPane::FocusDraw( inSubPane );
	
	// Restore the origin
	::SetOrigin( 0, 0 );
	
	LView::OutOfFocus( nil );
	
	return superFocused;
}


// ---------------------------------------------------------------------------
// DrawSelf
// ---------------------------------------------------------------------------

void C3DMapView::DrawSelf()
{
	if (fQD3DView)
	{
		TQ3Status theStatus;
		
		theStatus = ::Q3View_StartRendering( fQD3DView );
		ThrowIf_( theStatus == kQ3Failure );
		
		do
		{
			// Render the mesh
			if ( fQD3DMeshGroup )
			{
				theStatus = ::Q3DisplayGroup_Submit( fQD3DMeshGroup, fQD3DView );
				ThrowIf_( theStatus == kQ3Failure );
			}

			// Render the models
			if ( fModelManager && fShowModels )
			{
				MarkerArrayT	theModelMarkers 	= fMeshMarkerManager->GetMarkerArray( _marker_model );
				PaletteArrayT	theModelPalette 	= fMeshMarkerManager->GetPaletteArray( _marker_model );
				StArrayLocker	lockP( theModelMarkers );
				StArrayLocker	lockM( theModelPalette );
				
				ArrayIndexT		theMaxIndex			= theModelMarkers.GetCount();

				for ( ArrayIndexT i = LArray::index_First; i <= theMaxIndex; i++ )
				{
					marker_data		theMarker		= theModelMarkers[i];
					FourCharCode	theTag			= theModelPalette[theMarker.palette_index + 1].tag;

					TQ3Matrix4x4 theRotationMatrix;
					this->CalcModelRotation( theMarker.location.yaw, theRotationMatrix );

					TQ3Matrix4x4 theTranslationMatrix;
					this->CalcModelTranslation( theMarker.location.position.x, theMarker.location.position.y,
									  			theMarker.location.position.z, theTranslationMatrix );
					
					theStatus = ::Q3Push_Submit( fQD3DView );
					ThrowIf_( theStatus == kQ3Failure );
					{
						TQ3GroupObject theModel = fModelManager->GetModel( theTag );
						if ( theModel )
						{
							TQ3AttributeSet theAttribute = CPatManager::GetColorAttribute( kColorGray8 );
							theStatus = ::Q3AttributeSet_Submit( theAttribute, fQD3DView );
							ThrowIf_( theStatus == kQ3Failure );

							theStatus = ::Q3MatrixTransform_Submit( &theTranslationMatrix, fQD3DView );
							ThrowIf_( theStatus == kQ3Failure );

							theStatus = ::Q3MatrixTransform_Submit( &theRotationMatrix, fQD3DView );
							ThrowIf_( theStatus == kQ3Failure );

							theStatus = ::Q3DisplayGroup_Submit( theModel, fQD3DView );
//							ThrowIf_( theStatus == kQ3Failure );
						}
					}
					theStatus = ::Q3Pop_Submit( fQD3DView );
					ThrowIf_( theStatus == kQ3Failure );
				}
			}

			// Render the markers
			if ( fMeshMarkerManager && fShowMarkers)
			{
				PaletteArrayT&	thePaletteArray	= fMeshMarkerManager->GetPaletteArray();
				MarkerArrayT&	theMarkerArray	= fMeshMarkerManager->GetMarkerArray();
				StArrayLocker	lockP( thePaletteArray );
				StArrayLocker	lockM( theMarkerArray );
				
				ArrayIndexT		theMaxIndex			= theMarkerArray.GetCount();

				for ( ArrayIndexT i = LArray::index_First; i <= theMaxIndex; i++ )
				{
					marker_data		theMarker		= theMarkerArray[i];
					Boolean			theSelectedFlag	= theMarker.unused[0];
					Boolean			theVisibleFlag	= ( theMarker.flags & _marker_is_invisible_flag ) == 0;
					short			theTeamIndex	= thePaletteArray[theMarker.palette_index + 1].team_index;
					TQ3BoxData 		theMarkerBox;
					
					this->CalcMarkerBox( theMarker.location.position.x,
										  theMarker.location.position.y,
										  theMarker.location.position.z,
										  theMarkerBox );
								
					theMarkerBox.boxAttributeSet = this->GetColorAttribute( theSelectedFlag, theVisibleFlag, theTeamIndex );
					
					theStatus = ::Q3Box_Submit( &theMarkerBox, fQD3DView );
					ThrowIf_( theStatus == kQ3Failure );
				}
			}
			
			// Render the mesh selection
			if ( !::EmptyRgn( fMeshSelection ) )
			{
/*				TQ3BoxData theSelectionBox;
				theSelectionBox.faceAttributeSet = NULL;
				theSelectionBox.boxAttributeSet = CPatManager::GetColorAttribute( kColorBlack );
	
				::Q3Vector3D_Set( &theSelectionBox.orientation, 0.0, 0.2, 0.0 );
				::Q3Vector3D_Set( &theSelectionBox.minorAxis, 0.2, 0.0, 0.0 );
				::Q3Vector3D_Set( &theSelectionBox.majorAxis, 0.0, 0.0, 0.2 );
	
				for ()
				{	
					float tempX = WORLD_TO_FLOAT( inMarkerX );
					float tempY = WORLD_TO_FLOAT( inMarkerY );
					float tempZ = WORLD_TO_FLOAT( inMarkerZ );
					float mirroredXPos = fMeshCellManager->GetCellWidth() - tempX;

					// The marker Z axis is equivalent to QD3D's Y axis and vice versa
					::Q3Point3D_Set( &outBox.origin, mirroredXPos - 0.1, tempZ, tempY - 0.1 );
							
					theStatus = ::Q3Box_Submit( &theSelectionBox, fQD3DView );
					ThrowIf_( theStatus == kQ3Failure );
				}
*/			}

		}
		while ( ::Q3View_EndRendering( fQD3DView ) == kQ3ViewStatusRetraverse );
	}	
}


// ---------------------------------------------------------------------------
// ClickSelf
// ---------------------------------------------------------------------------

void C3DMapView::ClickSelf( const SMouseDownEvent &inMouseDown )
{
	this->FocusDraw();
	
	switch ( CToolManager::GetCurrentTool() )
	{
		case kSelectMeshTool:
			this->HandleSelectMeshToolClick( inMouseDown );
			break;
			
		case kSelectMarkerTool:
			this->HandleSelectMarkerToolClick( inMouseDown );
			break;
			
		case kPlaceMarkerTool:
			this->HandlePlaceMarkerToolClick( inMouseDown );
			break;
			
		case kTerrainTypeTool:
			this->HandleTerrainTypeToolClick( inMouseDown );
			break;

		default:
			break;
	}
}


// ---------------------------------------------------------------------------
// HandleSelectMeshToolClick
// ---------------------------------------------------------------------------

void C3DMapView::HandleSelectMeshToolClick( const SMouseDownEvent &inMouseDown )
{
	Point portPoint = inMouseDown.wherePort;
	
	if ( inMouseDown.macEvent.modifiers & shiftKey )	// Extend the selection
	{
		RgnHandle originalSelection = ::NewRgn();
		ThrowIfNULL_( originalSelection );
		RgnHandle currentSelection = ::NewRgn();
		ThrowIfNULL_( currentSelection );
		::CopyRgn( fMeshSelection, originalSelection );

		Point 			mousePoint;
		world_point3d 	portMythPoint;
		world_point3d 	mouseMythPoint;
		
		while( ::StillDown() )
		{
			::GetMouse( &mousePoint );

			Rect	theSelectionRect	= { MIN( portPoint.v, mousePoint.v ),
											MIN( portPoint.h, mousePoint.h ),
											MAX( portPoint.v, mousePoint.v ),
											MAX( portPoint.h, mousePoint.h ) };
			
			::PenPat( &qd.gray );
			::FrameRect( &theSelectionRect );
			
			this->SelectVerticesInRect( theSelectionRect );

			this->DrawSelf();
		}
		
		::DisposeRgn( originalSelection );
		::DisposeRgn( currentSelection );
	}
	else
	{
		TQ3Status theStatus;
		TQ3Point2D theQD3DPoint;
		theQD3DPoint.x = portPoint.h;
		theQD3DPoint.y = portPoint.v;
		
		theStatus = ::Q3WindowPointPick_SetPoint( fQD3DPointPicker, &theQD3DPoint );
		ThrowIf_( theStatus == kQ3Failure );
		
		::Q3View_StartPicking( fQD3DView, fQD3DPointPicker );
		ThrowIf_( theStatus == kQ3Failure );

		do
		{
			::Q3DisplayGroup_Submit( fQD3DMeshGroup, fQD3DView );
			ThrowIf_( theStatus == kQ3Failure );
		}
		while ( ::Q3View_EndPicking( fQD3DView ) == kQ3ViewStatusRetraverse );

		UInt32 theHitCount = 0L;

		theStatus = ::Q3Pick_GetNumHits( fQD3DPointPicker, &theHitCount );
		ThrowIf_( theStatus == kQ3Failure );

		if ( theHitCount > 0 )
		{
			TQ3PickDetail 	theHitDetail;
			TQ3Point3D 		theXYZPoint;

			theStatus = ::Q3Pick_GetPickDetailValidMask( fQD3DPointPicker, 0, &theHitDetail );
			ThrowIf_( theStatus == kQ3Failure );
			Assert_( theHitDetail & kQ3PickDetailMaskXYZ );

			UInt32 thePickID = 0L;
			theStatus = ::Q3Pick_GetPickDetailData( fQD3DPointPicker, 0, kQ3PickDetailMaskXYZ, &theXYZPoint );
			ThrowIf_( theStatus == kQ3Failure );
			
			// Build the world and frustum transforms
			TQ3CameraObject theCamera;
			TQ3Matrix4x4 	matrixWorldToFrustum;
			TQ3Matrix4x4 	matrixFrustumToWorld;
			
			theStatus = ::Q3View_GetCamera ( fQD3DView, &theCamera );
			ThrowIf_( theStatus == kQ3Failure );
			theStatus = ::Q3Camera_GetWorldToFrustum( theCamera, &matrixWorldToFrustum );
			ThrowIf_( theStatus == kQ3Failure );
			
			::Q3Matrix4x4_Invert( &matrixWorldToFrustum, &matrixFrustumToWorld );
			
			TQ3TriGridData theTriGridData;
			theStatus = ::Q3TriGrid_GetData( fQD3DTriGrid, &theTriGridData );
			ThrowIf_( theStatus == kQ3Failure );

			Point 		mousePoint;
			TQ3Point3D 	theFrustumPoint;
			SDimension16 frameSize;
			this->GetFrameSize( frameSize );

			// Get the mesh point corresponding to our hit
			Int16 theClickedRow = ( theXYZPoint.z + 0.5 );
			Int16 theClickedColumn = ( theXYZPoint.x + 0.5 );

			if ( theClickedColumn > 0 && theClickedRow < fMeshCellManager->GetCellHeight() )
			{
				while( ::StillDown() )
				{
					::GetMouse( &mousePoint );	// Which is in PORT coordinates IM::Mac Toolbox Essentials 2-108
					this->PortToLocalPoint( mousePoint );
					
					::Q3Point3D_Transform( &theXYZPoint, &matrixWorldToFrustum, &theFrustumPoint );

					theFrustumPoint.y = ( (float) mousePoint.v * -2.0 / frameSize.height ) + 1.0;
					
					::Q3Point3D_Transform( &theFrustumPoint, &matrixFrustumToWorld, &theXYZPoint );
					theXYZPoint.x = theClickedColumn;
					theXYZPoint.z = theClickedRow;
					
					::Q3TriGrid_SetVertexPosition( fQD3DTriGrid, theClickedRow, theClickedColumn, &theXYZPoint );
					this->DrawSelf();
				}
				
				Int32	theMeshCols		= fMeshCellManager->GetCellWidth();
				UInt32	theMeshIndex	= ( theMeshCols - theClickedColumn + 1 ) + theClickedRow * theMeshCols;

				// Get a reference to the mesh_cell array (rather than a copy of the array)
				CellArrayT& theMeshArray = fMeshCellManager->GetCellArray();
				StArrayLocker lock( theMeshArray );

				theMeshArray[theMeshIndex].height = FLOAT_TO_WORLD( theXYZPoint.y );
				fMeshCellManager->Changed( msg_MeshDataChanged );
			}

			::Q3Object_Dispose( theCamera );
		}
		
		::Q3Pick_EmptyHitList( fQD3DPointPicker );
	}
}


// ---------------------------------------------------------------------------
// HandleSelectMarkerToolClick
// ---------------------------------------------------------------------------

void C3DMapView::HandleSelectMarkerToolClick( const SMouseDownEvent &inMouseDown )
{
	Point portPoint = inMouseDown.wherePort;
	
	if ( sClickCount > 1 )
	{
		fMeshMarkerManager->EditMarker( this->FindMarkerHitBy( portPoint ) );
	}
	else
	{
		ArrayIndexT theHitMarker = this->FindMarkerHitBy( portPoint );
		
		if ( theHitMarker != LArray::index_Bad )
		{
			MarkerArrayT&	theMarkerArray	= fMeshMarkerManager->GetMarkerArray();
			UInt32			theMaxIndex		= theMarkerArray.GetCount();

			if ( ( inMouseDown.macEvent.modifiers & shiftKey ) != 0 )
			{
				if ( theMarkerArray[theHitMarker].unused[0] == false )
					fMeshMarkerManager->SelectMarker( theHitMarker, true );
				else
					fMeshMarkerManager->UnselectMarker( theHitMarker );
			}
			else
			{
				if ( theMarkerArray[theHitMarker].unused[0] == false )
					fMeshMarkerManager->SelectMarker( theHitMarker, false );
			}
			this->DrawSelf();
			
			Point			thePreviousLoc	= { inMouseDown.wherePort.v, inMouseDown.wherePort.h };
			world_point3d	thePreviousMythLoc;
			LocalToMythPoint( thePreviousLoc, thePreviousMythLoc);

			while ( ::StillDown() )
			{
				Point			theCurrentLoc;
				world_point3d	theCurrentMythLoc;
				::GetMouse( &theCurrentLoc );
				
				if ( thePreviousLoc.v != theCurrentLoc.v || thePreviousLoc.h != theCurrentLoc.h )
				{
					if ( LocalToMythPoint( theCurrentLoc, theCurrentMythLoc) )
					{
						world_distance	theWorldDeltaX	= theCurrentMythLoc.x - thePreviousMythLoc.x;
						world_distance	theWorldDeltaY	= theCurrentMythLoc.y - thePreviousMythLoc.y;
						world_distance	theWorldDeltaZ	= theCurrentMythLoc.z - thePreviousMythLoc.z;
						
						for ( ArrayIndexT i = LArray::index_First; i <= theMaxIndex; i++ )
						{
							marker_data		theMarker	= theMarkerArray[i];

							if ( theMarker.unused[0] == true )
							{
								theMarker.location.position.x	+= theWorldDeltaX;
								theMarker.location.position.y	+= theWorldDeltaY;
								theMarker.location.position.z	+= theWorldDeltaZ;
								theMarkerArray[i] = theMarker;
							}
						}
						
						this->DrawSelf();
					}
				
					thePreviousLoc = theCurrentLoc;
					thePreviousMythLoc = theCurrentMythLoc;
				}
			}
		

			this->Refresh();
		}
		else
		{
			fMeshMarkerManager->UnselectAllMarkers();
			this->DrawSelf();
					
			StColorPenState	theState;
			Point			thePreviousLoc	= { portPoint.v, portPoint.h };
			
			while ( ::StillDown() )
			{
				Point	theCurrentLoc;
				
				::GetMouse( &theCurrentLoc );
			
				Rect	theSelectionRect	= { MIN( portPoint.v, theCurrentLoc.v ),
												MIN( portPoint.h, theCurrentLoc.h ),
												MAX( portPoint.v, theCurrentLoc.v ),
												MAX( portPoint.h, theCurrentLoc.h ) };
				
				::PenPat( &qd.gray );
			
				::FrameRect( &theSelectionRect );
					
				if ( thePreviousLoc.v != theCurrentLoc.v || thePreviousLoc.h != theCurrentLoc.h )
				{
					fMeshMarkerManager->UnselectAllMarkers();
					
					this->SelectMarkersInRect( theSelectionRect );

					this->DrawSelf();
				}
				
				thePreviousLoc = theCurrentLoc;
			}
			
			this->Refresh();
		}
	}
}


// ---------------------------------------------------------------------------
// HandlePlaceMarkerToolClick
// ---------------------------------------------------------------------------

void C3DMapView::HandlePlaceMarkerToolClick( const SMouseDownEvent &inMouseDown )
{
	marker_data	theMarker;

	Boolean validLocation = this->LocalToMythPoint( inMouseDown.wherePort, theMarker.location.position );

	if ( validLocation )
	{
		theMarker.flags						= 0;
		theMarker.type						= fMeshMarkerManager->GetCurrentMarkerType();
		theMarker.palette_index				= CToolManager::GetCurrentOption() - 1;
		theMarker.identifier				= fMeshMarkerManager->GetUniqueIdentifier( theMarker.type );
		theMarker.minimum_difficulty_level	= 0;
		theMarker.location.velocity.i		= 0;
		theMarker.location.velocity.j		= 0;
		theMarker.location.velocity.k		= 0;
		theMarker.location.height			= 0;
		theMarker.location.yaw				= 0;
		theMarker.location.pitch			= 0;
		theMarker.user_data[0]				= 0;
		theMarker.user_data[1]				= 0;
		theMarker.user_data[2]				= 0;
		theMarker.user_data[3]				= 0;
		theMarker.user_data[4]				= 0;
		theMarker.user_data[5]				= 0;
		theMarker.user_data[6]				= 0;
		theMarker.user_data[7]				= 0;
		theMarker.user_data[8]				= 0;
		theMarker.user_data[9]				= 0;
		theMarker.user_data[10]				= 0;
		theMarker.user_data[11]				= 0;
		theMarker.user_data[12]				= 0;
		theMarker.user_data[13]				= 0;
		theMarker.user_data[14]				= 0;
		theMarker.user_data[15]				= 0;
		theMarker.unused[0]					= 0;
		theMarker.unused[1]					= 0;
		theMarker.unused[2]					= 0;
		theMarker.unused[3]					= 0;
		theMarker.data_index				= 0;
		theMarker.data_identifier			= 0;
		
		fMeshMarkerManager->AddMarker( theMarker );
		
		this->Refresh();
	}
}


// ---------------------------------------------------------------------------
// HandleTerrainTypeToolClick
// ---------------------------------------------------------------------------

void C3DMapView::HandleTerrainTypeToolClick( const SMouseDownEvent &inMouseDown )
{
	Int16	theToolOption	= CToolManager::GetCurrentOption();
	Int16	theTerrainType	= kConvertToolOptionToTerrainType[theToolOption];

//	if ( sClick > 1 )
//	{
		// Terrain tool double-click; make a terrain flag dialog?
//	}
	
	Point theLastPoint = { 0, 0 };
	
	// Set stuff up for redrawing if terrain is visible
	TQ3Status 			theStatus;
	TQ3AttributeSet		theAttribute;
	if ( fRenderType == kTerrainTypeRender )
	{
		// get the terrain attribute
		theAttribute = CPatManager::GetColorAttribute(kColorMediaDwarfDepth + theTerrainType );
	}


	Point thePortPoint = inMouseDown.wherePort;
	while ( ::StillDown() )
	{
		::GetMouse( &thePortPoint );

		if ( ( thePortPoint.h != theLastPoint.h ) || ( thePortPoint.v != theLastPoint.v ) )
		{
			world_point3d	theMythPoint;
			Boolean			validPoint = this->LocalToMythPoint( thePortPoint, theMythPoint );

			if ( validPoint )
			{
				world_point2d theMythPoint2d;
				theMythPoint2d.x 	= theMythPoint.x;
				theMythPoint2d.y 	= theMythPoint.y;
				
				fMeshCellManager->SetTerrainType( theMythPoint2d, theTerrainType );

				if ( fRenderType == kTerrainTypeRender )
				{
					Int32	theFacetIndex	= 2 * ( WORLD_TO_INTEGER( theMythPoint2d.y ) * fMeshCellManager->GetCellWidth()
											+ fMeshCellManager->GetCellWidth() - WORLD_TO_INTEGER( theMythPoint2d.x ) - 1 );

					if ( ( theMythPoint2d.x ^ theMythPoint2d.y ) & WORLD_ONE )
					{
						if ( ( theMythPoint2d.y & ( WORLD_ONE - 1 ) ) > ( theMythPoint2d.x & ( WORLD_ONE - 1 ) ) )
						{
							theFacetIndex += 1;
						}
					}
					else if ( ( theMythPoint2d.y & ( WORLD_ONE - 1 ) ) < ( WORLD_ONE - ( theMythPoint2d.x & ( WORLD_ONE - 1 ) ) ) )
					{
						theFacetIndex += 1;
					}

					theStatus = ::Q3TriGrid_SetFacetAttributeSet( fQD3DTriGrid, theFacetIndex, theAttribute );
					this->DrawSelf();
				}
			}
		}
		
		theLastPoint = thePortPoint;
	}
}


/*
// ---------------------------------------------------------------------------
// PointsAreClose
// ---------------------------------------------------------------------------

Boolean C3DMapView::PointsAreClose( Point inFirstPt, Point inSecondPt ) const
{
	if ( LView::PointsAreClose( inFirstPt, inSecondPt ) )
	{
		LocalToPortPoint( inSecondPt );
		if ( fSelectedMarker > 0 && this->GetMarkerHit( inSecondPt ) == fSelectedMarker )
			return true;
	}
	
	return false;
}
*/



// ---------------------------------------------------------------------------
// LocalToMythPoint
// ---------------------------------------------------------------------------

Boolean C3DMapView::LocalToMythPoint( const Point &inPortPoint, world_point3d &ioMythPoint )
{
	Boolean outValid = false;
	
	TQ3Status theStatus;
	TQ3Point2D theQD3DPoint;
	theQD3DPoint.x = inPortPoint.h;
	theQD3DPoint.y = inPortPoint.v;
	
	theStatus = ::Q3WindowPointPick_SetPoint( fQD3DPointPicker, &theQD3DPoint );
	ThrowIf_( theStatus == kQ3Failure );
	
	::Q3View_StartPicking( fQD3DView, fQD3DPointPicker );
	ThrowIf_( theStatus == kQ3Failure );

	do
	{
		::Q3DisplayGroup_Submit( fQD3DMeshGroup, fQD3DView );
		ThrowIf_( theStatus == kQ3Failure );
	}
	while ( ::Q3View_EndPicking( fQD3DView ) == kQ3ViewStatusRetraverse );

	UInt32 theHitCount = 0L;

	theStatus = ::Q3Pick_GetNumHits( fQD3DPointPicker, &theHitCount );
	ThrowIf_( theStatus == kQ3Failure );

	if ( theHitCount > 0 )
	{
		TQ3PickDetail 	theHitDetail;
		TQ3Point3D 		theXYZPoint;

		theStatus = ::Q3Pick_GetPickDetailValidMask( fQD3DPointPicker, 0, &theHitDetail );
		ThrowIf_( theStatus == kQ3Failure );
		Assert_( theHitDetail & kQ3PickDetailMaskXYZ );

		theStatus = ::Q3Pick_GetPickDetailData( fQD3DPointPicker, 0, kQ3PickDetailMaskXYZ, &theXYZPoint );
		ThrowIf_( theStatus == kQ3Failure );
		
		// QuickDraw3D to Myth
		ioMythPoint.x = FLOAT_TO_WORLD( fMeshCellManager->GetCellWidth() - theXYZPoint.x );
		ioMythPoint.y = FLOAT_TO_WORLD( theXYZPoint.z );
		ioMythPoint.z = FLOAT_TO_WORLD( theXYZPoint.y );
		
		outValid = true;
	}
	
	::Q3Pick_EmptyHitList( fQD3DPointPicker );
	
	return outValid;
}


// ---------------------------------------------------------------------------
// FindMarkerHitBy
// ---------------------------------------------------------------------------

ArrayIndexT C3DMapView::FindMarkerHitBy( Point &inPortPoint ) const
{
	ArrayIndexT theHitMarker = LArray::index_Bad;

	MarkerArrayT&	theMarkerArray	= fMeshMarkerManager->GetMarkerArray();
	UInt32			theMaxIndex		= theMarkerArray.GetCount();

	TQ3Status theStatus;
	TQ3Point2D theQD3DPoint;
	theQD3DPoint.x = inPortPoint.h;
	theQD3DPoint.y = inPortPoint.v;
	
	theStatus = ::Q3WindowPointPick_SetPoint( fQD3DPointPicker, &theQD3DPoint );
	ThrowIf_( theStatus == kQ3Failure );
	
	::Q3View_StartPicking( fQD3DView, fQD3DPointPicker );
	ThrowIf_( theStatus == kQ3Failure );

//	StArrayLocker	lock( theMarkerArray );
	
	do
	{
		for ( ArrayIndexT i = LArray::index_First; i <= theMaxIndex; i++ )
		{
			marker_data		theMarker		= theMarkerArray[i];
			TQ3BoxData 		theMarkerBox;
			
			this->CalcMarkerBox( theMarker.location.position.x,
								  theMarker.location.position.y,
								  theMarker.location.position.z,
								  theMarkerBox );

			theStatus = ::Q3PickIDStyle_Submit ( i, fQD3DView );
			ThrowIf_( theStatus == kQ3Failure );
						
			theStatus = ::Q3Box_Submit( &theMarkerBox, fQD3DView );
			ThrowIf_( theStatus == kQ3Failure );
		}
	}
	while ( ::Q3View_EndPicking( fQD3DView ) == kQ3ViewStatusRetraverse );

	UInt32 theHitCount = 0L;

	theStatus = ::Q3Pick_GetNumHits( fQD3DPointPicker, &theHitCount );
	ThrowIf_( theStatus == kQ3Failure );

	if ( theHitCount > 0 )
	{
		TQ3PickDetail theHitDetail;

		theStatus = ::Q3Pick_GetPickDetailValidMask( fQD3DPointPicker, 0, &theHitDetail );
		ThrowIf_( theStatus == kQ3Failure );
		Assert_( theHitDetail & kQ3PickDetailMaskPickID );

		UInt32 thePickID = 0L;
		theStatus = ::Q3Pick_GetPickDetailData( fQD3DPointPicker, 0, kQ3PickDetailMaskPickID, &thePickID );
		ThrowIf_( theStatus == kQ3Failure );
		
		theHitMarker = thePickID;
	}
	
	::Q3Pick_EmptyHitList( fQD3DPointPicker );
	
	return theHitMarker;
}


// ---------------------------------------------------------------------------
// SelectMarkersInRect
// ---------------------------------------------------------------------------

void C3DMapView::SelectMarkersInRect( const Rect &inPortRect )
{
	MarkerArrayT&	theMarkerArray	= fMeshMarkerManager->GetMarkerArray();
	UInt32			theMaxIndex		= theMarkerArray.GetCount();

	TQ3Status theStatus;
	TQ3Area theQD3DRect;
	theQD3DRect.min.x = topLeft(inPortRect).h;
	theQD3DRect.min.y = topLeft(inPortRect).v;
	theQD3DRect.max.x = botRight(inPortRect).h;
	theQD3DRect.max.y = botRight(inPortRect).v;
	
	theStatus = ::Q3WindowRectPick_SetRect( fQD3DRectPicker, &theQD3DRect );
	ThrowIf_( theStatus == kQ3Failure );
	
	::Q3View_StartPicking( fQD3DView, fQD3DRectPicker );
	ThrowIf_( theStatus == kQ3Failure );

	do
	{
		for ( ArrayIndexT i = LArray::index_First; i <= theMaxIndex; i++ )
		{
			marker_data		theMarker		= theMarkerArray[i];
			TQ3PointData	theMarkerPoint;

			theMarkerPoint.pointAttributeSet = NULL;
			
			float tempX = WORLD_TO_FLOAT( theMarker.location.position.x );
			float tempY = WORLD_TO_FLOAT( theMarker.location.position.y );
			float tempZ = WORLD_TO_FLOAT( theMarker.location.position.z );
			float mirroredXPos = fMeshCellManager->GetCellWidth() - tempX;

			// The marker Z axis is equivalent to QD3D's Y axis and vice versa
			::Q3Point3D_Set( &theMarkerPoint.point, mirroredXPos, tempZ, tempY );

			theStatus = ::Q3PickIDStyle_Submit ( i, fQD3DView );
			ThrowIf_( theStatus == kQ3Failure );
						
			theStatus = ::Q3Point_Submit( &theMarkerPoint, fQD3DView );
			ThrowIf_( theStatus == kQ3Failure );
		}
	}
	while ( ::Q3View_EndPicking( fQD3DView ) == kQ3ViewStatusRetraverse );

	UInt32 theHitCount = 0L;

	theStatus = ::Q3Pick_GetNumHits( fQD3DRectPicker, &theHitCount );
	ThrowIf_( theStatus == kQ3Failure );

	for (UInt16 hitIndex = 0; hitIndex < theHitCount; hitIndex++ )
	{
		TQ3PickDetail theHitDetail;

		theStatus = ::Q3Pick_GetPickDetailValidMask( fQD3DRectPicker, hitIndex, &theHitDetail );
		ThrowIf_( theStatus == kQ3Failure );
		Assert_( theHitDetail & kQ3PickDetailMaskPickID );

		UInt32 thePickID = 0L;
		theStatus = ::Q3Pick_GetPickDetailData( fQD3DRectPicker, hitIndex, kQ3PickDetailMaskPickID, &thePickID );
		ThrowIf_( theStatus == kQ3Failure );
		
		fMeshMarkerManager->SelectMarker( thePickID, true );
	}
	
	::Q3Pick_EmptyHitList( fQD3DRectPicker );
}


// ---------------------------------------------------------------------------
// SelectVerticesInRect
// ---------------------------------------------------------------------------

void C3DMapView::SelectVerticesInRect( const Rect &inPortRect )
{
	TQ3Status theStatus;
	TQ3Area theQD3DRect;
	theQD3DRect.min.x = topLeft(inPortRect).h;
	theQD3DRect.min.y = topLeft(inPortRect).v;
	theQD3DRect.max.x = botRight(inPortRect).h;
	theQD3DRect.max.y = botRight(inPortRect).v;
	
	theStatus = ::Q3WindowRectPick_SetRect( fQD3DRectPicker, &theQD3DRect );
	ThrowIf_( theStatus == kQ3Failure );
	
	::Q3View_StartPicking( fQD3DView, fQD3DRectPicker );
	ThrowIf_( theStatus == kQ3Failure );
	
	TQ3TriGridData theTrigridData;
	theStatus = ::Q3TriGrid_GetData( fQD3DTriGrid, &theTrigridData );
	ThrowIf_( theStatus == kQ3Failure );

	do
	{
		UInt32 vertexCount = theTrigridData.numRows * theTrigridData.numColumns;
		TQ3PointData thePointData;
		thePointData.pointAttributeSet = NULL;

		for ( UInt32 i = 0; i < vertexCount; i++ )
		{
			theStatus = ::Q3PickIDStyle_Submit ( i, fQD3DView );
			ThrowIf_( theStatus == kQ3Failure );
			
			thePointData.point = theTrigridData.vertices[i].point;
			theStatus = ::Q3Point_Submit( &thePointData, fQD3DView );
			ThrowIf_( theStatus == kQ3Failure );
		}
	}
	while ( ::Q3View_EndPicking( fQD3DView ) == kQ3ViewStatusRetraverse );
	
	// Release the trigrid data or leak like Netscape...
	theStatus = ::Q3TriGrid_EmptyData( &theTrigridData );
	ThrowIf_( theStatus == kQ3Failure );

	UInt32 theHitCount = 0L;

	theStatus = ::Q3Pick_GetNumHits( fQD3DRectPicker, &theHitCount );
	ThrowIf_( theStatus == kQ3Failure );

	for (UInt16 hitIndex = 0; hitIndex < theHitCount; hitIndex++ )
	{
		TQ3PickDetail theHitDetail;

		theStatus = ::Q3Pick_GetPickDetailValidMask( fQD3DRectPicker, hitIndex, &theHitDetail );
		ThrowIf_( theStatus == kQ3Failure );
		Assert_( theHitDetail & kQ3PickDetailMaskPickID );

		UInt32 thePickID = 0L;
		theStatus = ::Q3Pick_GetPickDetailData( fQD3DRectPicker, hitIndex, kQ3PickDetailMaskPickID, &thePickID );
		ThrowIf_( theStatus == kQ3Failure );
		
//		fMeshMarkerManager->SelectMarker( thePickID, true );
	}
	
	::Q3Pick_EmptyHitList( fQD3DRectPicker );
}


// ---------------------------------------------------------------------------
// CalcModelTranslation
// ---------------------------------------------------------------------------

inline void C3DMapView::CalcModelTranslation( long inMarkerX, long inMarkerY, long inMarkerZ, TQ3Matrix4x4 &outMatrix ) const
{
	float transX = WORLD_TO_FLOAT( inMarkerX );
	float transY = WORLD_TO_FLOAT( inMarkerY );
	float transZ = WORLD_TO_FLOAT( inMarkerZ );
	float mirroredXPos = fMeshCellManager->GetCellWidth() - transX;

	// The marker Z axis is equivalent to QD3D's Y axis and vice versa
	::Q3Matrix4x4_SetTranslate( &outMatrix, mirroredXPos, transZ, transY );
}


// ---------------------------------------------------------------------------
// CalcModelRotation
// ---------------------------------------------------------------------------

inline void C3DMapView::CalcModelRotation( long inMarkerAngle, TQ3Matrix4x4 &outMatrix ) const
{
	NORMALIZE_ANGLE( inMarkerAngle );
	float rotationAngle = inMarkerAngle / FULL_CIRCLE;
	rotationAngle = ((float)((rotationAngle) * kQ3Pi / 180.0));		// rotationAngle = ::Q3Math_DegreesToRadians( rotationAngle );

	::Q3Matrix4x4_SetRotate_Y( &outMatrix, rotationAngle );
}


// ---------------------------------------------------------------------------
// CalcMarkerBox
// ---------------------------------------------------------------------------

void C3DMapView::CalcMarkerBox( long inMarkerX, long inMarkerY, long inMarkerZ, TQ3BoxData &outBox ) const
{
	outBox.faceAttributeSet = NULL;
	outBox.boxAttributeSet 	= NULL;
	
	::Q3Vector3D_Set( &outBox.orientation, 0.0, 0.2, 0.0 );
	::Q3Vector3D_Set( &outBox.minorAxis, 0.2, 0.0, 0.0 );
	::Q3Vector3D_Set( &outBox.majorAxis, 0.0, 0.0, 0.2 );
	
	float tempX = WORLD_TO_FLOAT( inMarkerX );
	float tempY = WORLD_TO_FLOAT( inMarkerY );
	float tempZ = WORLD_TO_FLOAT( inMarkerZ );
	float mirroredXPos = fMeshCellManager->GetCellWidth() - tempX;

	// The marker Z axis is equivalent to QD3D's Y axis and vice versa
	::Q3Point3D_Set( &outBox.origin, mirroredXPos - 0.1, tempZ, tempY - 0.1 );
}


// ---------------------------------------------------------------------------
// GetColorAttribute
// ---------------------------------------------------------------------------

TQ3AttributeSet C3DMapView::GetColorAttribute( Boolean inIsSelected, Boolean inIsVisible, Int16 inTeamIndex )
{
	if ( inIsSelected )
	{
		return CPatManager::GetColorAttribute( kColorSelected );
	}
	else
	{
		switch ( inTeamIndex )
		{
			case 0:
			case 1:
			case 2:
			case 3:
			case 4:
				return CPatManager::GetColorAttribute( inIsVisible ? kColorTeam1 + inTeamIndex : kColorTeamInvis1 + inTeamIndex );
				break;
				
			default:
				return CPatManager::GetColorAttribute( inIsVisible ? kColorGray5 : kColorWhite );
				break;
		}
	}
}


// ---------------------------------------------------------------------------
//	HandleKeyPress (with TestKeyBit	utility)		
// ---------------------------------------------------------------------------

inline Boolean TestKeyBit(KeyMap& inKeyMap, UInt16 inTestBit)
{
	UInt8* theRealArray = (UInt8*) inKeyMap;
	
	return (theRealArray[inTestBit / 8] & (0x01 << (inTestBit % 8)));	
}

const UInt16 kKey_Control 	= 0x3B;
const UInt16 kKey_Shift		= 0x38;

Boolean C3DMapView::HandleKeyPress( const EventRecord&	inKeyEvent )
{
	Boolean		keyHandled 	= false;
	
	if ( inKeyEvent.modifiers & cmdKey )
	{
		keyHandled = LCommander::HandleKeyPress( inKeyEvent );
	}
	else
	{
		KeyMap theKeyState;
		Boolean movementKey = false;

		do
		{
			::GetKeys( theKeyState );
			movementKey 		= false;
			float modifierScale = 1.0;

			if ( TestKeyBit(theKeyState, kKey_Control) ) {
				modifierScale *= 2.0;
			}
			if ( TestKeyBit(theKeyState, kKey_Shift) ) {
				modifierScale *= 3.0;
			}

			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_PanLeft]) ) {
				CameraPan(3.0 * modifierScale);
				movementKey = true;
			}
							
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_PanRight]) ) {
				CameraPan(-3.0 * modifierScale);
				movementKey = true;
			}

			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_OrbitLeft]) ) {
				CameraOrbit(-3.0 * modifierScale);
				movementKey = true;
			}
				
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_OrbitRight]) ) {
				CameraOrbit(3.0 * modifierScale);
				movementKey = true;
			}

			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_OrbitUp]) ) {
				CameraOrbit(0.0, 3.0 * modifierScale);
				movementKey = true;
			}
				
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_OrbitDown]) ) {
				CameraOrbit(0.0, -3.0 * modifierScale);
				movementKey = true;
			}

			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_ZoomIn]) ) {
				CameraZoom(0.3 * modifierScale);
				movementKey = true;
			}
				
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_ZoomOut]) ) {
				CameraZoom(-0.3 * modifierScale);
				movementKey = true;
			}
				
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_MoveForward]) ) {
				CameraMove(0.3 * modifierScale, 0.0);
				movementKey = true;
			}
				
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_MoveBack]) ) {
				CameraMove(-0.3 * modifierScale, 0.0);
				movementKey = true;
			}
				
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_MoveLeft]) ) {
				CameraMove(0.0, 0.3 * modifierScale);
				movementKey = true;
			}
				
			if ( TestKeyBit(theKeyState, fCameraKeys[kCamera_MoveRight]) ) {
				CameraMove(0.0, -0.3 * modifierScale);
				movementKey = true;
			}
			 
			if ( movementKey )
			{
				this->Refresh();
				this->UpdatePort();
//				this->DrawSelf();
				keyHandled = true;
			}
		}
		while ( movementKey );
		
		if ( !keyHandled )
			keyHandled = LCommander::HandleKeyPress( inKeyEvent );
	}
	
	return keyHandled;
}


// ---------------------------------------------------------------------------
//	CameraOrbit						
// ---------------------------------------------------------------------------

void C3DMapView::CameraOrbit( float inDegrees, float inVerticalDegrees )
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera( fQD3DView, &theCamera );
	ThrowIf_( status == kQ3Failure );

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement( theCamera, &theCameraPlacement );
	ThrowIf_( status == kQ3Failure );

	float xAngle = 0.0;
	float yAngle = Q3Math_DegreesToRadians( inDegrees );
	float zAngle = 0.0;
	
	TQ3Matrix4x4 cameraMatrix;
	::Q3Matrix4x4_SetRotateAboutPoint( &cameraMatrix, &theCameraPlacement.pointOfInterest, xAngle, yAngle, zAngle );
	
	::Q3Point3D_Transform( &theCameraPlacement.cameraLocation, &cameraMatrix, &theCameraPlacement.cameraLocation );

	if ( inVerticalDegrees != 0.0 )
	{
		float verticalAngle = Q3Math_DegreesToRadians( inVerticalDegrees );
		TQ3Vector3D theCameraVector;
		::Q3Point3D_Subtract( &theCameraPlacement.pointOfInterest, &theCameraPlacement.cameraLocation, &theCameraVector );
		
		TQ3Vector3D theOrthogonalVector;
		theOrthogonalVector.x = theCameraVector.z;
		theOrthogonalVector.y = 0.0;
		theOrthogonalVector.z = 0.0 - theCameraVector.x;
		::Q3Vector3D_Normalize( &theOrthogonalVector, &theOrthogonalVector );

		TQ3Matrix4x4 verticalMatrix;
		::Q3Matrix4x4_SetRotateAboutAxis( &verticalMatrix, &theCameraPlacement.pointOfInterest, &theOrthogonalVector, verticalAngle );

		::Q3Point3D_Transform( &theCameraPlacement.cameraLocation, &verticalMatrix, &theCameraPlacement.cameraLocation );
	}

	status = ::Q3Camera_SetPlacement( theCamera, &theCameraPlacement );
	ThrowIf_( status == kQ3Failure );
	
	::Q3Object_Dispose( theCamera );
}


// ---------------------------------------------------------------------------
//	CameraPan						
// ---------------------------------------------------------------------------

void C3DMapView::CameraPan( float inDegrees )
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera( fQD3DView, &theCamera );
	ThrowIf_( status == kQ3Failure );

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement( theCamera, &theCameraPlacement );
	ThrowIf_( status == kQ3Failure );

	float xAngle = 0.0;
	float yAngle = Q3Math_DegreesToRadians( inDegrees );
	float zAngle = 0.0;
	
	TQ3Matrix4x4 cameraMatrix;
	::Q3Matrix4x4_SetRotateAboutPoint( &cameraMatrix, &theCameraPlacement.cameraLocation, xAngle, yAngle, zAngle );
	::Q3Point3D_Transform( &theCameraPlacement.pointOfInterest, &cameraMatrix, &theCameraPlacement.pointOfInterest );

	status = ::Q3Camera_SetPlacement( theCamera, &theCameraPlacement );
	ThrowIf_( status == kQ3Failure );
	
	::Q3Object_Dispose( theCamera );
}


// ---------------------------------------------------------------------------
//	CameraZoom						
// ---------------------------------------------------------------------------

void C3DMapView::CameraZoom( float inDistance )
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera( fQD3DView, &theCamera );
	ThrowIf_( status == kQ3Failure );

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement( theCamera, &theCameraPlacement );
	ThrowIf_( status == kQ3Failure );

	TQ3Vector3D theCameraVector;
	::Q3Point3D_Subtract( &theCameraPlacement.pointOfInterest, &theCameraPlacement.cameraLocation, &theCameraVector );
	
	float theCameraDistance = ::Q3Vector3D_Length( &theCameraVector );
	
	if ( inDistance < theCameraDistance )
	{
		::Q3Vector3D_Normalize( &theCameraVector, &theCameraVector );
		::Q3Vector3D_Scale( &theCameraVector, inDistance, &theCameraVector );
		
		::Q3Point3D_Vector3D_Add( &theCameraPlacement.cameraLocation, &theCameraVector, &theCameraPlacement.cameraLocation );
		
		status = ::Q3Camera_SetPlacement( theCamera, &theCameraPlacement );
		ThrowIf_( status == kQ3Failure );
	}
	
	::Q3Object_Dispose( theCamera );
}


// ---------------------------------------------------------------------------
//	CameraMove						
// ---------------------------------------------------------------------------

void C3DMapView::CameraMove( float inDistance, float inOrthogonalDistance )
{
	TQ3CameraObject theCamera;
	TQ3Status status = ::Q3View_GetCamera( fQD3DView, &theCamera );
	ThrowIf_( status == kQ3Failure );

	TQ3CameraPlacement theCameraPlacement;
	status = ::Q3Camera_GetPlacement( theCamera, &theCameraPlacement );
	ThrowIf_( status == kQ3Failure );

	TQ3Vector3D theCameraVector;
	::Q3Point3D_Subtract( &theCameraPlacement.pointOfInterest, &theCameraPlacement.cameraLocation, &theCameraVector );
	theCameraVector.y = 0.0;
	
	TQ3Vector3D theOrthogonalVector;
	::Q3Vector3D_Normalize( &theCameraVector, &theCameraVector );
	theOrthogonalVector.x = theCameraVector.z;
	theOrthogonalVector.y = 0.0;
	theOrthogonalVector.z = 0.0 - theCameraVector.x;
	
	::Q3Vector3D_Scale( &theCameraVector, inDistance, &theCameraVector );
	::Q3Vector3D_Scale( &theOrthogonalVector, inOrthogonalDistance, &theOrthogonalVector );
	::Q3Vector3D_Add( &theCameraVector, &theOrthogonalVector, &theCameraVector );
		
	::Q3Point3D_Vector3D_Add( &theCameraPlacement.pointOfInterest, &theCameraVector, &theCameraPlacement.pointOfInterest );
	::Q3Point3D_Vector3D_Add( &theCameraPlacement.cameraLocation, &theCameraVector, &theCameraPlacement.cameraLocation );
		
	status = ::Q3Camera_SetPlacement( theCamera, &theCameraPlacement );
	ThrowIf_( status == kQ3Failure );
	
	::Q3Object_Dispose( theCamera );
}