// =================================================================================
//	CcommandDocument.cp				1997 eli curtz. All rights reserved.
// =================================================================================

#include <LFile.h>
#include <LPlaceHolder.h>
#include <LPrintout.h>
#include <LString.h>
#include <LWindow.h>
#include <PP_Messages.h>
#include <UMemoryMgr.h>
#include <UWindows.h>

#include "Constants.h"

#include "CcommandDocument.h"


Int32
LCommandComparator::Compare(
	const void*		inItemOne,
	const void*		inItemTwo,
	Uint32			/* inSizeOne */,
	Uint32			/* inSizeTwo */) const
{
	command_Storage * command1 = (command_Storage *) inItemOne;
	command_Storage * command2 = (command_Storage *) inItemTwo;

	return ( command1->theCommandID - command2->theCommandID );
}

// ---------------------------------------------------------------------------------
//		 CcommandDocument(LCommander*, FSSpec*)
// ---------------------------------------------------------------------------------


CcommandDocument::CcommandDocument(
	LCommander	*inSuper,
	FSSpec		*inFileSpec )
		: LSingleDoc( inSuper ), LListener()
{
	mCommandArray = nil;
	
	// Create window for our document.
	mWindow = LWindow::CreateWindow( rPPob_Window, this );
	Assert_( mWindow != nil );
	
	// Find the active panes
	mOutputView = (LTextEditView *) mWindow->FindPaneByID( kMainView );
	Assert_( mOutputView != nil );
	
	if ( inFileSpec == nil ) {

		// Set the name of new window.
		NameNewDoc();
		
	} else {
	
		// Display contents of file in window.
		OpenFile( *inFileSpec );
	
	}
	
	// Make the window visible.
	mWindow->Show();
}

CcommandDocument::~CcommandDocument()
{
	delete mCommandArray;
}


// ---------------------------------------------------------------------------------
//		 NameNewDoc
// ---------------------------------------------------------------------------------

void
CcommandDocument::NameNewDoc()
{
	// Setup the window title. Start with the default title.
	LStr255	theTitle( "\pUntitled" );

	// Find the first available title. We could also check the window
	// pane id if we wanted to make sure we didn't collide with other
	// window types.
	Int32	theNumber = 0;
	while ( UWindows::FindNamedWindow( theTitle ) != nil ) {

		// An existing window has the current name
		// Increment counter and try again.
		++theNumber;
		theTitle = "\pUntitled ";
		theTitle += (LStr255) theNumber;

	}		
	
	// Finally, set window title.
	mWindow->SetDescriptor( theTitle );
}


// ---------------------------------------------------------------------------------
//		 OpenFile
// ---------------------------------------------------------------------------------

void
CcommandDocument::OpenFile(
	FSSpec	&inFileSpec )
{
	Try_ {

		// Create a new file object.
		mFileStream = new LFileStream( inFileSpec );
		mFile = mFileStream;	// In case the superclass is using it
		
		// Open the data fork.
		mFileStream->OpenDataFork( fsRdPerm );
		
		// Set the window title to the name of the file and
		// flag that the document has an associated file.
		mWindow->SetDescriptor( inFileSpec.name );
		mIsSpecified = true;

		// Skip the file header and read the mesh header
		dir_Header theFileHeader;
		Int32 byteCount = sizeof(dir_Header);
		mFileStream->GetBytes(&theFileHeader, byteCount);
		Assert_(theFileHeader.group == 'mesh');
		
		byteCount = sizeof(mesh_Header);
		mFileStream->GetBytes(&mMeshHeader, byteCount);
		mHeaderOffset = sizeof(dir_Header) + mMeshHeader.headerSize;
		
		mCommandArray = new LArray(sizeof(command_Storage));
		mCommandArray->AdjustAllocation(mMeshHeader.commandCount);

		mFileStream->SetMarker(mHeaderOffset + mMeshHeader.commandOffset, streamFrom_Start);
		command_Storage theNewCommand;
		for (Int16 commandCount = 0; commandCount < mMeshHeader.commandCount; commandCount++) {
			byteCount = sizeof(mesh_Command);
			mFileStream->GetBytes(&theNewCommand.theCommand, byteCount);
			theNewCommand.theCommandID = theNewCommand.theCommand.commandLabel;
			theNewCommand.theCommandName = nil;
			mCommandArray->InsertItemsAt(1, LArray::index_Last, &theNewCommand);
		}
		mCommandArray->SetComparator(new LCommandComparator);
		mCommandArray->SetKeepSorted(false);	// We want to keep the same order for labels
		
		Int32 argumentSize = mMeshHeader.commandSize - (mMeshHeader.commandCount * sizeof(mesh_Command));
		Ptr argumentData = ::NewPtr(argumentSize);
		ThrowIfMemFail_(argumentData);

		byteCount = argumentSize;
		mFileStream->GetBytes(argumentData, byteCount);

		mCommandArray->Lock();
		
		command_Storage* theCurrentCommand;
		Ptr theArgs = nil;
		mesh_CmdArgument * theTempArgs;
		for (Int16 commandCount = 1; commandCount <= mCommandArray->GetCount(); commandCount++) {
			theCurrentCommand = (command_Storage*) mCommandArray->GetItemPtr(commandCount);
			theArgs = argumentData + theCurrentCommand->theCommand.argDataOffset;
			for (Int16 argumentCount = 0; argumentCount < theCurrentCommand->theCommand.argumentCount; argumentCount++) {
				theTempArgs = (mesh_CmdArgument *) theArgs;
				if (theTempArgs->argumentType == 'name') {
					if (theTempArgs->dataUnitCount < 64)
						theCurrentCommand->theCommandName = new TString<Str63>(&theTempArgs->data[0], theTempArgs->dataUnitCount);
					else if (theTempArgs->dataUnitCount < 256)
						theCurrentCommand->theCommandName = new TString<Str255>(&theTempArgs->data[0], theTempArgs->dataUnitCount);
					else
						SignalPStr_("\pLonger name than expected.");
					break;
				}
				Int16 dataUnitSize;
				switch (theTempArgs->dataUnitType) {
					case 0:		// flags ?
					case 2:
					case 3:		// "action_identifier"
					case 4:		// "angle"
					case 14:	// "object_Identifier"
					case 15:	// "model_Identifier"
					case 16:	// "sound_source_identifier"
					case 18:	// "local_projectile_group_identifier"
						dataUnitSize = sizeof(Int16);
						break;

					case 1:
						dataUnitSize = sizeof(char);
						break;

					case 5:
						dataUnitSize = sizeof(Int32);
						break;

					case 6:		// "world_distance"
					case 8:
						dataUnitSize = sizeof(Fixed);
						break;

					case 7:		// "field_name"
					case 11:
						dataUnitSize = sizeof(FourCharCode);
						break;
						
					case 12:	// "world_point2d"
						dataUnitSize = 2 * sizeof(Fixed);
						break;

					default:
Assert_(false);
				}
				theArgs += (theTempArgs->dataUnitCount * dataUnitSize);
				theArgs += sizeof(mesh_CmdArgument);
				theArgs += ((Int32) theArgs & 0x00000001);
				theArgs += ((Int32) theArgs & 0x00000002);
			}
			if (theCurrentCommand->theCommandName == nil) {
				long scratch = theCurrentCommand->theCommandID;
				theCurrentCommand->theCommandName = new TString<Str15>(scratch);
			}
		}

		LHandleStream tempStorage;
		LStr255 scratchString;
		tempStorage.SetLength(32000);
		for (Int16 commandCount = 1; commandCount <= mCommandArray->GetCount(); commandCount++) {
			theCurrentCommand = (command_Storage*) mCommandArray->GetItemPtr(commandCount);
			theArgs = argumentData + theCurrentCommand->theCommand.argDataOffset;
			if (theCurrentCommand->theCommand.commandType == -1)
				scratchString = "\p__Label__";
			else
				scratchString = theCurrentCommand->theCommand.commandType;
			tempStorage.WriteBlock(&scratchString[1], scratchString[0]);
			tempStorage << '\t';
/*
			long scratchID = theCurrentCommand->theCommandID;
			scratchString = scratchID;
			tempStorage.WriteBlock(&scratchString[1], scratchString[0]);
			tempStorage << '\t';
*/
			scratchString = *theCurrentCommand->theCommandName;
			tempStorage.WriteBlock(&scratchString[1], scratchString[0]);
			tempStorage << "\p {\r";
			
			for (Int16 argumentCount = 0; argumentCount < theCurrentCommand->theCommand.argumentCount; argumentCount++) {
				theTempArgs = (mesh_CmdArgument *) theArgs;
				if (theTempArgs->argumentType != 'name') {
					tempStorage << '\t';
					scratchString = theTempArgs->argumentType;
					tempStorage.WriteBlock(&scratchString[1], scratchString[0]);
					tempStorage << '\t';
				}

				Int16 dataUnitSize;
				Ptr	 dataSubPtr = (Ptr) theTempArgs->data;
				scratchString = "\p";
				for (Int16 dataListCount = 1; dataListCount <= theTempArgs->dataUnitCount; dataListCount++) {
					switch (theTempArgs->dataUnitType) {
						case 0:
							dataUnitSize = sizeof(Int16);	// flags ?
							Int32 temp32 = *((Int16 *) dataSubPtr);
							scratchString += temp32;
							break;
						case 1:
							dataUnitSize = sizeof(char);
							scratchString.Append(dataSubPtr, theTempArgs->dataUnitCount);
							break;
						case 2:
							dataUnitSize = sizeof(Int16);
							temp32 = *((Int16 *) dataSubPtr);
							scratchString += temp32;
							break;
						case 3:
							dataUnitSize = sizeof(Int16);	// "action_identifier"
							command_Storage scratchCommand;
							scratchCommand.theCommandID = *((Int16 *) dataSubPtr);
							Int32 arrayIndex = mCommandArray->FetchIndexOf(&scratchCommand);
							if (mCommandArray->ValidIndex(arrayIndex)) {
								scratchString += *((command_Storage*) mCommandArray->GetItemPtr(arrayIndex))->theCommandName;
							}
							else {
								SignalPStr_("\pCouldn't find action_identifier");
								// Assert_(false);
							}
							break;
						case 4:
							dataUnitSize = sizeof(Int16);	// "angle"
							temp32 = *((Int16 *) dataSubPtr);
							scratchString += temp32;
							break;
						case 5:
							dataUnitSize = sizeof(Int32);
							scratchString += *((Int32 *) dataSubPtr);
							break;
						case 6:
							dataUnitSize = sizeof(Fixed);	// "world_distance"
							scratchString += *((Fixed *) dataSubPtr);
							break;
						case 7:
							dataUnitSize = sizeof(FourCharCode);	// "field_name"
							scratchString += TString<Str15>(*((FourCharCode *) dataSubPtr));
							break;
						case 8:
							dataUnitSize = sizeof(Fixed);
							scratchString += *((Fixed *) dataSubPtr);
							break;
						case 11:
							dataUnitSize = sizeof(FourCharCode);
							scratchString += TString<Str15>(*((FourCharCode *) dataSubPtr));
							break;
						case 12:
							dataUnitSize = 2 * sizeof(Fixed);		// "world_point2d"
							scratchString += '(';
							scratchString += *((Fixed *) dataSubPtr);
							scratchString += "\p, ";
							scratchString += *((Fixed *) (dataSubPtr + sizeof(Fixed)));
							scratchString += ')';
							break;
						case 14:
							dataUnitSize = sizeof(Int16);	// "object_Identifier"
							temp32 = *((Int16 *) dataSubPtr);
							scratchString += temp32;
							break;
						case 15:
							dataUnitSize = sizeof(Int16);	// "model_Identifier"
							temp32 = *((Int16 *) dataSubPtr);
							scratchString += temp32;
							break;
						case 16:
							dataUnitSize = sizeof(Int16);	// "sound_source_identifier"
							temp32 = *((Int16 *) dataSubPtr);
							scratchString += temp32;
							break;
						case 18:
							dataUnitSize = sizeof(Int16);	// "local_projectile_group_identifier"
							temp32 = *((Int16 *) dataSubPtr);
							scratchString += temp32;
							break;
							
							
						default:
	Assert_(false);
					}
					
					dataSubPtr += dataUnitSize;
					if (dataListCount < theTempArgs->dataUnitCount)
						scratchString += "\p,\t";
				}
				theArgs += (theTempArgs->dataUnitCount * dataUnitSize);
				theArgs += sizeof(mesh_CmdArgument);
				theArgs += ((Int32) theArgs & 0x00000001);
				theArgs += ((Int32) theArgs & 0x00000002);
				
				if (theTempArgs->argumentType != 'name') {
					tempStorage.WriteBlock(&scratchString[1], scratchString[0]);
					tempStorage << '\r';
				}
			}
			tempStorage << "}\r\r";
		}

		mCommandArray->Unlock();
		
		tempStorage.SetLength(tempStorage.GetMarker());
		mOutputView->SetTextHandle(tempStorage.GetDataHandle());

// read the markers to try and figure them out
		mFileStream->SetMarker(mHeaderOffset + mMeshHeader.objectTypeOffset, streamFrom_Start);
		mesh_ObjectType theNewObjectType;
		for (Int16 objectTypeCount = 0; objectTypeCount < mMeshHeader.objectTypeCount; objectTypeCount++) {
			byteCount = sizeof(mesh_ObjectType);
			mFileStream->GetBytes(&theNewObjectType, byteCount);
			switch (theNewObjectType.typeFlag) {
				case 0:
				case 1:
				case 3:
				case 5:
				case 6:
				case 9:
				case 10:
					break;
					
				default:
					Assert_(false);
			}
		}



	} Catch_( inErr ) {

		// Cleanup and rethrow the error.
		delete this;
		Throw_( inErr );
	
	} EndCatch_
}


// ---------------------------------------------------------------------------------
//		 IsModified
// ---------------------------------------------------------------------------------

Boolean
CcommandDocument::IsModified()
{
	return false;
}


// ---------------------------------------------------------------------------------
//		 DoAESave
// ---------------------------------------------------------------------------------

void
CcommandDocument::DoAESave(
	FSSpec	& /*  inFileSpec */,
	OSType	/* inFileType */)
{
/*
	// Delete the existing file object.
	// Note: this does nothing to the actual file on disk.
	delete mFileStream;
	
	// Make a new file object.
	mFileStream = new LFile( inFileSpec );
	
	// Get the proper file type.
	OSType	theFileType = 'TEXT';
	if ( inFileType != fileType_Default ) theFileType = inFileType;

	// Make new file on disk (we'll use
	// SimpleText's creator for this example).
	mFileStream->CreateNewDataFile( 'ttxt', theFileType );
	
	// Write out the data.
	DoSave();

	// Change window title to reflect the new name.
	mWindow->SetDescriptor( inFileSpec.name );

	// Document now has a specified file.
	mIsSpecified = true;
*/
}


// ---------------------------------------------------------------------------------
//		 DoSave
// ---------------------------------------------------------------------------------

void
CcommandDocument::DoSave()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Get the text from the text view.
	Handle	theTextH = mTextView->GetTextHandle();
	
	// Lock the text handle.
	StHandleLocker	theLock( theTextH );
	
	// Write the text to the file.
	mFileStream->WriteDataFork( *theTextH, ::GetHandleSize( theTextH ) );

	// Close the data fork.
	mFileStream->CloseDataFork();

	// Saving makes doc un-dirty.
	mTextView->SetDirty( false );
*/
}


// ---------------------------------------------------------------------------------
//		 DoRevert
// ---------------------------------------------------------------------------------

void
CcommandDocument::DoRevert()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Read the entire file contents and close the file.
	Handle	theTextH = mFileStream->ReadDataFork();
	mFileStream->CloseDataFork();
	
	// Put the contents in the text view
	// and clear the dirty flag.
	mTextView->SetTextHandle( theTextH );
	mTextView->SetDirty( false );
	
	// Dispose of the text.
	::DisposeHandle( theTextH );

	// Refresh the text view.
	mTextView->Refresh();
*/
}


// ---------------------------------------------------------------------------------
//		 DoPrint
// ---------------------------------------------------------------------------------

void
CcommandDocument::DoPrint()
{
/*
	// Create the printout.
	LPrintout	*thePrintout = LPrintout::CreatePrintout( rPPob_TextPrintout );
	Assert_( thePrintout != nil );
	
	// Set the print record.
	
	if (mPrintRecordH)
	{
		thePrintout->SetPrintRecord( mPrintRecordH );
	}
	
	// Get the text placeholder.
	LPlaceHolder	*thePlaceholder;
	thePlaceholder = (LPlaceHolder*) thePrintout->FindPaneByID( kTextPlaceholder );
	Assert_( thePlaceholder != nil );
	
	// Install the text view in the placeholder.
	thePlaceholder->InstallOccupant( mTextView, atNone );
	
	// Set the frame size.
	SetPrintFrameSize();
	
	// Print.
	thePrintout->DoPrintJob();
	
	// Delete the printout. The text view is returned
	// to the window when the placeholder is destroyed.
	delete thePrintout;
*/
}


// ---------------------------------------------------------------------------------
//		 SetPrintFrameSize
// ---------------------------------------------------------------------------------

void
CcommandDocument::SetPrintFrameSize()
{
/*
	// Get the frame size.
	SDimension16	theFrameSize;
	mTextView->GetFrameSize( theFrameSize );
	
	// Get the text edit record handle.
	TEHandle	theTextEditH = mTextView->GetMacTEH();
	
	// Calculate the number of lines per page.
	Int16	theLinesPerPage;
	theLinesPerPage = theFrameSize.height / (**theTextEditH).lineHeight;

	// Resize the frame to an integral number of lines.
	mTextView->ResizeFrameTo( theFrameSize.width,
		(**theTextEditH).lineHeight * theLinesPerPage, false );

*/
}


void
CcommandDocument::ListenToMessage(
	MessageT /* inMessage */,
	void * /* ioParam */)
{
}
