// =================================================================================
//	CmeshDocument.cp				1997 eli curtz. All rights reserved.
// =================================================================================

#include <LFile.h>
#include <LPlaceHolder.h>
#include <LPrintout.h>
#include <LString.h>
#include <LWindow.h>
#include <PP_Messages.h>
#include <UMemoryMgr.h>
#include <UWindows.h>

#include <QD3DGroup.h>
#include <QD3DGeometry.h>
#include <QD3DStorage.h>
#include <QD3DShader.h>

#include "Constants.h"
#include "CmeshView.h"

#include "CmeshDocument.h"

// ---------------------------------------------------------------------------------
//		 CmeshDocument(LCommander*, FSSpec*)
// ---------------------------------------------------------------------------------


CmeshDocument::CmeshDocument(
	LCommander	*inSuper,
	FSSpec		*inFileSpec )
		: LSingleDoc( inSuper ), LListener()
{
	// Create window for our document.
	mWindow = LWindow::CreateWindow( rPPob_Window, this );
	Assert_( mWindow != nil );
	
	// Find the active panes
	mMeshView = (CmeshView *) mWindow->FindPaneByID( kMainView );
	Assert_( mMeshView != nil );
	
	if ( inFileSpec == nil ) {

		// Set the name of new window.
		NameNewDoc();
		
	} else {
	
		// Display contents of file in window.
		OpenFile( *inFileSpec );
	
	}
	
	// Make the window visible.
	mWindow->Show();
}


// ---------------------------------------------------------------------------------
//		 NameNewDoc
// ---------------------------------------------------------------------------------

void
CmeshDocument::NameNewDoc()
{
	// Setup the window title. Start with the default title.
	LStr255	theTitle( "\pUntitled" );

	// Find the first available title. We could also check the window
	// pane id if we wanted to make sure we didn't collide with other
	// window types.
	Int32	theNumber = 0;
	while ( UWindows::FindNamedWindow( theTitle ) != nil ) {

		// An existing window has the current name
		// Increment counter and try again.
		++theNumber;
		theTitle = "\pUntitled ";
		theTitle += (LStr255) theNumber;

	}		
	
	// Finally, set window title.
	mWindow->SetDescriptor( theTitle );
}


// ---------------------------------------------------------------------------------
//		 OpenFile
// ---------------------------------------------------------------------------------

void
CmeshDocument::OpenFile(
	FSSpec	&inFileSpec )
{
	Try_ {

		// Create a new file object.
		mFileStream = new LFileStream( inFileSpec );
		mFile = mFileStream;	// In case the superclass is using it
		
		// Open the data fork.
		mFileStream->OpenDataFork( fsRdPerm );
		
		// Set the window title to the name of the file and
		// flag that the document has an associated file.
		mWindow->SetDescriptor( inFileSpec.name );
		mIsSpecified = true;

		// Skip the file header and read the mesh header
		dir_Header theFileHeader;
		Int32 byteCount = sizeof(dir_Header);
		mFileStream->GetBytes(&theFileHeader, byteCount);
		Assert_(theFileHeader.group == 'mesh');
		
		byteCount = sizeof(mesh_Header);
		mFileStream->GetBytes(&mMeshHeader, byteCount);
		mHeaderOffset = sizeof(dir_Header) + mMeshHeader.headerSize;
	
		// Read the mesh
		TQ3GroupObject qd3DNewDisplayGroup = ::Q3OrderedDisplayGroup_New();
		if (qd3DNewDisplayGroup == nil)
			SignalPStr_("\pCouldn't allocate display group.");
	
		Int32 meshRows = mMeshHeader.height * 16;
		Int32 meshCols = mMeshHeader.width * 16;

		Ptr theMeshData = ::NewPtr(mMeshHeader.meshSize);
		ThrowIfNil_(theMeshData);
		byteCount = mMeshHeader.meshSize;
		mFileStream->SetMarker(mHeaderOffset + mMeshHeader.meshOffset, streamFrom_Start);
		mFileStream->GetBytes(theMeshData, byteCount);
		Int32* tempPtr = (Int32*) theMeshData;

		TQ3TriGridData theTriGridData;
		TQ3AttributeSet theVertexAttributes;
		theTriGridData.numRows = meshRows;
		theTriGridData.numColumns = meshCols;
		theTriGridData.facetAttributeSet = nil;
		theTriGridData.triGridAttributeSet = nil;

		TQ3Vertex3D* theVertices = new TQ3Vertex3D[meshRows * meshCols];
		double rowScale = 0.125;
		double colScale = 0.125;
		for (unsigned long rowIndex = 0; rowIndex < meshRows; rowIndex++) {
			for (unsigned long colIndex = 0; colIndex < meshCols; colIndex++) {
				theVertices[colIndex + rowIndex * meshCols].point.x = rowIndex * rowScale;
				theVertices[colIndex + rowIndex * meshCols].point.y = *tempPtr * 0.000000005;
				theVertices[colIndex + rowIndex * meshCols].point.z = colIndex * colScale;
				theVertices[colIndex + rowIndex * meshCols].attributeSet = nil;
				tempPtr++;
				tempPtr++;	// skip the flags
			}
		}

		theTriGridData.vertices = theVertices;
		TQ3GeometryObject theNewTriGrid = ::Q3TriGrid_New(&theTriGridData);

		delete [] theVertices;

		// Parameterize the surface
		TQ3Param2D theVertexUV;
		for (unsigned long rowIndex = 0; rowIndex < meshRows; rowIndex++) {
			for (unsigned long colIndex = 0; colIndex < meshCols; colIndex++) {
				theVertexAttributes = ::Q3AttributeSet_New();
//				theVertexUV.u = rowIndex; theVertexUV.u /= (meshRows - 1); Assert_(theVertexUV.u <= 1.0);
//				theVertexUV.v = colIndex; theVertexUV.v /= (meshCols - 1); Assert_(theVertexUV.v <= 1.0);
				theVertexUV.v = meshRows - rowIndex; theVertexUV.v /= meshRows; Assert_(theVertexUV.v <= 1.0);
				theVertexUV.u = colIndex; theVertexUV.u /= (meshCols - 1); Assert_(theVertexUV.u <= 1.0);
				::Q3AttributeSet_Add(theVertexAttributes, kQ3AttributeTypeSurfaceUV, &theVertexUV);
				::Q3TriGrid_SetVertexAttributeSet(theNewTriGrid, rowIndex, colIndex, theVertexAttributes);
				::Q3Object_Dispose(theVertexAttributes);
			}
		}
		
		// Set the mesh
		::Q3Group_AddObject(qd3DNewDisplayGroup, theNewTriGrid);
		::Q3Object_Dispose(theNewTriGrid);

		// Try to find the correct texture
		LFileStream* theTextureFile = FindTextureFile(inFileSpec);
		if (theTextureFile) {
			AddMapTexture(qd3DNewDisplayGroup, theTextureFile);
		}
		
		mMeshView->SetGroup(qd3DNewDisplayGroup);
		
	} Catch_( inErr ) {

		// Cleanup and rethrow the error.
		delete this;
		Throw_( inErr );
	
	} EndCatch_
}


// ---------------------------------------------------------------------------------
//		 FindTextureFile
// ---------------------------------------------------------------------------------
// Most of this code was ripped out of DropShell
// I Believe that code reuse is the sincerest form of slack.
// ALL HAIL SLACK!

LFileStream*
CmeshDocument::FindTextureFile(
	FSSpec	&inFileSpec )
{
	LFileStream* theFilePtr = nil;

	OSErr		err = noErr;
	short		index, oldIndex, localIndex;
	FSSpec		localFSSpec, curFSSpec;
	CInfoPBRec	cipb;
	Str255		fName, vFName;
	long		dirID, origDirID;
	Boolean		foundPosition;
 
	// We want the ".256" folder inside the grandparent of our mesh file
 	// parent folder FSSpec
 	ThrowIfOSErr_(::FSMakeFSSpec(inFileSpec.vRefNum, inFileSpec.parID, nil, &curFSSpec));
 	// .256 folder FSSpec
 	ThrowIfOSErr_(::FSMakeFSSpec(curFSSpec.vRefNum, curFSSpec.parID, "\p.256", &localFSSpec));
 	
	//	get the dirID for .256 folder, not it's parent!
	BlockMoveData(localFSSpec.name, fName, 32);

	cipb.hFileInfo.ioCompletion	= 0L;
	cipb.hFileInfo.ioNamePtr	= fName;
	cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
	cipb.hFileInfo.ioFDirIndex	= 0;	// use the dir & vRefNum;
	cipb.hFileInfo.ioDirID		= localFSSpec.parID;
	ThrowIfOSErr_(PBGetCatInfoSync(&cipb));
	
	origDirID = cipb.dirInfo.ioDrDirID; // copy the sucker
	index = 1;
			
	// index through all contents of this folder
	Boolean foundItem = false;
	while ((err == noErr) && !foundItem) {
		dirID = origDirID;
		localIndex = index;
		fName [0] = 0;
		cipb.hFileInfo.ioCompletion	= 0L;
		cipb.hFileInfo.ioNamePtr	= fName;
		cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
		cipb.hFileInfo.ioFDirIndex	= localIndex;	// use a real index
		cipb.hFileInfo.ioDirID		= dirID;
		err = PBGetCatInfoSync(&cipb);

		if (!err) {
			BlockMoveData(fName, curFSSpec.name, 32);
			curFSSpec.vRefNum	= cipb.hFileInfo.ioVRefNum;
			curFSSpec.parID		= dirID;
		
			// Check to see if this entry is a folder.
			if (cipb.hFileInfo.ioFlAttrib & ioDirMask) {
				SignalPStr_("\pFolder found inside .256 folder!");
			}
			// Is it the file we're looking for?
			else {
				theFilePtr = new LFileStream(curFSSpec);
				theFilePtr->OpenDataFork(fsRdPerm);
				dir_Header theFileHeader;
				Int32 dataSize = sizeof(dir_Header);
				theFilePtr->GetBytes(&theFileHeader, dataSize);
				theFilePtr->CloseDataFork();
				
				if (theFileHeader.subGroup == mMeshHeader.textureRef) {
					foundItem = true;
				}
				else {
					delete theFilePtr;
					theFilePtr = nil;
				}
			}
			
			/*	If we've had an error, get out! */
			if (err)	break;

			// dirID = origDirID;	
			localIndex = index;	

			/*	
				Now take into account new files being created
				in the current directory & messing up our index.
				See Dev.CD Vol. XI:Tools & Apps (Moof!):Misc Utilities:
				Disinfectant & Source 2.5.1:Sample:Notes:Scan Alg	
			*/
			vFName [0] = 0;
			cipb.hFileInfo.ioCompletion	= 0L;
			cipb.hFileInfo.ioNamePtr	= vFName;
			cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
			cipb.hFileInfo.ioFDirIndex	= localIndex;	// use a real index
			cipb.hFileInfo.ioDirID		= dirID;
			err = PBGetCatInfoSync(&cipb);
			oldIndex = index;
			if (!err) {
				/*	If they're equal - same place, go to next */
				if (EqualString (vFName, fName, false, false))
					index++;
			}
			
			/*	If we didn't advance, then perhaps a file was created or deleted */
			if (oldIndex == index) {
				oldIndex		= index;	/* save off the old */
				index			= 0;		/* and start at the beginning */
				err				= noErr;
				vFName [0]		= 0;
				foundPosition	= false;
				
				while (!foundPosition) {
					index++;
					vFName [0] = 0;
					cipb.hFileInfo.ioCompletion	= 0L;
					cipb.hFileInfo.ioNamePtr	= vFName;
					cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
					cipb.hFileInfo.ioFDirIndex	= index;	/* now use a real index */
					cipb.hFileInfo.ioDirID		= dirID;
					err = PBGetCatInfoSync(&cipb);
					
					if (err == fnfErr) {  // we've just been deleted
						index = oldIndex;
						foundPosition = true;
						err = noErr;	// have to remember to reset this!
					}
					
				/*	found same file & same index position */
				/*	so try the next item */
					if ((!foundPosition) && EqualString(fName, vFName, false, false)) {
						index++;
						foundPosition = true;
					}
				}
			}
		}
	}
	
	ThrowIfOSErr_(err);
	return theFilePtr;
}


// ---------------------------------------------------------------------------
//	 AddMapTexture									 [private]
// ---------------------------------------------------------------------------

void
CmeshDocument::AddMapTexture(TQ3GroupObject& ioGroup, LFileStream* inFile)
{
	inFile->OpenDataFork(fsRdPerm);
	
	// Skip the file header and read the collection header
	dir_Header theFileHeader;
	Int32 byteCount = sizeof(dir_Header);
	inFile->GetBytes(&theFileHeader, byteCount);
	Assert_(theFileHeader.group == '.256');
	
	collection_header theCollectionHeader;
	byteCount = sizeof(collection_header);
	inFile->GetBytes(&theCollectionHeader, byteCount);
	Int32 textureHeaderOffset = inFile->GetMarker();
	Assert_(theCollectionHeader.user_data_flags & _collection_is_color_map_flag);

	// Read a color table
	Assert_(theCollectionHeader.color_table_count >= 1);
	if(theCollectionHeader.color_table_count > 1)
		SignalPStr_("\pMore than one color table in this file.");
	Ptr theColorData = ::NewPtr(theCollectionHeader.color_tables_size);
	ThrowIfNil_(theColorData);
	inFile->SetMarker(textureHeaderOffset + theCollectionHeader.color_tables_offset, streamFrom_Start);
	byteCount = theCollectionHeader.color_tables_size;
	inFile->GetBytes(theColorData, byteCount);
	
	CTabHandle theColorTable = ::GetCTable(40);				// default 8-bit grey
	::HLock((Handle) theColorTable);
	Ptr colorSourcePtr = theColorData + 32;		// skip header
	for (Int16 colorCount = 0; colorCount <= (*theColorTable)->ctSize; colorCount++) {
		(*theColorTable)->ctTable[colorCount].rgb = *((RGBColor*) colorSourcePtr);
		colorSourcePtr += 8;
	}
	::HUnlock((Handle) theColorTable);
	
	::CTabChanged(theColorTable);
	::DisposePtr(theColorData);
	
	Rect bitBounds = { 0, 0, 0, 0 };
	bitBounds.bottom = mMeshHeader.height * 256;	// 256 pixel sections
	bitBounds.right = mMeshHeader.width * 256;
	
	LGWorld* theNewGWorld = nil;
	
	theNewGWorld = new LGWorld(bitBounds, 8, 0, theColorTable);
	Assert_(theNewGWorld != nil);
	
	// Write data into the new GWorld
	GWorldPtr macGWorld = theNewGWorld->GetMacGWorld();	
	if (theNewGWorld->BeginDrawing()) {

		Int32 bitmapRefMarker = textureHeaderOffset + theCollectionHeader.bitmap_references_offset;
//		bitmapRefMarker += sizeof(bitmap_reference);	// read the tags instead of texture

		PixMapHandle thePixH = ::GetGWorldPixMap(macGWorld);
		UInt32 realPixRowBytes = (*thePixH)->rowBytes & 0x3FFF;

		for (Uint8 hCount = 0; hCount < mMeshHeader.height; hCount++) {
		for (Uint8 wCount = 0; wCount < mMeshHeader.width; wCount++) {
			// Read the bitmap "reference"
			bitmap_reference theBitmapRef;
			inFile->SetMarker(bitmapRefMarker, streamFrom_Start);
			byteCount = sizeof(bitmap_reference);
			inFile->GetBytes(&theBitmapRef, byteCount);
			
			// Read the bitmap data proper
			inFile->SetMarker(textureHeaderOffset + theBitmapRef.offset, streamFrom_Start);

			Ptr theBitmapData = ::NewPtr(theBitmapRef.size);
			ThrowIfNil_(theBitmapData);
			byteCount = theBitmapRef.size;
			inFile->GetBytes(theBitmapData, byteCount);
				
			bitmapData_Header *bitmapHeaderPtr = (bitmapData_Header*) theBitmapData;
			Assert_(bitmapHeaderPtr->height == 256);
			Assert_(bitmapHeaderPtr->width == 256);
			Ptr bitmapDataPtr = theBitmapData + sizeof(bitmapData_Header);
			bitmapDataPtr += 4;	// We know its map data (test above)
			Assert_(bitmapHeaderPtr->rowBytes != -1);	// No encoded textures

			Ptr realPixPtr = ::GetPixBaseAddr(thePixH);
			realPixPtr += hCount * 256 * realPixRowBytes;

			for (UInt16 rowCount = 1; rowCount <= bitmapHeaderPtr->height; rowCount++) {
				Ptr ScratchPtr = realPixPtr + (wCount * 256);
				::BlockMoveData(bitmapDataPtr, ScratchPtr, bitmapHeaderPtr->rowBytes);
				realPixPtr += realPixRowBytes;
				bitmapDataPtr += bitmapHeaderPtr->rowBytes;
			}

			::DisposePtr(theBitmapData);
			bitmapRefMarker += 2 * sizeof(bitmap_reference);
		}
		}
	}
	theNewGWorld->EndDrawing();
	
	// Force the GWorld to 16 bit for QD3D
	::UpdateGWorld(&macGWorld, 16, &bitBounds, nil, nil, clipPix);

	PixMapHandle theGWorldPixMap = ::GetGWorldPixMap(macGWorld);
	::LockPixels(theGWorldPixMap);
	unsigned long pixMapWidth = (*theGWorldPixMap)->bounds.right - (*theGWorldPixMap)->bounds.left;
	unsigned long pixMapHeight = (*theGWorldPixMap)->bounds.bottom - (*theGWorldPixMap)->bounds.top;
	unsigned long realRowBytes = (*theGWorldPixMap)->rowBytes & 0x3FFF;
	unsigned long dataSize = realRowBytes * pixMapHeight;
	
	TQ3StorageObject theNewStorage = ::Q3MemoryStorage_NewBuffer((unsigned char *) ::GetPixBaseAddr(theGWorldPixMap), 0, dataSize);
	TQ3StoragePixmap theNewPixmap;

	theNewPixmap.image = theNewStorage;

	theNewPixmap.width = pixMapWidth;
	theNewPixmap.height = pixMapHeight;
	theNewPixmap.rowBytes = (*theGWorldPixMap)->rowBytes  & 0x3FFF;

	Assert_((*theGWorldPixMap)->pixelSize == 16);
	theNewPixmap.pixelType = kQ3PixelTypeRGB16;

	theNewPixmap.bitOrder = kQ3EndianBig;
	theNewPixmap.byteOrder = kQ3EndianBig;
	
	TQ3TextureObject theNewTexture = ::Q3PixmapTexture_New(&theNewPixmap);
	::Q3Object_Dispose(theNewStorage);
	
	TQ3ShaderObject theNewShader = Q3TextureShader_New(theNewTexture);
	::Q3Object_Dispose(theNewTexture);

	::Q3Group_AddObject(ioGroup, theNewShader);
	::Q3Object_Dispose(theNewShader);
}


// ---------------------------------------------------------------------------------
//		 IsModified
// ---------------------------------------------------------------------------------

Boolean
CmeshDocument::IsModified()
{
	return false;
}


// ---------------------------------------------------------------------------------
//		 DoAESave
// ---------------------------------------------------------------------------------

void
CmeshDocument::DoAESave(
	FSSpec	&/*inFileSpec*/,
	OSType	/*inFileType*/ )
{
/*
	// Delete the existing file object.
	// Note: this does nothing to the actual file on disk.
	delete mFileStream;
	
	// Make a new file object.
	mFileStream = new LFile( inFileSpec );
	
	// Get the proper file type.
	OSType	theFileType = 'TEXT';
	if ( inFileType != fileType_Default ) theFileType = inFileType;

	// Make new file on disk (we'll use
	// SimpleText's creator for this example).
	mFileStream->CreateNewDataFile( 'ttxt', theFileType );
	
	// Write out the data.
	DoSave();

	// Change window title to reflect the new name.
	mWindow->SetDescriptor( inFileSpec.name );

	// Document now has a specified file.
	mIsSpecified = true;
*/
}


// ---------------------------------------------------------------------------------
//		 DoSave
// ---------------------------------------------------------------------------------

void
CmeshDocument::DoSave()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Get the text from the text view.
	Handle	theTextH = mTextView->GetTextHandle();
	
	// Lock the text handle.
	StHandleLocker	theLock( theTextH );
	
	// Write the text to the file.
	mFileStream->WriteDataFork( *theTextH, ::GetHandleSize( theTextH ) );

	// Close the data fork.
	mFileStream->CloseDataFork();

	// Saving makes doc un-dirty.
	mTextView->SetDirty( false );
*/
}


// ---------------------------------------------------------------------------------
//		 DoRevert
// ---------------------------------------------------------------------------------

void
CmeshDocument::DoRevert()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Read the entire file contents and close the file.
	Handle	theTextH = mFileStream->ReadDataFork();
	mFileStream->CloseDataFork();
	
	// Put the contents in the text view
	// and clear the dirty flag.
	mTextView->SetTextHandle( theTextH );
	mTextView->SetDirty( false );
	
	// Dispose of the text.
	::DisposeHandle( theTextH );

	// Refresh the text view.
	mTextView->Refresh();
*/
}


// ---------------------------------------------------------------------------------
//		 DoPrint
// ---------------------------------------------------------------------------------

void
CmeshDocument::DoPrint()
{
/*
	// Create the printout.
	LPrintout	*thePrintout = LPrintout::CreatePrintout( rPPob_TextPrintout );
	Assert_( thePrintout != nil );
	
	// Set the print record.
	
	if (mPrintRecordH)
	{
		thePrintout->SetPrintRecord( mPrintRecordH );
	}
	
	// Get the text placeholder.
	LPlaceHolder	*thePlaceholder;
	thePlaceholder = (LPlaceHolder*) thePrintout->FindPaneByID( kTextPlaceholder );
	Assert_( thePlaceholder != nil );
	
	// Install the text view in the placeholder.
	thePlaceholder->InstallOccupant( mTextView, atNone );
	
	// Set the frame size.
	SetPrintFrameSize();
	
	// Print.
	thePrintout->DoPrintJob();
	
	// Delete the printout. The text view is returned
	// to the window when the placeholder is destroyed.
	delete thePrintout;
*/
}


// ---------------------------------------------------------------------------------
//		 SetPrintFrameSize
// ---------------------------------------------------------------------------------

void
CmeshDocument::SetPrintFrameSize()
{
/*
	// Get the frame size.
	SDimension16	theFrameSize;
	mTextView->GetFrameSize( theFrameSize );
	
	// Get the text edit record handle.
	TEHandle	theTextEditH = mTextView->GetMacTEH();
	
	// Calculate the number of lines per page.
	Int16	theLinesPerPage;
	theLinesPerPage = theFrameSize.height / (**theTextEditH).lineHeight;

	// Resize the frame to an integral number of lines.
	mTextView->ResizeFrameTo( theFrameSize.width,
		(**theTextEditH).lineHeight * theLinesPerPage, false );

*/
}


void
CmeshDocument::ListenToMessage(
	MessageT /*inMessage*/,
	void * /* ioParam */)
{
}
