// =================================================================================
//	CobjectDocument.cp				1997 eli curtz. All rights reserved.
// =================================================================================

#include <LFile.h>
#include <LPlaceHolder.h>
#include <LPrintout.h>
#include <LString.h>
#include <LWindow.h>
#include <PP_Messages.h>
#include <UMemoryMgr.h>
#include <UWindows.h>
#include <UModalDialogs.h>
#include <LTableArrayStorage.h>
#include <LString.h>

#include <fp.h>

#include "Constants.h"
#include "CobjectView.h"
#include "CMiniIconTable.h"
#include "gorStructs.h"

#include "CobjectDocument.h"


Int32
LTagComparator::Compare(
	const void*		inItemOne,
	const void*		inItemTwo,
	Uint32			/* inSizeOne */,
	Uint32			/* inSizeTwo */) const
{
	dir_Header * dirHeader1 = (dir_Header *) inItemOne;
	dir_Header * dirHeader2 = (dir_Header *) inItemTwo;

	if (dirHeader1->group != dirHeader2->group)
		return ( dirHeader1->group - dirHeader2->group );
	else
		return ( dirHeader1->subGroup - dirHeader2->subGroup );
}

// ---------------------------------------------------------------------------------
//		 CobjectDocument(LCommander*, FSSpec*)
// ---------------------------------------------------------------------------------


CobjectDocument::CobjectDocument(
	LCommander	*inSuper,
	FSSpec		*inFileSpec )
		: LSingleDoc( inSuper ), LListener()
{
	for (Int16 typeCount = 0; typeCount < 16; typeCount++) {
		mObjectTypes[typeCount] = new LArray(sizeof(mesh_ObjectType));
	}
	
	mTags = new LArray(sizeof(dir_Header), new LTagComparator(), true);
	
	mObjectIndex = -1;
	
	// Create window for our document.
	mWindow = LWindow::CreateWindow( rPPob_Window, this );
	Assert_( mWindow != nil );
	
	mMapGWorld = nil;
	
	// Find the active panes
	mPreviousButton = (LStdControl *) mWindow->FindPaneByID( kPrevButton );
	Assert_(mPreviousButton); mPreviousButton->AddListener(this);
	mNextButton = (LStdControl *) mWindow->FindPaneByID( kNextButton );
	Assert_(mNextButton); mNextButton->AddListener(this);

	mTypeCaption = (LCaption *) mWindow->FindPaneByID( kObjectType ); Assert_(mTypeCaption);
	mTeamCaption = (LCaption *) mWindow->FindPaneByID( kObjectTeam ); Assert_(mTeamCaption);
	mXPosCaption = (LCaption *) mWindow->FindPaneByID( kObjectXPos ); Assert_(mXPosCaption);
	mYPosCaption = (LCaption *) mWindow->FindPaneByID( kObjectYPos ); Assert_(mYPosCaption);
	mZPosCaption = (LCaption *) mWindow->FindPaneByID( kObjectZPos ); Assert_(mZPosCaption);
	mFacingCaption = (LCaption *) mWindow->FindPaneByID( kObjectFacing ); Assert_(mFacingCaption);
	mObjectIDCaption = (LCaption *) mWindow->FindPaneByID( kObjectID ); Assert_(mObjectIDCaption);

	mHiddenCheck= (LStdControl *) mWindow->FindPaneByID( kObjectHidden ); Assert_(mHiddenCheck);
		
	if ( inFileSpec == nil ) {

		// Set the name of new window.
		NameNewDoc();
		
	} else {
	
		// Display contents of file in window.
		OpenFile( *inFileSpec );
	
	}
	
	// Make the window visible.
	mWindow->Show();
}

CobjectDocument::~CobjectDocument()
{
	for (Int16 typeCount = 0; typeCount < 16; typeCount++) {
		delete mObjectTypes[typeCount];
	}

	delete mTags;
	
	delete mMapGWorld;
}


// ---------------------------------------------------------------------------------
//		 NameNewDoc
// ---------------------------------------------------------------------------------

void
CobjectDocument::NameNewDoc()
{
	// Setup the window title. Start with the default title.
	LStr255	theTitle( "\pUntitled" );

	// Find the first available title. We could also check the window
	// pane id if we wanted to make sure we didn't collide with other
	// window types.
	Int32	theNumber = 0;
	while ( UWindows::FindNamedWindow( theTitle ) != nil ) {

		// An existing window has the current name
		// Increment counter and try again.
		++theNumber;
		theTitle = "\pUntitled ";
		theTitle += (LStr255) theNumber;

	}		
	
	// Finally, set window title.
	mWindow->SetDescriptor( theTitle );
}


// ---------------------------------------------------------------------------------
//		 OpenFile
// ---------------------------------------------------------------------------------

void
CobjectDocument::OpenFile(
	FSSpec	&inFileSpec )
{
	Try_ {

		// Create a new file object.
		mFileStream = new LFileStream( inFileSpec );
		mFile = mFileStream;	// In case the superclass is using it
		
		// Open the data fork.
		mFileStream->OpenDataFork( fsRdPerm );
		
		// Set the window title to the name of the file and
		// flag that the document has an associated file.
		mWindow->SetDescriptor( inFileSpec.name );
		mIsSpecified = true;

		// Skip the file header and read the mesh header
		dir_Header theFileHeader;
		Int32 byteCount = sizeof(dir_Header);
		mFileStream->GetBytes(&theFileHeader, byteCount);
		Assert_(theFileHeader.group == 'mesh');
		
		byteCount = sizeof(mesh_Header);
		mFileStream->GetBytes(&mMeshHeader, byteCount);
		mHeaderOffset = sizeof(dir_Header) + mMeshHeader.headerSize;
		
		mFileStream->SetMarker(mHeaderOffset + mMeshHeader.objectTypeOffset, streamFrom_Start);
		mesh_ObjectType theNewType;
		for (Int16 typeCount = 0; typeCount < mMeshHeader.objectTypeCount; typeCount++) {
			byteCount = sizeof(mesh_ObjectType);
			mFileStream->GetBytes(&theNewType, byteCount);
			mObjectTypes[theNewType.typeFlag]->InsertItemsAt(1, LArray::index_Last, &theNewType);
		}
		
/*
		Handle scratchH = NewHandle(mMeshHeader.mediaPolygonsSize);
		Assert_(scratchH);
		mFileStream->SetMarker(mHeaderOffset + mMeshHeader.mediaPolygonsOffset, streamFrom_Start);
		byteCount = GetHandleSize(scratchH);
		HLock(scratchH);
			mFileStream->GetBytes(*scratchH, byteCount);
		HUnlock(scratchH);
		Debugger();
		DisposeHandle(scratchH);
*/
		
		// Display troops on the map (temp)
		Rect mapBounds;
		mapBounds.top = 0; mapBounds.left = 0;
		mapBounds.bottom = mMeshHeader.height * 64; mapBounds.right = mMeshHeader.width * 64;
		mMapGWorld = new LGWorld(mapBounds, 8);
		
		Point theUnitLoc = { 0, 0 };
		Point scratch;
		Rect drawRect;
		mMapGWorld->BeginDrawing();
		mFileStream->SetMarker(mHeaderOffset + mMeshHeader.objectInstanceOffset, streamFrom_Start);
		mesh_ObjectInstance theObject;
		for (Int16 objectCount = 0; objectCount < mMeshHeader.objectInstanceCount; objectCount++) {
			byteCount = sizeof(mesh_ObjectInstance);
			mFileStream->GetBytes(&theObject, byteCount);
			if ((theObject.typeFlag == 3) && !(theObject.visibilityFlags)) {
				theUnitLoc.h = ((16384 * mMeshHeader.width) - theObject.locationX) / 256;
				theUnitLoc.v = theObject.locationY / 256;
				scratch.h = theUnitLoc.h + 2; scratch.v = theUnitLoc.v + 2;
				Pt2Rect(theUnitLoc, scratch, &drawRect);
				PaintRect(&drawRect);
			}
		}
		mMapGWorld->EndDrawing();
		LWindow* theMapWindow = LWindow::CreateWindow( rPPob_MapWindow, this );
		CobjectView* theMapView = (CobjectView*) theMapWindow->FindPaneByID(kMapView);
		Assert_(theMapView);
		theMapView->SetGWorld(mMapGWorld);
		
		ListenToMessage(msg_Next, nil);

	} Catch_( inErr ) {

		// Cleanup and rethrow the error.
		delete this;
		Throw_( inErr );
	
	} EndCatch_
}


// ---------------------------------------------------------------------------------
//		 IsModified
// ---------------------------------------------------------------------------------

Boolean
CobjectDocument::IsModified()
{
	return false;
}


// ---------------------------------------------------------------------------------
//		 DoAESave
// ---------------------------------------------------------------------------------

void
CobjectDocument::DoAESave(
	FSSpec	& /*  inFileSpec */,
	OSType	/* inFileType */)
{
/*
	// Delete the existing file object.
	// Note: this does nothing to the actual file on disk.
	delete mFileStream;
	
	// Make a new file object.
	mFileStream = new LFile( inFileSpec );
	
	// Get the proper file type.
	OSType	theFileType = 'TEXT';
	if ( inFileType != fileType_Default ) theFileType = inFileType;

	// Make new file on disk (we'll use
	// SimpleText's creator for this example).
	mFileStream->CreateNewDataFile( 'ttxt', theFileType );
	
	// Write out the data.
	DoSave();

	// Change window title to reflect the new name.
	mWindow->SetDescriptor( inFileSpec.name );

	// Document now has a specified file.
	mIsSpecified = true;
*/
}


// ---------------------------------------------------------------------------------
//		 DoSave
// ---------------------------------------------------------------------------------

void
CobjectDocument::DoSave()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Get the text from the text view.
	Handle	theTextH = mTextView->GetTextHandle();
	
	// Lock the text handle.
	StHandleLocker	theLock( theTextH );
	
	// Write the text to the file.
	mFileStream->WriteDataFork( *theTextH, ::GetHandleSize( theTextH ) );

	// Close the data fork.
	mFileStream->CloseDataFork();

	// Saving makes doc un-dirty.
	mTextView->SetDirty( false );
*/
}


// ---------------------------------------------------------------------------------
//		 DoRevert
// ---------------------------------------------------------------------------------

void
CobjectDocument::DoRevert()
{
/*
	// Open the data fork.
	mFileStream->OpenDataFork( fsRdWrPerm );

	// Read the entire file contents and close the file.
	Handle	theTextH = mFileStream->ReadDataFork();
	mFileStream->CloseDataFork();
	
	// Put the contents in the text view
	// and clear the dirty flag.
	mTextView->SetTextHandle( theTextH );
	mTextView->SetDirty( false );
	
	// Dispose of the text.
	::DisposeHandle( theTextH );

	// Refresh the text view.
	mTextView->Refresh();
*/
}


// ---------------------------------------------------------------------------------
//		 DoPrint
// ---------------------------------------------------------------------------------

void
CobjectDocument::DoPrint()
{
/*
	// Create the printout.
	LPrintout	*thePrintout = LPrintout::CreatePrintout( rPPob_TextPrintout );
	Assert_( thePrintout != nil );
	
	// Set the print record.
	
	if (mPrintRecordH)
	{
		thePrintout->SetPrintRecord( mPrintRecordH );
	}
	
	// Get the text placeholder.
	LPlaceHolder	*thePlaceholder;
	thePlaceholder = (LPlaceHolder*) thePrintout->FindPaneByID( kTextPlaceholder );
	Assert_( thePlaceholder != nil );
	
	// Install the text view in the placeholder.
	thePlaceholder->InstallOccupant( mTextView, atNone );
	
	// Set the frame size.
	SetPrintFrameSize();
	
	// Print.
	thePrintout->DoPrintJob();
	
	// Delete the printout. The text view is returned
	// to the window when the placeholder is destroyed.
	delete thePrintout;
*/
}


// ---------------------------------------------------------------------------------
//		 SetPrintFrameSize
// ---------------------------------------------------------------------------------

void
CobjectDocument::SetPrintFrameSize()
{
/*
	// Get the frame size.
	SDimension16	theFrameSize;
	mTextView->GetFrameSize( theFrameSize );
	
	// Get the text edit record handle.
	TEHandle	theTextEditH = mTextView->GetMacTEH();
	
	// Calculate the number of lines per page.
	Int16	theLinesPerPage;
	theLinesPerPage = theFrameSize.height / (**theTextEditH).lineHeight;

	// Resize the frame to an integral number of lines.
	mTextView->ResizeFrameTo( theFrameSize.width,
		(**theTextEditH).lineHeight * theLinesPerPage, false );

*/
}


// ---------------------------------------------------------------------------
//		 ShowPreferencesDialog
// ---------------------------------------------------------------------------

void
CobjectDocument::ShowPreferencesDialog()
{
	StDialogHandler theHandler( rPPob_PrefDialog, this );
	MessageT theDialogResult;
	
	LWindow	*theDialog = theHandler.GetDialog();
	CMiniIconTable *theTable = (CMiniIconTable*) theDialog->FindPaneByID(kGORTable); Assert_(theTable);
	theTable->InsertCols(1, 0);
	
	LStdButton *theAddButton = (LStdButton*) theDialog->FindPaneByID(kAddButton); Assert_(theAddButton);
	theAddButton->AddListener(&theHandler);
	LStdButton *theRemoveButton = (LStdButton*) theDialog->FindPaneByID(kRemoveButton); Assert_(theRemoveButton);
	theRemoveButton->AddListener(&theHandler);
	
	Boolean dialogDone = false;
	do {
		while ( (theDialogResult = theHandler.DoDialog()) == msg_Nothing ) {
			if ( theTable->GetValue() == -1 )
				theRemoveButton->Disable();
			else
				theRemoveButton->Enable();
		}; // wait for result
		
		if ( theDialogResult == msg_OK ) {
			dialogDone = true;
		}
		else if ( theDialogResult == msg_Cancel ) {
			dialogDone = true;
		}
		else if ( theDialogResult == msg_Add ) {
			// Deactivate the desktop.
			::UDesktop::Deactivate();

			// Browse for a document.
			SFTypeList			theTypeList = {'.gor'};
			StandardFileReply	theReply;
			::StandardGetFile( nil, 1, theTypeList, &theReply );

			// Activate the desktop.
			::UDesktop::Activate();
			
			// Add the new file
			if ( theReply.sfGood ) {
				MIconTableRec theNewItem;
				theNewItem.iconID = rICM_GORFileIcon;
				TString<Str31> fileName = theReply.sfFile.name;
				::BlockMoveData(&fileName[0], &theNewItem.name[0], sizeof(Str31));
				
				TableIndexT	rowCount, colCount;
				theTable->GetTableSize(rowCount, colCount);
				theTable->InsertRows(1, rowCount, &theNewItem, 0, true);
				
				// Read the tags
				LFileStream theGORFile(theReply.sfFile);
				theGORFile.OpenDataFork(fsRdPerm);
				gor_Header theFileHeader;
				Int32 byteCount = sizeof(gor_Header);
				
				theGORFile.GetBytes(&theFileHeader, byteCount);
				theGORFile.SetMarker(theFileHeader.tagHeaderOffset, streamFrom_Start);
				dir_Header* tagBuffer = new dir_Header[theFileHeader.headerCount]; Assert_(tagBuffer);
				byteCount = theFileHeader.headerCount * sizeof(dir_Header);
				theGORFile.GetBytes(tagBuffer, byteCount);
				
				// whomever wrote LArray needs a good ass whoopin'
				for (Int16 count = 0; count < theFileHeader.headerCount; count++) {
					mTags->InsertItemsAt(1, LArray::index_Last, &tagBuffer[count]);
				}
				mTags->Sort();

				delete tagBuffer;
				theGORFile.CloseDataFork;
			}
		}
		else if ( theDialogResult == msg_Remove ) {
			if (theTable->GetValue() >= 0) {
			}
		}
	} while(!dialogDone);
}


// ---------------------------------------------------------------------------
//		 ObeyCommand
// ---------------------------------------------------------------------------
//	Respond to commands

Boolean
CobjectDocument::ObeyCommand(
	CommandT	inCommand,
	void		*ioParam)
{
	Boolean		cmdHandled = true;

	switch (inCommand) {
	
		case cmd_Preferences:
			ShowPreferencesDialog();
			break;

		default:
			cmdHandled = LSingleDoc::ObeyCommand(inCommand, ioParam);
			break;
	}
	
	return cmdHandled;
}

// ---------------------------------------------------------------------------
//		 FindCommandStatus
// ---------------------------------------------------------------------------
//	Pass back whether a Command is enabled and/or marked (in a Menu)

void
CobjectDocument::FindCommandStatus(
	CommandT	inCommand,
	Boolean		&outEnabled,
	Boolean		&outUsesMark,
	Char16		&outMark,
	Str255		outName)
{
	switch (inCommand) {
	
		case cmd_Preferences:
			outEnabled = true;
			break;
			
		default:
			LSingleDoc::FindCommandStatus(inCommand, outEnabled,
									outUsesMark, outMark, outName);
			break;
	}
}


// ---------------------------------------------------------------------------
//		 ListenToMessage
// ---------------------------------------------------------------------------

void
CobjectDocument::ListenToMessage(
	MessageT inMessage,
	void * /* ioParam */)
{
	if (inMessage == msg_Next)
		mObjectIndex++;
	else if (inMessage == msg_Previous)
		mObjectIndex--;
	
	if (mObjectIndex <= 0)
		mPreviousButton->Disable();
	else
		mPreviousButton->Enable();
	
	if (mObjectIndex < mMeshHeader.objectInstanceCount)
		mNextButton->Enable();
	else
		mNextButton->Disable();
	
	// read the current object from the file
	mesh_ObjectInstance theObject;
	mFileStream->SetMarker(mHeaderOffset + mMeshHeader.objectInstanceOffset + (mObjectIndex * sizeof(mesh_ObjectInstance)), streamFrom_Start);
	Int32 byteCount = sizeof(mesh_ObjectInstance);
	mFileStream->GetBytes(&theObject, byteCount);	
	
	// get the associated object type, remembering LArray INDICES are 1 based (sheesh)
	mesh_ObjectType theObjectType;
	::BlockMoveData(mObjectTypes[theObject.typeFlag]->GetItemPtr(theObject.typeIndex + 1), &theObjectType, sizeof(mesh_ObjectType));
	
	// fill in the display
	FourCharCode objectTypeGroup = '????';
	switch (theObjectType.typeFlag) {
		case 0:
			break;
		case 1:
			objectTypeGroup = 'scen';
			break;
		case 3:
			objectTypeGroup = 'unit';
			break;
		case 5:
			objectTypeGroup = 'amso';
			break;
		case 6:
			objectTypeGroup = 'mode';
			break;
		case 9:
			objectTypeGroup = 'prgr';
			break;
		case 10:
			objectTypeGroup = 'lpgr';
			break;
		default:
			Assert_(false);
	}
	
	// try to get the name from our tags
	dir_Header theTestHeader;
	theTestHeader.group = objectTypeGroup;
	theTestHeader.subGroup = theObjectType.subGroupTag;
	ArrayIndexT foundIndex = mTags->FetchIndexOf(&theTestHeader);
	
	if (mTags->ValidIndex(foundIndex)) {
		::BlockMoveData(mTags->GetItemPtr(foundIndex), &theTestHeader, sizeof(dir_Header));
		mTypeCaption->SetDescriptor(LStr255(theTestHeader.name));
	}
	else
		mTypeCaption->SetDescriptor(LStr255(theObjectType.subGroupTag));
		
	long tempLong = theObjectType.teamGroupNumber;
	mTeamCaption->SetDescriptor(LStr255(tempLong));

	LStr255 tempString;
	float tempValue = theObject.locationX; tempValue /= 1024;
	tempString.Assign(tempValue, FIXEDDECIMAL, 3);
	mXPosCaption->SetDescriptor(tempString);
	tempValue = theObject.locationY; tempValue /= 1024;
	tempString.Assign(tempValue, FIXEDDECIMAL, 3);
	mYPosCaption->SetDescriptor(tempString);
	tempValue = theObject.locationZ; tempValue /= 1024;
	tempString.Assign(tempValue, FIXEDDECIMAL, 3);
	mZPosCaption->SetDescriptor(tempString);
	
	tempValue = theObject.facingAngle; tempValue /= 0xFFFF; tempValue *= 180;
	tempString.Assign(tempValue, FIXEDDECIMAL, 3);
	mFacingCaption->SetDescriptor(tempString);

	tempLong = theObject.objectID;
	mObjectIDCaption->SetDescriptor(LStr255(tempLong));

	mHiddenCheck->SetValue(theObject.visibilityFlags);
	
	
	/* no inherited to call */
}
