//
// The Gray Council
// Copyright 1996-1997 by Trygve Isaacson. All Rights Reserved.
//
// Before using any of the Gray Council source code, read and
// follow the licensing info in the accompanying documentation
// or contact:
//   <trygve@bombaydigital.com>
//   <http://www.bombaydigital.com>
//
// Also check the web site above to make sure you have the latest version!
//
// See the header file "GCNewControls.h" for usage.
//

#include "GCNewControls.h"

//
// MIconButtonObject ---------------------------------------------------------------
//
// This mixin class handles button drawing for the different types
// of icon button classes. This class draws the button, and subclasses
// for push/radio/checkbox that mix it in handle the behavioral aspects.
//
// Ideal button sizes, to match AGA specification:
// - Large icon button: 40x40 (32x32 iclx plus 4 pixel border on each side)
// - Small icon button: 22x22 (16x16 icsx plus 3 pixel border on each side)
// - Mini icon button:  20x16 (16x12 icmx plus 2 pixel border on each side)
// Note that mini icons are a rather rare resource that ResEdit 2.x doesn't grok.
//

MIconButtonObject::MIconButtonObject()
	{
	// Construct with no icon IDs.

	mOffIconID = kNoIconID;
	mOnIconID = kNoIconID;
	mFrameType = kAutoFrame;
	mImageType = kAutoIconFamily;
	}

MIconButtonObject::MIconButtonObject(SInt16 offIconID, SInt16 onIconID, ButtonFrameType frameType, ButtonImageType imageType)
	{
	// Construct with specified icon IDs.

	mOffIconID = offIconID;
	mOnIconID = onIconID;
	mFrameType = frameType;
	mImageType = imageType;
	}

MIconButtonObject::~MIconButtonObject()
	{
	}

void MIconButtonObject::SetIconIDs(SInt16 offIconID, SInt16 onIconID)
	{
	// Set icon IDs as specified.

	mOffIconID = offIconID;
	mOnIconID = onIconID;
	}

void MIconButtonObject::SetFrameType(ButtonFrameType frameType)
	{
	// Set button frame type as specified.

	mFrameType = frameType;
	}

void MIconButtonObject::SetImageType(ButtonImageType imageType)
	{
	// Set buttom image type as specified.

	mImageType = imageType;
	}

void MIconButtonObject::DrawIconButton(Rect* bounds,
										Boolean isOn,
										Boolean isPressed,
										Boolean isEnabled,
										Boolean deep)
	{
	// Draw the icon button in the appropriate state.

	Rect	r = *bounds;
	
	//
	// Draw the blank button with the appropriate frame type.
	//
	
	AGADrawBevelButtonAtDepth(deep, bounds, mFrameType, isOn, isPressed, isEnabled);

	//
	// Determine the target area into which the image must
	// be drawn.
	//

	ButtonImageType	imageType = mImageType;
	Point			imageSize;
	Point			imageCenter;
	SInt16			imageResourceID = isOn ? mOnIconID : mOffIconID;
	PicHandle		pictureHandle;
	CIconHandle		colorIconHandle;
	
	imageCenter.h = r.left + ((r.right - r.left) / 2);
	imageCenter.v = r.top + ((r.bottom - r.top) / 2);
	
	if (imageType == kAutoIconFamily)
		imageType = MIconButtonObject::GetDefaultImageType(r.right - r.left);

	switch (imageType)
		{
		case kPictureIcon:
			pictureHandle = this->GetPictureImage(&imageSize, imageResourceID);
			break;

		case kColorIcon:
			colorIconHandle = this->GetColorIconImage(&imageSize, imageResourceID);
			break;

		case kMiniIcon:
			::SetPt(&imageSize, 16, 12);	// 16x12 icm#
			break;

		case kSmallIcon:
			::SetPt(&imageSize, 16, 16);	// 16x16 ics#
			break;

		case kLargeIcon:
		default:
			::SetPt(&imageSize, 32, 32);	// 32x32 icl#
			break;
		}
	
	r.left = imageCenter.h - (imageSize.h / 2);
	r.right = r.left + imageSize.h;
	r.top = imageCenter.v - (imageSize.v / 2);
	r.bottom = r.top + imageSize.v;

	//
	// Draw the image onto the blank button.
	//

	IconTransformType	aTransform = isPressed ? ttSelected : (isEnabled ? ttNone : ttDisabled);

	if (imageResourceID != kNoIconID)
		{
		::RGBForeColor(&gAGARamp[rB]);

		switch (imageType)
			{
			case kPictureIcon:
				if (pictureHandle != NULL)
					{
					::DrawPicture(pictureHandle, &r);
					// The PicHandle is not explicitly released. Just let it get purged as needed.
					}
				break;

			case kColorIcon:
				if (colorIconHandle != NULL)
					{
					(void) ::PlotCIconHandle(&r, atNone, aTransform, colorIconHandle);
					::DisposeCIcon(colorIconHandle);
					}
				break;

			case kMiniIcon:
			case kSmallIcon:
			case kLargeIcon:
			default:
				(void) ::PlotIconID(&r, atNone, aTransform, imageResourceID);
				break;
			}
		}
	}

void MIconButtonObject::DrawMiniEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	// Draw the button edges as specified for a
	// "mini" sized button.

	enum { TLFrame, TLInner, BRFrame, BRInner, CornerFrame, CornerInner, Fill, kNumEdgeColors };

	UInt8	colorIndexes[kNumEdgeColors];

	Rect	r = *bounds;
	
	if (isPressed)
		{
		colorIndexes[TLFrame] = rA2;
		colorIndexes[BRFrame] = r11;
		colorIndexes[CornerFrame] = r12;
		colorIndexes[TLInner] = r10;
		colorIndexes[BRInner] = r5;
		colorIndexes[CornerInner] = r8;
		colorIndexes[Fill] = r7;
		}
	else if (isEnabled)
		{
		colorIndexes[TLFrame] = isOn ? rA2 : r9;
		colorIndexes[BRFrame] = isOn ? r11 : rA1;
		colorIndexes[CornerFrame] = isOn ? r12 : r10;
		colorIndexes[TLInner] = isOn ? r10 : rW;
		colorIndexes[BRInner] = isOn ? r5 : r7;
		colorIndexes[CornerInner] = isOn ? r8 : r3;
		colorIndexes[Fill] = isOn ? r7 : r3;
		}
	else	// disabled
		{
		colorIndexes[TLFrame] = isOn ? r8 : r5;
		colorIndexes[BRFrame] = r6;
		colorIndexes[CornerFrame] = isOn ? r7 : r5;
		colorIndexes[TLInner] = isOn ? r4 : r2;
		colorIndexes[BRInner] = isOn ? r4 : r2;
		colorIndexes[CornerInner] = isOn ? r4 : r2;
		colorIndexes[Fill] = isOn ? r4 : r2;
		}
	
	// Draw the outer frame.

	::RGBForeColor(&gAGARamp[colorIndexes[TLFrame]]);
	::MoveTo(r.left, r.bottom - 2);
	::LineTo(r.left, r.top);
	::LineTo(r.right - 1, r.top);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRFrame]]);
	::Move(0, 1);
	::LineTo(r.right - 1, r.bottom - 1);
	::LineTo(r.left, r.bottom - 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	// Draw the 1st pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner]]);
	::MoveTo(r.left + 1, r.bottom - 3);
	::LineTo(r.left + 1, r.top + 1);
	::LineTo(r.right - 2, r.top + 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner]]);
	::Move(0, 1);
	::LineTo(r.right - 2, r.bottom - 2);
	::LineTo(r.left + 1, r.bottom - 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner]]);
	::Line(0, 0);

	// Fill the interior.

	::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
	::InsetRect(&r, 2, 2);
	::PaintRect(&r);
	::InsetRect(&r, -2, -2);
	}

void MIconButtonObject::DrawSmallEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	// Draw the button edges as specified for a
	// "small" sized button.

	enum { TLFrame, TLInner1, TLInner2, BRFrame, BRInner1, BRInner2, CornerFrame, CornerInner1, CornerInner2, Fill, kNumEdgeColors };

	UInt8	colorIndexes[kNumEdgeColors];

	Rect	r = *bounds;
	
	if (isPressed)
		{
		colorIndexes[TLFrame] = rA2;
		colorIndexes[BRFrame] = r11;
		colorIndexes[CornerFrame] = rA2;
		colorIndexes[TLInner1] = r11;
		colorIndexes[BRInner1] = r4;
		colorIndexes[CornerInner1] = r8;
		colorIndexes[TLInner2] = r9;
		colorIndexes[BRInner2] = r6;
		colorIndexes[CornerInner2] = r7;
		colorIndexes[Fill] = r7;
		}
	else if (isEnabled)
		{
		colorIndexes[TLFrame] = isOn ? rA2 : r9;
		colorIndexes[BRFrame] = isOn ? r11 : rA1;
		colorIndexes[CornerFrame] = isOn ? rA2 : r9;
		colorIndexes[TLInner1] = isOn ? r11 : r3;
		colorIndexes[BRInner1] = isOn ? r4 : r8;
		colorIndexes[CornerInner1] = isOn ? r8 : r5;
		colorIndexes[TLInner2] = isOn ? r9 : rW;
		colorIndexes[BRInner2] = r6;
		colorIndexes[CornerInner2] = isOn ? r7 : r3;
		colorIndexes[Fill] = isOn ? r7 : r3;
		}
	else	// disabled
		{
		colorIndexes[TLFrame] = isOn ? r8 : r5;
		colorIndexes[BRFrame] = r6;
		colorIndexes[CornerFrame] = isOn ? r8 : r5;
		colorIndexes[TLInner1] = isOn ? r4 : r2;
		colorIndexes[BRInner1] = isOn ? r4 : r2;
		colorIndexes[CornerInner1] = isOn ? r4 : r2;
		colorIndexes[TLInner2] = isOn ? r4 : r2;
		colorIndexes[BRInner2] = isOn ? r4 : r2;
		colorIndexes[CornerInner2] = isOn ? r4 : r2;
		colorIndexes[Fill] = isOn ? r4 : r2;
		}
	
	// Draw the outer frame.

	::RGBForeColor(&gAGARamp[colorIndexes[TLFrame]]);
	::MoveTo(r.left, r.bottom - 2);
	::LineTo(r.left, r.top);
	::LineTo(r.right - 1, r.top);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRFrame]]);
	::Move(0, 1);
	::LineTo(r.right - 1, r.bottom - 1);
	::LineTo(r.left, r.bottom - 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	// Draw the 1st pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner1]]);
	::MoveTo(r.left + 1, r.bottom - 3);
	::LineTo(r.left + 1, r.top + 1);
	::LineTo(r.right - 2, r.top + 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner1]]);
	::Move(0, 1);
	::LineTo(r.right - 2, r.bottom - 2);
	::LineTo(r.left + 1, r.bottom - 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	// Draw the 2nd pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner2]]);
	::MoveTo(r.left + 2, r.bottom - 4);
	::LineTo(r.left + 2, r.top + 2);
	::LineTo(r.right - 3, r.top + 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner2]]);
	::Move(0, 1);
	::LineTo(r.right - 3, r.bottom - 3);
	::LineTo(r.left + 2, r.bottom - 3);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	// Fill the interior.

	::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
	::InsetRect(&r, 3, 3);
	::PaintRect(&r);
	::InsetRect(&r, -3, -3);
	}

void MIconButtonObject::DrawLargeEdges(Rect* bounds, Boolean isOn, Boolean isPressed, Boolean isEnabled)
	{
	// Draw the button edges as specified for a
	// "large" sized button.

	enum { TLFrame, TLInner1, TLInner2, TLInner3, BRFrame, BRInner1, BRInner2, BRInner3, CornerFrame, CornerInner1, CornerInner2, CornerInner3, Fill, kNumEdgeColors };

	UInt8	colorIndexes[kNumEdgeColors];

	Rect	r = *bounds;
	
	if (isPressed)
		{
		colorIndexes[TLFrame] = rA2;
		colorIndexes[BRFrame] = r11;
		colorIndexes[CornerFrame] = r12;
		colorIndexes[TLInner1] = rA1;
		colorIndexes[BRInner1] = r4;
		colorIndexes[CornerInner1] = r11;
		colorIndexes[TLInner2] = r11;
		colorIndexes[BRInner2] = r5;
		colorIndexes[CornerInner2] = r8;
		colorIndexes[TLInner3] = r9;
		colorIndexes[BRInner3] = r6;
		colorIndexes[CornerInner3] = r7;
		colorIndexes[Fill] = r7;
		}
	else if (isEnabled)
		{
		colorIndexes[TLFrame] = isOn ? rA2 : r9;
		colorIndexes[BRFrame] = isOn ? r11 : rA1;
		colorIndexes[CornerFrame] = isOn ? r12 : r10;
		colorIndexes[TLInner1] = isOn ? rA1 : r3;
		colorIndexes[BRInner1] = isOn ? r4 : r10;
		colorIndexes[CornerInner1] = isOn ? r11 : r5;
		colorIndexes[TLInner2] = isOn ? r11 : r1;
		colorIndexes[BRInner2] = isOn ? r5 : r8;
		colorIndexes[CornerInner2] = isOn ? r8 : r4;
		colorIndexes[TLInner3] = isOn ? r9 : rW;
		colorIndexes[BRInner3] = r6;
		colorIndexes[CornerInner3] = isOn ? r7 : r3;
		colorIndexes[Fill] = isOn ? r7 : r3;
		}
	else	// disabled
		{
		colorIndexes[TLFrame] = isOn ? r8 : r5;
		colorIndexes[BRFrame] = isOn ? r6 : r7;
		colorIndexes[CornerFrame] = isOn ? r7 : r6;
		colorIndexes[TLInner1] = isOn ? r4 : r2;
		colorIndexes[BRInner1] = isOn ? r4 : r2;
		colorIndexes[CornerInner1] = isOn ? r4 : r2;
		colorIndexes[TLInner2] = isOn ? r4 : r2;
		colorIndexes[BRInner2] = isOn ? r4 : r2;
		colorIndexes[CornerInner2] = isOn ? r4 : r2;
		colorIndexes[TLInner3] = isOn ? r4 : r2;
		colorIndexes[BRInner3] = isOn ? r4 : r2;
		colorIndexes[CornerInner3] = isOn ? r4 : r2;
		colorIndexes[Fill] = isOn ? r4 : r2;
		}
	
	// Draw the outer frame.

	::RGBForeColor(&gAGARamp[colorIndexes[TLFrame]]);
	::MoveTo(r.left, r.bottom - 2);
	::LineTo(r.left, r.top);
	::LineTo(r.right - 1, r.top);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRFrame]]);
	::Move(0, 1);
	::LineTo(r.right - 1, r.bottom - 1);
	::LineTo(r.left, r.bottom - 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerFrame]]);
	::Line(0, 0);

	// Draw the 1st pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner1]]);
	::MoveTo(r.left + 1, r.bottom - 3);
	::LineTo(r.left + 1, r.top + 1);
	::LineTo(r.right - 2, r.top + 1);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner1]]);
	::Move(0, 1);
	::LineTo(r.right - 2, r.bottom - 2);
	::LineTo(r.left + 1, r.bottom - 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner1]]);
	::Line(0, 0);

	// Draw the 2nd pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner2]]);
	::MoveTo(r.left + 2, r.bottom - 4);
	::LineTo(r.left + 2, r.top + 2);
	::LineTo(r.right - 3, r.top + 2);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner2]]);
	::Move(0, 1);
	::LineTo(r.right - 3, r.bottom - 3);
	::LineTo(r.left + 2, r.bottom - 3);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner2]]);
	::Line(0, 0);

	// Draw the 3rd pixel of shading.

	::RGBForeColor(&gAGARamp[colorIndexes[TLInner3]]);
	::MoveTo(r.left + 3, r.bottom - 5);
	::LineTo(r.left + 3, r.top + 3);
	::LineTo(r.right - 4, r.top + 3);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner3]]);
	::Line(0, 0);

	::RGBForeColor(&gAGARamp[colorIndexes[BRInner3]]);
	::Move(0, 1);
	::LineTo(r.right - 4, r.bottom - 4);
	::LineTo(r.left + 3, r.bottom - 4);

	::RGBForeColor(&gAGARamp[colorIndexes[CornerInner3]]);
	::Line(0, 0);

	// Fill the interior.

	::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
	::InsetRect(&r, 4, 4);
	::PaintRect(&r);
	::InsetRect(&r, -4, -4);
	}

PicHandle MIconButtonObject::GetPictureImage(Point* imageSize, SInt16 pictureResourceID)
	{
	PicHandle	pictureHandle = ::GetPicture(pictureResourceID);
					
	if (pictureHandle != NULL)
		{
		Rect	frame = (**pictureHandle).picFrame;

		::SetPt(imageSize, frame.right - frame.left, frame.bottom - frame.top);
		}

	return pictureHandle;
	}

CIconHandle MIconButtonObject::GetColorIconImage(Point* imageSize, SInt16 cicnResourceID)
	{
	CIconHandle	colorIconHandle = ::GetCIcon(cicnResourceID);
					
	if (colorIconHandle != NULL)
		{
		Rect	frame = (**colorIconHandle).iconBMap.bounds;

		::SetPt(imageSize, frame.right - frame.left, frame.bottom - frame.top);
		}

	return colorIconHandle;
	}

MIconButtonObject::ButtonImageType MIconButtonObject::GetDefaultImageType(SInt16 buttonWidth)
	{
	if (buttonWidth > 39)
		return kLargeIcon;
	else if (buttonWidth > 21)
		return kSmallIcon;
	else
		return kMiniIcon;
	}

//
// AGAIconPushButton ---------------------------------------------------------------
//
// This class implements an icon-based pushbutton object, such as
// a toolbar icon command button.
//

AGAIconPushButton::AGAIconPushButton(Rect* bounds)
: AGAPushButton(bounds, gAGAStdSystemStyle), MIconButtonObject()
	{
	// Just call through to mixed in constructors.
	}

AGAIconPushButton::AGAIconPushButton(Rect* bounds, SInt16 iconID, ButtonFrameType frameType, ButtonImageType imageType)
: AGAPushButton(bounds, gAGAStdSystemStyle), MIconButtonObject(iconID, iconID, frameType, imageType)
	{
	// Just call through to mixed in constructors.
	}

AGAIconPushButton::~AGAIconPushButton()
	{
	}

void AGAIconPushButton::DrawButtonNormal(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, false, kNotPressed, kEnabled, deep);
	}

void AGAIconPushButton::DrawButtonPressed(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, false, kPressed, kEnabled, deep);
	}

void AGAIconPushButton::DrawButtonDisabled(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, false, kNotPressed, kDisabled, deep);
	}

//
// AGAIconCheckBox ---------------------------------------------------------------
//
// This class implements an icon-based checkbox object, such as
// a toolbar icon on/off button.
//

AGAIconCheckBox::AGAIconCheckBox(Rect* bounds, Boolean automaticState)
: AGACheckBox(bounds, gAGAStdSystemStyle, automaticState), MIconButtonObject()
	{
	// Just call through to mixed in constructors.
	}

AGAIconCheckBox::AGAIconCheckBox(Rect* bounds, Boolean automaticState, SInt16 offIconID, SInt16 onIconID, ButtonFrameType frameType, ButtonImageType imageType)
: AGACheckBox(bounds, gAGAStdSystemStyle, automaticState), MIconButtonObject(offIconID, onIconID, frameType, imageType)
	{
	// Just call through to mixed in constructors.
	}

AGAIconCheckBox::~AGAIconCheckBox()
	{
	}

void AGAIconCheckBox::DrawButtonNormal(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, mValue != kCheckBoxOff, kNotPressed, kEnabled, deep);
	}

void AGAIconCheckBox::DrawButtonPressed(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, mValue != kCheckBoxOff, kPressed, kEnabled, deep);
	}

void AGAIconCheckBox::DrawButtonDisabled(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, mValue != kCheckBoxOff, kNotPressed, kDisabled, deep);
	}

//
// AGAIconRadioButton ---------------------------------------------------------------
//
// This class implements an icon-based pushbutton object, such as
// a toolbar icon XOR group member button.
//

AGAIconRadioButton::AGAIconRadioButton(Rect* bounds, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState)
: AGARadioButton(bounds, gAGAStdSystemStyle, groupIDPart1, groupIDPart2, automaticState), MIconButtonObject()
	{
	// Just call through to mixed in constructors.
	}

AGAIconRadioButton::AGAIconRadioButton(Rect* bounds, UInt32 groupIDPart1, UInt32 groupIDPart2, Boolean automaticState, SInt16 offIconID, SInt16 onIconID, ButtonFrameType frameType, ButtonImageType imageType)
: AGARadioButton(bounds, gAGAStdSystemStyle, groupIDPart1, groupIDPart2, automaticState), MIconButtonObject(offIconID, onIconID, frameType, imageType)
	{
	// Just call through to mixed in constructors.
	}

AGAIconRadioButton::~AGAIconRadioButton()
	{
	}

void AGAIconRadioButton::DrawButtonNormal(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, mValue != kRadioButtonOff, kNotPressed, kEnabled, deep);
	}

void AGAIconRadioButton::DrawButtonPressed(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, mValue != kRadioButtonOff, kPressed, kEnabled, deep);
	}

void AGAIconRadioButton::DrawButtonDisabled(Boolean deep)
	{
	// Call icon button parent class with appropriate parameters.

	MIconButtonObject::DrawIconButton(&mBounds, mValue != kRadioButtonOff, kNotPressed, kDisabled, deep);
	}

//
// AGASlider ---------------------------------------------------------------
//
// This class implements a standard AGA slider, with the additional
// ability to perform live tracking and a proportional thumb.
//

// These regions are used momentarily while tracking the slider
// indicator to clip and optimize redrawing.
RgnHandle AGASlider::mgSavedIndicatorClip = ::NewRgn();
RgnHandle AGASlider::mgOldIndicatorClip = ::NewRgn();
RgnHandle AGASlider::mgNewIndicatorClip = ::NewRgn();

AGASlider::AGASlider(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, const AGATextStyle& textStyle, SInt16 labelsID)
	: AGATrackingIndicator(bounds, minimum, maximum, initialValue)
	{
	// Construct with default options.
	mPageSize = 0;

	mLabelStringHandles = NULL;
	mJustification = teFlushDefault;
	mTextStyle = textStyle;

	if (labelsID == 0)
		{
		mSliderKind = kRectSlider;

		mLiveTracking = TestGrayCouncilDefault(kAGALiveRSliders);
		mIsProportional = TestGrayCouncilDefault(kAGAProportionalRSliders);
		}
	else
		{
		mSliderKind = kPointerSlider;
		this->SetLabelsFromResource(labelsID);

		mLiveTracking = TestGrayCouncilDefault(kAGALivePSliders);
		mIsProportional = TestGrayCouncilDefault(kAGAProportionalPSliders);
		}

	mInverseDirection = TestGrayCouncilDefault(mIsHorizontal ? kAGAInverseHSliderValues : kAGAInverseVSliderValues);
	}

AGASlider::AGASlider(Rect* bounds, SInt32 minimum, SInt32 maximum, SInt32 initialValue, const AGATextStyle& textStyle, SInt16 labelsID, SliderKind kind, Boolean liveTracking, Boolean proportional)
	: AGATrackingIndicator(bounds, minimum, maximum, initialValue, 1, liveTracking, proportional)
	{
	// Construct with specified options.

	mPageSize = 0;

	mLabelStringHandles = NULL;
	mJustification = teFlushDefault;
	mSliderKind = kind;
	mTextStyle = textStyle;
	mInverseDirection = TestGrayCouncilDefault(mIsHorizontal ? kAGAInverseHSliderValues : kAGAInverseVSliderValues);

	this->SetLabelsFromResource(labelsID);
	}

AGASlider::~AGASlider()
	{
	// Destruct by releasing memory allocated for labels.

	this->DisposeLabels();
	}

void AGASlider::SetSliderKind(SliderKind kind)
	{
	// Set the slider kind (pointy or rectangular).

	mSliderKind = kind;
	}

void AGASlider::SetInverseDirection(Boolean inverseDirection)
	{
	// Set the inverse value direction indicator.

	mInverseDirection = inverseDirection;
	}

SInt32 AGASlider::GetValue()
	{
	// Tweak the return value if using inverse values.

	SInt32	value = AGATrackingIndicator::GetValue();
	
	if (mInverseDirection)
		value = mMaximum - (value - mMinimum);
	
	return value;
	}

void AGASlider::SetValue(SInt32 newValue, Boolean redraw)
	{
	// Tweak the input value if using inverse values.

	SInt32	actualValue = newValue;
	
	if (mInverseDirection)
		actualValue = mMaximum - (newValue - mMinimum);
	
	AGATrackingIndicator::SetValue(actualValue, redraw);
	}

void AGASlider::SetLabelsFromResource(SInt16 stringListResourceID)
	{
	// Build the slider labels from the specified 'STR#' resource.
	// A label slot is allocated for each string in the resource;
	// there will be a visual label tick for each of these strings.
	// If the current min/max range is smaller than the number of
	// labels (i.e., it's uninitialized), the min/max range will
	// be set to match the number of labels.

	this->DisposeLabels();
	
	if (stringListResourceID != 0)
		{
		SInt16**	stringListHandle = (SInt16**) ::GetResource('STR#', stringListResourceID);
		SInt16		numLabels = (stringListHandle == NULL) ? 0 : **stringListHandle;
		
		if (numLabels != 0)
			{
			mLabelStringHandles = ::NewHandleClear(numLabels * sizeof(StringHandle));
			
			for (SInt16 i = 0; i < numLabels; i++)
				{
				Str255			aLabel;
				StringHandle	aStringHandle;
				
				::GetIndString(aLabel, stringListResourceID, i+1);
				aStringHandle = ::NewString(aLabel);
				
				((StringHandle*)(*mLabelStringHandles))[i] = aStringHandle;
				}
			}
	
		if ((mMaximum - mMinimum + 1) < numLabels)
			mMaximum = mMinimum + numLabels - 1;
		}
	}

void AGASlider::SetNumLabels(UInt32 numLabels)
	{
	// Allocate slots for the specified number of labels.
	// Any existing labels are thrown away.

	this->DisposeLabels();
	
	mLabelStringHandles = ::NewHandleClear(numLabels * sizeof(StringHandle));

	if ((mMaximum - mMinimum + 1) < numLabels)
		mMaximum = mMinimum + numLabels - 1;
	}

void AGASlider::SetLabel(UInt32 labelIndex, StringPtr itsLabel)
	{
	// Assign the specified string to the specified label slot.

	if (mLabelStringHandles != NULL)
		{
		UInt32	numLabels = ::GetHandleSize(mLabelStringHandles) / sizeof(StringHandle);
		
		if (labelIndex < numLabels)
			((StringHandle*)(*mLabelStringHandles))[labelIndex] = ::NewString(itsLabel);
		}
	}

void AGASlider::GetLabel(UInt32 labelIndex, StringPtr itsLabel)
	{
	// Return the label string of the specified label slot.

	itsLabel[0] = 0;

	if (mLabelStringHandles != NULL)
		{
		UInt32	numLabels = ::GetHandleSize(mLabelStringHandles) / sizeof(StringHandle);
		
		if (labelIndex < numLabels)
			{
			Ptr	sourceStringPtr = (Ptr) *(((StringHandle*)(*mLabelStringHandles))[labelIndex]);

			::BlockMoveData(sourceStringPtr, itsLabel, 1 + *sourceStringPtr);
			}
		}
	}

void AGASlider::SetLabelsStyle(const AGATextStyle& textStyle)
	{
	// Set the QuickDraw text style for the labels.

	mTextStyle = textStyle;
	}

void AGASlider::SetJustification(SInt32 justification)
	{
	// Set the justification of the slider. For vertical
	// sliders, default/left means the slider is on the
	// left and the labels are on the right; flush right
	// reverses this. For horizontal sliders, default/left
	// means the slider is above and the labels are below;
	// flush right reverses this.
	// In fact, for "teFlushDefault", the system script
	// direction is checked at runtime, so this is usually
	// what you should use. Use flush right or flush left
	// to force it one way or another.

	mJustification = justification;
	}

SInt32 AGASlider::TrackTestPart(Point mouseLocation)
	{
	// Return the indicator part value if the mouse location
	// would hit the indicator.

	Rect	indicatorRect;
	
	this->GetIndicatorBox(mValue, &indicatorRect);
	
	if (::PtInRect(mouseLocation, &indicatorRect))
		return kIndicatorPress;
	else
		return kNoTrackPress;
	}

Boolean AGASlider::IsValidIndicatorTrackMouse(Point mouseLocation)
	{
	// Return true if the mouse location is reasonably close
	// to the indicator track. This will depend on the
	// justification.

	Rect	okRect = mBounds;
	
	if (mIsHorizontal)
		{
		if (RuntimeJustify(mJustification) == teFlushLeft)
			okRect.bottom = okRect.top + kIndicatorWidth;
		else
			okRect.top = okRect.bottom - kIndicatorWidth;
		}
	else
		{
		if (RuntimeJustify(mJustification) == teFlushLeft)
			okRect.right = okRect.left + kIndicatorWidth;
		else
			okRect.left = okRect.right - kIndicatorWidth;
		}
	
	::InsetRect(&okRect, -32, -32);	// slop for mouse tracking

	return ::PtInRect(mouseLocation, &okRect);
	}

SInt32 AGASlider::GetValueFromMouseDelta(SInt32 originalValue, Point originalMouseLocation, Point newMouseLocation)
	{
	// Return the new potential indicator value based on the mouse
	// delta from the original click location and the original value.

	SInt32	dontCare;
	SInt32	valueRange = mMaximum - mMinimum;
	SInt32	pixelRange = this->GetIndicatorPixelRange(&dontCare);
	SInt32	pixelDelta;
	float	pixelFraction;
	
	if (mIsHorizontal)
		pixelDelta = newMouseLocation.h - originalMouseLocation.h;
	else
		pixelDelta = newMouseLocation.v - originalMouseLocation.v;

	pixelFraction = ((float) pixelDelta) / (float) pixelRange;
	
	float delta = pixelFraction * ((float) valueRange);
	delta += (float) 0.5;	// round to nearest integer, don't just truncate fraction
	
	return originalValue + delta;
	}

void AGASlider::DrawBackground()
	{
	AGADrawingEnvironment	env;

	// Paint the background and labels.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			this->ApplyBackgroundColor();
		else
			::RGBForeColor(&gAGARamp[rW]);
		}

	::PaintRect(&mBounds);
	
	iter.Cleanup();	// Restore clipping since not yet destructed.

	this->DrawLabels();
	}

void AGASlider::DrawIndicatorTrack()
	{
	AGADrawingEnvironment	env;

	// Draw the empty slider track.

	GDIterator	iter;
	Boolean		deep;
	Boolean		drawEnabled = mEnabled && mActive;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			Rect	trackFrame;
			
			this->GetTrackFrame(&trackFrame, kIncludeBackground);

			this->ApplyBackgroundColor();
			::PaintRect(&trackFrame);
			
			this->GetTrackFrame(&trackFrame, kTrackOnly);

			if (drawEnabled)
				::RGBForeColor(&gAGARamp[rB]);
			else
				::RGBForeColor(&gAGARamp[r8]);
			
			::InsetRect(&trackFrame, 1, 1);
			::MoveTo(trackFrame.left + 1, trackFrame.top);
			::LineTo(trackFrame.right - 2, trackFrame.top);
			::Move(1, 1);
			::LineTo(trackFrame.right - 1, trackFrame.bottom - 2);
			::Move(-1, 1);
			::LineTo(trackFrame.left + 1, trackFrame.bottom - 1);
			::Move(-1, -1);
			::LineTo(trackFrame.left, trackFrame.top + 1);

			::InsetRect(&trackFrame, 1, 1);

			if (drawEnabled)
				::RGBForeColor(&gAGARamp[r5]);
			else
				::RGBForeColor(&gAGARamp[r4]);
			
			::PaintRect(&trackFrame);

			if (drawEnabled)
				{
				::InsetRect(&trackFrame, -2, -2);
				::MoveTo(trackFrame.left, trackFrame.bottom - 3);
				::LineTo(trackFrame.left, trackFrame.top + 1);
				::Line(1, 0);
				::Line(0, -1);
				::LineTo(trackFrame.right - 3, trackFrame.top);
				
				if (mIsHorizontal)
					{
					::Move(1, 1);
					::Line(0, 0);
					}
				
				::RGBForeColor(&gAGARamp[rW]);
				::MoveTo(trackFrame.left + 1, trackFrame.bottom - 2);
				::Line(0, 0);
				::Move(1, 1);
				::LineTo(trackFrame.right - 2, trackFrame.bottom - 1);
				::Line(0, -1);
				::Line(1, 0);
				::LineTo(trackFrame.right -  1, trackFrame.top + 2);

				if (! mIsHorizontal)
					{
					::Move(-1, -1);
					::Line(0, 0);
					}
				}
			}
		else	// 1 -bit
			{
			Rect	trackFrame;
			
			this->GetTrackFrame(&trackFrame, kIncludeBackground);

			::RGBForeColor(&gAGARamp[rW]);
			::PaintRect(&trackFrame);
			
			this->GetTrackFrame(&trackFrame, kTrackOnly);

			::RGBForeColor(&gAGARamp[rB]);
			
			if (! drawEnabled)
				PenPatGray();
			
			::InsetRect(&trackFrame, 1, 1);
			
			::MoveTo(trackFrame.left + 1, trackFrame.top);
			::LineTo(trackFrame.right - 2, trackFrame.top);
			::Move(1, 1);
			::LineTo(trackFrame.right - 1, trackFrame.bottom - 2);
			::Move(-1, 1);
			::LineTo(trackFrame.left + 1, trackFrame.bottom - 1);
			::Move(-1, -1);
			::LineTo(trackFrame.left, trackFrame.top + 1);
			
			if (! drawEnabled)
				PenPatBlack();
			}
		}
	}

void AGASlider::DrawIndicator()
	{
	AGADrawingEnvironment	env;

	// Draw the slider indicator.

	Rect	indicatorRect;
	
	this->GetIndicatorBox(mValue, &indicatorRect);

	if (mSliderKind == kRectSlider)
		this->DrawRectIndicator(&indicatorRect, kNormalIndicator);
	else
		this->DrawPointerIndicator(&indicatorRect, kNormalIndicator);
	}

void AGASlider::DrawGhost()
	{
	AGADrawingEnvironment	env;

	// Draw the ghost slider indicator.

	Rect	indicatorRect;
	
	this->GetIndicatorBox(mGhostValue, &indicatorRect);

	if (mSliderKind == kRectSlider)
		this->DrawRectIndicator(&indicatorRect, kGhostIndicator);
	else
		this->DrawPointerIndicator(&indicatorRect, kGhostIndicator);
	}

void AGASlider::DrawLabels()
	{
	AGADrawingEnvironment	env;

	// Draw the slider labels and ticks.

	UInt32	numHandles = 0;
	Boolean	drawEnabled = mEnabled && mActive;

	if (mLabelStringHandles != NULL)
		numHandles = ::GetHandleSize(mLabelStringHandles) / sizeof(StringHandle);

	if (numHandles < 2)	// Gotta at least label the ends, man!
		return;
	
	SInt16		tickRange;
	Point		tickOrigin;
	Point		drawDirection;
	Point		moveDirection;
	Rect		trackFrame;
	Boolean		flushLeft = (RuntimeJustify(mJustification) == teFlushLeft);
	FontInfo	fontInfo;

	mTextStyle.PrepareForDrawing();
	::GetFontInfo(&fontInfo);
	
	this->GetTrackFrame(&trackFrame, kTrackOnly);

	if (mIsHorizontal)
		{
		tickRange = trackFrame.right - trackFrame.left - (2 * kLabelTickEndInset) - 1;
		drawDirection.h = 0;
		drawDirection.v = 1;
		moveDirection.h = 1;
		moveDirection.v = 0;
		tickOrigin.h = trackFrame.left + kLabelTickEndInset;
		
		if (flushLeft)
			tickOrigin.v = trackFrame.bottom + kLabelTickOffset;
		else
			tickOrigin.v = trackFrame.top - kLabelTickOffset - 2 - kLabelTickLength;
		}
	else
		{
		tickRange = trackFrame.bottom - trackFrame.top - (2 * kLabelTickEndInset) - 1;
		drawDirection.h = 1;
		drawDirection.v = 0;
		moveDirection.h = 0;
		moveDirection.v = 1;
		tickOrigin.v = trackFrame.top + kLabelTickEndInset;
		
		if (flushLeft)
			tickOrigin.h = trackFrame.right + kLabelTickOffset;
		else
			tickOrigin.h = trackFrame.left - kLabelTickOffset - 2 - kLabelTickLength;
		}
	
	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		for (UInt32 i = 0; i < numHandles; i++)
			{
			Point	tickLocation;
			float	fraction = ((float) i) / ((float) (numHandles - 1));
			
			tickLocation.h = tickOrigin.h + (moveDirection.h * tickRange * fraction);
			tickLocation.v = tickOrigin.v + (moveDirection.v * tickRange * fraction);

			::MoveTo(tickLocation.h, tickLocation.v);
			
			::RGBForeColor(&gAGARamp[rB]);
			
			if (! drawEnabled)
				{
				if (deep)
					::RGBForeColor(&gAGARamp[r7]);
				else
					PenPatGray();
				}

			::Line(drawDirection.h * (kLabelTickLength + 1), drawDirection.v * (kLabelTickLength + 1));

			::PenNormal();

			if (deep)
				{
				// If disabled, we still draw in order to erase the 7/W
				// shading pixels from prior enabled drawing.

				if (drawEnabled)
					::RGBForeColor(&gAGARamp[r7]);
				else
					this->ApplyBackgroundColor();

				::Line(moveDirection.h, moveDirection.v);
				::Line( - (kLabelTickLength * drawDirection.h), - (kLabelTickLength * drawDirection.v));
				
				if (drawEnabled)
					::RGBForeColor(&gAGARamp[rW]);
				
				::Move(-1, -1);
				::Line(-moveDirection.h, -moveDirection.v);
				::Line(drawDirection.h * kLabelTickLength, drawDirection.v * kLabelTickLength);
				}

			Str255	aLabel;
			Rect	labelBox;
			SInt16	labelHeight = fontInfo.ascent + fontInfo.descent;
			SInt16	labelWidth;
			SInt32	textJustification;

			this->GetLabel(i, aLabel);
			labelWidth = ::StringWidth(aLabel);
			
			if (mIsHorizontal)
				{
				textJustification = teCenter;
				labelBox.left = tickLocation.h - (labelWidth / 2);
				labelBox.right = labelBox.left + labelWidth;

				if (flushLeft)
					{
					labelBox.top = tickLocation.v + 9;
					labelBox.bottom = labelBox.top + labelHeight;
					}
				else
					{
					labelBox.bottom = tickLocation.v - 2;
					labelBox.top = labelBox.bottom - labelHeight;
					}
				}
			else
				{
				textJustification = mJustification;
				labelBox.top = tickLocation.v - (labelHeight / 2);
				labelBox.bottom = labelBox.top + labelHeight;

				if (flushLeft)
					{
					labelBox.left = tickLocation.h + 11;
					labelBox.right = labelBox.left + labelWidth;
					}
				else
					{
					labelBox.right = tickLocation.h - 2;
					labelBox.left = labelBox.right - labelWidth;
					}
				}

			//
			// To make view layout simpler, we assume that the labels may
			// extend outside the mBounds. So if this happens
			// we need to make sure we aren't being clipped out, but also
			// make sure we don't extend the unclipped area out to other
			// screens during device iteration.
			//
			
			RgnHandle	savedRegion = NULL;

			if ((labelBox.left < mBounds.left) ||
				(labelBox.right > mBounds.right))
				{
				savedRegion = ::NewRgn();
				::GetClip(savedRegion);
				::InsetRect(&labelBox, -1, -1);
				::ClipRect(&labelBox);
				::InsetRect(&labelBox, 1, 1);
				
				iter.ClipFurtherToCurrentDevice();
				}

			AGAStringOut(aLabel, &labelBox, kNoTruncation, textJustification, drawEnabled ? kNormalOutput : kDisabledOutput, deep, mTextStyle);

			if (savedRegion != NULL)
				{
				::SetClip(savedRegion);
				::DisposeRgn(savedRegion);
				}
			}
		}
	}

void AGASlider::DrawRectIndicator(Rect* indicatorRect, Boolean isGhost)
	{
	AGADrawingEnvironment	env;

	// Draw the rectangular slider indicator at the
	// specified location. The location will be sized
	// to account for a proportional indicator.

	enum { Frame, Fill, Corner, Light, TL, BR, Dark, Grip, BG, kNumIndicatorColors };

	UInt8	colorIndexes[kNumIndicatorColors];
	Rect	r = *indicatorRect;
	Point	drawDirection;
	Point	moveDirection;
	SInt16	lineLength;

	GDIterator	iter;
	Boolean		deep;
	Boolean		drawEnabled = mEnabled && mActive;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			colorIndexes[BG] = r2;

			if (isGhost)
				{
				colorIndexes[Frame] = r7;
				colorIndexes[Fill] = r2;
				colorIndexes[Corner] = r1;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = r4;
				colorIndexes[Dark] = r5;
				colorIndexes[Grip] = rW;
				}
			else if (! drawEnabled)
				{
				colorIndexes[Frame] = r8;
				colorIndexes[Fill] = r2;
				colorIndexes[Corner] = r2;
				colorIndexes[Light] = r2;
				colorIndexes[TL] = r2;
				colorIndexes[BR] = r2;
				// n/a colorIndexes[Dark] = r9;
				colorIndexes[Grip] = r2;
				}
			else if (mIsPressed)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = r8;
				colorIndexes[Corner] = r3;
				colorIndexes[Light] = r3;
				colorIndexes[TL] = r5;
				colorIndexes[BR] = r10;
				colorIndexes[Dark] = r12;
				colorIndexes[Grip] = r5;
				}
			else // normal
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = r5;
				colorIndexes[Corner] = r1;
				colorIndexes[Light] = r1;
				colorIndexes[TL] = r3;
				colorIndexes[BR] = r8;
				colorIndexes[Dark] = r10;
				colorIndexes[Grip] = r3;
				}
			}
		else	// 1-bit
			{
			colorIndexes[BG] = rW;

			if (isGhost)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rW;
				colorIndexes[Corner] = rW;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = rW;
				colorIndexes[Dark] = rW;
				colorIndexes[Grip] = rB;
				}
			else if (! drawEnabled)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rW;
				colorIndexes[Corner] = rW;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = rW;
				colorIndexes[Dark] = rW;
				colorIndexes[Grip] = rW;
				}
			else if (mIsPressed)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rB;
				colorIndexes[Corner] = rB;
				colorIndexes[Light] = rB;
				colorIndexes[TL] = rB;
				colorIndexes[BR] = rB;
				colorIndexes[Dark] = rB;
				colorIndexes[Grip] = rW;
				}
			else	// normal
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rW;
				colorIndexes[Corner] = rW;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[BR] = rW;
				colorIndexes[Dark] = rW;
				colorIndexes[Grip] = rB;
				}
			}
		
		::RGBForeColor(&gAGARamp[colorIndexes[Frame]]);
		::FrameRect(&r);
		::RGBForeColor(&gAGARamp[colorIndexes[BG]]);
		::MoveTo(r.left, r.top);
		::Line(0, 0);
		::MoveTo(r.right - 1, r.top);
		::Line(0, 0);
		::MoveTo(r.right - 1, r.bottom - 1);
		::Line(0, 0);
		::MoveTo(r.left, r.bottom - 1);
		::Line(0, 0);
		
		::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
		::InsetRect(&r, 1, 1);
		::PaintRect(&r);
		::InsetRect(&r, -1, -1);
		
		::RGBForeColor(&gAGARamp[colorIndexes[Corner]]);
		::MoveTo(r.left + 1, r.top + 1);
		::Line(0, 0);
		
		::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
		::MoveTo(r.left + 1, r.bottom - 3);
		::LineTo(r.left + 1, r.top + 2);
		::Move(1, -1);
		::LineTo(r.right - 3, r.top + 1);
		
		::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
		::Move(1, 1);
		::LineTo(r.right - 2, r.bottom - 2);
		::LineTo(r.left + 2, r.bottom - 2);

		if (mIsHorizontal)
			{
			::SetPt(&drawDirection, 0, 1);
			::SetPt(&moveDirection, 1, 0);
			lineLength = r.bottom - r.top - 10;
			}
		else
			{
			::SetPt(&drawDirection, 1, 0);
			::SetPt(&moveDirection, 0, 1);
			lineLength = r.right - r.left - 10;
			}

		if (drawEnabled)
			{
			::MoveTo(r.left + 3 + drawDirection.h + (moveDirection.h * (((r.right - r.left) / 2) - 6)),
					r.top + 3 + drawDirection.v + (moveDirection.v * (((r.bottom - r.top) / 2) - 6)));

			if (! deep)
				::Move(moveDirection.h, moveDirection.v);

			for (SInt32 i = 0; i < 3; i++)
				{
				::RGBForeColor(&gAGARamp[colorIndexes[Light]]);
				::Line(drawDirection.h, drawDirection.v);
				
				::RGBForeColor(&gAGARamp[colorIndexes[Grip]]);
				::Line(drawDirection.h * (lineLength-1), drawDirection.v * (lineLength-1));
				
				::Move(moveDirection.h, moveDirection.v);
				::Move(drawDirection.h, drawDirection.v);
				::RGBForeColor(&gAGARamp[colorIndexes[Dark]]);
				::Line(- (drawDirection.h * lineLength), - (drawDirection.v * lineLength));

				::Move(moveDirection.h, moveDirection.v);
				::Move(- drawDirection.h, - drawDirection.v);
				}
			}
		
		if (isGhost && deep)
			{
			// Draw the little see-thru track slides.
			// They run inverse draw direction from the grip lines.
			// To handle proportional indicators, we need to vary
			// the length of the track slides on the flat part.
			SInt32	wholeFlatLength;
			SInt32	firstFlatLength;
			SInt32	remainingFlatLength;
			
			if (mIsHorizontal)
				wholeFlatLength = r.right - r.left - 10;
			else
				wholeFlatLength = r.bottom - r.top - 10;

			firstFlatLength = wholeFlatLength / 2;
			remainingFlatLength = wholeFlatLength - firstFlatLength; // don't round down twice

			::RGBForeColor(&gAGARamp[r11]);
			::MoveTo(r.left + (drawDirection.h * 5), r.top + (drawDirection.v * 5));
			
			for (UInt32 i = 0; i < 2; i++)
				{
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r5]);
				::Line(moveDirection.h * firstFlatLength, moveDirection.v * firstFlatLength);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r9]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r9]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r9]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r5]);
				::Line(moveDirection.h * remainingFlatLength, moveDirection.v * remainingFlatLength);
				::RGBForeColor(&gAGARamp[r7]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r11]);
				::Line(0, 0);
				
				::MoveTo(r.left + (drawDirection.h * 9), r.top + (drawDirection.v * 9));
				}
			}
		}
	}

void AGASlider::DrawPointerIndicator(Rect* indicatorRect, Boolean isGhost)
	{
	AGADrawingEnvironment	env;

	// Draw the pointy slider indicator at the
	// specified location. The location will be sized
	// to account for a proportional indicator.

	enum { Frame, Fill, Corner, Light, TL, Grip, BR, Dark, kNumIndicatorColors };

	UInt8	colorIndexes[kNumIndicatorColors];
	Rect	r = *indicatorRect;
	Point	drawDirection;
	Point	moveDirection;
	SInt16	lineLength;
	SInt16	propExtra;	// extra pixels due to proportional indicator shape

	GDIterator	iter;
	Boolean		deep;
	Boolean		drawEnabled = mEnabled && mActive;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			if (isGhost)
				{
				colorIndexes[Frame] = r7;
				colorIndexes[Fill] = r2;
				colorIndexes[Corner] = r1;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[Grip] = rW;
				colorIndexes[BR] = r4;
				colorIndexes[Dark] = r5;
				}
			else if (! drawEnabled)
				{
				colorIndexes[Frame] = r8;
				colorIndexes[Fill] = r2;
				colorIndexes[Corner] = r2;
				colorIndexes[Light] = r2;
				colorIndexes[TL] = r2;
				colorIndexes[Grip] = r2;
				colorIndexes[BR] = r2;
				// n/a colorIndexes[Dark] = r9;
				}
			else if (mIsPressed)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = r8;
				colorIndexes[Corner] = r3;
				colorIndexes[Light] = r3;
				colorIndexes[TL] = r5;
				colorIndexes[Grip] = r5;
				colorIndexes[BR] = r10;
				colorIndexes[Dark] = r12;
				}
			else // normal
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = r5;
				colorIndexes[Corner] = r1;
				colorIndexes[Light] = r1;
				colorIndexes[TL] = r3;
				colorIndexes[Grip] = r3;
				colorIndexes[BR] = r8;
				colorIndexes[Dark] = r10;
				}
			}
		else	// 1-bit
			{
			if (isGhost)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rW;
				colorIndexes[Corner] = rW;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[Grip] = rW;
				colorIndexes[BR] = rW;
				colorIndexes[Dark] = rW;
				}
			else if (! drawEnabled)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rW;
				colorIndexes[Corner] = rW;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[Grip] = rW;
				colorIndexes[BR] = rW;
				colorIndexes[Dark] = rW;
				}
			else if (mIsPressed)
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rB;
				colorIndexes[Corner] = rB;
				colorIndexes[Light] = rB;
				colorIndexes[TL] = rB;
				colorIndexes[Grip] = rW;
				colorIndexes[BR] = rB;
				colorIndexes[Dark] = rB;
				}
			else // normal
				{
				colorIndexes[Frame] = rB;
				colorIndexes[Fill] = rW;
				colorIndexes[Corner] = rW;
				colorIndexes[Light] = rW;
				colorIndexes[TL] = rW;
				colorIndexes[Grip] = rB;
				colorIndexes[BR] = rW;
				colorIndexes[Dark] = rW;
				}
			}

		if (mIsHorizontal)
			{
			::SetPt(&drawDirection, 0, 1);
			::SetPt(&moveDirection, 1, 0);
			lineLength = r.bottom - r.top - 11;
			propExtra = MaxSInt16(0, r.right - r.left - 15);
			}
		else
			{
			::SetPt(&drawDirection, 1, 0);
			::SetPt(&moveDirection, 0, 1);
			lineLength = r.right - r.left - 11;
			propExtra = MaxSInt16(0, r.bottom - r.top - 15);
			}
		
		PolyHandle	pointerPoly = ::OpenPoly();
		
		this->BuildPointerIndicator(indicatorRect);
		
		::ClosePoly();
		
		::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
		::PaintPoly(pointerPoly);
		::RGBForeColor(&gAGARamp[colorIndexes[Frame]]);
		::FramePoly(pointerPoly);
		
		::KillPoly(pointerPoly);

		if (mIsHorizontal)
			{
			if (RuntimeJustify(mJustification) == teFlushLeft)
				{
				::MoveTo(r.left + 1, r.top + 1);
				::RGBForeColor(&gAGARamp[colorIndexes[Corner]]);
				::Line(0, 0);
				::MoveTo(r.right - 2, r.top + 2);
				::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
				::LineTo(r.right - 2, r.top + 9);
				::Line(-5, 5);
				::Line(-2 - propExtra, 0);
				::Line(-5, -5);
				::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
				::LineTo(r.left + 1, r.top + 2);
				::Line(1, -1);
				::LineTo(r.right - 3, r.top + 1);
				}
			else
				{
				::MoveTo(r.right - 6, r.top + 2);
				::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
				::Line(4, 4);
				::LineTo(r.right - 2, r.bottom - 2);
				::LineTo(r.left + 2, r.bottom - 2);
				::Move(-1, -1);
				::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
				::Line(0, -7);
				::Line(5, -5);
				::Line(1 + propExtra, 0);
				::MoveTo(r.left + 6, r.top + 1);
				::RGBForeColor(&gAGARamp[colorIndexes[Corner]]);
				::Line(0, 0);
				::Move(-5, 5);
				::Line(0, 0);
				}
			}
		else
			{
			if (RuntimeJustify(mJustification) == teFlushLeft)
				{
				::MoveTo(r.left + 1, r.top + 1);
				::RGBForeColor(&gAGARamp[colorIndexes[Corner]]);
				::Line(0, 0);
				::MoveTo(r.left + 1, r.bottom - 3);
				::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
				::LineTo(r.left + 1, r.top + 2);
				::Line(1, -1);
				::Line(7, 0);
				::Move(1, 1);
				::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
				::Line(4, 4);
				::Line(0, 2 + propExtra);
				::Line(-5, 5);
				::Line(-7, 0);
				}
			else
				{
				::MoveTo(r.left + 2, r.top + 9 + propExtra);
				::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
				::Line(4, 4);
				::LineTo(r.right - 2, r.bottom - 2);
				::LineTo(r.right - 2, r.top + 2);
				::Move(-1, -1);
				::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
				::Line(-7, 0);
				::Line(-5, 5);
				::Line(0, 1 + propExtra);
				::Move(0, - (1 + propExtra));
				::RGBForeColor(&gAGARamp[colorIndexes[Corner]]);
				::Line(0, 0);
				::Move(5, -5);
				::Line(0, 0);
				}
			}
		
		if (drawEnabled)
			{
			::MoveTo(r.left + 3 + (moveDirection.h * (((r.right - r.left) / 2) - 6)),
					r.top + 3 + (moveDirection.v * (((r.bottom - r.top) / 2) - 6)));

			if (RuntimeJustify(mJustification) == teFlushRight)
				::Move(2 * drawDirection.h, 2 * drawDirection.v);

			if (! deep)
				::Move(moveDirection.h, moveDirection.v);

			for (SInt32 i = 0; i < 3; i++)
				{
				::RGBForeColor(&gAGARamp[colorIndexes[Light]]);
				::Line(drawDirection.h, drawDirection.v);
				
				::RGBForeColor(&gAGARamp[colorIndexes[Grip]]);
				::Line(drawDirection.h * (lineLength-1), drawDirection.v * (lineLength-1));
				
				::Move(moveDirection.h, moveDirection.v);
				::Move(drawDirection.h, drawDirection.v);
				::RGBForeColor(&gAGARamp[colorIndexes[Dark]]);
				::Line(- (drawDirection.h * lineLength), - (drawDirection.v * lineLength));

				::Move(moveDirection.h, moveDirection.v);
				::Move(- drawDirection.h, - drawDirection.v);
				}
			}
		
		if (isGhost && deep)
			{
			// Draw the little see-thru track slides.
			// They run inverse draw direction from the grip lines.
			// To handle proportional indicators, we need to vary
			// the length of the track slides on the flat part.
			SInt32	wholeFlatLength;
			SInt32	firstFlatLength;
			SInt32	remainingFlatLength;
			
			if (mIsHorizontal)
				wholeFlatLength = r.right - r.left - 10;
			else
				wholeFlatLength = r.bottom - r.top - 10;

			firstFlatLength = wholeFlatLength / 2;
			remainingFlatLength = wholeFlatLength - firstFlatLength; // don't round down twice

			::RGBForeColor(&gAGARamp[r11]);
			::MoveTo(r.left + (drawDirection.h * 4), r.top + (drawDirection.v * 4));
			
			for (UInt32 i = 0; i < 2; i++)
				{
				if (RuntimeJustify(mJustification) == teFlushRight)
					::Move(3 * drawDirection.h, 3 * drawDirection.v);

				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r5]);
				::Line(moveDirection.h * firstFlatLength, moveDirection.v * firstFlatLength);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r9]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r9]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r2]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r9]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r5]);
				::Line(moveDirection.h * remainingFlatLength, moveDirection.v * remainingFlatLength);
				::RGBForeColor(&gAGARamp[r7]);
				::Line(moveDirection.h, moveDirection.v);
				::RGBForeColor(&gAGARamp[r11]);
				::Line(0, 0);
				
				::MoveTo(r.left + (drawDirection.h * 8), r.top + (drawDirection.v * 8));
				}
			}
		}
	}

void AGASlider::BuildPointerIndicator(Rect* indicatorRect)
	{
	// Perform the QuickDraw framing commands to build the
	// pointy indicator. This function is called while wrapped
	// in polygon or region building calls. The caller will
	// use the resulting polygon or region to frame, fill,
	// or clip.

	Rect	r = *indicatorRect;
	SInt16	propExtra;

	if (mIsHorizontal)
		{
		propExtra = MaxSInt16(0, r.right - r.left - 15);

		if (RuntimeJustify(mJustification) == teFlushLeft)
			{
			::MoveTo(r.right - 1, r.top + 1);
			::LineTo(r.right - 1, r.bottom - 7);
			::Line(-6, 6);
			::Line(-2 - propExtra, 0);
			::Line(-6, -6);
			::LineTo(r.left, r.top + 1);
			::Line(1, -1);
			::LineTo(r.right - 2, r.top);
			::Line(1, 1);
			}
		else
			{
			::MoveTo(r.left, r.bottom - 2);
			::LineTo(r.left, r.top + 6);
			::Line(6, -6);
			::Line(2 + propExtra, 0);
			::Line(6, 6);
			::LineTo(r.right - 1, r.bottom - 2);
			::Line(-1, 1);
			::LineTo(r.left + 1, r.bottom - 1);
			::Line(-1, -1);
			}
		}
	else
		{
		propExtra = MaxSInt16(0, r.bottom - r.top - 15);

		if (RuntimeJustify(mJustification) == teFlushLeft)
			{
			::MoveTo(r.left + 1, r.top);
			::LineTo(r.right - 7, r.top);
			::Line(6, 6);
			::Line(0, 2 + propExtra);
			::Line(-6, 6);
			::LineTo(r.left + 1, r.bottom - 1);
			::Line(-1, -1);
			::LineTo(r.left, r.top + 1);
			::Line(1, -1);
			}
		else
			{
			::MoveTo(r.right - 2, r.bottom - 1);
			::LineTo(r.left + 6, r.bottom - 1);
			::Line(-6, -6);
			::Line(0, -2 - propExtra);
			::Line(6, -6);
			::LineTo(r.right - 2, r.top);
			::Line(1, 1);
			::LineTo(r.right - 1, r.bottom - 2);
			::Line(-1, 1);
			}
		}
	}

void AGASlider::RemoveIndicator(SInt32 newValue)
	{
	// Draw the background to remove the indicator
	// at the specified value.

	//
	// Set clipping so we only draw the track to
	// wipe out the old indicator, but not where
	// the new indicator will get drawn anyway.
	// This reduces flicker. Polygon geometry makes
	// tweaking necessary and still imperfect.
	//

	::GetClip(mgSavedIndicatorClip);

	if (newValue != mValue)
		{
		Rect	oldIndicatorBox;
		Rect	newIndicatorBox;

		this->GetIndicatorBox(mValue, &oldIndicatorBox);
		this->GetIndicatorBox(newValue, &newIndicatorBox);
		
		if (mSliderKind == kRectSlider)
			{
			::RectRgn(mgOldIndicatorClip, &oldIndicatorBox);
			::RectRgn(mgNewIndicatorClip, &newIndicatorBox);
			}
		else
			{
			//
			// For optimal clipping, because of polygon and
			// region geometry, we have to tweak the regions
			// a little bit.
			//
			
			RgnHandle	tempRegion = ::NewRgn();

			::OpenRgn();
			this->BuildPointerIndicator(&oldIndicatorBox);
			::CloseRgn(mgOldIndicatorClip);
			::CopyRgn(mgOldIndicatorClip, tempRegion);
			::OffsetRgn(tempRegion, 1, 1);
			::InsetRgn(tempRegion, -1, -1);
			::UnionRgn(mgOldIndicatorClip, tempRegion, mgOldIndicatorClip);

			::OpenRgn();
			this->BuildPointerIndicator(&newIndicatorBox);
			::CloseRgn(mgNewIndicatorClip);
			::CopyRgn(mgNewIndicatorClip, tempRegion);
			::UnionRgn(mgNewIndicatorClip, tempRegion, mgNewIndicatorClip);
			
			::DisposeRgn(tempRegion);
			}

		::DiffRgn(mgOldIndicatorClip, mgNewIndicatorClip, mgOldIndicatorClip);

		::SetClip(mgOldIndicatorClip);
		}

	this->DrawIndicatorTrack();

	::SetClip(mgSavedIndicatorClip);
	}

void AGASlider::RemoveGhost(SInt32 /*newValue*/)
	{
	// Draw the background necessary to remove the
	// ghost indicator while not removing the solid
	// value indicator.

	Rect	r;
	RgnHandle	savedClip = ::NewRgn();
	
	::GetClip(savedClip);

	this->GetIndicatorBox(mGhostValue, &r);
	::ClipRect(&r);
	this->DrawIndicatorTrack();
	this->DrawIndicator();
	
	::SetClip(savedClip);
	::DisposeRgn(savedClip);
	}

void AGASlider::GetTrackFrame(Rect* trackFrame, Boolean includeBackground)
	{
	// Return the rectangle that bounds just the indicator
	// track, including shading pixels but not the area covered
	// by the indicator itself.

	Boolean	leftJustify = (RuntimeJustify(mJustification) == teFlushLeft);

	*trackFrame = mBounds;
	
	if (mIsHorizontal)
		{
		if (leftJustify)
			{
			trackFrame->top += 4;
			trackFrame->bottom = trackFrame->top + 7;
			}
		else
			{
			trackFrame->bottom -= 4;
			trackFrame->top = trackFrame->bottom - 7;
			}
		
		if (includeBackground)
			{
			trackFrame->top -= leftJustify ? 4 : 6;
			trackFrame->bottom += leftJustify ? 6 : 4;
			}
		}
	else
		{
		if (leftJustify)
			{
			trackFrame->left += 4;
			trackFrame->right = trackFrame->left + 7;
			}
		else
			{
			trackFrame->right -= 4;
			trackFrame->left = trackFrame->right - 7;
			}
		
		if (includeBackground)
			{
			trackFrame->left -= leftJustify ? 4 : 6;
			trackFrame->right += leftJustify ? 6 : 4;
			}
		}
	}

void AGASlider::GetIndicatorBox(SInt32 indicatorValue, Rect* indicatorBox)
	{
	// Return the box that defines the indicator's location
	// given the specified indicator value.

	Rect	r = mBounds;
	SInt32	startPt;
	SInt32	valuePt;
	SInt32	stopPt;

	this->GetIndicatorSpan(indicatorValue, &startPt, &valuePt, &stopPt);
	
	if (mIsHorizontal)
		{
		r.left = startPt;
		r.right = stopPt;

		if (RuntimeJustify(mJustification) == teFlushLeft)
			{
			if (mSliderKind == kPointerSlider)
				r.top++;

			r.bottom = r.top + kIndicatorWidth;
			}
		else
			{
			if (mSliderKind == kPointerSlider)
				r.bottom--;

			r.top = r.bottom - kIndicatorWidth;
			}
		}
	else
		{
		r.top = startPt;
		r.bottom = stopPt;

		if (RuntimeJustify(mJustification) == teFlushLeft)
			{
			if (mSliderKind == kPointerSlider)
				r.left++;

			r.right = r.left + kIndicatorWidth;
			}
		else
			{
			if (mSliderKind == kPointerSlider)
				r.right--;

			r.left = r.right - kIndicatorWidth;
			}
		}
	
	*indicatorBox = r;
	}

SInt32 AGASlider::GetIndicatorPixelRange(SInt32* proportionalIndicatorPixels)
	{
	// Return the pixel range that the center point of the indicator
	// can travel along the track, and also return the amount of
	// range reduction that is caused by proportional indicator growth.
	// Non-proportional indicators use the full track range, minus
	// the size of the indicator itself.

	SInt32	range;
	SInt32	indicatorInset = (mSliderKind == kRectSlider) ? kRectIndicatorPixelInset : kPointerIndicatorPixelInset;
	
	*proportionalIndicatorPixels = 0;

	if (mIsHorizontal)
		range = mBounds.right - mBounds.left - (2 * (kEndGapSize + indicatorInset)) + 1;
	else
		range = mBounds.bottom - mBounds.top - (2 * (kEndGapSize + indicatorInset)) + 1;

	if (mIsProportional)
		{
		SInt32	pagePixels;
		
		pagePixels =
			(range * (((float) mPageSize) / ((float) (mMaximum - mMinimum))));
		
		if (pagePixels > 0)
			{
			range -= pagePixels;
			*proportionalIndicatorPixels = pagePixels;
			}
		}
	
	return range;
	}

void AGASlider::GetIndicatorPixelEnds(SInt32* startPt, SInt32* stopPt)
	{
	// Return the start and end pixel locations of the indicator
	// track traveled by the center point of the indicator.

	SInt32	dontCare;
	SInt32	range = this->GetIndicatorPixelRange(&dontCare);

	if (mIsHorizontal)
		*startPt = ((mBounds.left + mBounds.right) / 2) - (range / 2) - 1;
	else
		*startPt = ((mBounds.top + mBounds.bottom) / 2) - (range / 2) - 1;

	*stopPt = *startPt + range;	// avoid rounding of 1/2 pixel twice
	}

void AGASlider::GetIndicatorSpan(SInt32 value, SInt32* startPt, SInt32* valuePt, SInt32* stopPt)
	{
	// Return the pixel locations of the indicator's ends and value.

	SInt32	proportionalIndicatorPixels;
	SInt32	trackStart;
	SInt32	trackStop;
	SInt32	trackRange = this->GetIndicatorPixelRange(&proportionalIndicatorPixels);
	SInt32	indicatorInset = (mSliderKind == kRectSlider) ? kRectIndicatorPixelInset : kPointerIndicatorPixelInset;
	
	this->GetIndicatorPixelEnds(&trackStart, &trackStop);
	
	*valuePt = trackStart +
		(trackRange * (((float) (value - mMinimum)) / ((float) (mMaximum - mMinimum))));

	*startPt = *valuePt - indicatorInset - (proportionalIndicatorPixels / 2);
	*stopPt = *startPt + (2 * indicatorInset) + proportionalIndicatorPixels + 1;
	}

void AGASlider::DisposeLabels()
	{
	// Release the memory allocated for slider labels.

	if (mLabelStringHandles != NULL)
		{
		UInt32	numHandles = ::GetHandleSize(mLabelStringHandles) / sizeof(StringHandle);
		
		for (UInt32 i = 0; i < numHandles; i++)
			::DisposeHandle((Handle) ((StringHandle*)(*mLabelStringHandles))[i]);

		::DisposeHandle(mLabelStringHandles);
		}
	}

//
// AGALittleArrows ---------------------------------------------------------------
//
// This class implements a little arrows control.
//

AGALittleArrows::AGALittleArrows(Rect* bounds)
: AGAObject(bounds)
	{
	mNotificationRoutine = NULL;
	mUserData = 0;
	}

AGALittleArrows::~AGALittleArrows()
	{
	}

void AGALittleArrows::InstallNotificationRoutine(AGALittleArrowsNotifyPtr notificationRoutine, void* userData)
	{
	// Install the supplied function pointer and user data to
	// be called during notification.

	mNotificationRoutine = notificationRoutine;
	mUserData = userData;
	}

void AGALittleArrows::DrawObject()
	{
	// Draw the arrows in normal unpressed state.

	this->DrawButton(kNoArrowPressed);
	}

Boolean AGALittleArrows::TrackMouse(Point mouseLocation)
	{
	if (! mEnabled)
		return false;

	// Track the mouse in the appropriate arrow. Little arrows
	// always return true because the tracking is always successful.

	return this->TrackPart(mouseLocation, this->TrackTestPart(mouseLocation));
	}

void AGALittleArrows::DrawButton(SInt32 pressedPart)
	{
	AGADrawingEnvironment	env;

	// Draw the arrows in the specified press state. The
	// state may indicate that either or no arrow is pressed.

	enum { Frame, Fill, TL, BR, Arrow, kNumArrowColors };

	UInt8	colorIndexes[kNumArrowColors];
	
	GDIterator	iter;
	Boolean		deep;
	Boolean		drawEnabled = mEnabled && mActive;

	while (iter.NextDevice(deep))
		{
		Rect	r = mBounds;
		
		r.right = r.left + kLittleArrowWidth;
		r.bottom = r.top + kLittleArrowHeight;
		
		for (UInt32 i = 0; i < 2; i++)
			{
			if (deep)
				{
				if (drawEnabled)
					{
					if (((pressedPart == kUpArrowPressed) && (i == 0)) ||
						((pressedPart == kDownArrowPressed) && (i == 1)))
						{
						colorIndexes[Frame] = rB;
						colorIndexes[Fill] = r8;
						colorIndexes[TL] = r10;
						colorIndexes[BR] = r6;
						colorIndexes[Arrow] = rW;
						}
					else if (((pressedPart != kUpArrowPressed) && (i == 0)) ||
						((pressedPart != kDownArrowPressed) && (i == 1)))
						{
						colorIndexes[Frame] = rB;
						colorIndexes[Fill] = r2;
						colorIndexes[TL] = rW;
						colorIndexes[BR] = r5;
						colorIndexes[Arrow] = rB;
						}
					}
				else	// disabled
					{
					colorIndexes[Frame] = r7;
					colorIndexes[Fill] = r2;
					colorIndexes[TL] = r2;
					colorIndexes[BR] = r2;
					colorIndexes[Arrow] = r7;
					}
				
				// These (1-i) and i uses here are to round off the top corners
				// of the top frame (i=0) and round off the bottom corners of
				// the bottom frame (i=1).
				::RGBForeColor(&gAGARamp[colorIndexes[Frame]]);
				::MoveTo(r.left + (1-i), r.top);
				::LineTo(r.right - 1 - (1-i), r.top);
				::Move(1-i, 1-i);
				::LineTo(r.right -1, r.bottom - 1 - i);
				::Move(-i, i);
				::LineTo(r.left + i, r.bottom - 1);
				::Move(-i, -i);
				::LineTo(r.left, r.top + (1-i));
				
				::InsetRect(&r, 1, 1);
				::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
				::PaintRect(&r);
				::InsetRect(&r, -1, -1);
				
				::RGBForeColor(&gAGARamp[colorIndexes[TL]]);
				::MoveTo(r.left + 1, r.bottom - 3);
				::LineTo(r.left + 1, r.top + 1);
				::LineTo(r.right - 3, r.top + 1);
				
				::RGBForeColor(&gAGARamp[colorIndexes[BR]]);
				::MoveTo(r.left + 2, r.bottom - 2);
				::LineTo(r.right - 2, r.bottom - 2);
				::LineTo(r.right - 2, r.top + 2);
				
				::RGBForeColor(&gAGARamp[colorIndexes[Arrow]]);
				}
			else	// 1-bit
				{
				if (((pressedPart == kUpArrowPressed) && (i == 0)) ||
					((pressedPart == kDownArrowPressed) && (i == 1)))
					{
					::RGBForeColor(&gAGARamp[rB]);
					::InsetRect(&r, 1, 1);
					::PaintRect(&r);
					::InsetRect(&r, -1, -1);
					::RGBForeColor(&gAGARamp[rW]);
					}
				else
					{
					::RGBForeColor(&gAGARamp[rW]);
					::PaintRect(&r);
					::RGBForeColor(&gAGARamp[rB]);

					::MoveTo(r.left + (1-i), r.top);
					::LineTo(r.right - 1 - (1-i), r.top);
					::Move(1-i, 1-i);
					::LineTo(r.right -1, r.bottom - 1 - i);
					::Move(-i, i);
					::LineTo(r.left + i, r.bottom - 1);
					::Move(-i, -i);
					::LineTo(r.left, r.top + (1-i));
					}
				}

			::MoveTo(r.left + 3, r.top + 7 - (i*3));
			::Line(6, 0);
			::Move(-1, -1 + (2*i));
			::Line(-4, 0);
			::Move(1, -1 + (2*i));
			::Line(2, 0);
			::Move(-1, -1 + (2*i));
			::Line(0, 0);
			
			::OffsetRect(&r, 0, kLittleArrowHeight - 1);
			}

		if ((! drawEnabled) && ! deep)
			{
			::RGBForeColor(&gAGARamp[rB]);
			PenPatGray();
			::PenMode(patBic);
			::PaintRect(&mBounds);

			::PenNormal();
			}
		}
	}

Boolean AGALittleArrows::TrackPart(Point /*mouseLocation*/, SInt32 partHit)
	{
	// Track the mouse in the specified arrow. Little arrows always
	// return true because the tracking is always successful.

	Boolean	wasIn = false;
	Boolean	isIn = true;
	Boolean	isButtonDown = true;	// ensure at least one time through
	Point	newMouseLocation;
	UInt32	accelerationDelay = 18;	// will divide by 2 each time
	
	while (isButtonDown)
		{
		if (isIn)
			this->NotifyDelta(partHit);

		::GetMouse(&newMouseLocation);	// gives local coordinates
		
		isIn = (this->TrackTestPart(newMouseLocation) == partHit);
		
		if (isIn != wasIn)
			this->DrawButton(isIn ? partHit : kNoArrowPressed);
		
		wasIn = isIn;
		
		if (isButtonDown && (accelerationDelay != 0))
			{
			DelayFutureWhileStillDown(accelerationDelay);
			accelerationDelay = accelerationDelay >> 1;	// divide by 2
			}

		isButtonDown = ::StillDown();
		}
	
	if (isIn)	// if finished IN, restore normal appearance
		this->DrawButton(kNoArrowPressed);
	
	return true;
	}

SInt32 AGALittleArrows::TrackTestPart(Point mouseLocation)
	{
	// Return the part value where the mouse would hit.

	if ((mouseLocation.h >= mBounds.left + kLittleArrowWidth) ||
		(mouseLocation.h < mBounds.left) ||
		(mouseLocation.v >= mBounds.top + kLittleArrowTotalHeight) ||
		(mouseLocation.v < mBounds.top))
		return kNoArrowPressed;
	else if (mouseLocation.v < mBounds.top + kLittleArrowHeight)
		return kUpArrowPressed;
	else
		return kDownArrowPressed;
	}

void AGALittleArrows::NotifyDelta(SInt32 partHit)
	{
	// Call the installed notification routine, if any, with
	// the installed user data and our part hit, which is the
	// value delta (-1 or 1).

	if (mNotificationRoutine != NULL)
		(*(mNotificationRoutine))(this, partHit, mUserData);
	}

//
// AGADisclosureTriangle ---------------------------------------------------------------
//
// This class implements a disclosure triangle.
//

AGADisclosureTriangle::AGADisclosureTriangle(Rect* bounds, Boolean automaticState)
: AGAObject(bounds)
	{
	// Construct as specified.

	mAutomaticState = automaticState;
	mIsDisclosed = kClosedState;
	}

AGADisclosureTriangle::~AGADisclosureTriangle()
	{
	}

void AGADisclosureTriangle::DrawObject()
	{
	// Draw the triangle in the current state.

	this->DrawTriangle(mIsDisclosed ? kDTDisclosed : kDTClosed);
	}

Boolean AGADisclosureTriangle::TrackMouse(Point mouseLocation)
	{
	// Let superclass handle tracking, and change our state
	// if we have automatic state change turned on.

	Boolean	wasItHit = AGAObject::TrackMouse(mouseLocation);

	if (wasItHit && mAutomaticState)
		this->SetStateAnimate(! mIsDisclosed);

	return wasItHit;
	}

Boolean AGADisclosureTriangle::GetState()
	{
	// Return the current state.

	return mIsDisclosed;
	}

void AGADisclosureTriangle::SetState(Boolean isDisclosed, Boolean redraw)
	{
	// Set the state and optionally redraw.

	if (isDisclosed != mIsDisclosed)
		{
		mIsDisclosed = isDisclosed;
		
		if (redraw)
			this->DrawTriangle(mIsDisclosed ? kDTDisclosed : kDTClosed);
		}
	}

void AGADisclosureTriangle::SetStateAnimate(Boolean isDisclosed)
	{
	// Set the state, but animate the triangle as it changes
	// to the new state.

	const SInt32 kAnimationDelay = 4;

	if (isDisclosed)
		{
		this->DrawTriangle(kDTPressedClosed);
		DelayFuture(kAnimationDelay);

		this->EraseBackground();
		this->DrawTriangle(kDTIntermediate);
		DelayFuture(kAnimationDelay);

		this->EraseBackground();
		this->DrawTriangle(kDTPressedDisclosed);
		DelayFuture(kAnimationDelay);

		this->EraseBackground();
		}
	else
		{
		this->DrawTriangle(kDTPressedDisclosed);
		DelayFuture(kAnimationDelay);

		this->EraseBackground();
		this->DrawTriangle(kDTIntermediate);
		DelayFuture(kAnimationDelay);

		this->EraseBackground();
		this->DrawTriangle(kDTPressedClosed);
		DelayFuture(kAnimationDelay);

		this->EraseBackground();
		}
	
	this->SetState(isDisclosed, kRedraw);
	}

void AGADisclosureTriangle::SetTrackingState(Boolean isIn)
	{
	// Draw the triangle pressed or unpressed as specified.

	DTAnimationState	state;
	
	if (isIn)
		state = mIsDisclosed ? kDTPressedDisclosed : kDTPressedClosed;
	else
		state = mIsDisclosed ? kDTDisclosed : kDTClosed;

	this->DrawTriangle(state);
	}

void AGADisclosureTriangle::DrawTriangle(DTAnimationState state)
	{
	AGADrawingEnvironment	env;

	// Draw the triangle at the specified animation frame.

	Point		fulcrum;
	Point		corners[3];
	Boolean		drawEnabled = mEnabled && mActive;
	
	::SetPt(&fulcrum, mBounds.left + 5, mBounds.top + 5);

	switch (state)
		{
		case kDTClosed:
		case kDTPressedClosed:
			::SetPt(&corners[0], fulcrum.h - 2, fulcrum.v - 5);
			::SetPt(&corners[1], fulcrum.h + 3, fulcrum.v);
			::SetPt(&corners[2], fulcrum.h - 2, fulcrum.v + 5);
			break;
		case kDTIntermediate:
			::SetPt(&corners[0], fulcrum.h + 3, fulcrum.v - 5);
			::SetPt(&corners[1], fulcrum.h + 3, fulcrum.v + 3);
			::SetPt(&corners[2], fulcrum.h - 5, fulcrum.v + 3);
			break;
		case kDTPressedDisclosed:
		case kDTDisclosed:
			::SetPt(&corners[0], fulcrum.h + 5, fulcrum.v - 2);
			::SetPt(&corners[1], fulcrum.h, fulcrum.v + 3);
			::SetPt(&corners[2], fulcrum.h - 5, fulcrum.v - 2);
			break;
		}
	
	// Little trick: tweak state so that disabled state (gray
	// fill) is handled by pressed case code below. We already set
	// the color so they only set pressed fill color if enabled.
	DTAnimationState	deepState = state;
	
	if (! drawEnabled)
		{
		if (deepState == kDTClosed)
			deepState = kDTPressedClosed;
		else
			deepState = kDTPressedDisclosed;
		}
			
	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			if (drawEnabled)
				::RGBForeColor(&gAGARamp[rB]);
			else
				::RGBForeColor(&gAGARamp[r5]);

			::MoveTo(corners[0].h, corners[0].v);
			::LineTo(corners[1].h, corners[1].v);
			::LineTo(corners[2].h, corners[2].v);
			::LineTo(corners[0].h, corners[0].v);
			
			switch (deepState)
				{
				case kDTClosed:
					::RGBForeColor(&gAGARamp[rP1]);
					::MoveTo(corners[0].h + 1, corners[0].v + 2);
					::Line(0, 6);
					::RGBForeColor(&gAGARamp[rP2]);
					::Line(1, -1);
					::Line(0, -4);
					::Line(1, 1);
					::Line(0, 1);
					::Move(1, 0);
					::RGBForeColor(&gAGARamp[rP3]);
					::Line(-2, 2);

					::RGBForeColor(&gAGARamp[r7]);
					::MoveTo(corners[2].h + 1, corners[2].v);
					::Line(4, -4);
					::RGBForeColor(&gAGARamp[r4]);
					::Move(1, 0);
					::Line(-5, 5);
					break;
				case kDTPressedClosed:
					if (drawEnabled)
						::RGBForeColor(&gAGARamp[rP4]);

					::MoveTo(corners[0].h + 1, corners[0].v + 2);
					::Line(0, 6);
					::Move(1, -1);
					::Line(0, -4);
					::Move(1, 1);
					::Line(0, 2);
					::Line(1, -1);

					::RGBForeColor(&gAGARamp[r2]);
					::MoveTo(corners[2].h + 1, corners[2].v);
					::Line(4, -4);
					::Move(1, 0);
					::Line(-5, 5);
					break;
				case kDTIntermediate:
					::RGBForeColor(&gAGARamp[rP4]);
					::MoveTo(corners[0].h - 1, corners[0].v + 2);
					::Line(-5, 5);
					::Move(1, 0);
					::Line(4, -4);
					::Move(0, 1);
					::Line(-3, 3);
					::Move(1, 0);
					::Line(2, -2);
					::Move(0, 1);
					::Line(-1, 1);
					::Line(1, 0);
					break;
				case kDTPressedDisclosed:
					if (drawEnabled)
						::RGBForeColor(&gAGARamp[rP4]);

					::MoveTo(corners[2].h + 2, corners[2].v + 1);
					::Line(6, 0);
					::Move(-1, 1);
					::Line(-4, 0);
					::Move(1, 1);
					::Line(2, 0);
					::Line(-1, 1);

					::RGBForeColor(&gAGARamp[r2]);
					::MoveTo(corners[0].h, corners[0].v + 1);
					::Line(-4, 4);
					::Move(0, 1);
					::Line(5, -5);
					break;
				case kDTDisclosed:
					::RGBForeColor(&gAGARamp[rP1]);
					::MoveTo(corners[2].h + 2, corners[2].v + 1);
					::Line(6, 0);
					::RGBForeColor(&gAGARamp[rP2]);
					::Line(-1, 1);
					::Line(-4, 0);
					::Line(1, 1);
					::Line(1, 0);
					::Move(0, 1);
					::RGBForeColor(&gAGARamp[rP3]);
					::Line(2, -2);

					::RGBForeColor(&gAGARamp[r7]);
					::MoveTo(corners[0].h, corners[0].v + 1);
					::Line(-4, 4);
					::RGBForeColor(&gAGARamp[r4]);
					::Move(0, 1);
					::Line(5, -5);
					break;
				}
			}
		else	// 1-bit
			{
			PolyHandle	theTriangle = ::OpenPoly();

			::MoveTo(corners[0].h, corners[0].v);
			::LineTo(corners[1].h, corners[1].v);
			::LineTo(corners[2].h, corners[2].v);
			::LineTo(corners[0].h, corners[0].v);
			
			::ClosePoly();
			
			switch (state)
				{
				case kDTClosed:
				case kDTDisclosed:
					::RGBForeColor(&gAGARamp[rW]);
					::PaintPoly(theTriangle);
					::RGBForeColor(&gAGARamp[rB]);
					::FramePoly(theTriangle);

					if (! drawEnabled)
						{
						PenPatGray();
						::PenMode(patBic);
						::PaintRect(&mBounds);

						::PenNormal();
						}
					break;
				case kDTPressedClosed:
				case kDTIntermediate:
				case kDTPressedDisclosed:
					::RGBForeColor(&gAGARamp[rB]);
					::PaintPoly(theTriangle);
					break;
				}
			
			::KillPoly(theTriangle);
			::PenNormal();
			}
		}
	}

void AGADisclosureTriangle::EraseBackground()
	{
	AGADrawingEnvironment	env;

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			this->ApplyBackgroundColor();
		else
			::RGBForeColor(&gAGARamp[rW]);

		::PaintRect(&mBounds);
		}
	}

//
// AGAProgressIndicator ---------------------------------------------------------------
//
// This class implements a progress indicator that can be either
// determinate or indeterminate, and also can have a moving start
// value.
//

Boolean AGAProgressIndicator::mImagesAllocated = false;

// We use an offscreen image for the indeterminate progress pattern.
AGAOffscreenImage*	AGAProgressIndicator::mgOffscreenIndeterminateProgressImage[kNumProgressPatterns];

// These are the color ramp indexes that are used to build the offscreen image.
SInt8				AGAProgressIndicator::kIndeterminateProgressImage[kNumProgressPatterns]
																[kProgressImageHeight]
																	[kProgressImageWidth] =
		{
		{	// k1BitPattern
			{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
			{  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
			{  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
			{  rW,  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW,  rW,  rW,  rW,  rW },
			{  rW,  rW,  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW,  rW,  rW,  rW },
			{  rW,  rW,  rW,  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW,  rW,  rW },
			{  rW,  rW,  rW,  rW,  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW,  rW },
			{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rW },
			{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
			{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		},
		{	// kDeepPattern
			{ r11, r11, r11, r11, r11, r11, r11, r11, r10, r10, r10, r10, r10, r10, r10, r10 },
			{  r8, r10, r10, r10, r10, r10, r10, r10, r10,  r8,  r8,  r8,  r8,  r8,  r8,  r8 },
			{  r5,  r5,  r8,  r8,  r8,  r8,  r8,  r8,  r8,  r8,  r5,  r5,  r5,  r5,  r5,  r5 },
			{  r2,  r2,  r2,  r5,  r5,  r5,  r5,  r5,  r5,  r5,  r5,  r2,  r2,  r2,  r2,  r2 },
			{  rW,  rW,  rW,  rW,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  rW,  rW,  rW,  rW },
			{  r2,  r2,  r2,  r2,  r2,  r5,  r5,  r5,  r5,  r5,  r5,  r5,  r5,  r2,  r2,  r2 },
			{  r4,  r4,  r4,  r4,  r4,  r4,  r8,  r8,  r8,  r8,  r8,  r8,  r8,  r8,  r4,  r4 },
			{  r6,  r6,  r6,  r6,  r6,  r6,  r6, r10, r10, r10, r10, r10, r10, r10, r10,  r6 },
			{  r8,  r8,  r8,  r8,  r8,  r8,  r8,  r8, r11, r11, r11, r11, r11, r11, r11, r11 },
			{ r12, r10, r10, r10, r10, r10, r10, r10, r10, r12, r12, r12, r12, r12, r12, r12 },
		},
	};

OSErr AGAProgressIndicator::AllocateProgressImages()
	{
	// Allocate the offscreen image for each radio button state.

	OSErr	result = noErr;
	Point	imageSize;
	
	imageSize.h = kProgressImageWidth;
	imageSize.v = kProgressImageHeight;
	
	for (UInt32 imageIndex = 0; imageIndex < kNumProgressPatterns; imageIndex++)
		{
		if (result == noErr)
			{
			mgOffscreenIndeterminateProgressImage[imageIndex] = new AGAOffscreenImage;
			
			result = mgOffscreenIndeterminateProgressImage[imageIndex]->CreateImageData(
						imageSize,
						&kIndeterminateProgressImage[imageIndex][0][0],
						NULL, 
						&gAGARamp[rB]);
			}
		}
		
	mImagesAllocated = true;

	return result;
	}

void AGAProgressIndicator::DisposeProgressImages()
	{
	for (UInt32 imageIndex = 0; imageIndex < kNumProgressPatterns; imageIndex++)
		if (mgOffscreenIndeterminateProgressImage[imageIndex] != NULL)
			delete mgOffscreenIndeterminateProgressImage[imageIndex];
	}

AGAProgressIndicator::AGAProgressIndicator(Rect* bounds, SInt32 minimum, SInt32 maximum)
: AGAObject(bounds)
	{
	// Construct with default options.
	// (minimum == maximum) means indeterminate indicator.

	mMinimum = minimum;
	mMaximum = maximum;
	mValue = minimum;
	mOriginValue = minimum;
	mAnimationIndex = 0;
	mLastAnimationTime = 0;
	
	if (! mImagesAllocated)
		(void) AllocateProgressImages();
	}

AGAProgressIndicator::~AGAProgressIndicator()
	{
	}

void AGAProgressIndicator::DrawObject()
	{
	// Draw at current gauge value or next animation step.

	this->DrawFrame();
	
	if (mMinimum == mMaximum)
		this->DrawAnimationStep();
	else
		this->DrawGauge();
	}

SInt32 AGAProgressIndicator::GetValue()
	{
	// Return current value.

	return mValue;
	}

void AGAProgressIndicator::SetValue(SInt32 newValue, Boolean redraw)
	{
	// Set gauge value, optionally redraw.

	// Constrain: mMinimum <= mValue <= mMaximum
	newValue = MaxSInt32(mMinimum, MinSInt32(newValue, mMaximum));
	
	if (newValue != mValue)
		{
		mValue = newValue;
		
		if (redraw)
			this->DrawGauge();
		}
	}

void AGAProgressIndicator::Increment(SInt32 delta, Boolean redraw)
	{
	// Increment gauge value, optionally redraw.

	this->SetValue(delta + this->GetValue(), redraw);
	}

void AGAProgressIndicator::Animate()
	{
	// Draw the next animation step.

	mAnimationIndex = (mAnimationIndex + 1) % kNumProgressAnimationSteps;
	
	this->DrawAnimationStep();
	}

void AGAProgressIndicator::QuantizedAnimate()
	{
	// Draw the next animation step if enough time has elapsed since the
	// last animation step.
	
	if (mMinimum == mMaximum)
		{
		SInt32	now = ::TickCount();
		
		if (now > mLastAnimationTime + kAnimationDelay)
			{
			mLastAnimationTime = now;
			this->Animate();
			}
		}
	}

void AGAProgressIndicator::GetRange(SInt32* minimum, SInt32* maximum)
	{
	// Return the gauge value range.

	*minimum = mMinimum;
	*maximum = mMaximum;
	}

void AGAProgressIndicator::SetRange(SInt32 newMinimum, SInt32 newMaximum, Boolean redraw)
	{
	// Set the gauge value range, optionally redraw.
	// (minimum == maximum) means indeterminate indicator.

	if (mOriginValue == mMinimum)
		mOriginValue = newMinimum;

	mMinimum = newMinimum;
	mMaximum = MaxSInt32(newMinimum, newMaximum);
	
	if (redraw)
		this->DrawObject();
	}

SInt32 AGAProgressIndicator::GetOriginValue()
	{
	// Return the optional gauge origin value.

	return mOriginValue;
	}

void AGAProgressIndicator::SetOriginValue(SInt32 newValue, Boolean redraw)
	{
	// Set the optional gauge origin value.

	// Constrain: mMinimum <= mOriginValue <= mValue
	newValue = MaxSInt32(mMinimum, MinSInt32(newValue, mValue));

	if (newValue != mOriginValue)
		{
		mOriginValue = newValue;
		
		if (redraw)
			this->DrawObject();
		}
	}

void AGAProgressIndicator::DrawFrame()
	{
	AGADrawingEnvironment	env;

	// Draw the constant frame around the indicator.
	Rect	r = mBounds;
	
	r.bottom = r.top + kProgressHeight;
	

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			::RGBForeColor(&gAGARamp[r5]);
			::MoveTo(r.left, r.bottom - 2);
			::LineTo(r.left, r.top);
			::LineTo(r.right - 2, r.top);
			
			::RGBForeColor(&gAGARamp[rW]);
			::MoveTo(r.left + 1, r.bottom - 1);
			::LineTo(r.right - 1, r.bottom - 1);
			::LineTo(r.right - 1, r.top + 1);
			}
		
		::InsetRect(&r, 1, 1);
		::RGBForeColor(&gAGARamp[rB]);
		::FrameRect(&r);
		::InsetRect(&r, -1, -1);
		}
	}

void AGAProgressIndicator::DrawAnimationStep()
	{
	AGADrawingEnvironment	env;

	// Draw the indeterminate progress indicator at the
	// current animation step.

	const UInt32 kNumRows = 10;

	Rect	r = mBounds;
	SInt16	patternHOrigin;
	
	r.bottom = r.top + kProgressHeight;
	::InsetRect(&r, 2, 2);
	
	//
	// Move the pen leftward so the pattern is offset for
	// the current animation step. This will mean we need
	// to explicitly clip to our gauge area, since our pattern
	// origin is outside it.
	//

	patternHOrigin = r.left - 32 + ((16 / kNumProgressAnimationSteps) * mAnimationIndex);
	
	RgnHandle	savedClip = ::NewRgn();
	::GetClip(savedClip);
	
	AGAClipFurther(&r);
	
	r.left = patternHOrigin;

	{
	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			mgOffscreenIndeterminateProgressImage[kDeepPattern]->DrawPattern(&r);
		else
			mgOffscreenIndeterminateProgressImage[k1BitPattern]->DrawPattern(&r);
		}
	}	// force GDIterator clip to clean up now, not after we set savedClip!

	::SetClip(savedClip);
	::DisposeRgn(savedClip);
	}

void AGAProgressIndicator::DrawGauge()
	{
	// Draw the determinate progress indicator.

	if (mOriginValue != mMinimum)
		this->DrawOriginArea();
	
	this->DrawProgressArea();
	this->DrawRemainderArea();
	}

void AGAProgressIndicator::DrawOriginArea()
	{
	// Draw the empty track between the minimum value
	// and the origin value.

	Rect	frame;
	
	if (this->GetOriginFrame(&frame))
		{
		GDIterator	iter;
		Boolean		deep;

		while (iter.NextDevice(deep))
			{
			Rect	r = frame;

			if (deep)
				{
				if (gAGAExtraControls)
					{
					SInt16	width = r.right - r.left;
					
					if (width >= 0)
						{
						::RGBForeColor(&gAGARamp[rB]);
						::MoveTo(r.right, r.top);
						::LineTo(r.right, r.bottom - 1);
						}

					if (width > 0)
						{
						::RGBForeColor(&gAGARamp[r10]);
						::MoveTo(r.left, r.top);
						::LineTo(r.left, r.bottom - 1);
						}
					
					if (width > 1)
						{
						::RGBForeColor(&gAGARamp[r4]);
						::MoveTo(r.right - 1, r.top);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r2]);
						::LineTo(r.right - 1, r.bottom - 1);
						}
					
					if (width > 2)
						{
						::RGBForeColor(&gAGARamp[r7]);
						::MoveTo(r.left + 1, r.top);
						::LineTo(r.left + 1, r.bottom - 1);
						}
					
					if (width > 3)
						{
						::RGBForeColor(&gAGARamp[r2]);
						::MoveTo(r.left + 2, r.bottom - 1);
						::LineTo(r.right - 1, r.bottom - 1);

						::RGBForeColor(&gAGARamp[r7]);
						::MoveTo(r.left + 2, r.top);
						::LineTo(r.right - 2, r.top);

						::RGBForeColor(&gAGARamp[r4]);
						r.left += 2;
						r.top++;
						r.bottom--;
						r.right--;
						::PaintRect(&r);
						}
					}
				else	// mimic System 7 Finder progress
					{
					::RGBForeColor(&gFinder7EmptyProgressColor);
					::PaintRect(&r);
					}
				}
			else	// 1-bit
				{
				::RGBForeColor(&gAGARamp[rW]);
				::PaintRect(&r);
				}
			}
		}
	}

void AGAProgressIndicator::DrawProgressArea()
	{
	AGADrawingEnvironment	env;

	// Draw the filled progress gauge area between
	// the origin and the current value.

	Rect	frame;
	
	if (this->GetProgressFrame(&frame))
		{
		GDIterator	iter;
		Boolean		deep;

		while (iter.NextDevice(deep))
			{
			Rect	r = frame;

			if (deep)
				{
				if (gAGAExtraControls)
					{
					SInt16	width = r.right - r.left;

					//
					// Because the endcaps of the gauge are shaded,
					// we need to carefully draw the ends horizontal
					// pixel by pixel, bailing out if there is not
					// enough width to accomodate the endcaps and the
					// midsection.
					//

					if (width > 0)
						{
						::RGBForeColor(&gAGARamp[rB]);
						::MoveTo(r.right - 1, r.top);
						::LineTo(r.right - 1, r.bottom - 1);
						}

					if (width > 1)
						{
						::RGBForeColor(&gAGARamp[r8]);
						::MoveTo(r.left, r.top);
						::LineTo(r.left, r.bottom - 1);
						}

					if (width > 2)
						{
						::RGBForeColor(&gAGARamp[r10]);
						::MoveTo(r.right - 2, r.top);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r12]);
						::LineTo(r.right - 2, r.bottom - 1);
						}

					if (width > 3)
						{
						::RGBForeColor(&gAGARamp[r8]);
						::MoveTo(r.left + 1, r.top);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r5]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r3]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r1]);
						::Line(0, 3);
						::RGBForeColor(&gAGARamp[r3]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r5]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r8]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r10]);
						::Line(0, 0);
						}

					if (width > 4)
						{
						::RGBForeColor(&gAGARamp[r10]);
						::MoveTo(r.right - 3, r.top);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r8]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r10]);
						::LineTo(r.right - 3, r.bottom - 1);
						::RGBForeColor(&gAGARamp[r12]);
						::Line(0, 0);
						}

					if (width > 5)
						{
						::RGBForeColor(&gAGARamp[r10]);
						::MoveTo(r.right - 4, r.top);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r8]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r5]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r3]);
						::Line(0, 3);
						::RGBForeColor(&gAGARamp[r5]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r8]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r10]);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r12]);
						::Line(0, 0);
						}

					if (width > 6)
						{
						SInt16	distance = width - 7;
						
						::RGBForeColor(&gAGARamp[r10]);
						::MoveTo(r.left + 2, r.top);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r8]);
						::MoveTo(r.left + 2, r.top + 1);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r5]);
						::MoveTo(r.left + 2, r.top + 2);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r3]);
						::MoveTo(r.left + 2, r.top + 3);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r1]);
						::MoveTo(r.left + 2, r.top + 4);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r3]);
						::MoveTo(r.left + 2, r.top + 5);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r5]);
						::MoveTo(r.left + 2, r.top + 6);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r8]);
						::MoveTo(r.left + 2, r.top + 7);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r10]);
						::MoveTo(r.left + 2, r.top + 8);
						::Line(distance, 0);
						::RGBForeColor(&gAGARamp[r12]);
						::MoveTo(r.left + 2, r.top + 9);
						::Line(distance, 0);
						}
					}
				else	// mimic System 7 Finder progress
					{
					::RGBForeColor(&gFinder7FilledProgressColor);
					::PaintRect(&r);
					}
				}
			else	// 1-bit
				{
				::RGBForeColor(&gAGARamp[rB]);
				::PaintRect(&r);
				}
			}
		}
	}

void AGAProgressIndicator::DrawRemainderArea()
	{
	AGADrawingEnvironment	env;

	// Draw the empty track area between the current
	// gauge value and the maximum.

	Rect	frame;
	
	if (this->GetRemainderFrame(&frame))
		{
		GDIterator	iter;
		Boolean		deep;

		while (iter.NextDevice(deep))
			{
			Rect	r = frame;

			if (deep)
				{
				if (gAGAExtraControls)
					{
					SInt16	width = r.right - r.left;
					
					if (width > 0)
						{
						::RGBForeColor(&gAGARamp[r10]);
						::MoveTo(r.left, r.top);
						::LineTo(r.left, r.bottom - 1);
						}
					
					if (width > 1)
						{
						::RGBForeColor(&gAGARamp[r4]);
						::MoveTo(r.right - 1, r.top);
						::Line(0, 1);
						::RGBForeColor(&gAGARamp[r2]);
						::LineTo(r.right - 1, r.bottom - 1);
						}
					
					if (width > 2)
						{
						::RGBForeColor(&gAGARamp[r7]);
						::MoveTo(r.left + 1, r.top);
						::LineTo(r.left + 1, r.bottom - 1);
						}
					
					if (width > 3)
						{
						::RGBForeColor(&gAGARamp[r2]);
						::MoveTo(r.left + 2, r.bottom - 1);
						::LineTo(r.right - 1, r.bottom - 1);

						::RGBForeColor(&gAGARamp[r7]);
						::MoveTo(r.left + 2, r.top);
						::LineTo(r.right - 2, r.top);

						::RGBForeColor(&gAGARamp[r4]);
						r.left += 2;
						r.top++;
						r.bottom--;
						r.right--;
						::PaintRect(&r);
						}
					}
				else	// mimic System 7 Finder progress
					{
					::RGBForeColor(&gFinder7EmptyProgressColor);
					::PaintRect(&r);
					}
				}
			else	// 1-bit
				{
				::RGBForeColor(&gAGARamp[rW]);
				::PaintRect(&r);
				}
			}
		}
	}

Boolean AGAProgressIndicator::GetOriginFrame(Rect* originFrame)
	{
	// Return true (and the area) if there is an area
	// to be drawn for the empty origin part of the track.

	Boolean	isNeeded = false;
	
	if (mOriginValue != mMinimum)
		{
		isNeeded = true;

		this->GetRangeFrame(mMinimum, mOriginValue, originFrame);
		
		if (mOriginValue != mMaximum)
			originFrame->right--;
		}

	return isNeeded;
	}

Boolean AGAProgressIndicator::GetProgressFrame(Rect* progressFrame)
	{
	// Return true (and the area) if there is an area
	// to be drawn for the progress section of the gauge.

	Boolean	isNeeded = false;
	
	if (mValue != mOriginValue)
		{
		isNeeded = true;

		this->GetRangeFrame(mOriginValue, mValue, progressFrame);
		
		if (mValue == mMaximum)
			progressFrame->right++;
		}

	return isNeeded;
	}

Boolean AGAProgressIndicator::GetRemainderFrame(Rect* remainderFrame)
	{
	// Return true (and the area) if there is an area
	// to be drawn for the uncompleted gauge remainder area.

	Boolean	isNeeded = false;
	
	if (mMaximum != mValue)
		{
		isNeeded = true;

		this->GetRangeFrame(mValue, mMaximum, remainderFrame);
		}

	return isNeeded;
	}

void AGAProgressIndicator::GetRangeFrame(SInt32 startValue, SInt32 endValue, Rect* rangeFrame)
	{
	// Return the rectangle that the indicator draws in,
	// which does not include the constant border frame area.

	*rangeFrame = mBounds;
	rangeFrame->bottom = rangeFrame->top + kProgressHeight;
	
	::InsetRect(rangeFrame, 2, 2);

	SInt32	divisor = mMaximum - mMinimum;
	
	if (divisor != 0)	// avoid divide by zero
		{
		SInt32	totalWidth = rangeFrame->right - rangeFrame->left;
		float	startFraction = ((float) (startValue - mMinimum)) / divisor;
		float	endFraction = ((float) (endValue - mMinimum)) / divisor;

		rangeFrame->right = rangeFrame->left + (endFraction * totalWidth);
		rangeFrame->left = rangeFrame->left + (startFraction * totalWidth);
		}
	}

//
// AGASeparator ---------------------------------------------------------------
//
// This class draws a separator line. It knowns whether it's horizontal or
// vertical based on the larger dimension.
//

AGASeparator::AGASeparator(Rect* bounds)
: AGAObject(bounds)
	{
	// Call superclass constructor.
	}

AGASeparator::~AGASeparator()
	{
	}

void AGASeparator::DrawObject()
	{
	AGADrawSeparator(&mBounds, kSeparatorAutomatic);
	}

//
// AGAGroupBox ---------------------------------------------------------------
//
// This class draws primary and secondary group boxes, with optional
// title label or title gap for a separate control.
//

AGAGroupBox::AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType)
: AGAObject(bounds)
	{
	// Construct without a title or gap.

	mIsPrimaryBox = groupBoxType;
	mTitleGap = 0;
	mTitle[0] = 0;
	mTextStyle = textStyle;
	}

AGAGroupBox::AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType, SInt16 titleGap)
: AGAObject(bounds)
	{
	// Construct with specified gap.

	mIsPrimaryBox = groupBoxType;
	mTitleGap = titleGap;
	mTitle[0] = 0;
	mTextStyle = textStyle;
	}

AGAGroupBox::AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType, StringPtr titleString)
: AGAObject(bounds)
	{
	// Construct with specified title.

	mIsPrimaryBox = groupBoxType;
	mTitleGap = 0;
	AGA_PLstrcpy(mTitle, titleString);
	mTextStyle = textStyle;
	}

AGAGroupBox::AGAGroupBox(Rect* bounds, const AGATextStyle& textStyle, Boolean groupBoxType, SInt16 titleStringListResourceID, SInt16 titleStringIndex)
: AGAObject(bounds)
	{
	// Construct by getting title from specified resource.

	mIsPrimaryBox = groupBoxType;
	mTitleGap = 0;
	::GetIndString(mTitle, titleStringListResourceID, titleStringIndex);
	mTextStyle = textStyle;
	}

AGAGroupBox::~AGAGroupBox()
	{
	}

void AGAGroupBox::SetTitleGap(SInt16 titleGap)
	{
	// Set the width of the title gap that will be left blank.

	mTitleGap = titleGap;
	}

void AGAGroupBox::SetTitle(StringPtr newTitle, Boolean redraw)
	{
	// Set the title, optionally redraw.

	AGA_PLstrcpy(mTitle, newTitle);

	if (redraw)
		this->DrawObject();
	}

void AGAGroupBox::DrawObject()
	{
	AGADrawingEnvironment	env;

	// Draw the group box.

	Rect		r = mBounds;
	SInt16		titleGap = 0;
	SInt16		topIndent = 0;
	FontInfo	fontInfo;

	if (mTitle[0] != 0)
		{
		mTextStyle.PrepareForDrawing();	// calculate gap for title's text style
		::GetFontInfo(&fontInfo);
		
		topIndent = fontInfo.ascent;	// text baseline should be at frame top
		
		titleGap = 5 + ::StringWidth(mTitle);
		}
	else
		titleGap = mTitleGap;

	if (mIsPrimaryBox)
		this->DrawPrimaryFrame(topIndent, titleGap);
	else
		this->DrawSecondaryFrame(topIndent, titleGap);

	if (mTitle[0] != 0)
		{
		Rect	titleRect;
		
		titleRect.left = r.left + 14;
		titleRect.top = r.top + 2;
		titleRect.right = titleRect.left + titleGap;
		titleRect.bottom = titleRect.top + fontInfo.ascent + fontInfo.descent;

		GDIterator	iter;
		Boolean		deep;

		while (iter.NextDevice(deep))
			AGAStringOut(mTitle, &titleRect, kNoTruncation, teFlushDefault, (mEnabled && mActive) ? kNormalOutput : kDisabledOutput, deep, mTextStyle);
		}
	}

void AGAGroupBox::DrawPrimaryFrame(SInt16 topIndent, SInt16 gapSize)
	{
	// Draw the group box as a primary group.

	Rect	r = mBounds;
	
	r.top += topIndent;

	GDIterator	iter;
	Boolean		deep;
	Boolean		drawEnabled = mEnabled && mActive;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			if (drawEnabled)
				::RGBForeColor(&gAGARamp[r7]);
			else
				::RGBForeColor(&gAGARamp[r4]);
			}
		else
			{
			::RGBForeColor(&gAGARamp[rB]);
			
			if (! drawEnabled)
				PenPatGray();
			}

		::MoveTo(r.right - 2, r.top);
		::LineTo(r.right - 2, r.bottom - 2);
		::LineTo(r.left, r.bottom - 2);
		::LineTo(r.left, r.top);
		::Line(10, 0);
		::Move(gapSize, 0);
		::LineTo(r.right - 2, r.top);
		
		if (deep)
			{
			if (drawEnabled)
				::RGBForeColor(&gAGARamp[rW]);
			else
				this->ApplyBackgroundColor();

			::MoveTo(r.right - 1, r.top);
			::LineTo(r.right - 1, r.bottom - 1);
			::LineTo(r.left, r.bottom - 1);
			::MoveTo(r.left + 1, r.bottom - 3);
			::LineTo(r.left + 1, r.top + 1);
			::Line(9, 0);
			::Move(gapSize, 0);
			::LineTo(r.right - 3, r.top + 1);
			}
		else
			PenPatBlack();
		}
	}

void AGAGroupBox::DrawSecondaryFrame(SInt16 topIndent, SInt16 gapSize)
	{
	// Draw the group box as a secondary group.

	Rect	r = mBounds;
	
	r.top += topIndent;

	GDIterator	iter;
	Boolean		deep;
	Boolean		drawEnabled = mEnabled && mActive;

	while (iter.NextDevice(deep))
		{
		if (deep)
			{
			::MoveTo(r.right - 1, r.top);

			if (drawEnabled)
				{
				::RGBForeColor(&gAGARamp[rW]);
				::Move(0, 1);
				}
			else
				::RGBForeColor(&gAGARamp[r4]);

			::LineTo(r.right - 1, r.bottom - 1);
			::LineTo(r.left + 1, r.bottom - 1);

			if (drawEnabled)
				{
				::Move(-1, -1);
				::RGBForeColor(&gAGARamp[r7]);
				}
			else
				::Line(-1, 0);

			::LineTo(r.left, r.top);
			::Line(10, 0);
			::Move(gapSize, 0);
			::LineTo(r.right - 2, r.top);
			}
		else
			{
			::RGBForeColor(&gAGARamp[rB]);
			
			if (! drawEnabled)
				PenPatGray();

			::MoveTo(r.right - 2, r.top);
			::LineTo(r.right - 2, r.bottom - 2);
			::LineTo(r.left, r.bottom - 2);
			::LineTo(r.left, r.top);
			::Line(10, 0);
			::Move(gapSize, 0);
			::LineTo(r.right - 2, r.top);
			
			if (! drawEnabled)
				PenPatBlack();
			}
		}
	}

//
// AGATabPanel ---------------------------------------------------------------
//
// AGATabPanel implements a folder tabs box, often referred to as a "tabbed
// dialog" interface.
//

AGATabPanel::AGATabPanel(Rect* bounds, SInt32 initialTabIndex, const AGATextStyle& textStyle, SInt16 labelsID)
	: AGAObject(bounds)
	{
	// Construct with default options.
	
	mNumTabs = 0;
	mCurrentTabIndex = initialTabIndex;
	mTabInfo = NULL;
	mTabsLocation = kTopTabs;
	mTabWidthKind = kVariableTabWidth;
	mDrawFrame = true;
	mTextStyle = textStyle;
	
	mTabSize = ((textStyle.mFontSize > 10) || (textStyle.mFontSize == 0)) ? kLargeTabs : kSmallTabs;

	this->CalculateTabSizeInfo();
	
	this->SetLabelsFromResource(labelsID);
	
	if (! mImagesAllocated)
		(void) AllocateTabImages();
	}

AGATabPanel::AGATabPanel(Rect* bounds, SInt32 initialTabIndex, const AGATextStyle& textStyle, SInt16 labelsID, TabWidthKind widthKind, TabsLocation location, Boolean drawFrame)
	: AGAObject(bounds)
	{
	// Construct with specified options.

	mNumTabs = 0;
	mCurrentTabIndex = initialTabIndex;
	mTabInfo = NULL;
	mTabsLocation = location;
	mTabWidthKind = widthKind;
	mDrawFrame = drawFrame;
	mTextStyle = textStyle;

	mTabSize = ((textStyle.mFontSize > 10) || (textStyle.mFontSize == 0)) ? kLargeTabs : kSmallTabs;

	this->CalculateTabSizeInfo();

	this->SetLabelsFromResource(labelsID);
	
	if (! mImagesAllocated)
		(void) AllocateTabImages();
	}

AGATabPanel::~AGATabPanel()
	{
	// Destruct by releasing memory allocated for labels.

	this->DisposeTabInfo();
	}

void AGATabPanel::DrawObject()
	{
	this->DrawBackground();
	this->DrawFrame();
	this->DrawTabs();
	}

Boolean AGATabPanel::TrackMouse(Point mouseLocation)
	{
	Boolean	wasItHit;
	Rect	tabRect;

	//
	// First we determine which tab, if any, was clicked.
	//

	mLastTabClicked = -1;
	
	for (SInt32 i = 0; i < mNumTabs; i++)
		{
		this->GetTabBounds(i, &tabRect);
		if (::PtInRect(mouseLocation, &tabRect) && this->IsTabEnabled(i))
			{
			mLastTabClicked = i;
			break;
			}
		}
	
	//
	// Now we let the standard in/out tracking occur.
	// Our hit test will compare with the mLastTabClicked
	// area.
	//

	wasItHit = AGAObject::TrackMouse(mouseLocation);
	
	if (! wasItHit)
		mLastTabClicked = -1;
		
	return wasItHit;
	}

void AGATabPanel::SetObjectBounds(Rect* bounds, Boolean redraw)
	{
	AGAObject::SetObjectBounds(bounds, redraw);

	this->CalculateTabWidths();
	this->CalculateTabSizeInfo();
	}

Boolean AGATabPanel::HitTest(Point mouseLocation)
	{
	Boolean	isIn;

	// Return true if the mouse is in the mLastTabClicked.
	
	if (mLastTabClicked == -1)
		isIn = false;
	else
		{
		Rect	tabRect;

		this->GetTabBounds(mLastTabClicked, &tabRect);
		
		isIn = ::PtInRect(mouseLocation, &tabRect);
		}

	return isIn;
	}

void AGATabPanel::SetTrackingState(Boolean isIn)
	{
	// Draw the mLastTabClicked pressed or unpressed as specified.

	if ((mLastTabClicked != -1) && (mLastTabClicked != mCurrentTabIndex))
		this->DrawOneTab(mLastTabClicked, isIn);
	}

SInt32 AGATabPanel::GetCurrentTab()
	{
	return mCurrentTabIndex;
	}

void AGATabPanel::SetCurrentTab(SInt32 tabIndex, Boolean redraw)
	{
	if (tabIndex != mCurrentTabIndex)
		{
		mCurrentTabIndex = tabIndex;
		
		if (redraw)
			this->DrawTabs();
		}
	}

SInt32 AGATabPanel::GetLastTabClicked()
	{
	return mLastTabClicked;
	}

void AGATabPanel::SetTabSize(TabSize size)
	{
	// Set the tabs size.

	mTabSize = size;
	this->CalculateTabSizeInfo();
	}

void AGATabPanel::SetTabsLocation(TabsLocation location)
	{
	// Set the tabs location.

	mTabsLocation = location;
	this->CalculateTabSizeInfo();
	}

void AGATabPanel::SetTabsWidth(TabWidthKind widthKind)
	{
	// Set the tabs width style.

	mTabWidthKind = widthKind;
	
	this->CalculateTabWidths();
	}

SInt16 AGATabPanel::GetTabHeight()
	{
	// Return the current tab image height (or width if oriented vertically).

	if ((mTabsLocation == kTopTabs) || (mTabsLocation == kBottomTabs))
		return mTabImageSize.v;
	else
		return mTabImageSize.h;
	}

Boolean AGATabPanel::HasFrame()
	{
	return mDrawFrame;
	}

void AGATabPanel::SetLabelsFromResource(SInt16 stringListResourceID)
	{
	// Build the tab labels from the specified 'STR#' resource.
	// A label slot is allocated for each string in the resource;
	// there will be a tab for each of these strings.

	this->DisposeTabInfo();
	
	if (stringListResourceID != 0)
		{
		SInt16**	stringListHandle = (SInt16**) ::GetResource('STR#', stringListResourceID);
		SInt32		numTabs = (stringListHandle == NULL) ? 0 : **stringListHandle;
		
		if (numTabs != 0)
			{
			this->AllocateTabInfo(numTabs, kShrinkToFit);

			for (SInt16 i = 0; i < mNumTabs; i++)
				{
				Str255	aLabel;
				
				::GetIndString(aLabel, stringListResourceID, i+1);
				this->SetTabLabel(i, aLabel);
				}
			
			this->CalculateTabWidths();
			}
		}
	}

void AGATabPanel::SetNumTabs(SInt32 numTabs, Boolean shrinkToFit)
	{
	// Allocate slots for the specified number of labels.
	
	this->AllocateTabInfo(numTabs, shrinkToFit);
	
	// In case the caller never informs us when the labels have
	// all been set (i.e., never calls CalculateTabWidths), we
	// init the widths to fill the bounds.
	
	this->CalculateTabWidthsByKind(kVariableTabWidth);
	}

SInt32 AGATabPanel::GetNumTabs()
	{
	return mNumTabs;
	}

void AGATabPanel::RemoveTab(SInt32 tabIndex)
	{
	if ((mTabInfo != NULL) && (mNumTabs > 0))
		{
		// Only the label handle is our responsibility to free.

		::DisposeHandle((Handle) (*mTabInfo)[tabIndex].mLabel);
		
		// Move the tab info struct for all higher-index tabs back
		// by 1, and then shrink the handle by 1 entry.

		Ptr		destPtr = (Ptr) &((*mTabInfo)[tabIndex]);
		Ptr		sourcePtr = (Ptr) &((*mTabInfo)[tabIndex + 1]);
		SInt32	numBytesToMove = (mNumTabs - 1 - tabIndex) * sizeof(TabInfo);

		::BlockMoveData(sourcePtr, destPtr, numBytesToMove);
		
		mNumTabs--;

		// We're shrinking, so no out-of-memory error possible.
		::SetHandleSize((Handle) mTabInfo, mNumTabs * sizeof(TabInfo));
		
		// If we just deleted the current and last tab, need to change
		// the current selection to be the new last tab.
		if ((mCurrentTabIndex == tabIndex) && (mCurrentTabIndex >= mNumTabs))
			mCurrentTabIndex = mNumTabs - 1;

		this->CalculateTabWidthsByKind(kVariableTabWidth);
		}
	}

void AGATabPanel::AllocateTabInfo(SInt32 numTabs, Boolean shrinkToFit)
	{
	// Make sure the tab info handle is sized to fit the
	// specified number of tabs.
	
	if (mTabInfo == NULL)
		{
		mTabInfo = (TabInfoHandle) ::NewHandleClear(numTabs * sizeof(TabInfo));
		
		if (mTabInfo == NULL)
			mNumTabs = 0;
		else
			mNumTabs = numTabs;
		}
	else if ((mNumTabs < numTabs) || (shrinkToFit && (mNumTabs > numTabs)))
		{
		::SetHandleSize((Handle) mTabInfo, numTabs * sizeof(TabInfo));
		
		if (::MemError() != noErr)
			return;
		
		// Make sure the new tab info elements are zeroed like NewHandleClear.

		for (SInt32 i = mNumTabs; i < numTabs; i++)
			{
			(*mTabInfo)[i].mLabel = ::NewString("\p");
			(*mTabInfo)[i].mTabWidth = 0;
			(*mTabInfo)[i].mDisabled = false;
			(*mTabInfo)[i].mUserData = NULL;
			}

		mNumTabs = numTabs;
		}
	}

void AGATabPanel::SetTabLabel(SInt32 tabIndex, StringPtr itsLabel)
	{
	// Assign the specified string to the specified label slot.

	if (mTabInfo != NULL)
		{
		if (tabIndex < mNumTabs)
			{
			if ((*mTabInfo)[tabIndex].mLabel != NULL)
				::DisposeHandle((Handle) (*mTabInfo)[tabIndex].mLabel);
			
			(*mTabInfo)[tabIndex].mLabel = ::NewString(itsLabel);
			}
		}
	}

void AGATabPanel::GetTabLabel(SInt32 tabIndex, StringPtr itsLabel)
	{
	// Return the label string of the specified label slot.

	itsLabel[0] = 0;

	if (mTabInfo != NULL)
		{
		if (tabIndex < mNumTabs)
			{
			StringHandle	labelStringHandle = (*mTabInfo)[tabIndex].mLabel;
			
			if (labelStringHandle != NULL)
				{
				Ptr	sourceStringPtr = (Ptr) *labelStringHandle;

				::BlockMoveData(sourceStringPtr, itsLabel, 1 + *sourceStringPtr);
				}
			}
		}
	}

void AGATabPanel::SetTabEnable(SInt32 tabIndex, Boolean enable)
	{
	// Enable or disable the specified tab.

	if (mTabInfo != NULL)
		{
		if (tabIndex < mNumTabs)
			(*mTabInfo)[tabIndex].mDisabled = ! enable;
		}
	}

Boolean AGATabPanel::IsTabEnabled(SInt32 tabIndex)
	{
	// Return the enable state of the specified tab.
	
	Boolean	isEnabled = true;

	if (mTabInfo != NULL)
		{
		if (tabIndex < mNumTabs)
			isEnabled = ! (*mTabInfo)[tabIndex].mDisabled;
		}
	
	return isEnabled;
	}

void AGATabPanel::SetTabUserData(SInt32 tabIndex, void* userData)
	{
	// Stash the userData in the specified tab's info.

	if (mTabInfo != NULL)
		{
		if (tabIndex < mNumTabs)
			(*mTabInfo)[tabIndex].mUserData = userData;
		}
	}

void* AGATabPanel::GetTabUserData(SInt32 tabIndex)
	{
	// Return the specified tab's mUserData.
	
	void*	userData = NULL;

	if (mTabInfo != NULL)
		{
		if (tabIndex < mNumTabs)
			userData = (*mTabInfo)[tabIndex].mUserData;
		}
	
	return userData;
	}

void AGATabPanel::SetLabelsStyle(const AGATextStyle& textStyle)
	{
	// Set the QuickDraw text style for the labels.

	mTextStyle = textStyle;
	
	this->CalculateTabWidths();
	this->CalculateTabSizeInfo();
	}

void AGATabPanel::CalculateTabWidths()
	{
	this->CalculateTabWidthsByKind(mTabWidthKind);
	}

void AGATabPanel::CalculateTabWidthsByKind(TabWidthKind widthKind)
	{
	AGADrawingEnvironment	env;

	SInt32		widthAvailable;
	Str255		aLabel;
	SInt32		tabWidth;
	SInt32		totalWidth;
	SInt32		overflow;
	SInt32		tabEndWidth;
	
	if ((mTabsLocation == kTopTabs) || (mTabsLocation == kBottomTabs))
		{
		widthAvailable = mBounds.right - mBounds.left - (2 * kPanelEndMargin);
		tabEndWidth = mTabImageSize.h;
		}
	else
		{
		widthAvailable = mBounds.bottom - mBounds.top - (2 * kPanelEndMargin);
		tabEndWidth = mTabImageSize.v;
		}

	mTextStyle.PrepareForDrawing();
	
	switch (widthKind)
		{
		case kVariableTabWidth:
		
			totalWidth = 0;

			for (SInt32 i = 0; i < mNumTabs; i++)
				{
				this->GetTabLabel(i, aLabel);
				tabWidth = ::StringWidth(aLabel) + (2 * tabEndWidth);

				totalWidth += tabWidth;
				(*mTabInfo)[i].mTabWidth = tabWidth;
				}
			
			// If we overflowed the available width, reduce the tab sizes.

			overflow = totalWidth - widthAvailable;

			if (overflow > 0)
				for (SInt32 i = 0; i < mNumTabs; i++)
					(*mTabInfo)[i].mTabWidth -= (1 + (overflow / mNumTabs));
			break;

		case kCommonTabWidth:

			tabWidth = 0;

			for (SInt32 i = 0; i < mNumTabs; i++)
				{
				this->GetTabLabel(i, aLabel);
				tabWidth = MaxSInt32(tabWidth, ::StringWidth(aLabel));
				}
			
			tabWidth += (2 * tabEndWidth);

			for (SInt32 i = 0; i < mNumTabs; i++)
				(*mTabInfo)[i].mTabWidth = tabWidth;

			// If we overflowed the available width, reduce the tab sizes.

			overflow = (tabWidth * mNumTabs) - widthAvailable;

			if (overflow > 0)
				for (SInt32 i = 0; i < mNumTabs; i++)
					(*mTabInfo)[i].mTabWidth -= (1 + (overflow / mNumTabs));

			break;
		}
	}

void AGATabPanel::DrawBackground()
	{
	AGADrawingEnvironment	env;

	// Paint the background.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		if (deep)
			this->ApplyBackgroundColor();
		else
			::RGBForeColor(&gAGARamp[rW]);

		::PaintRect(&mBounds);
		}
	}

void AGATabPanel::DrawFrame()
	{
	// If we have a frame, draw it; otherwise, draw the line/shading
	// for just the tabs edge.

	if (mDrawFrame)
		this->DrawFrameBox();
	else
		this->DrawFrameTop();
	}

void AGATabPanel::DrawFrameTop()
	{
	// Draw the line and shading for the tabs without a frame.

	AGADrawingEnvironment	env;

	Rect	frameBounds = mBounds;

	frameBounds.top = frameBounds.bottom - 3;
	
	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		Rect	r = frameBounds;
		
		if (deep && !mEnabled)
			::RGBForeColor(&gAGARamp[r10]);
		else
			::RGBForeColor(&gAGARamp[rB]);

		::MoveTo(r.left, r.top);
		::LineTo(r.right - 1, r.top);
		
		if (deep)
			{
			if (mEnabled)
				::RGBForeColor(&gAGARamp[r3]);
			else
				::RGBForeColor(&gAGARamp[r2]);

			::MoveTo(r.left + 1, r.top + 1);
			::LineTo(r.right - 2, r.top + 1);

			if (mEnabled)
				::RGBForeColor(&gAGARamp[rW]);
			else
				::RGBForeColor(&gAGARamp[r2]);

			::MoveTo(r.left + 2, r.top + 2);
			::LineTo(r.right - 3, r.top + 2);
			}
		}
	}

void AGATabPanel::DrawFrameBox()
	{
	// Draw the tab panel frame box.

	AGADrawingEnvironment	env;

	enum { Frame, Fill, LightOuter, LightInner, DarkOuter, DarkInner, kNumFrameColors };

	UInt8	colorIndexes[kNumFrameColors];
	Rect	frameBounds = mBounds;
	
	if (mEnabled)
		{
		colorIndexes[Frame] = rB;
		colorIndexes[Fill] = r1;
		colorIndexes[LightOuter] = r3;
		colorIndexes[LightInner] = rW;
		colorIndexes[DarkOuter] = r6;
		colorIndexes[DarkInner] = r4;
		}
	else
		{
		colorIndexes[Frame] = r10;
		colorIndexes[Fill] = r2;
		colorIndexes[LightOuter] = r2;
		colorIndexes[LightInner] = r2;
		colorIndexes[DarkOuter] = r2;
		colorIndexes[DarkInner] = r2;
		}

	frameBounds.top += (mTabImageSize.v - 3);
	
	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		Rect	r = frameBounds;
		
		if (deep)
			::RGBForeColor(&gAGARamp[colorIndexes[Frame]]);
		else
			{
			::RGBForeColor(&gAGARamp[rB]);
			
			if (! mEnabled)
				PenPatGray();
			}
		
		::FrameRect(&r);
		
		if (deep)
			::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
		else
			{
			::RGBForeColor(&gAGARamp[rW]);
			
			if (! mEnabled)	// put it back if we set to gray above
				PenPatBlack();
			}
			
		::InsetRect(&r, 1, 1);
		::PaintRect(&r);
		
		if (deep && mEnabled)	// disabled has no raised light/dark edges
			{
			::RGBForeColor(&gAGARamp[colorIndexes[LightOuter]]);
			::MoveTo(r.left, r.bottom - 2);
			::LineTo(r.left, r.top);
			::LineTo(r.right - 2, r.top);

			::RGBForeColor(&gAGARamp[colorIndexes[DarkOuter]]);
			::Move(1, 1);
			::LineTo(r.right - 1, r.bottom - 1);
			::LineTo(r.left + 1, r.bottom - 1);

			::RGBForeColor(&gAGARamp[colorIndexes[LightInner]]);
			::Move(0, -2);
			::LineTo(r.left + 1, r.top + 1);
			::LineTo(r.right - 3, r.top + 1);

			::RGBForeColor(&gAGARamp[colorIndexes[DarkInner]]);
			::Move(1, 1);
			::LineTo(r.right - 2, r.bottom - 2);
			::LineTo(r.left + 2, r.bottom - 2);
			
			::RGBForeColor(&gAGARamp[r2]);
			::MoveTo(r.left, r.bottom - 1);
			::Line(0, 0);
			::MoveTo(r.top, r.right - 1);
			::Line(0, 0);
			}
		}
	}

void AGATabPanel::DrawTabs()
	{
	// Draw the tabs, one at a time.

	for (SInt32 i = 0; i < mNumTabs; i++)
		this->DrawOneTab(i, kNotPressed);
	}

void AGATabPanel::DrawOneTab(SInt32 tabIndex, Boolean pressed)
	{
	// Draw the specified tab.

	AGADrawingEnvironment	env;

	enum { EdgeFrame, OuterEdge, InnerEdge, Fill, InteriorFrame, InteriorOuterEdge, InteriorInnerEdge, kNumTabColors };

	UInt8	colorIndexes[kNumTabColors];

	Boolean				selected = (tabIndex == mCurrentTabIndex);
	Rect				tabBounds;
	Str255				tabLabel;
	Point				tabOrigin;
	AGAStringOutColor	textColor;
	UInt32				leftImageIndex;
	UInt32				rightImageIndex;

	this->GetTabBounds(tabIndex, &tabBounds);
	this->GetTabLabel(tabIndex, tabLabel);
	
	tabBounds.left += mTabImageSize.h;
	tabBounds.right -= mTabImageSize.h;

	if ((! mEnabled) || ! this->IsTabEnabled(tabIndex))
		{
		textColor = kDisabledOutput;
		leftImageIndex = selected ? kTabDisabledSelectedL : kTabDisabledDeselectedL;
		rightImageIndex = selected ? kTabDisabledSelectedR : kTabDisabledDeselectedR;

		colorIndexes[EdgeFrame] = r10;
		colorIndexes[OuterEdge] = r2;
		colorIndexes[InnerEdge] = r2;
		colorIndexes[Fill] = r2;
		colorIndexes[InteriorFrame] = selected ? r2 : r10;
		colorIndexes[InteriorOuterEdge] = r2;
		colorIndexes[InteriorInnerEdge] = r2;
		}
	else if (pressed)
		{
		textColor = kInverseOutput;
		leftImageIndex = kTabPressedL;
		rightImageIndex = kTabPressedR;

		colorIndexes[EdgeFrame] = rB;
		colorIndexes[OuterEdge] = r11;
		colorIndexes[InnerEdge] = r10;
		colorIndexes[Fill] = r9;
		colorIndexes[InteriorFrame] = rB;
		colorIndexes[InteriorOuterEdge] = r3;
		colorIndexes[InteriorInnerEdge] = rW;
		}
	else if (selected)
		{
		textColor = kNormalOutput;
		leftImageIndex = kTabNormalSelectedL;
		rightImageIndex = kTabNormalSelectedR;

		colorIndexes[EdgeFrame] = rB;
		colorIndexes[OuterEdge] = r3;
		colorIndexes[InnerEdge] = rW;
		colorIndexes[Fill] = r1;
		colorIndexes[InteriorFrame] = r1;
		colorIndexes[InteriorOuterEdge] = r1;
		colorIndexes[InteriorInnerEdge] = r1;
		}
	else
		{
		textColor = kNormalOutput;
		leftImageIndex = kTabNormalDeselectedL;
		rightImageIndex = kTabNormalDeselectedR;

		colorIndexes[EdgeFrame] = rB;
		colorIndexes[OuterEdge] = r3;
		colorIndexes[InnerEdge] = r2;
		colorIndexes[Fill] = r3;
		colorIndexes[InteriorFrame] = rB;
		colorIndexes[InteriorOuterEdge] = r3;
		colorIndexes[InteriorInnerEdge] = rW;
		}
	
	// Note that at this point, tabBounds is inset to avoid
	// the corner images.

	GDIterator	iter;
	Boolean		deep;

	while (iter.NextDevice(deep))
		{
		// Draw the tab corner images.

		tabOrigin.h = tabBounds.left - mTabImageSize.h;
		tabOrigin.v = tabBounds.top;
		
		this->CopyTabImage(leftImageIndex, tabOrigin, deep);
		tabOrigin.h = tabBounds.right;
		this->CopyTabImage(rightImageIndex, tabOrigin, deep);
		
		// Draw the straight edge shading.

		Rect	r = tabBounds;

		::RGBForeColor(&gAGARamp[colorIndexes[EdgeFrame]]);
		::MoveTo(r.left, r.top);
		::LineTo(r.right - 1, r.top);
		
		if (deep)
			{
			r.top++;
			r.bottom--;

			::RGBForeColor(&gAGARamp[colorIndexes[OuterEdge]]);
			::MoveTo(r.left, r.top);
			::LineTo(r.right - 1, r.top);
			r.top++;

			::RGBForeColor(&gAGARamp[colorIndexes[InnerEdge]]);
			::MoveTo(r.left, r.top);
			::LineTo(r.right - 1, r.top);
			r.top++;
			
			::RGBForeColor(&gAGARamp[colorIndexes[InteriorInnerEdge]]);
			::MoveTo(r.left, r.bottom);
			::LineTo(r.right - 1, r.bottom);

			r.bottom--;
			::RGBForeColor(&gAGARamp[colorIndexes[InteriorOuterEdge]]);
			::MoveTo(r.left, r.bottom);
			::LineTo(r.right - 1, r.bottom);

			r.bottom--;
			::RGBForeColor(&gAGARamp[colorIndexes[InteriorFrame]]);
			::MoveTo(r.left, r.bottom);
			::LineTo(r.right - 1, r.bottom);

			::RGBForeColor(&gAGARamp[colorIndexes[Fill]]);
			::PaintRect(&r);
			}
		else
			{
			// For 1-bit drawing, there's a pixel at the bottom that
			// may need to be erased or re-drawn. It's either leftover
			// from the frame (need to erase), or missing from being
			// previously selected (need to draw).
			
			r.bottom -= 3;
			r.top++;

			if (selected)
				::RGBForeColor(&gAGARamp[rW]);
			else
				::RGBForeColor(&gAGARamp[rB]);

			::MoveTo(r.left, r.bottom);
			::LineTo(r.right - 1, r.bottom);
			
			// The enable state of the tabs as a whole, and this
			// tab in particular require us to paint a gray pattern over
			// some or all of the tab corner images.

			if ((! mEnabled) || ! this->IsTabEnabled(tabIndex))
				{
				Rect	rectToPaint = tabBounds;
				
				rectToPaint.left -= mTabImageSize.h;
				rectToPaint.right += mTabImageSize.h;
				
				// If whole thing is enabled but this tab is disabled,
				// make sure we don't gray out the box frame line segment.
				if (mEnabled && ! this->IsTabEnabled(tabIndex))
					rectToPaint.bottom -= 3;

				PenPatGray();
				::PenMode(patBic);
				::PaintRect(&rectToPaint);
				
				::PenNormal();
				}
			
			// We also need to paint the tab rectangular inside area
			// black or white, depending on whether the tab is pressed,
			// in preparation for the label text.
			
			if (pressed)
				::RGBForeColor(&gAGARamp[rB]);
			else
				::RGBForeColor(&gAGARamp[rW]);

			::PaintRect(&r);
			}
		
		AGAStringOut(tabLabel, &tabBounds, smTruncMiddle, teCenter, textColor, deep, mTextStyle);
		}
	}

void AGATabPanel::GetTabBounds(SInt32 tabIndex, Rect* tabBounds)
	{
	// Return the bounds of the specified tab.

	SInt16	tabOrigin = kPanelEndMargin;
	SInt16	tabWidth = (*mTabInfo)[tabIndex].mTabWidth;
	
	for (SInt32 i = 0; i < tabIndex; i++)
		tabOrigin += (*mTabInfo)[i].mTabWidth;;

	switch (mTabsLocation)
		{
		case kTopTabs:
		case kBottomTabs:

			tabBounds->left = mBounds.left + tabOrigin;
			tabBounds->right = tabBounds->left + tabWidth;
			
			if ((mDrawFrame && (mTabsLocation == kTopTabs)) ||
				(!mDrawFrame && (mTabsLocation == kBottomTabs)))
				{
				tabBounds->top = mBounds.top;
				tabBounds->bottom = mBounds.top + mTabImageSize.v;
				}
			else
				{
				tabBounds->top = mBounds.bottom - mTabImageSize.v;
				tabBounds->bottom = mBounds.bottom;
				}

			break;

		case kLeftTabs:
		case kRightTabs:

			tabBounds->top = mBounds.top + tabOrigin;
			tabBounds->bottom = tabBounds->top + tabWidth;
			
			if ((mDrawFrame && (mTabsLocation == kLeftTabs)) ||
				(!mDrawFrame && (mTabsLocation == kRightTabs)))
				{
				tabBounds->left = mBounds.left;
				tabBounds->right = mBounds.left + mTabImageSize.h;
				}
			else
				{
				tabBounds->left = mBounds.right - mTabImageSize.h;
				tabBounds->right = mBounds.right;
				}

			break;
		}
	}

void AGATabPanel::CalculateTabSizeInfo()
	{
	// Recalculate the tab image size based on the tab size kind and
	// the tab location orientation.
	// Decide whether our bounds are large enough to draw a frame.

	switch (mTabSize)
		{
		case kSmallTabs:
		
			switch (mTabsLocation)
				{
				case kTopTabs:
				case kBottomTabs:
					::SetPt(&mTabImageSize, kSmallTabImageWidth, kSmallTabImageHeight);
					break;

				case kLeftTabs:
				case kRightTabs:
					::SetPt(&mTabImageSize, kSmallTabImageHeight, kSmallTabImageWidth);
					break;
				}

			break;

		case kLargeTabs:
		
			switch (mTabsLocation)
				{
				case kTopTabs:
				case kBottomTabs:
					::SetPt(&mTabImageSize, kLargeTabImageWidth, kLargeTabImageHeight);
					break;

				case kLeftTabs:
				case kRightTabs:
					::SetPt(&mTabImageSize, kLargeTabImageHeight, kLargeTabImageWidth);
					break;
				}

			break;
		}
	
	//
	// If the bounds are small, we don't draw the frame. We allow some
	// leeway in layout so that the user doesn't have to make the bounds
	// exactly match the image size to suppress the frame. This is fine
	// because if there is a frame, it won't be that small.
	//

	mDrawFrame = ((mBounds.bottom - mBounds.top) > (mTabImageSize.v + 16));
	}

void AGATabPanel::DisposeTabInfo()
	{
	// Release the memory allocated for tab labels.

	if (mTabInfo != NULL)
		{
		for (SInt32 i = 0; i < mNumTabs; i++)
			::DisposeHandle((Handle) (*mTabInfo)[i].mLabel);

		::DisposeHandle((Handle) mTabInfo);
		mTabInfo = NULL;
		}

	mNumTabs = 0;
	}

Boolean AGATabPanel::mImagesAllocated = false;

// Each radio button drawing state has an offscreen image.
AGAOffscreenImage*	AGATabPanel::mgOffscreenTabImages [kNumTabImages] [kNumImageTypes];
AGAOffscreenImage*	AGATabPanel::mg1BitOffscreenTabImages [kNumTabImages] [kNumImageTypes];

// These are the color ramp indexes that are used to build the small tab offscreen images.
SInt8				AGATabPanel::kSmallTabImageValues [kNumTabImages] [kSmallTabImageHeight] [kSmallTabImageWidth] =
	{
		{	// kTabNormalDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB,  r5,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  r5,  r3,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r12,  r5,  r2,  r2,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, r12,  r3,  r2,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3 },
//		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3 },
//		{ OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3 },
//		{ OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
//		{ OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
//		{ OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalDeselectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r4, r11, r12, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r4,  r9, r12, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT, OUT, OUT },
//		{  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r4,  r9, r12, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT, OUT },
//		{  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r4,  r9, r12, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT },
//		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r4,  r9, r12, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT },
//		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r4,  r9, r12, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT },
//		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r4,  r9, r12 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB,  r4,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  r4,  r3,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r12,  r4,  rW,  rW,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1 },
//		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1 },
//		{ OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1 },
//		{ OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
//		{ OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
//		{ OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		},

		{	// kTabNormalSelectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r4,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT, OUT, OUT },
//		{  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT, OUT },
//		{  r1,  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT },
//		{  r1,  r1,  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT },
//		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r3, r11, r12, OUT },
//		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r7,  rB, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r2, r11,  rB },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r7,  rB },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r3,  r3 },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  rW },
		},

		{	// kTabPressedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB, r11, r11 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12, r11, r11, r10, r10 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r12, r11, r10, r10,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9 },
//		{ OUT, OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9 },
//		{ OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9 },
//		{ OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
//		{ OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
//		{ OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabPressedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{ r11,  r7,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{ r10, r10,  r7,  r8,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT, OUT },
//		{  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT },
//		{  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT },
//		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT },
//		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT },
//		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

		{	// kTabDisabledDeselectedR
		{ r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2, r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10 },
		{ r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

		{	// kTabDisabledSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
//		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

		{	// kTabDisabledSelectedR
		{ r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2, r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
//		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

	};

// These are the color ramp indexes that are used to build the large tab offscreen images.
SInt8				AGATabPanel::kLargeTabImageValues [kNumTabImages] [kLargeTabImageHeight] [kLargeTabImageWidth] =
	{
		{	// kTabNormalDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB,  r5,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  r5,  r3,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r12,  r5,  r2,  r2,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT, r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ OUT,  rB,  r3,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{ r12,  r5,  r2,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalDeselectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r4, r11, r12, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r4, r11, r12, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r4, r11, r12, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r4, r11, r12, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r4, r11, r12, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r4, r11, r12, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r7,  rB, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r5,  rB, OUT },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r4,  r9, r12 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB,  r4,  r3 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  r4,  r3,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r12,  r4,  r3,  rW,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT, r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ OUT,  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{ r12,  r4,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{  rB,  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{  r3,  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		{  rW,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1 },
		},

		{	// kTabNormalSelectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r3,  r4,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r4, r11, r12, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r3,  r7,  rB, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r3, r11, r12, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r2,  r7,  rB, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r7,  rB, OUT },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r2, r11,  rB },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r7,  rB },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r3,  r3 },
		{  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  r1,  rW },
		},

		{	// kTabPressedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12,  rB, r11, r11 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r12, r11, r11, r10, r10 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r12, r11, r11, r10,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT, r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ OUT,  rB, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{ r12, r11, r10,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabPressedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{ r11,  r7,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{ r10, r10,  r7,  r8,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r8,  rB, OUT },
		{  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r9,  r7, r11, r12 },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3,  r3 },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

		{	// kTabDisabledDeselectedR
		{ r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2, r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10 },
		{ r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

		{	// kTabDisabledSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10, r10,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ OUT, r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{ r10,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

		{	// kTabDisabledSelectedR
		{ r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2, r10, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10, OUT },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2, r10 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		{  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2,  r2 },
		},

	};

// These are the color ramp indexes that are used to build the small 1-bit tab offscreen images.
SInt8				AGATabPanel::k1BitSmallTabImageValues [kNumTabImages] [kSmallTabImageHeight] [kSmallTabImageWidth] =
	{
		{	// kTabNormalDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalDeselectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalSelectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabPressedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
//		{ OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
//		{ OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
//		{ OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
//		{ OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
//		{ OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabPressedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT },
//		{  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT },
//		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT },
//		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT },
//		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT },
//		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledDeselectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
//		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledSelectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
//		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

	};

// These are the color ramp indexes that are used to build the large 1-bit tab offscreen images.
SInt8				AGATabPanel::k1BitLargeTabImageValues [kNumTabImages] [kLargeTabImageHeight] [kLargeTabImageWidth] =
	{
		{	// kTabNormalDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalDeselectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabNormalSelectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabPressedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT, OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{ OUT,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabPressedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB, OUT },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledDeselectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledDeselectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledSelectedL
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rB,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT, OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{ OUT,  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rB,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

		{	// kTabDisabledSelectedR
		{  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rB,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB, OUT },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rB },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		{  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW,  rW },
		},

	};

OSErr AGATabPanel::AllocateTabImages()
	{
	// Allocate the offscreen images for the various sizes & states.

	OSErr	result = noErr;
	Point	imageSize;
	
	if (result == noErr)
		{
		SetPt(&imageSize, kLargeTabImageWidth, kLargeTabImageHeight);

		for (UInt32 imageIndex = 0; (imageIndex < kNumTabImages) && (result == noErr); imageIndex++)
			{
			mgOffscreenTabImages [imageIndex] [kLargeTabImages] = new AGAOffscreenImage;
		
			result = mgOffscreenTabImages [imageIndex] [kLargeTabImages] ->CreateImageData(
						imageSize,
						&kLargeTabImageValues[imageIndex][0][0],
						NULL, 
						&gAGARamp[r2]);
			}
		}
		
	if (result == noErr)
		{
		SetPt(&imageSize, kSmallTabImageWidth, kSmallTabImageHeight);

		for (UInt32 imageIndex = 0; (imageIndex < kNumTabImages) && (result == noErr); imageIndex++)
			{
			mgOffscreenTabImages [imageIndex] [kSmallTabImages] = new AGAOffscreenImage;
		
			result = mgOffscreenTabImages [imageIndex] [kSmallTabImages] ->CreateImageData(
						imageSize,
						&kSmallTabImageValues[imageIndex][0][0],
						NULL, 
						&gAGARamp[r2]);
			}
		}
		
	if (result == noErr)
		{
		SetPt(&imageSize, kLargeTabImageWidth, kLargeTabImageHeight);

		for (UInt32 imageIndex = 0; (imageIndex < kNumTabImages) && (result == noErr); imageIndex++)
			{
			mg1BitOffscreenTabImages [imageIndex] [kLargeTabImages] = new AGAOffscreenImage;
		
			result = mg1BitOffscreenTabImages [imageIndex] [kLargeTabImages] ->CreateImageData(
						imageSize,
						&k1BitLargeTabImageValues[imageIndex][0][0],
						NULL, 
						&gAGARamp[r2]);
			}
		}
		
	if (result == noErr)
		{
		SetPt(&imageSize, kSmallTabImageWidth, kSmallTabImageHeight);

		for (UInt32 imageIndex = 0; (imageIndex < kNumTabImages) && (result == noErr); imageIndex++)
			{
			mg1BitOffscreenTabImages [imageIndex] [kSmallTabImages] = new AGAOffscreenImage;
		
			result = mg1BitOffscreenTabImages [imageIndex] [kSmallTabImages] ->CreateImageData(
						imageSize,
						&k1BitSmallTabImageValues[imageIndex][0][0],
						NULL, 
						&gAGARamp[r2]);
			}
		}
		
	mImagesAllocated = true;

	return result;
	}

void AGATabPanel::CopyTabImage(UInt32 imageIndex, Point imageDestination, Boolean deep)
	{
	// Blast the specified offscreen image into the radio button's button area.

	if (deep)
		mgOffscreenTabImages [imageIndex] [(mTabSize == kSmallTabs) ? kSmallTabImages : kLargeTabImages]->DrawImage(imageDestination.h, imageDestination.v);
	else
		mg1BitOffscreenTabImages [imageIndex] [(mTabSize == kSmallTabs) ? kSmallTabImages : kLargeTabImages]->DrawImage(imageDestination.h, imageDestination.v);
	}

void AGATabPanel::DisposeTabImages()
	{
	for (UInt32 imageIndex = 0; imageIndex < kNumTabImages; imageIndex++)
		if (mgOffscreenTabImages [imageIndex] [kLargeTabImages] != NULL)
			delete mgOffscreenTabImages [imageIndex] [kLargeTabImages];

	for (UInt32 imageIndex = 0; imageIndex < kNumTabImages; imageIndex++)
		if (mgOffscreenTabImages [imageIndex] [kSmallTabImages] != NULL)
			delete mgOffscreenTabImages [imageIndex] [kSmallTabImages];

	for (UInt32 imageIndex = 0; imageIndex < kNumTabImages; imageIndex++)
		if (mg1BitOffscreenTabImages [imageIndex] [kLargeTabImages] != NULL)
			delete mg1BitOffscreenTabImages [imageIndex] [kLargeTabImages];

	for (UInt32 imageIndex = 0; imageIndex < kNumTabImages; imageIndex++)
		if (mg1BitOffscreenTabImages [imageIndex] [kSmallTabImages] != NULL)
			delete mg1BitOffscreenTabImages [imageIndex] [kSmallTabImages];
	}

